/* Driver for Realtek card reader
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>

#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include <linux/module.h>

#include "rts_cr.h"
#include "transport.h"
#include "pm.h"
#include "cardmanager.h"
#include "cards/xd.h"
#include "cr_regs.h"

/* cm_enable_card_clock - Enable card clock
 * @cm: card manager structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cm_enable_card_clock(struct card_manager *cm, u8 card)
{
	return 0;
}

/* cm_disable_card_clock - Disable card clock
 * @cm: card manager structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cm_disable_card_clock(struct card_manager *cm, u8 card)
{
	return 0;
}

/* cm_card_power_on - Power on card
 * @cm: card manager structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cm_card_power_on(struct card_manager *cm, u8 card)
{
	return 0;
}

/* cm_card_power_off - Power off card
 * @cm: card manager structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cm_card_power_off(struct card_manager *cm, u8 card)
{
	return 0;
}

/* do_remaining_work - perform "stop transmission" after sequential transfer
 * for specified card
 * @cm: pointes to struct card_manager
 *
 * Returns: Nothing
 */
void do_remaining_work(struct card *c)
{
	int ret = 0;

	if (c->cleanup) {
		ret = c->cleanup(c);
		if (ret)
			rts_trace(TRACE_CM, "card:%s cleanup error\n",
								c->name);
	}
}

/* cm_do_remaining_work - perform "stop transmission" after sequential
 * transfer for all cards
 * @cm: pointes to struct card_manager
 *
 * Returns: Nothing
 */
void cm_do_remaining_work(struct card_manager *cm)
{
	struct list_head *pos;


	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		do_remaining_work(c);
	}

}

void cm_status_runtime_save(struct card_manager *cm)
{
	struct list_head *pos;

	list_for_each(pos, &cm->card_list) {
		struct card *c = list_entry(pos, struct card, link);
		if (c->card_flags & READY)
			set_card_flags(c, RPM_READY);
	}
}
EXPORT_SYMBOL_GPL(cm_status_runtime_save);

void cm_status_runtime_clear(struct card_manager *cm)
{
	struct list_head *pos;

	list_for_each(pos, &cm->card_list) {
		struct card *c = list_entry(pos, struct card, link);
		clr_card_flags(c, RPM_READY);
	}
}
EXPORT_SYMBOL_GPL(cm_status_runtime_clear);

/* cm_release_cards - Release all the cards under control of card manager
 * @cm: card manager structure
 *
 * Returns: Nothing
 */
void cm_release_cards(struct card_manager *cm)
{
	struct list_head *pos;

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (c->card_flags & READY) {
			c->cleanup(c);
			c->release(c);
			clr_card_flags(c, READY);
		}
	}
}
EXPORT_SYMBOL_GPL(cm_release_cards);

static int check_cd_deglitch(struct card_reader *cr)
{
	u32 card_exist = 0;
	unsigned long need_reset = cr->need_reset;
	int deglitch = 0;

	if (!(cr->ops->cd_deglitch))
		return 0;

	/*if (hweight32(cr->need_reset) <= 1)
		return 0;*/

	card_exist = cr->ops->cd_deglitch(cr);

	if ((test_bit(MS_ID, &need_reset)) && (!(card_exist & MS_EXIST))) {
		clear_bit(MS_ID, &(cr->need_reset));
		deglitch = 1;
	}

	if ((test_bit(SD_ID, &need_reset)) && (!(card_exist & SD_EXIST))) {
		clear_bit(SD_ID, &(cr->need_reset));
		deglitch = 1;
	}

	if ((test_bit(XD_ID, &need_reset)) && (!(card_exist & XD_EXIST))) {
		clear_bit(XD_ID, &(cr->need_reset));
		deglitch = 1;
	}

	return deglitch;
}

int process_card_reset(struct card *c)
{
	struct card_reader *cr = c->cm->cr;
	int i = 0;
	int ret = 0;
	int sdio = 0;

	cr_inform_reset(c->cm->cr);

	if (!check_card_flag(c, READY | FAIL_RESET)) {
		rts_trace(TRACE_CM, "reset card:%s\n", c->name);
		set_card_flags(c, EXIST);
		if (!CHECK_STATUS_FLAG(c->cm->cr, LED_TEST_MODE))
			cr_enable_auto_blink(c->cm->cr);
		for (i = 0; i < c->max_reset_retry; i++) {
			c->cm->cr->tr->catch_link_error = 0;
			c->reset_retry = i;
			ret = c->reset_card(c);
			if (STATUS_SUCCESS == ret) {
				set_card_flags(c, READY);
				c->cm->cur_card = c;
				clear_bit(c->id, &(c->cm->cr->reset_map));
				break;
			} else {
				if (c->id != SD_ID)
					continue;
				if (CHK_SDIO_EXIST(c->cm->cr) &&
					(!c->cm->cr->sdio_func_init)) {
					cr_switch_to_sdio(c->cm->cr);
					sdio = 1;
					break;
				}
			}

			/* For some platform, there exist a long
			 * period (about 600ms) at the moment driver can not
			 * access internal register by BAR0,
			 * So dalay the reset procedure and retry it again.
			 */
			if (c->cm->cr->tr->catch_link_error) {
				msleep(600);
				c->cm->cr->tr->catch_link_error = 0;
			}
		}

		if (ret) {
			rts_trace(TRACE_CM, "Err reset card:%s\n", c->name);
			set_card_flags(c, FAIL_RESET);
			c->capacity = 0;
			c->disable_card(c);
			if (!sdio)
				cr_poweroff_card(c->cm->cr, c->id);
		}
	}

	if (cr->done_resume)
		complete(cr->done_resume);
	rts_trace(TRACE_SD,
		"-------------------- card reset done --------------------\n");
	return ret;
}

/* cm_release_cards - Perform card release operation based on card status
 * @c: card structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int process_card_release(struct card *c)
{
	rts_trace(TRACE_CM, "%s\n", c->name);

	clr_card_flags(c, EXIST | FAIL | FAIL_RELEASE | FAIL_RESET | FAIL_RW);
	c->cleanup(c);
	c->release(c);
	clr_card_flags(c, READY | EJECTED);
	clear_bit(c->id, &(c->cm->cr->release_map));
	c->cm->cur_card = NULL;

	return 0;
}

/* process_status_further - Reserved for further use
 * @c: card structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int process_status_further(struct card *c, int status)
{
	int ret = 0;

	return ret;
}

/* process_card_status - Main entry to process card status
 * @c: card structure
 * @status: card status
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int process_card_status(struct card *c, int status)
{
	unsigned long reset_map = c->cm->cr->reset_map;
	unsigned long release_map = c->cm->cr->release_map;
	unsigned long exist_map = c->cm->cr->exist_map;
	int ret = STATUS_FAIL;

	rts_trace(TRACE_POLL_THRD, "%s: reset_map: 0x%lx, release_map: 0x%lx,"
				"exist_map: 0x%lx, %s: status: 0x%x\n",
				c->cm->cr->name, reset_map, release_map,
				exist_map, c->name, c->card_flags);

	if (!check_card_flag(c, FAIL_RELEASE) &&
		test_bit(c->id, &release_map)) {
		ret = process_card_release(c);
		cr_notify_upper(c->cm->cr, c, ret);
	} else {
		rts_trace(TRACE_POLL_THRD, "%s: skip process_card_release!\n",
			c->name);
	}

	if (!check_card_flag(c, FAIL_RESET) && test_bit(c->id, &reset_map)) {
		ret = process_card_reset(c);
		cr_notify_upper(c->cm->cr, c, ret);
		/* It will consume long time for reset of false card,
		 * so update the auto delink timer here
		 */
		c->cm->cr->auto_delink_timer = cr_get_tick();
	} else {
		rts_trace(TRACE_POLL_THRD, "%s: skip process_card_reset!\n",
			c->name);
	}

	if (status)
		ret = process_status_further(c, status);

	return ret;
}

/*
 * cm_update_cards_status - Main entry to update status of all cards under
 * control of cm
 * @card: card structure
 * @options: customized conditions from ultimate user.
 *
 * Returns: 0 for success, otherwise nonzero
 *
 * Description: This function is used to monitor status of all cards, and makes
 * corresponding update.
 */
int cm_update_cards_status(struct card_manager *cm,
		struct cr_option options)
{
	struct list_head *pos;
	int status = 0;
	int ret = 0;

	if (cm->cr->need_reset)
		check_cd_deglitch(cm->cr);

	cr_cd_debounce(cm->cr);

	/* process OCP */
	cr_process_ocp(cm->cr);

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (!((1 << c->id) & (cm->cc_sw.card_spt_map)))
			continue;
		if (check_card_flag(c, EXIST) && c->update_report_status)
			status = c->update_report_status(c);
		ret = process_card_status(c, status);
	}

	return ret;
}

void cm_set_reset_map_directly(struct card_manager *cm)
{
	struct list_head *pos;
	unsigned long reset_map = 0;

	rts_trace(TRACE_CR, "<--\n");

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		rts_trace(TRACE_CM, "card:%s\n", c->name);
		if (test_bit(c->id, &(cm->cr->need_reset))) {
			set_card_flags(c, EXIST);
			set_bit(c->id, &reset_map);
			/* edwin added: after reset_map is set,
			 * clear up the bit */
			clear_bit(c->id, &(cm->cr->need_reset));
		}
	}

	cm->cr->reset_map = reset_map;
	rts_trace(TRACE_CR, "--> reset_map:0x%lx", cm->cr->reset_map);
}
/*
 * cm_process_card_reset_directly - Do card reset during initialisation stage
 * @cm: points to card manager structure
 * @reset_map: logical card reset map
 *
 * Returns: 0 for success, otherwise nonzero
 *
 * Description: due to senario that hw cannot trigger interrupt for the card
 * has been plugged into card slot in the initialisation stage, we have to
 * perform reset process directly, instead of doing it after debounce process
 * in polling thread.
 */
int cm_process_card_reset_directly(struct card_manager *cm,
						unsigned long reset_map)
{
	struct list_head *pos;
	int ret = 0;

	rts_trace(TRACE_CM, "reset_map: 0x%lx\n", reset_map);
	list_for_each(pos, &cm->card_list) {
		struct card *c;
		ret = STATUS_FAIL;
		c = list_entry(pos, struct card, link);
		if (test_bit(c->id, &reset_map)) {
			ret = process_card_reset(c);
			cr_notify_upper(cm->cr, c, ret);
		}
	}

	return ret;
}

/* cm_get_card - Find one card by CARD_ID
 * @cm: card manager structure
 * @id: card id
 *
 * Returns: if found, return pointer to the card struct, otherwise, NULL
 */
struct card *cm_get_card(struct card_manager *cm, enum CARD_ID id)
{
	struct list_head *pos;
	unsigned long flags = 0;

	spin_lock_irqsave(&cm->lock, flags);

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (id == c->id) {
			rts_trace(TRACE_CM, "match, card:%s with id=%d\n",
								c->name, id);
			spin_unlock_irqrestore(&cm->lock, flags);
			return c;
		}
	}

	spin_unlock_irqrestore(&cm->lock, flags);

	return NULL;
}
EXPORT_SYMBOL_GPL(cm_get_card);

/* cm_get_ready_card - Find one card by CARD_ID
 * @cm: card manager structure
 * @id: card id
 *
 * Returns: if found, return pointer to the card struct, otherwise, NULL
 */
struct card *cm_get_ready_card(struct card_manager *cm, enum CARD_ID id)
{
	struct list_head *pos;
	unsigned long flags = 0;


	spin_lock_irqsave(&cm->lock, flags);

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (check_card_flag(c, READY) && id == c->id) {
			rts_trace(TRACE_POLL_THRD, "match, card:%s with id=%d\n",
								c->name, id);
			spin_unlock_irqrestore(&cm->lock, flags);
			return c;
		} else {
			rts_trace(TRACE_POLL_THRD, "%s is matched but not ready\n",
								c->name);
		}
	}

	spin_unlock_irqrestore(&cm->lock, flags);

	return NULL;
}
EXPORT_SYMBOL_GPL(cm_get_ready_card);

/* cm_get_id_by_lun - Get CARD_ID card by lun
* @cm: card manager structure
* @lun: lun for the card
*
* Returns: if found, return card id, otherwise return INVAL_ID
*/
enum CARD_ID cm_get_id_by_lun(struct card_manager *cm, int lun)
{
	struct list_head *pos;
	unsigned long flags = 0;

	spin_lock_irqsave(&cm->lock, flags);

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (check_card_flag(c, READY) && (lun == c->lun)) {
			spin_unlock_irqrestore(&cm->lock, flags);
			return c->id;
		}
	}
	spin_unlock_irqrestore(&cm->lock, flags);

	return INVAL_ID;
}

/* cm_locate_card - Locate card based on LUN (obsolete)
 * @cm: card manager structure
 * @lun: card LUN(Logical Unit Number)
 *
 * Returns: if found, return pointer to the card struct, otherwise, NULL
 */
struct card *cm_locate_card(struct card_manager *cm, u8 lun)
{
	struct list_head *pos;
	unsigned long flags = 0;

	spin_lock_irqsave(&cm->lock, flags);

	if (NULL != cm->cur_card) {
		rts_trace(TRACE_POLL_THRD, "cm->cur_card is: %s, lun:%d\n",
			cm->cur_card->name, cm->cur_card->lun);
		if (lun == cm->cur_card->lun) {
			spin_unlock_irqrestore(&cm->lock, flags);
			return cm->cur_card;
		}
	}

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (lun == c->lun) {
			rts_trace(TRACE_CM, "match, card:%s with lun =%d\n",
				c->name, lun);
			cm->cur_card = c;
			spin_unlock_irqrestore(&cm->lock, flags);
			return c;
		} else {
			rts_trace(TRACE_CM, "%s is matched but not ready\n",
				c->name);
		}
	}

	spin_unlock_irqrestore(&cm->lock, flags);

	rts_trace(TRACE_POLL_THRD, "Card is not found to match lun: %d\n", lun);

	return NULL;
}

/* cm_locate_card - Locate card based on Channel, Trarget and LUN
 * @cm: card manager structure
 * @sdev: scsi_device structure
 *
 * Returns: if found, return pointer to the card struct, otherwise, NULL
 */
struct card *cm_locate_card_new(struct card_manager *cm,
						struct scsi_device *sdev)
{
	struct list_head *pos;
	struct card *c = NULL;
	struct card *cur_c = cm->cur_card;
	unsigned long flags = 0;

	rts_trace(TRACE_POLL_THRD, "<--\n");

	spin_lock_irqsave(&cm->lock, flags);

	if (NULL != cur_c) {
		if (cur_c->channel == sdev->channel &&
			cur_c->target == sdev->id &&
			cur_c->lun == sdev->lun) {
			rts_trace(TRACE_CM, "cur_card info: %s "
					"[channel:target:lun]:%d%d%d\n",
						cur_c->name, cur_c->channel,
						cur_c->target, cur_c->lun);
			if (check_card_flag(cur_c, READY)) {
				spin_unlock_irqrestore(&cm->lock, flags);
				return cur_c;
			}
		}
	}

	list_for_each(pos, &cm->card_list) {
		c = list_entry(pos, struct card, link);
		if (c->channel == sdev->channel && c->target == sdev->id &&
			c->lun == sdev->lun) {
			rts_trace(TRACE_CM, "card info: %s "
					"[channel:target:lun]:%d%d%d\n",
						c->name, c->channel,
						c->target, c->lun);
		} else {
			/* Store info from scsi host */
			c->channel = sdev->channel;
			c->target = sdev->id;
		}

		if (check_card_flag(c, READY)) {
			rts_trace(TRACE_CM, "Match, card info: %s "
					"[channel:target:lun]:%d%d%d\n",
						c->name, c->channel,
						c->target, c->lun);
			cur_c = c;
			spin_unlock_irqrestore(&cm->lock, flags);
			return c;
		} else {
			rts_trace(TRACE_CM, "%s is matched but not ready\n",
								c->name);
		}
	}

	rts_trace(TRACE_POLL_THRD, "Card is not found to match "
					"[channel:target:lun]:%d%d%d\n",
					c->channel, c->target, c->lun);
	spin_unlock_irqrestore(&cm->lock, flags);

	return NULL;
}

/* check_card_readiness - Check whether the card is reader to perform some
 * operation
 * @card: card structure
 * @options: options to be checked
 * @dir: DMA direction
 *
 * Returns: 0 for success, otherwise return sense type
 */
int check_card_readiness(struct card *card, u16 options,
		enum dma_data_direction dir)
{
	struct card_reader *cr = card->cm->cr;
	int sense_type = 0;

	if ((options & READY) && !check_card_flag(card, RPM_READY) &&
		(!check_card_flag(card, READY) || card->capacity == 0))
		sense_type = SENSE_TYPE_MEDIA_NOT_PRESENT;
	rts_trace(TRACE_POLL_THRD, "sense_type1 = 0x%x\n", sense_type);

	if (check_dev_flag(cr, CHECK_MEDIA_CHANGE)
			&& (options & MC) && (!check_card_flag(card, MC))) {
		set_card_flags(card, MC);
		sense_type = SENSE_TYPE_MEDIA_CHANGE;
	}
	rts_trace(TRACE_POLL_THRD, "sense_type2 = 0x%x\n", sense_type);

#ifdef SUPPORT_SD_LOCK
	if ((options & ERASING) && (check_card_flag(card, ERASING)))
		sense_type = SENSE_TYPE_MEDIA_READ_FORBIDDEN;
	rts_trace(TRACE_POLL_THRD, "sense_type3 = 0x%x\n", sense_type);

	if ((options & LOCKED) && (check_card_flag(card, LOCKED)))
		sense_type = SENSE_TYPE_MEDIA_READ_FORBIDDEN;
	rts_trace(TRACE_POLL_THRD, "sense_type4 = 0x%x\n", sense_type);
#endif

	if ((options & WP) && (DMA_TO_DEVICE == dir
	      && check_card_flag(card, WP))) {
		rts_trace(TRACE_CM, "Write protected card!\n");
		sense_type = SENSE_TYPE_MEDIA_WRITE_PROTECT;
	}

	rts_trace(TRACE_POLL_THRD, "sense_type5 = 0x%x\n", sense_type);

	return sense_type;
}

/* card_rw - Perform card read/write operation
 * @srb: SCSI cmd received
 * @card: card structure
 * @lba: Logical Block Address
 * @num: Block number
 *
 * Returns: 0 for success, otherwise nonzero
 */
int card_rw(struct scsi_cmnd *srb, struct card *card, u32 lba, u16 num)
{
	int sense_type_tmp, sense_type = 0;
	int ret = STATUS_SUCCESS;
	int i;
	struct card_reader *cr = card->cm->cr;

	rts_trace(TRACE_CM, "%s card, sector addr: 0x%x, sector cnt: %d\n",
		       (srb->sc_data_direction ==
			DMA_TO_DEVICE) ? "Write" : "Read", lba, num);

	rts_trace(TRACE_CM, "card:%s, lun: %d\n", card->name, (int)SCSI_LUN(srb));

	sense_type = check_card_readiness(card, OPTIONS_ALL,
			srb->sc_data_direction);
	if (sense_type)
		goto err_out;

	if (srb->sc_data_direction == DMA_FROM_DEVICE)
		sense_type_tmp = SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR;
	else
		sense_type_tmp = SENSE_TYPE_MEDIA_WRITE_ERR;

	if (cr->option.force_stop_after_rw_fail &&
		check_card_flag(card, FAIL_RW)) {
		rts_trace(TRACE_ERROR, "Stop card RW after RW failed ever\n");
		sense_type = sense_type_tmp;
		goto err_out;
	}

	for (i = 0; i < 4; i++) {
		card->rw_need_retry = 0;
		card->rw_retry_cnt = i;
		ret = card->rw(srb, card, lba, num);
		if (ret != STATUS_SUCCESS) {
			sense_type = sense_type_tmp;
			if (cr->option.force_stop_after_rw_fail) {
				rts_trace(TRACE_ERROR, "set flag FAIL_RW\n");
				set_card_flags(card, FAIL_RW);
				break;
			}
		} else {
			sense_type = 0;
		}

		if (!card->rw_need_retry)
			break;

		rts_trace(TRACE_CM, "Retry RW, (i = %d)\n", i);
	}

err_out:
	return sense_type;
}

/* cm_eject_card - Ejext card with LUN as @lun
 * @cm: card manager structure
 * @lun: card LUN number
 *
 * Returns: Nothing
 */
void cm_eject_card(struct card_manager *cm, u8 lun)
{
	struct card *card;

	rts_trace(TRACE_CM, "eject card\n");

	card = cm_locate_card(cm, lun);
	if (NULL == card) {
		rts_trace(TRACE_CM, "eject card: card cannot be found"
				"to match lun:%d\n", lun);
		return;
	}

	do_remaining_work(card);
	card->release(card);
	set_card_flags(card, EJECTED);
	clr_card_flags(card, READY);
	card->capacity = 0;
}

/* eject_card - Eject card
 * @card: card structure
 *
 * Returns: Nothing
 */
void eject_card(struct card *card)
{
	rts_trace(TRACE_CM, "eject card\n");

	do_remaining_work(card);
	card->release(card);
	set_card_flags(card, EJECTED);
	clr_card_flags(card, READY);
	card->capacity = 0;
	cr_notify_upper(card->cm->cr, card, 0);
}

/* init_card_specific - Intialise card attributions and callback
 * @card: card structure
 * @cm: card manager structure
 * @card_para: card configuration parameters
 *
 * Returns: 0 for success, otherwise nonzero
 */
int init_card_specific(struct card *card, struct card_manager *cm,
		struct card_config_parameter *card_para)
{
	int ret = 0;
	struct card_callback *cb;

	cb = &(card_para->callback);
	if (!card_para->name) {
		printk("Card name is not assigned, exit\n");
		return -EINVAL;
	}

	card->name = card_para->name;
	card->id = card_para->id;
	card->lun = card_para->lun;
	card->reset_card = cb->reset;
	card->disable_card = cb->disable;
	card->rw = cb->rw;
	card->cleanup = cb->cleanup;
	card->release = cb->release;
	card->update_report_status = cb->update_report_status;
	card->cm = cm;
	card->max_reset_retry = 5;

	return ret;
}

/* register_card - Register card with card manager
 * @cm: card manager structure
 * @card: card structure
 *
 * Returns: Nothing
 */
void register_card(struct card_manager *cm, struct card *card)
{
	list_add_tail(&card->link, &cm->card_list);
}

/* unregister_card - Unregister card from card manager
 * @cm: card manager structure
 * @card: card structure
 *
 * Returns: Nothing
 */
void unregister_card(struct card_manager *cm, struct card *card)
{
	list_del(&card->link);
}

/* cm_alloc - Allocate memory for card manager structure
 *
 * Returns: address to the card manager structure
 */
struct card_manager *cm_alloc(void)
{
	struct card_manager *cm = (struct card_manager *)
			kzalloc(sizeof(struct card_manager), GFP_KERNEL);
	if (cm)
		INIT_LIST_HEAD(&cm->card_list);

	return cm;
}

/* card_alloc - Allocate struct card
 * @preserve_size: extra size
 *
 * Returns: return address of struct card
 *
 * Description: if preserve_size is not equal to zero, and allocation is done
 * successfully the card->extra will be assigned with the address of the
 * preserved area
 */
struct card *card_alloc(u32 preserve_size)
{
	struct card *card = NULL;

	rts_trace(TRACE_CM, "<--\n");

	if (0 == preserve_size) {
		printk("extra size is ZERO for card\n");
		goto err_alloc_card;
	}

	card = kzalloc(sizeof(struct card), GFP_KERNEL);
	if (NULL == card) {
		printk("%s: ERROR on malloc for card\n", __func__);
		goto err_alloc_card;
	}

	card->extra = kzalloc(preserve_size, GFP_KERNEL);
	if (NULL == card->extra) {
		printk("%s: ERROR on malloc for extra\n", __func__);
		goto err_alloc_extra;
	}

	rts_trace(TRACE_CM, "card=%p\n", card);
	rts_trace(TRACE_CM, "card->extra=%p\n", card->extra);
	rts_trace(TRACE_CM, "-->\n");
	return card;

err_alloc_extra:
	kfree(card);
	card = NULL;
err_alloc_card:
	return card;
}

/* cm_release - Release card manager structure and all the card strutures
 * under control of cm
 * @cm: card manager structure
 *
 * Returns: Nothing
 */
void cm_release(struct card_manager *cm)
{
	struct list_head *pos;
	struct list_head *n;

	rts_trace(TRACE_CM, "<--\n");

	BUG_ON(NULL == cm);

	list_for_each_safe(pos, n, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (c->extra)
			kfree(c->extra);
		kfree(c);
	}

	kfree(cm);

	rts_trace(TRACE_CM, "-->\n");
}

/* cm_get_max_lun - Seach the maximum LUN number
 * @cm: card manager structure
 *
 * Returns: the max lun
 */
int cm_get_max_lun(struct card_manager *cm)
{
	struct list_head *pos;
	u8 lun = 0;

	rts_trace(TRACE_CM, "<--\n");

	if (NULL == cm) {
		rts_trace(TRACE_CM, "cm is NULL\n");
		return -EINVAL;
	}

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (c->lun > lun)
			lun = c->lun;
	}

	rts_trace(TRACE_CM, "max_lun = %d, -->\n", lun);

	return lun;
}

extern struct card_config_parameter card_sets[];

/* cm_init - Initialise struct card_manager with card configuration
 * @cm: card manager structure
 * @cc: configuration of support card
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cm_init(struct card_manager *cm, struct card_conf *cc)
{
	struct card *card = NULL;
	u8 card_spt_map = cc->card_spt_map;
	u32 card_lun_map = cc->card_lun_map;
	u8 lun = 0;
	u8 card_count = 0;
	int i, ret = 0;

	rts_trace(TRACE_CM, "<--\n");

	INIT_LIST_HEAD(&cm->card_list);
	spin_lock_init(&cm->lock);
	cm->cc_sw = *cc;

	for (i = 0; i < REALTEK_MAX_CARD_ID; i++) {
		if (!(card_spt_map & (0x01 << i)))
			continue;
		card = card_alloc(card_sets[i].extra_size);
		if (NULL == card) {
			ret = -ENOMEM;
			goto err_card_alloc;
		}
		ret = init_card_specific(card, cm, &card_sets[i]);
		if (ret)
			goto err_init_card_specific;

		lun = (card_lun_map >> i*4) & 0x0f;
		if (SIG_LUN == cc->lun_mod)
			card->lun = LUN0;
		else if (card->lun != lun && MUL_LUN == cc->lun_mod)
			card->lun = lun;
		else
			rts_trace(TRACE_CM, "Keep the default lun setting\n");
		register_card(cm, card);
		card_count++;
	};

	rts_trace(TRACE_CM, "Currently supported card num:%d\n", card_count);

	rts_trace(TRACE_CM, "-->\n");

	return 0;

err_init_card_specific:
err_card_alloc:
	return ret;
}

/* set_card_flags - Set card flags
 * @card: card structure
 * @bit: card flag bit to be set
 *
 * Returns: Nothing
 */
void set_card_flags(struct card *card, u16 bit)
{
	card->card_flags |= (bit);
}
EXPORT_SYMBOL_GPL(set_card_flags);

/* clr_card_flags - Clear card flags
 * @card: card structure
 * @bit: card flag bit to be cleared
 *
 * Returns: Nothing
 */
void clr_card_flags(struct card *card, u16 bit)
{
	card->card_flags &= ~(bit);
}
EXPORT_SYMBOL_GPL(clr_card_flags);

/* check_card_flag - Check whether card flags set or not
 * @card: card structure
 * @flag: card flag to be checked
 *
 * Returns: 1 for set, otherwise 0
 */
int check_card_flag(struct card *card, u16 flag)
{
	return (NULL != card && card->card_flags & flag);
}
EXPORT_SYMBOL_GPL(check_card_flag);

/* cm_check_card_flag - Check whether card flags set or not based on card LUN
 * @cm: card manager structure
 * @lun: card LUN
 * @flag: card flag to be checked
 *
 * Returns: 1 for set, otherwise 0
 */
int cm_check_card_flag(struct card_manager *cm, u8 lun, u16 flag)
{
	struct card *card = cm_locate_card(cm, lun);

	return (NULL != card && card->card_flags & flag);
}
EXPORT_SYMBOL_GPL(cm_check_card_flag);

/* cm_check_cards_flags - Check if there is one card with specified flag set
 * @cm: card manager structure
 * @flag: card flag to be checked
 *
 * Returns: 1 for set, otherwise 0
 */
int cm_check_cards_flags(struct card_manager *cm, u16 flag)
{
	struct list_head *pos;
	int ret = 0;

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (c->card_flags & flag) {
			ret = 1;
			break;
		}
	}

	return ret;
}
EXPORT_SYMBOL_GPL(cm_check_cards_flags);

/* check_fake_card_ready - Check if there is fake card ready or not
 * @cr: card reader structure
 * @lun: card LUN
 *
 * Returns: 1 for ready, otherwise 0
 */
int check_fake_card_ready(struct card_reader *cr, u8 lun)
{
#if 0 /* edwin: comment out currently */
	if (cr->fake_card_ready & cr->lun2card[lun])
		return 1;
#else
	if (cm_check_card_flag(cr->cm, lun, RPM_READY) ||
		cm_check_card_flag(cr->cm, lun, READY))
		return 1;
#endif
	return 0;
}

int cm_get_led_state_at_ss(struct card_manager *cm)
{
	struct cr_option *option = &(cm->cr->option);
	struct list_head *pos;
	int on = 0;

	if (!(option->led_always_on))
		return 0;

	list_for_each(pos, &cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (check_card_flag(c, READY)) {
			if (!check_card_flag(c, EJECTED)) {
				on = 1;
				break;
			}
		}
	}

	return on;
}
