/* Driver for Realtek card reader
 * Header file
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#ifndef RTS_CR_CARD_H
#define RTS_CR_CARD_H

#include <linux/spinlock.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>

#include "cards.h"

#define DEBOUNCE_CNT	5

enum LUN_MOD {
	SIG_LUN = 0,
	MUL_LUN,
};

/* Specify lun for each supported card */
enum LUNS_ALLOC {
	LUN0 = 0,
	LUN1,
	LUN2,
	LUN_MAX,
};

struct card_conf {
	enum LUN_MOD lun_mod;
	/* Each of the least 3 bits denotes whether the card is supported or
	 * not, the order is bit2: xd card, bit1: ms card, bit0: sd card */
	u8 card_spt_map;
	/* The lun number for each card is defined with 4 bits, the order is
	 * depicted as below: bit8~11: xd card lun, bit4~7: ms card lun, bit0~3:
	 * sd card lun, if card is not support, the corresponding value in
	 * card_lun_map makes nonsense */
	u32 card_lun_map;
	u32 sd_capability;
};

/* NOTE: the value of xd_lun, ms_lun, sd_lun cannot be larger than 8.  if a new
 * card needs to be supported, a new Marcro extended from this should be defined
 * as so on */
#define CARD_CONF(xd, ms, sd, xd_lun, ms_lun, sd_lun) \
{ \
	.card_spt_map = (((xd) << 2) | ((ms) << 1) | (sd)),\
	.card_lun_map = (((xd_lun) << 8) | ((ms_lun) << 4) | (sd_lun)), \
}

struct card_reader;
struct card_manager {
	char *name;
	/* list_head support_card_list; */
	struct list_head card_list;
	struct card *cur_card;
	spinlock_t lock; /* the lock is used to protect cur_card field */
	struct card_reader *cr;
	struct transport *tr;
	struct card_conf cc_sw;
};

struct card_callback {
	int (*reset)(struct card *c);
	int (*disable)(struct card *c);
	int (*rw)(struct scsi_cmnd *srb, struct card *c, u32 lba, u16 num);
	int (*cleanup)(struct card *c);
	int (*release)(struct card *c);
	int (*update_report_status)(struct card *c);
	int (*enable)(struct card *c);
};

struct card_config_parameter {
	char *name;
	/* The member is used to set the size of pre-allocated
	 * memory space to store card-specific parameters */
	u32 extra_size;
	/* Specify the card id */
	u8 id;
	/* Specify the lun for each card controller, if cards
	 * are armed with the same lun, which means, the card
	 * reader only support one card each time*/
	u8 lun;
	struct card_callback callback;
};

struct cr_capability;
struct card_config_parameter;
struct card_manager *cm_alloc(void);
int cm_get_max_lun(struct card_manager *cm);
int cm_init(struct card_manager *cm, struct card_conf *cc);
void cm_release(struct card_manager *cm);
void register_card(struct card_manager *cm, struct card *card);
void unregister_card(struct card_manager *cm, struct card *card);
void set_card_flags(struct card *card, u16 bit);
void clr_card_flags(struct card *card, u16 bit);
int check_card_flag(struct card *card, u16 flag);
int cm_check_cards_flags(struct card_manager *cm, u16 flag);
struct cr_option;
int cm_update_cards_status(struct card_manager *cm,
		struct cr_option options);
void cm_set_reset_map_directly(struct card_manager *cm);
int cm_process_card_reset_directly(struct card_manager *cm,
						unsigned long reset_map);
struct card *cm_locate_card(struct card_manager *cm, u8 lun);
struct card *cm_locate_card_new(struct card_manager *cm,
					struct scsi_device *sdev);
struct card *cm_get_card(struct card_manager *cm, enum CARD_ID id);
struct card *cm_get_ready_card(struct card_manager *cm, enum CARD_ID id);
int check_card_readiness(struct card *card, u16 options,
		enum dma_data_direction dir);
void do_remaining_work(struct card *c);
void cm_do_remaining_work(struct card_manager *cm);
void cm_status_runtime_save(struct card_manager *cm);
void cm_status_runtime_clear(struct card_manager *cm);
void cm_release_cards(struct card_manager *cm);
int card_rw(struct scsi_cmnd *srb, struct card *card,
		u32 sec_addr, u16 sec_cnt);
void eject_card(struct card *card);
void cm_eject_card(struct card_manager *cm, u8 lun);

int cm_enable_card_clock(struct card_manager *cm, u8 card);
int cm_disable_card_clock(struct card_manager *cm, u8 card);
int cm_card_power_on(struct card_manager *cm, u8 card);
int cm_card_power_off(struct card_manager *cm, u8 card);
int cm_check_card_flag(struct card_manager *cm, u8 lun, u16 flag);
int check_fake_card_ready(struct card_reader *cr, u8 lun);
inline int cr_enable_msi(struct card_reader *cr);
enum CARD_ID cm_get_id_by_lun(struct card_manager *cm, int lun);
int cm_get_led_state_at_ss(struct card_manager *cm);
int process_card_reset(struct card *c);
int cr_enable_oobs_polling(struct card_reader *cr);
int cr_disable_oobs_polling(struct card_reader *cr);

#endif /* RTS_CR_CARD_H */
