/* Driver for Realtek card reader
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>
#include <linux/dma-mapping.h>
#include <linux/list.h>
#include <scsi/scsi.h>
#include <linux/module.h>
#include <linux/pci.h>

#include "rts_cr.h"
#include "cardreader.h"
#include "pm.h"
#include "rts_scsi.h"
#include "operations.h"
#include "cr_regs.h"
#include "cards/sd.h"
#include "cards/ms.h"
#include "cards/xd.h"
#include "cr_system.h"

static const char *cr_stat_str[] = {
	"INIT",
	"IDLE",
	"RUN",
	"SS",
	"DELINK",
	"SUSPEND",
	"ABORT",
	"DISCONNECT",
};

static void cr_deal_extra_action(struct card_reader *cr);
/*
 * cr_set_stat - Set current state of card reader
 * @cr: card reader structure
 * @stat: one state of card reader state machine
 *
 * This function is used to set state of card reader, edwin suspicition: does it
 * need a lock to guarantee the set_stat() behavior ??
 */
void cr_set_stat(struct card_reader *cr, enum RTSX_STAT stat)
{
	if (stat != RTSX_STAT_IDLE)
		cr->idle_counter = 0;

	if (cr->rtsx_stat != stat) {
		rts_trace(TRACE_CR, "set stat from %s to %s\n",
			cr_stat_str[cr->rtsx_stat], cr_stat_str[stat]);
		cr->rtsx_stat = stat;
	}

	switch (stat) {
	case RTSX_STAT_RUN:
		cr->polling_interval = cr->option.polling_interval_run;
		break;
	case RTSX_STAT_IDLE:
		cr->polling_interval = cr->option.polling_interval_idle;
		break;
	case RTSX_STAT_SS:
		cr->polling_interval = cr->option.polling_interval_ss;
		break;
	default:
		cr->polling_interval = cr->option.polling_interval_run;
		break;
	}
}
EXPORT_SYMBOL_GPL(cr_set_stat);

/* cr_rd_reg - Read content of register with specified addr to data area
 * @cr: card reader structure
 * @addr: address of register
 * @data: store the content of the register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_rd_reg(struct card_reader *cr, u16 addr, u8 *data)
{
	if (cr->ops->rd_reg)
		return (cr->ops->rd_reg)(cr, addr, data);
	else
		return cr_common_rd_reg(cr, addr, data);

}

EXPORT_SYMBOL_GPL(cr_rd_reg);

/* cr_wr_reg - Write data to register with specified addr
 * @cr: card reader structure
 * @addr: address of register
 * @mask: effective bits mapping to the data
 * @data: value to be written to register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_wr_reg(struct card_reader *cr, u16 addr, u8 mask, u8 data)
{
	if (cr->ops->wr_reg)
		return (cr->ops->wr_reg)(cr, addr, mask, data);
	else
		return cr_common_wr_reg(cr, addr, mask, data);
}
EXPORT_SYMBOL_GPL(cr_wr_reg);

/* cr_rd_phy_u16 - Read content of phy register with specified addr to val for
 * PCIe interface
 * @cr: card reader structure
 * @addr: address of register
 * @val: store the content of the register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_rd_phy16(struct card_reader *cr, u8 addr, u16 *val)
{
	if (cr->ops->rd_phy16)
		return (cr->ops->rd_phy16)(cr, addr, val);
	else
		return cr_common_rd_phy16(cr, addr, val);

}
EXPORT_SYMBOL_GPL(cr_rd_phy16);

/* cr_wr_phy_u16 - Write val to phy register with specified addr for PCIe
 * interface
 * @cr: card reader structure
 * @addr: address of register
 * @val: value to be written to register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_wr_phy16(struct card_reader *cr, u8 addr, u16 val)
{
	if (cr->ops->wr_phy16)
		return (cr->ops->wr_phy16)(cr, addr, val);
	else
		return cr_common_wr_phy16(cr, addr, val);
}
EXPORT_SYMBOL_GPL(cr_wr_phy16);

void cr_patch_rtd3(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_PME_FORCE_CTL,
		FORCE_PME_CONTROL | SEND_PME_FOR_CARD, 0);
	cr_wr_reg(cr, AUTOLOAD_CFG3, U_PM_D3_EN, 0);
}
EXPORT_SYMBOL_GPL(cr_patch_rtd3);

/* cr_rd_reg_seq - Sequentially read registers
 * @cr: card reader structure
 * @addr: starting address of registers
 * @len: the count of the registers to be read
 * @data: starting address to store the content of registers
 *
 * Returns: 0 for success, otherwise nonzero
 * Description: read the addresses of registers starts from @addr to
 * @addr + @count to area
 * starting from the location pointing by @data
 */
int cr_rd_reg_seq(struct card_reader *cr, u16 addr, u16 len, u8 *data)
{
	int ret = 0;

	if (cr->ops->rd_reg_seq) {
		ret = (cr->ops->rd_reg_seq)(cr, addr, len, data);
	} else {
		printk("%s: rd_reg_seq callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}

/* cr_wr_reg_seq - Sequentially write registers
 * @cr: card reader structure
 * @addr: starting address of registers
 * @len: the count of the registers to be read
 * @data: starting address of data to be written to the registers
 *
 * Returns: 0 for success, otherwise nonzero
 * Description: write data starting from the location pointing by @data to
 * registers
 * starts from @addr to @addr + @count
 */
int cr_wr_reg_seq(struct card_reader *cr, u16 addr, u16 len, u8 *data)
{
	int ret = 0;

	if (cr->ops->wr_reg_seq) {
		ret = (cr->ops->wr_reg_seq)(cr, addr, len, data);
	} else {
		printk("%s: wr_reg_seq callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}

int cr_rd_cfg8(struct card_reader *cr, u16 addr, u8 *val)
{
	int ret;
	u16 tmp;
	u32 value;

	tmp = addr & 0xFFFC;
	ret = cr_rd_cfg32(cr, tmp, &value);
	*val = (u8)(value >> ((addr & 0x3) * 8));

	return ret;
}
EXPORT_SYMBOL_GPL(cr_rd_cfg8);

int cr_wr_cfg8(struct card_reader *cr, u16 addr, u8 val)
{
	int ret;
	u16 tmp;

	tmp = addr & 0xfffc;
	ret = cr_wr_cfg32(cr, tmp, 0xFF << ((addr & 0x03) * 8),
			((u32)val) << ((addr & 0x03) * 8));
	return ret;
}
EXPORT_SYMBOL_GPL(cr_wr_cfg8);

/* cr_rd_cfg32 - Read PCIe configuration register
 * @cr: card reader structure
 * @addr: address of register
 * @val: store the content of the register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_rd_cfg32(struct card_reader *cr, u16 addr, u32 *val)
{
	if (cr->ops->rd_cfg32)
		return (cr->ops->rd_cfg32)(cr, addr, val);
	else
		return cr_common_rd_cfg32(cr, addr, val);

}
EXPORT_SYMBOL_GPL(cr_rd_cfg32);

/* cr_wr_cfg32 - Write PCIe configuration register
 * @cr: card reader structure
 * @addr: address of register
 * @val: data toe be written to the register
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_wr_cfg32(struct card_reader *cr, u16 addr, u32 mask, u32 val)

{
	if (cr->ops->wr_cfg32)
		return (cr->ops->wr_cfg32)(cr, addr, mask, val);
	else
		return  cr_common_wr_cfg32(cr, addr, mask, val);
}
EXPORT_SYMBOL_GPL(cr_wr_cfg32);

/* cr_get_card_status - Obtain the status of card
 * @cr: card reader structure
 * @status: where to store the status of card
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_get_card_status(struct card_reader *cr, u16 *status)
{
	int ret = 0;

	if (cr->ops->get_card_status) {
		ret = (cr->ops->get_card_status)(cr, status);
	} else {
		printk("%s: get_card_status callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}

/* cr_init_cmd - Initialise a command following IC convension
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
inline void cr_init_cmd(struct card_reader *cr)
{
	tr_init_cmd(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_init_cmd);

/* cr_init_cmd - Add a command following IC convension
 * @cr: card reader structure
 * @cmd_type: one in cmd type READ_REG_CMD/WRITE_REG_CMD/CHECK_REG_CMD
 * @reg_addr: register address to be operated
 * @mask: effective bits mapping to @data
 * @data: make sense for WRITE_REG_CMD only
 *
 * Returns: Nothing
 */
inline void cr_add_cmd(struct card_reader *cr, u8 type,
		u16 addr, u8 mask, u8 data)
{
	tr_add_cmd(cr->tr, type, addr, mask, data);
}
EXPORT_SYMBOL_GPL(cr_add_cmd);

/* cr_send_cmd - Send command(s)
 * @cr: card reader structure
 * @flag: one in cmd stage mode MODE_C/MODE_CR/MODE_CDIR/MODE_CDOR
 * @timeout: limited value of timeout
 *
 * Returns: Nothing
 */
inline int cr_send_cmd(struct card_reader *cr, u8 flag, int timeout)
{
	return tr_send_cmd(cr->tr, flag, timeout);
}
EXPORT_SYMBOL_GPL(cr_send_cmd);

/* cr_send_cmd_no_wait - Send command(s) without wating
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
inline void cr_send_cmd_no_wait(struct card_reader *cr)
{
	tr_send_cmd_no_wait(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_send_cmd_no_wait);


/*
 * If cr_stop_cmd is inline function, Everest on windows
 * would give out compile error
 */
void cr_stop_cmd(struct card_reader *cr, int card)
{
	tr_stop_cmd(cr->tr, card);
}
EXPORT_SYMBOL_GPL(cr_stop_cmd);

inline u8 *cr_get_cmd_data(struct card_reader *cr)
{
	return tr_get_cmd_data(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_get_cmd_data);

/* cr_get_rsp - Get the response from card
 * @cr: card reader structure
 * @rsp_len: the lenght of response
 * @timeout: limited value of timeout
 *
 * Returns: STATUS_SUCCESS for success, otherwise nonzero
 *
 * Description: the response is stored in cr->rsp_buf
 */
int cr_get_rsp(struct card_reader *cr, int rsp_len, int timeout)
{
	int ret = 0;

	if (cr->ops->get_rsp) {
		ret = (cr->ops->get_rsp)(cr, rsp_len, timeout);
	} else {
		printk("%s: get_rsp callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}

/* cr_read_rsp_buf - Obtain data of interest from response
 * @cr: card reader structure
 * @offset: offset to the starting address of response
 * @buf: where to store the data of interest
 * @buf_len: lenght of data of interest
 *
 * Returns: Nothing
 *
 */
void cr_read_rsp_buf(struct card_reader *cr, int offset,
				       u8 *buf, int buf_len)
{
	memcpy(buf, cr->rsp_buf + offset, buf_len);
}

/* cr_get_rsp_data - Obtain whole response
 * @cr: card reader structure
 *
 * Returns: return the pointer to the response
 */
u8 *cr_get_rsp_data(struct card_reader *cr)
{
	return cr->rsp_buf;
}

/* cr_rd_ppbuf - Read Ping Pong buffer
 * @cr: card reader structure
 * @buf: where to store the data of interest
 * @buf_len: lenght of data
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_rd_ppbuf(struct card_reader *cr, u8 *buf, int buf_len)
{
	if (cr->ops->rd_ppbuf)
		return  (cr->ops->rd_ppbuf)(cr, buf, buf_len);
	else
		return cr_common_read_ppbuf(cr, buf, buf_len);
}
EXPORT_SYMBOL_GPL(cr_rd_ppbuf);

/* cr_wr_ppbuf - Write Ping Pong buffer
 * @cr: card reader structure
 * @buf: pointing to the start address of data to be written to Ping Pong buffer
 * @buf_len: lenght of data
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_wr_ppbuf(struct card_reader *cr, u8 *buf, int buf_len)
{
	if (cr->ops->wr_ppbuf)
		return (cr->ops->wr_ppbuf)(cr, buf, buf_len);
	else
		return cr_common_write_ppbuf(cr, buf, buf_len);
}
EXPORT_SYMBOL_GPL(cr_wr_ppbuf);

/* cr_prepare_run - operations performed before real operations to card reader
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_prepare_run(struct card_reader *cr)
{
	if (cr->ops->prepare_run)
		cr->ops->prepare_run(cr);
	else
		rts_trace(TRACE_CR, "%s: ops->prepare_run() not implemented\n",
			cr->name);
}

/* cr_pp_status - Obtain device status
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pp_status(struct card_reader *cr, struct card *card,
		u8 lun, u8 *status, u8 status_len)
{
	if (cr->ops->pp_status)
		cr->ops->pp_status(cr, card, lun, status, status_len);
	else
		printk("%s: pp_status callback is not implemented\n",
								cr->name);
}

void cr_read_status(struct card_reader *cr, struct card *card,
		u8 lun, u8 *status, u8 status_len)
{
	if (cr->ops->read_status)
		cr->ops->read_status(cr, card, lun, status, status_len);
	else
		printk("%s: read_status callback is not implemented\n",
								cr->name);
}

/* cr_transfer_data - Transfer whole data
 * @cr: card reader structure
 * @dir: direction
 * @buf: data location
 * @len: data length
 * @use_sg: whether use scatter-gater mode
 * @act_len: actual data transfered
 * @timeout: limited value for transfering these data
 * @stage_flag: transfer stage
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_transfer_data(struct card_reader *cr, u8 card, void *buf,
	size_t len, int use_sg, int dir, int timeout)

{
	return tr_transfer_data(cr->tr, card, buf, len, use_sg, dir, timeout);
}
EXPORT_SYMBOL_GPL(cr_transfer_data);

int cr_transfer_data_partial(struct card_reader *cr, u8 card,
	void *buf, size_t len, int use_sg, unsigned int *index,
	unsigned int *offset, int dir, int timeout)
{
	return tr_transfer_data_partial(cr->tr, card, buf, len, use_sg, index,
			offset, dir, timeout);
}
EXPORT_SYMBOL_GPL(cr_transfer_data_partial);

/* cr_transfer_data - Transfer data partially
 * @cr: card reader structure
 * @dir: direction
 * @buf: data location
 * @ptr: update the location of data transfered
 * @offset: where the data to be transfered
 * @len: data length
 * @use_sg: whether use scatter-gater mode
 * @act_len: actual data transfered
 * @timeout: limited value for transfering these data
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_transfer_data_03(struct card_reader *cr,
				enum dir_t dir, void *buf, void **ptr,
				unsigned int *offset, unsigned int len,
				int use_sg, unsigned int *act_len, int timeout)
{
	int ret = 0;

	if (cr->ops->transfer_data_03) {
		ret = (cr->ops->transfer_data_03)(cr, dir, buf, ptr, offset,
					len, use_sg, act_len, timeout);
	} else {
		printk("%s: transfer_data_03 callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}

/* cr_clr_card_err - Clear card error
 * @cr: card reader structure
 * @id: logical ID for card
 *
 * Returns: Nothing
 */
void cr_clr_card_err(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->clr_card_err)
		cr->ops->clr_card_err(cr, id);
	else
		printk("%s: clr_card_err callback is not implemented\n",
								cr->name);
}

/* cr_select_card - Select card to be operated
 * @cr: card reader structure
 * @card: card structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_select_card(struct card_reader *cr, struct card *card)
{

	if (cr->ops->select_card)
		return cr->ops->select_card(cr, card);
	else
		return cr_common_select_card(cr, card);
}
EXPORT_SYMBOL_GPL(cr_select_card);

/* cr_poweron_card - Power on the card to be operated
 * @cr: card reader structure
 * @id: logical ID for card
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_poweron_card(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->poweron_card)
		return (cr->ops->poweron_card)(cr, id);
	else
		return cr_common_poweron_card(cr, id);
}
EXPORT_SYMBOL_GPL(cr_poweron_card);


/* cr_poweroff_card - Power off the card to be operated
 * @cr: card reader structure
 * @id: logical ID for card
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_poweroff_card(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->poweroff_card)
		return (cr->ops->poweroff_card)(cr, id);
	else
		return cr_common_poweroff_card(cr, id);
}
EXPORT_SYMBOL_GPL(cr_poweroff_card);

/* cr_enable_dma - Enable card reader DMA
 * @cr: card reader structure
 * @dir: DMA direction
 * @byte_cnt: data count to be transfered in DMA mode
 * @pack_size: package size per transmission
 *
 * Returns: Nothing
 */
void cr_enable_dma(struct card_reader *cr,
		u32 byte_cnt, u8 pack_size, enum dma_data_direction dir)
{
	if (cr->ops->enable_dma)
		cr->ops->enable_dma(cr, byte_cnt, pack_size, dir);
	else
		cr_common_enable_dma(cr, byte_cnt, pack_size, dir);
}
EXPORT_SYMBOL_GPL(cr_enable_dma);

/* cr_switch_clk - Switch clock of card reader IC in terms of card
 * @cr: card reader structure
 * @c:  card structure
 * @clk: clock value to switch to
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_switch_clk(struct card_reader *cr, struct card *c, int clk)
{

	if (cr->ops->switch_clk)
		return (cr->ops->switch_clk)(cr, c, clk);
	else
		return cr_common_switch_clock(cr, c, clk);

}
EXPORT_SYMBOL_GPL(cr_switch_clk);

/* cr_enable_card_clk - Enable card clock
 * @cr: card reader structure
 * @id:  card logical ID
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_enable_card_clk(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->enable_card_clk)
		return (cr->ops->enable_card_clk)(cr, id);
	else
		return cr_common_enable_card_clock(cr, id);

}
EXPORT_SYMBOL_GPL(cr_enable_card_clk);

/* cr_disable_card_clk - Disable card clock
 * @cr: card reader structure
 * @id:  card logical ID
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_disable_card_clk(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->disable_card_clk)
		return (cr->ops->disable_card_clk)(cr, id);
	else
		return cr_common_disable_card_clock(cr, id);
}
EXPORT_SYMBOL_GPL(cr_disable_card_clk);

/* cr_clr_hw_err - Clear card reader hardware error
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_clr_hw_err(struct card_reader *cr)
{
	if (cr->ops->clr_hw_err)
		cr->ops->clr_hw_err(cr);
	else
		printk("%s: clr_hw_err callback is not implemented\n",
								cr->name);
}
EXPORT_SYMBOL_GPL(cr_clr_hw_err);

/* cr_init_hw - Initialise card reader hardware
 * @cr: card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_init_hw(struct card_reader *cr)
{
	int ret = 0;

	if (cr->ops->init) {
		ret = (cr->ops->init)(cr);
	} else {
		printk("%s: init callback is not implemented\n", cr->name);
		ret = -EINVAL;
	}

	return ret;
}
EXPORT_SYMBOL_GPL(cr_init_hw);

/* cr_flush_rb - Flush Ring Buffer
 * @cr: card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_flush_rb(struct card_reader *cr)
{

	if (cr->ops->flush_rb)
		return (cr->ops->flush_rb)(cr);
	else
		return cr_common_flush_rb(cr);
}
EXPORT_SYMBOL_GPL(cr_flush_rb);

/* cr_detect_card - detect whether the card exist or not
 * @cr: card reader structure
 * @id: logical card ID
 *
 * Returns: CD_EXIST/CD_NOT_EXIST
 */
int cr_detect_card(struct card_reader *cr, enum CARD_ID id)
{
	if (cr->ops->detect_card)
		return (cr->ops->detect_card)(cr, id);
	else
		return cr_common_detect_card(cr, id);
}
EXPORT_SYMBOL_GPL(cr_detect_card);


int cr_rd_cfg_seq(struct card_reader *cr, u16 addr, u8 *buf, int len)
{
	if (cr->ops->rd_cfg_seq)
		return cr->ops->rd_cfg_seq(cr, addr, buf, len);
	else
		return cr_common_rd_cfg_seq(cr, addr, buf, len);
}
EXPORT_SYMBOL_GPL(cr_rd_cfg_seq);

int cr_wr_cfg_seq(struct card_reader *cr, u16 addr, u8 *buf, int len)
{

	if (cr->ops->wr_cfg_seq)
		return cr->ops->wr_cfg_seq(cr, addr, buf, len);
	else
		return cr_common_wr_cfg_seq(cr, addr, buf, len);
}
EXPORT_SYMBOL_GPL(cr_wr_cfg_seq);

int cr_dump_register(struct card_reader *cr, u16 start, u16 end)
{
	u16 len = end - start + 1;
	u16 i;
	u8 data[8] = {0}, any;

	rts_trace(TRACE_CR, "dump registers: [%x, %x]----------\n", start, end);
	for (i = 0; i < len; i += 8, start += 8) {
		u16 j, n = min(8, len - i);

		memset(data, 0, 8);
		any = 0;
		for (j = 0; j < n; j++) {
			cr_rd_reg(cr, start + j, data + j);
			any |= *(data + j);
		}
		if (any)
			rts_trace(TRACE_CR, "0x%04X(%d): %02x %02x %02x %02x %02x %02x %02x %02x\n",
				start, n, data[0], data[1], data[2], data[3],
				data[4], data[5], data[6], data[7]);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(cr_dump_register);

int cr_dump_phy(struct card_reader *cr, u8 start, u8 end)
{
	u8 len = end - start + 1;
	u8 i;
	u16 data[8] = {0}, any;

	rts_trace(TRACE_CR, "dump phy: [%x, %x]----------------\n", start, end);
	for (i = 0; i < len; i += 8, start += 8) {
		u8 j, n = min(8, len - i);

		memset(data, 0, 16);
		any = 0;
		for (j = 0; j < n; j++) {
			cr_rd_phy16(cr, start + j, data + j);
			any |= *(data + j);
		}
		if (any)
			rts_trace(TRACE_CR, "0x%04X(%d): %04x %04x %04x %04x %04x %04x %04x %04x\n",
				start, n, data[0], data[1], data[2], data[3],
				data[4], data[5], data[6], data[7]);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(cr_dump_phy);

int cr_polling_func(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	int stat = rtsx_get_stat(cr);
	int ret = 0;

	switch (stat) {
	case RTSX_STAT_INIT:
	case RTSX_STAT_RUN:
	case RTSX_STAT_IDLE:
		/* cm_do_remaining_work(cr->cm); */
		break;
	case RTSX_STAT_SS:
		break;
	case RTSX_STAT_DELINK:
		break;
	case RTSX_STAT_SUSPEND:
		break;
	case RTSX_STAT_ABORT:
		break;
	case RTSX_STAT_DISCONNECT:
		break;
	default:
		rts_trace(TRACE_PM, "Unknonwn state: %d\n", stat);
	}

	cr_deal_extra_action(cr);

	if (!rtsx_chk_stat(cr, RTSX_STAT_DELINK)) {
#if RTS_PM
		cr_pm_update_pm_real(cr);
		/* cr_pm_update_pm(cr); */
#endif
		cm_update_cards_status(cr->cm, cr->option);
	}

	if (option->auto_delink_en && option->auto_delink_allowed &&
		!cm_check_cards_flags(cr->cm, READY) &&
		!cm_check_cards_flags(cr->cm, EJECTED)) {
		cr_auto_delink(cr);
	} else {
		CLEAR_STATUS_FLAG(cr, DELINK_STEP1_DONE);
		CLEAR_STATUS_FLAG(cr, DELINK_STEP2_DONE);
		CLEAR_STATUS_FLAG(cr, DELINK_STEP3_DONE);
		cr->auto_delink_timer = cr_get_tick();
	}

	return ret;
}
EXPORT_SYMBOL_GPL(cr_polling_func);

/* cr_alloc - Allocate struct card_reader
 * @preserve_size: extra size
 *
 * Returns: return address of struct card_reader
 * Description: if preserve_size is not equal to zero, and allocation is done
 * successfully
 * the cr->extra will be assigned with the address of the preserved area
 */
struct card_reader *cr_alloc(int preserve_size)
{
	struct card_reader *cr = NULL;

	rts_trace(TRACE_CR, "<--\n");
#if 0 /* FIXME edwin: no extra for card_reader also make sense */
	if (0 == preserve_size) {
		printk("extra size is ZERO for card reader\n");
		goto err_alloc_cr;
	}
#endif

	cr = kzalloc(sizeof(struct card_reader), GFP_KERNEL);
	if (NULL == cr) {
		printk("ERROR on malloc for card reader\n");
		goto err_alloc_cr;
	}

	cr->extra = kzalloc(preserve_size, GFP_KERNEL);
	if (NULL == cr->extra) {
		printk("ERROR on malloc for extra\n");
		goto err_alloc_extra;
	}

	rts_trace(TRACE_CR, "cr=%p\n", cr);
	rts_trace(TRACE_CR, "cr->extra=%p\n", cr->extra);

	rts_trace(TRACE_CR, "-->\n");
	return cr;

err_alloc_extra:
	kfree(cr);
	cr = NULL;
err_alloc_cr:
	return cr;
}
EXPORT_SYMBOL_GPL(cr_alloc);

/* cr_init_callback - Intialise callback member of struct card_reader
 * @cr: card reader structure
 * @cb: callback configuration
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int cr_init_callback(struct card_reader *cr, struct cr_callback cb)
{
	if (!cb.ops->init) {
		printk("%s: ERR, init callback is not implemented!\n",
				cr->name);
		return -EINVAL;
	}

	cr->specific_cmd_handlers = cb.specific_cmd_handlers;
	cr->sd_ops_prot = cb.sd_ops_prot;
	cr->ms_ops_prot = cb.ms_ops_prot;
	cr->xd_ops_prot = cb.xd_ops_prot;
	cr->ops = cb.ops;
	cr->pm_ops = cb.pm_ops;

	return 0;
}
EXPORT_SYMBOL_GPL(cr_init_callback);

/* cr_init_option - Initialise cr default parameters
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_init_option(struct card_reader *cr)
{
	if (cr->ops->init_options)
		cr->ops->init_options(cr);
	else
		printk("%s: init_options callback is not implemented\n",
								cr->name);
}
EXPORT_SYMBOL_GPL(cr_init_option);

static int get_max_lun_from_capability(struct cr_capability *capa)
{
	u32 max = 0;
	u32 lun_map;

	if (SIG_LUN == capa->card_conf.lun_mod)
		return 0;

	lun_map = capa->card_conf.card_lun_map;

	while (lun_map) {
		if ((lun_map & 0xF) > max)
			max = lun_map & 0xF;
		lun_map >>= 4;
	}

	return max;
}

/* cr_init - Initialise struct card_reader with card reader configuration and
 * armed with transport
 * @cr: card reader structure
 * @cr_capa: card reader configuration
 * @tr: tranport medium abstract type of the card reader
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_init(struct card_reader *cr, struct cr_capability *cr_capa,
		struct transport *tr)
{
	struct cr_capability *__cr_capa = cr_capa;
	int ret = 0;

	rts_trace(TRACE_CR, "<--\n");

	if (NULL == tr->iobuf) {
		printk("err: buf is NULL\n");
		ret = -ENOMEM;
		goto BadDevice;
	}
	cr->cmd_buf = cr->rsp_buf = tr->iobuf;

	cr->name = cr_capa->name;
	cr->vendor_id = cr_capa->vid;
	cr->product_id = cr_capa->pid;
	/* set the supported number of lun */
	cr->max_lun = get_max_lun_from_capability(cr_capa) + 1;
	cr->lun_mod = cr_capa->card_conf.lun_mod;
	cr->cur_card = INVAL_ID;
	cr->cc_hw = cr_capa->card_conf;
	spin_lock_init(&cr->pm_lock);

	init_completion(&cr->polling_exit);

	init_waitqueue_head(&cr->wait_q);

	/* init callback of card reader */
	ret = cr_init_callback(cr, __cr_capa->callback);
	if (ret)
		goto BadDevice;

	rts_trace(TRACE_CR, "-->\n");

	return 0;
BadDevice:
	return ret;
}
EXPORT_SYMBOL_GPL(cr_init);

/* cr_release - Release card readr resource
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_release(struct card_reader *cr)
{
	rts_trace(TRACE_CR, "<--\n");

	if (NULL != cr) {
		if (NULL != cr->extra) {
			kfree(cr->extra);
		}
		kfree(cr);
	}

	rts_trace(TRACE_CR, "-->\n");
}
EXPORT_SYMBOL_GPL(cr_release);

/* cr_cd_debounce - card debounce
 *
 */
void cr_cd_debounce(struct card_reader *cr)
{
	if (cr->ops->cd_debounce)
		cr->ops->cd_debounce(cr);
	else
		cr_common_cd_debounce(cr);
}
EXPORT_SYMBOL_GPL(cr_cd_debounce);


inline void cr_enable_host_aspm(struct card_reader *cr)
{
	tr_enable_host_aspm(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_enable_host_aspm);

inline void cr_disable_host_aspm(struct card_reader *cr)
{
	tr_disable_host_aspm(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_disable_host_aspm);

inline void cr_set_host_aspm(struct card_reader *cr, u8 val)
{
	tr_set_host_aspm(cr->tr, val);
}
EXPORT_SYMBOL_GPL(cr_set_host_aspm);

inline void cr_get_host_aspm(struct card_reader *cr, u8 *val)
{
	tr_get_host_aspm(cr->tr, val);
}
EXPORT_SYMBOL_GPL(cr_get_host_aspm);


inline void cr_notify_upper(struct card_reader *cr, struct card *c, int val)
{
	tr_notify_upper(cr->tr, c, val);
}
EXPORT_SYMBOL_GPL(cr_notify_upper);

inline u32 cr_ioread32(struct card_reader *cr, u32 reg)
{
	return tr_ioread32(cr->tr, reg);
}
EXPORT_SYMBOL_GPL(cr_ioread32);

inline void cr_iowrite32(struct card_reader *cr, u32 reg, u32 value)
{
	tr_iowrite32(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_iowrite32);

inline u16 cr_ioread16(struct card_reader *cr, u32 reg)
{
	return tr_ioread16(cr->tr, reg);
}
EXPORT_SYMBOL_GPL(cr_ioread16);

inline void cr_iowrite16(struct card_reader *cr, u32 reg, u16 value)
{
	tr_iowrite16(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_iowrite16);

inline u8 cr_ioread8(struct card_reader *cr, u32 reg)
{
	return tr_ioread8(cr->tr, reg);
}
EXPORT_SYMBOL_GPL(cr_ioread8);

inline void cr_iowrite8(struct card_reader *cr, u32 reg, u8 value)
{
	tr_iowrite8(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_iowrite8);

inline int cr_read_config8(struct card_reader *cr, int reg,  u8 *value)
{
	return tr_read_config8(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_read_config8);

inline int cr_read_config16(struct card_reader *cr, int reg,  u16 *value)
{
	return tr_read_config16(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_read_config16);

inline int cr_read_config32(struct card_reader *cr, int reg,  u32 *value)
{
	return tr_read_config32(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_read_config32);

inline int cr_write_config8(struct card_reader *cr, int reg, u8 value)
{
	return tr_write_config8(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_write_config8);

inline int cr_write_config16(struct card_reader *cr, int reg, u16 value)
{
	return tr_write_config16(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_write_config16);

inline int cr_write_config32(struct card_reader *cr, int reg, u32 value)
{
	return tr_write_config32(cr->tr, reg, value);
}
EXPORT_SYMBOL_GPL(cr_write_config32);


static void cr_print_flags(struct card_reader *cr)
{
	rts_trace(TRACE_CR, "TURN_ON_LED_AT_START:%d\n",
			!!check_dev_flag(cr, TURN_ON_LED_AT_START));
	rts_trace(TRACE_CR, "RTD3_TEST_EN:%d\n",
			!!check_dev_flag(cr, RTD3_TEST_EN));
	rts_trace(TRACE_CR, "ASPM_L1_1_EN:%d\n",
			!!check_dev_flag(cr, ASPM_L1_1_EN));
	rts_trace(TRACE_CR, "ASPM_L1_2_EN:%d\n",
			!!check_dev_flag(cr, ASPM_L1_2_EN));
	rts_trace(TRACE_CR, "PM_L1_1_EN:%d\n",
			!!check_dev_flag(cr, PM_L1_1_EN));
	rts_trace(TRACE_CR, "PM_L1_2_EN:%d\n",
			!!check_dev_flag(cr, PM_L1_2_EN));
	rts_trace(TRACE_CR, "HANDSHAKE_EN:%d\n",
			!!check_dev_flag(cr, HANDSHAKE_EN));
	rts_trace(TRACE_CR, "LTR_OBFF_SETBACK_EN:%d\n",
			!!check_dev_flag(cr, LTR_OBFF_SETBACK_EN));
	rts_trace(TRACE_CR, "LTR_L1SS_PWR_GATE_EN:%d\n",
			!!check_dev_flag(cr, LTR_L1SS_PWR_GATE_EN));
	rts_trace(TRACE_CR, "FORCE_HOST_ASPM_EN:%d\n",
			!!check_dev_flag(cr, FORCE_HOST_ASPM_EN));
	rts_trace(TRACE_CR, "SD_CD_WP_REVERSE_EN:%d\n",
			!!check_dev_flag(cr, SD_CD_WP_REVERSE_EN));
	rts_trace(TRACE_CR, "POLLING_OCP_EN:%d\n",
			!!check_dev_flag(cr, POLLING_OCP_EN));
	rts_trace(TRACE_CR, "LTR_EXT_SETBACK_EN:%d\n",
			!!check_dev_flag(cr, LTR_EXT_SETBACK_EN));
	rts_trace(TRACE_CR, "L1SS_SETBACK_EN:%d\n",
			!!check_dev_flag(cr, L1SS_SETBACK_EN));
	rts_trace(TRACE_CR, "S3_CLEAR_SETBACK_EN:%d\n",
			!!check_dev_flag(cr, S3_CLEAR_SETBACK_EN));
	rts_trace(TRACE_CR, "L1_SNOOZE_TEST_EN:%d\n",
			!!check_dev_flag(cr, L1_SNOOZE_TEST_EN));
	rts_trace(TRACE_CR, "CLEAR_0xFE8D_BIT0_EN:%d\n",
			!!check_dev_flag(cr, CLEAR_0xFE8D_BIT0_EN));
	rts_trace(TRACE_CR, "LTR_L1SS_PWR_GATE_CHECK_CARD_EN:%d\n",
			!!check_dev_flag(cr, LTR_L1SS_PWR_GATE_CHECK_CARD_EN));
	rts_trace(TRACE_CR, "RESET_CARD_AFTER_RESUME_EN:%d\n",
			!!check_dev_flag(cr, RESET_CARD_AFTER_RESUME_EN));
	rts_trace(TRACE_CR, "CONFIG_N_LSS_DIR:%d\n",
			!!check_dev_flag(cr, CONFIG_N_LSS_DIR));
}

/* cr_print_option - print options
 */
void cr_print_option(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
#define cr_dump_value(x)						\
	rts_trace(TRACE_CR, "%s = 0x%x\n", #x, x)

	cr_dump_value(cr->ic_version);
	cr_dump_value(cr->asic_code);
	cr_dump_value(cr->host_aspm_val);

	cr_dump_value(option->sd40_initial_retry_enable);
	cr_dump_value(option->aux_pwr_exist);
	cr_dump_value(option->force_clkreq_0);
	cr_dump_value(option->dev_aspm_val);
	cr_dump_value(option->dev_aspm_mode);
	cr_dump_value(option->auto_delink_en);
	cr_dump_value(option->ss_en);
	cr_dump_value(option->lun_mode);
	cr_dump_value(option->card_drive_sel);
	cr_dump_value(option->sd30_clk_drive_sel_1v8);
	cr_dump_value(option->sd30_cmd_drive_sel_1v8);
	cr_dump_value(option->sd30_dat_drive_sel_1v8);
	cr_dump_value(option->sd30_clk_drive_sel_3v3);
	cr_dump_value(option->sd30_cmd_drive_sel_3v3);
	cr_dump_value(option->sd30_dat_drive_sel_3v3);
	cr_dump_value(option->asic_sd_hs_clk);
	cr_dump_value(option->asic_mmc_52m_clk);
	cr_dump_value(option->asic_ms_hg_clk);
	cr_dump_value(option->asic_sd_sdr104_clk);
	cr_dump_value(option->asic_sd_sdr50_clk);
	cr_dump_value(option->asic_sd_ddr50_clk);
	cr_dump_value(option->ssc_depth_sd_sdr104);
	cr_dump_value(option->ssc_depth_sd_sdr50);
	cr_dump_value(option->ssc_depth_sd_ddr50);
	cr_dump_value(option->ssc_depth_sd_hs);
	cr_dump_value(option->ssc_depth_mmc_52m);
	cr_dump_value(option->ssc_depth_ms_hg);
	cr_dump_value(option->support_mmc);
	cr_dump_value(option->led_always_on);
	cr_dump_value(option->force_host_aspm_val);
	cr_dump_value(option->reg_0xfe76);
	cr_dump_value(option->ltr_l1off_latency);
	cr_dump_value(option->card_spt_map);
	cr_dump_value(option->sd_capability);

	cr_print_flags(cr);
}
EXPORT_SYMBOL_GPL(cr_print_option);


inline int cr_pci_rd_cfg_byte(struct card_reader *cr, u8 bus, u8 dev, u8 func,
		u8 offset, u8 *val)
{
	return tr_pci_rd_cfg_byte(cr->tr, bus, dev, func, offset, val);
}
EXPORT_SYMBOL_GPL(cr_pci_rd_cfg_byte);

inline int cr_enable_msi(struct card_reader *cr)
{
	return tr_enable_msi(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_enable_msi);

inline void cr_synchronize_irq(struct card_reader *cr)
{
	tr_synchronize_irq(cr->tr);
}
EXPORT_SYMBOL_GPL(cr_synchronize_irq);

inline void cr_clear_interrupt_flag(struct card_reader *cr, u32 flag)
{
	if (cr->option.write_clear_interrupt)
		cr_iowrite32(cr, BIPR, flag);
}
EXPORT_SYMBOL_GPL(cr_clear_interrupt_flag);

inline u8 cr_get_card_spt_map_hw(struct card_reader *cr)
{
	return cr->cc_hw.card_spt_map;
}
EXPORT_SYMBOL_GPL(cr_get_card_spt_map_hw);

inline u32 cr_get_card_lun_map_hw(struct card_reader *cr)
{
	return cr->cc_hw.card_lun_map;
}
EXPORT_SYMBOL_GPL(cr_get_card_lun_map_hw);

/* TODO:
 * Do we need to lock cm when r/w cc_sw ?
 * Consider it in furture.
 */
inline u8 cr_get_card_spt_map_sw(struct card_reader *cr)
{
	return cr->cm->cc_sw.card_spt_map;
}
EXPORT_SYMBOL_GPL(cr_get_card_spt_map_sw);

void cr_set_card_spt_map_sw(struct card_reader *cr, u8 map)
{
	u8 hw = cr_get_card_spt_map_hw(cr);

	cr->cm->cc_sw.card_spt_map = map & hw;
}
EXPORT_SYMBOL_GPL(cr_set_card_spt_map_sw);

inline u32 cr_get_card_lun_map_sw(struct card_reader *cr)
{
	return cr->cm->cc_sw.card_lun_map;
}
EXPORT_SYMBOL_GPL(cr_get_card_lun_map_sw);

inline void cr_set_card_lun_map_sw(struct card_reader *cr, u32 map)
{
	cr->cm->cc_sw.card_lun_map = map;
}
EXPORT_SYMBOL_GPL(cr_set_card_lun_map_sw);

inline u32 cr_get_sd_capability_hw(struct card_reader *cr)
{
	return cr->cc_hw.sd_capability;
}
EXPORT_SYMBOL_GPL(cr_get_sd_capability_hw);

inline u32 cr_get_sd_capability_sw(struct card_reader *cr)
{
	return cr->cm->cc_sw.sd_capability;
}
EXPORT_SYMBOL_GPL(cr_get_sd_capability_sw);

void cr_set_sd_capability_sw(struct card_reader *cr, u32 capability)
{
	u32 hw = cr_get_sd_capability_hw(cr);

	cr->cm->cc_sw.sd_capability = capability & hw;
}
EXPORT_SYMBOL_GPL(cr_set_sd_capability_sw);

void cr_switch_to_sdio(struct card_reader *cr)
{
	if (cr->ops->switch_to_sdio)
		cr->ops->switch_to_sdio(cr);
	else
		rts_trace(TRACE_ERROR, "switch_to_sdio was not implement!\n");
}
EXPORT_SYMBOL_GPL(cr_switch_to_sdio);


void cr_enable_bus_int(struct card_reader *cr, u32 interrupt)
{
	cr_iowrite32(cr, BIER, interrupt);
}
EXPORT_SYMBOL_GPL(cr_enable_bus_int);

void cr_disable_bus_int(struct card_reader *cr)
{
	cr_iowrite32(cr, BIER, 0);
}
EXPORT_SYMBOL_GPL(cr_disable_bus_int);

static void cr_handle_pm_dstate(struct card_reader *cr, u8 dstate)
{
	cr_write_config8(cr, 0x44, dstate);
	cr_write_config8(cr, 0x45, 0);
}

void cr_exit_L1(struct card_reader *cr)
{
	cr_handle_pm_dstate(cr, 0);
}
EXPORT_SYMBOL_GPL(cr_exit_L1);

void cr_enter_L1(struct card_reader *cr)
{
	cr_handle_pm_dstate(cr, 2);
}
EXPORT_SYMBOL_GPL(cr_enter_L1);

inline void cr_disable_sd_dat3_poll(struct card_reader *cr)
{
	cr_wr_reg(cr, SDDAT3_POLL_CFG, SDDAT3_POLL_MASK, 0);
}
EXPORT_SYMBOL_GPL(cr_disable_sd_dat3_poll);

inline void cr_enable_sd_dat3_poll(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_CARD_PWR_CTL, REG_PUPD_CFG | BPP_POWER_MASK,
		BPP_POWER_10_PERCENT_ON);
	cr_wr_reg(cr, REG_PWR_GATE_CTRL, LDO3318_PWR_MASK, LDO_ON);
	cr_wr_reg(cr, SDDAT3_POLL_CFG, SDDAT3_POLL_MASK, SDDAT3_POLL_EN);
}
EXPORT_SYMBOL_GPL(cr_enable_sd_dat3_poll);

/* If multi-luns, it's possible that when plugging/unplugging one card
 * there is another card which still exists in the slot. In this case,
 * all existed cards should be reset.
 */
int cr_process_card_interrupt(struct card_reader *cr)
{
	u32 interrupt = cr->int_reg;

	if (check_dev_flag(cr, RESET_CARD_MANUAL))
		return 0;

	if (interrupt & CARD_INT) {
		if (rtsx_get_stat(cr) == RTSX_STAT_SS)
			cr_set_stat(cr, RTSX_STAT_RUN);
		cr_wakeup_polling_thread(cr);
	}

	if (interrupt & SD_INT) {
		if (interrupt & SD_EXIST) {
			rts_trace(TRACE_INTR, "Insert SD\n");
			set_bit(SD_ID, &(cr->need_reset));
		} else {
			rts_trace(TRACE_INTR, "Remove SD\n");
			set_bit(SD_ID, &(cr->need_release));
			clear_bit(SD_ID, &(cr->need_reset));
		}
	}

	if (interrupt & SD_DAT3_INT) {
		rts_trace(TRACE_INTR, "Insert SD without CD\n");
		set_bit(SD_ID, &(cr->need_reset));
		cr_disable_sd_dat3_poll(cr);
		cr_wr_reg(cr, REG_CARD_PWR_CTL, REG_PUPD_CFG, REG_PUPD_CFG);
	}

	if (interrupt & MS_INT) {
		if (interrupt & MS_EXIST) {
			rts_trace(TRACE_INTR, "Insert MS\n");
			set_bit(MS_ID, &(cr->need_reset));
		} else {
			rts_trace(TRACE_INTR, "Remove MS\n");
			set_bit(MS_ID, &(cr->need_release));
			clear_bit(MS_ID, &(cr->need_reset));
		}
	}

	return 0;
}

int cr_process_transfer_interrupt(struct card_reader *cr)
{
	u32 interrupt = cr->int_reg;
	struct rts_pcie *pcie = cr->tr->extra;

	if (interrupt & (NEED_COMPLETE_INT | DELINK_INT)) {
		if (interrupt & (TRANS_FAIL_INT | DELINK_INT)) {
			if (interrupt & DELINK_INT)
				RTSX_SET_DELINK(cr);
			pcie->trans_result = TRANS_RESULT_FAIL;
			if (pcie->done)
				complete(pcie->done);
		} else if (interrupt & TRANS_OK_INT) {
			pcie->trans_result = TRANS_RESULT_OK;
			if (pcie->done)
				complete(pcie->done);
		} else if (interrupt & DATA_DONE_INT) {
			struct card *card = cm_get_ready_card(cr->cm, SD_ID);
			struct sd_info *sd_info = NULL;
			int pre_rw = 0;

			if (card)
				sd_info = card->extra;
			if ((sd_info) && CHK_SD_UHSII(sd_info)) {
				if (PRE_READ == sd_info->rw_mode
					|| PRE_WRITE == sd_info->rw_mode)
					pre_rw = 1;
			}

			if (pre_rw)
				pcie->trans_result = TRANS_RESULT_OK;
			else
				pcie->trans_result = TRANS_NOT_READY;

			if (pcie->done &&
				(pcie->trans_state == STATE_TRANS_SG))
				complete(pcie->done);
		}
	}

	if (interrupt & TRANS_FAIL_INT)
		rts_trace(TRACE_INTR, "TRANS FAIL. BIPR:%#x\n", interrupt);

	return 0;
}

int cr_process_ocp_interrupt(struct card_reader *cr)
{
	if (cr->option.ocp_en) {
		cr->ocp_int = SD_OC_INT & cr->int_reg;
		if (cr->ocp_int)
			cr_wakeup_polling_thread(cr);
	}

	return 0;
}

int cr_pre_handle_interrupt(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	u32 int_enable;
	u32 int_after_clear;

	if (option->ss_en) {
		cr->ss_counter = 0;
		if (rtsx_get_stat(cr) == RTSX_STAT_SS) {
			cr_exit_L1(cr);
			cr_set_stat(cr, RTSX_STAT_RUN);
		}
	}

	int_enable = cr_ioread32(cr, BIER);
	cr->int_reg = cr_ioread32(cr, BIPR);
	cr_clear_interrupt_flag(cr, cr->int_reg);
	int_after_clear = cr_ioread32(cr, BIPR);
	if (int_after_clear & cr->hw_param.interrupt_en) {
		rts_trace(TRACE_INTR, "before clear cr->int_reg = 0x%x, int_enable = 0x%x\n",
						     cr->int_reg, int_enable);
		rts_trace(TRACE_INTR, "after clear  cr->int_reg = 0x%x, int_enable = 0x%x\n",
						     int_after_clear, int_enable);
		cr_disable_bus_int(cr);
		cr_enable_bus_int(cr, cr->hw_param.interrupt_en);
	}
	if (((cr->int_reg & int_enable) == 0) ||
		(cr->int_reg == 0xFFFFFFFF)) {
		rts_trace(TRACE_INTR, "cr->int_reg = 0x%x, int_enable = 0x%x\n",
						     cr->int_reg, int_enable);
		return STATUS_FAIL;
	}

	cr->int_reg &= (int_enable | 0x7FFFFF);

	return 0;
}

int cr_clear_ocpstat(struct card_reader *cr)
{

	if (cr->ops->clear_ocpstat) {
		return cr->ops->clear_ocpstat(cr);
	} else {
		u8 mask = SD_OCP_INT_CLR | SD_OC_CLR;
		u8 val = SD_OCP_INT_CLR | SD_OC_CLR;
		cr_wr_reg(cr, REG_OCPCTL, mask, val);
		cr_wr_reg(cr, REG_OCPCTL, mask, 0);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(cr_clear_ocpstat);

int cr_enable_ocp(struct card_reader *cr)
{
	u8 val = SD_OCP_INT_EN | SD_DETECT_EN;

	if (cr->ops->enable_ocp)
		return cr->ops->enable_ocp(cr);
	else
		return cr_wr_reg(cr, REG_OCPCTL, 0xFF, val);

}
EXPORT_SYMBOL_GPL(cr_enable_ocp);

int cr_disable_ocp(struct card_reader *cr)
{
	u8 mask = SD_OCP_INT_EN | SD_DETECT_EN;

	if (cr->ops->disable_ocp)
		return cr->ops->disable_ocp(cr);
	else
		return cr_wr_reg(cr, REG_OCPCTL, mask, 0);
}
EXPORT_SYMBOL_GPL(cr_disable_ocp);

int cr_get_ocpstat(struct card_reader *cr, u8 *val)
{
	if (cr->ops->get_ocpstat)
		return cr->ops->get_ocpstat(cr, val);
	else
		return cr_rd_reg(cr, REG_OCPSTAT, val);

}
EXPORT_SYMBOL_GPL(cr_get_ocpstat);

int cr_init_ocp(struct card_reader *cr)
{
	if (cr->ops->init_ocp) {
		return cr->ops->init_ocp(cr);
	} else {
		struct cr_option *option = &(cr->option);
		if (option->ocp_en) {
			u8 val = option->sd_400mA_ocp_thd;

			cr_wr_reg(cr, REG_FPDCTL, OC_POWER_DOWN, 0);
			cr_wr_reg(cr, REG_OCPPARA1, SD_OCP_TIME_MASK,
					SD_OCP_TIME_800);
			cr_wr_reg(cr, REG_OCPPARA2, SD_OCP_THD_MASK, val);
			cr_wr_reg(cr, REG_OCPGLITCH, SD_OCP_GLITCH_MASK,
						cr->hw_param.ocp_glitch);
			cr_enable_ocp(cr);
		} else {
			/* OC power down */
			cr_wr_reg(cr, REG_FPDCTL, OC_POWER_DOWN, OC_POWER_DOWN);
		}
	}

	return 0;
}
EXPORT_SYMBOL_GPL(cr_init_ocp);

/*
 * cr_comm_process_ocp -- Process OCP once it's triggered
 */
void cr_comm_process_ocp(struct card_reader *cr)
{
	struct list_head *pos;

	if (!cr->option.ocp_en)
		return;

	list_for_each(pos, &cr->cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		if (cr->ocp_int) {
			cr_get_ocpstat(cr, &(cr->ocp_stat));
			if (SD_ID == c->id && check_card_flag(c, EXIST))
				sd_power_off_card3v3(cr);
			else if (MS_ID == c->id && check_card_flag(c, EXIST))
				ms_power_off_card3v3(cr);
			else if (XD_ID == c->id && check_card_flag(c, EXIST))
				xd_power_off_card3v3(cr);
			cr->ocp_int = 0;
		}
	}

	if (cr->release_map) {
		if (cr->ocp_stat & (SD_OC_NOW | SD_OC_EVER))
			cr_clear_ocpstat(cr);
		cr->ocp_stat = 0;
	}

	if (cr->ocp_stat & (SD_OC_NOW | SD_OC_EVER)) {
		rts_trace(TRACE_CR, "Over current, OCPSTAT is 0x%x\n",
				cr->ocp_stat);
		list_for_each(pos, &cr->cm->card_list) {
			struct card *c;
			c = list_entry(pos, struct card, link);
			if (SD_ID == c->id && check_card_flag(c, EXIST)) {
				cr_wr_reg(cr, REG_CARD_OE, SD_OUTPUT_EN, 0);
				set_card_flags(c, FAIL);
			} else if (MS_ID == c->id &&
					check_card_flag(c, EXIST)) {
				cr_wr_reg(cr, REG_CARD_OE, MS_OUTPUT_EN, 0);
				set_card_flags(c, FAIL);
			}
		}
		cr_poweroff_card(cr, SD_ID);
	}
}

void cr_process_ocp(struct card_reader *cr)
{
	if (cr->ops->process_ocp)
		cr->ops->process_ocp(cr);
	else
		cr_comm_process_ocp(cr);
}
EXPORT_SYMBOL_GPL(cr_process_ocp);

void cr_init_generic_sw_param(struct cr_option *option)
{
	option->adma_mode = 1;
	option->mspro_formatter_enable = 0;
	option->lun_mode = 0;
	option->auto_delink_en = 0;
	option->ss_en = 0;
	option->ss_idle_period = 50 * 1000;

	option->remote_wakeup_en = 0;
	option->dev_aspm_val = 2;
	option->dev_aspm_mode = DEV_ASPM_DYNAMIC;
	option->config_host_aspm = 1;

	option->fpga_sd_sdr104_clk = CLK_200;
	option->fpga_sd_ddr50_clk = CLK_100;
	option->fpga_sd_sdr50_clk = CLK_100;
	option->fpga_sd_hs_clk = CLK_100;
	option->fpga_mmc_52m_clk = CLK_80;
	option->fpga_ms_hg_clk = CLK_80;
	option->fpga_ms_4bit_clk = CLK_80;
	option->fpga_ms_1bit_clk = CLK_40;
	option->fpga_mmc_sdr200_clk = CLK_200;

	option->asic_sd_sdr104_clk = 203;
	option->asic_sd_sdr50_clk = 98;
	option->asic_sd_ddr50_clk = 98;
	option->asic_sd_hs_clk = 98;
	option->asic_mmc_52m_clk = 98;
	option->asic_ms_hg_clk = 117;
	option->asic_ms_4bit_clk = 78;
	option->asic_ms_1bit_clk = 39;
	option->asic_mmc_sdr200_clk = 195;

	option->ssc_depth_sd_sdr104 = SSC_DEPTH_2M;
	option->ssc_depth_sd_sdr50 = SSC_DEPTH_2M;
	option->ssc_depth_sd_ddr50 = SSC_DEPTH_1M;
	option->ssc_depth_sd_hs = SSC_DEPTH_1M;
	option->ssc_depth_mmc_52m = SSC_DEPTH_1M;
	option->ssc_depth_ms_hg = SSC_DEPTH_1M;
	option->ssc_depth_ms_4bit = SSC_DEPTH_512K;
	option->ssc_depth_low_speed = SSC_DEPTH_512K;
	option->ssc_en = 1;
	option->sd_speed_prior = 0;
	option->sd_current_prior = 0x00010203;

	option->sd_ctl = 0x49;

	option->mmc_ddr_tx_phase = 0;
	option->mmc_sdr200_tx_phase = 0;
	option->phase_min_cnt = 10;

	option->pmos_pwr_on_interval = 4000;
	option->sd_voltage_switch_delay = 1000;
	option->ms_power_class_en = 3;

	option->sd_400mA_ocp_thd = 1;	/* 550mA */
	option->sd_800mA_ocp_thd = 5;	/* 950mA */

	option->card_drive_sel = 0x55;
	option->sd30_drive_sel_1v8 = 3;
	option->sd30_drive_sel_3v3 = 3;

	option->do_delink_before_power_down = 0;
	option->auto_power_down = 1;
	option->polling_config = 0;

	option->force_clkreq_0 = 1;
	option->ft2_fast_mode = 0;

	option->sd_timeout = 10000;
	option->sd40_timeout = 10000;
	option->xd_timeout = 2000;
	option->ms_timeout = 2000;
	option->mspro_timeout = 15000;

	option->power_down_in_ss = 1;

	/* unit: ms */
	option->delink_stage1_step = 700 * 32;
	option->delink_stage2_step = 40 * 32;
	option->delink_stage3_step = 700 * 32;

	option->auto_delink_in_L1 = 0;
	option->blink_led = 1;
	option->msi_en = 1;
	option->hp_watch_bios_hotplug = 0;

	option->support_ms_8bit = 0;
	option->s3_pwr_off_delay = 0;

	option->write_clear_interrupt = 1;
	option->ocp_en = 1;
	option->rtd3_en = 0;

	memset(&option->dev_flags, 0, sizeof(option->dev_flags));

	option->sd_flags = 0;
	option->ms_flags = 0;
	option->polling_interval_run = 32;
	option->polling_interval_idle = 200;
	option->polling_interval_ss = 250;
	option->cd_debounce_time = 500;
	option->l1_snooze_delay = 5; /* ms */
	option->card_spt_map = (1 << MS_ID) | (1 << SD_ID) | (1 << XD_ID);
	option->sd_capability = SD_CAP_MMC | SD_CAP_SD;
	option->n_lss_dir = 0x00;
	option->patch_flags = 0x00;
	option->delay_for_rw_reg = 0;
	option->sd40_nfcu = 1;
	option->cprm_cmd_timeout = 60000;
	option->sd40_cache_rw_en = 0;
	option->ack_every_tlp = 1;
	option->thermal_sensor_en = 0;
	option->thermal_direct_to_step3 = 1;
	option->thermal_time_step1 = 1800000;
	option->thermal_time_step2 = 3600000;
	option->thermal_time_step3 = 7200000;
	option->thermal_time_recovery = 1800000;
	option->ignore_sd = 0;
	option->led_always_on = 0;
	option->force_host_aspm_val = 2;
	option->sd_ddr50_default_rx_phase = 5;
	option->sd_ddr50_default_tx_phase = 16;
	option->sd_sdr104_default_rx_phase = 24;
	option->sd_sdr104_default_tx_phase = 27;
	option->sd_sdr50_default_rx_phase = 6;
	option->sd_sdr50_default_tx_phase = 8;
	option->sdio_retry_cnt = 1;
	option->ww_enable = 0;
	option->use_hw_setting = 1;
	option->aux_clk_freq = 0;
	option->relink_time = 0x8FFFF;
	option->ltr_active_latency = 0x883C;
	option->ltr_idle_latency = 0x892C;
	option->ms_ocp_thd = 2;
	option->ltr_en = 1;
	option->reg_0xfe76 = 0xF7;
	option->sd40_ctl = 0x91821000;
	option->polling_oobs_en = 0;
	option->pcie_oobs_max_value = 8;
	option->sd40_hiber_pwr_off_delay = 30;
	option->idle_max_cnt = IDLE_MAX_COUNT;
	option->sdvcc_pd = 3;
	option->eios_patch_en = 0;
}
EXPORT_SYMBOL_GPL(cr_init_generic_sw_param);

void cr_init_generic_hw_param(struct hw_param *hw_param)
{
	hw_param->phy0x1e_val = 0x78EB;
	hw_param->phy0x1e_val_l1ss_off = 0x58EB;
	hw_param->phy0x08_val = 0x4FE4;
	hw_param->phy0x82_val = 0x2508;
	hw_param->phy0x85_val = 0x2497;
	hw_param->phy0x87_val = 0x7C69;
	hw_param->phy0x88_val = 0x4820;
	hw_param->phy0x90_val = 0x09C2;
	hw_param->phy0x92_val = 0xE21D;
	hw_param->av12s = 0x04;
	hw_param->stb_lh_time = 0xFF;
	hw_param->eidle_num = 0x07;
}
EXPORT_SYMBOL_GPL(cr_init_generic_hw_param);

void cr_init_generic_options(struct card_reader *cr)
{
	struct rts_pcie *pcie = cr->tr->extra;

	cr->ssvid = pcie->pci_dev->subsystem_vendor;
	cr->ssdid = pcie->pci_dev->subsystem_device;
	cr->lun_mc = 0;
	cr->driver_first_load = 1;
	cr->d3cold_supported = 0;

	cr_init_generic_sw_param(&cr->option);
	cr_init_generic_hw_param(&cr->hw_param);
}
EXPORT_SYMBOL_GPL(cr_init_generic_options);


static void cr_delink_stage1(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	int enter_L1 = option->auto_delink_in_L1
			&& (option->dev_aspm_val || option->ss_en);
	int step1 = option->delink_stage1_step;
	int step2 = step1 + option->delink_stage2_step;
	int step3 = step2 + option->delink_stage3_step;

	if (cr->pm_ops->do_before_auto_delink)
		cr->pm_ops->do_before_auto_delink(cr);

	cr_set_stat(cr, RTSX_STAT_DELINK);
	cr_clear_first_install_mark(cr);

	if (cm_check_cards_flags(cr->cm, EXIST)) {
		/*
		* False card
		*/
		rts_trace(TRACE_CR, "False card inserted,do force delink\n");
		if (enter_L1)
			cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03, 1);
		cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x0A, 0x0A);
		/* cr_wr_reg(cr, AUTOLOAD_CFG3, 0x40, 0x00); */
		if (enter_L1)
			cr_enter_L1(cr);
		cr->auto_delink_timer -= (step3 + 1);
	} else {
		rts_trace(TRACE_CR, "No card inserted, do delink\n");
		if (enter_L1)
			cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03, 1);

		cr_clear_interrupt_flag(cr, cr->int_reg);
		rts_trace(TRACE_CR, "BIPR: 0x%x\n", cr_ioread32(cr, BIPR));
		cr_disable_bus_int(cr);
		rts_trace(TRACE_CR, "BIER: 0x%x\n", cr_ioread32(cr, BIER));

		cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x02, 0x02);
		/* cr_wr_reg(cr, AUTOLOAD_CFG3, 0x40, 0x00); */

		if (enter_L1)
			cr_enter_L1(cr);
	}
}

static void cr_delink_stage2(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	int enter_L1 = option->auto_delink_in_L1 &&
				(option->dev_aspm_val || option->ss_en);

	rts_trace(TRACE_CR, "Try to do force delink\n");

	/*
	* wwang, 2010-12-03 If chip in D2 state ,we should exit L1
	* before we try to write register
	*/
	if (enter_L1)
		cr_exit_L1(cr);
	cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x0A, 0x0A);
}

static void cr_delink_stage3(struct card_reader *cr)
{

}

/* cr_auto_delink - Require to enable auto-delink function
 * @cr: card reader structure
 *
 * Returns: Nothing
 */

void cr_auto_delink(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	int step1 = option->delink_stage1_step;
	int step2 = step1 + option->delink_stage2_step;
	int step3 = step2 + option->delink_stage3_step;
	u64 elapse = cr_get_tick() - cr->auto_delink_timer;

	rts_trace(TRACE_CR, "elapse time before auto delink: %lldms\n", elapse);

	if (elapse >= step1 && !CHECK_STATUS_FLAG(cr, DELINK_STEP1_DONE)) {
		cr_delink_stage1(cr);
		SET_STATUS_FLAG(cr, DELINK_STEP1_DONE);
	} else if (elapse >= step2
			&& !CHECK_STATUS_FLAG(cr, DELINK_STEP2_DONE)) {
		cr_delink_stage2(cr);
		SET_STATUS_FLAG(cr, DELINK_STEP2_DONE);
	} else if (elapse >= step3
			&& !CHECK_STATUS_FLAG(cr, DELINK_STEP3_DONE)) {
		cr_delink_stage3(cr);
		SET_STATUS_FLAG(cr, DELINK_STEP3_DONE);
	}
}

void cr_undo_delink(struct card_reader *cr)
{
	cr->option.auto_delink_allowed = 0;
	cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x0A, 0x00);
}

static u8 lun_mode[4] = {
	0xFF,
	MS_LUN,
	SD_LUN,
	DEFAULT_SINGLE,
};

static u8 ssc_depth1[4] = {
	SSC_DEPTH_512K,
	SSC_DEPTH_1M,
	SSC_DEPTH_2M,
	SSC_DEPTH_4M,
};

static u8 ssc_depth2[4] = {
	SSC_DEPTH_256K,
	SSC_DEPTH_512K,
	SSC_DEPTH_1M,
	SSC_DEPTH_2M,
};

int cr_init_generic_options_from_hw(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	u32 lval;
	u8 val;
	u8 cap = 0;
	u32 sd_capability = cr_get_sd_capability_sw(cr);
	u8 card_spt_map = cr_get_card_spt_map_sw(cr);

	if (cr_rd_cfg32(cr, 0x724, &lval))
		TRACE_RET(cr, STATUS_FAIL);

	if (lval & (1 << 24))
		return 0;

	/* 0x727 */
	val = (u8)(lval >> 24);

	option->lun_mode = lun_mode[(val >> 6) & 0x03];
	cap = (val >> 6) & 0x03;
	option->dev_aspm_val = (val >> 4) & 0x03;
	option->force_host_aspm_val = (val >> 4) & 0x03;
	if (check_dev_flag(cr, FORCE_HOST_ASPM_EN))
		cr->host_aspm_val = option->force_host_aspm_val;
	option->card_drive_sel &= 0x3F;
	option->card_drive_sel |= ((val >> 1) & 0x01) << 6;

	/* 0x726 */
	val = (u8)(lval >> 16);
	if ((val & 0xC0) != 0xC0) {
		option->asic_sd_hs_clk = (49 - ((val >> 6) & 0x03) * 2) * 2;
		option->asic_mmc_52m_clk = option->asic_sd_hs_clk;
	}

	/* WARNING: keep option->sd_capability  constant
	 * when merge with branch rts5250
	 */
	if (val & (1 << 5))
		sd_capability |= SD_CAP_SDR50;
	else
		sd_capability &= ~SD_CAP_SDR50;
	if (val & (1 << 4))
		sd_capability |= SD_CAP_DDR50;
	else
		sd_capability &= ~SD_CAP_DDR50;
	if (val & (1 << 3))
		sd_capability |= SD_CAP_SDR104;
	else
		sd_capability &= ~SD_CAP_SDR104;
	if (sd_capability & SD_CAP_SD)
		card_spt_map |= 1 << SD_ID;
	else if (0 == sd_capability)
		card_spt_map &= ~(1 << SD_ID);


	if ((val & 0x07) != 0x07)
		option->asic_ms_hg_clk = (59 - (val & 0x07)) * 2;

	/* 0x725 */
	val = (u8)(lval >> 8);
	if ((val & 0xE0) != 0xE0)
		option->asic_sd_sdr104_clk = 206 - ((val >> 5) & 0x07) * 3;
	if ((val & 0x1C) != 0x1C)
		option->asic_sd_sdr50_clk = 98 - ((val >> 2) & 0x07) * 2;
	if ((val & 0x03) != 0x03)
		option->asic_sd_ddr50_clk = (48 - (val & 0x03) * 2) * 2;

	/* 0x724 */
	val = (u8)lval;

	option->ssc_depth_sd_sdr104 = ssc_depth1[(val >> 6) & 0x03];
	option->ssc_depth_sd_sdr50 = option->ssc_depth_sd_sdr104;
	option->ssc_depth_sd_ddr50 = ssc_depth1[(val >> 4) & 0x03];
	option->ssc_depth_sd_hs = ssc_depth2[(val >> 2) & 0x03];
	option->ssc_depth_mmc_52m = option->ssc_depth_sd_hs;
	option->ssc_depth_ms_hg = ssc_depth2[val & 0x03];

	/* 0x814 */
	if (cr_rd_cfg32(cr, 0x814, &lval))
		TRACE_RET(cr, STATUS_FAIL);

	val = (u8)lval;
	option->auto_delink_en = (val & 0x80) ? 1 : 0;
	if (val	& 0x10) {
		sd_capability &= ~SD_CAP_MMC;
		if (0 == sd_capability)
			card_spt_map &= ~(1 << SD_ID);
	} else {
		card_spt_map |= 1 << SD_ID;
		sd_capability |= SD_CAP_MMC;
	}

	if (0x03 == cap)
		card_spt_map = (1 << SD_ID) | (1 << MS_ID);
	else if (0x02 == cap)
		card_spt_map = 1 << SD_ID;
	else if (0x01 == cap)
		card_spt_map = 1 << MS_ID;
	else
		card_spt_map = 0;

	cr_set_card_spt_map_sw(cr, card_spt_map);
	cr_set_sd_capability_sw(cr, sd_capability);

	option->led_always_on = !!(val & 0x08);

	return 0;
}
EXPORT_SYMBOL_GPL(cr_init_generic_options_from_hw);

static void cr_deal_extra_action(struct card_reader *cr)
{
	if (cr->ops->deal_extra_action)
		cr->ops->deal_extra_action(cr);
}

void cr_deal_generic_extra_action(struct card_reader *cr)
{
	u8 val;

	if (!rtsx_chk_stat(cr, RTSX_STAT_IDLE)
		&& !rtsx_chk_stat(cr, RTSX_STAT_SS))
		return;
	if (cm_check_cards_flags(cr->cm, READY)) {
		if (check_dev_flag(cr, POLLING_OCP_EN)) {
			cr_rd_reg(cr, REG_OCPSTAT, &val);

			/* some chip cannot issue OCP interrupt */
			if (val & (SD_OC_NOW | SD_OC_EVER
					| SDVIO_OC_NOW | SDVIO_OC_EVER))
				cr->ocp_stat = val;

			msleep(1);
		}
	}

}
EXPORT_SYMBOL_GPL(cr_deal_generic_extra_action);

/* cr_save_reg_to_autoload - save register to autoload space
 *
 * @cr: pointer of cardreader
 * @reg: register addr
 * @mask: wirte mask
 * @data: value writed to reg
 * @autoload: autoload addr
 */
int cr_save_reg_to_autoload(struct card_reader *cr,
		u16 reg, u8 mask, u8 data, u16 autoload)
{
	u8 val = (u8)((reg >> 8) & 0x3F);

	/* SAVE REGISTER */
	val |= 0xC0;

	cr_wr_reg(cr, autoload + 0, 0xFF, data);
	cr_wr_reg(cr, autoload + 1, 0xFF, mask);
	cr_wr_reg(cr, autoload + 2, 0xFF, (u8)(reg & 0xFF));
	cr_wr_reg(cr, autoload + 3, 0xFF, val);

	return 0;
}
EXPORT_SYMBOL_GPL(cr_save_reg_to_autoload);

int cr_save_cfg_to_epcore(struct card_reader *cr, u16 reg, u16 cfg, u8 func)
{
	u8 val, mode;
	u32 lval;

	cr_rd_cfg32(cr, (cfg & 0xFFFC), &lval);
	val = (u8) (lval >> ((cfg & 0x03) * 8));
	cr_wr_reg(cr, reg, 0xFF, val);

	mode = (u8) (1 << (cfg & 0x03));

	/* the value of CFG ADDR saved in autoload space must be
	 * aligned by 4.
	 */
	cfg &= 0xFFFC;
	val = (mode & 0x0F) | ((func & 0x03) << 4) | (((u8)cfg & 0x03) << 6);
	cr_wr_reg(cr, reg + 1, 0xFF, val);
	cr_wr_reg(cr, reg + 2, 0xFF, (u8)(cfg >> 2));
	cr_wr_reg(cr, reg + 3, 0xFF, ((u8)(cfg >> 10) & 0x3F) | 0x80);

	return 0;
}
EXPORT_SYMBOL_GPL(cr_save_cfg_to_epcore);

/* cr_read_host_sd40_iospace - read data from host sd40 iospace
 *
 * @cr
 * @addr: it's BYTE ADDRESS mode, not IOADDR mode.
 * @data
 * @len
 */
int cr_read_host_sd40_iospace(struct card_reader *cr,
				u16 addr, u8 *data, int len)
{
	if (cr->ops->read_host_sd40_iospace)
		return cr->ops->read_host_sd40_iospace(cr, addr, data, len);
	else
		return cr_comm_read_host_sd40_iospace(cr, addr, data, len);
}
EXPORT_SYMBOL_GPL(cr_read_host_sd40_iospace);

/* cr_write_host_sd40_iospace - write data to sd40 iospace
 *
 * @cr
 * @addr: it's BYTE ADDRESS mode, not IOADDR mode.
 * @data
 * @len
 */
int cr_write_host_sd40_iospace(struct card_reader *cr,
				u16 addr, u8 *data, int len)
{
	if (cr->ops->write_host_sd40_iospace)
		return cr->ops->write_host_sd40_iospace(cr, addr, data, len);
	else
		return cr_comm_write_host_sd40_iospace(cr, addr, data, len);

}
EXPORT_SYMBOL_GPL(cr_write_host_sd40_iospace);

/* cr_check_card_in_init - check card if exist in init procedure
 *
 * @cr:
 *
 * note : It can only be called in initilization procedure
 */
int cr_check_card_in_init(struct card_reader *cr)
{
	struct list_head *pos;

	if (check_dev_flag(cr, RESET_CARD_MANUAL))
		return 0;

	cr->need_reset = 0;

	cr->int_reg = cr_ioread32(cr, BIPR);

	if (cr->int_reg & SD_EXIST)
		set_bit(SD_ID, &(cr->need_reset));
	if (cr->int_reg & MS_EXIST)
		set_bit(MS_ID, &(cr->need_reset));

	if (cr->int_reg & (MS_EXIST | SD_EXIST))
		cr_wr_reg(cr, REG_SSC_CTL1, SSC_RSTB, SSC_RSTB);

	rts_trace(TRACE_PROBE, "need_reset:%#lx\n", cr->need_reset);

	if (CHECK_STATUS_FLAG(cr, ENTERED_RTD3)) {
		list_for_each(pos, &cr->cm->card_list) {
			struct card *c;
			c = list_entry(pos, struct card, link);
			if (!((1 << c->id) & (cr->cm->cc_sw.card_spt_map)))
				continue;
			rts_trace(TRACE_PROBE, "card %d flag is %#x\n", c->id, c->card_flags);
			if (!test_bit(c->id, &(cr->need_reset))) {
				if (check_card_flag(c, EXIST)) {
					set_bit(c->id, &(cr->need_release));
					msleep(200);
				}
			}
		}
		cr->int_reg = cr_ioread32(cr, BIPR);
	}
	cr_clear_interrupt_flag(cr, cr->int_reg);

	return 0;
}
EXPORT_SYMBOL_GPL(cr_check_card_in_init);

/* cr_find_pci_capability
 *
 * @cr:
 * @cap:
 *
 */
int cr_find_pci_capability(struct card_reader *cr, int cap)
{
	return tr_find_pci_capability(cr->tr, cap);
}
EXPORT_SYMBOL_GPL(cr_find_pci_capability);

int cr_find_host_pci_capability(struct card_reader *cr, int cap)
{
	return tr_find_host_pci_capability(cr->tr, cap);
}
EXPORT_SYMBOL_GPL(cr_find_host_pci_capability);

int cr_swap_sdclk_polarity(struct card_reader *cr)
{
	u8 val = 0;
	u8 mask = SD_CLK_POLARITY_SWAP;

	cr_rd_reg(cr, REG_SD_PUSH_POINT_CTL, &val);
	val &= mask;
	val ^= mask;
	return cr_wr_reg(cr, REG_SD_PUSH_POINT_CTL, mask, val);
}
EXPORT_SYMBOL_GPL(cr_swap_sdclk_polarity);

inline void cr_read_host_config8(struct transport *tr, u16 addr, u8 *val)
{
	tr_read_host_config8(tr, addr, val);
}
EXPORT_SYMBOL_GPL(cr_read_host_config8);

inline void cr_write_host_config8(struct transport *tr, u16 addr, u8 val)
{
	tr_write_host_config8(tr, addr,  val);
}
EXPORT_SYMBOL_GPL(cr_write_host_config8);

int cr_comm_enable_oobs_polling(struct card_reader *cr)
{
	u16 val = 0;
	rts_trace(TRACE_CR, "Enable Polling OOBS \n");
	cr_rd_phy16(cr, 0x01, &val);
	val |= 1<<9;
	cr_wr_phy16(cr, 0x01, val);
	cr_wr_reg(cr, REG_CFG_OOBS_OFF_TIMER, 0xFF, 0x32);
	cr_wr_reg(cr, REG_CFG_OOBS_ON_TIMER, 0xFF, 0x05);
	cr_wr_reg(cr, REG_CFG_VCM_ON_TIMER, 0xFF, 0x83);
	cr_wr_reg(cr, REG_CFG_OOBS_POLLING, 0xFF, 0xDE);
	return 0;
}

int cr_enable_oobs_polling(struct card_reader *cr)
{
	if (cr->ops->enable_oobs_polling)
		return cr->ops->enable_oobs_polling(cr);
	else
		return cr_comm_enable_oobs_polling(cr);
}
EXPORT_SYMBOL_GPL(cr_enable_oobs_polling);

int cr_comm_disable_oobs_polling(struct card_reader *cr)
{
	u16 val = 0;
	rts_trace(TRACE_CR, "Disable Polling OOBS \n");
	cr_rd_phy16(cr, 0x01, &val);
	val &= ~(1<<9);
	cr_wr_phy16(cr, 0x01, val);
	cr_wr_reg(cr, REG_CFG_VCM_ON_TIMER, 0xFF, 0x03);
	cr_wr_reg(cr, REG_CFG_OOBS_POLLING, 0xFF, 0x00);
	return 0;
}

int cr_disable_oobs_polling(struct card_reader *cr)
{
	if (cr->ops->disable_oobs_polling)
		return cr->ops->disable_oobs_polling(cr);
	else
		return cr_comm_disable_oobs_polling(cr);
}
EXPORT_SYMBOL_GPL(cr_disable_oobs_polling);

int cr_patch_dma(struct card_reader *cr)
{
	int ret = 0;
	if(cr->ops->patch_dma) {
		ret = cr->ops->patch_dma(cr);
	} else {
		printk("%s: patch_dma callback is not implemented\n", cr->name);
		ret = -EINVAL;
	}
	return ret;
}
EXPORT_SYMBOL_GPL(cr_patch_dma);

void cr_get_host_devctl(struct card_reader *cr, u8 *val)
{
	cr_read_host_config8(cr->tr,
		(u16)(cr->host_pci_cap_offset + PCI_EXP_DEVCTL), val);
}
EXPORT_SYMBOL_GPL(cr_get_host_devctl);

void cr_get_dev_devctl(struct card_reader *cr, u8 *val)
{
	cr_read_config8(cr, cr->dev_pci_cap_offset + PCI_EXP_DEVCTL, val);
}
EXPORT_SYMBOL_GPL(cr_get_dev_devctl);

void cr_set_dev_devctl(struct card_reader *cr, u8 val)
{
	cr_write_config8(cr, cr->dev_pci_cap_offset + PCI_EXP_DEVCTL, val);
}
EXPORT_SYMBOL_GPL(cr_set_dev_devctl);

