/* Driver for Realtek card reader
 * Header file
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/kernel.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include <linux/module.h>
#include <linux/vmalloc.h>

#include "ms.h"
#include "cardreader.h"
#include "trace.h"
#include "rts_cr.h"
#include "rts_scsi.h"
#include "scsi_cmd.h"
#include "cr_regs.h"


int ms_send_cmd(struct card_reader *cr, struct card *card, u8 cmd, u8 cfg)
{
	u8 data[2];

	data[0] = cmd;
	data[1] = 0;

	return ms_write_bytes(cr, card, PRO_SET_CMD, 1, cfg, data, 1);
}
EXPORT_SYMBOL_GPL(ms_send_cmd);

static int ms_comm_prepare_init(struct card_reader *cr, struct card *c)
{
	if (cr_enable_card_clk(cr, MS_ID))
		TRACE_RET(cr, STATUS_FAIL);
	if (cr_select_card(cr, c))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

/* ms_prepare_init - Prepare for the kickoff of ms protocol
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 *
 * Description: currently , work in this function, revolves in enabling clock
 * and selecting card.
 */
int ms_prepare_init(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->prepare_init)
		return cr->ms_ops_prot->prepare_init(cr, c);
	else
		return ms_comm_prepare_init(cr, c);
}
EXPORT_SYMBOL_GPL(ms_prepare_init);

static int ms_comm_pull_ctl_disable(struct card_reader *cr)
{
	/* This function would be called in enter_ss
	 * which is mutual exclusive to interrupt handler.
	 * So we can't use command buffer to access internal register */
	cr_wr_reg(cr, REG_CARD_PULL_CTL5, 0xFF, 0x55);
	cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0xFF, 0x15);

	return STATUS_SUCCESS;
}

int ms_pull_ctl_disable(struct card_reader *cr)
{
	if (cr->ms_ops_prot->pull_ctl_disable)
		return cr->ms_ops_prot->pull_ctl_disable(cr);
	else
		return ms_comm_pull_ctl_disable(cr);
}
EXPORT_SYMBOL_GPL(ms_pull_ctl_disable);

static int ms_comm_pull_ctl_enable(struct card_reader *cr)
{
	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL5, 0xFF, 0x55);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0xFF, 0x15);
	if (cr_send_cmd(cr, MS_CARD, 100))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int ms_pull_ctl_enable(struct card_reader *cr)
{
	if (cr->ms_ops_prot->pull_ctl_enable)
		return cr->ms_ops_prot->pull_ctl_enable(cr);
	else
		return ms_comm_pull_ctl_enable(cr);
}
EXPORT_SYMBOL_GPL(ms_pull_ctl_enable);

/*
 * ms_power_off_card3v3 shoud be placed in cardreader.c
 */
int ms_power_off_card3v3(struct card_reader *cr)
{
	if (cr_disable_card_clk(cr, MS_ID))
		TRACE_RET(cr, STATUS_FAIL);

	if (cr->asic_code) {
		if (ms_pull_ctl_disable(cr))
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		cr_wr_reg(cr, REG_FPGA_PULL_CTL, FPGA_MS_PULL_CTL_BIT | 0x20,
							FPGA_MS_PULL_CTL_BIT);
	}

	cr_wr_reg(cr, REG_CARD_OE, MS_OUTPUT_EN, 0);
	cr_poweroff_card(cr, MS_ID);

	return 0;
}
EXPORT_SYMBOL_GPL(ms_power_off_card3v3);

static int ms_comm_prepare_reset(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	ms_card->ms_type = 0;
	ms_card->check_ms_flow = 0;
	ms_card->switch_8bit_fail = 0;
	ms_card->delay_write.delay_write_flag = 0;
	ms_card->pro_under_formatting = 0;

	if (ms_power_off_card3v3(cr))
		TRACE_RET(cr, STATUS_FAIL);
	msleep(250);

	if (cr_enable_card_clk(cr, MS_ID))
		TRACE_RET(cr, STATUS_FAIL);

	if (cr->asic_code) {
		if (ms_pull_ctl_enable(cr))
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		cr_wr_reg(cr, REG_FPGA_PULL_CTL,
				FPGA_MS_PULL_CTL_BIT | 0x20, 0);
	}

	if (cr_poweron_card(cr, MS_ID))
		TRACE_RET(cr, STATUS_FAIL);
	msleep(150);

	if (cr->option.ocp_en) {
		u8 ocp;
		cr_get_ocpstat(cr, &ocp);
		if (ocp) {
			struct card *card = cm_get_card(cr->cm, MS_ID);

			if (card)
				card->reset_flags |= RESET_MEET_OCP;
			rts_trace(TRACE_SD, "OCP! OCPSTAT:%#x\n", ocp);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	cr_wr_reg(cr, REG_CARD_OE, MS_OUTPUT_EN, MS_OUTPUT_EN);

	/* Reset Registers */
	if (cr->asic_code) {
		cr_wr_reg(cr, REG_MS_CFG, 0xFF, SAMPLE_TIME_RISING
			| PUSH_TIME_DEFAULT | NO_EXTEND_TOGGLE
			| MS_BUS_WIDTH_1);
	} else {
		cr_wr_reg(cr, REG_MS_CFG, 0xFF, SAMPLE_TIME_FALLING
			| PUSH_TIME_DEFAULT | NO_EXTEND_TOGGLE
			| MS_BUS_WIDTH_1);
	}

	cr_wr_reg(cr, REG_MS_TRANS_CFG, 0xFF, NO_WAIT_INT |
						NO_AUTO_READ_INT_REG);
	cr_wr_reg(cr, REG_CARD_STOP, MS_STOP | MS_CLR_ERR,
						MS_STOP | MS_CLR_ERR);
	if (ms_set_init_para(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

/*
 * ms_prepare_reset -- Prepare for card reset operation
 */
int ms_prepare_reset(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->prepare_reset)
		return cr->ms_ops_prot->prepare_reset(cr, c);
	else
		return ms_comm_prepare_reset(cr, c);
}
EXPORT_SYMBOL_GPL(ms_prepare_reset);

static int ms_comm_switch_bus_width_4(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_MS_CFG, 0x18, MS_BUS_WIDTH_4);
	return 0;
}

/*
 * ms_switch_bus_width_4 -- Switch bus width to 4 bits
 */
int ms_switch_bus_width_4(struct card_reader *cr)
{
	if (cr->ms_ops_prot->switch_bus_width_4)
		return cr->ms_ops_prot->switch_bus_width_4(cr);
	else
		return ms_comm_switch_bus_width_4(cr);
}
EXPORT_SYMBOL_GPL(ms_switch_bus_width_4);

static int ms_comm_set_push_timming_odd(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_MS_CFG, PUSH_TIME_ODD, PUSH_TIME_ODD);

	return 0;
}

/*
 * ms_set_push_timming_odd -- Set MS-PRO Card timing 40MHz, and set Push Timing
 * as "ODD"
 */
int ms_set_push_timming_odd(struct card_reader *cr)
{
	if (cr->ms_ops_prot->set_push_timming_odd)
		return cr->ms_ops_prot->set_push_timming_odd(cr);
	else
		return ms_comm_set_push_timming_odd(cr);
}

static int ms_send_cmd_4_write_bytes(struct card_reader *cr,
		u8 tpc, u8 cnt, u8 cfg, u8 *data, int data_len)
{
	int i = 0;

	cr_init_cmd(cr);

	for (i = 0; i < cnt; i++)
		cr_add_cmd(cr, WRITE_REG_CMD, PPBUF_BASE2 + i, 0xFF, data[i]);
	if (cnt % 2)
		cr_add_cmd(cr, WRITE_REG_CMD, PPBUF_BASE2 + i, 0xFF, 0xFF);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, tpc);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_BYTE_CNT, 0xFF, cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, cfg);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
						0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER, 0xFF,
			REG_MS_TRANSFER_START | MS_TM_WRITE_BYTES);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
			REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);

	return cr_send_cmd(cr, MS_CARD, 5000);

}

static int ms_comm_write_bytes(struct card_reader *cr,
	struct card *card, u8 tpc, u8 cnt, u8 cfg, u8 *data, int data_len)
{
	struct ms_info *ms_card = card->extra;

	if (!data || (data_len < cnt))
		TRACE_RET(cr, STATUS_ERROR);

	if (ms_send_cmd_4_write_bytes(cr, tpc, cnt, cfg, data, data_len)) {
		u8 val = 0;

		cr_rd_reg(cr, REG_MS_TRANS_CFG, &val);
		rts_trace(TRACE_MS, "REG_MS_TRANS_CFG: 0x%02x\n", val);
		ms_clear_error(cr);
		if (!(tpc & 0x08)) {
			if (val & MS_CRC16_ERR) {
				ms_set_error_code(cr, card, MS_CRC16_ERROR);
				TRACE_RET(cr, ms_parse_error_code(cr));
			}
		} else {
			if (CHK_MSPRO(ms_card) && !(val & 0x80)) {
				if (val & (MS_INT_ERR | MS_INT_CMDNK)) {
					ms_set_error_code(cr, card, MS_CMD_NK);
					TRACE_RET(cr, ms_parse_error_code(cr));
				}
			}
		}

		if (val & MS_RDY_TIMEOUT) {
			ms_set_error_code(cr, card, MS_TO_ERROR);
			TRACE_RET(cr, ms_parse_error_code(cr));
		}
		ms_set_error_code(cr, card, MS_TO_ERROR);
		TRACE_RET(cr, ms_parse_error_code(cr));
	}

	return 0;
}

/*
 * ms_write_bytes --  Write bytes by submiting TPC to MS card
 * @cr: points to card reader structure
 * @c: points to card struct
 * @tpc: TPC(Transfer Protocol Command)
 * @cnt: data count to be written
 * @__cfg: Logical MS transaction configuration, which will be converted into
 * cr-dependent value in implement
 * @data: points to the data tobe written
 * @data_len: Length of expected data to be written, which is supposed to
 * larger than @cnt
 */
int ms_write_bytes(struct card_reader *cr, struct card *c, u8 tpc,
		u8 cnt, u8 __cfg, u8 *data, int data_len)
{
	if (cr->ms_ops_prot->write_bytes) {
		return cr->ms_ops_prot->write_bytes(cr, c, tpc, cnt,
						__cfg, data, data_len);
	} else {
		return ms_comm_write_bytes(cr, c, tpc, cnt,
						__cfg, data, data_len);
	}
}
EXPORT_SYMBOL_GPL(ms_write_bytes);

static int ms_send_cmd_4_read_bytes(struct card_reader *cr,
		u8 tpc, u8 cnt, u8 cfg, u8 *data, int data_len)
{
	int i = 0;

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, tpc);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_BYTE_CNT, 0xFF, cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, cfg);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
						0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER, 0xFF,
			REG_MS_TRANSFER_START | MS_TM_READ_BYTES);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
			REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);

	for (i = 0; i < data_len - 1; i++)
	       cr_add_cmd(cr, READ_REG_CMD, PPBUF_BASE2 + i, 0, 0);

	if (data_len % 2)
		cr_add_cmd(cr, READ_REG_CMD, PPBUF_BASE2 + data_len, 0, 0);
	else
		cr_add_cmd(cr, READ_REG_CMD, PPBUF_BASE2 + data_len - 1, 0, 0);

	return cr_send_cmd(cr, MS_CARD, 5000);
}

static int ms_comm_read_bytes(struct card_reader *cr, struct card *card,
				u8 tpc, u8 cnt, u8 cfg, u8 *data, int data_len)
{
	struct ms_info *ms_card = card->extra;
	u8 *ptr;

	if (!data)
		TRACE_RET(cr, STATUS_ERROR);
	if (ms_send_cmd_4_read_bytes(cr, tpc, cnt, cfg, data, data_len)) {
		u8 val = 0;

		cr_rd_reg(cr, REG_MS_TRANS_CFG, &val);
		ms_clear_error(cr);
		if (!(tpc & 0x08)) {
			if (val & MS_CRC16_ERR) {
				ms_set_error_code(cr, card, MS_CRC16_ERROR);
				TRACE_RET(cr, ms_parse_error_code(cr));
			}
		} else {
			if (CHK_MSPRO(ms_card) && !(val & 0x80)) {
				if (val & (MS_INT_ERR | MS_INT_CMDNK)) {
					ms_set_error_code(cr, card, MS_CMD_NK);
					TRACE_RET(cr, ms_parse_error_code(cr));
				}
			}
		}

		if (val & MS_RDY_TIMEOUT) {
			ms_set_error_code(cr, card, MS_TO_ERROR);
			TRACE_RET(cr, ms_parse_error_code(cr));
		}
		ms_set_error_code(cr, card, MS_TO_ERROR);
		TRACE_RET(cr, ms_parse_error_code(cr));
	}

	ptr = cr_get_cmd_data(cr) + 1;
	memcpy(data, ptr, data_len);

	if ((tpc == PRO_READ_SHORT_DATA) && (data_len == 8)) {
		rts_trace(TRACE_MS, "Read format progress:\n");
		RTS_DUMP(ptr, cnt);
	}

	return 0;
}


/*
 * ms_read_bytes --  Read bytes by submiting TPC to MS card
 * @cr: points to card reader structure
 * @c: points to card struct
 * @tpc: TPC(Transfer Protocol Command)
 * @cnt: data count to be read
 * @__cfg: Logical MS transaction configuration, which will be converted into
 * cr-dependent value in implement
 * @data: points to the data tobe read
 * @data_len: Length of expected data to be read, which is supposed to larger
 * than @cnt
 */
int ms_read_bytes(struct card_reader *cr, struct card *c, u8 tpc,
		u8 cnt, u8 __cfg, u8 *data, int data_len)
{

	if (cr->ms_ops_prot->read_bytes) {
		return cr->ms_ops_prot->read_bytes(cr, c, tpc, cnt,
						__cfg, data, data_len);
	} else {
		return ms_comm_read_bytes(cr, c, tpc, cnt,
						__cfg, data, data_len);
	}
}
EXPORT_SYMBOL_GPL(ms_read_bytes);

static int ms_send_cmd_4_transfer_data(struct card_reader *cr,
		u8 trans_mode, u8 tpc, u16 sec_cnt,
		u8 cfg, int mode_2k, int use_sg, void *buf, int buf_len)
{
	enum dma_data_direction dir;

	if (trans_mode == MS_TM_AUTO_READ)
		dir = DMA_FROM_DEVICE;
	else if (trans_mode == MS_TM_AUTO_WRITE)
		dir = DMA_TO_DEVICE;
	else {
		rts_trace(TRACE_SD, "unknown transfer mode %d\n", trans_mode);
		return -EINVAL;
	}

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, tpc);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_SECTOR_CNT_H,
			0xFF, (u8)(sec_cnt >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_SECTOR_CNT_L, 0xFF, (u8)sec_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, cfg);

	if (mode_2k) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_CFG,
				MS_2K_SECTOR_MODE, MS_2K_SECTOR_MODE);
	} else {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_CFG, MS_2K_SECTOR_MODE, 0);
	}

	cr_enable_dma(cr, sec_cnt * 512, DMA_512, dir);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER,
			0xFF, REG_MS_TRANSFER_START | trans_mode);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
			REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);
	cr_send_cmd_no_wait(cr);

	return 0;
}

static int ms_comm_transfer_data(struct card_reader *cr,
		struct card *card, u8 trans_mode, u8 tpc, u16 sec_cnt,
		u8 cfg, int mode_2k, int use_sg, void *buf, int buf_len)
{
	int retval;
	u8 val, err_code = 0;
	enum dma_data_direction dir;
	struct cr_option *option = &(cr->option);

	if (!buf || !buf_len)
		TRACE_RET(cr, STATUS_FAIL);

	if (trans_mode == MS_TM_AUTO_READ) {
		dir = DMA_FROM_DEVICE;
		err_code = MS_FLASH_READ_ERROR;
	} else if (trans_mode == MS_TM_AUTO_WRITE) {
		dir = DMA_TO_DEVICE;
		err_code = MS_FLASH_WRITE_ERROR;
	} else {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_send_cmd_4_transfer_data(cr, trans_mode, tpc, sec_cnt, cfg,
			mode_2k, use_sg, buf, buf_len);
	retval = cr_transfer_data(cr, MS_CARD, buf,
			buf_len, use_sg, dir, option->mspro_timeout);
	if (retval < 0) {
		ms_set_error_code(cr, card, err_code);
		if (retval == -ETIMEDOUT)
			retval = STATUS_TIMEDOUT;
		else
			retval = STATUS_FAIL;
		TRACE_RET(cr, retval);
	}

	cr_rd_reg(cr, REG_MS_TRANS_CFG, &val);
	if (val & (MS_INT_CMDNK | MS_INT_ERR | MS_CRC16_ERR | MS_RDY_TIMEOUT))
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}


/*
 * ms_transfer_data --  Transfer data between MS card and card reader
 * @cr: points to card reader structure
 * @c: points to card struct
 * @__mod: ms transfer mode, one of u8
 * @tpc: TPC(Transfer Protocol Command)
 * @sec_cnt: sector count
 * @__cfg: Logical MS transaction configuration, which will be converted into
 * cr-dependent value in implement
 * @mode_2k: 2KB per sector mode
 * @use_sg: Use scatter-gather mode or not
 * @buf: points to the data tobe read
 * @buf_len: Length of to be read
 */
int ms_transfer_data(struct card_reader *cr, struct card *c,
			u8 __mod, u8 tpc, u16 sec_cnt,
			u8 __cfg, int mode_2k, int use_sg,
						void *buf, int buf_len)
{
	if (cr->ms_ops_prot->transfer_data) {
		return cr->ms_ops_prot->transfer_data(cr, c, __mod, tpc,
				sec_cnt, __cfg, mode_2k, use_sg, buf, buf_len);
	} else {
		return ms_comm_transfer_data(cr, c, __mod, tpc,
				sec_cnt, __cfg, mode_2k, use_sg, buf, buf_len);
	}
}
EXPORT_SYMBOL_GPL(ms_transfer_data);

static void ms_comm_clear_error(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_CARD_STOP, MS_STOP | MS_CLR_ERR,
					MS_STOP | MS_CLR_ERR);
}

/*
 * ms_clear_error - Clear MS card error
 */
void ms_clear_error(struct card_reader *cr)
{
	if (cr->ms_ops_prot->clr_err)
		cr->ms_ops_prot->clr_err(cr);
	else
		ms_comm_clear_error(cr);
}
EXPORT_SYMBOL_GPL(ms_clear_error);


static int ms_comm_transfer_tpc(struct card_reader *cr, struct card *card,
					u8 trans_mode, u8 tpc, u8 cnt, u8 cfg)
{
	struct ms_info *ms_card = card->extra;
	u8 *ptr;

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, tpc);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_BYTE_CNT, 0xFF, cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, cfg);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
						0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER, 0xFF,
				REG_MS_TRANSFER_START | trans_mode);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
			REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);
	cr_add_cmd(cr, READ_REG_CMD, REG_MS_TRANS_CFG, 0, 0);

	if (cr_send_cmd(cr, MS_CARD, 5000)) {
		ms_clear_error(cr);
		ms_set_error_code(cr, card, MS_TO_ERROR);
		TRACE_RET(cr, ms_parse_error_code(cr));
	}
	ptr = cr_get_cmd_data(cr) + 1;
	if (!(tpc & 0x08)) { /* read packet */
		/*Check CRC16 & Ready Timeout*/
		if (*ptr & MS_CRC16_ERR) {
			ms_set_error_code(cr, card, MS_CRC16_ERROR);
			TRACE_RET(cr, ms_parse_error_code(cr));
		}
	} else { /* Write Packet */
		if (CHK_MSPRO(ms_card) && !(*ptr & 0x80)) {
			if (*ptr & (MS_INT_ERR |
				MS_INT_CMDNK)) {
				ms_set_error_code(cr, card, MS_CMD_NK);
				TRACE_RET(cr, ms_parse_error_code(cr));
			}
		}
	}

	if (*ptr & MS_RDY_TIMEOUT) {
		ms_clear_error(cr);
		ms_set_error_code(cr, card, MS_TO_ERROR);
		TRACE_RET(cr, ms_parse_error_code(cr));
	}

	return 0;
}


/*
 * ms_transfer_tpc --  Submit a TPC
 * @cr: points to card reader structure
 * @c: points to card struct
 * @__mod: ms transfer mode, one of u8
 * @tpc: TPC(Transfer Protocol Command)
 * @cnt: the byte count read/write from MS card registers
 * @__cfg: Logical MS transaction configuration, which will be converted into
 * cr-dependent value in implement
 */
int ms_transfer_tpc(struct card_reader *cr, struct card *c,
		u8 __mod, u8 tpc, u8 cnt, u8 __cfg)
{
	if (cr->ms_ops_prot->transfer_tpc)
		return cr->ms_ops_prot->transfer_tpc(cr, c, __mod, tpc, cnt,
				__cfg);
	else
		return ms_comm_transfer_tpc(cr, c, __mod, tpc, cnt, __cfg);

}
EXPORT_SYMBOL_GPL(ms_transfer_tpc);

static inline int ms_comm_set_sample_time_falling(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_MS_CFG, 0x98, MS_BUS_WIDTH_8 | SAMPLE_TIME_FALLING);

	return 0;
}

/*
 * ms_set_sample_time_falling - Set sample clock at the falling edge of hw MS
 * module clock
 */
int ms_set_sample_time_falling(struct card_reader *cr)
{
	if (cr->ms_ops_prot->set_sample_time_falling)
		return cr->ms_ops_prot->set_sample_time_falling(cr);
	else
		return ms_comm_set_sample_time_falling(cr);
}
EXPORT_SYMBOL_GPL(ms_set_sample_time_falling);

static inline int ms_comm_get_int_status(struct card_reader *cr, u8 *val)
{
	return cr_rd_reg(cr, REG_MS_TRANS_CFG, val);
}

/*
 * ms_get_int_status - Get MS card Interrupt status
 */
int ms_get_int_status(struct card_reader *cr, u8 *val)
{
	if (cr->ms_ops_prot->get_int_status)
		return cr->ms_ops_prot->get_int_status(cr, val);
	else
		return ms_comm_get_int_status(cr, val);
}
EXPORT_SYMBOL_GPL(ms_get_int_status);

int ms_set_rw_reg_addr(struct card_reader *cr, struct card *c,
		u8 read_start, u8 read_cnt, u8 write_start, u8 write_cnt)
{
	int retval, i;
	u8 data[4];

	rts_trace(TRACE_MS, "<--\n");

	data[0] = read_start;
	data[1] = read_cnt;
	data[2] = write_start;
	data[3] = write_cnt;

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, SET_RW_REG_ADRS, 4,
						NO_WAIT_INT, data, 4);
		if (retval == STATUS_SUCCESS)
			return STATUS_SUCCESS;

		ms_clear_error(cr);
	}

	rts_trace(TRACE_MS, "-->\n");

	TRACE_RET(cr, STATUS_FAIL);
}
EXPORT_SYMBOL_GPL(ms_set_rw_reg_addr);

static int ms_comm_read_regs_to_buf(struct card_reader *cr,
						u8 *buf, int offset, int count)
{
	int i;

	for (i = 0; i < count; i++)
		cr_rd_reg(cr, PPBUF_BASE2 + offset + i, buf++);

	return 0;
}

/*
 * ms_read_regs_to_buf - Sequentially read @count registers address from
 * PPBUF_BASE2 + @offset
 * @cr: points to card reader structure
 * @buf: where to store data
 * @offset: offset to PPBUF_BASE2
 * @count: count of registers to be read
 */
int ms_read_regs_to_buf(struct card_reader *cr, u8 *buf, int offset, int cnt)
{
	if (cr->ms_ops_prot->read_regs_to_buf)
		return cr->ms_ops_prot->read_regs_to_buf(cr, buf, offset, cnt);
	else
		return ms_comm_read_regs_to_buf(cr, buf, offset, cnt);
}
EXPORT_SYMBOL_GPL(ms_read_regs_to_buf);


static int ms_comm_set_init_para(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;
	struct cr_option *option = &(cr->option);

	if (CHK_MSHG(ms_card)) {
		if (cr->asic_code)
			ms_card->ms_clock = option->asic_ms_hg_clk;
		else
			ms_card->ms_clock = option->fpga_ms_hg_clk;
	} else if (CHK_MSPRO(ms_card) ||
			CHK_MS4BIT(ms_card)) {
		if (cr->asic_code)
			ms_card->ms_clock = option->asic_ms_4bit_clk;
		else
			ms_card->ms_clock = option->fpga_ms_4bit_clk;
	} else {
		if (cr->asic_code)
			ms_card->ms_clock = option->asic_ms_1bit_clk;
		else
			ms_card->ms_clock = option->fpga_ms_1bit_clk;
	}

	if (cr_switch_clk(cr, card, ms_card->ms_clock))
		TRACE_RET(cr, STATUS_FAIL);

	if (cr_select_card(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

/*
 * ms_set_init_para - Set inital parameters including "switch clock" and "select
 * card"
 */
int ms_set_init_para(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->set_init_para)
		return cr->ms_ops_prot->set_init_para(cr, c);
	else
		return ms_comm_set_init_para(cr, c);
}
EXPORT_SYMBOL_GPL(ms_set_init_para);

int ms_switch_clock(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	if (cr_select_card(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	if (cr_switch_clk(cr, card, ms_card->ms_clock))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}
EXPORT_SYMBOL_GPL(ms_switch_clock);

static int ms_comm_auto_tune_clock(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	if (cr->asic_code) {
		if (ms_card->ms_clock > 30)
			ms_card->ms_clock -= 20;
	} else {
		if (ms_card->ms_clock == CLK_80)
			ms_card->ms_clock = CLK_60;
		else if (ms_card->ms_clock == CLK_60)
			ms_card->ms_clock = CLK_40;
	}

	if (ms_switch_clock(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

/*
 * ms_auto_tune_clock - Auto decrease clock frequency following some convension
 */
int ms_auto_tune_clock(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->auto_tune_clk)
		return cr->ms_ops_prot->auto_tune_clk(cr, c);
	else
		return ms_comm_auto_tune_clock(cr, c);
}
EXPORT_SYMBOL_GPL(ms_auto_tune_clock);

int ms_read_status_reg(struct card_reader *cr, struct card *c)
{
	u8 val[2];

	if (ms_set_rw_reg_addr(cr, c, StatusReg0, 2, 0, 0))
		TRACE_RET(cr, STATUS_FAIL);

	if (ms_read_bytes(cr, c, READ_REG, 2, NO_WAIT_INT, val, 2))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_MS, "ms statusReg0=0x%x, statusReg1=0x%x\n",
							val[0], val[1]);

	/* Unable to correct error */
	if (val[1] & (STS_UCDT | STS_UCEX | STS_UCFG)) {
		ms_set_error_code(cr, c, MS_FLASH_READ_ERROR);
		TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(ms_read_status_reg);

static int ms_read_extra_data(struct card_reader *cr,
			struct card *c, u16 block_addr, u8 page_num,
							u8 *buf, int buf_len)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 val, data[10];

	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
					MS_COMM_EXTRA_SIZE, SystemParm, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Write REG */
	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(block_addr >> 8);
	data[3] = (u8)block_addr;
	/* Page Number */
	/* Extra data access mode */
	data[4] = 0x40;
	data[5] = page_num;

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, WRITE_REG, 6,
						NO_WAIT_INT, data, 6);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_set_error_code(cr, c, MS_NO_ERROR);

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_send_cmd(cr, c, BLOCK_READ, WAIT_INT);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_read_bytes(cr, c, GET_INT, 1, NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	if (val & INT_REG_CMDNK) {
		/* Command can not be executed */
		ms_set_error_code(cr, c, MS_CMD_NK);
		TRACE_RET(cr, STATUS_FAIL);
	}
	/* Command error termination */
	if (val & INT_REG_CED) {
		if (val & INT_REG_ERR) {
			retval = ms_read_status_reg(cr, c);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
			retval = ms_set_rw_reg_addr(cr, c,
					OverwriteFlag, MS_COMM_EXTRA_SIZE,
							SystemParm, 6);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
	}
	retval = ms_read_bytes(cr, c, READ_REG,
				MS_COMM_EXTRA_SIZE, NO_WAIT_INT,
						data, MS_COMM_EXTRA_SIZE);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (buf && buf_len) {
		if (buf_len > MS_COMM_EXTRA_SIZE)
			buf_len = MS_COMM_EXTRA_SIZE;
		memcpy(buf, data, buf_len);
	}

	return STATUS_SUCCESS;
}


static int ms_write_extra_data(struct card_reader *cr,
				struct card *c, u16 block_addr, u8 page_num,
							u8 *buf, int buf_len)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 val, data[16];

	if (!buf || (buf_len < MS_COMM_EXTRA_SIZE)) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
					MS_COMM_EXTRA_SIZE, SystemParm,
						6 + MS_COMM_EXTRA_SIZE);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Write REG */
	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(block_addr >> 8);
	data[3] = (u8)block_addr;
	/* Page Number */
	/* Extra data access mode */
	data[4] = 0x40;
	data[5] = page_num;

	for (i = 6; i < MS_COMM_EXTRA_SIZE + 6; i++) {
		data[i] = buf[i - 6];
	}
	retval = ms_write_bytes(cr, c, WRITE_REG ,
						(6+MS_COMM_EXTRA_SIZE),
						NO_WAIT_INT, data, 16);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_send_cmd(cr, c, BLOCK_WRITE, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	if (val & INT_REG_CMDNK) {
		/* Command can not be executed */
		ms_set_error_code(cr, c, MS_CMD_NK);
		TRACE_RET(cr, STATUS_FAIL);
	}
	/* Command error termination */
	if (val & INT_REG_CED) {
		if (val & INT_REG_ERR) {
			ms_set_error_code(cr, c, MS_FLASH_WRITE_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int ms_comm_read_page(struct card_reader *cr, struct card *c,
						u16 block_addr, u8 page_num)
{
	struct ms_info *ms_card = c->extra;
	int retval;
	u8 val, data[6];

	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
					MS_COMM_EXTRA_SIZE, SystemParm, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Write REG */
	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(block_addr >> 8);
	data[3] = (u8)block_addr;
	/* Page Number */
	/* Single page access mode */
	data[4] = 0x20;
	data[5] = page_num;
	retval = ms_write_bytes(cr, c, WRITE_REG , 6,
						NO_WAIT_INT, data, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_send_cmd(cr, c, BLOCK_READ, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	if (val & INT_REG_CMDNK) {
		/* Command can not be executed */
		ms_set_error_code(cr, c, MS_CMD_NK);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_CED) {
		if (val & INT_REG_ERR) {
			if (!(val & INT_REG_BREQ)) {
				ms_set_error_code(cr, c,
						MS_FLASH_READ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
			retval = ms_read_status_reg(cr, c);
			if (retval != STATUS_SUCCESS) {
				ms_set_error_code(cr, c,
						MS_FLASH_WRITE_ERROR);
			}
		} else {
			if (!(val & INT_REG_BREQ)) {
				ms_set_error_code(cr, c,
						MS_BREQ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
	}
	retval = ms_transfer_tpc(cr, c, MS_TM_NORMAL_READ,
			READ_PAGE_DATA, 0, NO_WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Uncorrectable error */
	/* Notice that the data of the page which has uncorrectable error is
	 * read out also, just return failed */
	if (ms_check_error_code(cr, c, MS_FLASH_WRITE_ERROR)) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int ms_set_bad_block(struct card_reader *cr,
						struct card *c, u16 phy_blk)
{
	struct ms_info *ms_card = c->extra;
	int retval;
	u8 val, data[8], extra[MS_COMM_EXTRA_SIZE];

	retval = ms_read_extra_data(cr, c, phy_blk, 0, extra,
						MS_COMM_EXTRA_SIZE);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
				MS_COMM_EXTRA_SIZE, SystemParm, 7);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_set_error_code(cr, c, MS_NO_ERROR);

	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(phy_blk >> 8);
	data[3] = (u8)phy_blk;
	data[4] = 0x80;
	data[5] = 0;
	data[6] = extra[0] & 0x7F;
	data[7] = 0xFF;
	retval = ms_write_bytes(cr, c, WRITE_REG , 7,
						NO_WAIT_INT, data, 7);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_send_cmd(cr, c, BLOCK_WRITE, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_CMDNK) {
		/* Command can not be executed */
		ms_set_error_code(cr, c, MS_CMD_NK);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_CED) {
		if (val & INT_REG_ERR) {
			ms_set_error_code(cr, c,
					MS_FLASH_WRITE_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int ms_erase_block(struct card_reader *cr, struct card *c,
								u16 phy_blk)
{
	struct ms_info *ms_card = c->extra;
	int retval, i = 0;
	u8 val, data[6];

	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
				MS_COMM_EXTRA_SIZE, SystemParm, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_set_error_code(cr, c, MS_NO_ERROR);

	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(phy_blk >> 8);
	data[3] = (u8)phy_blk;
	data[4] = 0;
	data[5] = 0;

	retval = ms_write_bytes(cr, c, WRITE_REG, 6,
						NO_WAIT_INT, data, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

ERASE_RTY:
	retval = ms_send_cmd(cr, c, BLOCK_ERASE, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_CMDNK) {
		if (i < 3) {
			i++;
			goto ERASE_RTY;
		}

		/* Command can not be executed */
		ms_set_error_code(cr, c, MS_CMD_NK);
		ms_set_bad_block(cr, c, phy_blk);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_CED) {
		if (val & INT_REG_ERR) {
			ms_set_error_code(cr, c, MS_FLASH_WRITE_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static void ms_comm_set_page_status(u16 log_blk, u8 type,
						u8 *extra, int extra_len)
{
	if (!extra || (extra_len < MS_COMM_EXTRA_SIZE))
		return;

	memset(extra, 0xFF, MS_COMM_EXTRA_SIZE);

	if (type == setPS_NG) {
		/* set page status as 1:NG,and block status keep 1:OK */
		extra[0] = 0xB8;
	} else {
		/* set page status as 0:Data Error,and block status keep 1:OK */
		extra[0] = 0x98;
	}

	extra[2] = (u8)(log_blk >> 8);
	extra[3] = (u8)log_blk;
}

static int ms_init_page(struct card_reader *cr, struct card *c,
			u16 phy_blk, u16 log_blk, u8 start_page, u8 end_page)
{
	int retval;
	u8 extra[MS_COMM_EXTRA_SIZE], i;

	memset(extra, 0xff, MS_COMM_EXTRA_SIZE);

	extra[0] = 0xf8; /* Block, page OK, data erased */
	extra[1] = 0xff;
	extra[2] = (u8)(log_blk >> 8);
	extra[3] = (u8)log_blk;

	for (i = start_page; i < end_page; i++) {
		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = ms_write_extra_data(cr, c, phy_blk, i,
						extra, MS_COMM_EXTRA_SIZE);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int ms_copy_page(struct card_reader *cr, struct card *c,
					u16 old_blk, u16 new_blk, u16 log_blk,
						u8 start_page, u8 end_page)
{
	struct ms_info *ms_card = c->extra;
	int retval, rty_cnt, uncorrect_flag = 0;
	u8 extra[MS_COMM_EXTRA_SIZE], val, i, j, data[16];

	rts_trace(TRACE_MS,
		"Copy page from 0x%x to 0x%x, logical block is 0x%x\n",
		old_blk, new_blk, log_blk);
	rts_trace(TRACE_MS, "start_page = %d, end_page = %d\n",
		start_page, end_page);

	/* Extra data of the last block of one segment may be failed read, and
	 * it result in status register 1 keeping failure informations.  So when
	 * the following check status register 1 f/w will return fail.  To avoid
	 * this add reading extra data of new block before function
	 * ms_read_status_reg which is for checking new block extra data
	 * status. */
	retval = ms_read_extra_data(cr, c, new_blk, 0, extra,
							MS_COMM_EXTRA_SIZE);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_read_status_reg(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_read_regs_to_buf(cr, &val, 0, 1);
	if (retval)
		return STATUS_FAIL;
	if (val & BUF_FULL) {
		/* Clear Buffer */
		retval = ms_send_cmd(cr, c, CLEAR_BUF,
							WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* GET_INT Register */
		retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (!(val & INT_REG_CED)) {
			ms_set_error_code(cr, c,
					MS_FLASH_WRITE_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	for (i = start_page; i < end_page; i++) {
		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* for MS check procedure check the page status of the source
		 * page is [1,1]:OK, if not OK,means that the page is
		 * uncorrectable error and the page status og targe page shall
		 * not be set [1,1]:OK */
		DISABLE_TRIGGER();
		ms_read_extra_data(cr, c, old_blk, i, extra,
							MS_COMM_EXTRA_SIZE);
		ENABLE_TRIGGER();
		retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
					MS_COMM_EXTRA_SIZE, SystemParm, 6);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* Write REG */
		ms_set_error_code(cr, c, MS_NO_ERROR);

		if (CHK_MS4BIT(ms_card)) {
			/* Parallel interface */
			data[0] = 0x88;
		} else {
			/* Serial interface */
			data[0] = 0x80;
		}
		/* Block Address */
		data[1] = 0;
		data[2] = (u8)(old_blk >> 8);
		data[3] = (u8)old_blk;
		data[4] = 0x20;
		data[5] = i;
		retval = ms_write_bytes(cr, c, WRITE_REG, 6,
						NO_WAIT_INT, data, 6);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* read page data to internal buffer by command */
		retval = ms_send_cmd(cr, c, BLOCK_READ,
							WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* GET_INT Register */
		ms_set_error_code(cr, c, MS_NO_ERROR);
		retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (val & INT_REG_CMDNK) {
			/* Command can not be executed */
			ms_set_error_code(cr, c, MS_CMD_NK);
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (val & INT_REG_CED) {
			if (val & INT_REG_ERR) {
				retval = ms_read_status_reg(cr, c);
				if (retval != STATUS_SUCCESS) {
					/* for MS check procedure
					 * if occurs uncorrectable error,
					 * after read(dump) the buffer data,
					 * don't write back */
					uncorrect_flag = 1;
					rts_trace(TRACE_MS,
						"Uncorrectable error\n");
				} else {
					uncorrect_flag = 0;
				}

				/* Read Sector Data of Source Block Page to
				 * SRAM2 */
				retval = ms_transfer_tpc(cr, c,
						MS_TM_NORMAL_READ,
						READ_PAGE_DATA, 0, NO_WAIT_INT);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}

				/* Uncorrectable data in old physical block
				 * means something wrong with read/write In new
				 * physical block, the data is not guaranteed,
				 * but no problem with read/write */
				if (uncorrect_flag) {
					/* set the page status of the source
					 * page as [0.1]:NG */
					ms_comm_set_page_status(log_blk,
							setPS_NG, extra,
							MS_COMM_EXTRA_SIZE);
					if (i == 0) {
						/* Set Update Status to 0 when
						 * the bad page is the first
						 * page of the block as it
						 * happens */
						extra[0] &= 0xEF;
					}
					ms_write_extra_data(cr, c,
							old_blk, i, extra,
							MS_COMM_EXTRA_SIZE);
					rts_trace(TRACE_MS,
						"page %d : extra[0] = 0x%x\n",
						i, extra[0]);
					MS_SET_BAD_BLOCK_FLG(ms_card);

					/* set the page status of the targe page
					 * as [0.0]:ERROR */
					ms_comm_set_page_status(log_blk,
						setPS_Error, extra,
						MS_COMM_EXTRA_SIZE);
					ms_write_extra_data(cr, c,
							new_blk, i, extra,
							MS_COMM_EXTRA_SIZE);
					continue;
				}

				for (rty_cnt = 0;
					rty_cnt < MS_COMM_MAX_RETRY_COUNT;
					rty_cnt++) {
					retval = ms_transfer_tpc(cr, c,
						MS_TM_NORMAL_WRITE,
						WRITE_PAGE_DATA, 0,
						NO_WAIT_INT);
					if (retval == STATUS_SUCCESS)
						break;
				}
				if (rty_cnt == MS_COMM_MAX_RETRY_COUNT) {
					TRACE_RET(cr, STATUS_FAIL);
				}
			}

			if (!(val & INT_REG_BREQ)) {
				ms_set_error_code(cr, c,
							MS_BREQ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
		retval = ms_set_rw_reg_addr(cr, c,
				OverwriteFlag, MS_COMM_EXTRA_SIZE,
				SystemParm, (6+MS_COMM_EXTRA_SIZE));

		/* Write REG */
		ms_set_error_code(cr, c, MS_NO_ERROR);

		if (CHK_MS4BIT(ms_card)) {
			/* Parallel interface */
			data[0] = 0x88;
		} else {
			/* Serial interface */
			data[0] = 0x80;
		}
		/* Block Address */
		data[1] = 0;
		data[2] = (u8)(new_blk >> 8);
		data[3] = (u8)new_blk;
		data[4] = 0x20;
		data[5] = i;

		/* for MS check procedure */
		if ((extra[0] & 0x60) != 0x60)
			data[6] = extra[0];
		else
			data[6] = 0xF8;
		data[6 + 1] = 0xFF;
		data[6 + 2] = (u8)(log_blk >> 8);
		data[6 + 3] = (u8)log_blk;

		for (j = 4; j <= MS_COMM_EXTRA_SIZE; j++)
			data[6 + j] = 0xFF;
		retval = ms_write_bytes(cr, c, WRITE_REG,
						(6 + MS_COMM_EXTRA_SIZE),
						NO_WAIT_INT, data, 16);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = ms_send_cmd(cr, c, BLOCK_WRITE,
							WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* GET_INT Register */
		ms_set_error_code(cr, c, MS_NO_ERROR);
		retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (val & INT_REG_CMDNK) {
			/* Command can not be executed */
			ms_set_error_code(cr, c, MS_CMD_NK);
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (val & INT_REG_CED) {
			if (val & INT_REG_ERR) {
				ms_set_error_code(cr, c,
						MS_FLASH_WRITE_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		if (i == 0) {
			retval = ms_set_rw_reg_addr(cr, c,
					OverwriteFlag, MS_COMM_EXTRA_SIZE,
							SystemParm, 7);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}

			ms_set_error_code(cr, c, MS_NO_ERROR);

			if (CHK_MS4BIT(ms_card)) {
				/* Parallel interface */
				data[0] = 0x88;
			} else {
				/* Serial interface */
				data[0] = 0x80;
			}
			/* Block Address */
			data[1] = 0;
			data[2] = (u8)(old_blk >> 8);
			data[3] = (u8)old_blk;
			data[4] = 0x80;
			data[5] = 0;
			data[6] = 0xEF;
			data[7] = 0xFF;
			retval = ms_write_bytes(cr, c,
						WRITE_REG, 7,
						NO_WAIT_INT, data, 8);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
			retval = ms_send_cmd(cr, c, BLOCK_WRITE,
							WAIT_INT);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}

			/* GET_INT Register */
			ms_set_error_code(cr, c, MS_NO_ERROR);
			retval = ms_read_bytes(cr, c,
						GET_INT, 1,
						NO_WAIT_INT, &val, 1);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}

			if (val & INT_REG_CMDNK) {
				/* Command can not be executed */
				ms_set_error_code(cr, c, MS_CMD_NK);
				TRACE_RET(cr, STATUS_FAIL);
			}

			if (val & INT_REG_CED) {
				if (val & INT_REG_ERR) {
					ms_set_error_code(cr, c,
						MS_FLASH_WRITE_ERROR);
					TRACE_RET(cr, STATUS_FAIL);
				}
			}
		}
	}

	return STATUS_SUCCESS;
}

/* Procedure to identify Media Typ
 * Read register: type, category, class */
static int ms_real_identify_media_type(struct card_reader *cr,
					struct card *c, int switch_8bit_bus)
{
	struct ms_info *ms_card = c->extra;
	u8 buf[16];
	u8 val;
	int ret = 0;

	rts_trace(TRACE_MS, "<--\n");

	ret = ms_read_regs_to_buf(cr, buf, 0, 6);
	if (ret)
		return STATUS_FAIL;

	/* Type Register */
	/* cr_rd_reg(cr, PPBUF_BASE2 + 2, &val); */
	val = *(buf + 2);
	rts_trace(TRACE_MS, "Type register: 0x%x\n", val);
	if (val != 0x01) {
		/* For MS-PRO certification 2-7-5: VTE tester will set "Type
		 * Register = 0x02" We shall Power-Off MS_3V3, and we can't do
		 * initializing MS too.  If not 2, we need check as if Memory
		 * Stick Card!*/
		if (val != 0x02)
			ms_card->check_ms_flow = 1;
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Category Register */
	/* cr_rd_reg(cr, PPBUF_BASE2 + 4, &val); */
	val = *(buf + 4);
	rts_trace(TRACE_MS, "Category register: 0x%x\n", val);
	if (val != 0) {
		ms_card->check_ms_flow = 1;
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Class Register */
	/* cr_rd_reg(cr, PPBUF_BASE2 + 5, &val); */
	val = *(buf + 5);
	rts_trace(TRACE_MS, "Class register: 0x%x\n", val);
	if (val == 0) {
		/* cr_rd_reg(cr, PPBUF_BASE2, &val); */
		val = *(buf);
		if (val & MS_COMM_WRT_PRTCT)
			set_card_flags(c, WP);
		else
			clr_card_flags(c, WP);
	} else if ((val == 0x01) || (val == 0x02) || (val == 0x03)) {
		set_card_flags(c, WP);
	} else {
		ms_card->check_ms_flow = 1;
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_card->ms_type |= TYPE_MSPRO;

	/* Check MSPro-HG Card, use IF Mode Register to distinguish */
	/* cr_rd_reg(cr, PPBUF_BASE2 + 3, &val); */
	val = *(buf + 3);
	rts_trace(TRACE_MS, "IF Mode register: 0x%x\n", val);
	if (val == 0) {
		ms_card->ms_type &= 0x0F;
	} else if (val == 7) {
		if (switch_8bit_bus) {
			ms_card->ms_type |= MS_HG;
		} else {
			ms_card->ms_type &= 0x0F;
		}
	} else {
		TRACE_RET(cr, STATUS_FAIL);
	}

	rts_trace(TRACE_MS, "-->\n");

	return ret;
}

int ms_identify_media_type(struct card_reader *cr,
					struct card *c, int switch_8bit_bus)
{
	int retval, i;

	rts_trace(TRACE_MS, "<--\n");

	retval = ms_set_rw_reg_addr(cr, c, Pro_StatusReg,
						6, SystemParm, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	/* Get Register form MS-PRO card */
	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_transfer_tpc(cr, c, MS_TM_READ_BYTES,
					READ_REG, 6, NO_WAIT_INT);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	retval = ms_real_identify_media_type(cr, c, switch_8bit_bus);

	rts_trace(TRACE_MS, "-->\n");

	/* end Procedure to identify Media Type */
	return retval;
}
EXPORT_SYMBOL_GPL(ms_identify_media_type);


static int ms_confirm_cpu_startup(struct card_reader *cr,
							struct card *c)
{
	int retval, i, k;
	u8 val;

	/* Confirm CPU StartUp */
	k = 0;
	do {
		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}

		for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
			retval = ms_read_bytes(cr, c, GET_INT, 1,
							NO_WAIT_INT, &val, 1);
			if (retval == STATUS_SUCCESS)
				break;
		}
		if (i == MS_COMM_MAX_RETRY_COUNT) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (k > 100) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		k++;
		msleep(100);
	} while (!(val & INT_REG_CED));

	/* For MSPro Certification 2-7-9, MS-PRO Fatal error occured(CED=1 &
	 * ERR=1).  We shall Power-Off MS_3V3, and we can't do initializing MS
	 * too. */
	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_read_bytes(cr, c, GET_INT, 1,
							NO_WAIT_INT, &val, 1);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (val & INT_REG_ERR) {
		/* For MSPro Certification 2-7-10, Write-Disabled Status  */
		if (val & INT_REG_CMDNK) { /* CMDNK = 1	 */
			set_card_flags(c, WP);
		} else { /* CMDNK = 0 */
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	/* --  end confirm CPU startup */

	return STATUS_SUCCESS;
}


static int ms_switch_parallel_bus(struct card_reader *cr,
							struct card *c)
{
	int retval, i;
	u8 data[2];

	/* For MS-Pro Certification 2-8-3, if switch 4-bit bus fail(after
	 * retry), ++ set ms parallel 4-bit interface */
	data[0] = PARALLEL_4BIT_IF;
	data[1] = 0;
	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, WRITE_REG, 1,
						NO_WAIT_INT, data, 2);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (retval != STATUS_SUCCESS) {
		/* For MS-Pro Certification 2-8-3:
		 * We shall power reset MS-PRo Card if switched-bus fail. */
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int ms_switch_8bit_bus(struct card_reader *cr,
							struct card *c)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 data[2];

	data[0] = PARALLEL_8BIT_IF;
	data[1] = 0;
	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, WRITE_REG, 1,
						NO_WAIT_INT, data, 2);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (retval != STATUS_SUCCESS) {
		/* For MS-Pro Certification 2-8-3:
		 * We shall power reset MS-PRo Card if switched-bus fail. */
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Switch MS-PRO into 8 bits Parallel mode
	 * Because HG run as 60MHz, and we shall guarantee the tPDod 10ns,
	 * So we shall set "SAMPLE_TIME_FALLING" to delay sample the output data
	 * from HG card. */
	retval = ms_set_sample_time_falling(cr);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	ms_card->ms_type |= MS_8BIT;
	retval = ms_set_init_para(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* To solve the issue that the socket is 4-bit only,
	 * We shall try some command under 8-bit here, such as GET_INT TPC... */
	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		/* At least 1 GET_INT TPC Fail, bus not match,
		 * shall reset to run with 4-bit */
		retval = ms_transfer_tpc(cr, c, MS_TM_READ_BYTES,
					GET_INT, 1, NO_WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int ms_mspro_reset_flow(struct card_reader *cr,
					struct card *c,	int switch_8bit_bus)
{
	struct ms_info *ms_card = c->extra;
	struct cr_option *option = &(cr->option);
	int ret, i;

	rts_trace(TRACE_MS, "<--\n");

	for (i = 0; i < 3; i++) {
		ret = ms_prepare_reset(cr, c);
		if (ret != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		ret = ms_identify_media_type(cr, c, switch_8bit_bus);
		if (ret != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		ret = ms_confirm_cpu_startup(cr, c);
		if (ret != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		ret = ms_switch_parallel_bus(cr, c);
		if (ret != STATUS_SUCCESS) {
			if (cr_detect_card(cr, MS_ID) !=
				STATUS_SUCCESS) {
				ms_set_error_code(cr, c, MS_NO_CARD);
				TRACE_RET(cr, STATUS_FAIL);
			}
			continue;
		} else {
			break;
		}
	}

	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Switch MS-PRO into Parallel mode */
	ret = ms_switch_bus_width_4(cr);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Set MS-PRO Card timing 40MHz, and set Push Timing as "ODD" */
	ret = ms_set_push_timming_odd(cr);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	ret = ms_set_init_para(cr, c);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* If MSPro HG Card, We shall try to switch to 8-bit bus */
	if (CHK_MSHG(ms_card) && option->support_ms_8bit && switch_8bit_bus) {
		ret = ms_switch_8bit_bus(cr, c);
		if (ret != STATUS_SUCCESS) {
			ms_card->switch_8bit_fail = 1;
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	rts_trace(TRACE_MS, "-->\n");

	return STATUS_SUCCESS;
}


static int ms_read_attr_info(struct card_reader *cr,
							struct card *c)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 val, *buf, class_code, device_type, sub_class, data[16];
	u16 total_blk = 0, blk_size = 0;
#ifdef SUPPORT_MSXC
	u32 xc_total_blk = 0, xc_blk_size = 0;
#endif
	u32 sys_info_addr = 0, sys_info_size;
#ifdef SUPPORT_PCGL_1P18
	u32 model_name_addr = 0, model_name_size;
	int found_sys_info = 0, found_model_name = 0;
#endif
	retval = ms_set_rw_reg_addr(cr, c, Pro_IntReg, 2,
						Pro_SystemParm, 7);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (CHK_MS8BIT(ms_card)) {
		data[0] = PARALLEL_8BIT_IF;
	} else {
		data[0] = PARALLEL_4BIT_IF;
	}
	data[1] = 0;

	/* For MS-Pro Certification 2-7-15
	 * The Maximun sector count of Attribute Data is 0x40 sectors */
	data[2] = 0x40;
	data[3] = 0;
	data[4] = 0;
	data[5] = 0;
	/* Start address 0 */
	data[6] = 0;
	data[7] = 0;

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, PRO_WRITE_REG, 7,
						NO_WAIT_INT, data, 8);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	buf = (u8 *)cr_alloc_dma_buf(cr, 64 * 512, GFP_KERNEL);
	if (buf == NULL) {
		TRACE_RET(cr, STATUS_ERROR);
	}

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_send_cmd(cr, c, PRO_READ_ATRB,
							WAIT_INT);
		if (retval != STATUS_SUCCESS)
			continue;
		/* For MS-Pro Certification 2-7-15,
		 * We shall check BREQ after sending "READ_ATTRIBUTE" cmd */
		retval = ms_get_int_status(cr, &val);
		if (retval != STATUS_SUCCESS) {
			cr_free_dma_buf(cr, buf);
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (!(val & MS_COMM_INT_BREQ)) {
			cr_free_dma_buf(cr, buf);
			TRACE_RET(cr, STATUS_FAIL);
		}
		/* For MS-PRO Certification 2-7-15, we wait INT signal before
		 * transfer completed */
		retval = ms_transfer_data(cr, c, MS_TM_AUTO_READ,
						PRO_READ_LONG_DATA, 0x40,
					WAIT_INT, 0, 0, buf, 64 * 512);
		if (retval == STATUS_SUCCESS)
			break;
		else
			ms_clear_error(cr);

	}
	if (retval != STATUS_SUCCESS) {
		cr_free_dma_buf(cr, buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	i = 0;
	/* Due to many Sandisk's MS-PRO Card, they will still put BREQ to '1'
	 * after "STOP" command. */
	do {
		retval = ms_get_int_status(cr, &val);
		if (retval != STATUS_SUCCESS) {
			cr_free_dma_buf(cr, buf);
			TRACE_RET(cr, STATUS_FAIL);
		}
		if ((val & MS_COMM_INT_CED) || !(val & MS_COMM_INT_BREQ))
			break;
		retval = ms_transfer_tpc(cr, c, MS_TM_NORMAL_READ,
				PRO_READ_LONG_DATA, 0, WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			cr_free_dma_buf(cr, buf);
			TRACE_RET(cr, STATUS_FAIL);
		}

		i++;
	} while (i < 1024);

	if (retval != STATUS_SUCCESS) {
		cr_free_dma_buf(cr, buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if ((buf[0] != 0xa5) && (buf[1] != 0xc3)) {
		/* Signature code is wrong */
		cr_free_dma_buf(cr, buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* For MsPro Certification 2-5-2. We shall judge wheather
	 * "Device Informantion Entry Count" <= 12 and >=1. */
	if ((buf[4] < 1) || (buf[4] > 12)) {
		cr_free_dma_buf(cr, buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	for (i = 0; i < buf[4]; i++) {
		int cur_addr_off = 16 + i * 12;

#ifdef SUPPORT_MSXC
		if ((buf[cur_addr_off + 8] == 0x10) ||
					(buf[cur_addr_off + 8] == 0x13))
#else
		if (buf[cur_addr_off + 8] == 0x10)
#endif
		{
			/* Device Information ID = 0x10 means
			 * this entry is "System Information" Entry */
			sys_info_addr = ((u32)buf[cur_addr_off + 0] << 24) |
					((u32)buf[cur_addr_off + 1] << 16) |
					((u32)buf[cur_addr_off + 2] << 8) |
							buf[cur_addr_off + 3];
			sys_info_size = ((u32)buf[cur_addr_off + 4] << 24) |
					((u32)buf[cur_addr_off + 5] << 16) |
					((u32)buf[cur_addr_off + 6] << 8) |
							buf[cur_addr_off + 7];
			rts_trace(TRACE_MS, "sys_info_addr = 0x%x,"
						"sys_info_size = 0x%x\n",
						sys_info_addr, sys_info_size);
			if (sys_info_size != 96) {/* Size shall equal to 96 */
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if (sys_info_addr < 0x1A0) {
				/* Address shall be equal to
				 * or greater than 0x1A0 */
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if ((sys_info_size + sys_info_addr) > 0x8000) {
				/* Address + Size can't be over than 0x8000 */
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}

#ifdef SUPPORT_MSXC
			if (buf[cur_addr_off + 8] == 0x13) {
				ms_card->ms_type |= MS_XC;
			}
#endif
#ifdef SUPPORT_PCGL_1P18
			found_sys_info = 1;
#else
			break;
#endif
		}
#ifdef SUPPORT_PCGL_1P18
		if (buf[cur_addr_off + 8] == 0x15) {
			/* Device Information ID = 0x15 means
			 * this entry is "Model Name" Entry */
			model_name_addr = ((u32)buf[cur_addr_off + 0] << 24) |
					((u32)buf[cur_addr_off + 1] << 16) |
					((u32)buf[cur_addr_off + 2] << 8) |
							buf[cur_addr_off + 3];
			model_name_size = ((u32)buf[cur_addr_off + 4] << 24) |
					((u32)buf[cur_addr_off + 5] << 16) |
					((u32)buf[cur_addr_off + 6] << 8) |
							buf[cur_addr_off + 7];
			rts_trace(TRACE_MS, "model_name_addr = 0x%x,"
						"model_name_size = 0x%x\n",
					model_name_addr, model_name_size);
			if (model_name_size != 48)  {
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if (model_name_addr < 0x1A0) {
				/* Address shall be equal to
				 * or greater than 0x1A0 */
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if ((model_name_size + model_name_addr) > 0x8000) {
				/* Address + Size can't be over than 0x8000 */
				cr_free_dma_buf(cr, buf);
				TRACE_RET(cr, STATUS_FAIL);
			}

			found_model_name = 1;
		}

		if (found_sys_info && found_model_name) {
			break;
		}
#endif
	}

	if (i == buf[4]) {
		/* Can't find System Information Entry and Model Name Entry */
		cr_free_dma_buf(cr, buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	class_code =  buf[sys_info_addr + 0];
	device_type = buf[sys_info_addr + 56];
	sub_class = buf[sys_info_addr + 46];
#ifdef SUPPORT_MSXC
	if (CHK_MSXC(ms_card)) {
		xc_total_blk = ((u32)buf[sys_info_addr + 6] << 24) |
					((u32)buf[sys_info_addr + 7] << 16) |
					((u32)buf[sys_info_addr + 8] << 8) |
							buf[sys_info_addr + 9];
		xc_blk_size = ((u32)buf[sys_info_addr + 32] << 24) |
					((u32)buf[sys_info_addr + 33] << 16) |
					((u32)buf[sys_info_addr + 34] << 8) |
							buf[sys_info_addr + 35];
		rts_trace(TRACE_MS, "xc_total_blk = 0x%x, xc_blk_size = 0x%x\n",
						xc_total_blk, xc_blk_size);
	} else {
		total_blk = ((u16)buf[sys_info_addr + 6] << 8) |
							buf[sys_info_addr + 7];
		blk_size = ((u16)buf[sys_info_addr + 2] << 8) |
							buf[sys_info_addr + 3];
		rts_trace(TRACE_MS, "total_blk = 0x%x, blk_size = 0x%x\n",
							total_blk, blk_size);
	}
#else
	total_blk = ((u16)buf[sys_info_addr + 6] << 8) |
						buf[sys_info_addr + 7];
	blk_size = ((u16)buf[sys_info_addr + 2] << 8) |
						buf[sys_info_addr + 3];
	rts_trace(TRACE_MS, "total_blk = 0x%x, blk_size = 0x%x\n",
							total_blk, blk_size);
#endif

	rts_trace(TRACE_MS, "class_code = 0x%x, device_type = 0x%x,"
					"sub_class = 0x%x\n", class_code,
						device_type, sub_class);

	memcpy(ms_card->raw_sys_info, buf + sys_info_addr, 96);
	memcpy(ms_card->raw_ms_id, ms_card->raw_sys_info + 64, 16);
	rts_trace(TRACE_MS, "MSPro ID:\n");
	RTS_DUMP(ms_card->raw_ms_id, 16);

#ifdef SUPPORT_PCGL_1P18
	memcpy(ms_card->raw_model_name, buf + model_name_addr, 48);
#endif

	cr_free_dma_buf(cr, buf);

	/* Confirm System Information */
#ifdef SUPPORT_MSXC
	if (CHK_MSXC(ms_card)) {
		if (class_code != 0x03) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else {
		if (class_code != 0x02) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
#else
	if (class_code != 0x02) {
		TRACE_RET(cr, STATUS_FAIL);
	}
#endif

	if (device_type != 0x00) {
		if ((device_type == 0x01) || (device_type == 0x02)
			|| (device_type == 0x03)) {
			set_card_flags(c, WP);
		} else {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	/* For MsPro Certification 2-5-3 */
	if (sub_class & 0xC0) {
		/* MsPro enter the "Read/Write-Protect Status",
		 * we return "card not exist" */
		TRACE_RET(cr, STATUS_FAIL);
	}

	rts_trace(TRACE_MS, "class_code: 0x%x, device_type: 0x%x,"
					"sub_class: 0x%x\n", class_code,
						device_type, sub_class);

#ifdef SUPPORT_MSXC
	if (CHK_MSXC(ms_card))
		c->capacity = xc_total_blk * xc_blk_size;
	else
		c->capacity = total_blk * blk_size;
#else
	c->capacity =  total_blk * blk_size;
#endif

	return STATUS_SUCCESS;
}

int ms_cleanup(struct card_reader *cr, struct card *c);
static int ms_msxc_chge_power(struct card_reader *cr,
						struct card *c, u8 mode)
{
	int retval;
	u8 buf[6];

	ms_cleanup(cr, c);

	/* Set Parameter Register */
	retval = ms_set_rw_reg_addr(cr, c, 0, 0,
						Pro_DataCount1, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	buf[0] = 0;
	buf[1] = mode;
	buf[2] = 0;
	buf[3] = 0;
	buf[4] = 0;
	buf[5] = 0;
	retval = ms_write_bytes(cr, c, PRO_WRITE_REG , 6,
						NO_WAIT_INT, buf, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Set "CHG_POWER" Command by "SET_CMD" TPC, Maybe try with
	 * "EX_SET_CMD" TPC. */
	retval = ms_send_cmd(cr, c, XC_CHG_POWER, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_get_int_status(cr, buf);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	if (buf[0] & (MS_COMM_INT_CMDNK | MS_COMM_INT_ERR)) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

/**
  * Set TPC Parameter Register to 00
  * It sets the data size of a READ_SHORT_DATA and a WRITE_SHORT_DATA TPC
  * as 32 bytes.
  */
static int ms_mg_set_tpc_para_sub(struct card_reader *cr,
				struct card *c, int type, u8 mg_entry_num)
{
	int retval;
	u8 buf[6];

	rts_trace(TRACE_MS, "<--\n");

	if (type == 0)
		retval = ms_set_rw_reg_addr(cr, c, 0, 0,
						Pro_TPCParm, 1);
	else
		retval = ms_set_rw_reg_addr(cr, c, 0, 0,
						Pro_DataCount1, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	buf[0] = 0;
	buf[1] = 0;
	if (type == 1) {
		buf[2] = 0;
		buf[3] = 0;
		buf[4] = 0;
		buf[5] = mg_entry_num;
	}
	retval = ms_write_bytes(cr, c, PRO_WRITE_REG,
			(type == 0) ? 1 : 6, NO_WAIT_INT, buf, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int ms_reset_mspro(struct card_reader *cr, struct card *c)
{
	struct ms_info *ms_card = c->extra;
	struct cr_option *option = &(cr->option);
	int ret = 0;
/* #ifdef XC_POWERCLASS */
	u8 change_power_class;

	rts_trace(TRACE_MS, "<--\n");

	if (option->ms_power_class_en & 0x02)
		change_power_class = 2;
	else if (option->ms_power_class_en & 0x01)
		change_power_class = 1;
	else
		change_power_class = 0;
/* #endif */

/* #ifdef XC_POWERCLASS */
Retry:
/* #endif */
	ret = ms_mspro_reset_flow(cr, c, 1);
	if (ret != STATUS_SUCCESS) {
		if (ms_card->switch_8bit_fail) {
			ret = ms_mspro_reset_flow(cr, c, 0);
			if (ret != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	ret = ms_read_attr_info(cr, c);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

/* #ifdef XC_POWERCLASS */
	/* If this card is HG8Bit card, we can't change power class */
	if (CHK_HG8BIT(ms_card))
		change_power_class = 0;

	if (change_power_class && CHK_MSXC(ms_card)) {
		u8 power_class_en = option->ms_power_class_en;

		rts_trace(TRACE_MS, "power_class_en = 0x%x\n", power_class_en);
		rts_trace(TRACE_MS, "change_power_class = %d\n",
							change_power_class);

		if (change_power_class)
			power_class_en &= (1 << (change_power_class - 1));
		else
			power_class_en = 0;

		if (power_class_en) {
			/* power_class_mode = (sub_class & 0x18) >> 3 */
			u8 power_class_mode = (ms_card->raw_sys_info[46] &
								0x18) >> 3;
			rts_trace(TRACE_MS, "power_class_mode = 0x%x",
							power_class_mode);
			if (change_power_class > power_class_mode)
				change_power_class = power_class_mode;
			if (change_power_class) {
				ret = ms_msxc_chge_power(cr, c,
							change_power_class);
				if (ret != STATUS_SUCCESS) {
					change_power_class--;
					goto Retry;
				}
			}
		}
	}
/* #endif */

/* #ifdef SUPPORT_MAGIC_GATE */
	/* Set READ/WRITE_SHORT_DATA length as 32 Bytes */
	ret = ms_mg_set_tpc_para_sub(cr, c, 0, 0);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
/* #endif */

	if (CHK_HG8BIT(ms_card))
		c->bus_width = 8;
	else
		c->bus_width = 4;

	rts_trace(TRACE_MS, "-->\n");

	return STATUS_SUCCESS;
}

static int ms_comm_read_sysinfo(struct card_reader *cr, u8 *sysinfo)
{
	u8 *ptr = NULL;
	int i = 0;

	cr_init_cmd(cr);
	for (i = 0; i < 96; i++)
		cr_add_cmd(cr, READ_REG_CMD, PPBUF_BASE2 + 0x1A0 + i, 0, 0);

	if (cr_send_cmd(cr, MS_CARD, 100))
		TRACE_RET(cr, STATUS_FAIL);

	ptr = cr_get_cmd_data(cr);
	memcpy(sysinfo, ptr, 96);

	return 0;
}
/*
 * ms_read_sysinfo -  Read MS system information as sys_info
 */
int ms_read_sysinfo(struct card_reader *cr, u8 *sysinfo)
{
	if (cr->ms_ops_prot->read_sysinfo)
		return cr->ms_ops_prot->read_sysinfo(cr, sysinfo);
	else
		return ms_comm_read_sysinfo(cr, sysinfo);
}
EXPORT_SYMBOL_GPL(ms_read_sysinfo);

static int ms_comm_read_block_contents(struct card_reader *cr, u8 **__ptr)
{
	u8 *ptr = NULL;
	u16 reg_addr;

	cr_init_cmd(cr);
	cr_add_cmd(cr, READ_REG_CMD, HEADER_ID0, 0, 0);
	cr_add_cmd(cr, READ_REG_CMD, HEADER_ID1, 0, 0);
	for (reg_addr = DISABLED_BLOCK0;
		reg_addr <= DISABLED_BLOCK3; reg_addr++)
		cr_add_cmd(cr, READ_REG_CMD, reg_addr, 0, 0);

	for (reg_addr = BLOCK_SIZE_0;
		reg_addr <= PAGE_SIZE_1; reg_addr++)
		cr_add_cmd(cr, READ_REG_CMD, reg_addr, 0, 0);
	cr_add_cmd(cr, READ_REG_CMD, MS_Device_Type, 0, 0);
	cr_add_cmd(cr, READ_REG_CMD, MS_4bit_Support, 0, 0);

	if (cr_send_cmd(cr, MS_CARD, 100))
		TRACE_RET(cr, STATUS_FAIL);

	ptr = cr_get_cmd_data(cr);
	*__ptr = ptr;

	return 0;
}

/*
 * ms_read_block_contents -  Read useful block contents
 *
 * @__ptr: points to command data
 */
int ms_read_block_contents(struct card_reader *cr, u8 **__ptr)
{
	if (cr->ms_ops_prot->read_block_contents)
		return cr->ms_ops_prot->read_block_contents(cr, __ptr);
	else
		return ms_comm_read_block_contents(cr, __ptr);
}
EXPORT_SYMBOL_GPL(ms_read_block_contents);

static int ms_comm_switch_intf_mode_4bit(struct card_reader *cr, struct card *c)
{

	if (ms_set_rw_reg_addr(cr, c, 0, 0, SystemParm, 1))
		TRACE_RET(cr, STATUS_FAIL);

	/* set Parallel IF. see spec. V1.4 TABLE5.7.10 */
	cr_wr_reg(cr, PPBUF_BASE2, 0xFF, 0x88);
	cr_wr_reg(cr, PPBUF_BASE2 + 1, 0xFF, 0);
	if (ms_transfer_tpc(cr, c, MS_TM_WRITE_BYTES,
				WRITE_REG , 1, NO_WAIT_INT))
		TRACE_RET(cr, STATUS_FAIL);

	/* No check INT if the card is 4-bits */
	cr_wr_reg(cr, REG_MS_CFG, 0x58 | MS_NO_CHECK_INT,
			MS_BUS_WIDTH_4 | PUSH_TIME_ODD | MS_NO_CHECK_INT);
	return 0;
}
/*
 * ms_switch_intf_mode_4bit - Switch bus width of both card reader controller
 * and card to 4 bits
 *
 * TODO: the func can be used to switch but width to 1bit/4bit
 */
int ms_switch_intf_mode_4bit(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->switch_intf_mode_4bit)
		return cr->ms_ops_prot->switch_intf_mode_4bit(cr, c);
	else
		return ms_comm_switch_intf_mode_4bit(cr, c);
}
EXPORT_SYMBOL_GPL(ms_switch_intf_mode_4bit);

static int ms_comm_reset_ms(struct card_reader *cr, struct card *c)
{
	struct ms_info *ms_card = c->extra;
	int retval;
	u16 i, /*reg_addr,*/ block_size;
	u8 val, extra[MS_COMM_EXTRA_SIZE], j, *ptr;
#ifndef SUPPORT_MAGIC_GATE
	u16 eblock_cnt;
#endif

	rts_trace(TRACE_MS, "<--\n");

	retval = ms_prepare_reset(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_card->ms_type |= TYPE_MS;

	retval = ms_send_cmd(cr, c, MS_RESET, NO_WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* detect write protect status */
	retval = ms_read_status_reg(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	retval = ms_read_regs_to_buf(cr, &val, 0, 1);
	if (retval)
		return STATUS_FAIL;
	if (val & MS_COMM_WRT_PRTCT)
		set_card_flags(c, WP);
	else
		clr_card_flags(c, WP);

	i = 0;

RE_SEARCH:
	/* Search For Boot Block */
	while (i < (MAX_DEFECTIVE_BLOCK + 2)) {
		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = ms_read_extra_data(cr, c, i, 0, extra,
							MS_COMM_EXTRA_SIZE);
		if (retval != STATUS_SUCCESS) {
			i++;
			continue;
		}

		if (extra[0] & BLOCK_OK) {
			if (!(extra[1] & NOT_BOOT_BLOCK)) {
				ms_card->boot_block = i;
				break;
			}
		}
		i++;
	}
	if (i == (MAX_DEFECTIVE_BLOCK + 2)) {
		rts_trace(TRACE_MS, "No boot block found!");
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* for MS chcek Procedure TEST7009 (7-9-5)
	 * Each Boot Block has 3 meaningful pages(p0, p1 ans p2),
	 * check if they are uncorrectable */
	for (j = 0; j < 3; j++) {
		retval = ms_comm_read_page(cr, c, ms_card->boot_block, j);
		if (retval != STATUS_SUCCESS) {
			/* use MS_FLASH_WRITE_ERROR to indicate
			 * uncorrectable ECC error */
			if (ms_check_error_code(cr, c,
				MS_FLASH_WRITE_ERROR)) {
				/* due to g_byBootBlockNum occur uncorrectable
				 * ecc error, try to find the 2nd BootBlock if
				 * the 2nd BootBlock is also occur uncorrectable
				 * ecc error, return 0 to PC else, we should use
				 * the data of 2nd Bootblock */
				i = ms_card->boot_block + 1;
				ms_set_error_code(cr, c, MS_NO_ERROR);
				rts_trace(TRACE_MS, "%d\n", __LINE__);
				goto RE_SEARCH;
			}
		}
	}

	/* Read boot block contents */
	retval = ms_comm_read_page(cr, c, ms_card->boot_block, 0);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Read MS system information as sys_info */
	retval = ms_read_sysinfo(cr, ms_card->raw_sys_info);
	if (retval < 0) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* memcpy(ms_card->raw_sys_info, ptr, 96); */
	memcpy(ms_card->raw_ms_id, ms_card->raw_sys_info + 0x0C, 16);
	rts_trace(TRACE_MS, "MS ID:\n");
	RTS_DUMP(ms_card->raw_ms_id, 16);

	/* Read useful block contents */
	retval = ms_read_block_contents(cr, &ptr);
	if (retval < 0) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	rts_trace(TRACE_MS, "Boot block data:\n");
	RTS_DUMP(ptr, 16);

	/* Block ID error */
	/* HEADER_ID0, HEADER_ID1 */
	if (ptr[0] != 0x00 || ptr[1] != 0x01) {
		i = ms_card->boot_block + 1;
		rts_trace(TRACE_MS, "%d\n", __LINE__);
		goto RE_SEARCH;
	}

	/* Page size error */
	/* PAGE_SIZE_0, PAGE_SIZE_1 */
	if (ptr[12] != 0x02 || ptr[13] != 0x00) {
		i = ms_card->boot_block + 1;
		rts_trace(TRACE_MS, "%d\n", __LINE__);
		goto RE_SEARCH;
	}

	/* for MS check procedure,if device type is ROM1 or ROM3,
	 * treat as Read only stick MS_Device_Type */
	if ((ptr[14] == 1) || (ptr[14] == 3))
		set_card_flags(c, WP);

	/* BLOCK_SIZE_0, BLOCK_SIZE_1 */
	block_size = ((u16)ptr[6] << 8) | ptr[7];
	if (block_size == 0x0010) {
		/* Block size 16KB */
		ms_card->block_shift = 5;
		ms_card->page_off = 0x1F;
	} else if (block_size == 0x0008) {
		/* Block size 8KB */
		ms_card->block_shift = 4;
		ms_card->page_off = 0x0F;
	}

	/* BLOCK_COUNT_0, BLOCK_COUNT_1 */
	ms_card->total_block = ((u16)ptr[8] << 8) | ptr[9];

#ifdef SUPPORT_MAGIC_GATE
	/* According to PCGL, the Number of blocks of Memory Stick classic
	 * series media should be fixed as follows.  Shink certain block size in
	 * order to match the volume capacity by calculating CHS addressing and
	 * that by calculating LBA addressing, S capacity is smaller than LBA
	 * capacity in case of 64MB KB == 2 Erase Blocks smaller) and 128MB(96KB
	 * == 6 Erase Blocks smaller). */
	j = ptr[10]; /* Higher byte of effective block */

	if (ms_card->block_shift == 4)	{ /* 4MB or 8MB */
		if (j < 2) /* Effective block for 4MB: 0x1F0 */
			c->capacity = 0x1EE0;
		else /* Effective block for 8MB: 0x3E0 */
			c->capacity = 0x3DE0;
	} else	{ /* 16MB, 32MB, 64MB or 128MB */
		if (j < 5) /* Effective block for 16MB: 0x3E0 */
			c->capacity = 0x7BC0;
		else if (j < 0xA) /* Effective block for 32MB: 0x7C0 */
			c->capacity = 0xF7C0;
		else if (j < 0x11) /* Effective block for 64MB: 0xF80 */
			c->capacity = 0x1EF80;
		else /* Effective block for 128MB: 0x1F00 */
			c->capacity = 0x3DF00;
	}
#else
	/* EBLOCK_COUNT_0, EBLOCK_COUNT_1 */
	eblock_cnt = ((u16)ptr[10] << 8) | ptr[11];
	c->capacity = ((u32)eblock_cnt - 2) << ms_card->block_shift;
#endif

	/* Switch I/F Mode, MS_4bit_Support */
	if (ptr[15]) {
		retval = ms_switch_intf_mode_4bit(cr, c);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		ms_card->ms_type |= MS_4BIT;
	}

	if (CHK_MS4BIT(ms_card))
		c->bus_width = 4;
	else
		c->bus_width = 1;

	rts_trace(TRACE_MS, "card capacity=%d\n", c->capacity);
	rts_trace(TRACE_MS, "-->\n");

	return STATUS_SUCCESS;
}


static int ms_comm_init_l2p_tbl(struct card_reader *cr, struct card *c)
{
	struct ms_info *ms_card = c->extra;
	int size, i, seg_no, retval;
	u16 defect_block, reg_addr;
	u8 val1, val2;

	ms_card->segment_cnt = ms_card->total_block >> 9;
	rts_trace(TRACE_MS, "ms_card->segment_cnt = %d\n",
		ms_card->segment_cnt);

	size = ms_card->segment_cnt * sizeof(struct zone_entry);
	ms_card->segment = (struct zone_entry *)vmalloc(size);
	if (ms_card->segment == NULL) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	memset(ms_card->segment, 0, size);
	retval = ms_comm_read_page(cr, c, ms_card->boot_block, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_GOTO(cr, INIT_FAIL);
	}
	/* reg_addr = PPBUF_BASE2; */
	reg_addr = 0; /* the offset to PPBUF_BASE2 is set to 0 */
	for (i = 0; i < (((ms_card->total_block >> 9) * 10)+1); i++) {
		/* max defect count + information block */
		/* retval = cr_rd_reg(cr, reg_addr++, &val1); */
		retval = ms_read_regs_to_buf(cr, &val1, reg_addr++, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_GOTO(cr, INIT_FAIL);
		}
		/* retval = cr_rd_reg(cr, reg_addr++, &val2); */
		retval = ms_read_regs_to_buf(cr, &val2, reg_addr++, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_GOTO(cr, INIT_FAIL);
		}

		defect_block = ((u16)val1 << 8) | val2;
		if (defect_block == 0xFFFF)
			break;
		seg_no = defect_block / 512;
		ms_card->segment[seg_no].defect_list[ms_card->segment[seg_no].disable_count++] = defect_block;
	}

	for (i = 0; i < ms_card->segment_cnt; i++) {
		ms_card->segment[i].build_flag = 0;
		ms_card->segment[i].l2p_table = NULL;
		ms_card->segment[i].free_table = NULL;
		ms_card->segment[i].get_index = 0;
		ms_card->segment[i].set_index = 0;
		ms_card->segment[i].unused_blk_cnt = 0;

		rts_trace(TRACE_MS,
			"defective block count of segment %d is %d\n",
			i, ms_card->segment[i].disable_count);
	}

	return STATUS_SUCCESS;

INIT_FAIL:
	if (ms_card->segment) {
		vfree(ms_card->segment);
		ms_card->segment = NULL;
	}

	return STATUS_FAIL;
}

static u16 ms_comm_get_l2p_tbl(struct card_reader *cr, struct card *c,
						int seg_no, u16 log_off)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;

	if (ms_card->segment == NULL)
		return 0xFFFF;

	segment = &(ms_card->segment[seg_no]);

	if (segment->l2p_table)
		return segment->l2p_table[log_off];

	return 0xFFFF;
}

static void ms_comm_set_l2p_tbl(struct card_reader *cr, struct card *c,
					int seg_no, u16 log_off, u16 phy_blk)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;

	if (ms_card->segment == NULL)
		return;

	segment = &(ms_card->segment[seg_no]);
	if (segment->l2p_table)
		segment->l2p_table[log_off] = phy_blk;
}

static void ms_comm_set_unused_block(struct card_reader *cr,
					struct card *c, u16 phy_blk)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;
	int seg_no;

	seg_no = (int)phy_blk >> 9;
	segment = &(ms_card->segment[seg_no]);
	segment->free_table[segment->set_index++] = phy_blk;
	if (segment->set_index >= MS_FREE_TABLE_CNT)
		segment->set_index = 0;
	segment->unused_blk_cnt++;
}

static u16 ms_comm_get_unused_block(struct card_reader *cr,
						struct card *c, int seg_no)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;
	u16 phy_blk;

	segment = &(ms_card->segment[seg_no]);
	if (segment->unused_blk_cnt <= 0)
		return 0xFFFF;

	phy_blk = segment->free_table[segment->get_index];
	segment->free_table[segment->get_index++] = 0xFFFF;
	if (segment->get_index >= MS_FREE_TABLE_CNT)
		segment->get_index = 0;
	segment->unused_blk_cnt--;

	return phy_blk;
}

const unsigned short ms_start_idx[MS_COMM_IDX_LEN] = {0, 494, 990, 1486, 1982,
					2478, 2974, 3470, 3966, 4462, 4958,
					5454, 5950, 6446, 6942, 7438, 7934};
EXPORT_SYMBOL_GPL(ms_start_idx);

static int ms_comm_arbitrate_l2p(struct card_reader *cr, struct card *c,
				u16 phy_blk, u16 log_off, u8 us1, u8 us2)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;
	int seg_no;
	u16 tmp_blk;

	seg_no = (int)phy_blk >> 9;
	segment = &(ms_card->segment[seg_no]);
	tmp_blk = segment->l2p_table[log_off];

	if (us1 != us2) {
		if (us1 == 0) {
			if (!check_card_flag(c, WP))
				ms_erase_block(cr, c, tmp_blk);
			ms_comm_set_unused_block(cr, c, tmp_blk);
			segment->l2p_table[log_off] = phy_blk;
		} else {
			if (!check_card_flag(c, WP))
				ms_erase_block(cr, c, phy_blk);
			ms_comm_set_unused_block(cr, c, phy_blk);
		}
	} else {
		if (phy_blk < tmp_blk) {
			if (!check_card_flag(c, WP))
				ms_erase_block(cr, c, phy_blk);
			ms_comm_set_unused_block(cr, c, phy_blk);
		} else {
			if (!check_card_flag(c, WP))
				ms_erase_block(cr, c, tmp_blk);
			ms_comm_set_unused_block(cr, c, tmp_blk);
			segment->l2p_table[log_off] = phy_blk;
		}
	}

	return STATUS_SUCCESS;
}

static int ms_comm_build_l2p_tbl(struct card_reader *cr, struct card *c,
								int seg_no)
{
	struct ms_info *ms_card = c->extra;
	struct zone_entry *segment;
	int retval, table_size, disable_cnt, defect_flag, i;
	u16 start, end, phy_blk, log_blk, tmp_blk;
	u8 extra[MS_COMM_EXTRA_SIZE], us1, us2;

	rts_trace(TRACE_MS, "<--\n");

	rts_trace(TRACE_MS, "ms_build_l2p_tbl: %d\n", seg_no);

	if (ms_card->segment == NULL) {
		retval = ms_comm_init_l2p_tbl(cr, c);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, retval);
		}
	}

	if (ms_card->segment[seg_no].build_flag) {
		rts_trace(TRACE_MS, "l2p table of segment %d has been built\n",
								seg_no);
		return STATUS_SUCCESS;
	}

	if (seg_no == 0)
		table_size = 494;
	else
		table_size = 496;

	segment = &(ms_card->segment[seg_no]);
	if (segment->l2p_table == NULL) {
		segment->l2p_table = (u16 *)vmalloc(table_size * 2);
		if (segment->l2p_table == NULL) {
			TRACE_GOTO(cr, BUILD_FAIL);
		}
	}
	memset((u8 *)(segment->l2p_table), 0xff, table_size * 2);

	if (segment->free_table == NULL) {
		segment->free_table = (u16 *)vmalloc(MS_FREE_TABLE_CNT * 2);
		if (segment->free_table == NULL) {
			TRACE_GOTO(cr, BUILD_FAIL);
		}
	}
	memset((u8 *)(segment->free_table), 0xff, MS_FREE_TABLE_CNT * 2);

	start = (u16)seg_no << 9;
	end = (u16)(seg_no + 1) << 9;

	disable_cnt = segment->disable_count;

	segment->get_index = segment->set_index = 0;
	segment->unused_blk_cnt = 0;

	for (phy_blk = start; phy_blk < end; phy_blk++) {
		if (disable_cnt) {
			defect_flag = 0;
			for (i = 0; i < segment->disable_count; i++) {
				if (phy_blk == segment->defect_list[i]) {
					defect_flag = 1;
					break;
				}
			}
			if (defect_flag) {
				disable_cnt--;
				continue;
			}
		}

		/* rts_trace(TRACE_MS, "==> phy_blk = 0x%x\n", phy_blk)); */
		retval = ms_read_extra_data(cr, c, phy_blk,
					0, extra, MS_COMM_EXTRA_SIZE);
		if (retval != STATUS_SUCCESS) {
			rts_trace(TRACE_MS, "read extra data fail\n");
			ms_set_bad_block(cr, c, phy_blk);
			continue;
		}

		/* rts_trace(TRACE_MS, "extra[0] = 0x%x\n", extra[0]); */
		/* rts_trace(TRACE_MS, "extra[1] = 0x%x\n", extra[1]); */
		if (seg_no == ms_card->segment_cnt - 1) {
			if (!(extra[1] & NOT_TRANSLATION_TABLE)) {
				if (!check_card_flag(c, WP)) {
					retval = ms_erase_block(cr, c,
								phy_blk);
					if (retval != STATUS_SUCCESS)
						/* This block is a bad block,
						 * we continue to the next */
						continue;
					/* make sure that this block
					 * would be put to free table */
					extra[2] = 0xff;
					extra[3] = 0xff;
				}
			}
		}

		if (!(extra[0] & BLOCK_OK))
			continue;
		if (!(extra[1] & NOT_BOOT_BLOCK))
			continue;
		if ((extra[0] & PAGE_OK) != PAGE_OK)
			continue;

		log_blk = ((u16)extra[2] << 8) | extra[3];

		/* rts_trace(TRACE_MS, "log_blk = 0x%x\n", log_blk)); */

		if (log_blk == 0xFFFF) {
			if (!check_card_flag(c, WP)) {
				retval = ms_erase_block(cr, c, phy_blk);
				if (retval != STATUS_SUCCESS)
					continue;
			}
			ms_comm_set_unused_block(cr, c, phy_blk);
			continue;
		}

		if ((log_blk < ms_start_idx[seg_no]) ||
				(log_blk >= ms_start_idx[seg_no+1])) {
			if (!check_card_flag(c, WP)) {
				retval = ms_erase_block(cr, c, phy_blk);
				if (retval != STATUS_SUCCESS)
					continue;
			}
			ms_comm_set_unused_block(cr, c, phy_blk);
			continue;
		}

		if (segment->l2p_table[log_blk - ms_start_idx[seg_no]] ==
			0xFFFF) {
			segment->l2p_table[log_blk - ms_start_idx[seg_no]] =
									phy_blk;
			continue;
		}

		us1 = extra[0] & 0x10;
		tmp_blk = segment->l2p_table[log_blk - ms_start_idx[seg_no]];
		retval = ms_read_extra_data(cr, c, tmp_blk, 0,
						extra, MS_COMM_EXTRA_SIZE);
		if (retval != STATUS_SUCCESS)
			continue;
		us2 = extra[0] & 0x10;
		(void)ms_comm_arbitrate_l2p(cr, c, phy_blk,
				log_blk-ms_start_idx[seg_no], us1, us2);
		continue;
	}

	segment->build_flag = 1;

	rts_trace(TRACE_MS, "unused block count: %d\n",
		segment->unused_blk_cnt);

	/* Logical Address Confirmation Process */
	if (seg_no == ms_card->segment_cnt - 1) {
		if (segment->unused_blk_cnt < 2)
			set_card_flags(c, WP);
	} else {
		if (segment->unused_blk_cnt < 1)
			set_card_flags(c, WP);
	}

	if (check_card_flag(c, WP))
		return STATUS_SUCCESS;

	for (log_blk = ms_start_idx[seg_no];
		log_blk < ms_start_idx[seg_no + 1]; log_blk++) {
		if (segment->l2p_table[log_blk-ms_start_idx[seg_no]] ==
			0xFFFF) {
			phy_blk = ms_comm_get_unused_block(cr, c, seg_no);
			if (phy_blk == 0xFFFF) {
				set_card_flags(c, WP);
				return STATUS_SUCCESS;
			}
			retval = ms_init_page(cr, c, phy_blk,
							log_blk, 0, 1);
			if (retval != STATUS_SUCCESS) {
				TRACE_GOTO(cr, BUILD_FAIL);
			}
			segment->l2p_table[log_blk-ms_start_idx[seg_no]] =
								phy_blk;
			if (seg_no == ms_card->segment_cnt - 1) {
				if (segment->unused_blk_cnt < 2) {
					set_card_flags(c, WP);
					return STATUS_SUCCESS;
				}
			} else {
				if (segment->unused_blk_cnt < 1) {
					set_card_flags(c, WP);
					return STATUS_SUCCESS;
				}
			}
		}
	}

	/* Make boot block be the first normal block */
	if (seg_no == 0) {
		for (log_blk = 0; log_blk < 494; log_blk++) {
			tmp_blk = segment->l2p_table[log_blk];
			if (tmp_blk < ms_card->boot_block) {
				rts_trace(TRACE_MS,
					"Boot block is not the first normal block.\n");
				if (check_card_flag(c, WP))
					break;
				phy_blk = ms_comm_get_unused_block(cr, c, 0);
				retval = ms_copy_page(cr, c, tmp_blk,
							phy_blk, log_blk, 0,
							ms_card->page_off + 1);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}

				segment->l2p_table[log_blk] = phy_blk;
				retval = ms_set_bad_block(cr, c, tmp_blk);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}
			}
		}
	}
#if 0
	rts_trace(TRACE_MS, "L2P table:\n");
	for (i = 0; i < table_size; i++) {
		rts_trace(TRACE_MS, "%03x : %03x\n", i, segment->l2p_table[i]);
	}
#endif
	rts_trace(TRACE_MS, "-->\n");

	return STATUS_SUCCESS;

BUILD_FAIL:
	segment->build_flag = 0;
	if (segment->l2p_table) {
		vfree(segment->l2p_table);
		segment->l2p_table = NULL;
	}
	if (segment->free_table) {
		vfree(segment->free_table);
		segment->free_table = NULL;
	}

	return STATUS_FAIL;
}

static int ms_comm_reset(struct card_reader *cr, struct card *c)
{
	struct ms_info *ms_card = c->extra;
	int ret = 0;

	rts_trace(TRACE_MS, "<--\n");

	memset(ms_card, 0, sizeof(struct ms_info));

	if (c->reset_flags & RESET_MEET_OCP)
		return -1;

	ret = ms_prepare_init(cr, c);
	if (ret)
		return STATUS_FAIL;

	ms_card->ms_type = 0;
	ret = ms_reset_mspro(cr, c);
	if (ret != STATUS_SUCCESS) {
		if (ms_card->check_ms_flow) {
			ret = ms_comm_reset_ms(cr, c);
			if (ret != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	ret = ms_set_init_para(cr, c);
	if (ret != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (!CHK_MSPRO(ms_card)) {
		/* Build table for the last segment,
		 * to check if L2P talbe block exist,erasing it */
		ret = ms_comm_build_l2p_tbl(cr, c,
					ms_card->total_block / 512 - 1);
		if (ret != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	rts_trace(TRACE_MS, "ms_card->ms_type = 0x%x\n", ms_card->ms_type);

	return ret;
}

int ms_reset(struct card_reader *cr, struct card *c)
{
	u32 card_cap = cr_get_card_spt_map_sw(cr);

	if (!(card_cap & (1 << MS_ID)))
		return -1;

	if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
		rts_trace(TRACE_MS, "MS card not exist\n");
		ms_set_error_code(cr, c, MS_NO_CARD);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (cr->ms_ops_prot->reset)
		return cr->ms_ops_prot->reset(cr, c);
	else
		return ms_comm_reset(cr, c);
}
EXPORT_SYMBOL_GPL(ms_reset);

static int ms_comm_disable(struct card_reader *cr, struct card *c)
{
	/* TODO: Implement */
#if 0
	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, CARD_OE, MS_OUTPUT_EN, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, CARD_PWR_CTL, POWER_MASK,
								POWER_OFF);
	cr_add_cmd(cr, WRITE_REG_CMD, CARD_CLK_EN, MS_CLK_EN, 0);
	cr_send_cmd(cr, MODE_C, 100);
#endif
	return -1;
}

int ms_disable(struct card_reader *cr, struct card *c)
{

	if (cr->ms_ops_prot->disable)
		return cr->ms_ops_prot->disable(cr, c);
	else
		return ms_comm_disable(cr, c);
}
EXPORT_SYMBOL_GPL(ms_disable);

int ms_mspro_set_rw_cmd(struct card_reader *cr,
			struct card *c,	u32 start_sec, u16 sec_cnt, u8 cmd)
{
	int retval, i;
	u8 data[8];

	data[0] = cmd;
	data[1] = (u8)(sec_cnt >> 8);
	data[2] = (u8)sec_cnt;
	data[3] = (u8)(start_sec >> 24);
	data[4] = (u8)(start_sec >> 16);
	data[5] = (u8)(start_sec >> 8);
	data[6] = (u8)start_sec;
	data[7] = 0;

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, PRO_EX_SET_CMD, 7,
						WAIT_INT, data, 8);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(ms_mspro_set_rw_cmd);

static int ms_mspro_rw_multi_sector(struct scsi_cmnd *srb,
				struct card_reader *cr, struct card *c,
					u32 start_sector, u16 sector_cnt)
{
	struct ms_info *ms_card = c->extra;
	int retval, mode_2k = 0;
	u16 count;
	u8 val, trans_mode, rw_tpc, rw_cmd;


	ms_set_error_code(cr, c, MS_NO_ERROR);

	ms_card->cleanup_counter = 0;

	if (CHK_MSHG(ms_card)) {
		if ((start_sector % 4) || (sector_cnt % 4)) {
			if (srb->sc_data_direction == DMA_FROM_DEVICE) {
				rw_tpc = PRO_READ_LONG_DATA;
				rw_cmd = PRO_READ_DATA;
			} else {
				rw_tpc = PRO_WRITE_LONG_DATA;
				rw_cmd = PRO_WRITE_DATA;
			}
		} else {
			if (srb->sc_data_direction == DMA_FROM_DEVICE) {
				rw_tpc = PRO_READ_QUAD_DATA;
				rw_cmd = PRO_READ_2K_DATA;
			} else {
				rw_tpc = PRO_WRITE_QUAD_DATA;
				rw_cmd = PRO_WRITE_2K_DATA;
			}
			mode_2k = 1;
		}
	} else {
		if (srb->sc_data_direction == DMA_FROM_DEVICE) {
			rw_tpc = PRO_READ_LONG_DATA;
			rw_cmd = PRO_READ_DATA;
		} else {
			rw_tpc = PRO_WRITE_LONG_DATA;
			rw_cmd = PRO_WRITE_DATA;
		}
	}
	retval = ms_switch_clock(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (srb->sc_data_direction == DMA_FROM_DEVICE) {
		/* trans_mode = MS_TM_AUTO_READ; */
		trans_mode = MS_TM_AUTO_READ;
	} else {
		/* trans_mode = MS_TM_AUTO_WRITE; */
		trans_mode = MS_TM_AUTO_WRITE;
	}
	retval = ms_get_int_status(cr, &val);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	if (ms_card->seq_mode) {
		if ((ms_card->prev_dir != srb->sc_data_direction) ||
			((ms_card->prev_sec_addr + ms_card->prev_sec_cnt) !=
			start_sector) || (mode_2k &&
			(ms_card->seq_mode & MODE_512_SEQ)) ||
			(!mode_2k &&
			(ms_card->seq_mode & MODE_2K_SEQ)) ||
			!(val & MS_COMM_INT_BREQ) ||
			((ms_card->total_sec_cnt + sector_cnt) > 0xFE00)) {
			ms_card->seq_mode = 0;
			ms_card->total_sec_cnt = 0;
			if (val & MS_COMM_INT_BREQ) {
				retval = ms_send_cmd(cr, c,
						PRO_STOP, WAIT_INT);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}
				cr_flush_rb(cr);
			}
		}
	}

	if (!ms_card->seq_mode) {
		ms_card->total_sec_cnt = 0;
		if (sector_cnt >= MS_COMM_SEQ_START_CRITERIA) {
			if ((c->capacity - start_sector) > 0xFE00) {
				count = 0xFE00;
			} else {
				count = (u16)(c->capacity - start_sector);
			}
			if (count > sector_cnt) {
				if (mode_2k)
					ms_card->seq_mode |= MODE_2K_SEQ;
				else
					ms_card->seq_mode |= MODE_512_SEQ;
			}
		} else {
			count = sector_cnt;
		}
		retval = ms_mspro_set_rw_cmd(cr, c, start_sector,
							count, rw_cmd);
		if (retval != STATUS_SUCCESS) {
			ms_card->seq_mode = 0;
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	retval = ms_transfer_data(cr, c, trans_mode, rw_tpc,
					sector_cnt, WAIT_INT, mode_2k,
					scsi_sg_count(srb), scsi_sglist(srb),
					scsi_bufflen(srb));
	if (retval != STATUS_SUCCESS) {
		ms_card->seq_mode = 0;
		ms_get_int_status(cr, &val);
		ms_clear_error(cr);
		/* In card rw, if card unplugged surprinsingly, don't need to
		 * retry rw */
		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			c->rw_need_retry = 0;
			rts_trace(TRACE_MS, "No card exist,"
					"exit mspro_rw_multi_sector\n");
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (val & MS_COMM_INT_BREQ)
			ms_send_cmd(cr, c, PRO_STOP, WAIT_INT);
		if (val & (MS_COMM_CRC16_ERR | MS_COMM_RDY_TIMEOUT)) {
			rts_trace(TRACE_MS, "MSPro CRC error, tune clock!\n");
			c->rw_need_retry = 1;
			ms_auto_tune_clock(cr, c);
		}
		TRACE_RET(cr, retval);
	}

	if (ms_card->seq_mode) {
		ms_card->prev_sec_addr = start_sector;
		ms_card->prev_sec_cnt = sector_cnt;
		ms_card->prev_dir = srb->sc_data_direction;
		ms_card->total_sec_cnt += sector_cnt;
	}

	return STATUS_SUCCESS;
}

static int ms_comm_read_page_data(struct card_reader *cr, struct card *c,
		struct scsi_cmnd *srb, u8 __cfg, u8 *buf, unsigned int **index,
							unsigned int *offset)
{
	struct cr_option *option = &(cr->option);
	u8 cfg = __cfg;


	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, READ_PAGE_DATA);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, cfg);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE, 0x01, RING_BUFFER);
	cr_enable_dma(cr, 512, DMA_512, DMA_FROM_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER,
			0xFF, REG_MS_TRANSFER_START | MS_TM_NORMAL_READ);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
				REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);
	cr_send_cmd_no_wait(cr);

	return cr_transfer_data_partial(cr, MS_CARD,
			buf, 512, scsi_sg_count(srb), *index, offset,
			DMA_FROM_DEVICE, option->ms_timeout);
}


/*
 * ms_read_page_data
 *
 * To unify infterfaces for different transport mechanisms, we have to adopt
 * pointer(index), which points to a pointer that points to a unsigned int
 * variable.
 *
 * NOTE: pay special attention to the intialization to parameter "index".
 */
int ms_read_page_data(struct card_reader *cr, struct card *c,
			struct scsi_cmnd *srb, u8 trans_cfg,
			u8 *buf, unsigned int **index, unsigned int *offset)
{
	if (cr->ms_ops_prot->read_page_data) {
		return cr->ms_ops_prot->read_page_data(cr, c, srb,
					trans_cfg, buf, index, offset);
	} else {
		return ms_comm_read_page_data(cr, c, srb,
					trans_cfg, buf, index, offset);
	}
}
EXPORT_SYMBOL_GPL(ms_read_page_data);

static int ms_comm_write_page_data(struct card_reader *cr, struct card *c,
			struct scsi_cmnd *srb, u8 *buf, unsigned int **index,
							unsigned int *offset)
{
	struct cr_option *option = &(cr->option);

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TPC, 0xFF, WRITE_PAGE_DATA);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANS_CFG, 0xFF, WAIT_INT);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE, 0x01, RING_BUFFER);
	cr_enable_dma(cr, 512, DMA_512, DMA_TO_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_MS_TRANSFER,
			0xFF, REG_MS_TRANSFER_START | MS_TM_NORMAL_WRITE);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_MS_TRANSFER,
				REG_MS_TRANSFER_END, REG_MS_TRANSFER_END);
	cr_send_cmd_no_wait(cr);

	return cr_transfer_data_partial(cr, MS_CARD,
			buf, 512, scsi_sg_count(srb), *index, offset,
				DMA_TO_DEVICE, option->ms_timeout);
}

/*
 * ms_write_page_data
 *
 * To unify infterfaces for different transport mechanisms, we have to adopt
 * pointer(index), which points to a pointer that points to a unsigned int
 * variable.
 *
 * NOTE: pay special attention to the intialization to parameter "index".
 */
int ms_write_page_data(struct card_reader *cr, struct card *c,
			struct scsi_cmnd *srb, u8 *buf, unsigned int **index,
							unsigned int *offset)
{
	if (cr->ms_ops_prot->write_page_data) {
		return cr->ms_ops_prot->write_page_data(cr, c, srb,
							buf, index, offset);
	} else {
		return ms_comm_write_page_data(cr, c, srb,
							buf, index, offset);
	}
}
EXPORT_SYMBOL_GPL(ms_write_page_data);

int mspro_read_format_progress(struct card_reader *cr,
				struct card *card, const int short_data_len)
{
	struct ms_info *ms_card = card->extra;
	int i;
	u32 total_progress, cur_progress;
	u8 cnt, tmp;
	u8 data[8];

	rts_trace(TRACE_MS, "mspro_read_format_progress, short_data_len = %d\n",
								short_data_len);

	if (ms_switch_clock(cr, card)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (cr_rd_reg(cr, REG_MS_TRANS_CFG, &tmp)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (!(tmp & MS_INT_BREQ)) {
		if ((tmp &  (MS_INT_CED | MS_INT_BREQ |
			MS_INT_CMDNK | MS_INT_ERR)) ==
			MS_INT_CED) {
			ms_card->format_status = FORMAT_SUCCESS;
			return STATUS_SUCCESS;
		}
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (short_data_len >= 256)
		cnt = 0;
	else
		cnt = (u8)short_data_len;
	if (cr_wr_reg(cr, REG_MS_CFG, MS_NO_CHECK_INT, MS_NO_CHECK_INT)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (ms_read_bytes(cr, card, PRO_READ_SHORT_DATA,
					cnt, WAIT_INT, data, 8)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	total_progress = (data[0] << 24) | (data[1] << 16) |
						(data[2] << 8) | data[3];
	cur_progress = (data[4] << 24) | (data[5] << 16) |
						(data[6] << 8) | data[7];
	rts_trace(TRACE_MS, "total_progress = %d, cur_progress = %d\n",
						total_progress, cur_progress);
	if (total_progress == 0) {
		ms_card->progress = 0;
		card->format_progress = 0;
	} else {
		u64 ulltmp = (u64)cur_progress * (u64)65535;
		do_div(ulltmp, total_progress);
		ms_card->progress = (u16)ulltmp;
		card->format_progress = (u16)ulltmp;
	}

	rts_trace(TRACE_MS, "progress = %d\n", ms_card->progress);
	rts_trace(TRACE_MS, "card->format_progress = %d\n",
		card->format_progress);

	/* H/W READ_BYTES mode can't support "WAIT_INT" function,
	 * so we should polling to wait INT here
	 */
	for (i = 0; i < 5000; i++) {
		if (cr_rd_reg(cr, REG_MS_TRANS_CFG, &tmp)) {
			ms_card->format_status = FORMAT_FAIL;
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (tmp & (MS_INT_CED | MS_INT_CMDNK |
				MS_INT_BREQ | MS_INT_ERR))
			break;
		msleep(1);
	}

	if (cr_wr_reg(cr, REG_MS_CFG, MS_NO_CHECK_INT, 0)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (i == 5000) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (tmp & (MS_INT_CMDNK | MS_INT_ERR)) {
		ms_card->format_status = FORMAT_FAIL;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (tmp & MS_INT_CED) {
		ms_card->format_status = FORMAT_SUCCESS;
		ms_card->pro_under_formatting = 0;
	} else if (tmp & MS_INT_BREQ) {
		ms_card->format_status = FORMAT_IN_PROGRESS;
	} else {
		ms_card->format_status = FORMAT_FAIL;
		ms_card->pro_under_formatting = 0;
		TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(mspro_read_format_progress);


static void mspro_comm_polling_format_status(struct card_reader *cr,
							struct card *card)
{
	struct ms_info *ms_card = card->extra;
	int i;

	if (ms_card->pro_under_formatting &&
			(rtsx_get_stat(cr) != RTSX_STAT_SS)) {
		for (i = 0; i < 65535; i++) {
			mspro_read_format_progress(cr, card, MS_SHORT_DATA_LEN);
			if (ms_card->format_status != FORMAT_IN_PROGRESS)
				break;
		}
	}
}


/*
 * ms_polling_format_progress -- Polling for MSPro card format progress
 */
void ms_polling_format_progress(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->polling_format_progress)
		cr->ms_ops_prot->polling_format_progress(cr, c);
	else
		mspro_comm_polling_format_status(cr, c);
}
EXPORT_SYMBOL_GPL(ms_polling_format_progress);

/*
 * ms_read_multiple_pages
 *
 * NOTE: pay special attention for the assignment of parameter "unsigned int
 * **index" if the transport type is PCIe, the parameter should be assigned with
 * address for the pointer which points to variable with type of unsigned int,
 * eg, "index", which MUST be assigned with "0", but, if the transport layer
 * adopts USB protocol, the parameter should be assigned with address of a
 * pointer, which is initialized with NULL.
 */
static int ms_read_multiple_pages(struct card_reader *cr,
		struct card *c, struct scsi_cmnd *srb, u16 phy_blk,
			u16 log_blk, u8 start_page, u8 end_page, u8 *buf,
				unsigned int **index, unsigned int *offset)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 extra[MS_COMM_EXTRA_SIZE], page_addr, val, trans_cfg, data[6];
	u8 *ptr;

	/* for Digital Read Protect bit issue */
	retval = ms_read_extra_data(cr, c, phy_blk, start_page,
						extra, MS_COMM_EXTRA_SIZE);
	if (retval == STATUS_SUCCESS) {
		/* if COPY bit & Access bit not set [1,1],then read protect */
		if ((extra[1] & 0x30) != 0x30) {
			ms_set_error_code(cr, c, MS_FLASH_READ_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
				MS_COMM_EXTRA_SIZE, SystemParm, 6);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Write REG */
	if (CHK_MS4BIT(ms_card)) {
		/* Parallel interface */
		data[0] = 0x88;
	} else {
		/* Serial interface */
		data[0] = 0x80;
	}
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(phy_blk >> 8);
	data[3] = (u8)phy_blk;
	/* Page Number */
	/* Extra data access mode */
	data[4] = 0;
	data[5] = start_page;

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, WRITE_REG, 6,
						NO_WAIT_INT, data, 6);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_set_error_code(cr, c, MS_NO_ERROR);
	retval = ms_send_cmd(cr, c, BLOCK_READ, WAIT_INT);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ptr = buf;

	for (page_addr = start_page; page_addr < end_page; page_addr++) {
		ms_set_error_code(cr, c, MS_NO_ERROR);

		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* GET_INT Register */
		retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (val & INT_REG_CMDNK) {
			/* Command can not be executed */
			ms_set_error_code(cr, c, MS_CMD_NK);
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (val & INT_REG_ERR) {
			if (val & INT_REG_BREQ) {
				retval = ms_read_status_reg(cr, c);
				if (retval != STATUS_SUCCESS) {
					/* for MS check procedure when reading
					 * data,if uncorrectable error occurs,
					 * such a page status is set to 1:NG,
					 * and block status in not set to 0:NG,
					 * when W/P OFF, besides, page status
					 * and block status are not changed when
					 * W/P ON
					 */
					if (!check_card_flag(c, WP)) {
						/* because BREQ may still set,
						 * we can't do as below
						 * directly.  After reset MS ,we
						 * can do as below. */
						ms_comm_reset_ms(cr, c);
						ms_comm_set_page_status(log_blk,
							setPS_NG, extra,
							MS_COMM_EXTRA_SIZE);
						ms_write_extra_data(cr,
							c, phy_blk,
							page_addr, extra,
							MS_COMM_EXTRA_SIZE);
					}
					ms_set_error_code(cr, c,
						MS_FLASH_READ_ERROR);
					TRACE_RET(cr, STATUS_FAIL);
				}
			} else {
				ms_set_error_code(cr, c,
						MS_FLASH_READ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			if (!(val & INT_REG_BREQ)) {
				ms_set_error_code(cr, c, MS_BREQ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		if (page_addr == (end_page - 1)) {
			if (!(val & INT_REG_CED)) {
				retval = ms_send_cmd(cr, c,
					BLOCK_END, WAIT_INT);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}
			}

			/* GET_INT Register Here We can't check CMDNK here,
			  * because MS Card sometimes will return CMDNK=1 and
			  * CED=1 while last page */
			retval = ms_read_bytes(cr, c,
						GET_INT, 1,
						NO_WAIT_INT, &val, 1);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
			if (!(val & INT_REG_CED)) {
				ms_set_error_code(cr, c,
						MS_FLASH_READ_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
			trans_cfg = NO_WAIT_INT;
		} else {
			trans_cfg = WAIT_INT;
		}
		retval = ms_read_page_data(cr, c, srb, trans_cfg, ptr,
							index, offset);
		if (retval < 0) {
			if (retval == -ETIMEDOUT) {
				ms_set_error_code(cr, c, MS_TO_ERROR);
				ms_clear_error(cr);
				TRACE_RET(cr, STATUS_TIMEDOUT);
			}
			retval = ms_get_int_status(cr, &val);
			if (retval != STATUS_SUCCESS) {
				ms_set_error_code(cr, c, MS_TO_ERROR);
				ms_clear_error(cr);
				TRACE_RET(cr, STATUS_TIMEDOUT);
			}
			if (val & (MS_COMM_CRC16_ERR |
				MS_COMM_RDY_TIMEOUT)) {
				ms_set_error_code(cr, c, MS_CRC16_ERROR);
				ms_clear_error(cr);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		if (scsi_sg_count(srb) == 0)
			ptr += 512;
	}

	return STATUS_SUCCESS;
}

static int ms_write_multiple_pages(struct card_reader *cr,
		struct card *c, struct scsi_cmnd *srb, u16 old_blk,
		u16 new_blk, u16 log_blk, u8 start_page, u8 end_page, u8 *buf,
				unsigned int **index, unsigned int *offset)
{
	struct ms_info *ms_card = c->extra;
	int retval, i;
	u8 page_addr, val, data[16];
	u8 *ptr;

	if (!start_page) {
		retval = ms_set_rw_reg_addr(cr, c,
				OverwriteFlag, MS_COMM_EXTRA_SIZE,
							SystemParm, 7);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (CHK_MS4BIT(ms_card))
			/* Parallel interface */
			data[0] = 0x88;
		else
			/* Serial interface */
			data[0] = 0x80;
		/* Block Address */
		data[1] = 0;
		data[2] = (u8)(old_blk >> 8);
		data[3] = (u8)old_blk;
		data[4] = 0x80;
		data[5] = 0;
		data[6] = 0xEF;
		data[7] = 0xFF;
		retval = ms_write_bytes(cr, c, WRITE_REG, 7,
						NO_WAIT_INT, data, 8);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = ms_send_cmd(cr, c, BLOCK_WRITE,
							WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* GET_INT Register */
		ms_set_error_code(cr, c, MS_NO_ERROR);
		retval = ms_transfer_tpc(cr, c, MS_TM_READ_BYTES,
				GET_INT, 1, NO_WAIT_INT);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	retval = ms_set_rw_reg_addr(cr, c, OverwriteFlag,
				MS_COMM_EXTRA_SIZE, SystemParm,
						(6 + MS_COMM_EXTRA_SIZE));
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ms_set_error_code(cr, c, MS_NO_ERROR);

	if (CHK_MS4BIT(ms_card))
		/* Parallel interface */
		data[0] = 0x88;
	else
		/* Serial interface */
		data[0] = 0x80;
	/* Block Address */
	data[1] = 0;
	data[2] = (u8)(new_blk >> 8);
	data[3] = (u8)new_blk;
	/* Page Number */
	/* Extra data access mode */
	if ((end_page - start_page) == 1)
		/* Single page access mode */
		data[4] = 0x20;
	else
		/* Block access mode */
		data[4] = 0;
	data[5] = start_page;
	data[6] = 0xF8;
	data[7] = 0xFF;
	data[8] = (u8)(log_blk >> 8);
	data[9] = (u8)log_blk;

	for (i = 0x0A; i < 0x10; i++) {
		/* ECC */
		data[i] = 0xFF;
	}

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_write_bytes(cr, c, WRITE_REG,
				6 + MS_COMM_EXTRA_SIZE, NO_WAIT_INT,
								data, 16);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	for (i = 0; i < MS_COMM_MAX_RETRY_COUNT; i++) {
		retval = ms_send_cmd(cr, c, BLOCK_WRITE,
							WAIT_INT);
		if (retval == STATUS_SUCCESS)
			break;
	}
	if (i == MS_COMM_MAX_RETRY_COUNT) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* GET_INT Register */
	retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	ptr = buf;
	for (page_addr = start_page; page_addr < end_page; page_addr++) {
		ms_set_error_code(cr, c, MS_NO_ERROR);

		if (cr_detect_card(cr, MS_ID) != STATUS_SUCCESS) {
			ms_set_error_code(cr, c, MS_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (val & INT_REG_CMDNK) {
			/* Command can not be executed */
			ms_set_error_code(cr, c, MS_CMD_NK);
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (val & INT_REG_ERR) {
			ms_set_error_code(cr, c, MS_FLASH_WRITE_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}
		if (!(val & INT_REG_BREQ)) {
			ms_set_error_code(cr, c, MS_BREQ_ERROR);
			TRACE_RET(cr, STATUS_FAIL);
		}

		udelay(30);
		retval = ms_write_page_data(cr, c, srb, ptr, index, offset);
		if (retval < 0) {
			ms_set_error_code(cr, c, MS_TO_ERROR);
			ms_clear_error(cr);
			if (retval == -ETIMEDOUT) {
				TRACE_RET(cr, STATUS_TIMEDOUT);
			} else {
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
		/* GET_INT Register */
		retval = ms_read_bytes(cr, c, GET_INT, 1,
						NO_WAIT_INT, &val, 1);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if ((end_page - start_page) == 1) {
			if (!(val & INT_REG_CED)) {
				/* Command can not be executed */
				ms_set_error_code(cr, c, MS_FLASH_WRITE_ERROR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			if (page_addr == (end_page - 1)) {
				if (!(val & INT_REG_CED)) {
					retval = ms_send_cmd(cr, c,
							BLOCK_END,
							WAIT_INT);
					if (retval != STATUS_SUCCESS) {
						TRACE_RET(cr, STATUS_FAIL);
					}
				}

				/* GET_INT Register */
				retval = ms_read_bytes(cr, c,
						GET_INT, 1,
						NO_WAIT_INT, &val, 1);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}
			}

			if ((page_addr == (end_page - 1)) ||
					(page_addr == ms_card->page_off)) {
				if (!(val & INT_REG_CED)) {
					/* Command can not be executed */
					ms_set_error_code(cr, c,
						MS_FLASH_WRITE_ERROR);
					TRACE_RET(cr, STATUS_FAIL);
				}
			}
		}

		if (scsi_sg_count(srb) == 0)
			ptr += 512;
	}

	return STATUS_SUCCESS;
}

static int ms_finish_write(struct card_reader *cr, struct card *c,
			u16 old_blk, u16 new_blk, u16 log_blk, u8 page_off)
{
	struct ms_info *ms_card = c->extra;
	int retval, seg_no;

	retval = ms_copy_page(cr, c, old_blk, new_blk, log_blk,
					page_off, ms_card->page_off + 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	seg_no = old_blk >> 9;

	if (MS_TST_BAD_BLOCK_FLG(ms_card)) {
		MS_CLR_BAD_BLOCK_FLG(ms_card);
		ms_set_bad_block(cr, c, old_blk);
	} else {
		retval = ms_erase_block(cr, c, old_blk);
		if (retval == STATUS_SUCCESS)
			ms_comm_set_unused_block(cr, c, old_blk);
	}
	ms_comm_set_l2p_tbl(cr, c, seg_no, log_blk - ms_start_idx[seg_no],
								new_blk);

	return STATUS_SUCCESS;
}

static int ms_prepare_write(struct card_reader *cr,
				struct card *c, u16 old_blk, u16 new_blk,
						u16 log_blk, u8 start_page)
{
	int retval;

	if (start_page) {
		retval = ms_copy_page(cr, c, old_blk, new_blk,
						log_blk, 0, start_page);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

/* #ifdef MS_DELAY_WRITE */
int ms_delay_write(struct card_reader *cr, struct card *c)
{
	struct ms_info *ms_card = c->extra;
	struct ms_delay_write_tag *delay_write = &(ms_card->delay_write);
	int retval;

	if (delay_write->delay_write_flag) {
		retval = ms_set_init_para(cr, c);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		delay_write->delay_write_flag = 0;
		retval = ms_finish_write(cr, c,
						delay_write->old_phyblock,
						delay_write->new_phyblock,
						delay_write->logblock,
						delay_write->pageoff);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}
/* #endif */

static inline void ms_comm_rw_fail(struct scsi_cmnd *srb,
						struct card_reader *cr)
{
	if (srb->sc_data_direction == DMA_FROM_DEVICE)
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
	else
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_WRITE_ERR);
}

static int ms_ms_rw_multi_sector(struct scsi_cmnd *srb,
			struct card_reader *cr, struct card *c,
					u32 start_sector, u16 sector_cnt)
{
	struct ms_info *ms_card = c->extra;
	u8 lun = SCSI_LUN(srb);
	int retval, seg_no;
	unsigned int index = 0, offset = 0;
	unsigned int *__index = NULL; /* default for USB transport */
	u16 old_blk = 0, new_blk = 0, log_blk, total_sec_cnt = sector_cnt;
	u8 start_page, end_page = 0, page_cnt;
	u8 *ptr;
#ifdef MS_DELAY_WRITE
	struct ms_delay_write_tag *delay_write = &(ms_card->delay_write);
#endif


	if (!strcmp(cr->tr->name, "PCIe"))
		__index = &index;

	ms_set_error_code(cr, c, MS_NO_ERROR);

	ms_card->cleanup_counter = 0;

	ptr = (u8 *)scsi_sglist(srb);

	retval = ms_switch_clock(cr, c);
	if (retval != STATUS_SUCCESS) {
		ms_comm_rw_fail(srb, cr);
		TRACE_RET(cr, STATUS_FAIL);
	}

	log_blk = (u16)(start_sector >> ms_card->block_shift);
	start_page = (u8)(start_sector & ms_card->page_off);

	for (seg_no = 0; seg_no < sizeof(ms_start_idx)/2 - 1; seg_no++) {
		if (log_blk < ms_start_idx[seg_no+1])
			break;
	}

	if (ms_card->segment[seg_no].build_flag == 0) {
		retval = ms_comm_build_l2p_tbl(cr, c, seg_no);
		if (retval != STATUS_SUCCESS) {
			set_card_flags(c, FAIL);
			set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	if (srb->sc_data_direction == DMA_TO_DEVICE) {
#ifdef MS_DELAY_WRITE
		if (delay_write->delay_write_flag &&
				(delay_write->logblock == log_blk) &&
				(start_page > delay_write->pageoff)) {
			delay_write->delay_write_flag = 0;
			retval = ms_copy_page(cr, c,
					delay_write->old_phyblock,
					delay_write->new_phyblock, log_blk,
					delay_write->pageoff, start_page);
			if (retval != STATUS_SUCCESS) {
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
			old_blk = delay_write->old_phyblock;
			new_blk = delay_write->new_phyblock;
		} else if (delay_write->delay_write_flag &&
				(delay_write->logblock == log_blk) &&
				(start_page == delay_write->pageoff)) {
			delay_write->delay_write_flag = 0;
			old_blk = delay_write->old_phyblock;
			new_blk = delay_write->new_phyblock;
		} else {
			retval = ms_delay_write(cr, c);
			if (retval != STATUS_SUCCESS) {
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
#endif
			old_blk = ms_comm_get_l2p_tbl(cr, c, seg_no,
					log_blk - ms_start_idx[seg_no]);
			new_blk = ms_comm_get_unused_block(cr, c, seg_no);
			if ((old_blk == 0xFFFF) || (new_blk == 0xFFFF)) {
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
			retval = ms_prepare_write(cr, c, old_blk, new_blk,
							log_blk, start_page);
			if (retval != STATUS_SUCCESS) {
				if (cr_detect_card(cr, MS_ID) !=
					STATUS_SUCCESS) {
					set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
					TRACE_RET(cr, STATUS_FAIL);
				}
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
#ifdef MS_DELAY_WRITE
		}
#endif
	} else {
#ifdef MS_DELAY_WRITE
		retval = ms_delay_write(cr, c);
		if (retval != STATUS_SUCCESS) {
			if (cr_detect_card(cr, MS_ID) !=
				STATUS_SUCCESS) {
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
				TRACE_RET(cr, STATUS_FAIL);
			}
			set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			TRACE_RET(cr, STATUS_FAIL);
		}
#endif
		old_blk = ms_comm_get_l2p_tbl(cr, c, seg_no,
					log_blk - ms_start_idx[seg_no]);
		if (old_blk == 0xFFFF) {
			set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	rts_trace(TRACE_MS, "seg_no = %d, old_blk = 0x%x, new_blk = 0x%x\n",
						seg_no, old_blk, new_blk);

	while (total_sec_cnt) {
		if ((start_page + total_sec_cnt) > (ms_card->page_off + 1))
			end_page = ms_card->page_off + 1;
		else
			end_page = start_page + (u8)total_sec_cnt;
		page_cnt = end_page - start_page;

		rts_trace(TRACE_MS, "start_page = %d, end_page = %d,"
						"page_cnt = %d\n", start_page,
							end_page, page_cnt);

		if (srb->sc_data_direction == DMA_FROM_DEVICE)
			retval = ms_read_multiple_pages(cr, c, srb,
					old_blk, log_blk, start_page, end_page,
							ptr, &__index, &offset);
		else
			retval = ms_write_multiple_pages(cr, c, srb,
					old_blk, new_blk, log_blk, start_page,
					end_page, ptr, &__index, &offset);
		if (retval != STATUS_SUCCESS) {
			if (cr_detect_card(cr, MS_ID) !=
				STATUS_SUCCESS) {
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
				TRACE_RET(cr, STATUS_FAIL);
			}
			ms_comm_rw_fail(srb, cr);
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* Update L2P table if need */
		if (srb->sc_data_direction == DMA_TO_DEVICE) {
			if (end_page == (ms_card->page_off + 1)) {
				retval = ms_erase_block(cr, c, old_blk);
				if (retval == STATUS_SUCCESS) {
					ms_comm_set_unused_block(cr, c,
								old_blk);
				}
				ms_comm_set_l2p_tbl(cr, c, seg_no,
					log_blk - ms_start_idx[seg_no],
					new_blk);
			}
		}

		total_sec_cnt -= page_cnt;
		if (scsi_sg_count(srb) == 0)
			ptr += page_cnt * 512;

		if (total_sec_cnt == 0)
			/* Data transfer finished */
			break;

		log_blk++;

		for (seg_no = 0; seg_no < sizeof(ms_start_idx)/2 - 1;
			seg_no++) {
			if (log_blk < ms_start_idx[seg_no+1])
				break;
		}

		if (ms_card->segment[seg_no].build_flag == 0) {
			retval = ms_comm_build_l2p_tbl(cr, c, seg_no);
			if (retval != STATUS_SUCCESS) {
				set_card_flags(c, FAIL);
				set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
		old_blk = ms_comm_get_l2p_tbl(cr, c, seg_no,
					log_blk - ms_start_idx[seg_no]);
		if (old_blk == 0xFFFF) {
			ms_comm_rw_fail(srb, cr);
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (srb->sc_data_direction == DMA_TO_DEVICE) {
			new_blk = ms_comm_get_unused_block(cr, c, seg_no);
			if (new_blk == 0xFFFF) {
				ms_comm_rw_fail(srb, cr);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		rts_trace(TRACE_MS,
			"seg_no = %d, old_blk = 0x%x, new_blk = 0x%x\n",
			seg_no, old_blk, new_blk);

		start_page = 0;
	}

	if (srb->sc_data_direction == DMA_TO_DEVICE) {
		if (end_page < (ms_card->page_off + 1)) {
#ifdef MS_DELAY_WRITE
			delay_write->delay_write_flag = 1;
			delay_write->old_phyblock = old_blk;
			delay_write->new_phyblock = new_blk;
			delay_write->logblock = log_blk;
			delay_write->pageoff = end_page;
#else
			retval = ms_finish_write(cr, c, old_blk, new_blk,
							log_blk, end_page);
			if (retval != STATUS_SUCCESS) {
				if (cr_detect_card(cr, MS_ID) !=
					STATUS_SUCCESS) {
					set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
					TRACE_RET(cr, STATUS_FAIL);
				}
				ms_comm_rw_fail(srb, cr);
				TRACE_RET(cr, STATUS_FAIL);
			}
#endif
		}
	}

	scsi_set_resid(srb, 0);


	return STATUS_SUCCESS;
}

static int ms_comm_rw(struct card_reader *cr, struct card *c,
				struct scsi_cmnd *srb, u32 lba, u16 num)
{
	struct ms_info *ms_card = c->extra;
	u32 start_sector = lba;
	u16 sector_cnt = num;


	if (CHK_MSPRO(ms_card)) {
		return ms_mspro_rw_multi_sector(srb, cr, c,
						start_sector, sector_cnt);
	} else {
		return ms_ms_rw_multi_sector(srb, cr, c,
						start_sector, sector_cnt);
	}
}


int ms_rw(struct card_reader *cr, struct card *c,
			struct scsi_cmnd *srb, u32 lba, u16 num)
{
	if (cr->ms_ops_prot->rw)
		return cr->ms_ops_prot->rw(cr, c, srb, lba, num);
	else
		return ms_comm_rw(cr, c, srb, lba, num);
}
EXPORT_SYMBOL_GPL(ms_rw);

static void mspro_stop_seq_mode(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	if (ms_card->seq_mode) {
		if (ms_switch_clock(cr, card))
			return;
		ms_card->seq_mode = 0;
		ms_card->total_sec_cnt = 0;
		ms_send_cmd(cr, card, PRO_STOP, WAIT_INT);
		cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
	}
}

static int ms_comm_cleanup_work(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	if (CHK_MSPRO(ms_card)) {
		if (ms_card->seq_mode) {
			rts_trace(TRACE_MS, "MS Pro: stop transmission\n");
			mspro_stop_seq_mode(cr, card);
			ms_card->cleanup_counter = 0;
		}
		if (CHK_MSHG(ms_card)) {
			/* clear 2K mode */
			cr_wr_reg(cr, REG_MS_CFG, MS_2K_SECTOR_MODE, 0x00);
		}
	}
#ifdef MS_DELAY_WRITE
	else if ((!CHK_MSPRO(ms_card)) &&
				ms_card->delay_write.delay_write_flag) {
		rts_trace(TRACE_MS, "MS: delay write\n");
		rts5227_ms_delay_write(cr, card);
		ms_card->cleanup_counter = 0;
	}
#endif

	return 0;
}


int ms_cleanup(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->cleanup)
		return cr->ms_ops_prot->cleanup(cr, c);
	else
		return ms_comm_cleanup_work(cr, c);
}
EXPORT_SYMBOL_GPL(ms_cleanup);


void ms_free_l2p_tbl(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;
	int i = 0;

	if (ms_card->segment != NULL) {
		for (i = 0; i < ms_card->segment_cnt; i++) {
			if (ms_card->segment[i].l2p_table != NULL) {
				vfree(ms_card->segment[i].l2p_table);
				ms_card->segment[i].l2p_table = NULL;
			}
			if (ms_card->segment[i].free_table != NULL) {
				vfree(ms_card->segment[i].free_table);
				ms_card->segment[i].free_table = NULL;
			}
		}
		vfree(ms_card->segment);
		ms_card->segment = NULL;
	}
}
EXPORT_SYMBOL_GPL(ms_free_l2p_tbl);

static int ms_comm_release(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;

	rts_trace(TRACE_MS, "%s\n", __func__);

#ifdef MS_DELAY_WRITE
	ms_card->delay_write.delay_write_flag = 0;
#endif
	ms_card->pro_under_formatting = 0;
	clr_card_flags(card, (READY | FAIL | WP));
	ms_free_l2p_tbl(cr, card);
	memset(ms_card->raw_sys_info, 0, sizeof(ms_card->raw_sys_info));
	memset(ms_card->raw_ms_id, 0, sizeof(ms_card->raw_ms_id));
#ifdef SUPPORT_PCGL_1P18
	memset(ms_card->raw_model_name, 0, sizeof(ms_card->raw_model_name));
#endif

	if (ms_power_off_card3v3(cr))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int ms_release(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->release)
		return cr->ms_ops_prot->release(cr, c);
	else
		return ms_comm_release(cr, c);

}
EXPORT_SYMBOL_GPL(ms_release);

static int ms_comm_update_status(struct card_reader *cr, struct card *c)
{
	/* TODO: implement */
	return -1;
}

int ms_update_status(struct card_reader *cr, struct card *c)
{
	if (cr->ms_ops_prot->update_status)
		return cr->ms_ops_prot->update_status(cr, c);
	else
		return ms_comm_update_status(cr, c);
}
EXPORT_SYMBOL_GPL(ms_update_status);

int ms_card_reset(struct card *c)
{
	return ms_reset(c->cm->cr, c);
}

int ms_card_disable(struct card *c)
{
	return ms_disable(c->cm->cr, c);
}

int ms_card_rw(struct scsi_cmnd *srb, struct card *c, u32 lba, u16 num)
{
	return ms_rw(c->cm->cr, c, srb, lba, num);
}

int ms_card_cleanup(struct card *c)
{
	return ms_cleanup(c->cm->cr, c);
}

int ms_card_release(struct card *c)
{
	return ms_release(c->cm->cr, c);
}

int ms_card_update_status(struct card *c)
{
	return ms_update_status(c->cm->cr, c);
}
