/* Driver for Realtek card reader
 * Header file
 *
 * Copyright(c) 2013 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 */

#include <linux/kernel.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include <linux/module.h>

#include "cardreader.h"
#include "cardmanager.h"
#include "rts_cr.h"
#include "trace.h"
#include "cr_regs.h"
#include "cards.h"
#include "sd.h"
#include "sd40.h"

static void sd40_err_st_print(struct card_reader *cr, u8 val);
static void sd40_tmout0_print(struct card_reader *cr, u8 val);
static void sd40_tmout1_print(struct card_reader *cr, u8 val);
static u8 get_dma_packet_size_from_n_fcu(u8 n_fcu);

static void add_cmd_4_native_dcmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	u32 byte_cnt;
	u8 *p = tlp->payload;

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0, 0xFF, tlp->header[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF, tlp->header[1]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, ARG_DIR, tlp->arg[0]);

	rts_trace(TRACE_SD, "Native DCMD\n");
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG1, 0xFF, tlp->arg[1]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0x78, tlp->arg[0]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0, 0xFF, *p++);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_3, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_2, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0, 0xFF, *p++);

	/* enable dma */
	/* how should we decide DMA direction for NONE native packet ? */
	byte_cnt = (p[4] << 24) | (p[5] << 16) | (p[6] << 8) | p[7];
	if (ARG_DIR_R == (tlp->arg[0] & ARG_DIR))
		cr_enable_dma(cr, byte_cnt, DMA_512, DMA_FROM_DEVICE);
	else
		cr_enable_dma(cr, byte_cnt, DMA_512, DMA_TO_DEVICE);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN,
			SD_TRANSFER_END, SD_TRANSFER_END);

}

static void add_cmd_4_sdtran_dcmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 blk_cnt;
	u32 byte_cnt;
	u8 tmode = tlp->arg[0] & 0x78;
	u8 *p = tlp->payload;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0, 0xFF, tlp->header[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF, tlp->header[1]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, ARG_DIR, tlp->arg[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0xFF, tlp->arg[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3, 0xFF, tlp->arg[1]);

	byte_cnt = (p[4] << 24) | (p[5] << 16) | (p[6] << 8) | p[7];
	blk_cnt = byte_cnt / 512 + ((byte_cnt % 512) ? 1 : 0);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0, 0xFF, *p++);

	if (tmode & LM_LEN_SPECIF) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0,
				0xFF, (u8)(blk_cnt));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1,
				0xFF, (u8)(blk_cnt >> 8));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_2,
				0xF0, 0x00);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_3,
				0xFF, 0x20);
	} else {
		if (byte_cnt & 0x1ff) {
			cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_2,
				0xF0, (u8)(byte_cnt << 4));
			cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_3,
				0xFF, (u8)(byte_cnt >> 4));
		} else {
			cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_2,
					0xF0, 0x00);
			cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_LINK_SETTING_3,
					0xFF, 0x20);
		}

		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_0,
				0xFF, (u8)(blk_cnt));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_1,
				0xFF, (u8)(blk_cnt >> 8));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_2, 0xFF, 0);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_3, 0xFF, 0);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0, 0xFF, 0);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1, 0xFF, 0);
	}
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_3, 0xFF, 0);

	/* enable dma */
	/* how should we decide DMA direction for NONE native packet ? */
	if (ARG_DIR_R == (tlp->arg[0] & ARG_DIR))
		cr_enable_dma(cr, byte_cnt, ps, DMA_FROM_DEVICE);
	else
		cr_enable_dma(cr, byte_cnt, ps, DMA_TO_DEVICE);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN,
			SD_TRANSFER_END, SD_TRANSFER_END);

}

static void add_cmd_4_dcmd(struct card *card, struct sd40_tlp *tlp)
{
	if (tlp->header[0] & NATIVE_PACKET)
		add_cmd_4_native_dcmd(card, tlp);
	else
		add_cmd_4_sdtran_dcmd(card, tlp);
}

static int transfer_sg_4_dcmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	enum dma_data_direction dir;
	u8 *p = tlp->payload;
	int len = (p[4] << 24) | (p[5] << 16) | (p[6] << 8) | p[7];
	int ret;


	if (ARG_DIR_W == (tlp->arg[0] & ARG_DIR))
		dir = DMA_TO_DEVICE;
	else
		dir = DMA_FROM_DEVICE;

	ret = cr_transfer_data(cr, SD_CARD, tlp->sg, len, 1, dir, tlp->timeout);

	return ret;
}

static int transfer_single_data_4_dcmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	enum dma_data_direction dir;
	u8 *p = tlp->payload;
	int len = (p[4] << 24) | (p[5] << 16) | (p[6] << 8) | p[7];
	int ret;
	u8 *buf;

	buf = (u8 *)cr_alloc_dma_buf(cr, len, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	if (ARG_DIR_W == (tlp->arg[0] & ARG_DIR)) {
		memcpy(buf, tlp->data, len);
		dir = DMA_TO_DEVICE;
	} else {
		dir = DMA_FROM_DEVICE;
	}

	ret = cr_transfer_data(cr, SD_CARD, buf, len, 0, dir, tlp->timeout);
	if (ret)
		goto out;
	if (DMA_FROM_DEVICE == dir)
		memcpy(tlp->data, buf, len);
out:
	if (buf)
		cr_free_dma_buf(cr, buf);
	return ret;
}

static int sd40_send_dcmd_packet(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	u32 len = 0;
	int ret = 0;

	sd_clear_error_code(cr, card);

	/* enable U_SHORT_DAM_EN for CPRM when data_len < 512. */
	len = tlp->payload[7];
	len |= tlp->payload[6] << 8;
	len |= tlp->payload[5] << 16;
	len |= tlp->payload[4] << 24;
	if (len & 0x1FF)
		cr_wr_reg(cr, REG_DMACTL, U_SHORT_DMA_EN, U_SHORT_DMA_EN);

	cr_wr_reg(cr, REG_SD40_EBSY, 0xC3,
		SD40_HAS_EBSY | SD40_WAIT_EBSY | SD40_STAT_EBSY_TMOUT);

	add_cmd_4_dcmd(card, tlp);

	cr_send_cmd_no_wait(cr);

	if (tlp->scatter_gather)
		ret = transfer_sg_4_dcmd(card, tlp);
	else
		ret = transfer_single_data_4_dcmd(card, tlp);

	cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, 0);

	/* disable U_SHORT_DAM_EN */
	if (len & 0x1FF)
		cr_wr_reg(cr, REG_DMACTL, U_SHORT_DMA_EN, 0);

	/* do we need process error for transfer fail? */
	if (ret < 0)
		dump_all_sd40_register(cr);

	return ret;
}

static int sd40_get_payload_len(u8 arg0)
{
	int len = 0;

	switch (arg0 & ARG_PLEN_MASK) {
	case PLEN_0B:
		len = 0;
		break;
	case PLEN_4B:
		len = 4;
		break;
	case PLEN_8B:
		len = 8;
		break;
	case PLEN_16B:
		len = 16;
		break;
	default:
		len = 0;
		break;
	}

	return len;
}

static void add_cmd_4_native_ccmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	int len = 0;

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0, 0xFF, tlp->header[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF, tlp->header[1]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, 0xFF, tlp->arg[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG1, 0xFF, tlp->arg[1]);
	len = sd40_get_payload_len(tlp->arg[0]);

	if (ARG_DIR_W == (tlp->arg[0] & ARG_DIR)) {
		u8 *p = tlp->payload;

		if (len >= 16) {
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW3_3, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW3_2, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW3_1, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW3_0, 0xFF, *p++);

			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW2_3, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW2_2, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW2_1, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW2_0, 0xFF, *p++);
		}

		if (len >= 8) {
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW1_3, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW1_2, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW1_1, 0xFF, *p++);
			cr_add_cmd(cr, WRITE_REG_CMD,
					REG_SD40_PLD_DW1_0, 0xFF, *p++);
		}

		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3, 0xFF, *p++);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2, 0xFF, *p++);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1, 0xFF, *p++);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0, 0xFF, *p++);
	}

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN,
			SD_TRANSFER_END, SD_TRANSFER_END);

	/* should we read RXPLD for all native ccmd packet ? */
	/* if (ARG_DIR_R == (tlp->arg[0] & ARG_DIR)) { */

	if (len >= 16) {
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_3, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_2, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_1, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_0, 0xFF, 0);

		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_3, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_2, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_1, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_0, 0xFF, 0);
	}

	if (len >= 8) {
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_3, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_2, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_1, 0xFF, 0);
		cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_0, 0xFF, 0);
	}

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_0, 0xFF, 0);
	/* } */
}

static void add_cmd_4_sdtran_ccmd(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	u8 *p = tlp->payload;

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0, 0xFF, tlp->header[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF, tlp->header[1]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, ARG_PLEN_MASK, PLEN_4B);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0xFF, tlp->arg[0]);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3, 0xFF, tlp->arg[1]);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1, 0xFF, *p++);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0, 0xFF, *p++);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN,
			SD_TRANSFER_END, SD_TRANSFER_END);



	/* to in accordance with legacy cmd response, read one more
	 * byte
	 */
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_RX_ARG1, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_0, 0xFF, 0);
}

static void add_cmd_4_ccmd(struct card *card, struct sd40_tlp *tlp)
{
	if (tlp->header[0] & NATIVE_PACKET)
		add_cmd_4_native_ccmd(card, tlp);
	else
		add_cmd_4_sdtran_ccmd(card, tlp);
}

static void sd40_dump_st(struct card_reader *cr)
{
	u8 val;

	cr_rd_reg(cr, REG_SD40_TRAN, &val);
	rts_trace(TRACE_SD, "SD40_TRAN: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS0, &val);
	rts_trace(TRACE_SD, "SD40_STATUS0(TPSM/TBSM): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS1, &val);
	rts_trace(TRACE_SD, "SD40_STATUS1(ERR/TSSM): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS2, &val);
	rts_trace(TRACE_SD, "SD40_STATUS2(TX/RX_STAT): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS3, &val);
	rts_trace(TRACE_SD, "SD40_STATUS3(DLSM): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS4, &val);
	rts_trace(TRACE_SD, "SD40_STATUS4(LPBK/PKTRX): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_TMOUT_0, &val);
	rts_trace(TRACE_SD, "SD40_TMOUT_0: 0x%X\n", val);
	sd40_tmout0_print(cr, val);

	cr_rd_reg(cr, REG_SD40_TMOUT_1, &val);
	rts_trace(TRACE_SD, "SD40_TMOUT_1: 0x%X\n", val);
	sd40_tmout1_print(cr, val);

	cr_rd_reg(cr, REG_SD40_ERR_ST, &val);
	rts_trace(TRACE_SD, "SD40_ERR_ST: 0x%X\n", val);
	sd40_err_st_print(cr, val);
}

static void sd40_status1_print(struct card_reader *cr, u8 val)
{
	switch (val & 0x70) {
	case SD40_HEAD_ERR:
		rts_trace(TRACE_SD, "header error\n");
		break;
	case SD40_FRAM_ERR:
		rts_trace(TRACE_SD, "frame error\n");
		break;
	case SD40_CRC_ERR:
		rts_trace(TRACE_SD, "CRC error\n");
		break;
	}

}

static void sd40_tmout0_print(struct card_reader *cr, u8 val)
{
	switch (val) {
	case SD40_T_EIDLE_STB_OUT:
		rts_trace(TRACE_SD, "eidle STB timeout\n");
		break;
	case SD40_T_ACTIVE_INIT_OUT:
		rts_trace(TRACE_SD, "active init timeout\n");
		break;
	case SD40_T_LIDLE_LIDLE_OUT:
		rts_trace(TRACE_SD, "lidle timeout\n");
		break;
	case SD40_T_LIDLE_FCREQ_BOOT_OUT:
		rts_trace(TRACE_SD, "lidel fcreq boot timeout\n");
		break;
	case SD40_T_DINI_CMD_OUT:
		rts_trace(TRACE_SD, "device init CCMD timeout\n");
		break;
	case SD40_T_CMD_RSP_OUT:
		rts_trace(TRACE_SD, "CCMD or DCMD response timeout\n");
		break;
	case SD40_T_FWD_BCMD_OUT:
		rts_trace(TRACE_SD, "broadcast CCMD response timeout\n");
		break;
	case SD40_T_RES_FCREQ_OUT:
		rts_trace(TRACE_SD,
			"response fcreq timeout in read transaction\n");
		break;
	}
}

static void sd40_tmout1_print(struct card_reader *cr, u8 val)
{
	switch (val&0x7f) {
	case SD40_T_STAT_FCREQ_OUT:
		rts_trace(TRACE_SD,
			"stat fcreq timeout in write transaction\n");
		break;
	case SD40_T_FCREQ_RDY_OUT:
		rts_trace(TRACE_SD, "fcreq fcrdy timeout\n");
		break;
	case SD40_T_FCRDY_EDB_OUT:
		rts_trace(TRACE_SD, "fcrdy edb timeout of data burst\n");
		break;
	case SD40_T_EDB_STAT_OUT:
		rts_trace(TRACE_SD, "edb stat timeout of data burst\n");
		break;
	}
}

static void sd40_err_st_print(struct card_reader *cr, u8 val)
{
	switch (val) {
	case SD40_LINK_MEM_ERR:
		rts_trace(TRACE_SD, "link message mem err\n");
		break;
	case SD40_LINK_UNRECV_ERR:
		rts_trace(TRACE_SD, "link message unrecoverable err\n");
		break;
	case SD40_LINK_RECV_ERR:
		rts_trace(TRACE_SD, "link message recoverable err\n");
		break;
	case SD40_RETRY_EXP_R_ERR:
		rts_trace(TRACE_SD, "retry number expire when read\n");
		break;
	case SD40_PACKET_ERR:
		rts_trace(TRACE_SD, "header or frame or CRC error\n");
		break;
	case SD40_RETRY_EXP_W_ERR:
		rts_trace(TRACE_SD, "retry number expire when write\n");
		break;
	case SD40_HEADER_ERR:
		rts_trace(TRACE_SD, "header error\n");
		break;
	case SD40_NACK_ERR:
		rts_trace(TRACE_SD, "response is NACK\n");
		break;
	}
}

void dump_all_sd40_register(struct card_reader *cr)
{
	u8 val;

	rts_trace(TRACE_SD, "###Dump All Register!!!!\n");

	cr_rd_reg(cr, REG_SD40_TRAN, &val);
	rts_trace(TRACE_SD, "SD40_TRAN: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_CTRL, &val);
	rts_trace(TRACE_SD, "SD40_CTRL: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS0, &val);
	rts_trace(TRACE_SD, "SD40_STATUS0(TPSM/TBSM): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS1, &val);
	rts_trace(TRACE_SD, "SD40_STATUS1(ERR/TSSM): 0x%X\n", val);
	sd40_status1_print(cr, val);

	cr_rd_reg(cr, REG_SD40_STATUS2, &val);
	rts_trace(TRACE_SD, "SD40_STATUS2(TX/RX_STAT): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS3, &val);
	rts_trace(TRACE_SD, "SD40_STATUS3(DLSM): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_STATUS4, &val);
	rts_trace(TRACE_SD, "SD40_STATUS4(LPBK/PKTRX): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_Dummy0, &val);
	rts_trace(TRACE_SD, "SD40_Dummy0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_Dummy1 , &val);
	rts_trace(TRACE_SD, "SD40_Dummy1 : 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_MISC, &val);
	rts_trace(TRACE_SD, "SD40_MISC: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_TMOUT_0, &val);
	rts_trace(TRACE_SD, "SD40_TMOUT_0: 0x%X\n", val);
	sd40_tmout0_print(cr, val);

	cr_rd_reg(cr, REG_SD40_TMOUT_1, &val);
	rts_trace(TRACE_SD, "SD40_TMOUT_1: 0x%X\n", val);
	sd40_tmout1_print(cr, val);

	cr_rd_reg(cr, REG_SD40_ERR_ST, &val);
	rts_trace(TRACE_SD, "SD40_ERR_ST: 0x%X\n", val);
	sd40_err_st_print(cr, val);

	cr_rd_reg(cr, REG_SD40_ARG0, &val);
	rts_trace(TRACE_SD, "SD40_ARG0(RW,PLEN,IOADR_H): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_ARG1, &val);
	rts_trace(TRACE_SD, "SD40_ARG1(IOADR_L): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_ARG2, &val);
	rts_trace(TRACE_SD, "SD40_ARG2(TMOD): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_ARG3, &val);
	rts_trace(TRACE_SD, "SD40_ARG3(CMD_INDEX): 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_ST_REG_2, &val);
	rts_trace(TRACE_SD, "SD40_ST_REG_2(rty/rcov/dev/illegal err): 0x%X\n",
		val);

	cr_rd_reg(cr, REG_SD40_RX_HDR0, &val);
	rts_trace(TRACE_SD, "SD40_RX_HDR0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RX_HDR1, &val);
	rts_trace(TRACE_SD, "SD40_RX_HDR1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RX_ARG0, &val);
	rts_trace(TRACE_SD, "SD40_RX_ARG0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RX_ARG1, &val);
	rts_trace(TRACE_SD, "SD40_RX_ARG1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW0_3, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW0_3: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW0_2, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW0_2: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW0_1, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW0_1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW0_0, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW0_0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW1_3, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW1_3: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW1_2, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW1_2: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW1_1, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW1_1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW1_0, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW1_0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW2_3, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW2_3: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW2_2, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW2_2: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW2_1, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW2_1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW2_0, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW2_0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW3_3, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW3_3: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW3_2, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW3_2: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW3_1, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW3_1: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_RXPLD_DW3_0, &val);
	rts_trace(TRACE_SD, "SD40_RXPLD_DW3_0: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_LINK_SETTING_2, &val);
	rts_trace(TRACE_SD, "SD40_LINK_SETTING_2: 0x%X\n", val);

	cr_rd_reg(cr, REG_SD40_LINK_SETTING_3, &val);
	rts_trace(TRACE_SD, "SD40_LINK_SETTING_3: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_CLK_EN, &val);
	rts_trace(TRACE_SD, "CARD_CLK_EN: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_SELECT, &val);
	rts_trace(TRACE_SD, "CARD_SELECT: 0x%X\n", val);

	cr_rd_reg(cr, REG_CLK_CTL, &val);
	rts_trace(TRACE_SD, "CLK_CTL: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_PWR_CTL, &val);
	rts_trace(TRACE_SD, "CARD_PWR_CTL: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_PULL_CTL1, &val);
	rts_trace(TRACE_SD, "CARD_PULL_CTL1: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_PULL_CTL2, &val);
	rts_trace(TRACE_SD, "CARD_PULL_CTL2: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_PULL_CTL3, &val);
	rts_trace(TRACE_SD, "CARD_PULL_CTL3: 0x%X\n", val);

	cr_rd_reg(cr, REG_CARD_PULL_CTL4, &val);
	rts_trace(TRACE_SD, "CARD_PULL_CTL4: 0x%X\n", val);

	rts_trace(TRACE_SD, "###Dump End!!!!\n");
}
EXPORT_SYMBOL_GPL(dump_all_sd40_register);

static int sd40_clear_error(struct card_reader *cr)
{
	sd40_dump_st(cr);

	if (CHECK_PID(cr, 0x524a) || CHECK_PID(cr, 0x525a))
		cr_wr_reg(cr, SD40_STOP_CTL , SD40_LINK_RST, SD40_LINK_RST);

	cr_wr_reg(cr, REG_SD40_CTRL, SD40_MAC_RST, SD40_MAC_RST);
	cr_wr_reg(cr, REG_DMACTL, DMA_RST, DMA_RST);

	rts_trace(TRACE_SD, "sd40 clear dma tc\n");
	cr_wr_reg(cr, REG_DMATC0, 0xff, 0x00);
	cr_wr_reg(cr, REG_DMATC1, 0xff, 0x00);
	cr_wr_reg(cr, REG_DMATC2, 0xff, 0x00);
	cr_wr_reg(cr, REG_DMATC3, 0xff, 0x00);

	return 0;
}

static int process_error_4_ccmd(struct card *card, int error)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int ret = error;
	u8 val;

	rts_trace(TRACE_SD, "!SD40 CMD Fail !!!!\n");

	dump_all_sd40_register(cr);

	if (-ETIMEDOUT == error) {
		rts_trace(TRACE_SD, "!!Timeout\n");
		sd_card->trans_abort = 1;
		sd_set_error_code(cr, card, SD_RSP_TIMEOUT);
		ret = STATUS_TIMEDOUT;
	} else {
		ret = STATUS_FAIL;
	}

	cr_rd_reg(cr, REG_SD40_TRAN, &val);
	if (val & SD40_TRANS_ERR) {
		sd_set_error_code(cr, card, SD_STS_ERR);
		cr_rd_reg(cr, REG_SD40_ST_REG_2, &val);
		if (val & RETRY_EXPIRE_ERROR) {
			rts_trace(TRACE_SD, "**Retry expire error!*\n");
			sd_card->trans_abort = 1;
		}

		cr_rd_reg(cr, REG_SD40_TMOUT_0, &val);
		if (val & SD40_T_CMD_RSP_OUT) {
			rts_trace(TRACE_SD, "**CMD Response Timeout!*\n");
			sd_card->trans_abort = 1;
		}

		cr_rd_reg(cr, REG_SD40_STATUS1, &val);
		if (val & (SD40_HEAD_ERR | SD40_FRAM_ERR | SD40_CRC_ERR))
			sd_set_error_code(cr, card, SD_CRC_ERR);

		cr_rd_reg(cr, REG_SD40_RX_ARG0, &val);
		if (val & RES_NAK)
			rts_trace(TRACE_SD, "!!CCMD NAK Fail\n");
	}

	sd40_clear_error(cr);

	return ret;
}

static int legacy_cmd_has_r1b_rsp(u8 cmd)
{
	int ret = 0;

	switch (cmd) {
	case READ_MULTIPLE_BLOCK:
	case WRITE_MULTIPLE_BLOCK:
	case STOP_TRANSMISSION:
	case SET_WRITE_PROT:
	case CLR_WRITE_PROT:
	case ERASE_CMD:
		ret = 1;
		break;
	default:
		ret = 0;
		break;
	};

	return ret;
}

void dump_payload(struct card_reader *cr)
{
	u8 *ptr;

	cr_init_cmd(cr);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW3_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW2_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW1_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_RXPLD_DW0_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW3_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW3_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW3_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW3_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW2_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW2_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW2_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW2_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW1_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW1_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW1_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW0_3, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW0_2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW0_1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD,  REG_SD40_PLD_DW0_0, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_RX_HDR0, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_RX_HDR1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_RX_ARG0, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_RX_ARG1, 0xFF, 0);

	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_HDR0, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_HDR1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_ARG0, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_ARG1, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_ARG2, 0xFF, 0);
	cr_add_cmd(cr, READ_REG_CMD, REG_SD40_ARG3, 0xFF, 0);

	cr_send_cmd(cr, SD_ID, 200);
	ptr = cr_get_cmd_data(cr);

	rts_trace(TRACE_SD, "dump payload\n");
	RTS_DUMP(ptr, 42);

}

static int sd40_send_ccmd_packet(struct card *card, struct sd40_tlp *tlp)
{
	struct card_reader *cr = card->cm->cr;
	int ret = 0;
	u8 *ptr;

	sd_clear_error_code(cr, card);

	if (!(NATIVE_PACKET & tlp->header[0])) {
		u8 opcode = tlp->arg[1] & 0x3F;
		if (legacy_cmd_has_r1b_rsp(opcode)) {
			cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, SD40_HAS_EBSY
				| SD40_WAIT_EBSY | SD40_EBSY_WRITE_RES_TMOUT);
		}
	}

	add_cmd_4_ccmd(card, tlp);

	ret = cr_send_cmd(cr, SD_ID, tlp->timeout);
	if (!(NATIVE_PACKET & tlp->header[0]))
		cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, 0x00);

	if (ret) {
		process_error_4_ccmd(card, ret);
		TRACE_RET(cr, ret);
	}

	ptr = cr_get_cmd_data(cr) + 1;
	memcpy(tlp->response, ptr, sizeof(tlp->response));

	return 0;
}

int sd40_send_tlp(struct card *card, struct sd40_tlp *tlp)
{
	u8 cmd = tlp->header[0] & 0x70;

	if (HD_CCMD == cmd)
		return sd40_send_ccmd_packet(card, tlp);
	else if (HD_DCMD == cmd)
		return sd40_send_dcmd_packet(card, tlp);

	return -EINVAL;
}

int sd40_send_sdtran_dcmd(struct card *card, struct sd_command *cmd)
{
	struct sd_info *sd_card = card->extra;
	struct sd40_tlp tlp;
	int ret = 0;
	u8 cmd_code = cmd->opcode & 0x3F;
	u8 tmode = 0;

	memset(&tlp, 0, sizeof(tlp));

	tlp.header[0] = HD_DCMD | APP_PACKET | sd_card->sd40_did;
	tlp.header[1] = sd_card->sd40_sid_tid;

	if ((READ_MULTIPLE_BLOCK == cmd_code)
			|| (WRITE_MULTIPLE_BLOCK == cmd_code)) {
		/* Use LM_LEN_UNSPECIF mode for ACMD18 and ACMD25 */
		if (cmd->flags & SD_APP_CMD) {
			tmode = (DAM_INC_ADDR | TLUM_BLOCK | LM_LEN_UNSPECIF | 0);
		} else {
			tmode = DAM_INC_ADDR | TLUM_BLOCK;
			tmode |=  LM_LEN_SPECIF | sd_card->lan_mode;
		}
	} else {
		tmode = (DAM_INC_ADDR | TLUM_BLOCK | LM_LEN_UNSPECIF | 0);
	}

	tlp.arg[0] = tmode;
	if (TO_SD == cmd->data.trans_dir)
		tlp.arg[0] |= ARG_DIR_W;
	else
		tlp.arg[0] |= ARG_DIR_R;
	tlp.arg[1] = cmd->opcode & 0x3F;
	if (cmd->flags & SD_APP_CMD)
		tlp.arg[1] |= (1 << 6);

	rts_trace(TRACE_SD, "arg:%x bytes:%d\n", cmd->arg, cmd->data.bytes);

	tlp.payload[0] = (u8) (cmd->arg >> 24);
	tlp.payload[1] = (u8) (cmd->arg >> 16);
	tlp.payload[2] = (u8) (cmd->arg >> 8);
	tlp.payload[3] = (u8) cmd->arg;

	tlp.payload[4] = (u8) (cmd->data.bytes >> 24);
	tlp.payload[5] = (u8) (cmd->data.bytes >> 16);
	tlp.payload[6] = (u8) (cmd->data.bytes >> 8);
	tlp.payload[7] = (u8) cmd->data.bytes;

	tlp.scatter_gather = 0;
	tlp.data = cmd->data.buf;
	tlp.timeout = (cmd->timeout > 300) ? (cmd->timeout) : 300;

	ret = sd40_send_tlp(card, &tlp);
	if (!ret) {
		int rsp_len;
		rsp_len = (sizeof(cmd->rsp) > sizeof(tlp.response))
				? sizeof(tlp.response) : sizeof(cmd->rsp);
		memcpy(cmd->rsp, tlp.response, rsp_len);
	}

	return ret;
}

static int sd40_process_sdtran_ccmd_error(struct card *card, int error)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;

	rts_trace(TRACE_ERROR, "!SD40 SD-TRAN CMD Fail !\n");
	dump_all_sd40_register(cr);

	if (error == -ETIMEDOUT) {
		rts_trace(TRACE_ERROR, "****E-Timeout\n");
		sd_card->trans_abort = 1;
		sd_clear_sd_error(cr);
		sd40_trans_abort(card);
		sd_set_error_code(cr, card, SD_RSP_TIMEOUT);
	} else {
		u8 val, val1;
		cr_rd_reg(cr, REG_SD40_TRAN, &val);
		if ((val & SD40_TRANS_ERR)) {
			rts_trace(TRACE_ERROR, "**SD40_TRANS_ERR\n");
			cr_rd_reg(cr, REG_SD40_STATUS1, &val);
			val &= SD40_HEAD_ERR | SD40_FRAM_ERR | SD40_CRC_ERR;
			if (val) {
				sd_set_error_code(cr, card, SD_CRC_ERR);
				return error;
			}

			/* Check Command NAK */
			cr_rd_reg(cr, REG_SD40_RX_ARG0, &val);
			if (val & RES_NAK) {
				rts_trace(TRACE_ERROR, "Command NAK:%X\n", val);
				return error;
			}

			/* UNRECOVERABLE ERROR: Issue TRANS_ABORT or CMD12. */
			cr_rd_reg(cr, REG_SD40_TMOUT_0, &val);
			cr_rd_reg(cr, REG_SD40_TMOUT_1, &val1);

			val1 &= SD40_T_STAT_FCREQ_OUT | SD40_T_FCREQ_RDY_OUT;
			val1 &= SD40_T_FCRDY_EDB_OUT | SD40_T_EDB_STAT_OUT;
			if (val || val1) {
				sd_card->trans_abort = 1;
				sd40_trans_abort(card);
			}

			sd_set_error_code(cr, card, SD_STS_ERR);
			return error;
		}
	}

	return error;
}

int sd40_send_sdtran_ccmd(struct card *card, struct sd_command *cmd)
{
	struct sd_info *sd_card = card->extra;
	struct sd40_tlp tlp;
	int ret = 0;

	memset(&tlp, 0, sizeof(tlp));
	tlp.header[0] = HD_CCMD | APP_PACKET | sd_card->sd40_did;
	tlp.header[1] = sd_card->sd40_sid_tid;

	tlp.arg[0] = 0;
	tlp.arg[1] = cmd->opcode & 0x3F;
	if (cmd->flags & SD_APP_CMD)
		tlp.arg[1] |= (1 << 6);

	tlp.payload[0] = (u8) (cmd->arg >> 24);
	tlp.payload[1] = (u8) (cmd->arg >> 16);
	tlp.payload[2] = (u8) (cmd->arg >> 8);
	tlp.payload[3] = (u8) cmd->arg;
	tlp.timeout = (cmd->timeout > 300) ? (cmd->timeout) : 300;

	rts_trace(TRACE_SD, "sd_commad:cmd->arg:%x\n", cmd->arg);
	rts_trace(TRACE_SD, "%x %x %x %x\n", tlp.payload[0],
			tlp.payload[1], tlp.payload[2], tlp.payload[3]);

	ret = sd40_send_tlp(card, &tlp);
	if (!ret) {
		int rsp_len;
		rsp_len = (sizeof(cmd->rsp) > sizeof(tlp.response))
				? sizeof(tlp.response) : sizeof(cmd->rsp);
		memcpy(cmd->rsp, tlp.response, rsp_len);
	} else {
		sd40_process_sdtran_ccmd_error(card, ret);
	}

	return ret;
}

/* sd40_device_init - sent CCMD DEVICE_INIT to card
 *
 * @card
 * @payload_h: sent by host
 * @payload_d: received from device
 *
 */
int sd40_send_device_init(struct card *card, u16 payload_h, u16 *payload_d)
{
	struct sd_info *sd_card = card->extra;
	u16 header, arg;
	u8 payload[4] = {0};
	int ret = 0;

	rts_trace(TRACE_SD, "send DEVICE_INIT CCMD");


	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;
	arg = (ARG_DIR_W | PLEN_4B | SD40_IOADR_CMD_BASE) << 8;
	arg |= SD40_DEVICE_INIT;

	payload[0] = (u8)(payload_h >> 8);
	payload[1] = (u8)payload_h;

	ret = sd40_send_native_ccmd(card, header, arg, payload);
	*payload_d = (u16)(payload[0] << 8);
	*payload_d |= payload[1];

	return ret;
}
EXPORT_SYMBOL_GPL(sd40_send_device_init);

/* sd40_enumerate_device - send ENUMERATE ccmd
 *
 * @card
 * @node_id_h - node id set by host
 * @node_id_d - node id set by device
 *
 */
int sd40_enumerate_device(struct card *card, u8 node_id_h, u8 *node_id_d)
{
	struct sd_info *sd_card = card->extra;
	u16 header, arg;
	u8 payload[4] = {0};
	int ret = 0;

	rts_trace(TRACE_SD, "send ENUMERATE CCMD\n");

	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;

	arg = (ARG_DIR_W | PLEN_4B | SD40_IOADR_CMD_BASE) << 8;
	arg |= SD40_ENUMERATE;

	payload[0] = node_id_h;

	ret = sd40_send_native_ccmd(card, header, arg, payload);
	*node_id_d = payload[0];

	return ret;
}
EXPORT_SYMBOL_GPL(sd40_enumerate_device);

/* sd40_go_dormant_state - send GO_DORMANT_STATE ccmd
 *
 * @card
 * @dor_mode
 *
 */
int sd40_go_dormant_state(struct card *card, int dor_mode)
{
	struct sd_info *sd_card = card->extra;
	u16 header, arg;
	u8 payload[4] = {0};

	rts_trace(TRACE_SD, "send GO_DORMANT_STATE\n");

	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;

	arg = (ARG_DIR_W | PLEN_4B | SD40_IOADR_CMD_BASE) << 8;
	arg |= SD40_GO_DORMANT_STATE;

	payload[0] = (u8) dor_mode;

	return sd40_send_native_ccmd(card, header, arg, payload);
}
EXPORT_SYMBOL_GPL(sd40_go_dormant_state);

/* sd40_trans_abort - send TRANS_ABORT ccmd
 *
 * @card
 *
 */
int sd40_trans_abort(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u16 header, arg;

	rts_trace(TRACE_SD, "send TRANS_ABORT\n");

	if (CHECK_PID(cr, 0x524a) || CHECK_PID(cr, 0x525a))
		cr_wr_reg(cr, SD40_STOP_CTL , SD40_LINK_RST, SD40_LINK_RST);

	cr_wr_reg(cr, REG_SD40_CTRL, SD40_MAC_RST, SD40_MAC_RST);

	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;

	arg = (ARG_DIR_W | PLEN_0B | SD40_IOADR_CMD_BASE) << 8;
	arg |= SD40_TRANS_ABORT;

	return sd40_send_native_ccmd(card, header, arg, NULL);
}
EXPORT_SYMBOL_GPL(sd40_trans_abort);

/* sd40_read_cfg_reg - read IOADDR CFG
 *
 * @card
 * @addr
 * @cfg
 * @len
 */
int sd40_read_cfg_reg(struct card *card, u8 addr, u8 *cfg, int len)
{
	struct sd_info *sd_card = card->extra;
	u16 header, arg;

	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;

	switch (len) {
	case 16:
		arg = (ARG_DIR_R | PLEN_16B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 8:
		arg = (ARG_DIR_R | PLEN_8B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 4:
		arg = (ARG_DIR_R | PLEN_4B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 0:
	default:
		return -EINVAL;
	}

	arg |= addr;

	return sd40_send_native_ccmd(card, header, arg, cfg);
}
EXPORT_SYMBOL_GPL(sd40_read_cfg_reg);

/* sd40_write_cfg_reg - write IOADDR CFG
 *
 * @card
 * @addr
 * @cfg
 * @len
 */
int sd40_write_cfg_reg(struct card *card, u8 addr, u8 *cfg, int len)
{
	struct sd_info *sd_card = card->extra;
	u16 header, arg;

	header = (NATIVE_PACKET | HD_CCMD | sd_card->sd40_did) << 8;
	header |= sd_card->sd40_sid_tid;

	switch (len) {
	case 16:
		arg = (ARG_DIR_W | PLEN_16B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 8:
		arg = (ARG_DIR_W | PLEN_8B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 4:
		arg = (ARG_DIR_W | PLEN_4B | SD40_IOADR_CFG_BASE) << 8;
		break;
	case 0:
	default:
		return -EINVAL;
	}

	arg |= addr;

	return sd40_send_native_ccmd(card, header, arg, cfg);
}
EXPORT_SYMBOL_GPL(sd40_write_cfg_reg);

/* how to get response for write CCMD */
/* sd40_send_native_ccmd - send native CCMD
 *
 * @card:
 * @header:
 * @arg:
 * @payload:
 *
 */
int sd40_send_native_ccmd(struct card *card, u16 header, u16 arg, u8 *payload)
{
	struct sd40_tlp tlp;
	int len = 0;
	int ret = 0;
	u8 dir = ((u8)(arg >> 8)) & ARG_DIR;

	memset(&tlp, 0, sizeof(tlp));

	tlp.header[0] = (u8)(header >> 8);
	tlp.header[1] = (u8)header;
	tlp.arg[0] = (u8)(arg >> 8);
	tlp.arg[1] = (u8)arg;
	tlp.timeout = 300;

	len = sd40_get_payload_len((u8)(arg >> 8));
	if (len > MAX_SD40_PAYLOAD_LEN)
		return -EINVAL;

	if (ARG_DIR_W == dir && len > 0)
		memcpy(tlp.payload, payload, len);

	ret = sd40_send_tlp(card, &tlp);
	/* if (!ret && ARG_DIR_R == dir && len > 0) */
	if (len > 0)
		memcpy(payload, tlp.response, len);

	return ret;
}
EXPORT_SYMBOL_GPL(sd40_send_native_ccmd);

static int sd40_get_host_capabilitilies(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 val = 0;
	u16 addr = 0;

	addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_GEN_CAP_L) << 2;
	addr += 1;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	sd_card->cap.host.lan_mode = val & 0x3F;
	if (val & SD40_LANE_MODE_2L_HD)
		sd_card->cap.host.lan_2L_HD = 1;


	addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_PHY_CAP_H) << 2;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	sd_card->cap.host.n_lss = val;

	addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_CAP_H) << 2;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	sd_card->cap.host.gap = val;

	addr = (((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_CAP_L) << 2) + 1;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	sd_card->cap.host.n_fcu = val;

	return 0;
}

static int sd40_get_dev_generic_cap(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 payload[16];

	rts_trace(TRACE_SD, "Get Generic Capabilities\n");

	memset(payload, 0, sizeof(payload));
	if (sd40_read_cfg_reg(card, SD40_IOADR_GEN_CAP_L, payload, 8)) {
		rts_trace(TRACE_SD, "Get Generic Capabilities Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	sd_card->cap.dev.lan_mode = payload[6];
	if (payload[6] & SD40_LANE_MODE_2L_HD) {
		/* support 2L-HD mode. */
		rts_trace(TRACE_SD, "Device is support 2L-HD mode.\n");
		sd_card->cap.dev.lan_2L_HD = 1;
	}

	memcpy(sd_card->cap.dev.raw_generic, payload, 8);

	if ((payload[5] & (SD40_APP_TYPE_MASK)) != SD40_APP_TYPE_SD_MEMORY) {
		rts_trace(TRACE_SD, "SD40 application type: %x\n", payload[5]);
		rts_trace(TRACE_SD, "Device is not SD memory!\n");
	}

	return 0;
}

static int sd40_get_dev_phy_cap(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 payload[16];

	rts_trace(TRACE_SD, "Get PHY Capabilities\n");

	memset(payload, 0, sizeof(payload));
	if (sd40_read_cfg_reg(card, SD40_IOADR_PHY_CAP_L, payload, 8)) {
		rts_trace(TRACE_SD, "Get Generic Capabilities Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	sd_card->cap.dev.hibernate =
		payload[6] & SD40_DEV_PHY_HIBERNATE_SUPPORT;
	sd_card->cap.dev.n_lss = payload[3];

	memcpy(sd_card->cap.dev.raw_phy, payload, 8);

	rts_trace(TRACE_SD, "dev hibernate support:%x\n",
					sd_card->cap.dev.hibernate);
	rts_trace(TRACE_SD, "dev lss num:%x\n", sd_card->cap.dev.n_lss);

	return 0;
}

static int sd40_get_dev_link_cap(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 payload[16];

	rts_trace(TRACE_SD, "Get LINK/TRAN Capabilities\n");
	memset(payload, 0, sizeof(payload));

	if (sd40_read_cfg_reg(card, SD40_IOADR_LINK_CAP_L, payload, 8)) {
		rts_trace(TRACE_SD, "Get LINK/TRAN Capabilities Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	sd_card->cap.dev.gap = payload[3];
	sd_card->cap.dev.n_fcu = payload[6];

	memcpy(sd_card->cap.dev.raw_link, payload, 8);

	return 0;
}

static int sd40_get_device_capabilitilies(struct card *card)
{
	struct card_reader *cr = card->cm->cr;

	if (sd40_get_dev_generic_cap(card))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd40_get_dev_phy_cap(card))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd40_get_dev_link_cap(card))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static u8 get_n_lss(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct cr_option *option = &cr->option;
	struct sd_info *sd_card = card->extra;
	u8 lss_dir, lss_syn;
	u8 dev_lss, host_lss;

	dev_lss = sd_card->cap.dev.n_lss;
	host_lss = sd_card->cap.host.n_lss;

	if (((dev_lss & 0x0F) == 0) || ((host_lss & 0x0F) == 0)) {
		lss_syn = 0;
	} else {
		if ((dev_lss & 0x0F) > (host_lss & 0x0F))
			lss_syn = dev_lss & 0x0F;
		else
			lss_syn = host_lss & 0x0F;
	}

	if (check_dev_flag(cr, CONFIG_N_LSS_DIR)) {
		lss_dir = (((u8)option->n_lss_dir) & 0x0F) << 4;
	} else {
		if (CHECK_VERSION(cr, 0x5249, IC_VER_A)) {
			/* TODO: how to move this check to HOST layler */
			lss_dir  = 0;
		} else {
			if (((dev_lss & 0xF0) == 0)
					|| ((host_lss & 0xF0) == 0)) {
				lss_dir = 0;
			} else {
				if ((dev_lss & 0xF0) > (host_lss & 0xF0))
					lss_dir = dev_lss & 0xF0;
				else
					lss_dir = host_lss & 0xF0;
			}

		}
	}

	return (lss_dir | lss_syn);
}

static int sd40_configure_phy(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	u8 lss = 0;
	u8 payload[16];

	rts_trace(TRACE_SD, "Config PHY setting\n");

	lss = get_n_lss(card);

	memset(payload, 0, sizeof(payload));
	payload[3] = lss;

	if (sd40_write_cfg_reg(card, SD40_IOADR_PHY_SET_H, payload, 4)) {
		rts_trace(TRACE_SD, "Config PHY setting Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	} else {
		u16 addr;
		rts_trace(TRACE_SD, "Config PHY setting SUCCESS\n");
		addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_PHY_SET_H) << 2;
		cr_write_host_sd40_iospace(cr, addr, &lss, 1);
		rts_trace(TRACE_SD, "Host and dev LSS configed to %#x\n", lss);
	}

	return 0;
}

static int sd40_configure_link(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 gap = sd_card->cap.dev.gap;
	u8 payload[16];

	rts_trace(TRACE_SD, "Config Link/Tran setting\n");

	if (gap && gap < sd_card->cap.host.gap)
		gap = sd_card->cap.host.gap;

	/* TO fix bug of rts5249: refer to mantis: 3990 and 3659 */
	/* TODO: These will influence successor */
	if (2 == gap)
		gap = 3;

	memset(payload, 0, sizeof(payload));
	payload[3] = gap;
	if (sd40_write_cfg_reg(card, SD40_IOADR_LINK_SET_H, payload, 4)) {
		rts_trace(TRACE_SD, "Config Link/Tran setting Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	} else {
		u16 addr;

		rts_trace(TRACE_SD, "Config Link/Tran setting SUCCESS\n");

		addr = (SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_SET_H;
		addr <<= 2;
		cr_write_host_sd40_iospace(cr, addr, &gap, 1);
		rts_trace(TRACE_SD, "Host and dev gap configed to %#x\n", gap);
	}

	return 0;
}

static int sd40_set_lan_mode(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int host = !!(sd_card->cap.host.lan_mode & SD40_LANE_MODE_2L_HD);
	int device = !!(sd_card->cap.dev.lan_mode & SD40_LANE_MODE_2L_HD);

	if (check_sd40_flag(cr, SD40_HD_MODE) && host && device)
		sd_card->lan_mode = DM_2L_HD;

	if (card->reset_flags & FORBIT_USHII_HD_MODE)
		sd_card->lan_mode = DM_FD;

	/* TODO: set device lan mode */

	/* TODO: set host lan mode */

	return 0;
}

static int sd40_comm_set_power_control_mode(struct card *card, u8 mode)
{
	struct card_reader *cr = card->cm->cr;
	u8 payload[16], val;
	u16 addr;

	rts_trace(TRACE_SD, "SD40 Set Power Mode %#x\n", mode);

	memset(payload, 0, sizeof(payload));
	payload[3] = mode;

	if (sd40_write_cfg_reg(card, SD40_IOADR_GEN_SET_L, payload, 4)) {
		rts_trace(TRACE_SD, "sd40 set power mode Fail\n");
		dump_all_sd40_register(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}

	addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_GEN_SET_L) << 2;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	val &= 0xFE;
	val |= mode;
	cr_write_host_sd40_iospace(cr, addr, &val, 1);

	return 0;
}

static int sd40_set_power_control_mode(struct card *card, u8 mode)
{
	struct card_reader *cr = card->cm->cr;

	if (cr->sd_ops_prot->set_power_control_mode)
		return cr->sd_ops_prot->set_power_control_mode(card, mode);
	else
		return sd40_comm_set_power_control_mode(card, mode);
}

static int sd40_configure_generic(struct card *card)
{
	struct card_reader *cr = card->cm->cr;

	if (check_sd40_flag(cr, SD40_LOW_PWR_MODE)) {
		if (sd40_set_power_control_mode(card, LOW_PWR_MODE))
			TRACE_RET(cr, STATUS_FAIL);
	}

	sd40_set_lan_mode(cr, card);

	return 0;
}

/* update_setting_info - save setting info
 *
 * These info was used by scsi_get_uhsii_status
 */
static void update_setting_info(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	u8 payload[8];

	memset(payload, 0, sizeof(payload));
	sd40_read_cfg_reg(card, SD40_IOADR_GEN_SET_L, payload, 8);
	memcpy(sd_card->setting.dev.raw_generic, payload, 8);

	memset(payload, 0, sizeof(payload));
	sd40_read_cfg_reg(card, SD40_IOADR_LINK_SET_L, payload, 8);
	memcpy(sd_card->setting.dev.raw_link, payload, 8);

	memset(payload, 0, sizeof(payload));
	sd40_read_cfg_reg(card, SD40_IOADR_PHY_SET_L, payload, 8);
	memcpy(sd_card->setting.dev.raw_phy, payload, 8);
}

static int sd40_configure(struct card *card)
{

	if (sd40_configure_phy(card))
		return -1;

	if (sd40_configure_link(card))
		return -1;

	if (sd40_configure_generic(card))
		return -1;

	update_setting_info(card);

	return 0;
}

static inline u8 sd40_get_host_dap(struct card *card)
{
	return SD40_DAP_1000mA;
}

int sd40_common_take_dev_init_process(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	u8 group_desc = 0;
	int i = 0;
	int nack = 0;

	do {
		u16 payload = 0;
		u8 gap = SD40_GAP_1000mA;

		payload = ((group_desc << 4) | gap) << 8;
		payload |= ((sd40_get_host_dap(card)) | SD40_CF);

		if ((sd40_send_device_init(card, payload, &payload))) {
			rts_trace(TRACE_ERROR, "SD40 DEVICE_INIT CCMD Fail\n");
			if (++nack >= 5)
				TRACE_RET(cr, STATUS_FAIL);
			else
				continue;
		}

		if (payload & SD40_CF) {
			rts_trace(TRACE_SD, "CF' = 1\n");
			break;
		}

		rts_trace(TRACE_SD, "CF' = 0, retry: %d\n", i + 1);

		if (((payload >> 8) & SD40_GAP_MASK) == gap)
			group_desc++;

		/* delay 100 ms since 2nd time */
		if (i)
			msleep(100);

	} while ((++i < DEVICE_INIT_RETRY_CNT));

	/* normal sd40 card should check retry cnt,
	 * but realsil fpga card module don't need
	 */
	if (!(check_sd40_flag(cr, SD40_RS_FPGA_CARD))) {
		if (i >= DEVICE_INIT_RETRY_CNT) {
			rts_trace(TRACE_ERROR,
				"DEVICE_INIT retry times: %d\n", i);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return 0;
}
EXPORT_SYMBOL_GPL(sd40_common_take_dev_init_process);

int sd40_take_dev_init_process(struct card *card)
{
	struct card_reader *cr = card->cm->cr;

	if (cr->sd_ops_prot->take_sd40_init)
		return cr->sd_ops_prot->take_sd40_init(card);
	else
		return sd40_common_take_dev_init_process(card);
}
EXPORT_SYMBOL_GPL(sd40_take_dev_init_process);

static int sd40_take_enumeration_process(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	int i = 0;

	do {
		u8 node_id_h = 0;
		u8 node_id_d = 0;

		/* ID_F: 0, ID_L: 0 */
		if (sd40_enumerate_device(card, node_id_h, &node_id_d)) {
			rts_trace(TRACE_SD, "SD40 ENUMERATE CCMD Fail\n");
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* Host issues ENUMERATE CCMD with (ID_F, ID_L) = (0h, 0h),
		 * Device sets its Node ID to ID_F + 1
		 */
		if ((node_id_d & SD40_ID_F_MASK) == 0x10)
			break;

		/* retry, do nothing if retry fail. */
	} while (i++ < 3);

	return 0;
}

static int sd40_comm_set_n_fcu(struct card *c, u8 n_fcu)
{
	struct card_reader *cr = c->cm->cr;
	struct sd_info *sd_card = c->extra;
	u8 payload[16];
	u16 addr = 0;
	u8 retry = 0;
	u8 val = 0;


	rts_trace(TRACE_SD, "Set N_FCU CCMD Test\n");

	memset(payload, 0, sizeof(payload));
	if (sd40_read_cfg_reg(c, SD40_IOADR_LINK_SET_L, payload, 4)) {
		rts_trace(TRACE_ERROR, "read CFG register fail!\n");
		TRACE_RET(cr, STATUS_FAIL);
	}
	retry = payload[1];

	memset(payload, 0, sizeof(payload));
	payload[0] = 0x20;
	payload[1] = 64 == n_fcu ? 0 : retry;
	payload[2] = (u8)n_fcu;

	if (sd40_write_cfg_reg(c, SD40_IOADR_LINK_SET_L, payload, 4)) {
		rts_trace(TRACE_ERROR, "set N_FCU fail!!\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	addr = (((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_SET_L) << 2) + 1;
	cr_write_host_sd40_iospace(cr, addr, &n_fcu, 1);

	/* set host MAX_RETRY_NUM of Data Burst Retry */
	addr = (((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_SET_L) << 2) + 2;
	cr_read_host_sd40_iospace(cr, addr, &val, 1);
	val &= 0xFC;
	val |= 64 == n_fcu ? 0 : (retry & 0x3);
	cr_write_host_sd40_iospace(cr, addr, &val, 1);

	sd_card->setting.host.n_fcu = n_fcu;
	sd_card->setting.dev.n_fcu = n_fcu;

	return 0;
}

int sd40_set_n_fcu(struct card *c, u8 n_fcu)
{
	struct card_reader *cr = c->cm->cr;

	if (cr->sd_ops_prot->set_sd40_n_fcu)
		return cr->sd_ops_prot->set_sd40_n_fcu(c, n_fcu);
	else
		return sd40_comm_set_n_fcu(c, n_fcu);
}

/* TODO: sd40_configure_n_fcu should be called in sd40_configure_link ? */
static int sd40_configure_n_fcu(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 n_fcu = (u8)(cr->option.sd40_nfcu);
	u8 dev_n_fcu = sd_card->cap.dev.n_fcu;
	/* u8 host_n_fcu = sd_card->cap.host.n_fcu; */

	n_fcu = (u8)min(dev_n_fcu, n_fcu);
	/* n_fcu = (u8)min(host_n_fcu, n_fcu); */

	/* TODO: check these */
	if (2 == n_fcu || 4 == n_fcu || 8 == n_fcu || 16 == n_fcu
			|| 32 == n_fcu || 64 == n_fcu) {
		if (!sd40_set_n_fcu(card, n_fcu)) {
			rts_trace(TRACE_SD, "set N_FCU to %d\n", n_fcu);
			sd_card->n_fcu = n_fcu;
		}
	} else {
		sd_card->n_fcu = 1;
		rts_trace(TRACE_SD, "N_FCU remain 1");
	}

	return 0;
}

static int sd40_set_speed_range(struct card *card, u8 range)
{
	struct card_reader *cr = card->cm->cr;
	u8 payload[16];
	u16 addr;

	rts_trace(TRACE_SD, "Send set speed range CCMD\n");
	rts_trace(TRACE_SD, "range: %x\n", (u8)range);

	memset(payload, 0, sizeof(payload));
	payload[3] = (u8)range;

	if (sd40_write_cfg_reg(card, SD40_IOADR_PHY_SET_L, payload, 4)) {
		rts_trace(TRACE_SD, "Send set speed range CCMD Fail!!\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_PHY_SET_L) << 2;
	cr_write_host_sd40_iospace(cr, addr, &range, 1);

	update_setting_info(card);

	return 0;
}

/* TODO:
 * should sd40_configure_speed_range be called in sd40_configure_phy?
 */
static int sd40_configure_speed_range(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int ret;

	if(!(card->reset_flags & UHSII_TO_RANGE_A) &&
		(check_sd40_flag(cr, SD40_RANGE_B_EN))) {
		if (!sd40_set_speed_range(card, SPEED_SEL_RANGE_B)) {
			ret = sd40_process_domant_state(cr,
				card, ENTRY_DORMANT, DOR_NORMAL_MODE);
			msleep(10);
			ret = sd40_process_domant_state(cr,
						card, EXIT_DORMANT, 3);
			if (!ret) {
				rts_trace(TRACE_SD,
					"SD40 switch to RangeB success\n");
				sd_card->setting.host.speed_range =
					SPEED_SEL_RANGE_B;
			}
		}
	}

	return 0;
}

static int sd40_set_configuration_complete(struct card_reader *cr,
						struct card *c, u8 cfg_ok)
{
	u8 payload[16];
	struct sd_info *sd_card = c->extra;

	rts_trace(TRACE_SD, "******Send Set Config Completion***********\n");
	rts_trace(TRACE_SD, "*cfg_ok: %d\n", cfg_ok);

	memset(payload, 0, sizeof(payload));
	payload[0] = cfg_ok;
	if (sd40_write_cfg_reg(c, SD40_IOADR_GEN_SET_H, payload, 4)) {
		rts_trace(TRACE_SD, "Send Set Config Completion Fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	} else {
		u16 addr;
		u8 val;

		addr = ((SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_GEN_SET_H) << 2;
		addr += 3;

		cr_read_host_sd40_iospace(cr, addr, &val, 1);
		val &= 0x7F;

		if (cfg_ok == SD40_CONFIG_COMPLETE) {
			val |= SD40_CONFIG_COMPLETE;
			cr_write_host_sd40_iospace(cr, addr, &val, 1);
			sd_card->sd40_config = SD40_CONFIG_COMPLETE;
		} else {
			val |= SD40_CONFIG_UNCOMPLETE;
			cr_write_host_sd40_iospace(cr, addr, &val, 1);
			sd_card->sd40_config = SD40_CONFIG_UNCOMPLETE;
		}
	}

	return 0;
}

int sd40_initialization(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;

	rts_trace(TRACE_SD, "%s.....\n", __func__);

	sd_card->sd40_did = 0;
	sd_card->sd40_sid_tid = 0;
	sd_card->cap.dev.lan_mode = 0;
	sd_card->cap.dev.hibernate = 0;
	sd_card->n_fcu = 1;
	sd_card->trans_abort = 0;
	sd_card->cap.dev.lan_2L_HD = 0;

	if (check_sd40_flag(cr,  DO_NOT_SD40_INIT)) {
		rts_trace(TRACE_SD, "Ignore sd40 initialization\n");
		/* for device debug */
		sd_card->sd40_did = 1;
		goto IGNORE_SD40_INIT;
	}

	if (sd40_init_hw(cr, c)) {
		dump_all_sd40_register(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (check_sd40_flag(cr,  SD40_PHY_TEST_MODE))
		return STATUS_SUCCESS;

	/* Check SD40 present */
	msleep(10);

	if (check_sd40_flag(cr,  SKIP_SD40_INIT_CMD)) {
		/* go to config_completion. */
		rts_trace(TRACE_SD, "Ignore SD40 Init CMD.....\n");
		sd_card->sd40_did = 1;
		goto SET_CFG_COMPLETE;
	}

	/* for phison test and normal sd40 card, did should be 0
	 * but for our own fpga card module, did should be 1
	 */
	 /* set Device NODE ID	 debug */
	if (check_sd40_flag(cr,  SD40_RS_FPGA_CARD))
		sd_card->sd40_did = 1;

	if (sd40_get_host_capabilitilies(c))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "SD40 card DID:%d\n", sd_card->sd40_did);

	/* DEVICE_INIT */
	if (sd40_take_dev_init_process(c))
		TRACE_RET(cr, STATUS_FAIL);

	/* ENUMERATE */
	if (sd40_take_enumeration_process(c))
		TRACE_RET(cr, STATUS_FAIL);

	c->reset_flags |= CONCRETE_SD_CARD;

	sd_card->sd40_did = 1;

	if (sd40_get_device_capabilitilies(c))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd40_configure(c))
		TRACE_RET(cr, STATUS_FAIL);

SET_CFG_COMPLETE:

	sd_card->sd40_did = 1;
	sd40_configure_n_fcu(c);

	/* TODO: move out */
	if (check_sd40_flag(cr,  SD40_LOW_PWR_MODE))
		cr_wr_reg(cr, REG_SD40_TMOUT_1, SD40_TIME_OUT_ERR_EN, 0);

	if (!sd40_set_configuration_complete(cr, c, SD40_CONFIG_COMPLETE))
		rts_trace(TRACE_SD, "SD40 config success\n");

	/* TODO: move out */
	if (check_sd40_flag(cr,  SD40_LOW_PWR_MODE)) {
		u8 mask = SD40_TIME_OUT_ERR_EN;
		u8 val = SD40_TIME_OUT_ERR_EN;
		cr_wr_reg(cr, REG_SD40_TMOUT_1, mask, val);
	}

	/* sd40_configure_n_fcu(c); */
	sd40_configure_speed_range(c);

	sd40_init_hw_m(cr, c);

	SET_SD_HCXC(sd_card);
	rts_trace(TRACE_SD, "SD40 Init Success!!!");
	rts_trace(TRACE_SD, "SD40_RW_CM_TRAN: %x\n",
		check_sd40_flag(cr, SD40_RW_CM_TRAN));

	return 0;

IGNORE_SD40_INIT:
	return STATUS_FAIL;

}

int sd40_init_hw(struct card_reader *cr, struct card *card)
{
	if (cr->sd_ops_prot->init_sd40_hw) {
		return cr->sd_ops_prot->init_sd40_hw(cr, card);
	} else {
		rts_trace(TRACE_ERROR, "%s was not defined\n", __func__);
		return -1;
	}
}

int sd40_init_hw_m(struct card_reader *cr, struct card *card)
{
	if (cr->sd_ops_prot->init_sd40_hw)
		return cr->sd_ops_prot->init_sd40_hw_m(cr, card);

	return 0;
}

int sd40_uninit_hw(struct card_reader *cr, struct card *card)
{
	if (cr->sd_ops_prot->uninit_sd40_hw) {
		return cr->sd_ops_prot->uninit_sd40_hw(cr, card);
	} else {
		rts_trace(TRACE_ERROR, "%s was not defined\n", __func__);
		return -1;
	}
}
EXPORT_SYMBOL_GPL(sd40_uninit_hw);

int sd40_process_domant_state(struct card_reader *cr, struct card *c,
		int dir, int dor_mode)
{

	if (cr->sd_ops_prot->process_dormannt_state) {
		return cr->sd_ops_prot->process_dormannt_state(cr, c,
								dir, dor_mode);
	} else {
		rts_trace(TRACE_ERROR, "%s was not defined\n", __func__);
		return -1;
	}
}
EXPORT_SYMBOL_GPL(sd40_process_domant_state);


int scsi_get_sector_info(struct scsi_cmnd *srb, u32 *lba, u16 *num);

static u8 get_dma_packet_size_from_n_fcu(u8 n_fcu)
{
	switch (n_fcu) {
	case 2:
		return DMA_1024;
	case 4:
		return DMA_2048;
	case 8:
		return DMA_4096;
	case 16:
		return DMA_8192;
	case 32:
		return DMA_16384;
	case 64:
		return DMA_32768;
	case 0:
	case 1:
	default:
		return DMA_512;
	}

	return DMA_512;
}

static int continue_pre_read(struct card *card, struct scsi_cmnd *srb)
{
	u32 start_sector;
	u16 sector_cnt;
	u32 reserved = PRE_READ_RESERVED;
	struct sd_info * sd_card = card->extra;

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (!sd_check_seq(card, srb))
		return 0;

	if (DMA_FROM_DEVICE != srb->sc_data_direction)
		return 0;

	if ((sd_card->read_sectors + sector_cnt) > PRE_READ_LEN)
		return 0;

	if (card->cm->cr->option.sd40_seq_speedup_en) {
		if (sector_cnt % 2)
			return 0;
	} else {
		if (sd_card->prev_sec_cnt % sd_card->n_fcu)
			return 0;

		if (sector_cnt % sd_card->n_fcu)
			return 0;
	}

	if ((start_sector + sector_cnt + reserved) >= sd_card->capacity)
		return 0;

	if (card->cm->cr->option.sd40_seq_speedup_en)
		sd_card->sd40_seq_sectors += sector_cnt;

	rts_trace(TRACE_SD, "sd_card->sd40_seq_sectors=%#x \n",
		sd_card->sd40_seq_sectors);

	return 1;
}

static int continue_pre_write(struct card *card, struct scsi_cmnd *srb)
{
	u32 start_sector;
	u16 sector_cnt;
	u32 reserved = PRE_READ_RESERVED;
	struct sd_info * sd_card = card->extra;

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (!sd_check_seq(card, srb))
		return 0;

	if (DMA_TO_DEVICE != srb->sc_data_direction)
		return 0;

	if ((sd_card->write_sectors + sector_cnt) > PRE_READ_LEN)
		return 0;

	if (sd_card->prev_sec_cnt % sd_card->n_fcu)
		return 0;

	if (sector_cnt % sd_card->n_fcu)
		return 0;

	if ((start_sector + sector_cnt + reserved) >= sd_card->capacity)
		return 0;

	return 1;
}


static int sd40_prepare_pre_read(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);
	u32 start_sector, sector_cnt, data_addr;
	u8 val;
	int ret;

	rts_trace(TRACE_SD, "**SD/MMC CMD18\n");

	ret = scsi_get_sector_info(srb, &start_sector, (u16 *)&sector_cnt);
	if (ret)
		return ret;

	sector_cnt = PRE_READ_LEN;
	cr_wr_reg(cr, REG_PRE_RW_MODE, EN_INFINITE_MODE, EN_INFINITE_MODE);

	if (!CHK_SD_HCXC(sd_card) && !CHK_MMC_SECTOR_MODE(sd_card))
		data_addr = start_sector << 9;
	else
		data_addr = start_sector;

	cr_init_cmd(cr);

	/* debug Native DCMD for RLE0510; */
	if (check_sd40_flag(cr,  SD40_RW_CM_TRAN)) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, NATIVE_PACKET | HD_DCMD
				| (sd_card->sd40_did));
	} else {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, APP_PACKET |
				HD_DCMD | (sd_card->sd40_did));
	}

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF,
			sd_card->sd40_sid_tid);

	val = DAM_INC_ADDR | TLUM_BLOCK;
	val |= LM_LEN_UNSPECIF | sd_card->lan_mode;
	sd_card->tmode = val;

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0xFF, val);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0,
			0xFF, (u8)data_addr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1,
			0xFF, (u8)(data_addr >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2,
			0xFF, (u8)(data_addr >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3,
			0xFF, (u8)(data_addr >> 24));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_0,
			0xFF, (u8)sector_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_1,
			0xFF, (u8)(sector_cnt >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_2,
			0xFF, (u8)(sector_cnt >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_3,
			0xFF, (u8)(sector_cnt >> 24));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_3, 0xFF, 0);

	rts_trace(TRACE_SD, "**SD40_PLD_DW0(addr): %#x %#x %#x %#x\n",
			(u8)(data_addr >> 24), (u8)(data_addr >> 16),
			(u8)(data_addr >> 8), (u8)data_addr);
	rts_trace(TRACE_SD, "TMODE:%#x\n", sd_card->tmode);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, 0x80, ARG_DIR_R);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3, 0xFF, READ_MULTIPLE_BLOCK);
	cr_enable_dma(cr, sector_cnt * 512, ps, DMA_FROM_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);

	return cr_send_cmd(cr, SD_ID, 200);
}

static int sd40_prepare_pre_write(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);
	u32 start_sector, sector_cnt, data_addr;
	u8 val;
	int ret;

	rts_trace(TRACE_SD, "**SD/MMC CMD25\n");

	ret = scsi_get_sector_info(srb, &start_sector, (u16 *)&sector_cnt);
	if (ret)
		return ret;

	sector_cnt = PRE_READ_LEN;
	cr_wr_reg(cr, REG_PRE_RW_MODE, EN_INFINITE_MODE, EN_INFINITE_MODE);

	if (!CHK_SD_HCXC(sd_card) && !CHK_MMC_SECTOR_MODE(sd_card))
		data_addr = start_sector << 9;
	else
		data_addr = start_sector;

	cr_init_cmd(cr);

	/* debug Native DCMD for RLE0510; */
	if (check_sd40_flag(cr,  SD40_RW_CM_TRAN)) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, NATIVE_PACKET | HD_DCMD
				| (sd_card->sd40_did));
	} else {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, APP_PACKET |
				HD_DCMD | (sd_card->sd40_did));
	}

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF,
			sd_card->sd40_sid_tid);

	val = DAM_INC_ADDR | TLUM_BLOCK;
	val |= LM_LEN_UNSPECIF | sd_card->lan_mode;
	sd_card->tmode = val;

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0xFF, val);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0,
			0xFF, (u8)data_addr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1,
			0xFF, (u8)(data_addr >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2,
			0xFF, (u8)(data_addr >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3,
			0xFF, (u8)(data_addr >> 24));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_0,
			0xFF, (u8)sector_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_1,
			0xFF, (u8)(sector_cnt >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_2,
			0xFF, (u8)(sector_cnt >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN_LEN_3,
			0xFF, (u8)(sector_cnt >> 24));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_3, 0xFF, 0);

	rts_trace(TRACE_SD, "**SD40_PLD_DW0(addr): %#x %#x %#x %#x\n",
			(u8)(data_addr >> 24), (u8)(data_addr >> 16),
			(u8)(data_addr >> 8), (u8)data_addr);
	rts_trace(TRACE_SD, "TMODE:%#x\n", sd_card->tmode);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, 0x80, ARG_DIR_W);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3,
					0xFF, WRITE_MULTIPLE_BLOCK);
	cr_enable_dma(cr, sector_cnt * 512, ps, DMA_TO_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);

	return cr_send_cmd(cr, SD_ID, 200);
}

static int sd40_prepare_rw_base(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int ret = 0;

	if (PRE_READ == sd_card->rw_mode) {
		if (!continue_pre_read(card, srb)) {
			ret = sd_stop_sequential(card);
			sd_card->rw_mode = NORMAL_RW;
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
		}
	} else if (PRE_WRITE == sd_card->rw_mode) {
		if (!continue_pre_write(card, srb)) {
			ret = sd_stop_sequential(card);
			sd_card->rw_mode = NORMAL_RW;
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
		}
	} else {
		if (card->cm->cr->option.sd40_seq_speedup_en)
			sd_check_seq(card, srb);
		cr_wr_reg(cr, REG_SD40_LINK_SETTING_2, 0xF0, 0x00);
		cr_wr_reg(cr, REG_SD40_LINK_SETTING_3, 0xFF, 0x20);
	}

	return 0;
}

static int sd40_prepare_rw_sequential(struct card *card, struct scsi_cmnd *srb)
{
	/* TODO: */
	return -1;

}

static void sd40_trace_data_tlp(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u16 sector_cnt;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (check_sd40_flag(cr,  SD40_RW_CM_TRAN)) {
		rts_trace(TRACE_SD, "**SD40_HDR: 0x%X,0x%X\n",
			(u8)(NATIVE_PACKET | HD_DCMD | (sd_card->sd40_did)),
			(u8)(sd_card->sd40_sid_tid));
	} else {
		rts_trace(TRACE_SD, "**SD40_HDR: 0x%X,0x%X\n",
			(u8)(APP_PACKET | HD_DCMD | (sd_card->sd40_did)),
			(u8)(sd_card->sd40_sid_tid));
	}

	if (!(sd_card->tmode & LM_LEN_SPECIF)) {
		rts_trace(TRACE_SD, "**SD40_ARG2(TOMD): 0x%x\n",
				sd_card->lan_mode);
		rts_trace(TRACE_SD, "**SD40_PLD_DW1(len): 0x0,0x0\n");
	} else {
		rts_trace(TRACE_SD, "**SD40_ARG2(TOMD): 0x%x\n",
				0x20 | sd_card->lan_mode);
		rts_trace(TRACE_SD, "**SD40_PLD_DW1(len): 0x%X,0x%X\n",
				(u8)(sector_cnt >> 8), (u8)sector_cnt);
	}

	rts_trace(TRACE_SD, "**DMA size: 0x%X\n", ps);
}

static int sd40_prepare_normal_read(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u16 sector_cnt;
	int ret;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);

	rts_trace(TRACE_SD, "**SD/MMC CMD18\n");

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, 0x80, ARG_DIR_R);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3, 0xFF, READ_MULTIPLE_BLOCK);
	cr_enable_dma(cr, sector_cnt * 512, ps, DMA_FROM_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN, SD_TRANSFER_END,
							SD_TRANSFER_END);

	sd40_trace_data_tlp(card, srb);

	cr_send_cmd_no_wait(cr);

	return 0;
}

static int sd40_prepare_normal_write(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u16 sector_cnt;
	int ret;
	u8 ps = get_dma_packet_size_from_n_fcu(sd_card->n_fcu);

	rts_trace(TRACE_SD, "***SD/MMC CMD25\n");

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	if ((card->runtime_flags & PATCH_TOSHIBA_UHSII)
		&& (cr->option.sd40_ctl & FIX_CMD12_ISSUE)) {
		if ((sd_card->prev_dir == DMA_TO_DEVICE) &&
			(sector_cnt >= 2048))
			sd_card->write_at_LM_1++;
		else
			sd_card->write_at_LM_1 = 0;
	}

	cr_init_cmd(cr);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG0, 0x80, ARG_DIR_W);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG3, 0xFF,
						WRITE_MULTIPLE_BLOCK);
	cr_enable_dma(cr, sector_cnt * 512, ps, DMA_TO_DEVICE);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_TRAN, 0xFF, SD40_TRANS_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD40_TRAN,
					SD_TRANSFER_END, SD_TRANSFER_END);

	sd40_trace_data_tlp(card, srb);

	cr_send_cmd_no_wait(cr);

	return 0;
}

static int sd40_prepare_normal_rw_base(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int ret = 0;
	u32 start_sector, data_addr;
	u16 sector_cnt;
	u8 val;

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	if (!CHK_SD_HCXC(sd_card) && !CHK_MMC_SECTOR_MODE(sd_card))
		data_addr = start_sector << 9;
	else
		data_addr = start_sector;

	cr_wr_reg(cr, REG_SD40_EBSY, 0xC3,
		SD40_HAS_EBSY | SD40_WAIT_EBSY | SD40_STAT_EBSY_TMOUT);
	rts_trace(TRACE_SD, "Normal RW, wait EBSY\n");

	cr_init_cmd(cr);

	/* TODO: CHECK SD40_RW_CM_TRAN */
	/* debug Native DCMD for RLE0510; */
	if (check_sd40_flag(cr,  SD40_RW_CM_TRAN)) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, NATIVE_PACKET | HD_DCMD
				| (sd_card->sd40_did));
	} else {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR0,
				0xFF, APP_PACKET |
				HD_DCMD | (sd_card->sd40_did));
	}

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_HDR1, 0xFF,
			sd_card->sd40_sid_tid);

	val = DAM_INC_ADDR | TLUM_BLOCK;
	val |= LM_LEN_SPECIF | sd_card->lan_mode;
	sd_card->tmode = val;

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_ARG2, 0xFF, val);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_0,
			0xFF, (u8)data_addr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_1,
			0xFF, (u8)(data_addr >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_2,
			0xFF, (u8)(data_addr >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW0_3,
			0xFF, (u8)(data_addr >> 24));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_0,
			0xFF, (u8)sector_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_1,
			0xFF, (u8)(sector_cnt >> 8));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_2, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD40_PLD_DW1_3, 0xFF, 0);

	rts_trace(TRACE_SD, "data_addr: %#x\n", data_addr);

	return cr_send_cmd(cr, SD_CARD, 200);
}

static int sd40_prepare_normal_rw(struct card *card, struct scsi_cmnd *srb)
{
	int ret = 0;

	ret = sd40_prepare_normal_rw_base(card, srb);
	if (ret) {
		sd_clear_sd_error(card->cm->cr);
		card->rw_need_retry = 1;
		sd_set_error_code(card->cm->cr, card, SD_TO_ERR);
		return ret;
	}

	if (srb->sc_data_direction == DMA_FROM_DEVICE)
		return sd40_prepare_normal_read(card, srb);
	else
		return sd40_prepare_normal_write(card, srb);
}

static int can_go_pre_read_mode(struct card *card, struct scsi_cmnd *srb)
{
	u32 start_sector;
	u16 sector_cnt;
	u32 reserved = PRE_READ_RESERVED;
	struct sd_info *sd_card = card->extra;

	if (card->cm->cr->option.sd40_seq_speedup_en)
		rts_trace(TRACE_SD, "sd_card->seq_num=%#x \n", sd_card->seq_num);

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (!(card->cm->cr->option.adma_mode))
		return 0;

	if (DMA_FROM_DEVICE != srb->sc_data_direction)
		return 0;

	if (card->cm->cr->option.sd40_seq_speedup_en) {
		if (sector_cnt <= PRE_READ_TH
			&& sd_card->seq_num < 8)
			return 0;
	} else {
		if (sector_cnt <= PRE_READ_TH)
			return 0;

		if (sector_cnt % sd_card->n_fcu)
			return 0;
	}

	if ((start_sector + sector_cnt + reserved) >= sd_card->capacity)
		return 0;

	if (card->runtime_flags & FORBIT_UHSII_UNSPECIFIED_LEN_MODE)
		return 0;

	if (card->cm->cr->option.sd40_seq_speedup_en)
		sd_card->sd40_seq_sectors = sector_cnt;

	rts_trace(TRACE_SD, "sd_card->sd40_seq_sectors=%#x \n",
		sd_card->sd40_seq_sectors);

	return 1;
}

static int can_go_pre_write_mode(struct card *card, struct scsi_cmnd *srb)
{
	u32 start_sector;
	u16 sector_cnt;
	u32 reserved = PRE_READ_RESERVED;
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (!(card->cm->cr->option.adma_mode))
		return 0;

	if (DMA_TO_DEVICE != srb->sc_data_direction)
		return 0;

	if (card->cm->cr->option.sd40_seq_speedup_en) {
		if (sector_cnt <= PRE_READ_TH
			&& sd_card->seq_num < 8)
			return 0;
	} else {
		if (sector_cnt <= PRE_READ_TH)
			return 0;
	}

	if (sector_cnt % sd_card->n_fcu)
		return 0;

	if ((start_sector + sector_cnt + reserved) >= sd_card->capacity)
		return 0;

	if (card->runtime_flags & FORBIT_UHSII_UNSPECIFIED_LEN_MODE)
		return 0;

	if ((card->runtime_flags & PATCH_TOSHIBA_UHSII) &&
		(cr->option.sd40_ctl & FIX_CMD12_ISSUE)) {
		if (sd_card->write_at_LM_1 < 10)
			return 0;
	}

	return 1;
}


int sd_prepare_uhsii_rw(struct card *card, struct scsi_cmnd *srb)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int ret = 0;

	ret = sd40_prepare_rw_base(card, srb);
	if (ret) {
		sd_clear_sd_error(cr);
		card->rw_need_retry = 1;
		sd_set_error_code(cr, card, SD_TO_ERR);
		return ret;
	}

	if (SEQ_RW == sd_card->rw_mode) {
		return sd40_prepare_rw_sequential(card, srb);
	} else if ((PRE_READ == sd_card->rw_mode)
			|| (PRE_WRITE == sd_card->rw_mode)) {
		/* TODO: */
		/* DO NOTHING NOW */
	} else {
		if (check_sd40_flag(cr, SD40_PRE_READ_EN)
			&& can_go_pre_read_mode(card, srb)) {
			sd_card->rw_mode = PRE_READ;
			sd_card->read_sectors = 0;
			ret = sd40_prepare_pre_read(card, srb);
			cr_enable_bus_int(cr,
				cr->hw_param.interrupt_en | DATA_DONE_INT_EN);
		} else if (check_sd40_flag(cr, SD40_PRE_WRITE_EN)
			&& can_go_pre_write_mode(card, srb)) {
			sd_card->rw_mode = PRE_WRITE;
			sd_card->write_sectors = 0;
			ret = sd40_prepare_pre_write(card, srb);
			cr_enable_bus_int(cr,
				cr->hw_param.interrupt_en | DATA_DONE_INT_EN);
		} else {
			/* For UHSII, SEQ_RW was not supported by HW */
			/* if(sd_card->sd40_no_len_mode)
				sd_card->rw_mode = SEQ_RW;*/
			return sd40_prepare_normal_rw(card, srb);
		}
	}

	return 0;
}

int sd40_process_transfer_error(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	u8 stat;
	u8 val, val2, val3, val4;
	int retval;

	if (check_sd40_flag(cr,  SD40_NO_ERR_HANDLING)) {
		c->rw_need_retry = 0;
		rts_trace(TRACE_SD, "SD40 RW error, Skip all...\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (PRE_WRITE == sd_card->rw_mode || PRE_READ == sd_card->rw_mode) {
		u32 reg, value;

		value = cr_ioread32(cr, HCBCTLR);
		cr_iowrite32(cr, HCBCTLR, value | STOP_CMD);
		value = cr_ioread32(cr, HDBCTLR);
		cr_iowrite32(cr, HDBCTLR, value | STOP_DMA);

		sd40_dump_st(cr);
		if (CHECK_PID(cr, 0x524a) || CHECK_PID(cr, 0x525a))
			cr_wr_reg(cr, SD40_STOP_CTL , SD40_LINK_RST, SD40_LINK_RST);
		cr_wr_reg(cr, REG_SD40_CTRL, SD40_MAC_RST, SD40_MAC_RST);

		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_DMACTL, DMA_RST, DMA_RST);

		cr_wr_reg(cr, REG_PRE_RW_MODE, EN_INFINITE_MODE, 0);
		reg = cr->hw_param.interrupt_en & (~DATA_DONE_INT_EN);
		cr_enable_bus_int(cr, reg);
	}

	cr_rd_reg(cr, REG_SD40_TRAN, &stat);
	if (stat & SD40_TRANS_ERR) {
		u8 rsp[16];

		rts_trace(TRACE_SD, "RTS5249_SD40_TRAN error.\n");
		cr_rd_reg(cr, REG_SD40_TMOUT_0, &val);
		cr_rd_reg(cr, REG_SD40_TMOUT_1, &val2);
		cr_rd_reg(cr, REG_SD40_ST_REG_2, &val3);
		cr_rd_reg(cr, REG_SD40_ERR_ST, &val4);

		sd40_clear_error(cr);
		cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, 0x0);

		/* Send TRANS_ABORT and read ST_REG
		 * Tstat_fcreq(r), Tfcreq_fcrdy(w), Tcmd_res,
		 * UNRECOVERABLE_ERROR bit in STAT MSG or FCRDY MSG
		 */
		val2 &= SD40_T_STAT_FCREQ_OUT | SD40_T_FCREQ_RDY_OUT | SD40_T_FCRDY_EDB_OUT
			| SD40_T_RES_FCREQ_OUT;
		val3 &= RETRY_EXPIRE_ERROR;
		if (val || val2 || val3) {
			rts_trace(TRACE_SD,
				"TRANS_ABORT. val: 0x%X,0x%X,0x%X\n",
				val, val2, val3);
			retval = sd40_trans_abort(c);
		}

		if (!(val4 & SD40_NACK_ERR)) {
			rts_trace(TRACE_SD, "Send CMD12.");
			retval = sd_stop_transmission(c);
			if (retval)
				sd_set_error_code(cr, c, SD_STS_ERR);
			else
				rts_trace(TRACE_SD, "... OK\n");
		}

		sd_send_status(c, rsp);
	} else {
		sd40_clear_error(cr);
		if (!(sd_card->tmode & LM_LEN_SPECIF))
			sd_stop_seq_mode(cr, c);
	}

	return 0;
}
