/* Driver for Realtek card reader
 * Header file
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/kernel.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include <linux/vmalloc.h>

#include "cardreader.h"
#include "xd.h"
#include "trace.h"
#include "rts_scsi.h"
#include "scsi_cmd.h"

#define IMPLEMENT_XD_PROTOCOL

inline void xd_comm_set_err_code(struct card *c, u8 err_code)
{
	struct xd_info *xd_card = c->extra;

	xd_card->err_code = err_code;
}

inline int xd_comm_check_err_code(struct card *c, u8 err_code)
{
	struct xd_info *xd_card = c->extra;

	return (xd_card->err_code == err_code);
}


/*
 * cr_xd_prepare_init -- Prepare for xD initialization including enable clock,
 * etc.
 */
static int cr_xd_prepare_init(struct card_reader *cr)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot->prepare_init)
		ret = cr->xd_ops_prot->prepare_init(cr);
	else
		ret = STATUS_FAIL;

	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

/*
 * cr_xd_prepare_reset -- Prepare for xD reset
 */
static int cr_xd_prepare_reset(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	if (cr->xd_ops_prot->prepare_reset)
		ret = cr->xd_ops_prot->prepare_reset(cr, c);
	else
		ret = STATUS_FAIL;

	return ret;
}

/*
 * cr_xd_set_init_para - Set inital parameters including "switch clock"
 */
static int cr_xd_set_init_para(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	if (cr->xd_ops_prot->set_init_para)
		ret = cr->xd_ops_prot->set_init_para(cr, c);
	else
		ret = STATUS_FAIL;

	return ret;
}

/*
 * cr_xd_assign_phy_addr - Specify phy address
 * @addr: address specified
 * @mode: assign address mode, one of XD_RW_ADDR and XD_ERASE_ADDR
 */
int cr_xd_assign_phy_addr(struct card_reader *cr,
					struct card *c, u32 addr, u8 mode)
{
	return cr->xd_ops_prot->assign_phy_addr(cr, c, addr, mode);
}

/*
 * cr_xd_read_id - Read ID by id command READ_ID
 */
int cr_xd_read_id(struct card_reader *cr, struct card *c, u8 id_cmd,
						u8 *id_buf, u8 buf_len)
{
	return cr->xd_ops_prot->read_id(cr, c, id_cmd, id_buf, buf_len);
}

/*
 * cr_xd_read_id - Read 10-bytes redundant (no ECC bytes) and one-byte parity
 */
int cr_xd_read_redundant(struct card_reader *cr, struct card *c,
				u32 page_addr, u8 *buf, int buf_len)
{
	return cr->xd_ops_prot->read_redundant(cr, c, page_addr, buf, buf_len);
}

/*
 * cr_xd_read_cis - Read card initial space
 */
int cr_xd_read_cis(struct card_reader *cr, struct card *c,
				u32 page_addr, u8 *buf, int buf_len)
{
	return cr->xd_ops_prot->read_cis(cr, c, page_addr, buf, buf_len);
}

/*
 * cr_xd_search_cis - Search CIS block
 */
int cr_xd_search_cis(struct card_reader *cr, struct card *c)
{
	return cr->xd_ops_prot->search_cis(cr, c);
}

/*
 * cr_xd_tune_timing - Tune timinig
 */
int cr_xd_tune_timing(struct card_reader *cr, struct card *c,
								 int step)
{
	return cr->xd_ops_prot->tune_timing(cr, c, step);
}

/*
 * cr_xd_check_data_blank - Check whether the data is blank
 */
int cr_xd_check_data_blank(struct card_reader *cr,
								u8 *redunt)
{
	return cr->xd_ops_prot->check_data_blank(redunt);
}

/*
 * cr_xd_load_log_block_addr - Return log block address
 */
u16 cr_xd_load_log_block_addr(struct card_reader *cr,
								u8 *redunt)
{
	return cr->xd_ops_prot->load_log_block_addr(redunt);
}

/*
 * cr_xd_mark_bad_block - Mark bad block
 */
int cr_xd_mark_bad_block(struct card_reader *cr, struct card *c,
							u32 phy_blk)
{
	return cr->xd_ops_prot->mark_bad_block(cr, c, phy_blk);
}

/*
 * cr_xd_init_page - Initalise new pages from @start_page to @end_page
 */
int cr_xd_init_page(struct card_reader *cr, struct card *c,
			u32 phy_blk, u16 logoff, u8 start_page, u8 end_page)
{
	return cr->xd_ops_prot->init_page(cr, c, phy_blk, logoff,
							start_page, end_page);
}

/*
 * cr_xd_copy_page - Copy data of pages from @start_page to @end_page from old
 * block to new block
 */
int cr_xd_copy_page(struct card_reader *cr, struct card *c,
		u32 old_blk, u32 new_blk, u8 start_page, u8 end_page)
{
	return cr->xd_ops_prot->copy_page(cr, c, old_blk, new_blk,
							start_page, end_page);
}

/*
 * cr_xd_auto_copy_page - An speedup mode for page copy operation
 */
int cr_xd_auto_copy_page(struct card_reader *cr, struct card *c,
		     u32 old_blk, u32 new_blk, u8 start_page, u8 end_page)
{
	return cr->xd_ops_prot->auto_copy_page(cr, c, old_blk, new_blk,
							start_page, end_page);
}

/*
 * cr_xd_reset_cmd - Reset xd command
 */
int cr_xd_reset_cmd(struct card_reader *cr, struct card *c)
{
	return cr->xd_ops_prot->reset_cmd(cr, c);
}

/*
 * cr_xd_reset_cmd - Send xd command
 */
int cr_xd_send_cmd(struct card_reader *cr, struct card *c, u8 cmd)
{
	return cr->xd_ops_prot->send_cmd(cr, c, cmd);
}

/*
 * cr_xd_erase_block - Erase block @phy_blk
 */
static int cr_xd_erase_block(struct card_reader *cr, struct card *c,
								u32 phy_blk)
{
	return cr->xd_ops_prot->erase_block(cr, c, phy_blk);
}

/*
 * cr_xd_read_multiple_pages - Read multiple pages to @buf
 * @cr: points to card reader structure
 * @c: points to card structure
 * @srb: SCSI command
 * @phy_blk: phyical block address
 * @log_blk: logical block address
 * @start_page: starting page
 * @end_page: ending page
 * @buf: where to store data
 * @ptr: points to data for the next transfer
 * @offset: data of interest
 */
static int cr_xd_read_multiple_pages(struct card_reader *cr,
			struct card *c, struct scsi_cmnd *srb,
			u32 phy_blk, u32 log_blk, u8 start_page, u8 end_page,
			u8 *buf, void **ptr, unsigned int *offset)
{
	return cr->xd_ops_prot->read_multiple_pages(cr, c, srb, phy_blk,
			log_blk, start_page, end_page, buf, ptr, offset);
}

/*
 * cr_xd_write_multiple_pages - Write @buf to multiple pages
 * @cr: points to card reader structure
 * @c: points to card structure
 * @srb: SCSI command
 * @phy_blk: phyical block address
 * @log_blk: logical block address
 * @start_page: starting page
 * @end_page: ending page
 * @buf: data to be written
 * @ptr: points to data for the next transfer
 * @offset: data of interest
 */
static int cr_xd_write_multiple_pages(struct card_reader *cr,
			struct card *c, struct scsi_cmnd *srb,
			u32 old_blk, u32 new_blk, u32 log_blk, u8 start_page,
			u8 end_page, u8 *buf, void **ptr, unsigned int *offset)
{
	return cr->xd_ops_prot->write_multiple_pages(cr, c, srb, old_blk,
				new_blk, log_blk, start_page, end_page,
							buf, ptr, offset);
}

/* Read ID to check if the timing setting is right */
static int xd_comm_check_timing(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	int i, j;
	u8 id_buf[4];
	int retval = 0;

	for (i = 0; i < 4; i++) {
		/* u8 xd_dat, xd_ctl; */
		if (cr_detect_card(cr, XD_ID) == CD_NOT_EXIST)
			TRACE_RET(cr, STATUS_FAIL);
		retval = cr_xd_tune_timing(cr, c, i);
		if (retval != STATUS_SUCCESS)
			continue;
		retval = cr_xd_read_id(cr, c, READ_ID, id_buf, 4);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, retval);

		rts_trace(TRACE_XD, "READ_ID: 0x%x 0x%x 0x%x 0x%x\n",
			       id_buf[0], id_buf[1], id_buf[2], id_buf[3]);

		xd_card->device_code = id_buf[1];

		/* Check if the xD card is supported */
		switch (xd_card->device_code) {
		case XD_4M_X8_512_1:
		case XD_4M_X8_512_2:
			xd_card->block_shift = 4;	/* 16 pages per block */
			xd_card->page_off = 0x0F;
			xd_card->addr_cycle = 3;
			xd_card->zone_cnt = 1;
			xd_card->capacity = 8000;	/* 500 * 2 ^ 4 */
			XD_SET_4MB(xd_card);
			break;
		case XD_8M_X8_512:
			xd_card->block_shift = 4;
			xd_card->page_off = 0x0F;
			xd_card->addr_cycle = 3;
			xd_card->zone_cnt = 1;
			xd_card->capacity = 16000;	/* 1000 * 2 ^ 4 */
			break;
		case XD_16M_X8_512:
			XD_PAGE_512(xd_card);	/* 32 pages per block */
			xd_card->addr_cycle = 3;
			xd_card->zone_cnt = 1;
			xd_card->capacity = 32000;	/* 1000 * 2 ^ 5 */
			break;
		case XD_32M_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 3;
			xd_card->zone_cnt = 2;
			xd_card->capacity = 64000;	/* 2000 * 2 ^ 5 */
			break;
		case XD_64M_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 4;
			xd_card->capacity = 128000;	/* 4000 * 2 ^ 5 */
			break;
		case XD_128M_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 8;
			xd_card->capacity = 256000;	/* 8000 * 2 ^ 5 */
			break;
		case XD_256M_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 16;
			xd_card->capacity = 512000;	/* 16000 * 2 ^ 5 */
			break;
		case XD_512M_X8:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 32;
			xd_card->capacity = 1024000;	/* 32000 * 2 ^ 5 */
			break;
		case xD_1G_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 64;
			xd_card->capacity = 2048000;	/* 64000 * 2 ^ 5 */
			break;
		case xD_2G_X8_512:
			XD_PAGE_512(xd_card);
			xd_card->addr_cycle = 4;
			xd_card->zone_cnt = 128;
			xd_card->capacity = 4096000;	/* 128000 * 2 ^ 5 */
			break;
		default:
			continue;
		}

		/* Confirm timing setting */
		for (j = 0; j < 10; j++) {
			retval = cr_xd_read_id(cr, c, READ_ID, id_buf, 4);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, retval);

			if (id_buf[1] != xd_card->device_code)
				break;
		}

		/* Current timing pass */
		if (j == 10)
			break;
	}

	if (i == 4) {
		xd_card->block_shift = 0;
		xd_card->page_off = 0;
		xd_card->addr_cycle = 0;
		xd_card->capacity = 0;

		TRACE_RET(cr, STATUS_FAIL);
	}

	return retval;
}


#if 0 /* edwin: comment out due to the Macro XD_GPG/XD_GBLK isn't
       * hw-independent */
/* Search CIS block */
static int xd_comm_search_cis(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	u8 redunt[11];
	int i, j;
	int retval = 0;

	for (i = 0; i < 24; i++) {
		u32 page_addr;
#if 0
		if (cr_monitor_card(cr, XD_CARD) == CD_NOT_EXIST)
#else
		if (cr_detect_card(cr, XD_ID) == CD_NOT_EXIST)
#endif
			TRACE_RET(cr, STATUS_FAIL);

		page_addr = (u32) i << xd_card->block_shift;

		for (j = 0; j < 3; j++) {
#if 0
			retval = xd_read_redundant(c, page_addr, redunt, 11);
#else
			retval = cr_xd_read_redundant(cr, c, page_addr,
								redunt, 11);
#endif
			if (retval == STATUS_SUCCESS)
				break;
		}
		if (j == 3)
			continue;
		/* Skip bad blocks, Search 1st good block */
		if (redunt[BLOCK_STATUS] != XD_GBLK)
			continue;

		j = 0;
		/* Check page status */
		if (redunt[PAGE_STATUS] != XD_GPG) {
			for (j = 1; j <= 8; j++) {
#if 0
				retval =
				    xd_read_redundant(c, page_addr + j,
						      redunt, 11);
#else
				retval = cr_xd_read_redundant(cr, c,
						page_addr + j, redunt, 11);
#endif
				if (retval == STATUS_SUCCESS) {
					if (redunt[PAGE_STATUS] == XD_GPG)
						break;
				}
			}

			/* No good page in 1st 8 pages of 1st good block
			 * Return can not find CIS block */
			if (j == 9)
				break;
		}
		/* Check CIS data */
		if ((redunt[BLOCK_STATUS] == XD_GBLK)
		    && (redunt[PARITY] & XD_BA1_ALL0)) {
			u8 buf[10];

			page_addr += j;
#if 0
			retval = xd_read_cis(c, page_addr, buf, 10);
#else
			retval = cr_xd_read_cis(cr, c, page_addr, buf, 10);
#endif
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, retval);

			if ((buf[0] == 0x01) && (buf[1] == 0x03)
			    && (buf[2] == 0xD9)
			    && (buf[3] == 0x01) && (buf[4] == 0xFF)
			    && (buf[5] == 0x18) && (buf[6] == 0x02)
			    && (buf[7] == 0xDF) && (buf[8] == 0x01)
			    && (buf[9] == 0x20)) {
				xd_card->cis_block = (u16) i;
			}
		}

		break;
	}

	rts_trace(TRACE_XD, "CIS block: 0x%x\n", xd_card->cis_block);
	if (xd_card->cis_block == 0xFFFF)
		TRACE_RET(cr, STATUS_FAIL);

	return retval;

}
#endif
static int xd_comm_reset_xd(struct card_reader *cr, struct card *c)
{
	u8 id_buf[4];/* , redunt[11]; */
	int retval = 0;

	rts_trace(TRACE_XD, "<--\n");

	retval = cr_xd_prepare_reset(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	retval = cr_xd_set_init_para(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	/* Read ID to check if the timing setting is right */
	retval = xd_comm_check_timing(cr, c);

	/* Send Read ID 0x9A for xD certification */
	retval = cr_xd_read_id(cr, c, READ_xD_ID, id_buf, 4);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);
	rts_trace(TRACE_XD, "READ_xD_ID: 0x%x 0x%x 0x%x 0x%x\n",
		       id_buf[0], id_buf[1], id_buf[2], id_buf[3]);
	if (id_buf[2] != XD_ID_CODE)
		TRACE_RET(cr, STATUS_FAIL);
	/* Search CIS block */
	/* retval = xd_comm_search_cis(cr, c); */
	retval = cr_xd_search_cis(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);

	rts_trace(TRACE_XD, "-->\n");

	return STATUS_SUCCESS;
}

static inline void xd_comm_free_zone(struct zone_entry *zone)
{
	rts_trace(TRACE_XD, "free_zone\n");
	if (!zone)
		return;
	zone->build_flag = 0;
	zone->set_index = 0;
	zone->get_index = 0;
	zone->unused_blk_cnt = 0;
	if (zone->l2p_table) {
		vfree(zone->l2p_table);
		zone->l2p_table = NULL;
	}
	if (zone->free_table) {
		vfree(zone->free_table);
		zone->free_table = NULL;
	}
}

void xd_comm_set_unused_block(struct card_reader *cr,
					struct card *c,	u32 phy_blk)
{
	struct xd_info *xd_card = c->extra;
	struct zone_entry *zone;
	int zone_no;

	zone_no = (int)phy_blk >> 10;
	if (zone_no >= xd_card->zone_cnt) {
		rts_trace(TRACE_XD, "Set unused block to invalid zone"
					"(zone_no = %d, zone_cnt = %d)\n",
					zone_no, xd_card->zone_cnt);
		return;
	}
	zone = &(xd_card->zone[zone_no]);

	if (zone->free_table == NULL) {
		if (xd_comm_build_l2p_tbl(cr, c, zone_no) != STATUS_SUCCESS)
			return;
	}

	if ((zone->set_index >= XD_FREE_TABLE_CNT)
	    || (zone->set_index < 0)) {
		xd_comm_free_zone(zone);
		rts_trace(TRACE_XD,
			"Set unused block fail, invalid set_index\n");
		return;
	}

	rts_trace(TRACE_XD, "Set unused block to index %d\n", zone->set_index);

	zone->free_table[zone->set_index++] = (u16) (phy_blk & 0x3ff);
	if (zone->set_index >= XD_FREE_TABLE_CNT)
		zone->set_index = 0;
	zone->unused_blk_cnt++;
}

u32 xd_comm_get_unused_block(struct card_reader *cr,
						struct card *c, int zone_no)
{
	struct xd_info *xd_card = c->extra;
	struct zone_entry *zone;
	u32 phy_blk;

	if (zone_no >= xd_card->zone_cnt) {
		rts_trace(TRACE_XD, "Get unused block from invalid zone"
					"(zone_no = %d, zone_cnt = %d)\n",
					zone_no, xd_card->zone_cnt);
		TRACE_RET(cr, BLK_NOT_FOUND);
	}
	zone = &(xd_card->zone[zone_no]);

	if ((zone->unused_blk_cnt == 0) ||
			(zone->set_index == zone->get_index)) {
		xd_comm_free_zone(zone);
		rts_trace(TRACE_XD, "Get unused block fail,"
					"no unused block available\n");
		TRACE_RET(cr, BLK_NOT_FOUND);
	}
	if ((zone->get_index >= XD_FREE_TABLE_CNT) || (zone->get_index < 0)) {
		xd_comm_free_zone(zone);
		rts_trace(TRACE_XD,
			"Get unused block fail, invalid get_index\n");
		TRACE_RET(cr, BLK_NOT_FOUND);
	}

	rts_trace(TRACE_XD, "Get unused block from index %d\n",
		zone->get_index);

	phy_blk = zone->free_table[zone->get_index];
	zone->free_table[zone->get_index++] = 0xFFFF;
	if (zone->get_index >= XD_FREE_TABLE_CNT)
		zone->get_index = 0;
	zone->unused_blk_cnt--;

	phy_blk += ((u32) (zone_no) << 10);
	return phy_blk;
}

int xd_comm_prepare_write(struct card_reader *cr, struct card *c,
			    u32 old_blk, u32 new_blk, u32 log_blk, u8 page_off)
{
	int retval;

	rts_trace(TRACE_XD, "xd_prepare_write, old_blk = 0x%x, new_blk = 0x%x,"
				"log_blk = 0x%x, page_off = %d\n",
				old_blk, new_blk, log_blk, (int)page_off);

	if (page_off) {
#ifdef XD_SPEEDUP
		retval = cr_xd_auto_copy_page(cr, c, old_blk, new_blk, 0,
								page_off);
#else
		retval = cr_xd_copy_page(cr, c, old_blk, new_blk, 0, page_off);
#endif
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, retval);
	}

	return STATUS_SUCCESS;
}

int xd_comm_finish_write(struct card_reader *cr, struct card *c,
			   u32 old_blk, u32 new_blk, u32 log_blk, u8 page_off)
{
	struct xd_info *xd_card = c->extra;
	int retval, zone_no;
	u16 log_off;

	rts_trace(TRACE_XD, "xd_finish_write, old_blk = 0x%x, new_blk = 0x%x,"
				"log_blk = 0x%x\n", old_blk, new_blk, log_blk);

	if (page_off > xd_card->page_off)
		TRACE_RET(cr, STATUS_FAIL);

	zone_no = (int)(log_blk / 1000);
	log_off = (u16) (log_blk % 1000);

	if (old_blk == BLK_NOT_FOUND) {
		retval = cr_xd_init_page(cr, c, new_blk, log_off,
				      page_off, xd_card->page_off + 1);
		if (retval != STATUS_SUCCESS) {
			retval = cr_xd_erase_block(cr, c, new_blk);
			if (retval == STATUS_SUCCESS)
				xd_comm_set_unused_block(cr, c, new_blk);
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else {
#ifdef XD_SPEEDUP
		retval = cr_xd_auto_copy_page(cr, c, old_blk, new_blk,
					   page_off, xd_card->page_off + 1);
#else
		retval = cr_xd_copy_page(cr, c, old_blk, new_blk,
				      page_off, xd_card->page_off + 1);
#endif
		if (retval != STATUS_SUCCESS) {
			if (!XD_CHK_BAD_NEWBLK(xd_card)) {
				retval = cr_xd_erase_block(cr, c, new_blk);
				if (retval == STATUS_SUCCESS)
					xd_comm_set_unused_block(cr, c,
								new_blk);
			}
			XD_CLR_BAD_NEWBLK(xd_card);
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = cr_xd_erase_block(cr, c, old_blk);
		if (retval == STATUS_SUCCESS) {
			if (XD_CHK_BAD_OLDBLK(xd_card)) {
				cr_xd_mark_bad_block(cr, c, old_blk);
				XD_CLR_BAD_OLDBLK(xd_card);
			} else {
				/* Add source block to unused block */
				xd_comm_set_unused_block(cr, c, old_blk);
			}
		} else {
			xd_comm_set_err_code(c, XD_NO_ERROR);
			XD_CLR_BAD_OLDBLK(xd_card);
		}
	}

	/* Add target block to L2P table */
	xd_comm_set_l2p_tbl(c, zone_no, log_off, (u16) (new_blk & 0x3FF));

	return STATUS_SUCCESS;
}

int xd_comm_switch_clock(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	int retval;

	retval = cr_select_card(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);

	retval = cr_switch_clk(cr, c, xd_card->xd_clock);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);

	return STATUS_SUCCESS;
}

int xd_comm_delay_write(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	struct xd_delay_write_tag *delay_write = &(xd_card->delay_write);
	int retval;

	if (delay_write->delay_write_flag) {
		rts_trace(TRACE_XD, "xd_delay_write\n");
		retval = xd_comm_switch_clock(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, retval);

		delay_write->delay_write_flag = 0;
		retval = xd_comm_finish_write(cr, c,
					 delay_write->old_phyblock,
					 delay_write->new_phyblock,
					 delay_write->logblock,
					 delay_write->pageoff);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, retval);
	}

	return STATUS_SUCCESS;
}

void xd_comm_set_l2p_tbl(struct card *c, int zone_no,
						u16 log_off, u16 phy_off)
{
	struct xd_info *xd_card = c->extra;
	struct zone_entry *zone;

	zone = &(xd_card->zone[zone_no]);
	zone->l2p_table[log_off] = phy_off;
}

u32 xd_comm_get_l2p_tbl(struct card_reader *cr, struct card *c,
						int zone_no, u16 log_off)
{
	struct xd_info *xd_card = c->extra;
	struct zone_entry *zone;
	int retval;

	zone = &(xd_card->zone[zone_no]);
	if (zone->l2p_table[log_off] == 0xFFFF) {
		u32 phy_blk = 0;
		int i;
		retval = xd_comm_delay_write(cr, c);
		if (retval != STATUS_SUCCESS) {
			rts_trace(TRACE_XD, "In xd_get_l2p_tbl,"
						"delay write fail!\n");
			TRACE_RET(cr, BLK_NOT_FOUND);
		}

		if (zone->unused_blk_cnt <= 0) {
			rts_trace(TRACE_XD, "No unused block!\n");
			TRACE_RET(cr, BLK_NOT_FOUND);
		}

		for (i = 0; i < zone->unused_blk_cnt; i++) {
			phy_blk = xd_comm_get_unused_block(cr, c, zone_no);
			if (phy_blk == BLK_NOT_FOUND) {
				rts_trace(TRACE_XD,
					"No unused block available!\n");
				TRACE_RET(cr, BLK_NOT_FOUND);
			}
			retval = cr_xd_init_page(cr, c, phy_blk, log_off, 0,
							xd_card->page_off + 1);
			if (retval == STATUS_SUCCESS)
				break;
		}
		if (i >= zone->unused_blk_cnt) {
			rts_trace(TRACE_XD,
				"No good unused block available!\n");
			TRACE_RET(cr, BLK_NOT_FOUND);
		}
		xd_comm_set_l2p_tbl(c, zone_no, log_off,
						(u16) (phy_blk & 0x3FF));
		return phy_blk;
	}

	return (u32) zone->l2p_table[log_off] + ((u32) (zone_no) << 10);
}

static int xd_comm_init_l2p_tbl(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	int size, i;

	rts_trace(TRACE_XD, "xd_comm_init_l2p_tbl: zone_cnt = %d\n",
							xd_card->zone_cnt);

	if (xd_card->zone_cnt < 1)
		TRACE_RET(cr, STATUS_FAIL);

	size = xd_card->zone_cnt * sizeof(struct zone_entry);
	rts_trace(TRACE_XD, "Buffer size for l2p table is %d\n", size);

	xd_card->zone = vmalloc(size);
	if (!xd_card->zone)
		TRACE_RET(cr, STATUS_NOMEM);

	for (i = 0; i < xd_card->zone_cnt; i++) {
		xd_card->zone[i].build_flag = 0;
		xd_card->zone[i].l2p_table = NULL;
		xd_card->zone[i].free_table = NULL;
		xd_card->zone[i].get_index = 0;
		xd_card->zone[i].set_index = 0;
		xd_card->zone[i].unused_blk_cnt = 0;
	}

	return STATUS_SUCCESS;
}

int xd_comm_build_l2p_tbl(struct card_reader *cr, struct card *c,
								int zone_no)
{
	struct xd_info *xd_card = c->extra;
	struct zone_entry *zone;
	int retval;
	u32 start, end, i;
	u16 max_logoff, cur_fst_page_logoff, cur_lst_page_logoff,
	    ent_lst_page_logoff;
	u8 redunt[11];

	rts_trace(TRACE_XD, "xd_comm_build_l2p_tbl: %d\n", zone_no);

	if (xd_card->zone == NULL) {
		retval = xd_comm_init_l2p_tbl(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, retval);
	}

	if (xd_card->zone[zone_no].build_flag) {
		rts_trace(TRACE_XD, "l2p table of zone %d has been built\n",
			       zone_no);
		return STATUS_SUCCESS;
	}

	zone = &(xd_card->zone[zone_no]);

	if (zone->l2p_table == NULL) {
		zone->l2p_table = vmalloc(2000);
		if (zone->l2p_table == NULL)
			TRACE_GOTO(cr, Build_Fail);
	}
	memset((u8 *) (zone->l2p_table), 0xff, 2000);

	if (zone->free_table == NULL) {
		zone->free_table = vmalloc(XD_FREE_TABLE_CNT * 2);
		if (zone->free_table == NULL)
			TRACE_GOTO(cr, Build_Fail);
	}
	memset((u8 *) (zone->free_table), 0xff, XD_FREE_TABLE_CNT * 2);

	if (zone_no == 0) {
		/* If we didn't find CIS block, we build L2P table
		 * from block 0. Or else, we build L2P table from
		 * the block just after CIS block.
		 */
		if (xd_card->cis_block == 0xFFFF)
			start = 0;
		else
			start = xd_card->cis_block + 1;
		if (XD_CHK_4MB(xd_card)) {
			end = 0x200;
			max_logoff = 499;
		} else {
			end = 0x400;
			max_logoff = 999;
		}
	} else {
		start = (u32) (zone_no) << 10;
		end = (u32) (zone_no + 1) << 10;
		max_logoff = 999;
	}

	rts_trace(TRACE_XD, "start block 0x%x, end block 0x%x\n", start, end);

	zone->set_index = zone->get_index = 0;
	zone->unused_blk_cnt = 0;

	for (i = start; i < end; i++) {
		u32 page_addr = i << xd_card->block_shift;
		u32 phy_block;
		retval = cr_xd_read_redundant(cr, c, page_addr, redunt, 11);
		if (retval != STATUS_SUCCESS)
			continue;

		if (redunt[BLOCK_STATUS] != 0xFF) {
			rts_trace(TRACE_XD, "bad block\n");
			continue;
		}
		if (cr_xd_check_data_blank(cr, redunt)) {
			rts_trace(TRACE_XD, "blank block\n");
			xd_comm_set_unused_block(cr, c, i);
			continue;
		}
		cur_fst_page_logoff = cr_xd_load_log_block_addr(cr, redunt);
		/* rts_trace(TRACE_XD, "cur_fst_page_logoff = 0x%x\n",
						cur_fst_page_logoff); */
		if ((cur_fst_page_logoff == 0xFFFF)
		    || (cur_fst_page_logoff > max_logoff)) {
			/* rts_trace(TRACE_XD, "erase invalid block\n"); */
			retval = cr_xd_erase_block(cr, c, i);
			if (retval == STATUS_SUCCESS)
				xd_comm_set_unused_block(cr, c, i);
			continue;
		}
		/* For MBR not found or MBR data status=0  */
		if ((zone_no == 0) && (cur_fst_page_logoff == 0)
		    && (redunt[PAGE_STATUS] != XD_GPG))
			XD_SET_MBR_FAIL(xd_card);
		/* This logical block has not mapped to a physical block,
		 * put current block into this logical block entry */
		if (zone->l2p_table[cur_fst_page_logoff] == 0xFFFF) {
			zone->l2p_table[cur_fst_page_logoff] =
			    (u16) (i & 0x3FF);
			continue;
		}
		/* You have get the logical address, but in Log2Phy mapping
		 * table, The entry for it has been occupied */

		phy_block =
		    zone->l2p_table[cur_fst_page_logoff] +
		    ((u32) ((zone_no) << 10));

		/* rts_trace(TRACE_XD, "Arbitrate: cur_blk = 0x%x,
					entry_blk = 0x%x\n", i, phy_block); */

		/* Read the last page redundant data */
		page_addr = ((i + 1) << xd_card->block_shift) - 1;
		retval = cr_xd_read_redundant(cr, c, page_addr, redunt, 11);
		if (retval != STATUS_SUCCESS)
			continue;
		cur_lst_page_logoff = cr_xd_load_log_block_addr(cr, redunt);
		if (cur_lst_page_logoff == cur_fst_page_logoff) {
			int m;

			/* Read the last page redundant data */
			page_addr =
			    ((phy_block + 1) << xd_card->block_shift) - 1;

			for (m = 0; m < 3; m++) {
				retval = cr_xd_read_redundant(cr, c, page_addr,
								redunt, 11);
				if (retval == STATUS_SUCCESS)
					break;
			}

			/* Out of retry put current block into this entry */
			if (m == 3) {
				zone->l2p_table[cur_fst_page_logoff] =
				    (u16) (i & 0x3FF);
				retval = cr_xd_erase_block(cr, c, phy_block);
				if (retval == STATUS_SUCCESS)
					xd_comm_set_unused_block(cr, c,
								phy_block);
				continue;
			}
			ent_lst_page_logoff = cr_xd_load_log_block_addr(cr,
								redunt);
			if (ent_lst_page_logoff != cur_fst_page_logoff) {
				zone->l2p_table[cur_fst_page_logoff] =
				    (u16) (i & 0x3FF);
				retval = cr_xd_erase_block(cr, c, phy_block);
				if (retval == STATUS_SUCCESS)
					xd_comm_set_unused_block(cr, c,
								phy_block);
				continue;
			} else {
				retval = cr_xd_erase_block(cr, c, i);
				if (retval == STATUS_SUCCESS)
					xd_comm_set_unused_block(cr, c, i);
			}
		} else {
			/* The addrress in the the first and last page are not
			 * the same in current block, Erase current block */
			retval = cr_xd_erase_block(cr, c, i);
			if (retval == STATUS_SUCCESS)
				xd_comm_set_unused_block(cr, c, i);
		}
	}

	if (XD_CHK_4MB(xd_card))
		end = 500;
	else
		end = 1000;

	i = 0;
	for (start = 0; start < end; start++) {
		if (zone->l2p_table[start] == 0xFFFF)
			i++;
	}

	rts_trace(TRACE_XD, "Block count %d, invalid L2P entry %d\n", end, i);
	rts_trace(TRACE_XD, "Total unused block: %d\n", zone->unused_blk_cnt);

	if ((zone->unused_blk_cnt - i) < 1) {
		/* No enough free block to support xD write */
		set_card_flags(c, WP);
	}

	zone->build_flag = 1;

	return STATUS_SUCCESS;

Build_Fail:
	if (zone->l2p_table) {
		vfree(zone->l2p_table);
		zone->l2p_table = NULL;
	}
	if (zone->free_table) {
		vfree(zone->free_table);
		zone->free_table = NULL;
	}

	return STATUS_FAIL;
}

void xd_comm_free_l2p_tbl(struct card *c)
{
	struct xd_info *xd_card = c->extra;
	int i = 0;

	if (xd_card->zone != NULL) {
		for (i = 0; i < xd_card->zone_cnt; i++) {
			if (xd_card->zone[i].l2p_table != NULL) {
				vfree(xd_card->zone[i].l2p_table);
				xd_card->zone[i].l2p_table = NULL;
			}
			if (xd_card->zone[i].free_table != NULL) {
				vfree(xd_card->zone[i].free_table);
				xd_card->zone[i].free_table = NULL;
			}
		}
		vfree(xd_card->zone);
		xd_card->zone = NULL;
	}
}

static int cr_xd_reset_new(struct card_reader *cr, struct card *c)
{
	struct xd_info *xd_card = c->extra;
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (NULL == xd_card) {
		rts_trace(TRACE_XD, "xd_card is NULL\n");
		return STATUS_FAIL;
	}
	memset(xd_card, 0, sizeof(struct xd_info));

	xd_card->block_shift = 0;
	xd_card->page_off = 0;
	xd_card->addr_cycle = 0;
	xd_card->capacity = 0;
	xd_card->zone_cnt = 0;
	xd_card->cis_block = 0xFFFF;
	xd_card->delay_write.delay_write_flag = 0;
	cr_xd_prepare_init(cr);
	ret = xd_comm_reset_xd(cr, c);
	if (ret != STATUS_SUCCESS) {
#if 0 /* edwin: comment out, temporarily, not error handling */
		if (cr->option.reset_or_rw_fail_set_pad_drive) {
			/* sangdy2010-07-26:if reset or rw fail,
			 * then set SD20 PAD drive to 8mA. */
			cr_wr_reg(cr, CARD_DRIVE_SEL,
					      SD20_DRIVE_MASK, DRIVE_8mA);
		}
#endif
		TRACE_RET(cr, ret);
	}
	ret = xd_comm_init_l2p_tbl(cr, c);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, ret);

	c->capacity = xd_card->capacity;

	rts_trace(TRACE_XD, "-->\n");

	return STATUS_SUCCESS;
}

static int cr_xd_reset(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	if (cr->xd_ops_prot) {
		ret = cr_xd_reset_new(cr, c);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;
}

static int cr_xd_disable_new(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot->disable) {
		ret = cr->xd_ops_prot->disable(cr, c);
	} else {
		printk("%s: xd_disable callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	rts_trace(TRACE_XD, "-->\n");

	return ret;
}


static int cr_xd_disable(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot) {
		ret = cr_xd_disable_new(cr, c);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;


	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

static int cr_xd_rw_new(struct card_reader *cr, struct card *c,
				struct scsi_cmnd *srb, u32 lba, u16 num)
{
	struct xd_info *xd_card = c->extra;
	struct xd_delay_write_tag *delay_write = &(xd_card->delay_write);
	u8 lun = SCSI_LUN(srb);
	int zone_no;
	u32 log_blk, old_blk = 0, new_blk = 0;
	u32 start_sector = lba;
	u16 sector_cnt = num;
	u16 log_off, total_sec_cnt = sector_cnt;
	u8 start_page, end_page = 0, page_cnt;
	u8 *buf;
	unsigned int index = 0;
	void *ptr = NULL; /* default for USB transport */
	unsigned int offset = 0;
	int retval = STATUS_SUCCESS;

	rts_trace(TRACE_XD, "<--\n");

	/* NOTE: for PCIe, it's important to initialise "__index" to the address
	 * of "index" which is assign to zero, as to USB, it should be assigned
	 * with NULL, due to the difference in transport layer */
	if (!strcmp(cr->tr->name, "PCIe")) ptr = &index;
	xd_comm_set_err_code(c, XD_NO_ERROR);
	xd_card->counter = 0;
	rts_trace(TRACE_XD, "xd_rw: scsi_bufflen = %d, scsi_sg_count = %d\n",
		       scsi_bufflen(srb), scsi_sg_count(srb));
	rts_trace(TRACE_XD, "Data direction: %s\n",
		       (srb->sc_data_direction == DMA_TO_DEVICE) ?
							"write" : "read");
	buf = (u8 *) scsi_sglist(srb);
	retval = xd_comm_switch_clock(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);

	log_blk = start_sector >> xd_card->block_shift;
	start_page = (u8) start_sector & xd_card->page_off;
	zone_no = (int)(log_blk / 1000);
	log_off = (u16) (log_blk % 1000);

	rts_trace(TRACE_XD, "log_blk = 0x%x\n", log_blk);

	if (xd_card->zone[zone_no].build_flag == 0) {
		retval = xd_comm_build_l2p_tbl(cr, c, zone_no);
		if (retval != STATUS_SUCCESS) {
			set_card_flags(c, FAIL);
			set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
			TRACE_RET(cr, retval);
		}
	}

	if (srb->sc_data_direction == DMA_TO_DEVICE) {
		if (delay_write->delay_write_flag &&
		    (delay_write->logblock == log_blk) &&
		    (start_page > delay_write->pageoff)) {
			delay_write->delay_write_flag = 0;
			if (delay_write->old_phyblock != BLK_NOT_FOUND) {
#ifdef XD_SPEEDUP
				retval = cr_xd_auto_copy_page(cr, c,
					delay_write->old_phyblock,
					delay_write->new_phyblock,
					delay_write->pageoff, start_page);
#else
				retval = cr_xd_copy_page(cr, c,
						      delay_write->old_phyblock,
						      delay_write->new_phyblock,
						      delay_write->pageoff,
						      start_page);
#endif
				if (retval != STATUS_SUCCESS) {
					set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_WRITE_ERR);
					TRACE_RET(cr, retval);
				}
			}
			old_blk = delay_write->old_phyblock;
			new_blk = delay_write->new_phyblock;
		} else if (delay_write->delay_write_flag &&
			   (delay_write->logblock == log_blk) &&
			   (start_page == delay_write->pageoff)) {
			delay_write->delay_write_flag = 0;
			old_blk = delay_write->old_phyblock;
			new_blk = delay_write->new_phyblock;
		} else {
			retval = xd_comm_delay_write(cr, c);
			if (retval != STATUS_SUCCESS) {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, retval);
			}
			old_blk = xd_comm_get_l2p_tbl(cr, c, zone_no, log_off);
			new_blk = xd_comm_get_unused_block(cr, c, zone_no);
			if ((old_blk == BLK_NOT_FOUND)
			    || (new_blk == BLK_NOT_FOUND)) {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, retval);
			}
			retval = xd_comm_prepare_write(cr, c, old_blk,
						new_blk, log_blk, start_page);
			if (retval != STATUS_SUCCESS) {
				if (cr_detect_card(cr, XD_ID) == CD_NOT_EXIST) {
					set_sense_type(cr, lun,
						SENSE_TYPE_MEDIA_NOT_PRESENT);
					TRACE_RET(cr, STATUS_FAIL);
				}
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, retval);
			}
		}
	} else {
		retval = xd_comm_delay_write(cr, c);
		if (retval != STATUS_SUCCESS) {
			if (cr_detect_card(cr, XD_ID) == CD_NOT_EXIST) {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_NOT_PRESENT);
				TRACE_RET(cr, STATUS_FAIL);
			}
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			TRACE_RET(cr, retval);
		}
		old_blk = xd_comm_get_l2p_tbl(cr, c, zone_no, log_off);
		if (old_blk == BLK_NOT_FOUND) {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	rts_trace(TRACE_XD, "old_blk = 0x%x\n", old_blk);
	if (srb->sc_data_direction == DMA_TO_DEVICE)
		rts_trace(TRACE_XD, "new_blk = 0x%x\n", new_blk);

	while (total_sec_cnt) {
		if ((start_page + total_sec_cnt) > (xd_card->page_off + 1))
			end_page = xd_card->page_off + 1;
		else
			end_page = start_page + (u8) total_sec_cnt;
		page_cnt = end_page - start_page;
		if (srb->sc_data_direction == DMA_FROM_DEVICE) {
			retval = cr_xd_read_multiple_pages(cr, c, srb,
							old_blk, log_blk,
							start_page, end_page,
							buf, &ptr, &offset);
			if (retval != STATUS_SUCCESS) {
				set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			retval =
			    cr_xd_write_multiple_pages(cr, c, srb,
						old_blk, new_blk, log_blk,
						start_page, end_page,
						buf, &ptr, &offset);
			if (retval != STATUS_SUCCESS) {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		total_sec_cnt -= page_cnt;
		if (scsi_sg_count(srb) == 0)
			buf += page_cnt * 512;
		if (total_sec_cnt == 0) {
			/* Data transfer finished */
			break;
		}

		log_blk++;
		zone_no = (int)(log_blk / 1000);
		log_off = (u16) (log_blk % 1000);

		if (xd_card->zone[zone_no].build_flag == 0) {
			retval = xd_comm_build_l2p_tbl(cr, c, zone_no);
			if (retval != STATUS_SUCCESS) {
				set_card_flags(c, FAIL);
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_NOT_PRESENT);
				TRACE_RET(cr, retval);
			}
		}
		old_blk = xd_comm_get_l2p_tbl(cr, c, zone_no, log_off);
		if (old_blk == BLK_NOT_FOUND) {
			if (srb->sc_data_direction == DMA_FROM_DEVICE) {
				set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			} else {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
			}
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (srb->sc_data_direction == DMA_TO_DEVICE) {
			new_blk = xd_comm_get_unused_block(cr, c, zone_no);
			if (new_blk == BLK_NOT_FOUND) {
				set_sense_type(cr, lun,
					       SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		start_page = 0;
	}

	if ((srb->sc_data_direction == DMA_TO_DEVICE) &&
	    (end_page != (xd_card->page_off + 1))) {
		delay_write->delay_write_flag = 1;
		delay_write->old_phyblock = old_blk;
		delay_write->new_phyblock = new_blk;
		delay_write->logblock = log_blk;
		delay_write->pageoff = end_page;
	}

	scsi_set_resid(srb, 0);

	rts_trace(TRACE_XD, "-->\n");

	return retval;
}

static int cr_xd_rw(struct card_reader *cr, struct card *c,
				struct scsi_cmnd *srb, u32 lba, u16 num)
{
	int ret = 0;

	if (cr->xd_ops_prot) {
		ret = cr_xd_rw_new(cr, c, srb, lba, num);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;
}

static int cr_xd_cleanup_new(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	if (cr->xd_ops_prot->cleanup) {
		ret = cr->xd_ops_prot->cleanup(cr, c);
	} else {
		printk("%s: xd_cleanup callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	return ret;
}


static int cr_xd_cleanup(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot) {
		ret = cr_xd_cleanup_new(cr, c);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;


	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

static int cr_xd_release_new(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot->release) {
		ret = cr->xd_ops_prot->release(cr, c);
	} else {
		printk("%s: xd_release callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

static int cr_xd_release(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot) {
		ret = cr_xd_release_new(cr, c);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;


	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

static int cr_xd_update_status_new(struct card_reader *cr,
							struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot->update_status) {
		ret = cr->xd_ops_prot->update_status(cr, c);
	} else {
		printk("%s: xd_update_status callback is not implemented\n",
								cr->name);
		ret = -EINVAL;
	}

	rts_trace(TRACE_XD, "-->\n");

	return ret;

}

static int cr_xd_update_status(struct card_reader *cr, struct card *c)
{
	int ret = 0;

	rts_trace(TRACE_XD, "<--\n");

	if (cr->xd_ops_prot) {
		ret = cr_xd_update_status_new(cr, c);
	} else {
		printk("Either cr->xd_ops_prot or cr->xd_ops is assigned\n");
		ret = -EINVAL;
	}

	return ret;


	rts_trace(TRACE_XD, "-->\n");

	return ret;
}

int xd_comm_reset(struct card *c)
{
	return cr_xd_reset(c->cm->cr, c);
}

int xd_comm_disable(struct card *c)
{
	return cr_xd_disable(c->cm->cr, c);
}

int xd_comm_rw(struct scsi_cmnd *srb, struct card *c, u32 lba, u16 num)
{
	return cr_xd_rw(c->cm->cr, c, srb, lba, num);
}

int xd_comm_cleanup(struct card *c)
{
	return cr_xd_cleanup(c->cm->cr, c);
}

int xd_comm_release(struct card *c)
{
	return cr_xd_release(c->cm->cr, c);
}

int xd_comm_update_status(struct card *c)
{
	return cr_xd_update_status(c->cm->cr, c);
}

int xd_power_off_card3v3(struct card_reader *cr)
{
	return 0;
}
