/* Driver for Realtek card reader
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>
#include <linux/dma-mapping.h>
#include <linux/list.h>
#include <scsi/scsi.h>
#include <linux/module.h>
#include <linux/pci.h>

#include "rts_cr.h"
#include "cardreader.h"
#include "pm.h"
#include "rts_scsi.h"
#include "operations.h"
#include "cards/ms.h"
#include "cr_regs.h"
#include "rts_rpm.h"

void cr_wakeup_polling_thread(struct card_reader *cr)
{
	if (!strcmp(cr->tr->name, "PCIe")) {
		if (waitqueue_active(&cr->wait_q)) {
			cr_set_stat(cr, RTSX_STAT_RUN);
			rts_trace(TRACE_PM, "wakeup for PCIe\n");
			wake_up_interruptible(&cr->wait_q);
		}
	}

}
EXPORT_SYMBOL_GPL(cr_wakeup_polling_thread);



#ifdef SCSI_SCAN_DELAY
extern unsigned int delay_use;
#endif
static int cr_polling_thread(void *__cr)
{
	struct card_reader *cr = (struct card_reader *)__cr;
	struct rts_scsi *scsi = cr->scsi;
	struct transport *tr = cr->tr;
	struct card *cur_card = cr->cm->cur_card;

#ifdef SCSI_SCAN_DELAY
	/* Wait until SCSI scan finished */
	msleep((delay_use + 5) * HZ);
#endif

	for (;;) {
		msleep(cr->polling_interval);

		/* rts_trace(TRACE_POLL_THRD, "scsi->dflags=0x%lx\n",
		 * scsi->dflags); */
		/* if the device has disconnected, we are free to exit */
		if (test_bit(FLIDX_DISCONNECTING, &scsi->dflags)) {
			rts_trace(TRACE_CR, "exiting for disconnection\n");
			break;
		}
		/* rts_trace(TRACE_POLL_THRD, "scsi->dflags=0x%lx\n",
		 * scsi->dflags); */

		if (cur_card && MS_ID == cur_card->id)
			ms_polling_format_progress(cr, cur_card);

		if (rtsx_chk_stat(cr, RTSX_STAT_SS) &&
			!strcmp(cr->tr->name, "PCIe") &&
			!check_dev_flag(cr, POLLING_OCP_EN)) {
			rts_trace(TRACE_PM, "RTSX_STAT_SS state, hold on!\n");
			cr_runtime_put(cr);
			wait_event_interruptible(cr->wait_q,
					rtsx_chk_stat(cr, RTSX_STAT_RUN));
			cr_runtime_get(cr);
		}

		/* lock the device pointers */
		mutex_lock(&(tr->trans_mutex));

		cr_polling_func(cr);

		/* unlock the device pointers */
		mutex_unlock(&tr->trans_mutex);
	}

	complete(&cr->polling_exit);

	return 0;
}

int start_polling_thread(struct card_reader *cr)
{
	struct task_struct *th;
	int ret = 0;

	/* Start up our polling thread */
	th = kthread_run(cr_polling_thread, cr, RTS_CR_POLLING_THREAD);
	if (IS_ERR(th)) {
		printk(KERN_WARNING RTS_TIP
		       "Unable to start polling thread\n");
		ret = PTR_ERR(th);
		goto BadDevice;
	}
	cr->polling_thread = th;
BadDevice:

	return ret;
}

static irqreturn_t cr_irq(int irq, void *dev_id)
{
	struct card_reader *cr = dev_id;
	struct rts_pcie *pcie = cr->tr->extra;

	spin_lock(&pcie->reg_lock);

	if (STATUS_FAIL == cr_pre_handle_interrupt(cr)) {
		spin_unlock(&pcie->reg_lock);
		if (cr->int_reg == 0xFFFFFFFF)
			return IRQ_HANDLED;
		else
			return IRQ_NONE;
	}

	cr_process_ocp_interrupt(cr);
	cr_process_card_interrupt(cr);
	cr_process_transfer_interrupt(cr);

	spin_unlock(&pcie->reg_lock);

	return IRQ_HANDLED;
}

int cr_request_irq(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	struct rts_pcie *pcie = (struct rts_pcie *)cr->tr->extra;

	if (option->msi_en) {
		rts_trace(TRACE_CR, "enable MSI\n");
		if (cr_enable_msi(cr) < 0) {
			option->msi_en = 0;
			rts_trace(TRACE_CR, "enable MSI failed, using INTX\n");
		}
	}

	if (request_irq(pcie->pci_dev->irq, cr_irq,
		option->msi_en ? 0 : IRQF_SHARED, CR_DRIVER_NAME, cr)) {
		rts_trace(TRACE_CR, "get IRQ %d failed, disabling device\n",
			pcie->pci_dev->irq);
		return -1;
	}

	cr_synchronize_irq(cr);
	pcie->irq = pcie->pci_dev->irq;

	return 0;
}
EXPORT_SYMBOL_GPL(cr_request_irq);

void cr_free_irq(struct card_reader *cr)
{
	struct rts_pcie *pcie = (struct rts_pcie *)cr->tr->extra;
	struct cr_option *option = &cr->option;

	if (pcie->irq >= 0) {
		rts_trace(TRACE_CR, "freeing IRQ %d\n", pcie->irq);
		synchronize_irq(pcie->irq);
		free_irq(pcie->irq, (void *)cr);
		pcie->irq = -1;
	}

	if (option->msi_en) {
		rts_trace(TRACE_CR, "disable MSI\n");
		pci_disable_msi(pcie->pci_dev);
	}
}
EXPORT_SYMBOL_GPL(cr_free_irq);

void cr_clear_first_install_mark(struct card_reader *cr)
{

}

u64 cr_get_tick(void)
{
	struct timespec now;

	getnstimeofday(&now);

	return (u64)(now.tv_sec) * 1000 + now.tv_nsec / 1000000;
}
EXPORT_SYMBOL_GPL(cr_get_tick);

void cr_inform_reset(struct card_reader *cr)
{
	/* TODO */
}

void cr_patch_hw_setting(struct card_reader *cr)
{
	/* TODO */
}
EXPORT_SYMBOL_GPL(cr_patch_hw_setting);
