/* Driver for Realtek PCIe card reader
 *
 * Copyright(c) 2009 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   wwang (wei_wang@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/vmalloc.h>

#include "operations.h"
#include "cr_regs.h"
#include "cards.h"
#include "trace.h"
#include "cardmanager.h"
#include "trans_pcie.h"
#include "cards/sd.h"
#include "cards/sd40.h"
#include "cards/ms.h"

int cr_common_wr_reg(struct card_reader *cr, u16 addr, u8 mask, u8 data)
{
	int i;
	u32 val = 3 << 30;

	val |= (u32)(addr & 0x3FFF) << 16;
	val |= (u32)mask << 8;
	val |= (u32)data;

	cr_iowrite32(cr, HAIMR, val);

	for (i = 0; i < MAX_RW_REG_CNT; i++) {
		val = cr_ioread32(cr, HAIMR);
		if ((val & (1 << 31)) == 0) {
			if (data != (u8)val)
				trace_ret(cr, STATUS_FAIL, "write 0x%04x to 0x%x failed\n",
					addr, data);
			return STATUS_SUCCESS;
		}

		if (cr->option.delay_for_rw_reg)
			mdelay(cr->option.delay_for_rw_reg);
	}

	catch_trigger(cr);
	cr->tr->catch_link_error = 1;

	trace_ret(cr, STATUS_TIMEDOUT, "write 0x%04x to 0x%x failed\n",
		addr, data);
}

int cr_common_rd_reg(struct card_reader *cr, u16 addr, u8 *data)
{
	int i;
	u32 val = 2 << 30;

	if (data)
		*data = 0;

	val |= (u32)(addr & 0x3FFF) << 16;
	cr_iowrite32(cr, HAIMR, val);

	for (i = 0; i < MAX_RW_REG_CNT; i++) {
		val = cr_ioread32(cr, HAIMR);
		if ((val & (1 << 31)) == 0)
			break;
		if (cr->option.delay_for_rw_reg)
			mdelay(cr->option.delay_for_rw_reg);
	}

	if (i >= MAX_RW_REG_CNT) {
		cr->tr->catch_link_error = 1;
		trace_ret(cr, STATUS_TIMEDOUT, "read 0x%04x failed\n", addr);
	}

	if (data)
		*data = (u8)(val & 0xFF);

	return STATUS_SUCCESS;
}

int cr_common_wr_cfg32(struct card_reader *cr, u16 addr, u32 mask, u32 val)
{
	struct rts_pcie *pcie = cr->tr->extra;
	u8 func_no = pcie->func_no;
	u8 mode = 0, tmp;
	int i;

	for (i = 0; i < 4; i++) {
		if (mask & 0xFF) {
			cr_wr_reg(cr, REG_CFGDATA0 + i, 0xFF,
				(u8)(val & mask & 0xFF));
			mode |= (1 << i);
		}
		mask >>= 8;
		val >>= 8;
	}

	if (mode) {
		cr_wr_reg(cr, REG_CFGADDR0, 0xFF, (u8)addr);
		cr_wr_reg(cr, REG_CFGADDR1, 0xFF, (u8)(addr >> 8));

		cr_wr_reg(cr, REG_CFGRWCTL,
				0xFF, 0x80 | mode | ((func_no & 0x03) << 4));

		/* Wait transfer end */
		for (i = 0; i < MAX_RW_REG_CNT; i++) {
			cr_rd_reg(cr, REG_CFGRWCTL, &tmp);
			if ((tmp & 0x80) == 0)
				break;
		}
	}

	return STATUS_SUCCESS;
}

int cr_common_rd_cfg32(struct card_reader *cr, u16 addr, u32 *val)
{
	struct rts_pcie *pcie = cr->tr->extra;
	u8 func_no = pcie->func_no;
	int i;
	u8 tmp;
	u32 data = 0;

	cr_wr_reg(cr, REG_CFGADDR0, 0xFF, (u8)addr);
	cr_wr_reg(cr, REG_CFGADDR1, 0xFF, (u8)(addr >> 8));
	cr_wr_reg(cr, REG_CFGRWCTL, 0xFF, 0x80 |
						((func_no & 0x03) << 4));

	/* Wait transfer end */
	for (i = 0; i < MAX_RW_REG_CNT; i++) {
		cr_rd_reg(cr, REG_CFGRWCTL, &tmp);
		if ((tmp & 0x80) == 0)
			break;
	}

	for (i = 0; i < 4; i++) {
		cr_rd_reg(cr, REG_CFGDATA0 + i, &tmp);
		data |= (u32)tmp << (i * 8);
	}

	if (val)
		*val = data;

	return STATUS_SUCCESS;
}

int cr_common_wr_cfg_seq(struct card_reader *cr, u16 addr,
							u8 *buf, int len)
{
	u32 *data, *mask;
	u16 offset = addr % 4;
	u16 aligned_addr = addr - offset;
	int dw_len, i, j;
	int retval;

	rts_trace(TRACE_CR_PROTOCOL, "<--\n");

	if (!buf)
		TRACE_RET(cr, STATUS_NOMEM);

	if ((len + offset) % 4)
		dw_len = (len + offset) / 4 + 1;
	else
		dw_len = (len + offset) / 4;
	rts_trace(TRACE_CR_PROTOCOL, "dw_len = %d\n", dw_len);

	data = (u32 *)vmalloc(dw_len * 4);
	if (!data)
		TRACE_RET(cr, STATUS_NOMEM);
	memset(data, 0, dw_len * 4);

	mask = (u32 *)vmalloc(dw_len * 4);
	if (!mask) {
		vfree(data);
		TRACE_RET(cr, STATUS_NOMEM);
	}
	memset(mask, 0, dw_len * 4);

	j = 0;
	for (i = 0; i < len; i++) {
		mask[j] |= 0xFF << (offset * 8);
		data[j] |= buf[i] << (offset * 8);
		if (++offset == 4) {
			j++;
			offset = 0;
		}
	}
#if 0
	RTSX_DUMP(mask, dw_len * 4);
	RTSX_DUMP(data, dw_len * 4);
#endif
	for (i = 0; i < dw_len; i++) {
		retval = cr_wr_cfg32(cr, aligned_addr + i * 4,
							mask[i], data[i]);
		if (retval != STATUS_SUCCESS) {
			vfree(data);
			vfree(mask);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	vfree(data);
	vfree(mask);

	return STATUS_SUCCESS;
}

int cr_common_rd_cfg_seq(struct card_reader *cr, u16 addr,
							u8 *buf, int len)
{
	u32 *data;
	u16 offset = addr % 4;
	u16 aligned_addr = addr - offset;
	int dw_len, i, j;
	int retval;

	rts_trace(TRACE_CR_PROTOCOL, "<--\n");

	if ((len + offset) % 4)
		dw_len = (len + offset) / 4 + 1;
	else
		dw_len = (len + offset) / 4;
	rts_trace(TRACE_CR_PROTOCOL, "dw_len = %d\n", dw_len);

	data = (u32 *)vmalloc(dw_len * 4);
	if (!data)
		TRACE_RET(cr, STATUS_NOMEM);

	for (i = 0; i < dw_len; i++) {
		retval = cr_rd_cfg32(cr, aligned_addr + i * 4, data + i);
		if (retval != STATUS_SUCCESS) {
			vfree(data);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	if (buf) {
		j = 0;
		for (i = 0; i < len; i++) {
			buf[i] = (u8)(data[j] >> (offset * 8));
			if (++offset == 4) {
				j++;
				offset = 0;
			}
		}
	}

	vfree(data);

	return STATUS_SUCCESS;
}


int cr_common_wr_phy16(struct card_reader *cr, u8 addr, u16 val)
{
	int i, finished = 0;
	u8 tmp;

	cr_wr_reg(cr, REG_PHYDATA0, 0xFF, (u8)val);
	cr_wr_reg(cr, REG_PHYDATA1, 0xFF, (u8)(val >> 8));
	cr_wr_reg(cr, REG_PHYADDR, 0xFF, addr);

	cr_wr_reg(cr, REG_PHYRWCTL, 0xFF, 0x81);

	for (i = 0; i < 100000; i++) {
		cr_rd_reg(cr, REG_PHYRWCTL, &tmp);
		if (!(tmp & 0x80)) {
			finished = 1;
			break;
		}
	}

	if (!finished)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(cr_common_wr_phy16);

int cr_common_rd_phy16(struct card_reader *cr, u8 addr, u16 *val)
{
	int i, finished = 0;
	u16 data = 0;
	u8 tmp;

	cr_wr_reg(cr, REG_PHYADDR, 0xFF, addr);
	cr_wr_reg(cr, REG_PHYRWCTL, 0xFF, 0x80);

	for (i = 0; i < 100000; i++) {
		cr_rd_reg(cr, REG_PHYRWCTL, &tmp);
		if (!(tmp & 0x80)) {
			finished = 1;
			break;
		}
	}

	if (!finished)
		TRACE_RET(cr, STATUS_FAIL);

	cr_rd_reg(cr, REG_PHYDATA0, &tmp);
	data = tmp;
	cr_rd_reg(cr, REG_PHYDATA1, &tmp);
	data |= (u16)tmp << 8;

	if (val)
		*val = data;

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(cr_common_rd_phy16);

int cr_common_read_ppbuf(struct card_reader *cr, u8 *buf, int buf_len)
{
	int retval;
	int i;
	u16 reg_addr;
	u8 *ptr;

	if (!buf)
		TRACE_RET(cr, STATUS_ERROR);
	ptr = buf;
	reg_addr = PPBUF_BASE2;
#ifdef USING_PPBUF
	for (i = 0; i < buf_len/256; i++) {
		int j;
		cr_init_cmd(cr);
		for (j = 0; j < 256; j++)
			cr_add_cmd(cr, READ_REG_CMD, reg_addr++, 0, 0);
		retval = cr_send_cmd(cr, 0, 250);
		if (retval < 0)
			TRACE_RET(cr, STATUS_FAIL);
		memcpy(ptr, cr_get_cmd_data(cr), 256);
		ptr += 256;
	}

	if (buf_len%256) {
		cr_init_cmd(cr);
		for (i = 0; i < buf_len%256; i++)
			cr_add_cmd(cr, READ_REG_CMD, reg_addr++, 0, 0);
		retval = cr_send_cmd(cr, 0, 250);
		if (retval < 0)
			TRACE_RET(cr, STATUS_FAIL);
	}
#else
	cr_init_cmd(cr);
	for (i = 0; i < buf_len%256; i++)
		cr_add_cmd(cr, READ_REG_CMD, reg_addr++, 0, 0);
	retval = cr_send_cmd(cr, 0, 250);
	if (retval < 0)
		TRACE_RET(cr, STATUS_FAIL);
#endif
	memcpy(ptr, cr_get_cmd_data(cr), buf_len%256);

	return STATUS_SUCCESS;
}

int cr_common_write_ppbuf(struct card_reader *cr, u8 *buf, int buf_len)
{
	int retval;
	int i;
	u16 reg_addr;
	u8 *ptr;

	if (!buf)
		TRACE_RET(cr, STATUS_ERROR);
	ptr = buf;
	reg_addr = PPBUF_BASE2;
#ifdef USING_PPBUF
	for (i = 0; i < buf_len/256; i++) {
		int j;
		cr_init_cmd(cr);
		for (j = 0; j < 256; j++) {
			cr_add_cmd(cr, WRITE_REG_CMD, reg_addr++,
								0xFF, *ptr);
			ptr++;
		}
		retval = cr_send_cmd(cr, 0, 250);
		if (retval < 0)
			TRACE_RET(cr, STATUS_FAIL);
	}

	if (buf_len%256) {
		cr_init_cmd(cr);
		for (i = 0; i < buf_len%256; i++) {
			cr_add_cmd(cr, WRITE_REG_CMD, reg_addr++,
								0xFF, *ptr);
			ptr++;
		}
		retval = cr_send_cmd(cr, 0, 250);
		if (retval < 0)
			TRACE_RET(cr, STATUS_FAIL);
	}
#else
	cr_init_cmd(cr);
	for (i = 0; i < buf_len%256; i++) {
		cr_add_cmd(cr, WRITE_REG_CMD, reg_addr++, 0xFF, *ptr);
		ptr++;
	}
	retval = cr_send_cmd(cr, 0, 250);
	if (retval < 0)
		TRACE_RET(cr, STATUS_FAIL);
#endif

	return STATUS_SUCCESS;
}

int cr_common_detect_card(struct card_reader *cr, enum CARD_ID id)
{
	u32 card_cd, status;

	if (cr->option.sd_dat3_cd_en && SD_ID == id) {
		rts_trace(TRACE_CR, "SDD3CDEnable=1, return STATUS_SUCCESS \n");
		return STATUS_SUCCESS;
	}

	if (id == SD_ID) {
		card_cd = SD_EXIST;
	} else if (id == MS_ID) {
		card_cd = MS_EXIST;
	} else {
		rts_trace(TRACE_CR_PROTOCOL, "Wrong card type: 0x%x\n", id);
		TRACE_RET(cr, STATUS_FAIL);
	}

	status = cr_ioread32(cr, BIPR);
	if (!(status & card_cd))
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

void cr_common_enable_dma(struct card_reader *cr, u32 byte_cnt,
		u8 pack_size, enum dma_data_direction dir)
{
	if (pack_size > DMA_1024)
		pack_size = DMA_512;

	cr_add_cmd(cr, WRITE_REG_CMD, REG_IRQSTAT0,
				DMA_DONE_INT, DMA_DONE_INT);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_DMATC3, 0xFF,
						(u8)(byte_cnt >> 24));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_DMATC2, 0xFF,
						(u8)(byte_cnt >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_DMATC1, 0xFF,
						(u8)(byte_cnt >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_DMATC0, 0xFF,
						(u8)byte_cnt);

	if (dir == DMA_FROM_DEVICE)
		cr_add_cmd(cr, WRITE_REG_CMD, REG_DMACTL, 0x03 |
			DMA_PACK_SIZE_MASK, DMA_DIR_FROM_CARD |
			DMA_EN | pack_size);
	else
		cr_add_cmd(cr, WRITE_REG_CMD, REG_DMACTL, 0x03 |
			DMA_PACK_SIZE_MASK, DMA_DIR_TO_CARD |
			DMA_EN | pack_size);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE, 0x01,
							RING_BUFFER);

}

void cr_common_cd_debounce(struct card_reader *cr)
{
	struct list_head *pos;
	unsigned long release_map = 0, reset_map = 0;

	list_for_each(pos, &cr->cm->card_list) {
		struct card *c;
		c = list_entry(pos, struct card, link);
		rts_trace(TRACE_POLL_THRD, "card:%s\n", c->name);
		if (test_bit(c->id, &(cr->need_release))) {
			rts_trace(TRACE_CR, "release: %s\n", c->name);
			clr_card_flags(c, EXIST);
			set_bit(c->id, &release_map);
			cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
			clear_bit(c->id, &(cr->need_release));
			c->cd_debounce_begin = 0;
		}

		if (test_bit(c->id, &(cr->need_reset))) {
			if (c->cd_debounce_begin) {
				u64 elapse = cr_get_tick() - c->cd_timestamp;
				rts_trace(TRACE_CR, "CD debounce time:%lld\n",
									elapse);
				if (elapse >= cr->option.cd_debounce_time) {
					cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
					set_bit(c->id, &reset_map);
					clear_bit(c->id, &(cr->need_reset));
					c->cd_debounce_begin = 0;
					if (!cr->option.ft2_fast_mode)
						c->reset_flags = 0;
					c->runtime_flags = 0;
					c->uhsii_hd_rw_fail = 0;
					c->uhsii_fd_rw_fail = 0;
				}
				cr_set_stat(cr, RTSX_STAT_RUN);
			} else {
				rts_trace(TRACE_CR, "reset: %s\n", c->name);
				c->cd_debounce_begin = 1;
				c->cd_timestamp = cr_get_tick();
				cr_set_stat(cr, RTSX_STAT_RUN);
			}
		}
	}

	cr->reset_map = reset_map;
	cr->release_map = release_map;
	rts_trace(TRACE_POLL_THRD, "reset_map:0x%lx, release_map=0x%lx\n",
						cr->reset_map, cr->release_map);
}


int cr_common_flush_rb(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);

	return 0;
}

int cr_common_poweron_card(struct card_reader *cr, enum CARD_ID id)
{
	int retval;

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PWR_CTL,
			SD_POWER_MASK, SD_PARTIAL_POWER_ON);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_PWR_GATE_CTRL,
			LDO3318_PWR_MASK, LDO_SUSPEND);
	retval = cr_send_cmd(cr, 0, 100);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	udelay(cr->option.pmos_pwr_on_interval);

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PWR_CTL,
			SD_POWER_MASK, SD_POWER_ON);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_PWR_GATE_CTRL,
			LDO3318_PWR_MASK, LDO_ON);
	retval = cr_send_cmd(cr, 0, 100);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

int cr_common_poweroff_card(struct card_reader *cr, enum CARD_ID id)
{
	/* 2010-10-22, wwang
	 * Can't use command buffer here,
	 * because this function may be called in enter_ss
	 */
	cr_wr_reg(cr, REG_CARD_PWR_CTL, SD_POWER_MASK
		 | PMOS_STRG_MASK, SD_POWER_OFF | PMOS_STRG_400mA);
	cr_wr_reg(cr, REG_PWR_GATE_CTRL, LDO3318_PWR_MASK, LDO_OFF);

	return STATUS_SUCCESS;
}


int cr_common_enable_card_clock(struct card_reader *cr, enum CARD_ID id)
{
	u8 clk_en = 0;

	if (SD_ID == id)
		clk_en |= SD_CLK_EN;
	if (MS_ID == id)
		clk_en |= MS_CLK_EN;
	if (XD_ID == id)
		clk_en |= XD_CLK_EN;

	cr_wr_reg(cr, REG_CARD_CLK_EN, clk_en, clk_en);

	return STATUS_SUCCESS;
}

int cr_common_disable_card_clock(struct card_reader *cr, enum CARD_ID id)
{
	/* disable all card's clock */
	return cr_wr_reg(cr, REG_CARD_CLK_EN, 0x1E, 0);

}

static int cr_switch_ssc_clock(struct card_reader *cr,
						struct card *card, int clk)
{
	struct sd_info *sd_card = NULL;
	struct ms_info *ms_card = NULL;
	struct cr_option *option = &(cr->option);
	int retval;
	u8 N, min_N, max_N;
	u8 mcu_cnt, div, max_div, ssc_depth;
	int sd_vpclk_phase_reset = 0;

	if (cr->cur_clk == clk)
		return STATUS_SUCCESS;

	if (cr->ops->calc_ssc_clock_N) {
		struct ssc_clock_N ssc = {clk, 0, 0, 0};
		cr->ops->calc_ssc_clock_N(cr, card, &ssc);
		N = ssc.N;
		min_N = ssc.min_N;
		max_N = ssc.max_N;
	} else {
		N = (u8)(clk - 2);
		min_N = 80;
		max_N = 208;
	}

	max_div = REG_CLK_DIV_8;

	if (SD_ID == card->id) {
		sd_card = card->extra;
		if (CHK_SD30_SPEED(sd_card) || CHK_MMC_DDR52(sd_card))
			sd_vpclk_phase_reset = 1;
	}

	rts_trace(TRACE_CR, "Switch SSC clock to %dMHz (cur_clk = %d)\n",
							clk, cr->cur_clk);

	if ((clk <= 2) || (N > max_N))
		TRACE_RET(cr, STATUS_FAIL);

	mcu_cnt = (u8)(125/clk + 3);
	if (mcu_cnt > 15)
		mcu_cnt = 15;

	/* To make sure that the SSC clock div_n is equal or greater than
	 * min_N */
	div = REG_CLK_DIV_1;
	while ((N < min_N) && (div < max_div)) {
		N = (N + 2) * 2 - 2;
		div++;
	}
	rts_trace(TRACE_CR, "N = %d, div = %d\n", N, div);

	if (option->ssc_en) {
		if (SD_ID == card->id) {
			sd_card = card->extra;
			if (CHK_SD_SDR104(sd_card))
				ssc_depth = option->ssc_depth_sd_sdr104;
			else if (CHK_SD_SDR50(sd_card))
				ssc_depth = option->ssc_depth_sd_sdr50;
			else if (CHK_SD_DDR50(sd_card))
				ssc_depth = double_depth(option->ssc_depth_sd_ddr50);
			else if (CHK_SD_HS(sd_card))
				ssc_depth = double_depth(option->ssc_depth_sd_hs);
			else if (CHK_MMC_52M(sd_card) || CHK_MMC_DDR52(sd_card))
				ssc_depth = double_depth(option->ssc_depth_mmc_52m);
			else
				ssc_depth = double_depth(option->ssc_depth_low_speed);
		} else if (MS_ID == card->id) {
			ms_card = card->extra;
			if (CHK_MSPRO(ms_card)) {
				if (CHK_HG8BIT(ms_card))
					ssc_depth = double_depth(option->ssc_depth_ms_hg);
				else
					ssc_depth = double_depth(option->ssc_depth_ms_4bit);
			} else {
				if (CHK_MS4BIT(ms_card))
					ssc_depth = double_depth(option->ssc_depth_ms_4bit);
				else
					ssc_depth = double_depth(option->ssc_depth_low_speed);
			}

		} else {
			ssc_depth = double_depth(option->ssc_depth_low_speed);
		}

		if (ssc_depth) {
			if (div == REG_CLK_DIV_2) {
				/* If clock divided by 2, ssc depth must
				 * be multiplied by 2 */
				if (ssc_depth > 1)
					ssc_depth -= 1;
				else
					ssc_depth = SSC_DEPTH_4M;
			} else if (div == REG_CLK_DIV_4) {
				/* If clock divided by 4, ssc depth must
				 * be multiplied by 4 */
				if (ssc_depth > 2)
					ssc_depth -= 2;
				else
					ssc_depth = SSC_DEPTH_4M;
			} else if (div == REG_CLK_DIV_8) {
				/* If clock divided by 8, ssc depth must
				 * be multiplied by 8 */
				if (ssc_depth > 3)
					ssc_depth -= 3;
				else
					ssc_depth = SSC_DEPTH_4M;
			}
		}
	} else {
		/* Disable SSC */
		ssc_depth = 0;
	}

	rts_trace(TRACE_CR, "ssc_depth = %d\n", ssc_depth);

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CLK_CTL,
				CLK_LOW_FREQ, CLK_LOW_FREQ);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CLK_DIV,
						0xFF, (div << 4) | mcu_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SSC_CTL1,
						SSC_RSTB, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SSC_CTL2,
					SSC_DEPTH_MASK, ssc_depth);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SSC_DIV_N_0, 0xFF, N);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SSC_CTL1, SSC_RSTB,
							SSC_RSTB);
	if (sd_vpclk_phase_reset) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_VPCLK0_CTL,
						PHASE_NOT_RESET, 0);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_VPCLK0_CTL,
						PHASE_NOT_RESET,
						PHASE_NOT_RESET);
	}

	retval = cr_send_cmd(cr, 0, WAIT_TIME);
	if (retval < 0)
		TRACE_RET(cr, STATUS_ERROR);

	udelay(10);
	cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, 0);

	cr->cur_clk = clk;

	return STATUS_SUCCESS;
}

static int cr_switch_normal_clock(struct card_reader *cr,
						struct card *card, int clk)
{
	u8 sel, div, mcu_cnt;

	if (cr->cur_clk == clk)
		return STATUS_SUCCESS;


	switch (clk) {
	case CLK_20:
		rts_trace(TRACE_CR, "Switch clock to 20MHz\n");
		sel = SSC_80;
		div = REG_CLK_DIV_4;
		mcu_cnt = 7;
		break;

	case CLK_30:
		rts_trace(TRACE_CR, "Switch clock to 30MHz\n");
		sel = SSC_120;
		div = REG_CLK_DIV_4;
		mcu_cnt = 7;
		break;

	case CLK_40:
		rts_trace(TRACE_CR, "Switch clock to 40MHz\n");
		sel = SSC_80;
		div = REG_CLK_DIV_2;
		mcu_cnt = 7;
		break;

	case CLK_50:
		rts_trace(TRACE_CR, "Switch clock to 50MHz\n");
		sel = SSC_100;
		div = REG_CLK_DIV_2;
		mcu_cnt = 6;
		break;

	case CLK_60:
		rts_trace(TRACE_CR, "Switch clock to 60MHz\n");
		sel = SSC_120;
		div = REG_CLK_DIV_2;
		mcu_cnt = 6;
		break;

	case CLK_80:
		rts_trace(TRACE_CR, "Switch clock to 80MHz\n");
		sel = SSC_80;
		div = REG_CLK_DIV_1;
		mcu_cnt = 5;
		break;

	case CLK_100:
		rts_trace(TRACE_CR, "Switch clock to 100MHz\n");
		sel = SSC_100;
		div = REG_CLK_DIV_1;
		mcu_cnt = 5;
		break;

	case CLK_120:
		rts_trace(TRACE_CR, "Switch clock to 120MHz\n");
		sel = SSC_120;
		div = REG_CLK_DIV_1;
		mcu_cnt = 5;
		break;

	case CLK_150:
		rts_trace(TRACE_CR, "Switch clock to 150MHz\n");
		sel = SSC_150;
		div = REG_CLK_DIV_1;
		mcu_cnt = 4;
		break;

	case CLK_200:
		rts_trace(TRACE_CR, "Switch clock to 200MHz\n");
		sel = SSC_200;
		div = REG_CLK_DIV_1;
		mcu_cnt = 4;
		break;

	default:
		rts_trace(TRACE_CR, "Try to switch to an illegal clock (%d)\n",
			clk);
		TRACE_RET(cr, STATUS_FAIL);
	}

	cr_wr_reg(cr, REG_CLK_CTL, 0xFF, CLK_LOW_FREQ);
	cr_wr_reg(cr, REG_SD_VPCLK0_CTL, PHASE_NOT_RESET, 0);
	cr_wr_reg(cr, REG_SD_VPCLK1_CTL, PHASE_NOT_RESET, 0);
	cr_wr_reg(cr, REG_CLK_DIV, 0xFF, (div << 4) | mcu_cnt);
	cr_wr_reg(cr, REG_CLK_SEL, 0xFF, sel);

	udelay(200);
	cr_wr_reg(cr, REG_SD_VPCLK0_CTL, PHASE_NOT_RESET,
			PHASE_NOT_RESET);
	cr_wr_reg(cr, REG_SD_VPCLK1_CTL, PHASE_NOT_RESET,
			PHASE_NOT_RESET);
	udelay(200);
	cr_wr_reg(cr, REG_CLK_CTL, 0xFF, 0);

	cr->cur_clk = clk;

	return STATUS_SUCCESS;
}

int cr_common_switch_clock(struct card_reader *cr, struct card *card, int clk)
{

	if (cr->asic_code)
		return cr_switch_ssc_clock(cr, card, clk);
	else
		return cr_switch_normal_clock(cr, card, clk);

}

static int set_card_share_mode(struct card_reader *cr, struct card *card)
{
	u8 mode = 0;
	u8 val = 0;
	enum CARD_ID id = card->id;

	if (id == SD_ID)
		mode = CARD_SHARE_SD;
	else if (id == MS_ID)
		mode = CARD_SHARE_MS;
	else if (id == XD_ID)
		mode = CARD_SHARE_XD;
	else
		TRACE_RET(cr, STATUS_FAIL);

	cr_rd_reg(cr, REG_CARD_SHARE_MODE, &val);

	if (SIG_LUN == cr->lun_mod) {
		if (mode != (val & CARD_SHARE_MASK)) {
			cr_wr_reg(cr, REG_CARD_SHARE_MODE,
					CARD_SHARE_MASK, mode);
		}
	} else {
		/* TODO: multi-LUN */
	}


	return 0;
}

int cr_common_select_card(struct card_reader *cr, struct card *card)
{
	enum CARD_ID id = card->id;
	u8 mod = 0, val = 0;

	if (id == SD_ID)
		mod = SD_MOD_SEL;
	else if (id == MS_ID)
		mod = MS_MOD_SEL;
	else if (id == XD_ID)
		mod = XD_MOD_SEL;
	else
		TRACE_RET(cr, STATUS_FAIL);

	cr_rd_reg(cr, REG_CARD_SELECT, &val);

	if (SIG_LUN == cr->lun_mod) {
		if (mod != (val & CARD_MOD_MASK)) {
			cr_wr_reg(cr, REG_CARD_SELECT, CARD_MOD_MASK, mod);
			cr->cur_card = id;
		}
	} else {
		/*TODO:for multi-lun*/
	}

	set_card_share_mode(cr, card);

	return STATUS_SUCCESS;
}


int cr_comm_read_host_sd40_iospace(struct card_reader *cr,
					u16 addr, u8 *data, int len)
{
	u16 reg = SD40_BASE_REGISTER + addr;
	u8 *p = data;
	int i = 0;

	for (i = 0; i < len; i++)
		cr_rd_reg(cr, reg++, p++);

	return 0;
}

int cr_comm_write_host_sd40_iospace(struct card_reader *cr,
					u16 addr, u8 *data, int len)
{
	u16 reg = SD40_BASE_REGISTER + addr;
	u8 *p = data;
	int i = 0;

	for (i = 0; i < len; i++)
		cr_wr_reg(cr, reg++, 0xFF, *p++);

	return 0;
}
