/* Driver for Realtek PCIe card reader
 * Header file
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/pci.h>
#include <linux/module.h>

#include "pm.h"
#include "cardreader.h"
#include "trans_pcie.h"
#include "rts_cr.h"
#include "cr_regs.h"
#include "cards/sd.h"
#include "cr_system.h"


void cr_turn_on_led(struct card_reader *cr)
{
	if (cr->ops->turn_on_led)
		cr->ops->turn_on_led(cr);
	else
		cr_wr_reg(cr, REG_GPIO_CTL, 0x02, 0x02);
}
EXPORT_SYMBOL_GPL(cr_turn_on_led);

void cr_turn_off_led(struct card_reader *cr)
{
	if (cr->ops->turn_off_led)
		cr->ops->turn_off_led(cr);
	else
		cr_wr_reg(cr, REG_GPIO_CTL, 0x02, 0x00);
}
EXPORT_SYMBOL_GPL(cr_turn_off_led);

void cr_enable_auto_blink(struct card_reader *cr)
{
	if (cr->ops->enable_auto_blink)
		cr->ops->enable_auto_blink(cr);
	else
		cr_wr_reg(cr, REG_OLT_LED_CTL, LED_SHINE_EN, LED_SHINE_EN);
}
EXPORT_SYMBOL_GPL(cr_enable_auto_blink);

void cr_disable_auto_blink(struct card_reader *cr)
{
	if (cr->ops->disable_auto_blink)
		cr->ops->disable_auto_blink(cr);
	else
		cr_wr_reg(cr, REG_OLT_LED_CTL, LED_SHINE_EN, 0x00);
}
EXPORT_SYMBOL_GPL(cr_disable_auto_blink);

void cr_blink_led(struct card_reader *cr)
{
	/* deprecated: It's needed only by 5209  */
}
EXPORT_SYMBOL_GPL(cr_blink_led);

void cr_trigger_led(struct card_reader *cr, enum led_switch s)
{
	if (ON == s)
		cr_turn_on_led(cr);
	else if (OFF == s)
		cr_turn_off_led(cr);
}


/* cr_pm_set_flags - Set PM flags
 * @cr: card reader structure
 * @flags: flags to be set
 *
 * Returns: Nothing
 */
void cr_pm_set_flags(struct card_reader *cr, u16 flags)
{
	cr->pm_flags = (flags);
}
EXPORT_SYMBOL_GPL(cr_pm_set_flags);

/* cr_pm_check_flags - Check whether the flag is set or not
 * @cr: card reader structure
 * @flags: flags to be checked
 *
 * Returns: 1 for set, otherwise 0
 */

int cr_pm_check_flags(struct card_reader *cr, u16 flags)
{
	return (cr->pm_flags == flags);
}
EXPORT_SYMBOL_GPL(cr_pm_check_flags);

/* cr_pm_setup - Set up environment for entering PM domain
 * @cr: card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_pm_setup(struct card_reader *cr)
{
	int ret = 0;

	if (cr->pm_ops->setup)
		ret = cr->pm_ops->setup(cr);
	else
		rts_trace(TRACE_PM, "cr->pm_ops->setup is not implemented\n");

	return ret;
}

/* cr_pm_bef_pow_dwn - Store context before power down operation
 * @cr: card reader structure
 * @pm_stat: PM state(PM_S1/PM_S3)
 *
 * Returns: Nothing
 */
void cr_pm_bef_pow_dwn(struct card_reader *cr, int pm_stat)
{
	if (cr->pm_ops->bef_pow_dwn)
		cr->pm_ops->bef_pow_dwn(cr, PM_S3);
	else
		rts_trace(TRACE_PM,
			"cr->pm_ops->bef_pow_dwn is not implemented\n");
}
EXPORT_SYMBOL_GPL(cr_pm_bef_pow_dwn);

/* cr_pm_do_resume - Perform PM resume operation
 * @cr: card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int cr_pm_do_resume(struct card_reader *cr)
{
	int ret = 0;

	if (cr->pm_ops->do_resume)
		ret = cr->pm_ops->do_resume(cr);
	else
		rts_trace(TRACE_PM,
			"cr->pm_ops->do_resume is not implemented\n");

	return ret;
}
EXPORT_SYMBOL_GPL(cr_pm_do_resume);
/* cr_pm_update_pm - All operations for updating PM(for debug and obsolete)
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pm_update_pm(struct card_reader *cr)
{
	if (cr->pm_ops->update_pm)
		cr->pm_ops->update_pm(cr);
	else
		rts_trace(TRACE_PM, "update_pm callback is not implemented\n");
}

int cr_set_l1off_sub(struct card_reader *cr, u32 val)
{
	if (cr->pm_ops->set_l1off_sub) {
		return cr->pm_ops->set_l1off_sub(cr, val);
	} else {
		cr_wr_reg(cr, REG_L1SUB_CFG3, 0xFF, (u8) val);
		return 0;
	}

	return 0;
}
EXPORT_SYMBOL_GPL(cr_set_l1off_sub);

static void cr_comm_set_l1off_cfg_sub_D0(struct card_reader *cr, int active)
{
	struct cr_option *option = &(cr->option);
	u32 interrupt = cr_ioread32(cr, BIPR);
	int card_exist = (interrupt & SD_EXIST) | (interrupt & MS_EXIST);
	int aspm_L1_1, aspm_L1_2;

	aspm_L1_1 = check_dev_flag(cr, ASPM_L1_1_EN);
	aspm_L1_2 = check_dev_flag(cr, ASPM_L1_2_EN);

	if (active) {
		/* run, latency: 60us */
		if (aspm_L1_1) {
			u8 val = option->ltr_l1off_snooze_sspwrgate;
			if (check_dev_flag(cr,
					LTR_L1SS_PWR_GATE_CHECK_CARD_EN)) {
				if (card_exist)
					val &= ~(1 << 4);
				else
					val |= (1 << 4);
			}

			if (card_exist) {
				struct card *card = cm_get_card(cr->cm, SD_ID);
				struct sd_info *sd_card = NULL;

				if (card)
					sd_card = card->extra;
				if (sd_card && CHK_SD_UHSII(sd_card))
					val &= ~(1 << 3);
			}

			cr_set_l1off_sub(cr, val);
			rts_trace(TRACE_PM, "ltr active and aspm L1.1,"
							"L1SUB:%#x\n", val);
		} else {
			cr_set_l1off_sub(cr, 0);
			rts_trace(TRACE_PM, "ltr active and not aspm L1.1,"
							"L1SUB:%#x\n", 0x00);
		}
	} else {
		/* l1off, latency: 300us */
		if (aspm_L1_2) {
			u8 val = option->ltr_l1off_sspwrgate;
			struct card *card = cm_get_card(cr->cm, SD_ID);

			if (check_dev_flag(cr,
					LTR_L1SS_PWR_GATE_CHECK_CARD_EN)) {
				if (card_exist)
					val &= ~(1 << 4);
				else
					val |= (1 << 4);
			}

			if (card_exist && card) {
				struct sd_info *sd_card = card->extra;
				if (CHK_SD_UHSII(sd_card))
					val &= ~(1 << 3);
			}

			cr_set_l1off_sub(cr, val);
			rts_trace(TRACE_PM, "ltr idle and aspm L1.2,"
							"L1SUB:%#x\n", val);
		} else {
			cr_set_l1off_sub(cr, 0);
			rts_trace(TRACE_PM, "ltr idle and not aspm L1.2,"
							"L1SUB:%#x\n", 0x00);
		}
	}
}

void cr_set_l1off_sub_cfg_D0(struct card_reader *cr, int active)
{
	if (cr->pm_ops->set_l1off_cfg_sub_D0)
		cr->pm_ops->set_l1off_cfg_sub_D0(cr, active);
	else
		cr_comm_set_l1off_cfg_sub_D0(cr, active);
}
EXPORT_SYMBOL_GPL(cr_set_l1off_sub_cfg_D0);

void cr_set_l1off_sub_cfg_D3(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	int pm_L1_1, pm_L1_2;
	u8 val;

	pm_L1_1 = check_dev_flag(cr, PM_L1_1_EN);
	pm_L1_2 = check_dev_flag(cr, PM_L1_2_EN);

	if (pm_L1_2) {
		val = option->ltr_l1off_sspwrgate;
		cr_set_l1off_sub(cr, val);
	} else {
		val = option->ltr_l1off_snooze_sspwrgate;
		if (pm_L1_1)
			cr_set_l1off_sub(cr, val);
		else
			cr_set_l1off_sub(cr, 0);
	}

}
EXPORT_SYMBOL_GPL(cr_set_l1off_sub_cfg_D3);

int cr_comm_set_ltr_latency(struct card_reader *cr, u32 latency)
{
	cr_wr_reg(cr, REG_MSGTXDATA0, 0xFF, (u8)(latency & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA1, 0xFF, (u8)((latency >> 8) & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA2, 0xFF, (u8)((latency >> 16) & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA3, 0xFF, (u8)((latency >> 24) & 0xFF));
	cr_wr_reg(cr, LTR_CTL, 0xC0, 0xC0);

	return 0;
}

int cr_set_ltr_latency(struct card_reader *cr, u32 latency)
{
	if (cr->pm_ops->set_ltr_latency)
		return cr->pm_ops->set_ltr_latency(cr, latency);
	else
		return cr_comm_set_ltr_latency(cr, latency);

	return 0;
}
EXPORT_SYMBOL_GPL(cr_set_ltr_latency);

static void cr_comm_pm_full_on(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	cr_force_power_on(cr, SSC_PDCTL | OC_PDCTL);
	cr_disable_aspm(cr);

	if (option->ltr_enabled) {
		u32 latency = option->ltr_active_latency;

		rts_trace(TRACE_PM, "Set active latency%#x\n", latency);
		cr_set_ltr_latency(cr, latency);
	}

	if (check_dev_flag(cr, LTR_L1SS_PWR_GATE_EN))
		cr_set_l1off_sub_cfg_D0(cr, 1);
}

/* cr_pm_full_on - Keep the card reader IC in working status, maximum power
 * consuming
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pm_full_on(struct card_reader *cr)
{
	if (cr->pm_ops->full_on)
		cr->pm_ops->full_on(cr);
	else
		cr_comm_pm_full_on(cr);
}

static void cr_comm_pm_power_saving(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	if (option->ltr_enabled) {
		u32 latency = option->ltr_l1off_latency;

		if (check_dev_flag(cr, L1_SNOOZE_TEST_EN))
			mdelay(option->l1_snooze_delay);

		rts_trace(TRACE_PM, "Set L1OFF latency:%#x\n", latency);
		cr_set_ltr_latency(cr, latency);
	}

	if (check_dev_flag(cr, LTR_L1SS_PWR_GATE_EN))
		cr_set_l1off_sub_cfg_D0(cr, 0);

	cr_enable_aspm(cr);
 }

/* cr_pm_power_saving - Perform some operations to save power to some extent
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pm_power_saving(struct card_reader *cr)
{
	if (cr->pm_ops->power_saving)
		cr->pm_ops->power_saving(cr);
	else
		cr_comm_pm_power_saving(cr);
}

static void cr_comm_pm_power_saving_max(struct card_reader *cr)
{
	/* TODO */
}

/* cr_pm_power_saving - Perform some operations to save power to the max
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pm_power_saving_max(struct card_reader *cr)
{
	if (cr->pm_ops->power_saving_max)
		cr->pm_ops->power_saving_max(cr);
	else
		cr_comm_pm_power_saving_max(cr);
}

extern int ss_delay;
/* cr_pm_update_pm_real - Implement PM state machine
 * @cr: card reader structure
 *
 * Returns: Nothing
 */
void cr_pm_update_pm_real(struct card_reader *cr)
{
	int stat = rtsx_get_stat(cr);
	int ss_counter;
	struct cr_option *option = &cr->option;

	switch (stat) {
	case RTSX_STAT_INIT:
	case RTSX_STAT_RUN:
		rts_trace(TRACE_POLL_THRD, "enter RTSX_STAT_RUN: ss_counter=%d, idle_counter=%d\n",
			cr->ss_counter, cr->idle_counter);
		if (!cr_pm_check_flags(cr, PM_FULL_ON)) {
			cr_pm_full_on(cr);
			cr_pm_set_flags(cr, PM_FULL_ON);
			cr->auto_delink_timer = cr_get_tick();
		}
		if (!CHECK_STATUS_FLAG(cr, LED_TEST_MODE))
			cr_blink_led(cr);
		cr->ss_counter = 0;
		if (cr->idle_counter < (int)option->idle_max_cnt) {
			cr->idle_counter++;
			break;
		}
	case RTSX_STAT_IDLE:
		rts_trace(TRACE_POLL_THRD, "enter RTSX_STAT_IDLE: ss_counter=%d\n",
			cr->ss_counter);
		cm_do_remaining_work(cr->cm);
		ss_counter = (ss_delay * 1000) / cr->polling_interval;
		cr_set_stat(cr, RTSX_STAT_IDLE);
		if (!cr_pm_check_flags(cr, PM_POWR_SAVING)) {
			if (!CHECK_STATUS_FLAG(cr, LED_TEST_MODE)) {
				cr_turn_off_led(cr);
				cr_disable_auto_blink(cr);
			}
			cr_pm_power_saving(cr);
			cr_pm_set_flags(cr, PM_POWR_SAVING);
		}
		if (cr->ss_counter < ss_counter) {
			/* cr_pm_trigger_led(cr, ON); */
			cr->ss_counter++;
			break;
		}
	case RTSX_STAT_SS:
		rts_trace(TRACE_POLL_THRD, "enter RTSX_STAT_SS\n");
		if (!cr_pm_check_flags(cr, PM_POWR_SAVING_MAX)) {
			if (!CHECK_STATUS_FLAG(cr, LED_TEST_MODE)) {
				int on = cm_get_led_state_at_ss(cr->cm);

				if (on)
					cr_turn_on_led(cr);
				else
					cr_turn_off_led(cr);
			}
			cr_pm_power_saving_max(cr);
			cr_pm_set_flags(cr, PM_POWR_SAVING_MAX);
		}
		if (!rtsx_chk_stat(cr, RTSX_STAT_SS))
			cr_set_stat(cr, RTSX_STAT_SS);
		break;
	case RTSX_STAT_DELINK:
		break;
	case RTSX_STAT_SUSPEND:
		break;
	case RTSX_STAT_ABORT:
		break;
	case RTSX_STAT_DISCONNECT:
		break;
	default:
		rts_trace(TRACE_PM, "Unknonwn state: %d\n", stat);
	}
}

static int cr_comm_force_power_on(struct card_reader *cr, u8 ctl)
{
	int retval;
	u8 mask = 0;

	if (ctl & SSC_PDCTL)
		mask |= SSC_POWER_DOWN;

	if (cr->option.ocp_en && (ctl & OC_PDCTL))
		mask |= SD_OC_POWER_DOWN;
	if (mask) {
		retval = cr_wr_reg(cr, REG_FPDCTL, mask, 0);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

int cr_force_power_on(struct card_reader *cr, u8 ctl)
{
	if (cr->pm_ops->force_power_on)
		return cr->pm_ops->force_power_on(cr, ctl);
	else
		return cr_comm_force_power_on(cr, ctl);
}
EXPORT_SYMBOL_GPL(cr_force_power_on);

static int cr_comm_force_power_down(struct card_reader *cr, u8 ctl)
{
	int retval;
	u8 mask = 0, val = 0;

	if (ctl & SSC_PDCTL)
		mask |= SSC_POWER_DOWN;
	if (cr->option.ocp_en && (ctl & OC_PDCTL))
		mask |= SD_OC_POWER_DOWN;

	if (mask) {
		val = mask;
		retval = cr_wr_reg(cr, REG_FPDCTL, mask, val);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

int cr_force_power_down(struct card_reader *cr, u8 ctl)
{
	if (cr->pm_ops->force_power_down)
		return cr->pm_ops->force_power_down(cr, ctl);
	else
		return cr_comm_force_power_down(cr, ctl);
}
EXPORT_SYMBOL_GPL(cr_force_power_down);

static void cr_comm_enable_aspm(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	rts_trace(TRACE_PM, "dev_aspm_val = %#x\n", option->dev_aspm_val);
	rts_trace(TRACE_PM, "dev_aspm_mode = %#x\n", option->dev_aspm_mode);

	if (!cr->aspm_enabled) {
		if (option->dev_aspm_mode != DEV_ASPM_DISABLE) {
			u8 val = 0;

			rts_trace(TRACE_PM, "Try to enable ASPM\n");
			if (option->dev_aspm_mode == DEV_ASPM_DYNAMIC) {
				int lctlr = cr->dev_pci_cap_offset + PCI_EXP_LNKCTL;
				cr_read_config8(cr, lctlr, &val);
				val &= 0xFC;
				val |= (option->dev_aspm_val & 0x03);
				cr_write_config8(cr, lctlr, val);
			} else if (option->dev_aspm_mode == DEV_ASPM_BACKDOOR) {
				u8 mask = FORCE_ASPM_VAL_MASK;
				val = option->dev_aspm_val;
				cr_wr_reg(cr, REG_ASPM_FORCE_CTL, mask, val);
			}
		}
		if (option->config_host_aspm)
			cr_set_host_aspm(cr, cr->host_aspm_val);
		cr->aspm_enabled = 1;
	}
}

void cr_enable_aspm(struct card_reader *cr)
{
	if (cr->pm_ops->enable_aspm)
		cr->pm_ops->enable_aspm(cr);
	else
		cr_comm_enable_aspm(cr);
}
EXPORT_SYMBOL_GPL(cr_enable_aspm);

static void cr_comm_disable_aspm(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	if (cr->aspm_enabled) {
		if (option->config_host_aspm)
			cr_disable_host_aspm(cr);
		if (DEV_ASPM_DISABLE != option->dev_aspm_mode) {
			rts_trace(TRACE_PM, "Try to disable ASPM\n");
			if (DEV_ASPM_DYNAMIC == option->dev_aspm_mode) {
				int lctlr = cr->dev_pci_cap_offset + PCI_EXP_LNKCTL;
				u8 val = 0;
				cr_read_config8(cr, lctlr, &val);
				val &= 0xFC;
				cr_write_config8(cr, lctlr, val);
			} else if (DEV_ASPM_BACKDOOR == option->dev_aspm_mode) {
				u8 mask = FORCE_ASPM_VAL_MASK;
				cr_wr_reg(cr, REG_ASPM_FORCE_CTL, mask, 0);
			}
			msleep(10);
		}
		cr->aspm_enabled = 0;
	}
}

void cr_disable_aspm(struct card_reader *cr)
{
	if (cr->pm_ops->disable_aspm)
		cr->pm_ops->disable_aspm(cr);
	else
		cr_comm_disable_aspm(cr);
}
EXPORT_SYMBOL_GPL(cr_disable_aspm);
