/* Driver for Realtek card reader
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>
#include <linux/errno.h>
#include <linux/freezer.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <linux/utsname.h>
#include <linux/pci.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_devinfo.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_host.h>

#include "rts_cr.h"
#include "option.h"
#include "pm.h"
#include "cardmanager.h"
#include "rts_scsi.h"
#include "transport.h"
#include "trans_pcie.h"
#include "cr_driver.h"
#include "rts_rpm.h"

MODULE_DESCRIPTION(RTS_CR_DESC);
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);

uint32_t rts_trace_mask = RTSX_TRACE_MASK;

module_param_named(trace_mask, rts_trace_mask, uint, S_IWUSR | S_IRUGO);
EXPORT_SYMBOL_GPL(rts_trace_mask);
MODULE_PARM_DESC(trace_mask, "mask for print trace massege");

#ifdef SCSI_SCAN_DELAY
static unsigned int delay_use = 5;
module_param(delay_use, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(delay_use, "seconds to delay before using a new device");
#endif

int auto_delink_en;
module_param(auto_delink_en, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(auto_delink_en, "enable auto delink");

int ss_en;
module_param(ss_en, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(ss_en, "enable selective suspend");

int ss_delay = 3;/* 50; */
module_param(ss_delay, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(ss_delay,
		 "seconds to delay before entering selective suspend");

int needs_remote_wakeup = 1;
module_param(needs_remote_wakeup, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(needs_remote_wakeup, "ss state needs remote wakeup supported");

static DEFINE_RWLOCK(cr_rwlock);
static struct list_head cr_capability_list;

/*
 * cr_find_capability - find cr_capability by pid and vid
 *
 * note: cr_find_capability don't lock cr_rwlock
 */
static struct cr_capability *cr_find_capability(u16 pid, u16 vid)
{
	struct list_head *pos;
	struct cr_capability *capability;

	list_for_each(pos, &cr_capability_list) {
		capability = list_entry(pos, struct cr_capability, next);
		if (capability->pid == pid && capability->vid == vid)
			return capability;
	}

	return NULL;
}

/*
 * cr_get_capability - get cr_capability by pid and vid
 *
 */
static struct cr_capability *cr_get_capability(u16 pid, u16 vid)
{
	struct cr_capability *capability;

	read_lock(&cr_rwlock);
	capability = cr_find_capability(pid, vid);
	read_unlock(&cr_rwlock);

	return capability;
}


/*
 * cr_register_capability - add capability to cr_capability_list
 *
 */
int cr_register_capability(struct cr_capability *capability)
{
	int ret = 0;

	write_lock(&cr_rwlock);
	if (cr_find_capability(capability->pid, capability->vid)) {
		ret = -EBUSY;
		goto exit;
	}

	list_add_tail(&capability->next, &cr_capability_list);

exit:
	write_unlock(&cr_rwlock);
	return ret;
}
EXPORT_SYMBOL_GPL(cr_register_capability);

/*
 *cr_ungister_capabililty - remove cr_capability from cr_capability_list
 *by pid and vid
 */
void cr_ungister_capabililty(u16 pid, u16 vid)
{
	struct cr_capability *capability;
	write_lock(&cr_rwlock);
	capability = cr_find_capability(pid, vid);
	if (!capability)
		goto exit;
	list_del(&capability->next);
exit:
	write_unlock(&cr_rwlock);
}
EXPORT_SYMBOL_GPL(cr_ungister_capabililty);

/* Second stage of disconnect processing: deallocate all resources */
static void release_everything(struct card_reader *cr)
{
	rts_trace(TRACE_CR, "<--\n");

	/* If the device is really gone, cut short reset delays */
	rts_trace(TRACE_CR, "Telling the tranport medium is disconnected\n");
	set_bit(FLIDX_DISCONNECTING, &cr->scsi->dflags);
	/* Wait polling thread exits, for scsi->dflags will be cleaned up
	 *after SCSI Host is removed */
	if (cr->polling_thread) {
		/*
		 * wake up polling thread first
		 */
		cr_wakeup_polling_thread(cr);
		rts_trace(TRACE_CR, "Waiting polling thread exit ...\n");
		wait_for_completion(&cr->polling_exit);
	}
	quiesce_and_remove_host(cr->scsi);
	cr_pm_bef_pow_dwn(cr, PM_S3);
	transport_release(cr->tr);
	cm_release(cr->cm);
	cr_release(cr);

	rts_trace(TRACE_CR, "-->\n");
}

/* Build up the relationship among each other */
static void rts_cr_bind_together(struct card_reader *cr,
	struct card_manager *cm, struct transport *tr,
	struct rts_scsi *scsi)
{
	rts_trace(TRACE_CR, "<--\n");

	BUG_ON(!cr);
	BUG_ON(!cm);
	BUG_ON(!tr);
	BUG_ON(!scsi);

	cr->cm = cm;
	cr->tr = tr;
	cr->scsi = scsi;
	cm->cr = cr;
	cm->tr = tr;
	tr->cr = cr;
	tr->cm = cm;
	scsi->cr = cr;
	rts_trace(TRACE_CR, "cr=%p, cm =%p, tr=%p, scsi=%p\n"
			"cr->cm=%p, cr->tr=%p, cr->scsi=%p\n"
			"cm->cr=%p, cm->tr=%p\n scsi->cr=%p\n",
			cr, cm, tr, scsi, cr->cm, cr->tr, cr->scsi,
			cm->cr, cm->tr,	scsi->cr);

	rts_trace(TRACE_CR, "-->\n");
}

static int rts_cr_init_comm(struct card_reader **__cr,  struct device *dev,
	__u16 pid, __u16 vid, int tr_preserve_size, void *tr_extra)
{
	struct cr_capability *cr_capa = NULL;
	struct card_reader *cr = NULL;
	struct card_manager *cm = NULL;
	struct transport *tr = NULL;
	struct rts_scsi *scsi = NULL;
	struct cr_driver *cr_driver = NULL;
	int ret = 0;

	rts_trace(TRACE_CR, "%s detected\n", RTS_CR_NAME);

	cr_capa = cr_get_capability(pid, vid);
	if (!cr_capa) {
		rts_trace(TRACE_ERROR, "driver for %#x %#x was not load\n",
				pid, vid);
		ret = -EINVAL;
		goto err_no_match_cr;
	}
	cr_driver = cr_capa->drv_data;

	tr = transport_alloc(tr_preserve_size);
	if (NULL == tr) {
		ret = -ENOMEM;
		goto err_transport_alloc;
		return ret;
	}

	cr = cr_alloc(cr_capa->extra_size);
	if (NULL == cr) {
		ret = -ENOMEM;
		goto err_cr_alloc;
	}

	cm = cm_alloc();
	if (NULL == cm) {
		ret = -ENOMEM;
		goto err_cm_alloc;
	}

	scsi = scsi_alloc();
	if (NULL == scsi) {
		ret = -ENOMEM;
		goto err_scsi_alloc;
	}

	rts_cr_bind_together(cr, cm, tr, scsi);

	ret = transport_init(tr, cr_capa->tr, tr_extra);
	if (ret)
		goto err_transport_init;


	ret = cm_init(cm, &cr_capa->card_conf);
	if (ret)
		goto err_cm_init;

	ret = cr_init(cr, cr_capa, tr);
	if (ret)
		goto err_cr_init;

	/* init param from sysfs */
	memcpy(&cr->option, &cr_driver->sw_param, sizeof(cr->option));
	memcpy(&cr->hw_param, &cr_driver->hw_param, sizeof(cr->hw_param));

	cr_init_option(cr);

	cr_set_card_spt_map_sw(cr, cr->option.card_spt_map);
	cr_set_sd_capability_sw(cr, cr->option.sd_capability);
	cr->polling_interval = cr->option.polling_interval_run;

	ret = cr_init_hw(cr);
	if (ret != STATUS_SUCCESS)
		goto err_cr_init_hw;

	/* Reset card if the card is plugged in during the initialised stage */
	/* Reset card in Polling thread */
	/* cm_set_reset_map_directly(cr->cm);
	if (cr->reset_map)
		cm_process_card_reset_directly(cr->cm, cr->reset_map); */

	/* NOTE: Setup PM, which MUST be placed after real hw initialization */
#if RTS_PM
	cr_pm_setup(cr);
#endif

	ret = scsi_init(scsi, dev);
	if (ret) {
		goto err_scsi_init;
	}

	ret = start_polling_thread(cr);
	if (ret)
		goto err_start_polling_thread;

	*__cr = cr;

	return 0;
err_start_polling_thread:
err_scsi_init:
err_cr_init_hw:
err_cm_init:
err_cr_init:
err_transport_init:
	scsi_release(scsi);
err_scsi_alloc:
	cm_release(cm);
err_cm_alloc:
	cr_release(cr);
err_cr_alloc:
	transport_release(tr);
err_transport_alloc:
err_no_match_cr:
	return ret;
}

int rts_cr_pcie_probe(struct pci_dev *pci,
	const struct pci_device_id *id)
{
	struct card_reader *cr = NULL;
	int ret = 0;

	rts_trace(TRACE_CR, "<--\n");

	ret = rts_cr_init_comm(&cr, &pci->dev, id->device, id->vendor,
		sizeof(struct rts_pcie), pci);
	if (ret) {
		printk("%s: Err!!!\n", __func__);
		return ret;
	}
	rts_trace(TRACE_CR, "cr=%p\n", cr);

	/* Store our private data in the pci */
	pci_set_drvdata(pci, cr);
	if (cr->option.rtd3_en)
		cr_pci_runtime_pm_allow(cr);

	rts_trace(TRACE_CR, "-->\n");

	return 0;
}
EXPORT_SYMBOL_GPL(rts_cr_pcie_probe);

void rts_cr_pcie_remove(struct pci_dev *pci)
{

	struct card_reader *cr = (struct card_reader *)pci_get_drvdata(pci);

	rts_trace(TRACE_CR, "<--\n");

	if (cr->option.rtd3_en)
		cr_pci_runtime_pm_forbid(cr);

	if (cr)
		release_everything(cr);
	pci_set_drvdata(pci, NULL);

	rts_trace(TRACE_CR, "-->\n");
}
EXPORT_SYMBOL_GPL(rts_cr_pcie_remove);

static int __init rts_cr_init(void)
{
	INIT_LIST_HEAD(&cr_capability_list);

	rts_trace(TRACE_INFO, "init realtek cardreader driver, trace mask 0x%x\n",
		rts_trace_mask);

	return 0;
}

static void __exit rts_cr_exit(void)
{
	rts_trace(TRACE_INFO, "exit realtek cardreader driver\n");
}

module_init(rts_cr_init);
module_exit(rts_cr_exit);
