/* Driver for Realtek card reader
 *
 * Copyright(c) 2009 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   wwang (wei_wang@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/module.h>
#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/freezer.h>
#include <linux/pci.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>

#include "rts_scsi.h"
#include "trace.h"
#include "rts_cr.h"
#include "pm.h"
#include "cardmanager.h"
#include "cards/ms.h"
#include "transport.h"
#include "scsi_cmd.h"


void scsi_invoke_transport(struct scsi_cmnd *srb, struct card_reader *cr)
{
	int result = 0;

#if RTS_PM
	result = scsi_pm(srb, cr);
	if (result < 0)
		goto out;
#endif
	result = scsi_handler(srb, cr);
	/* if there is a transport error, reset and don't auto-sense */
	if (result == TRANSPORT_ERROR) {
		rts_trace(TRACE_SCSI, "-- transport indicates error, resetting\n");
		srb->result = DID_ERROR << 16;
		goto Handle_Errors;
	}

	srb->result = SAM_STAT_GOOD;

	/*
	 * If we have a failure, we're going to do a REQUEST_SENSE
	 * automatically.  Note that we differentiate between a command
	 * "failure" and an "error" in the transport mechanism.
	 */
	if (result == TRANSPORT_FAILED) {
		/* set the result so the higher layers expect this data */
		srb->result = SAM_STAT_CHECK_CONDITION;
		memcpy(srb->sense_buffer,
		       (unsigned char *)&(cr->scsi->sense_buffer[SCSI_LUN(srb)]),
		       sizeof(struct sense_data_t));
	}
out:
	return;

	/* Error and abort processing: try to resynchronize with the device
	 * by issuing a port reset.  If that fails, try a class-specific
	 * device reset. */
Handle_Errors:
	return;
}

/***********************************************************************
 * Host functions
 ***********************************************************************/

int slave_alloc(struct scsi_device *sdev)
{
	/*
	 * Set the INQUIRY transfer length to 36.  We don't use any of
	 * the extra data and many devices choke if asked for more or
	 * less than 36 bytes.
	 */
	sdev->inquiry_len = 36;
	return 0;
}

int slave_configure(struct scsi_device *sdev)
{
	/* Scatter-gather buffers (all but the last) must have a length
	 * divisible by the bulk maxpacket size.  Otherwise a data packet
	 * would end up being short, causing a premature end to the data
	 * transfer.  Since high-speed bulk pipes have a maxpacket size
	 * of 512, we'll use that as the scsi device queue's DMA alignment
	 * mask.  Guaranteeing proper alignment of the first buffer will
	 * have the desired effect because, except at the beginning and
	 * the end, scatter-gather buffers follow page boundaries. */
	blk_queue_dma_alignment(sdev->request_queue, (512 - 1));

	/* Set the SCSI level to at least 2.  We'll leave it at 3 if that's
	 * what is originally reported.  We need this to avoid confusing
	 * the SCSI layer with devices that report 0 or 1, but need 10-byte
	 * commands (ala ATAPI devices behind certain bridges, or devices
	 * which simply have broken INQUIRY data).
	 *
	 * NOTE: This means /dev/sg programs (ala cdrecord) will get the
	 * actual information.  This seems to be the preference for
	 * programs like that.
	 *
	 * NOTE: This also means that /proc/scsi/scsi and sysfs may report
	 * the actual value or the modified one, depending on where the
	 * data comes from.
	 */
	if (sdev->scsi_level < SCSI_2)
		sdev->scsi_level = sdev->sdev_target->scsi_level = SCSI_2;

	return 0;
}

/***********************************************************************
 * /proc/scsi/ functions
 ***********************************************************************/

/* we use this macro to help us write into the buffer */
#undef SPRINTF
#define SPRINTF(args...) \
	do { if (pos < buffer+length) pos += sprintf(pos, ## args); } while (0)

/* queue a command */
/* This is always called with scsi_lock(host) held */
int queuecommand_lck(struct scsi_cmnd *srb, void (*done) (struct scsi_cmnd *))
{
	struct rts_scsi *scsi = host_to_scsi(srb->device->host);

	/* check for state-transition errors */
	if (scsi->srb != NULL) {
		rts_trace(TRACE_SCSI,
				"Error: cr->srb = %p\n", scsi->srb);
		return SCSI_MLQUEUE_HOST_BUSY;
	}

	/* fail the command if we are disconnecting */
	if (test_bit(FLIDX_DISCONNECTING, &scsi->dflags)) {
		rts_trace(TRACE_SCSI, "Fail command during disconnect\n");
		srb->result = DID_NO_CONNECT << 16;
		done(srb);
		return 0;
	}

	/* enqueue the command and wake up the control thread */
	srb->scsi_done = done;
	scsi->srb = srb;
	complete(&scsi->cmnd_ready);

	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 37)
int queuecommand(struct scsi_cmnd *srb, void (*done)(struct scsi_cmnd *))
{
	return queuecommand_lck(srb, done);
}
#else
DEF_SCSI_QCMD(queuecommand)
#endif	/* KERNEL_VERSION(2, 6, 37) */

/***********************************************************************
 * Error handling functions
 ***********************************************************************/
/* Command timeout and abort */
int command_abort(struct scsi_cmnd *srb)
{
	struct rts_scsi *scsi = host_to_scsi(srb->device->host);

	rts_trace(TRACE_SCSI, "called\n");

	/* us->srb together with the TIMED_OUT, RESETTING, and ABORTING
	 * bits are protected by the host lock. */
	scsi_lock(scsi_to_host(scsi));

	/* Is this command still active? */
	if (scsi->srb != srb) {
		scsi_unlock(scsi_to_host(scsi));
		rts_trace(TRACE_SCSI, "-- nothing to abort\n");
		return FAILED;
	}

	/* Set the TIMED_OUT bit.  Also set the ABORTING bit, but only if
	 * a device reset isn't already in progress (to avoid interfering
	 * with the reset).  Note that we must retain the host lock while
	 * calling usb_stor_stop_transport(); otherwise it might interfere
	 * with an auto-reset that begins as soon as we release the lock. */
	set_bit(FLIDX_TIMED_OUT, &scsi->dflags);
	rts_trace(TRACE_SCSI, "%d, set timeout flag\n", __LINE__);
	if (!test_bit(FLIDX_RESETTING, &scsi->dflags)) {
		rts_trace(TRACE_SCSI, "%d, set abort flag\n", __LINE__);
		set_bit(FLIDX_ABORTING, &scsi->dflags);
		/* rts51x_stop_transport(us); */
	}
	scsi_unlock(scsi_to_host(scsi));

	/* Wait for the aborted command to finish */
	wait_for_completion(&scsi->notify);
	return SUCCESS;
}

/* This invokes the transport reset mechanism to reset the state of the
 * device */
int device_reset(struct scsi_cmnd *srb)
{
	int result = 0;

	rts_trace(TRACE_SCSI, "called\n");

	return result < 0 ? FAILED : SUCCESS;
}

/* Simulate a SCSI bus reset by resetting the device's USB port. */
int bus_reset(struct scsi_cmnd *srb)
{
	int result = 0;

	rts_trace(TRACE_SCSI, "called\n");

	return result < 0 ? FAILED : SUCCESS;
}

static const char *rts_info(struct Scsi_Host *host)
{
	return "SCSI emulation for Realtek PCIE/USB card reader";
}

struct scsi_host_template rts_host_template = {
	/* basic userland interface stuff */
	.name = RTS_CR_NAME,
	.proc_name = RTS_CR_NAME,
	.info = rts_info,

	/* command interface -- queued only */
	.queuecommand = queuecommand,

	/* error and abort handlers */
	.eh_abort_handler = command_abort,
	.eh_device_reset_handler = device_reset,
	.eh_bus_reset_handler = bus_reset,

	/* queue commands only, only one command per LUN */
	.can_queue = 1,
	.cmd_per_lun = 1,

	/* unknown initiator id */
	.this_id = -1,

	.slave_alloc = slave_alloc,
	.slave_configure = slave_configure,

	/* lots of sg segments can be handled */
	.sg_tablesize = SG_ALL,

	/* limit the total size of a transfer to 1 MB */
	.max_sectors = 2048,

	/* merge commands... this seems to help performance, but
	 * periodically someone should test to see which setting is more
	 * optimal.
	 */
	.use_clustering = 1,

	/* emulated HBA */
	.emulated = 1,

	/* we do our own delay after a device or bus reset */
	.skip_settle_delay = 1,

	/* sysfs device attributes */
	/* .sdev_attrs = sysfs_device_attr_list, */

	/* module management */
	.module = THIS_MODULE
};

/* cm_release - Release rts_scsi structure
 *
 * Returns: Nothing
 */
void scsi_release(struct rts_scsi *scsi)
{
	rts_trace(TRACE_SCSI, "<--\n");

	rts_trace(TRACE_SCSI, "-->\n");
}

int scsi_control_thread(void *__scsi)
{
	struct rts_scsi *scsi = (struct rts_scsi *)__scsi;
	struct card_reader *cr = scsi->cr;
	struct transport *tr = cr->tr;
	struct Scsi_Host *host = scsi_to_host(scsi);

	for (;;) {
		if (wait_for_completion_interruptible(&scsi->cmnd_ready))
			break;

		if (test_bit(FLIDX_DISCONNECTING, &scsi->dflags)) {
			rts_trace(TRACE_SCSI, "exiting for disconnection\n");
			break;
		}

		if (need_normal_power(scsi->srb)) {
			if (CHECK_STATUS_FLAG(cr, ENTERED_RTD3)) {
				struct completion done_resume;
				cr->done_resume = &done_resume;
				init_completion(cr->done_resume);
				cr_wakeup_polling_thread(cr);
				rts_trace(TRACE_SD,
					"-------------------- wait reset card --------------------\n");
				wait_for_completion(cr->done_resume);
				cr->done_resume = NULL;
			} else if (rtsx_chk_stat(cr, RTSX_STAT_SS)) {
				cr_wakeup_polling_thread(cr);
			}
		}

		/* lock the device pointers */
		mutex_lock(&tr->trans_mutex);

		/* lock access to the state */
		scsi_lock(host);

		/* When we are called with no command pending, we're done */
		if (scsi->srb == NULL) {
			scsi_unlock(host);
			mutex_unlock(&tr->trans_mutex);
			rts_trace(TRACE_SCSI, "exiting for srb is NULL\n");
			break;
		}

		/* has the command timed out *already* ? */
		if (test_bit(FLIDX_TIMED_OUT, &scsi->dflags)) {
			scsi->srb->result = DID_ABORT << 16;
			goto SkipForAbort;
		}

		scsi_unlock(host);

		/* reject the command if the direction indicator
		 * is UNKNOWN
		 */
		if (scsi->srb->sc_data_direction == DMA_BIDIRECTIONAL) {
			rts_trace(TRACE_SCSI, "UNKNOWN data direction\n");
			scsi->srb->result = DID_ERROR << 16;
		}

		/* reject if target != 0 or if LUN is higher than
		 * the maximum known LUN
		 */
		else if (scsi->srb->device->id) {
			rts_trace(TRACE_SCSI, "Bad target number (%d:%d)\n",
				scsi->srb->device->id,
				(int)scsi->srb->device->lun);
			scsi->srb->result = DID_BAD_TARGET << 16;
		}

		else if (scsi->srb->device->lun > cr->max_lun) {
			rts_trace(TRACE_SCSI, "Bad LUN (%d:%d)\n",
				scsi->srb->device->id,
				(int)scsi->srb->device->lun);
			scsi->srb->result = DID_BAD_TARGET << 16;
		}
		/* we've got a command, let's do it! */
		else {
			scsi_show_command(scsi->srb);
			scsi_invoke_transport(scsi->srb, cr);
		}

		/* lock access to the state */
		scsi_lock(host);

		/* indicate that the command is done */
		if (scsi->srb->result != DID_ABORT << 16)
			scsi->srb->scsi_done(scsi->srb);
		else
SkipForAbort :
			rts_trace(TRACE_SCSI, "scsi command aborted\n");

		/* If an abort request was received we need to signal that
		 * the abort has finished.  The proper test for this is
		 * the TIMED_OUT flag, not srb->result == DID_ABORT, because
		 * the timeout might have occurred after the command had
		 * already completed with a different result code. */
		if (test_bit(FLIDX_TIMED_OUT, &scsi->dflags)) {
			rts_trace(TRACE_SCSI, "notify abort has finished\n");
			complete(&(scsi->notify));

			/* Allow USB transfers to resume */
			clear_bit(FLIDX_ABORTING, &scsi->dflags);
			clear_bit(FLIDX_TIMED_OUT, &scsi->dflags);
		}

		/* finished working on this command */
		scsi->srb = NULL;
		scsi_unlock(host);

		/* unlock the device pointers */
		mutex_unlock(&tr->trans_mutex);
	}			/* for (;;) */

	complete(&scsi->control_exit);

	/* Wait until we are told to stop */
/*	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);
		if (kthread_should_stop())
			break;
		schedule();
	}
	__set_current_state(TASK_RUNNING);*/
	return 0;
}

#ifdef SCSI_SCAN_DELAY
extern unsigned int delay_use;
/* Thread to carry out delayed SCSI-device scanning */
static int scsi_scan_thread(void *__scsi)
{
	struct rts_scsi *scsi = (struct rts_scsi *)__scsi;
	struct card_reader *cr = scsi->cr;
	struct Scsi_Host *host = scsi_to_host(scsi);

	set_freezable();
	/* Wait for the timeout to expire or for a disconnect */
	if (delay_use > 0) {
		printk("%s: waiting for device to settle before"
							"scanning\n", cr->name);
		wait_event_freezable_timeout(scsi->delay_wait,
			test_bit(FLIDX_DONT_SCAN, &scsi->dflags), delay_use*HZ);
	}

	/* If the device is still connected, perform the scanning */
	if (!test_bit(FLIDX_DONT_SCAN, &scsi->dflags)) {
		scsi_scan_host(host);
		printk("%s: device scan complete\n", cr->name);
	}

	complete_and_exit(&scsi->scanning_done, 0);
}
#endif

/* scsi_init - Initialise struct rts_scsi
 *
 * Returns: 0 for success, otherwise nonzero
 */
int scsi_init(struct rts_scsi *scsi, struct device *dev)
{
	struct Scsi_Host *shost = scsi_to_host(scsi);
	struct task_struct *ctrl_th;
#ifdef SCSI_SCAN_DELAY
	struct task_struct *scan_th;
#endif
	int ret = 0;

	rts_trace(TRACE_SCSI, "<--\n");

	init_completion(&scsi->cmnd_ready);
	init_completion(&scsi->control_exit);
	init_completion(&(scsi->notify));
#ifdef SCSI_SCAN_DELAY
	init_waitqueue_head(&scsi->delay_wait);
	init_completion(&scsi->scanning_done);
#endif
	/* Allow 16-byte CDBs and thus > 2TB */
	shost->max_cmd_len = 16;

	/* set the supported max_lun and max_id for the scsi host
	 * NOTE: the minimal value of max_id is 1 */
	shost->max_id = 1;
	shost->max_lun = scsi->cr->max_lun;
#if 1
	/* Start up our control thread */
	ctrl_th = kthread_run(scsi_control_thread, scsi, RTS_CR_CTL_THREAD);
	if (IS_ERR(ctrl_th)) {
		printk(KERN_WARNING RTS_TIP
		       "Unable to start control thread\n");
		ret = PTR_ERR(ctrl_th);
		goto BadDevice;
	}

	scsi->ctl_thread = ctrl_th;
#endif

	ret = scsi_add_host(scsi_to_host(scsi), dev);
	if (ret) {
		printk(KERN_WARNING RTS_TIP "Unable to add the scsi host\n");
		goto BadDevice;
	}
#ifdef SCSI_SCAN_DELAY
	/* Start up the thread for delayed SCSI-device scanning */
	scan_th = kthread_create(scsi_scan_thread, scsi, RTS_CR_SCAN_THREAD);
	if (IS_ERR(scan_th)) {
		printk(KERN_WARNING RTS_TIP
		       "Unable to start the device-scanning thread\n");
		complete(&scsi->scanning_done);
		quiesce_and_remove_host(scsi);
		ret = PTR_ERR(scan_th);
		goto BadDevice;
	}

	wake_up_process(scan_th);
#else
	scsi_scan_host(scsi_to_host(scsi));
#endif

	rts_trace(TRACE_SCSI, "-->\n");

BadDevice:
	return ret;

}

/* First stage of disconnect processing: stop SCSI scanning,
 * remove the host, and stop accepting new commands
 */
void quiesce_and_remove_host(struct rts_scsi *scsi)
{
	struct Scsi_Host *host = scsi_to_host(scsi);

	rts_trace(TRACE_SCSI, "<--\n");

	/* Tell the control thread to exit.  The SCSI host must
	 * already have been removed and the DISCONNECTING flag set
	 * so that we won't accept any more commands.
	 */
	rts_trace(TRACE_SCSI, "sending exit command to thread\n");
	complete(&scsi->cmnd_ready);
	if (scsi->ctl_thread) {
		rts_trace(TRACE_CR, "Waiting control thread exit ...\n");
		wait_for_completion(&scsi->control_exit);
	}

#ifdef SCSI_SCAN_DELAY
	/* Prevent SCSI-scanning (if it hasn't started yet)
	 * and wait for the SCSI-scanning thread to stop.
	 */
	set_bit(FLIDX_DONT_SCAN, &scsi->dflags);
	wake_up(&scsi->delay_wait);
	wait_for_completion(&scsi->scanning_done);
#endif
	/* Removing the host will perform an orderly shutdown: caches
	 * synchronized, disks spun down, etc.
	 */
	scsi_remove_host(host);

#ifdef SCSI_SCAN_DELAY
	wake_up(&scsi->delay_wait);
#endif
	/* Drop our reference to the host; the SCSI core will free it
	 * (and "cr" along with it) when the refcount becomes 0. */
	scsi_host_put(host);

	rts_trace(TRACE_SCSI, "-->\n");
}

/* scsi_alloc - Allocate memory for rts_scsi structure
 *
 * Returns: address to the rts_scsi structure
 */
struct rts_scsi *scsi_alloc(void)
{

	struct Scsi_Host *host = NULL;
	struct rts_scsi *scsi = NULL;

	rts_trace(TRACE_SCSI, "<--\n");

	/*
	 * Ask the SCSI layer to allocate a host structure, with extra
	 * space at the end for rts_scsi structure.
	 */
	host = scsi_host_alloc(&rts_host_template, sizeof(struct rts_scsi));
	if (!host) {
		printk(KERN_WARNING "Unable to allocate the scsi host\n");
		return NULL;
	}

	scsi = host_to_scsi(host);

	rts_trace(TRACE_SCSI, "-->\n");

	return scsi;
}
