/* Driver for Realtek PCIe card reader
 *
 * Copyright(c) 2009 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   wwang (wei_wang@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/freezer.h>
#include <linux/pci.h>
#include <linux/cdrom.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include "cardreader.h"
#include "operations.h"
#include "pm.h"
#include "cardmanager.h"
#include "cards.h"
#include "rts_scsi.h"
#include "scsi_cmd.h"
#include "scsi_xfer.h"
#include "cards/sd.h"
#include "cards/ms.h"
#include "cards/scsi_sd.h"
#include "cards/ms_mg.h"
#include "cr_regs.h"

int scsi_common_get_dev_sts(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);
	int buf_len;
	u8 status[32];

	if (!card)
		return TRANSPORT_FAILED;

	memset(status, 0, sizeof(status));

	/* HW/FW version (IC Version instead) */
	status[0] = (u8)((cr->ic_version) | (cr->ic_version<<4));
	status[1] = 0;

	/* Auto delink mode */
	if (CHK_AUTODELINK_EN(cr))
		status[2] = 0x10;
	else
		status[2] = 0x00;

	/* Spec version */
	status[3] = 20;
	status[4] = 10;
	status[5] = 5;
	status[6] = 21;

	/* Card WP */
	if (card && check_card_flag(card, WP))
		status[7] = 0x20;
	else
		status[7] = 0x00;

	/* Over current status */
	if (cr->option.ocp_en) {
		status[8] = 0;
		if (cr->ocp_stat & SD_OC_NOW)
			status[8] |= 0x02;
		if (cr->ocp_stat & SD_OC_EVER)
			status[8] |= 0x01;
	}

#ifdef SUPPORT_SD_LOCK
	/* SD Lock/Unlock */
	if (card && SD_ID == card->id) {
		struct sd_info *sd_card = card->extra;
		status[0x17] = 0x80;
		if (sd_card->sd_erase_status)
			status[0x17] |= 0x01;	/* Under erasing */
		if (sd_card->sd_lock_status & SD_LOCKED) {
			status[0x17] |= 0x02;	/* Locked */
			status[0x07] |= 0x40;	/* Read protected */
		}
		if (sd_card->sd_lock_status & SD_PWD_EXIST)
			status[0x17] |= 0x04;	/* Contain PWD */
	} else {
		status[0x17] = 0x00;
	}

	rts_trace(TRACE_SCSI, "status[0x17] = 0x%x\n", status[0x17]);
#endif

	/* Function 0 Support Magic Gate, CPRM and PhyRegister R/W */
	status[0x18] = 0x8A;

	/* Function 2 Support OC LUN status & WP LUN status */
	status[0x1A] = 0x28;

	/* Function 7 */
#ifdef SUPPORT_SD_LOCK
	status[0x1F] = 0x01;
#endif

	buf_len = min(scsi_bufflen(srb), (unsigned int)sizeof(status));
	/* rtsx_stor_set_xfer_buf(status, buf_len, srb); */
	scsi_stor_set_xfer_buf(status, buf_len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - buf_len);


	return TRANSPORT_GOOD;
}

int scsi_common_rd_mem(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned short addr, len, i;
	int retval;
	u8 *buf;
#if 0 /* edwin: comment out */
	cr_disable_aspm(cr);

	if (cr->ss_en && (rtsx_get_stat(cr) == RTSX_STAT_SS)) {
		rtsx_exit_ss(cr);
		msleep(100);
	}
	cr_set_stat(cr, RTSX_STAT_RUN);
#endif
	addr = ((u16)srb->cmnd[2] << 8) | srb->cmnd[3];
	len = ((u16)srb->cmnd[4] << 8) | srb->cmnd[5];

	if (addr < 0xC000) {
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	buf = (u8 *)vmalloc(len);
	if (!buf)
		TRACE_RET(cr, TRANSPORT_ERROR);
#if 0 /* edwin: comment out */
	retval = cr_force_power_on(cr, SSC_PDCTL);
	if (retval != STATUS_SUCCESS) {
		vfree(buf);
		set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_WRITE_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
#endif
	for (i = 0; i < len; i++) {
		retval = cr_rd_reg(cr, addr + i, buf + i);
		if (retval != STATUS_SUCCESS) {
			vfree(buf);
			set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
	}

	len = (unsigned short)min(scsi_bufflen(srb), (unsigned int)len);
	scsi_stor_set_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);

	vfree(buf);

	return TRANSPORT_GOOD;
}

int scsi_common_wr_mem(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned short addr, len, i;
	int retval;
	u8 *buf;

#if 0 /* edwin: comment out */
	cr_disable_aspm(cr);

	if (chip->ss_en && (rtsx_get_stat(cr) == RTSX_STAT_SS)) {
		rtsx_exit_ss(cr);
		msleep(100);
	}
	cr_set_stat(cr, RTSX_STAT_RUN);
#endif
	addr = ((u16)srb->cmnd[2] << 8) | srb->cmnd[3];
	len = ((u16)srb->cmnd[4] << 8) | srb->cmnd[5];

	if (addr < 0xC000) {
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	len = (unsigned short)min(scsi_bufflen(srb), (unsigned int)len);
	buf = (u8 *)vmalloc(len);
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	scsi_stor_get_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);
#if 0 /* edwin: comment out */
	retval = cr_force_power_on(cr, SSC_PDCTL);
	if (retval != STATUS_SUCCESS) {
		vfree(buf);
		set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_WRITE_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
#endif
	for (i = 0; i < len; i++) {
		retval = cr_wr_reg(cr, addr + i, 0xFF, buf[i]);
		if (retval != STATUS_SUCCESS) {
			vfree(buf);
			set_sense_type(cr, SCSI_LUN(srb),
						SENSE_TYPE_MEDIA_WRITE_ERR);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
	}

	vfree(buf);

	return TRANSPORT_GOOD;
}

int scsi_common_rd_sts(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 status[16];
	int buf_len;
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);

	/* VID */
	status[0] = (u8)(cr->vendor_id >> 8);
	status[1] = (u8)(cr->vendor_id);

	/* PID */
	status[2] = (u8)(cr->product_id >> 8);
	status[3] = (u8)(cr->product_id);

	/* gbLUN */
	status[4] = (u8)lun;

	/* Lun Card Number */
	if (NULL != card && SD_ID == card->id)
		status[5] = 2; /* SD Card */
	else if (NULL != card && MS_ID == card->id)
		status[5] = 3; /* MS Card */
	else
		status[5] = 7; /* Multi */

	/* Total LUNs */
	status[6] = 1;

	/* IC Version */
	status[7] = (u8)(cr->product_id);
	status[8] = cr->ic_version;

	/* Physical Exist
	 * 2008-05-12
	 * For MAC style, we hope that disk icon won't appear
	 * when a dummy card inserted, so we set physical exist
	 * bit equal to logical exist bit
	 * 2010-01-22
	 * MAC style depends on logical exist bit now, so we change it back */
	if (check_card_flag(card, EXIST) || check_card_flag(card, EJECTED))
		status[9] = 1;
	else
		status[9] = 0;

	/* Multi Flag */
	status[10] = 1;

	/* LUN Valid Map */
#if 0 /* edwin: changed here, tobe implemented in the future */
	if (CHECK_LUN_MODE(chip, DEFAULT_SINGLE))
		status[11] = SD_CARD | MS_CARD;
	else if (CHECK_LUN_MODE(chip, SD_LUN))
		status[11] = SD_CARD;
	else if (CHECK_LUN_MODE(chip, MS_LUN))
		status[11] = MS_CARD;
	else
		status[11] = 0;
#else
	status[11] = SD_CARD | MS_CARD;
#endif
	/* Logical Exist */
	if (card && check_card_flag(card, READY))
		status[12] = 1;
	else
		status[12] = 0;

	/* Detailed Type */
	if (NULL != card && SD_ID == card->id) {
		struct sd_info *sd_card = card->extra;

		status[13] = 0x20;
		if (CHK_SD(sd_card)) {
			if (CHK_SD_HCXC(sd_card))
				status[13] |= 0x04;	/* Hi capacity SD */
			if (CHK_SD_HS(sd_card))
				status[13] |= 0x02;	/* Hi speed SD */
		} else {
			status[13] |= 0x08;		/* MMC card */
			if (CHK_MMC_52M(sd_card))
				status[13] |= 0x02;	/* Hi speed */
			if (CHK_MMC_SECTOR_MODE(sd_card))
				status[13] |= 0x04;	/* Hi capacity */
		}
	} else if (NULL != card && MS_ID == card->id) {
		struct ms_info *ms_card = card->extra;

		if (CHK_MSPRO(ms_card)) {
			status[13] = 0x38;		/* MS Pro */
			if (CHK_HG8BIT(ms_card))
				status[13] |= 0x04;	/* HG */
#ifdef SUPPORT_MSXC
			if (CHK_MSXC(ms_card))
				status[13] |= 0x01;	/* MSXC */
#endif
		} else {
			status[13] = 0x30;
		}
	} else {
		status[13] = 0x70;
	}

	/* Support OC, auto delink, vendor r/w, get bus width */
	status[14] = 0x78;
	/* Support detailed type 2 information, PP Info */
	status[15] = 0x82;

	buf_len = min(scsi_bufflen(srb), (unsigned int)sizeof(status));
	scsi_stor_set_xfer_buf(status, buf_len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - buf_len);

	return TRANSPORT_GOOD;
}

/* TODO: compoment about scsi should be moved to SCSI layer */

int scsi_common_rd_phy_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned short addr, len, i;
	int retval;
	u8 *buf;
	u16 val;

	addr = ((u16)srb->cmnd[4] << 8) | srb->cmnd[5];
	len = ((u16)srb->cmnd[6] << 8) | srb->cmnd[7];
	if (len % 2)
		len -= len % 2;
	if (len) {
		buf = (u8 *)vmalloc(len);
		if (!buf)
			TRACE_RET(cr, TRANSPORT_ERROR);

		for (i = 0; i < len / 2; i++) {
			retval = cr_rd_phy16(cr, addr + i, &val);
			if (retval != STATUS_SUCCESS) {
				vfree(buf);
				set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
				TRACE_RET(cr, TRANSPORT_FAILED);
			}

			buf[2*i] = (u8)(val >> 8);
			buf[2*i+1] = (u8)val;
		}

		len = (unsigned short)min(scsi_bufflen(srb),
							(unsigned int)len);
		scsi_stor_set_xfer_buf(buf, len, srb);
		scsi_set_resid(srb, scsi_bufflen(srb) - len);

		vfree(buf);
	}

	return TRANSPORT_GOOD;
}

int scsi_common_wr_phy_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned short addr, len, i;
	int retval;
	u8 *buf;
	u16 val;

	addr = ((u16)srb->cmnd[4] << 8) | srb->cmnd[5];
	len = ((u16)srb->cmnd[6] << 8) | srb->cmnd[7];
	if (len % 2)
		len -= len % 2;
	if (len) {
		len = (unsigned short)min(scsi_bufflen(srb), (unsigned int)len);
		buf = (u8 *)vmalloc(len);
		if (buf == NULL)
			TRACE_RET(cr, TRANSPORT_ERROR);
		scsi_stor_get_xfer_buf(buf, len, srb);
		scsi_set_resid(srb, scsi_bufflen(srb) - len);

		for (i = 0; i < len / 2; i++) {
			val = ((u16)buf[2*i] << 8) | buf[2*i+1];
			retval = cr_wr_phy16(cr, addr + i, val);
			if (retval != STATUS_SUCCESS) {
				vfree(buf);
				set_sense_type(cr, SCSI_LUN(srb),
						SENSE_TYPE_MEDIA_WRITE_ERR);
				TRACE_RET(cr, TRANSPORT_FAILED);
			}
		}

		vfree(buf);
	}

	return TRANSPORT_GOOD;
}

int scsi_common_suit_rw_mem_cmd_buf(struct card_reader *cr,
						struct scsi_cmnd *srb)
{
	int retval =  STATUS_SUCCESS;
	u8 lun = SCSI_LUN(srb);
	u8 cmd_type, mask, value, idx;
	u16 addr;

	switch (srb->cmnd[3]) {
	case INIT_BATCHCMD:
		cr_init_cmd(cr);
		break;

	case ADD_BATCHCMD:
		cmd_type = srb->cmnd[4];
		if (cmd_type > 2) {
			set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		addr = (srb->cmnd[5] << 8) | srb->cmnd[6];
		mask = srb->cmnd[7];
		value = srb->cmnd[8];
		cr_add_cmd(cr, cmd_type, addr, mask, value);
		break;

	case SEND_BATCHCMD:
		retval = cr_send_cmd(cr, 0, 1000);
		break;

	case GET_BATCHRSP:
		idx = srb->cmnd[4];
		value = *(cr_get_cmd_data(cr) + idx);
		if (scsi_bufflen(srb) < 1) {
			set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		scsi_stor_set_xfer_buf(&value, 1, srb);
		scsi_set_resid(srb, 0);
		break;

	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (retval != STATUS_SUCCESS) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_WRITE_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return TRANSPORT_GOOD;
}

int scsi_common_dma_rb(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int retval;
	u8 lun = SCSI_LUN(srb);
	u16 len;

	len = ((u16)(srb->cmnd[4]) << 8) | srb->cmnd[5];
	len = min(len, (u16)scsi_bufflen(srb));

	if (srb->sc_data_direction == DMA_FROM_DEVICE)
		rts_trace(TRACE_SCSI, "Read from device\n");
	else
		rts_trace(TRACE_SCSI, "Write to device\n");

	retval = cr_transfer_data(cr, 0, scsi_sglist(srb), len,
			scsi_sg_count(srb), srb->sc_data_direction, 1000);
	if (retval < 0) {
		if (srb->sc_data_direction == DMA_FROM_DEVICE)
			set_sense_type(cr, lun,
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
		else
			set_sense_type(cr, lun, SENSE_TYPE_MEDIA_WRITE_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;
}

int scsi_common_rd_cfg_byte(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int retval;
	u8 func, cfg_mode;
	u16 addr, len;
	u8 *buf;

	/* bit [2-1]: func number
	 * bit [3]: 1 -> cfg mode, 0 -> back door mode
	 * bit [7-4]: RFU
	 */
	func = srb->cmnd[3] & 0x07;
	if (srb->cmnd[3] & 0x08)
		cfg_mode = 1;
	else
		cfg_mode = 0;

	addr = ((u16)(srb->cmnd[4]) << 8) | srb->cmnd[5];
	len = ((u16)(srb->cmnd[6]) << 8) | srb->cmnd[7];

	rts_trace(TRACE_SCSI, "%s: func = %d, addr = 0x%x, len = %d\n",
						__func__, func, addr, len);

	if (func > 0) {
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	buf = (u8 *)vmalloc(len);
	if (!buf)
		TRACE_RET(cr, TRANSPORT_ERROR);

	if (cfg_mode) {
		int i;
		for (i = 0; i < len; i++) {
			retval = cr_read_config8(cr, addr + i, &buf[i]);
			if (retval < 0) {
				set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
				vfree(buf);
				TRACE_RET(cr, TRANSPORT_FAILED);
			}
		}
	} else {
		retval = cr_rd_cfg_seq(cr, addr, buf, len);
		if (retval != STATUS_SUCCESS) {
			set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
			vfree(buf);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
	}

	len = (u16)min(scsi_bufflen(srb), (unsigned int)len);
	scsi_stor_set_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);

	vfree(buf);

	return TRANSPORT_GOOD;
}

int scsi_common_wr_cfg_byte(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int retval;
	u8 func, cfg_mode;
	u16 addr, len;
	u8 *buf;

	/* bit [2-1]: func number
	 * bit [3]: 1 -> cfg mode, 0 -> back door mode
	 * bit [7-4]: RFU
	 */
	func = srb->cmnd[3] & 0x07;
	if (srb->cmnd[3] & 0x08)
		cfg_mode = 1;
	else
		cfg_mode = 0;
	addr = ((u16)(srb->cmnd[4]) << 8) | srb->cmnd[5];
	len = ((u16)(srb->cmnd[6]) << 8) | srb->cmnd[7];

	rts_trace(TRACE_SCSI, "%s: func = %d, addr = 0x%x\n", __func__, func, addr);

	if (func > 0) {
		set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	len = (unsigned short)min(scsi_bufflen(srb), (unsigned int)len);
	buf = (u8 *)vmalloc(len);
	if (!buf)
		TRACE_RET(cr, TRANSPORT_ERROR);
	scsi_stor_get_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);

	if (cfg_mode) {
		int i;
		for (i = 0; i < len; i++) {
			retval = cr_write_config8(cr, addr + i, buf[i]);
			if (retval < 0) {
				set_sense_type(cr, SCSI_LUN(srb),
						SENSE_TYPE_MEDIA_WRITE_ERR);
				vfree(buf);
				TRACE_RET(cr, TRANSPORT_FAILED);
			}
		}
	} else {
		retval = cr_wr_cfg_seq(cr, addr, buf, len);
		if (retval != STATUS_SUCCESS) {
			set_sense_type(cr, SCSI_LUN(srb),
						SENSE_TYPE_MEDIA_WRITE_ERR);
			vfree(buf);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
	}

	vfree(buf);

	return TRANSPORT_GOOD;
}

int scsi_common_rd_host_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 addr, buf[4];
	u32 val;
	unsigned int len;

	addr = srb->cmnd[4];
	val = cr_ioread32(cr, addr);
	rts_trace(TRACE_SCSI, "Host register (0x%x): 0x%x\n", addr, val);

	buf[0] = (u8)(val >> 24);
	buf[1] = (u8)(val >> 16);
	buf[2] = (u8)(val >> 8);
	buf[3] = (u8)val;

	len = min(scsi_bufflen(srb), (unsigned int)4);
	scsi_stor_set_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);

	return TRANSPORT_GOOD;
}

int scsi_common_wr_host_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 addr, buf[4];
	u32 val;
	unsigned int len;

	addr = srb->cmnd[4];
	len = min(scsi_bufflen(srb), (unsigned int)4);
	scsi_stor_get_xfer_buf(buf, len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - len);
	val = ((u32)buf[0] << 24) | ((u32)buf[1] << 16) |
						((u32)buf[2] << 8) | buf[3];
	cr_iowrite32(cr, addr, val);

	return TRANSPORT_GOOD;
}

int scsi_common_set_chip_mode(struct card_reader *cr,
						struct scsi_cmnd *srb)
{
	set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
	TRACE_RET(cr, TRANSPORT_FAILED);
}

int scsi_control_thread(void *__scsi);
static int scsi_get_var(struct card_reader *cr, struct scsi_cmnd *srb);
static int scsi_set_var(struct card_reader *cr, struct scsi_cmnd *srb);

/***********************************************************************
 * Scatter-gather transfer buffer access routines
 ***********************************************************************/

/* Copy a buffer of length buflen to/from the srb's transfer buffer.
 * Update the **sgptr and *offset variables so that the next copy will
 * pick up from where this one left off.
 */

void scsi_show_command(struct scsi_cmnd *srb)
{
	int i, unknown_cmd = 1;
	char *name = "UNKNOWN_CMD";

#define cr_set_cmd_name(cmd)			\
	do {					\
		if (srb->cmnd[0] == cmd) {	\
			unknown_cmd = 0;	\
			name = #cmd;		\
			goto show_cmd;		\
		}				\
	} while (0)
	cr_set_cmd_name(TEST_UNIT_READY);
	cr_set_cmd_name(READ_FORMAT_CAPACITIES); /* 0x23 */
	cr_set_cmd_name(REZERO_UNIT);
	cr_set_cmd_name(REQUEST_SENSE);
	cr_set_cmd_name(FORMAT_UNIT);
	cr_set_cmd_name(READ_BLOCK_LIMITS);
	cr_set_cmd_name(REASSIGN_BLOCKS); /* 0x07 */
	cr_set_cmd_name(READ_6);
	cr_set_cmd_name(WRITE_6);
	cr_set_cmd_name(SEEK_6);
	cr_set_cmd_name(READ_REVERSE);
	cr_set_cmd_name(WRITE_FILEMARKS);
	cr_set_cmd_name(SPACE);
	cr_set_cmd_name(INQUIRY);
	cr_set_cmd_name(RECOVER_BUFFERED_DATA);
	cr_set_cmd_name(MODE_SELECT);
	cr_set_cmd_name(RESERVE);
	cr_set_cmd_name(RELEASE);
	cr_set_cmd_name(COPY);
	cr_set_cmd_name(ERASE);
	cr_set_cmd_name(MODE_SENSE);
	cr_set_cmd_name(START_STOP);
	cr_set_cmd_name(RECEIVE_DIAGNOSTIC);
	cr_set_cmd_name(SEND_DIAGNOSTIC);
	cr_set_cmd_name(ALLOW_MEDIUM_REMOVAL);
	cr_set_cmd_name(SET_WINDOW);
	cr_set_cmd_name(READ_CAPACITY);
	cr_set_cmd_name(READ_10);
	cr_set_cmd_name(WRITE_10);
	cr_set_cmd_name(SEEK_10);
	cr_set_cmd_name(WRITE_VERIFY);
	cr_set_cmd_name(VERIFY);
	cr_set_cmd_name(SEARCH_HIGH);
	cr_set_cmd_name(SEARCH_EQUAL);
	cr_set_cmd_name(SEARCH_LOW);
	cr_set_cmd_name(SET_LIMITS);
	cr_set_cmd_name(READ_POSITION);
	cr_set_cmd_name(SYNCHRONIZE_CACHE);
	cr_set_cmd_name(LOCK_UNLOCK_CACHE);
	cr_set_cmd_name(READ_DEFECT_DATA);
	cr_set_cmd_name(MEDIUM_SCAN);
	cr_set_cmd_name(COMPARE);
	cr_set_cmd_name(COPY_VERIFY);
	cr_set_cmd_name(WRITE_BUFFER);
	cr_set_cmd_name(READ_BUFFER);
	cr_set_cmd_name(UPDATE_BLOCK);
	cr_set_cmd_name(READ_LONG);
	cr_set_cmd_name(WRITE_LONG);
	cr_set_cmd_name(CHANGE_DEFINITION);
	cr_set_cmd_name(WRITE_SAME);
	cr_set_cmd_name(GPCMD_READ_SUBCHANNEL);
	cr_set_cmd_name(READ_TOC);
	cr_set_cmd_name(GPCMD_READ_HEADER);
	cr_set_cmd_name(GPCMD_PLAY_AUDIO_10);
	cr_set_cmd_name(GPCMD_PLAY_AUDIO_MSF);
	cr_set_cmd_name(GPCMD_GET_EVENT_STATUS_NOTIFICATION);
	cr_set_cmd_name(GPCMD_PAUSE_RESUME);
	cr_set_cmd_name(LOG_SELECT);
	cr_set_cmd_name(LOG_SENSE);
	cr_set_cmd_name(GPCMD_STOP_PLAY_SCAN);
	cr_set_cmd_name(GPCMD_READ_DISC_INFO);
	cr_set_cmd_name(GPCMD_READ_TRACK_RZONE_INFO);
	cr_set_cmd_name(GPCMD_RESERVE_RZONE_TRACK);
	cr_set_cmd_name(GPCMD_SEND_OPC);
	cr_set_cmd_name(MODE_SELECT_10);
	cr_set_cmd_name(GPCMD_REPAIR_RZONE_TRACK);
	cr_set_cmd_name(READ_BUFFER_CAPACITY); /* 0x5C */
	cr_set_cmd_name(READ_MASTER_CUE); /* 0x59 */
	cr_set_cmd_name(MODE_SENSE_10);
	cr_set_cmd_name(GPCMD_CLOSE_TRACK);
	cr_set_cmd_name(SEND_CUE_SHEET); /* 0x5D */
	cr_set_cmd_name(GPCMD_BLANK);
	cr_set_cmd_name(REPORT_LUNS);
	cr_set_cmd_name(MOVE_MEDIUM);
	cr_set_cmd_name(READ_12);
	cr_set_cmd_name(WRITE_12);
	cr_set_cmd_name(WRITE_VERIFY_12);
	cr_set_cmd_name(SEARCH_HIGH_12);
	cr_set_cmd_name(SEARCH_EQUAL_12);
	cr_set_cmd_name(SEARCH_LOW_12);
	cr_set_cmd_name(SEND_VOLUME_TAG);
	cr_set_cmd_name(READ_ELEMENT_STATUS);
	cr_set_cmd_name(GPCMD_READ_CD_MSF);
	cr_set_cmd_name(GPCMD_SCAN);
	cr_set_cmd_name(GPCMD_SET_SPEED);
	cr_set_cmd_name(GPCMD_MECHANISM_STATUS);
	cr_set_cmd_name(GPCMD_READ_CD);
	cr_set_cmd_name(WRITE_CONTINUE); /* 0xE1 */
	cr_set_cmd_name(WRITE_LONG_2);
	cr_set_cmd_name(VENDOR_CMND);
#undef cr_set_cmd_name

show_cmd:
	if (srb->cmnd[0] != TEST_UNIT_READY)
		rts_trace(TRACE_SCSI, "Command %s (%d bytes)\n", name, srb->cmd_len);
	else
		rts_trace(TRACE_POLL_THRD, "Command %s (%d bytes)\n", name, srb->cmd_len);
	if (unknown_cmd) {
		for (i = 0; i < srb->cmd_len && i < 16; i++)
			rts_trace(TRACE_SCSI, "0x%02x, ", srb->cmnd[i]);
		rts_trace(TRACE_SCSI, "\n");
	}
}

void set_sense_type(struct card_reader *cr, u8 lun, int sense_type)
{
	switch (sense_type) {
	case SENSE_TYPE_MEDIA_CHANGE:
		set_sense_data(cr, lun, CUR_ERR, 0x06, 0, 0x28, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_NOT_PRESENT:
		set_sense_data(cr, lun, CUR_ERR, 0x02, 0, 0x3A, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_LBA_OVER_RANGE:
		set_sense_data(cr, lun, CUR_ERR, 0x05, 0, 0x21, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT:
		set_sense_data(cr, lun, CUR_ERR, 0x05, 0, 0x25, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_WRITE_PROTECT:
		set_sense_data(cr, lun, CUR_ERR, 0x07, 0, 0x27, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR:
		set_sense_data(cr, lun, CUR_ERR, 0x03, 0, 0x11, 0, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_WRITE_ERR:
		set_sense_data(cr, lun, CUR_ERR, 0x03, 0, 0x0C, 0x02, 0, 0);
		break;

	case SENSE_TYPE_MEDIA_INVALID_CMD_FIELD:
		set_sense_data(cr, lun, CUR_ERR, ILGAL_REQ, 0,
			       ASC_INVLD_CDB, ASCQ_INVLD_CDB, CDB_ILLEGAL, 1);
		break;

	case SENSE_TYPE_FORMAT_IN_PROGRESS:
		set_sense_data(cr, lun, CUR_ERR, 0x02, 0, 0x04, 0x04, 0, 0);
		break;

	case SENSE_TYPE_FORMAT_CMD_FAILED:
		set_sense_data(cr, lun, CUR_ERR, 0x03, 0, 0x31, 0x01, 0, 0);
		break;

	case SENSE_TYPE_OVERLAPPED_CMD_ATTEMPTED:
		set_sense_data(cr, lun, CUR_ERR, 0x0B, 0, 0x4E, 0x00, 0, 0);
		break;

#ifdef SUPPORT_MAGIC_GATE
	case SENSE_TYPE_MG_KEY_FAIL_NOT_ESTAB:
		set_sense_data(cr, lun, CUR_ERR, 0x05, 0, 0x6F, 0x02, 0, 0);
		break;

	case SENSE_TYPE_MG_KEY_FAIL_NOT_AUTHEN:
		set_sense_data(cr, lun, CUR_ERR, 0x05, 0, 0x6F, 0x00, 0, 0);
		break;

	case SENSE_TYPE_MG_INCOMPATIBLE_MEDIUM:
		set_sense_data(cr, lun, CUR_ERR, 0x02, 0, 0x30, 0x00, 0, 0);
		break;

	case SENSE_TYPE_MG_WRITE_ERR:
		set_sense_data(cr, lun, CUR_ERR, 0x03, 0, 0x0C, 0x00, 0, 0);
		break;
#endif

#ifdef SUPPORT_SD_LOCK
	case SENSE_TYPE_MEDIA_READ_FORBIDDEN:
		set_sense_data(cr, lun, CUR_ERR, 0x07, 0, 0x11, 0x13, 0, 0);
		break;
#endif

	case SENSE_TYPE_NO_SENSE:
	default:
		set_sense_data(cr, lun, CUR_ERR, 0, 0, 0, 0, 0, 0);
		break;
	}
}
EXPORT_SYMBOL_GPL(set_sense_type);

void set_sense_data(struct card_reader *cr, u8 lun, u8 err_code,
		    u8 sense_key, u32 info, u8 asc, u8 ascq, u8 sns_key_info0,
		    u16 sns_key_info1)
{
	struct sense_data_t *sense = &(cr->scsi->sense_buffer[lun]);

	sense->err_code = err_code;
	sense->sense_key = sense_key;
	sense->info[0] = (u8) (info >> 24);
	sense->info[1] = (u8) (info >> 16);
	sense->info[2] = (u8) (info >> 8);
	sense->info[3] = (u8) info;

	sense->ad_sense_len = sizeof(struct sense_data_t) - 8;
	sense->asc = asc;
	sense->ascq = ascq;
	if (sns_key_info0 != 0) {
		sense->sns_key_info[0] = SKSV | sns_key_info0;
		sense->sns_key_info[1] = (sns_key_info1 & 0xf0) >> 8;
		sense->sns_key_info[2] = sns_key_info1 & 0x0f;
	}
}

static int test_unit_ready_new(struct card_reader *cr, struct scsi_cmnd *srb)
{
	struct list_head *pos;
	u8 lun = SCSI_LUN(srb);
	struct card_manager *cm = cr->cm;
	struct card *c = NULL;
	int ret = 0;

	list_for_each(pos, &cm->card_list) {
		c = list_entry(pos, struct card, link);
		if (c->lun == lun) {
			ret = check_card_readiness(c, OPTIONS_TEST_UNIT_READY,
						srb->sc_data_direction);
			if (ret)
				continue;
			else
				return TRANSPORT_GOOD;
		}
	}

	set_sense_type(cr, lun, ret);
	return TRANSPORT_FAILED;
}


unsigned char formatter_inquiry_str[20] = {
	'M', 'E', 'M', 'O', 'R', 'Y', 'S', 'T', 'I', 'C', 'K',
#ifdef SUPPORT_MAGIC_GATE
	'-', 'M', 'G', /* Byte[47:49] */
#else
	0x20, 0x20, 0x20,  /* Byte[47:49] */
#endif

#ifdef SUPPORT_MAGIC_GATE
	0x0B,  /* Byte[50]: MG, MS, MSPro, MSXC */
#else
	0x09,  /* Byte[50]: MS, MSPro, MSXC */
#endif
	0x00,  /* Byte[51]: Category Specific Commands */
	0x00,  /* Byte[52]: Access Control and feature */
	0x20, 0x20, 0x20, /* Byte[53:55] */
};

/* For WHQL1.6 */
/* Caching mode page, page code 0x08 */
/* Ref to spec sbc2r16.pdf, 6.3.3, page 126 */
static unsigned char inquiry_caching_mode[] = {
	0x00|0x08,	/* PS | Page Code */
	0x12,		/* Page Length */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
}; /* Caching Mode Page */

/* For WHQL1.6 */
/* Control mode page, Page Code 0x0A */
/* Ref to spec spc3r23.pdf, section 7.4.6, page 310 */
static unsigned char inquiry_control_mode[] = {
	0x00 | 0x0A,		/* PS | Page Code */
	0x0A,		/* Page Length */
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00
}; /* Control Mode Page */


/* For WHQL1.6 */
/* Information Exceptions mode page, Page Code 0x1C */
/* Ref to spec spc3r23.pdf, section 7.4.11, page 319 */
static unsigned char inquiry_info_exception_mode[] = {
	0x00|0x1C,	/* PS | Page Code */
	0x0A,		/* Page Length */
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00
}; /* Information Exceptions Mode Page */

static unsigned char inquiry_standard_header[] = {
	QULIFIRE|DRCT_ACCESS_DEV,

	/* RMB_DISC, */
	RMB_DISC|0x0D,

	/* For WHQL 1.6 */
	/* The device must return a valid VERSION field of 0x4,0x5 or 0x6 */
	/* Expectation: VERSION is 0x4(SPC-2) or later for non-SCSI bus type and */
	/* VERSION is 0x3(SPC) or later for SCSI bus type */
	/* Ref to SPC4, 6.4.2, page 310 */
	/* 0x04, */
	0x06,	/* VERSION for USB3.0 */

	/* For WHQL 1.6 */
	/* Checking that RESPONSE DATA FORMAT == 2 */
	/* Ref to SPC4, 6.4.2, page 310 */
	0x02,

	0x5B,	/* (n-4) */

	0x02,	/* Byte[5]: This byte is set as 0x01 initially, but bit 0 has been defined as PROTECT in SPC-4. */

	0,
	REL_ADR|WBUS_32|WBUS_16|SYNC|LINKED|CMD_QUE|SFT_RE
};

/* For WHQL1.6 */
/* Unit Serial Number VPD page, page code 0x80 */
/* Ref to SPC4, 7.8.14, page 653 */
static unsigned char inquiry_unit_serial_num[] = {
	QULIFIRE|DRCT_ACCESS_DEV,
	0x80,/* PAGE CODE */
	0x00,/* Reserved */
	0xFB,/* Page Length(n-3) */
	'2', '0', '1', '2', '0', '6', '2', '9', '1', '4', '3', '4', '5', '3', '0', '0',
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20
}; /* Unit Serial Number VPD page   */

static unsigned char inquiry_supported_vpd_pages[] = {
	QULIFIRE|DRCT_ACCESS_DEV,
	0x00,
	0x00,	/* (n-3), MSB */
	0x03,	/* (n-3), LSB */
	0x00,
	0x80,
	0x83
};

/* For WHQL1.6 */
/* Ref to SPC3r23.pdf, 7.6.3, page 350 */
/* Device Identification VPD page, page code 0x83 */
static unsigned char inquiry_device_identification[] = {
	QULIFIRE|DRCT_ACCESS_DEV,
	0x83,/* PAGE CODE */
	0x00,/* Page Length(n-3),MSB */
	0x34,/* Page Length(n-3), LSB */
	/* Identification Descriptor List */

	/* Identification Descriptor Second(T10 Vendor ID based) */
	/* Protocol Identifier | Code Set */
	0x00 | 0x02,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x01,
	0x00,/* Reserved */
	0x18,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x72, 0x65, 0x61, 0x6c, 0x74, 0x65, 0x6b, 0x20,
	'2', '0', '1', '2', '0', '6', '2', '9', '1', '4', '3', '4', '5', '3', '0', '0',

	/* Identification Descriptor Second(NAA) */
	/* Protocol Identifier | Code Set */
	0x00 | 0x01,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x02,
	0x00,/* Reserved */
	0x08,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x00, 0xE0, 0x4C, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,

	/* SCSI name string */
	/* Protocol Identifier | Code Set */
	0x00 | 0x03,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x08,
	0x00,/* Reserved */
	0x08,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x72, 0x65, 0x61, 0x6c, 0x74, 0x65, 0x6b, 0x20,
}; /* Device Identification VPD page   */

/* For WHQL1.6 support Gamma/EMID*/
/* Ref to SPC3r23.pdf, 7.6.3, page 350 */
/* Device Identification VPD page, page code 0x83 */
static unsigned char inquiry_device_identification_ge[] = {
	QULIFIRE|DRCT_ACCESS_DEV,

	0x83,/* PAGE CODE */
	0x00,/* Page Length(n-3),MSB */
	0x34 + 0x0D,/* Page Length(n-3), LSB */
	/* Identification Descriptor List */

	/* LEN = 0x0D, GAMMA device identification */
	0xF0 | 0x02,			/* Protocol ID | Code set */
	/* PIV | Rsv | Association | Design type */
	0x00 | 0x00 | 0x10 | 0x00,
	0x00,				/* Reserved */
	0x09,				/* Designator length */
	/* GAMMA-CPS: 0x47, 0x41, 0x4D, 0x4D, 0x41, 0x2D, 0x43, 0x50, 0x53 */
	0x47, 0x41, 0x4D, 0x4D, 0x41, 0x2D, 0x43, 0x50, 0x53,

	/* Identification Descriptor Second(T10 Vendor ID based) */
	/* Protocol Identifier | Code Set */
	0xF0 | 0x02,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x01,
	0x00,/* Reserved */
	0x18,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x72, 0x65, 0x61, 0x6c, 0x74, 0x65, 0x6b, 0x20,
	'2', '0', '1', '2', '0', '6', '2', '9', '1', '4', '3', '4', '5', '3', '0', '0',

	/* Identification Descriptor Second(NAA) */
	/* Protocol Identifier | Code Set */
	0xF0 | 0x01,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x02,
	0x00,/* Reserved */
	0x08,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x00, 0xE0, 0x4C, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,

	/* SCSI name string */
	/* Protocol Identifier | Code Set */
	0xF0 | 0x02,
	/* PIV |Association | Identifier Type */
	0x00 | 0x00 | 0x08,
	0x00,/* Reserved */
	0x08,/* Identifier Length is set to 08h */
	/* Byte0~Byte7, T10 Vendor Identification */
	0x72, 0x65, 0x61, 0x6C, 0x74, 0x65, 0x6B, 0x20,

}; /* Device Identification VPD page   */

static unsigned char inquiry_extended[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, /* CLOCKING/QAS/IUS */
	0, /* RESERVED */
	0x04, 0x60,	/* SPC-4 */
	0x04, 0xC0,	/* SBC-3 */
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


#define INQUIRY_DEFAULT  "Generic-xD/SD/M.S.      1.00 "
#define INQUIRY_SDMS     "Generic-SD/MemoryStick  1.00 "
#define INQUIRY_SDMMC    "Generic-SD/MMC          1.00 "
#define INQUIRY_SD       "Generic-SD              1.00 "
#define INQUIRY_MS       "Generic-MemoryStick     1.00 "

static int inquiry(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned int lun = SCSI_LUN(srb);
	const char *inquiry_string;
	unsigned char sendbytes;
	unsigned char *buf, *ptr, *vendor_str;
	int pro_formatter_flag = 0;
	enum CARD_ID id = cm_get_id_by_lun(cr->cm, lun);
	struct cr_option *option = &(cr->option);
	u8 page_code, evpd;
	u8 card_spt_map = cr->cc_hw.card_spt_map;
	u32 card_lun_map = cr->cc_hw.card_lun_map;
	int finish_inquiry = 0;
	struct card* card = NULL;
	struct sd_info* sd_card = NULL;
	card = cm_get_card(cr->cm, id);
	if (SD_ID == id && card)
		sd_card = card->extra;

	inquiry_string = INQUIRY_SDMS;

	evpd = srb->cmnd[1] & 0x01;
	page_code = srb->cmnd[2];
	rts_trace(TRACE_SCSI, "EVPD: 0x%02x, PAGE CODE: 0x%02x\n", evpd, page_code);

	/* SPC4, 6.4.1, page 308
	 * If the page code field is not set to zero when the EVPD bit is set to zero,
	 * the command shall be terminated with CHECK CONDITION STATUS, with the
	 * sense key set to ILLEGAL REQUEST, and the addtional sense code set to
	 * INVALID FIELD IN CDB
	 */

	if ((0 == evpd) && (0 != page_code)) {
		set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (1 == evpd) {
		if ((page_code != 0x80) && (page_code != 0x83) && (page_code != 0x00)) {
			set_sense_type(cr, SCSI_LUN(srb),
					SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
	}

	buf = kzalloc(scsi_bufflen(srb), GFP_KERNEL);
	if (buf == NULL)
		return TRANSPORT_ERROR;/* TRACE_RET(cr, TRANSPORT_ERROR); */

#ifdef SUPPORT_MAGIC_GATE
	if ((option->mspro_formatter_enable)
		&& (((cr->lun_mod == SIG_LUN) && (card_spt_map & (1 << MS_ID)))
 || ((cr->lun_mod == MUL_LUN) && (card_spt_map & (1 << MS_ID))
			   && (lun == ((card_lun_map >> 4) & 0xF))
			   )
			)
		) {
#else
	if (option->mspro_formatter_enable) {
#endif
		if (INVAL_ID == id || (MS_ID == id))
			pro_formatter_flag = 1;
	}
	rts_trace(TRACE_SCSI, "mspro_formatter_enable: %d, lun_mode: %d, card_spt_map: 0x%x\n"
					  "card_lun_map: 0x%x, pro_formatter_flag: 0x%02x\n",
					  option->mspro_formatter_enable, cr->lun_mod, card_spt_map,
					  card_lun_map, pro_formatter_flag);
	switch (page_code) {
	case 0x83:
		if (sd_card && sd_card->ge.ge_card) {
			sendbytes = sizeof(inquiry_device_identification_ge);
			if (scsi_bufflen(srb) < sendbytes)
				sendbytes = (u8)scsi_bufflen(srb);
			memcpy(buf, inquiry_device_identification_ge, sendbytes);
		} else {
			sendbytes = sizeof(inquiry_device_identification);
			if (scsi_bufflen(srb) < sendbytes)
				sendbytes = (u8)scsi_bufflen(srb);
			memcpy(buf, inquiry_device_identification, sendbytes);
		}
		finish_inquiry = 1;
		break;

	case 0x80:
		if (scsi_bufflen(srb) < 0xFF)
			sendbytes = (unsigned char)(scsi_bufflen(srb));
		else
			sendbytes = 0xFF;
		memcpy(buf, inquiry_unit_serial_num, sendbytes);
		finish_inquiry = 1;
		break;

	case 0x00:
		if (evpd) {
			if (scsi_bufflen(srb) < 7)
				sendbytes = (unsigned char)(scsi_bufflen(srb));
			else
				sendbytes = 7;
			memcpy(buf, inquiry_supported_vpd_pages, sendbytes);
			finish_inquiry = 1;
		} else {
			rts_trace(TRACE_SCSI, "Standard inquiry flow\n");
		}
		break;

	default:
		rts_trace(TRACE_SCSI, "Standard inquiry flow\n");
	}

	if (finish_inquiry)
		goto finish;

	if (scsi_bufflen(srb) < 96)
		sendbytes = (unsigned char)(scsi_bufflen(srb));
	else
		sendbytes = 96;
	ptr = buf;

	/* Send byte 0~7 */
	if (sendbytes > 8) {
		memcpy(ptr, inquiry_standard_header, 8);
		ptr += 8;
	} else {
		memcpy(ptr, inquiry_standard_header, sendbytes);
		goto finish;
	}

	/* Send byte 8~35 */
	if (sendbytes > 36) {
		memcpy(ptr, inquiry_string, 28);
		ptr += 28;
	} else {
		memcpy(ptr, inquiry_string, sendbytes - 8);
		goto finish;
	}

	/* Send byte 36~55 */
	if (pro_formatter_flag)
		vendor_str = formatter_inquiry_str;
	else
		vendor_str = inquiry_extended;
	if (sendbytes > 56) {
		memcpy(ptr, vendor_str, 20);
		ptr += 20;
	} else {
		memcpy(ptr, vendor_str, sendbytes - 36);
		goto finish;
	}

	/* Send byte 56~95 */
	memcpy(ptr, inquiry_extended + 20, sendbytes - 56);

finish:
	/* scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb); */
	scsi_set_xfer_buf(buf, sendbytes, srb);
	scsi_set_resid(srb, 0);
	kfree(buf);

	return TRANSPORT_GOOD;
}

static int start_stop_unit(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card_manager *cm = cr->cm;
	struct card *card = cm_locate_card(cm, lun);

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	scsi_set_resid(srb, scsi_bufflen(srb));

	if (srb->cmnd[1] == 1)
		return TRANSPORT_GOOD;

	switch (srb->cmnd[0x4]) {
	case STOP_MEDIUM:
		/* Media disabled */
		return TRANSPORT_GOOD;

	case UNLOAD_MEDIUM:
		/* Media shall be unload */
		if (check_card_flag(card, READY))
			eject_card(card);
		return TRANSPORT_GOOD;

	case MAKE_MEDIUM_READY:
	case LOAD_MEDIUM:
		if (check_card_flag(card, READY)) {
			return TRANSPORT_GOOD;
		} else {
			set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}

		break;
	}

	TRACE_RET(cr, TRANSPORT_ERROR);
}

static int allow_medium_removal(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int prevent;

	prevent = srb->cmnd[4] & 0x1;

	scsi_set_resid(srb, 0);

	if (prevent) {
		set_sense_type(cr, SCSI_LUN(srb),
			       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return TRANSPORT_GOOD;
}

static void ms_mode_sense(struct card *card, u8 cmd,
			  u8 lun, u8 *buf, int buf_len)
{
	struct card_reader *cr = card->cm->cr;
	struct ms_info *ms_card = card->extra;
	int sys_info_offset;
	int data_size = buf_len;
	int support_format = 0;
	int i = 0;

	if (cmd == MODE_SENSE) {
		sys_info_offset = 8;
		if (data_size > 0x68)
			data_size = 0x68;
		buf[i++] = 0x67;	/* Mode Data Length */
	} else {
		sys_info_offset = 12;
		if (data_size > 0x6C)
			data_size = 0x6C;
		buf[i++] = 0x00;	/* Mode Data Length (MSB) */
		buf[i++] = 0x6A;	/* Mode Data Length (LSB) */
	}

	/* Medium Type Code */
	if (cm_check_card_flag(cr->cm, lun, READY)) {
		if (CHK_MSXC(ms_card)) {
			support_format = 1;
			buf[i++] = 0x40;
		} else if (CHK_MSPRO(ms_card)) {
			support_format = 1;
			buf[i++] = 0x20;
		} else {
			buf[i++] = 0x10;
		}

		/* WP */
		if (cm_check_card_flag(cr->cm, lun, WP))
			buf[i++] = 0x80;
		else
			buf[i++] = 0x00;
	} else {
		buf[i++] = 0x00;	/* MediaType */
		buf[i++] = 0x00;	/* WP */
	}

	buf[i++] = 0x00;	/* Reserved */

	if (cmd == MODE_SENSE_10) {
		buf[i++] = 0x00;	/* Reserved */
		buf[i++] = 0x00;	/* Block descriptor length(MSB) */
		buf[i++] = 0x00;	/* Block descriptor length(LSB) */

		/* The Following Data is the content of "Page 0x20" */
		if (data_size >= 9)
			buf[i++] = 0x20;	/* Page Code */
		if (data_size >= 10)
			buf[i++] = 0x62;	/* Page Length */
		if (data_size >= 11)
			buf[i++] = 0x00;	/* No Access Control */
		if (data_size >= 12) {
			if (support_format)
				buf[i++] = 0xC0;	/* SF, SGM */
			else
				buf[i++] = 0x00;
		}
	} else {
		/* The Following Data is the content of "Page 0x20" */
		if (data_size >= 5)
			buf[i++] = 0x20;	/* Page Code */
		if (data_size >= 6)
			buf[i++] = 0x62;	/* Page Length */
		if (data_size >= 7)
			buf[i++] = 0x00;	/* No Access Control */
		if (data_size >= 8) {
			if (support_format)
				buf[i++] = 0xC0;	/* SF, SGM */
			else
				buf[i++] = 0x00;
		}
	}

	if (data_size > sys_info_offset) {
		/* 96 Bytes Attribute Data */
		int len = data_size - sys_info_offset;
		len = (len < 96) ? len : 96;

		memcpy(buf + sys_info_offset, ms_card->raw_sys_info, len);
	}
}

static int fill_mode_sense(struct card *card, unsigned int lun,
		unsigned char *buf, int buf_len, unsigned char page_code)
{
	int size = 0;
	unsigned char *ptr;

	switch (page_code) {
	case 0x3F:
	case 0x00:
		size = 48;
		ptr = buf;
		*ptr++ = 0x2F;
		*ptr++ = 0x00;
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size) {
			/* Caching Mode Page, page code 0x08 */
			memcpy(ptr, inquiry_caching_mode, 20);
			ptr += 20;
			/* Control Mode Page, page code 0x0A */
			memcpy(ptr, inquiry_control_mode, 12);
			ptr += 12;
			/* Information Exception Mode Page, page code 0x1C */
			memcpy(ptr, inquiry_info_exception_mode, 12);
		} else {
			size = 4;
		}
		break;

	case 0x08:
		size = 24;
		ptr = buf;
		*ptr++ = 0x17;
		*ptr++ = 0x00;
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_caching_mode, 20);
		else
			size = 4;
		break;

	case 0x0A:
		size = 16;
		ptr = buf;
		*ptr++ = 0x0F;
		*ptr++ = 0x00;
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_control_mode, 12);
		else
			size = 4;
		break;

	case 0x1C:
		size = 16;
		ptr = buf;
		*ptr++ = 0x0F;
		*ptr++ = 0x00;
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_info_exception_mode, 12);
		else
			size = 4;
		break;

	default:
		size = 4;
		buf[0] = 0x03;
		buf[1] = 0x00;
		if (check_card_flag(card, WP))
			buf[2] = 0x80;
		else
			buf[2] = 0x00;
		buf[3] = 0x00;
	}

	return size;
}

static int fill_mode_sense_10(struct card *card, unsigned int lun,
		unsigned char *buf, int buf_len, unsigned char page_code)
{
	int size = 0;
	unsigned char *ptr;

	switch (page_code) {
	case 0x3F:
	case 0x00:
		size = 52;
		ptr = buf;
		/* mode data length */
		*ptr++ = 0x00;
		*ptr++ = 0x32;
		/* medium type */
		*ptr++ = 0x00;
		/* device-specific parameter */
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		/* reverse */
		*ptr++ = 0x00;
		*ptr++ = 0x00;
		/* block descriptor length */
		*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size) {
			/* Caching Mode Page, page code 0x08 */
			memcpy(ptr, inquiry_caching_mode, 20);
			ptr += 20;
			/* Control Mode Page, page code 0x0A */
			memcpy(ptr, inquiry_control_mode, 12);
			ptr += 12;
			/* Information Exception Mode Page, page code 0x1C */
			memcpy(ptr, inquiry_info_exception_mode, 12);
		} else {
			size = 8;
		}
		break;

	case 0x08:
		size = 28;
		ptr = buf;
		/* mode data length */
		*ptr++ = 0x00;
		*ptr++ = 0x17;
		/* medium type */
		*ptr++ = 0x00;
		/* device-specific parameter */
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		/* reverse */
		*ptr++ = 0x00;
		*ptr++ = 0x00;
		/* block descriptor length */
		*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_caching_mode, 20);
		else
			size = 8;
		break;

	case 0x0A:
		size = 20;
		ptr = buf;
		/* mode data length */
		*ptr++ = 0x00;
		*ptr++ = 0x0F;
		/* medium type */
		*ptr++ = 0x00;
		/* device-specific parameter */
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		/* reverse */
		*ptr++ = 0x00;
		*ptr++ = 0x00;
		/* block descriptor length */
		*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_control_mode, 12);
		else
			size = 8;
		break;

	case 0x1C:
		size = 20;
		ptr = buf;
		/* mode data length */
		*ptr++ = 0x00;
		*ptr++ = 0x0F;
		/* medium type */
		*ptr++ = 0x00;
		/* device-specific parameter */
		if (check_card_flag(card, WP))
			*ptr++ = 0x80;
		else
			*ptr++ = 0x00;
		/* reverse */
		*ptr++ = 0x00;
		*ptr++ = 0x00;
		/* block descriptor length */
		*ptr++ = 0x00;
		*ptr++ = 0x00;

		if (buf_len >= size)
			memcpy(ptr, inquiry_info_exception_mode, 12);
		else
			size = 8;
		break;

	default:
		size = 8;
		/* mode data length */
		buf[0] = 0x00;
		buf[1] = 0x03;
		/* medium type */
		buf[2] = 0x00;
		/* device-specific parameter */
		if (check_card_flag(card, WP))
			buf[3] = 0x80;
		else
			buf[3] = 0x00;
		/* reverse */
		buf[4] = 0x00;
		buf[5] = 0x00;
		/* block descriptor length */
		buf[6] = 0x00;
		buf[7] = 0x00;
	}

	return size;
}

static int mode_sense(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);
	unsigned int dataSize;
	int status;
	int pro_formatter_flag;
	unsigned char pageCode, *buf;

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		scsi_set_resid(srb, scsi_bufflen(srb));
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	pro_formatter_flag = 0;
	dataSize = 108;
#ifdef SUPPORT_MAGIC_GATE
		/* In Combo mode, device responses ModeSense command as a MS LUN
		 * when no card is inserted */
		if (MS_ID == card->id) {
			if (!card || (card->id == MS_ID)) {
				if (cr->option.mspro_formatter_enable)
					pro_formatter_flag = 1;
			}
		}
#else
		if (MS_ID == card->id) {
			if (cr->option.mspro_formatter_enable)
				pro_formatter_flag = 1;
		}
#endif

	buf = kmalloc(dataSize, GFP_KERNEL);
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	pageCode = srb->cmnd[2] & 0x3f;

	if ((pageCode == 0x3F) || (pageCode == 0x1C)
		|| (pageCode == 0x0A) || (pageCode == 0x08)
		|| (pageCode == 0x00)
		|| (pro_formatter_flag && (pageCode == 0x20))) {
		if (srb->cmnd[0] == MODE_SENSE) {
			if (pro_formatter_flag &&
				((pageCode == 0x3F) || (pageCode == 0x20))) {
				ms_mode_sense(card, srb->cmnd[0], lun, buf,
					      dataSize);
			} else {
				dataSize = fill_mode_sense(card, lun, buf,
						dataSize, pageCode);
			}
		} else {
			if (pro_formatter_flag &&
				((pageCode == 0x3F) || (pageCode == 0x20))) {
				ms_mode_sense(card, srb->cmnd[0], lun, buf,
					      dataSize);
			} else {
				dataSize = fill_mode_sense_10(card, lun, buf,
						dataSize, pageCode);
			}
		}
		status = TRANSPORT_GOOD;
	} else {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		scsi_set_resid(srb, scsi_bufflen(srb));
		status = TRANSPORT_FAILED;
	}

	if (status == TRANSPORT_GOOD) {
		unsigned int len = min(scsi_bufflen(srb), dataSize);
		scsi_set_xfer_buf(buf, len, srb);
		scsi_set_resid(srb, scsi_bufflen(srb) - len);
	}
	kfree(buf);

	return status;
}

static int request_sense(struct card_reader *cr, struct scsi_cmnd *srb)
{
	struct sense_data_t *sense;
	u8 lun = SCSI_LUN(srb);
	unsigned char *tmp, *buf;

	sense = &(cr->scsi->sense_buffer[lun]);
#if 0 /* comment out --edwin */
	if ((MS_ID == card->id)
	    && PRO_UNDER_FORMATTING(ms_card)) {
		mspro_format_sense(card, card, lun);
	}
#endif

	buf = vmalloc(scsi_bufflen(srb));
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	tmp = (unsigned char *)sense;
	memcpy(buf, tmp, scsi_bufflen(srb));

	scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	vfree(buf);

	scsi_set_resid(srb, 0);
	/* Reset Sense Data */
	set_sense_type(cr, lun, SENSE_TYPE_NO_SENSE);
	return TRANSPORT_GOOD;
}

int scsi_get_sector_info(struct scsi_cmnd *srb, u32 *lba, u16 *num)
{
	int ret = 0;
	unsigned char *cmd = srb->cmnd;

	if (cmd[0] == READ_10 || cmd[0] == WRITE_10) {
		*lba = ((u32) cmd[2] << 24) | ((u32) cmd[3] << 16) |
			((u32) cmd[4] << 8) | ((u32) cmd[5]);
		*num = ((u16) cmd[7] << 8) | cmd[8];
	} else if (cmd[0] == READ_6 || cmd[0] == WRITE_6) {
		*lba = ((u32) (cmd[1] & 0x1F) << 16) |
			((u32) cmd[2] << 8) | ((u32) cmd[3]);
		*num = cmd[4];
	} else if (cmd[0] == VENDOR_CMND && cmd[1] == SCSI_APP_CMD &&
			(cmd[2] == PP_READ10 || cmd[2] == PP_WRITE10)) {
		*lba = ((u32) cmd[4] << 24) |
			((u32) cmd[5] << 16) | ((u32) cmd[6] << 8) |
			((u32) cmd[7]);
		*num = ((u16) (cmd[9]) << 8) | cmd[10];
	} else {
		ret = -1; /* command is not support */
	}

	return ret;
}

static int read_write(struct card_reader *cr, struct scsi_cmnd *srb)
{
	struct card *card;
	u8 lun = SCSI_LUN(srb);
	u32 start_sec;
	u16 sec_cnt;
	int retval;

	card = cm_locate_card(cr->cm, lun);
	if (!card || !check_card_flag(card, READY) || card->capacity == 0) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	retval = scsi_get_sector_info(srb, &start_sec, &sec_cnt);
	if (retval) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	rts_trace(TRACE_SCSI, "lba=0x%x, num=0x%x, card->capacity=0x%x\n",
				start_sec, sec_cnt, card->capacity);
	/* sangdy2010-06-23:
	 * modify for device path exercise of DTM test will send lba 0xfffffffff
	 * and num 1. */
	if ((start_sec > card->capacity) ||
	    ((start_sec + sec_cnt) > card->capacity)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LBA_OVER_RANGE);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (0 == sec_cnt)
		goto out;

	retval = card_rw(srb, card, start_sec, sec_cnt);
	if (retval) {
		set_sense_type(cr, lun, retval);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
out:
	scsi_set_resid(srb, 0);
	return TRANSPORT_GOOD;
}

static int read_format_capacity(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	unsigned char *buf;
	unsigned int buf_len;
	int desc_cnt;
	int i = 0;
	int ret = 0;
	struct card_manager *cm = cr->cm;
	struct card *card = cm_locate_card(cm, lun);

	if (!card) {
		rts_trace(TRACE_SCSI, "media not present\n");
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	ret = check_card_readiness(card, OPTIONS_READ_FMT_CPCTY,
			srb->sc_data_direction);
	if (ret) {
		rts_trace(TRACE_SCSI, "media not ready\n");
		set_sense_type(cr, lun, ret);
		return TRANSPORT_FAILED;
	}

	buf_len = (scsi_bufflen(srb) > 12) ? 0x14 : 12;

	buf = kmalloc(buf_len, GFP_KERNEL);
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	buf[i++] = 0;
	buf[i++] = 0;
	buf[i++] = 0;

	/* Capacity List Length */
	if ((buf_len > 12) && cr->option.mspro_formatter_enable &&
			(!card || MS_ID == card->id)) {
		buf[i++] = 0x10;
		desc_cnt = 2;
	} else {
		buf[i++] = 0x08;
		desc_cnt = 1;
	}

	while (desc_cnt) {
		if (check_card_flag(card, READY)) {
			buf[i++] = (unsigned char)(card->capacity >> 24);
			buf[i++] = (unsigned char)(card->capacity >> 16);
			buf[i++] = (unsigned char)(card->capacity >> 8);
			buf[i++] = (unsigned char)(card->capacity);

			if (desc_cnt == 2)
				/* Byte[8]: Descriptor Type: Formatted medium */
				buf[i++] = 2;
			else
				buf[i++] = 0;	/* Byte[16] */
		} else {
			/* If no media is present, then the Current/Maximum/
			 * Formattable Capacity Descriptor parameters reports
			 * Number of Blocks set to the maximum capacity of
			 * a media that the Drive is capable of accessing
			 * --- MMC-6 */
			buf[i++] = 0xFF;
			buf[i++] = 0xFF;
			buf[i++] = 0xFF;
			buf[i++] = 0xFF;

			if (desc_cnt == 2)
				/* Byte[8]: Descriptor Type: No medium */
				buf[i++] = 3;
			else
				buf[i++] = 0;	/*Byte[16] */
		}

		buf[i++] = 0x00;
		buf[i++] = 0x02;
		buf[i++] = 0x00;

		desc_cnt--;
	}

	buf_len = min(scsi_bufflen(srb), buf_len);
	scsi_set_xfer_buf(buf, buf_len, srb);
	kfree(buf);

	scsi_set_resid(srb, scsi_bufflen(srb) - buf_len);

	return TRANSPORT_GOOD;
}

static int read_capacity(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned char *buf;
	u8 lun = SCSI_LUN(srb);
	struct card_manager *cm = cr->cm;
	struct card *card = cm_locate_card(cm, lun);
	int ret = 0;

	if (!card) {
		rts_trace(TRACE_SCSI, "media not present\n");
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	ret = check_card_readiness(card, OPTIONS_READ_CPCTY,
					srb->sc_data_direction);
	if (ret) {
		rts_trace(TRACE_SCSI, "media not ready\n");
		set_sense_type(cr, lun, ret);
		return TRANSPORT_FAILED;
	}
	buf = kmalloc(8, GFP_KERNEL);
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	buf[0] = (unsigned char)((card->capacity - 1) >> 24);
	buf[1] = (unsigned char)((card->capacity - 1) >> 16);
	buf[2] = (unsigned char)((card->capacity - 1) >> 8);
	buf[3] = (unsigned char)(card->capacity - 1);
	buf[4] = 0x00;
	buf[5] = 0x00;
	buf[6] = 0x02;
	buf[7] = 0x00;

	rts_trace(TRACE_SCSI, "card capacity %d\n", card->capacity);
	scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	scsi_set_resid(srb, 0);

	kfree(buf);

	return TRANSPORT_GOOD;
}

static int read_capacity16(struct card_reader *cr, struct scsi_cmnd *srb)
{
	unsigned char *buf;
	u8 lun = SCSI_LUN(srb);
	struct card_manager *cm = cr->cm;
	struct card *card = cm_locate_card(cm, lun);
	int ret = 0;

	if (!card) {
		rts_trace(TRACE_SCSI, "media not present\n");
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	ret = check_card_readiness(card, OPTIONS_READ_CPCTY,
			srb->sc_data_direction);
	if (ret) {
		rts_trace(TRACE_SCSI, "media not ready\n");
		set_sense_type(cr, lun, ret);
		return TRANSPORT_FAILED;
	}

	buf = kmalloc(32, GFP_KERNEL);
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	buf[0] = 0x00;
	buf[1] = 0x00;
	buf[2] = 0x00;
	buf[3] = 0x00;
	buf[4] = (unsigned char)((card->capacity - 1) >> 24);
	buf[5] = (unsigned char)((card->capacity - 1) >> 16);
	buf[6] = (unsigned char)((card->capacity - 1) >> 8);
	buf[7] = (unsigned char)(card->capacity - 1);
	buf[8] = 0x00;
	buf[9] = 0x00;
	buf[10] = 0x02;
	buf[11] = 0x00;
	buf[12] = 0x00;

	rts_trace(TRACE_SCSI, "card capacity %d\n", card->capacity);
	scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	scsi_set_resid(srb, 0);

	kfree(buf);

	return TRANSPORT_GOOD;
}

static int verify(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card_manager *cm = cr->cm;
	struct card *card = cm_locate_card(cm, lun);
	u32 start_sec = 0;
	u32 sec_cnt = 0;

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	start_sec = ((u32)srb->cmnd[2]<<24) | ((u32)srb->cmnd[3]<<16)
		| ((u32)srb->cmnd[4]<<8) | ((u32)srb->cmnd[5]);
	sec_cnt = ((u32)srb->cmnd[7]<<8) | ((u32)srb->cmnd[8]);

	if ((start_sec > card->capacity) ||
			((start_sec + sec_cnt) > card->capacity)) {
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return TRANSPORT_GOOD;
}

/* scsi_get_dev_sts - Get device status
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_get_dev_sts(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_get_dev_sts)
		ret = cr->specific_cmd_handlers->app_get_dev_sts(cr, srb);
	else
		ret = scsi_common_get_dev_sts(cr, srb);

	return ret;
}


/* scsi_rd_host_reg - Read Host Register
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_rd_host_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_rd_host_reg)
		ret = cr->specific_cmd_handlers->app_rd_host_reg(cr, srb);
	else
		ret = scsi_common_rd_host_reg(cr, srb);

	return ret;
}


/* scsi_wr_host_reg - Write Host Register
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_wr_host_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_wr_host_reg)
		ret = cr->specific_cmd_handlers->app_wr_host_reg(cr, srb);
	else
		ret = scsi_common_wr_host_reg(cr, srb);

	return ret;
}

static int scsi_set_card_clock(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 clock = srb->cmnd[5];
	struct card *card = NULL;

	if (SD_CARD == srb->cmnd[4]) {
		struct sd_info *sd_card;
		card = cm_get_card(cr->cm, SD_ID);
		if (!card)
			goto fail;
		sd_card = card->extra;
		sd_card->sd_clock = (int) clock;
	} else if (MS_CARD == srb->cmnd[4]) {
		struct ms_info *ms_card;
		card = cm_get_card(cr->cm, MS_ID);
		if (!card)
			goto fail;
		ms_card = card->extra;
		ms_card->ms_clock = (int) clock;
	} else {
		goto fail;
	}

	return TRANSPORT_GOOD;

fail:
	set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
	return TRANSPORT_FAILED;
}

static int scsi_get_card_clock(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 clock = 0;
	struct card *card = NULL;

	if (SD_CARD == srb->cmnd[4]) {
		struct sd_info *sd_card;
		card = cm_get_card(cr->cm, SD_ID);
		if (!card)
			goto fail;
		sd_card = card->extra;
		clock = (u8) (sd_card->sd_clock);
	} else if (MS_CARD == srb->cmnd[4]) {
		struct ms_info *ms_card;
		card = cm_get_card(cr->cm, MS_ID);
		if (!card)
			goto fail;
		ms_card = card->extra;
		clock = (u8) (ms_card->ms_clock);
	} else {
		goto fail;
	}

	scsi_stor_set_xfer_buf(&clock, 1, srb);
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;

fail:
	set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
	return TRANSPORT_FAILED;
}

static int scsi_set_card_reset_flags(struct card_reader *cr,
						struct scsi_cmnd *srb)
{
	u32 flags;
	struct card *card = NULL;

	card = cm_get_card(cr->cm, SD_ID);
	if (!card)
		goto fail;

	flags = ((u32)(srb->cmnd[7]) << 24) |
			((u32)(srb->cmnd[6]) << 16) |
			((u32)(srb->cmnd[5]) << 8) |
			(srb->cmnd[4]);

	card->reset_flags = flags;

	rts_trace(TRACE_SCSI, "SD set flags:%#x\n", flags);

	return TRANSPORT_GOOD;
fail:
	set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
	return TRANSPORT_FAILED;
}

static int scsi_get_card_reset_flags(struct card_reader *cr,
						struct scsi_cmnd *srb)
{
	u32 flags = 0;
	struct card *card = NULL;

	card = cm_get_card(cr->cm, SD_ID);
	if (!card)
		goto fail;

	flags = card->reset_flags;
	rts_trace(TRACE_SCSI, "SD get flags:%#x\n", flags);

	scsi_stor_set_xfer_buf((unsigned char *)&flags, 4, srb);
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;
fail:
	set_sense_type(cr, SCSI_LUN(srb), SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
	return TRANSPORT_FAILED;
}

/* scsi_set_var - Set variable
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_set_var(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	switch (srb->cmnd[3]) {
	case 1:
		ret = scsi_set_card_clock(cr, srb);
		break;
	case 2:
		ret = scsi_set_card_reset_flags(cr, srb);
		break;
	case 3:
	case 4:
	default:
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		ret = TRANSPORT_FAILED;
		break;
	}

	return ret;
}



/* scsi_get_var - Get variable
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_get_var(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	switch (srb->cmnd[3]) {
	case 1:
		ret = scsi_get_card_clock(cr, srb);
		break;
	case 2:
		ret = scsi_get_card_reset_flags(cr, srb);
		break;
	case 3:
	case 4:
	default:
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		ret = TRANSPORT_FAILED;
		break;
	}

	return ret;
}


/* scsi_dma_rb - Access Ring Buffer via DMA
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_dma_rb(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_dma_rb)
		ret = cr->specific_cmd_handlers->app_dma_rb(cr, srb);
	else
		ret = scsi_common_dma_rb(cr, srb);

	return ret;
}


/* scsi_rd_cfg_byte - Read PCIe configuration
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_rd_cfg_byte(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_rd_cfg_byte)
		ret = cr->specific_cmd_handlers->app_rd_cfg_byte(cr, srb);
	else
		ret = scsi_common_rd_cfg_byte(cr, srb);

	return ret;
}


/* scsi_wr_cfg_byte - Write PCIe configuration
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_wr_cfg_byte(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_wr_cfg_byte)
		ret = cr->specific_cmd_handlers->app_wr_cfg_byte(cr, srb);
	else
		ret = scsi_common_wr_cfg_byte(cr, srb);

	return ret;
}


/* scsi_set_chip_mode - Set chip mode
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_set_chip_mode(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_set_chip_mode)
		ret = cr->specific_cmd_handlers->app_set_chip_mode(cr, srb);
	else
		ret = scsi_common_set_chip_mode(cr, srb);

	return ret;
}


/* scsi_sd40_rw_partial_bytes - Read/Write for SD4.0 in partical way
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_sd40_rw_partial_bytes(struct card_reader *cr,
						struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_sd40_rw_partial_bytes) {
		ret = cr->specific_cmd_handlers->app_sd40_rw_partial_bytes(
								cr, srb);
	} else {
		printk("%s: app_sd40_rw_partial_bytes callback is not"
						"implemented\n", cr->name);
		ret = -EINVAL;
	}

	return ret;
}



/* scsi_rd_sts - Read status
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_rd_sts(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->vendor_rd_sts)
		ret = cr->specific_cmd_handlers->vendor_rd_sts(cr, srb);
	else
		ret = scsi_common_rd_sts(cr, srb);

	return ret;
}

static int scsi_set_led(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 buf[2] = {0x55, 0x55};
	u8 led_mode, led_val;

	led_mode = srb->cmnd[4];
	led_val = srb->cmnd[5];

	rts_trace(TRACE_SCSI, "led_mode:%d,led_val:%d\n", led_mode, led_val);

	if (led_mode) {
		SET_STATUS_FLAG(cr, LED_TEST_MODE);
		cr_disable_auto_blink(cr);
		if (led_val)
			cr_turn_on_led(cr);
		else
			cr_turn_off_led(cr);
	} else {
		CLEAR_STATUS_FLAG(cr, LED_TEST_MODE);
	}

	scsi_stor_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;
}

/* scsi_rd_mem - Read memory
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_rd_mem(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->vendor_rd_mem)
		ret = cr->specific_cmd_handlers->vendor_rd_mem(cr, srb);
	else
		ret = scsi_common_rd_mem(cr, srb);

	return ret;
}

/* scsi_wr_mem - Write memory
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_wr_mem(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->vendor_wr_mem)
		ret = cr->specific_cmd_handlers->vendor_wr_mem(cr, srb);
	else
		ret = scsi_common_wr_mem(cr, srb);

	return ret;
}


static int get_sd_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	u8 regtype = srb->cmnd[2];
	struct card *card = cm_locate_card(cr->cm, lun);
	struct sd_info *sd_card;
	unsigned char *buf = NULL;

	if (!card || !check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	sd_card = card->extra;

	if (SD_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	switch (regtype) {
	case 0: /* CSD */
		buf = sd_card->raw_csd;
		break;
	case 1: /* CID */
		buf = sd_card->raw_cid;
		break;
	case 2: /* OCR */
		rts_trace(TRACE_SCSI, "OCR is not support!\n");
		set_sense_type(cr, lun, SENSE_TYPE_FORMAT_CMD_FAILED);
		TRACE_RET(cr, TRANSPORT_FAILED);
		break;
	case 3: /* RCA */
		buf = (unsigned char *)&sd_card->sd_addr;
		break;
	case 4: /* SCR */
		buf = sd_card->raw_scr;
		break;
	case 5: /* Ext CSD */
		rts_trace(TRACE_SCSI, "Ext CSD is not support!\n");
		set_sense_type(cr, lun, SENSE_TYPE_FORMAT_CMD_FAILED);
		TRACE_RET(cr, TRANSPORT_FAILED);
		break;
	default:
		rts_trace(TRACE_SCSI, "regtpye: %x is not support!\n", regtype);
	}

	scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;
}


/* scsi_suit_cmd - Perfrom suit command operations
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_suit_cmd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_suit_cmd)
		ret = cr->specific_cmd_handlers->app_suit_cmd(cr, srb);
	else
		ret = scsi_common_suit_rw_mem_cmd_buf(cr, srb);

	return ret;

}

/* scsi_rd_phy_reg - Read phy register
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_rd_phy_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_rd_phy_reg)
		ret = cr->specific_cmd_handlers->app_rd_phy_reg(cr, srb);
	else
		ret = scsi_common_rd_phy_reg(cr, srb);

	return ret;
}

/* scsi_wr_phy_reg - Write phy register
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int scsi_wr_phy_reg(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	if (cr->specific_cmd_handlers->app_wr_phy_reg)
		ret = cr->specific_cmd_handlers->app_wr_phy_reg(cr, srb);
	else
		ret = scsi_common_wr_phy_reg(cr, srb);

	return ret;
}


static int get_card_bus_width(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	u8 bus_width;
	struct card *card = cm_locate_card(cr->cm, lun);

	if (!card || !check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (SD_ID == card->id || MS_ID == card->id) {
		bus_width = card->bus_width;
	} else {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	scsi_set_xfer_buf(&bus_width, scsi_bufflen(srb), srb);
	scsi_set_resid(srb, 0);

	return TRANSPORT_GOOD;
}

#ifdef _MSG_TRACE
static int trace_msg_cmd(struct scsi_cmnd *srb, struct card_reader *cr)
{
	unsigned char *buf = NULL;
	u8 clear;
	unsigned int buf_len;

	buf_len =
	    4 +
	    ((2 + MSG_FUNC_LEN + MSG_FILE_LEN + TIME_VAL_LEN) * TRACE_ITEM_CNT);

	if ((scsi_bufflen(srb) < buf_len) || (scsi_sglist(srb) == NULL)) {
		set_sense_type(cr, SCSI_LUN(srb),
			       SENSE_TYPE_MEDIA_UNRECOVER_READ_ERR);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	clear = srb->cmnd[2];

	buf = vmalloc(scsi_bufflen(srb));
	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	trace_msg(cr, buf, clear);

	scsi_set_xfer_buf(buf, scsi_bufflen(srb), srb);
	vfree(buf);

	scsi_set_resid(srb, 0);
	return TRANSPORT_GOOD;
}
#endif

static int suit_cmd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int result;

	switch (srb->cmnd[3]) {
	case INIT_BATCHCMD:
	case ADD_BATCHCMD:
	case SEND_BATCHCMD:
	case GET_BATCHRSP:
		result = scsi_suit_cmd(cr, srb);
		break;
	default:
		result = TRANSPORT_ERROR;
	}

	return result;
}

static int scsi_reset_card(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);

	set_bit(card->id, &(cr->need_reset));

	return 0;
}

static int scsi_release_card(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);

	set_bit(card->id, &(cr->need_release));

	return 0;
}

/* scsi_get_uhsii_status - get UHSII capability and setting
 *
 * It returns the info saved in sd_card, while maybe that is not
 * a good idea. But if it sends CCMD to get capability and setting
 * of UHSII, pre-read or pre-wirte should be breaked firstly.
 */
static int scsi_get_uhsii_status(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int buf_len;
	u8 lun = SCSI_LUN(srb);
	u8 status[32] = {0};
	u8 *payload;
	struct card *card = cm_locate_card(cr->cm, lun);
	struct sd_info *sd_card;

	if (!card)
		return TRANSPORT_FAILED;
	sd_card = card->extra;
	if (!sd_card)
		return TRANSPORT_FAILED;
	if (!CHK_SD_UHSII(sd_card))
		return TRANSPORT_FAILED;

	/* Byte0~15, UHSII CFG_REG capability information */
	payload = sd_card->cap.dev.raw_generic;

	/* Byte0, SD Capacity Info */
	if (CHK_SD_XC(sd_card))
		status[0] = 1 << PP_SD_XC;
	else if (CHK_SD_HC(sd_card))
		status[0] = 1 << PP_SD_HC;
	else
		status[0] = 1 << PP_SD_SC;

	/* bit0~7: Application type */
	status[1] = payload[5];
	/* bit0~3: Device-specific Number of Lanes and Functionality
	 * bit6: DADR Length 0b = 4Byte, 1b = Reserved.
	 */
	status[2] = payload[6];

	/* Speed capacity */
	status[4] |= 1 << 5;
	if (payload[6] & 0x01)
		status[4] |= 1 << 6;

	payload = sd_card->cap.dev.raw_phy;

	/* bit0~3: Device-specific N_LSS_SYN
	 * bit4~7: Device-specific N_LSS_DIR
	 */
	status[5] = payload[3];
	/*bit7: Support Hibernate Mode */
	status[6] = payload[6];
	/* bit0~3: PHY Minor Revision
	 * bit4~5: PHY Major Revision
	 */
	status[7] = payload[7];

	payload = sd_card->cap.dev.raw_link;

	/* bit0~7: Device-specific N_DATA_GAP */
	status[9] = payload[3];
	/* bit0~7: Device-specific MAX_BLKLEN[4-11] */
	status[10] = payload[4];
	/* bit4~7: Device-specific MAX_BLKLEN[0~3]
	 * bit0~3: Device Type
	 */
	status[11] = payload[5];
	/* Device-specific N_FCU */
	status[12] = payload[6];
	/* bit0~3: LINK/TRAN Minor Revision
	 * bit4~5: LINK/TRAN Major Revision
	 */
	status[13] = payload[7];

	/* Byte16~31, UHSII CFG_REG setting information */
	payload = sd_card->setting.dev.raw_generic;

	/* Byte16, SD Capacity Info */
	if (CHK_SD_XC(sd_card))
		status[16] = 1 << PP_SD_XC;
	else if (CHK_SD_HC(sd_card))
		status[16] = 1 << PP_SD_HC;
	else
		status[16] = 1 << PP_SD_SC;

	/* bit7: Cofig Completion */
	status[17] = payload[0];
	/* bit0~3: Number of Lanes and Functionality */
	status[18] = payload[6];
	/* bit0: Powr Control Mode in Active Control */
	status[19] = payload[7];

	/* Speed mode */
	if (DM_2L_HD == sd_card->lan_mode)
		status[20] |= 1 << 6;
	else
		status[20] |= 1 << 5;


	payload = sd_card->setting.dev.raw_phy;

	/* bit0~3: N_LSS_SYN, bit4~7: N_LSS_DIR */
	status[21] = payload[3];
	/* bit4~5: Selected PHY Major Revision
	 * bit6~7: Selected Transmission Speed Range
	 */
	status[23] = payload[7];

	payload = sd_card->setting.dev.raw_link;

	/* N_DATA_GAP */
	status[25] = payload[3];
	/* bit0~7: Device-specific MAX_BLKLEN[4-11] */
	status[26] = payload[4];
	/* bit4~7: Device-specific MAX_BLKLEN[0~3]
	 * bit0~1: MAX_RETYR_NUM
	 */
	status[27] = payload[5];
	/* Device-specific N_FCU */
	status[28] = payload[6];

	buf_len = min(scsi_bufflen(srb), (unsigned int)sizeof(status));
	scsi_stor_set_xfer_buf(status, buf_len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - buf_len);

	return TRANSPORT_GOOD;
}

static int scsi_get_card_status(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int buf_len;
	u8 lun = SCSI_LUN(srb);
	u8 status[32] = {0};
	u8 current_limit[4] = {0x01, 0x02, 0x04, 0x08};
	struct card *card = cm_locate_card(cr->cm, lun);

	if (!card)
		return TRANSPORT_FAILED;

	if (SD_ID == card->id) {
		struct sd_info *sd_card = card->extra;
		if (sd_card && CHK_SD_UHSII(sd_card)) {
			return scsi_get_uhsii_status(cr, srb);
		} else if (sd_card && CHK_SD(sd_card)) {
			/* Byte0~15, static SD information */

			/* Byte0, SD Capacity Info */
			if (CHK_SD_XC(sd_card))
				status[0] = 1 << PP_SD_XC;
			else if (CHK_SD_HC(sd_card))
				status[0] = 1 << PP_SD_HC;
			else
				status[0] = 1 << PP_SD_SC;

			/* Byte1~3, reserved */

			/* Byte4, SD function group 1 (SD Access Mode)) */
			status[4] = sd_card->func_group1_mask;

			/* Byte5, SD functin Group 2(SD Command system) */
			status[5] = sd_card->func_group2_mask;

			/* Byte6, SD functin Group 3(SD Driver Strength) */
			status[6] = sd_card->func_group3_mask;

			/* Byte7, SD functin Group 4(SD Current system) */
			status[7] = sd_card->func_group4_mask;

			/* Byte8~15, reserved */

			/* Byte16~31, dynamic SD information */

			/* Byte16, SD Capacity Info */
			if (CHK_SD_XC(sd_card))
				status[16] = 1 << PP_SD_XC;
			else if (CHK_SD_HC(sd_card))
				status[16] = 1 << PP_SD_HC;
			else
				status[16] = 1 << PP_SD_SC;

			/* Byte17~19, reserved */

			/* Byte20, SD function group 1 (SD Access Mode)) */
			if (CHK_SD_DDR50(sd_card))
				status[20] = 0x10;
			else if (CHK_SD_SDR104(sd_card))
				status[20] = 0x08;
			else if (CHK_SD_SDR50(sd_card))
				status[20] = 0x04;
			else if (CHK_SD_HS(sd_card))
				status[20] = 0x02;
			else
				status[20] = 0x01;

			/* Byte21, SD functin Group 2(SD Command system) */
			status[21] = 0x00;

			/* Byte22, SD functin Group 3(SD Driver Strength) */
			status[22] = 0x00;

			/* Byte23, SD functin Group 4(SD Current system) */
			if (sd_card->current_limit < 4)
				status[23] = current_limit[sd_card->current_limit];
			else
				status[23] = 0;

			/* Byte24~31, reserved */
		} else {
			/* Byte0~15, static MMC information */

			/* Byte0, MMC Capacity Info */
			if (CHK_MMC_SECTOR_MODE(sd_card))
				status[0] = PP_MMC_HC;
			else
				status[0] = 0;

			/* Byte1~15, reserved */

			/* Byte16~31, dynamic MMC information */

			/* Byte16, MMC Capacity Info */
			if (CHK_MMC_SECTOR_MODE(sd_card))
				status[16] = PP_MMC_HC;
			else
				status[16] = 0;

			/* Byte17~31, reserved */
		}
	}

	buf_len = min(scsi_bufflen(srb), (unsigned int)sizeof(status));
	scsi_stor_set_xfer_buf(status, buf_len, srb);
	scsi_set_resid(srb, scsi_bufflen(srb) - buf_len);

	return TRANSPORT_GOOD;
}

static int scsi_set_sd_speed_prior(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;
	struct cr_option *option = &cr->option;
	switch (srb->cmnd[4]) {
	case HS_SUPPORT_MASK:
		option->sd_speed_prior = 0x01010101;
		break;
	case SDR50_SUPPORT_MASK:
		option->sd_speed_prior = 0x02020202;
		break;
	case SDR104_SUPPORT_MASK:
		option->sd_speed_prior = 0x03030303;
		break;
	case DDR50_SUPPORT_MASK:
		option->sd_speed_prior = 0x04040404;
		break;
	default:
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		ret = TRANSPORT_FAILED;
		break;
	}
	rts_trace(TRACE_SCSI, "scsi_set_sd_speed_prior return %#x\n", ret);
	return ret;
}

static int scsi_set_card_mode(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int ret = 0;

	switch (srb->cmnd[3]) {
	case CARD_MODE_CF:
	case CARD_MODE_SM:
	case CARD_MODE_MS:
	case CARD_MODE_XD:
	case CARD_MODE_MMC:
	case CARD_MODE_MSPRO:
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		ret = TRANSPORT_FAILED;
		break;
	case CARD_MODE_SD:
	case CARD_MODE_SDHC:
		/* group 1 sd speed prior */
		ret = scsi_set_sd_speed_prior(cr, srb);
		break;
	default:
		set_sense_type(cr, SCSI_LUN(srb),
				SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		ret = TRANSPORT_FAILED;
		break;
	}
	return ret;
}

static int sd_scsi_rw_csd(struct card_reader *cr, struct card *card, struct scsi_cmnd *srb)
{
	int retval = 0;
	struct sd_info *sd_card = card->extra;
	struct sd_command cmd;
	unsigned int rsp_len = 0;
	u32 opcode = 0;
	u32 arg = 0;
	u32 csd_len = 16;
	u8 rsp_type = 0;
	u8 dir = HOST_TO_DEVICE;
	u8 bus_width = 0;
	u8 lun = SCSI_LUN(srb);
	u8 *buf = (u8 *)vmalloc(csd_len);

	if (buf == NULL)
		TRACE_RET(cr, TRANSPORT_ERROR);

	if (srb->cmnd[2] == 0x70) {
		opcode = SEND_CSD;
		arg = sd_card->sd_addr;
		rsp_type = SD_RSP_TYPE_R2;
		rsp_len = 16;
		dir = DEVICE_TO_HOST;
	} else {
		opcode = 0x40 | PROGRAM_CSD;
		arg = 0;
		rsp_type = SD_RSP_TYPE_R1;
		rsp_len = 0;
		scsi_stor_get_xfer_buf(buf, csd_len, srb);
		dir = HOST_TO_DEVICE;
	}

	if (DEVICE_TO_HOST == dir) {
		retval = sd_select_card(cr, card, 0);
		if (retval != STATUS_SUCCESS)
			TRACE_GOTO(cr, SD_SCSI_RW_CSD_Failed);
	}

	if ((sd_card->sd_lock_status & SD_LOCK_1BIT_MODE) == 0) {
		if (CHK_MMC_8BIT(sd_card))
			bus_width = SD_BUS_WIDTH_8;
		else if (CHK_SD(sd_card) || CHK_MMC_4BIT(sd_card))
			bus_width = SD_BUS_WIDTH_4;
		else
			bus_width = SD_BUS_WIDTH_1;
	} else {
		bus_width = SD_BUS_WIDTH_4;
	}

	INIT_SD_CMD(cmd, opcode, arg, 0, rsp_type, rsp_len);
	if (HOST_TO_DEVICE == dir)
		init_sd_dcmd_data(cmd, buf, csd_len, csd_len, 1, bus_width,
			SD_TM_AUTO_WRITE_2, TO_SD, 150);
	retval = sd_send_cmd(cr, card, &cmd);
	if (retval) {
		/* Wait till SD DATA IDLE */
		sd_wait_data_idle(cr);
		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);
		TRACE_GOTO(cr, SD_SCSI_RW_CSD_Failed);
	}

	if (DEVICE_TO_HOST == dir) {
		memcpy(buf, cmd.rsp + 1, 15);

		/* Get CRC7 */
		retval = sd_get_crc7(cr, buf + 15);
		if (STATUS_SUCCESS != retval)
			TRACE_RET(cr, retval);

		rts_trace(TRACE_SD, "CSD Response:\n");
		hex_dump(buf, 16);
		scsi_stor_set_xfer_buf(buf, csd_len, srb);

		retval = sd_select_card(cr, card, 1);
		if (retval != STATUS_SUCCESS)
			TRACE_GOTO(cr, SD_SCSI_RW_CSD_Failed);
	} else {
		sd_clear_status(card);
	}

	scsi_set_resid(srb, 0);

	return retval;

SD_SCSI_RW_CSD_Failed:
	set_sense_type(cr, lun, SENSE_TYPE_NO_SENSE);
	sd_release(cr, card);
	clr_card_flags(card, READY);
	rts_trace(TRACE_SD, "Reset SD for SD CSD RW fail!\n");
	process_card_reset(card);
	if (!check_card_flag(card, READY))
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);

	TRACE_RET(cr, TRANSPORT_FAILED);
}

static int scsi_rw_csd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);
	struct sd_info *sd_card = NULL;

	if (!card || !check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (SD_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	} else {
		sd_card = card->extra;
	}

	cr_prepare_run(cr);

	sd_cleanup_work(cr, card);

	return sd_scsi_rw_csd(cr, card, srb);
}

static int app_cmd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int result = TRANSPORT_GOOD;

	switch (srb->cmnd[2]) {
	case PP_READ10:
	case PP_WRITE10:
		result = read_write(cr, srb);
		break;

	case SUIT_CMD:
		result = suit_cmd(cr, srb);
		break;

	case READ_PHY:
		result = scsi_rd_phy_reg(cr, srb);
		break;

	case WRITE_PHY:
		result = scsi_wr_phy_reg(cr, srb);
		break;

	case GET_DEV_STATUS:
		result = scsi_get_dev_sts(cr, srb);
		break;

	case READ_HOST_REG:
		result = scsi_rd_host_reg(cr, srb);
		break;

	case WRITE_HOST_REG:
		result = scsi_wr_host_reg(cr, srb);
		break;

	case GET_VAR:
		result = scsi_get_var(cr, srb);
		break;

	case SET_VAR:
		result = scsi_set_var(cr, srb);
		break;

	case DMA_READ:
	case DMA_WRITE:
		result = scsi_dma_rb(cr, srb);
		break;

	case READ_CFG:
		result = scsi_rd_cfg_byte(cr, srb);
		break;

	case WRITE_CFG:
		result = scsi_wr_cfg_byte(cr, srb);
		break;

	case SET_CHIP_MODE:
		result = scsi_set_chip_mode(cr, srb);
		break;

	case RESET_CARD:
		result = scsi_reset_card(cr, srb);
		break;
	case RELEASE_CARD:
		result = scsi_release_card(cr, srb);
		break;

	case READ10_BYTE:
	case WRITE10_BYTE:
		result = scsi_sd40_rw_partial_bytes(cr, srb);
		break;
	case GET_CARD_STATUS:
		result = scsi_get_card_status(cr, srb);
		break;

	case CARD_MODE:
		result = scsi_set_card_mode(cr, srb);
		break;

	case READ_CSD:
	case WRITE_CSD:
		result = scsi_rw_csd(cr, srb);
		break;

	default:
		set_sense_type(cr, SCSI_LUN(srb),
			       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return result;
}

static int vendor_cmnd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	int result = TRANSPORT_GOOD;

	switch (srb->cmnd[1]) {
	case READ_STATUS:
		result = scsi_rd_sts(cr, srb);
		break;
	case SET_LED:
		result = scsi_set_led(cr, srb);
		break;
	case READ_MEM:
		result = scsi_rd_mem(cr, srb);
		break;

	case WRITE_MEM:
		result = scsi_wr_mem(cr, srb);
		break;

	case GET_BUS_WIDTH:
		result = get_card_bus_width(cr, srb);
		break;

	case GET_SD_REG:
		result = get_sd_reg(cr, srb);
		break;
#ifdef _MSG_TRACE
	case TRACE_MSG:
		result = trace_msg_cmd(srb, cr);
		break;
#endif
	case SCSI_APP_CMD:
		result = app_cmd(cr, srb);
		break;

	default:
		set_sense_type(cr, SCSI_LUN(srb),
			       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return result;
}

static int ms_format_cmnd(struct card *card, struct scsi_cmnd *srb)
{
	struct ms_info *ms_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	u8 lun = SCSI_LUN(srb);
	int retval, quick_format;

	if (MS_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if ((srb->cmnd[3] != 0x4D) || (srb->cmnd[4] != 0x47)
	    || (srb->cmnd[5] != 0x66) || (srb->cmnd[6] != 0x6D)
	    || (srb->cmnd[7] != 0x74)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (srb->cmnd[8] & 0x01)
		quick_format = 0;
	else
		quick_format = 1;

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (check_card_flag(card, WP)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_WRITE_PROTECT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (!CHK_MSPRO(ms_card)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	cr_prepare_run(cr);

	retval = ms_format(cr, card, srb, MS_SHORT_DATA_LEN,
								quick_format);
	if (retval != STATUS_SUCCESS) {
		set_sense_type(cr, lun, SENSE_TYPE_FORMAT_CMD_FAILED);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	scsi_set_resid(srb, 0);
	return TRANSPORT_GOOD;
}

#ifdef SUPPORT_PCGL_1P18
int ms_get_information(struct card *card, struct scsi_cmnd *srb)
{
	struct ms_info *ms_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	u8 lun = SCSI_LUN(srb);
	u8 dev_info_id, data_len;
	u8 *buf;
	unsigned int buf_len;
	int i;

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (MS_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if ((srb->cmnd[2] != 0xB0) || (srb->cmnd[4] != 0x4D) ||
	    (srb->cmnd[5] != 0x53) || (srb->cmnd[6] != 0x49) ||
	    (srb->cmnd[7] != 0x44)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	dev_info_id = srb->cmnd[3];
	if ((CHK_MSXC(ms_card) && (dev_info_id == 0x10)) ||
	    (!CHK_MSXC(ms_card) && (dev_info_id == 0x13)) ||
	    !CHK_MSPRO(ms_card)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (dev_info_id == 0x15)
		buf_len = data_len = 0x3A;
	else
		buf_len = data_len = 0x6A;

	buf = kmalloc(buf_len, GFP_KERNEL);
	if (!buf)
		TRACE_RET(cr, TRANSPORT_ERROR);

	i = 0;
	/* GET Memory Stick Media Information Response Header */
	buf[i++] = 0x00;	/* Data length MSB */
	buf[i++] = data_len;	/* Data length LSB */
	/* Device Information Type Code */
	if (CHK_MSXC(ms_card))
		buf[i++] = 0x03;
	else
		buf[i++] = 0x02;
	/* SGM bit */
	buf[i++] = 0x01;
	/* Reserved */
	buf[i++] = 0x00;
	buf[i++] = 0x00;
	buf[i++] = 0x00;
	/* Number of Device Information */
	buf[i++] = 0x01;

	/*  Device Information Body
	 *  Device Information ID Number */
	buf[i++] = dev_info_id;
	/* Device Information Length */
	if (dev_info_id == 0x15)
		data_len = 0x31;
	else
		data_len = 0x61;
	buf[i++] = 0x00;	/* Data length MSB */
	buf[i++] = data_len;	/* Data length LSB */
	/* Valid Bit */
	buf[i++] = 0x80;
	if ((dev_info_id == 0x10) || (dev_info_id == 0x13)) {
		/* System Information */
		memcpy(buf + i, ms_card->raw_sys_info, 96);
	} else {
		/* Model Name */
		memcpy(buf + i, ms_card->raw_model_name, 48);
	}

	scsi_set_xfer_buf(buf, buf_len, srb);

	if (dev_info_id == 0x15)
		scsi_set_resid(srb, scsi_bufflen(srb) - 0x3C);
	else
		scsi_set_resid(srb, scsi_bufflen(srb) - 0x6C);

	kfree(buf);
	return STATUS_SUCCESS;
}
#endif

static int ms_sp_cmnd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);
	int retval = TRANSPORT_ERROR;

	if (!card || !check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		return TRANSPORT_FAILED;
	}

	if (srb->cmnd[2] == MS_FORMAT)
		retval = ms_format_cmnd(card, srb);
#ifdef SUPPORT_PCGL_1P18
	else if (srb->cmnd[2] == GET_MS_INFORMATION)
		retval = ms_get_information(card, srb);
#endif

	return retval;
}

static int sd_cprm_check_secure_cmnd(struct scsi_cmnd *srb)
{
	u8 cmd_idx = 0;
	u8 acmd = 0;
	int res = 0;
	cmd_idx = srb->cmnd[2] & 0x3F;
	if (srb->cmnd[1] & 0x01)
		acmd = 1;
	if (0 == acmd)
		return res;
	switch (cmd_idx) {
	case 18:	/* SECURE_READ_MULTI_BLOCK */
	case 25:	/* SECURE_WRITE_MULTI_BLOCK */
	case 26:	/* SECURE_WRITE_MKB */
	case 38:	/* SECURE_ERASE */
	case 43:	/* GET_MKB */
	case 44:	/* GET_MID */
	case 45:	/* SET_CER_RN1 */
	case 46:	/* GET_CER_RN2 */
	case 47:	/* SET_CER_RES2 */
	case 48:	/* GET_CER_RES1 */
	case 49:	/* CHANGE_SECURE_AREA */
		res = 1;
		break;
	default:
		break;
	}
	return res;
}

static int sd_cprm_cmnd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	int result;
	struct card *card = cm_locate_card(cr->cm, lun);
	struct sd_info *sd_card = NULL;

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	cr_prepare_run(cr);

	sd_cleanup_work(cr, card);

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (SD_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	} else {
		sd_card = card->extra;
	}

	if (0 == sd_card->sd_cprm_sp &&
		sd_cprm_check_secure_cmnd(srb)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	rts_trace(TRACE_SCSI, "CPRM CMD: %2x %2x %2x %2x "
			"%2x %2x %2x %2x %2x %2x %2x %2x\n",
		srb->cmnd[0], srb->cmnd[1], srb->cmnd[2], srb->cmnd[3],
		srb->cmnd[4], srb->cmnd[5], srb->cmnd[6], srb->cmnd[7],
		srb->cmnd[8], srb->cmnd[9], srb->cmnd[10], srb->cmnd[11]);
	switch (srb->cmnd[0]) {
	case SD_PASS_THRU_MODE:
		result = sd_cprm_pass_thru_mode(cr, card, srb);
		break;

	case SD_EXECUTE_NO_DATA:
		result = sd_cprm_no_data(cr, card, srb);
		break;

	case SD_EXECUTE_READ:
		result = sd_cprm_read_data(cr, card, srb);
		break;

	case SD_EXECUTE_WRITE:
		result = sd_cprm_write_data(cr, card, srb);
		break;

	case SD_GET_RSP:
		result = sd_cprm_get_cmd_rsp(cr, card, srb);
		break;

	case SD_HW_RST:
		result = sd_cprm_hw_rst(cr, card, srb);
		break;

	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return result;
}

static int sd_ge_cmnd(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	int err = 0;
	struct card *card = cm_locate_card(cr->cm, lun);

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	cr_prepare_run(cr);
	sd_cleanup_work(cr, card);

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (SD_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	rts_trace(TRACE_SCSI, "Gamma/EMID CMD: %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x\n",
		srb->cmnd[0], srb->cmnd[1], srb->cmnd[2], srb->cmnd[3],
		srb->cmnd[4], srb->cmnd[5], srb->cmnd[6], srb->cmnd[7],
		srb->cmnd[8], srb->cmnd[9], srb->cmnd[10], srb->cmnd[11]);

	switch (srb->cmnd[0]) {
	case CHECK_MEDIA_TYPE:
		err = sd_ge_check_media_type(cr, card, srb);
		break;
	case MEDIA_SPEC_FUNCTION:
		err = sd_ge_media_spec_function(cr, card, srb);
		break;
	case SECURITY_PROTOCOL_IN:
	case SECURITY_PROTOCOL_OUT:
		err = sd_ge_security_transfer(cr, card, srb);
		break;
	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	return err;
}

#ifdef SUPPORT_MAGIC_GATE
static int mg_report_key(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	int retval;
	u8 key_format;
	struct card *card = cm_locate_card(cr->cm, lun);
	struct ms_info *ms_card;

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	ms_card = card->extra;
	cr_prepare_run(cr);
	ms_cleanup(cr, card);

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (MS_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (srb->cmnd[7] != KC_MG_R_PRO) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	/* The Media is not MSPro, is MemoryStick,
	 * we don't support this type. */
	if (!CHK_MSPRO(ms_card)) {
		set_sense_type(cr, lun, SENSE_TYPE_MG_INCOMPATIBLE_MEDIUM);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	key_format = srb->cmnd[10] & 0x3F;

	switch (key_format) {
	case KF_GET_LOC_EKB:
		if ((scsi_bufflen(srb) == 0x41C) &&
		    (srb->cmnd[8] == 0x04) && (srb->cmnd[9] == 0x1C)) {
			retval = ms_mg_get_local_EKB(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	case KF_RSP_CHG:
		if ((scsi_bufflen(srb) == 0x24) &&
		    (srb->cmnd[8] == 0x00) && (srb->cmnd[9] == 0x24)) {
			retval = ms_mg_get_rsp_chg(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	case KF_GET_ICV:
		ms_card->mg_entry_num = srb->cmnd[5];
		if ((scsi_bufflen(srb) == 0x404) &&
		    (srb->cmnd[8] == 0x04) &&
		    (srb->cmnd[9] == 0x04) &&
		    (srb->cmnd[2] == 0x00) &&
		    (srb->cmnd[3] == 0x00) &&
		    (srb->cmnd[4] == 0x00) && (srb->cmnd[5] < 32)) {
			retval = ms_mg_get_ICV(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	scsi_set_resid(srb, 0);
	return TRANSPORT_GOOD;
}

static int mg_send_key(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	int retval;
	u8 key_format;
	struct card *card = cm_locate_card(cr->cm, lun);
	struct ms_info *ms_card;

	if (!card) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	ms_card = card->extra;
	cr_prepare_run(cr);
	ms_cleanup(cr, card);

	if (!check_card_flag(card, READY)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_NOT_PRESENT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (check_card_flag(card, WP)) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_WRITE_PROTECT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	if (MS_ID != card->id) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_LUN_NOT_SUPPORT);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	if (srb->cmnd[7] != KC_MG_R_PRO) {
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}
	/* The Media is not MSPro, is MemoryStick,
	 * we don't support this type. */
	if (!CHK_MSPRO(ms_card)) {
		set_sense_type(cr, lun, SENSE_TYPE_MG_INCOMPATIBLE_MEDIUM);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	key_format = srb->cmnd[10] & 0x3F;

	switch (key_format) {
	case KF_SET_LEAF_ID:
		if ((scsi_bufflen(srb) == 0x0C) &&
		    (srb->cmnd[8] == 0x00) && (srb->cmnd[9] == 0x0C)) {
			retval = ms_mg_set_leaf_id(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	case KF_CHG_HOST:
		if ((scsi_bufflen(srb) == 0x0C) &&
		    (srb->cmnd[8] == 0x00) && (srb->cmnd[9] == 0x0C)) {
			retval = ms_mg_chg(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	case KF_RSP_HOST:
		if ((scsi_bufflen(srb) == 0x0C) &&
		    (srb->cmnd[8] == 0x00) && (srb->cmnd[9] == 0x0C)) {
			retval = ms_mg_rsp(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	case KF_SET_ICV:
		ms_card->mg_entry_num = srb->cmnd[5];
		if ((scsi_bufflen(srb) == 0x404) &&
		    (srb->cmnd[8] == 0x04) &&
		    (srb->cmnd[9] == 0x04) &&
		    (srb->cmnd[2] == 0x00) &&
		    (srb->cmnd[3] == 0x00) &&
		    (srb->cmnd[4] == 0x00) && (srb->cmnd[5] < 32)) {
			retval = ms_mg_set_ICV(cr, card, srb);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, TRANSPORT_FAILED);
		} else {
			set_sense_type(cr, lun,
				       SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
			TRACE_RET(cr, TRANSPORT_FAILED);
		}
		break;

	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		TRACE_RET(cr, TRANSPORT_FAILED);
	}

	scsi_set_resid(srb, 0);
	return TRANSPORT_GOOD;
}
#endif

static int check_ms_format_progress(struct card_reader *cr, struct scsi_cmnd *srb)
{
	u8 lun = SCSI_LUN(srb);
	struct card *card = cm_locate_card(cr->cm, lun);

	/* NOTE: Once sd_card->sd_erase_status or
	 * ms_card->format_status = FORMAT_IN_PROGRESS is set,
	 * QUIESCENCE of card_flags in struct card should be set as well
	 * with set_card_flags(card, QUIESCENCE); or cleared with
	 * clr_card_flags(card, QUIESCENCE).
	 * Currently, the set/clr operationes to SD&MS card both are not
	 * finished.
	 */
	if (card && MS_ID == card->id) {
		struct ms_info *ms_card = card->extra;

		if ((ms_card->format_status == FORMAT_IN_PROGRESS)) {
			if ((srb->cmnd[0] != REQUEST_SENSE) && (srb->cmnd[0] != INQUIRY)) {
				/* Logical Unit Not Ready Format in Progress */
				set_sense_data(cr, lun, CUR_ERR, 0x02, 0, 0x04, 0x04,
						0, (u16)(ms_card->progress));
				return 1;
			}
		}
	}

	return 0;
}

int scsi_handler(struct scsi_cmnd *srb, struct card_reader *cr)
{
	u8 lun = SCSI_LUN(srb);
	int result = TRANSPORT_GOOD;

	if ((check_ms_format_progress(cr, srb)))
		TRACE_RET(cr, TRANSPORT_FAILED);

	scsi_set_resid(srb, scsi_bufflen(srb));

	switch (srb->cmnd[0]) {
	case INQUIRY:
		result = inquiry(cr, srb);
		break;

	case TEST_UNIT_READY:
		result = test_unit_ready_new(cr, srb);
		break;
	case READ_10:
	case WRITE_10:
	case READ_6:
	case WRITE_6:
		result = read_write(cr, srb);
		break;
	case READ_CAPACITY:
		result = read_capacity(cr, srb);
		break;

	case SERVICE_ACTION_IN_16:
		result = read_capacity16(cr, srb);
		break;

	case START_STOP:
		result = start_stop_unit(cr, srb);
		break;

	case ALLOW_MEDIUM_REMOVAL:
		result = allow_medium_removal(cr, srb);
		break;

	case REQUEST_SENSE:
		result = request_sense(cr, srb);
		break;

	case MODE_SENSE:
	case MODE_SENSE_10:
		result = mode_sense(cr, srb);
		break;

	case READ_FORMAT_CAPACITIES:
		result = read_format_capacity(cr, srb);
		break;

	case VENDOR_CMND:
		result = vendor_cmnd(cr, srb);
		break;

	case MS_SP_CMND:
		result = ms_sp_cmnd(cr, srb);
		break;

	case SD_PASS_THRU_MODE:
	case SD_EXECUTE_NO_DATA:
	case SD_EXECUTE_READ:
	case SD_EXECUTE_WRITE:
	case SD_GET_RSP:
	case SD_HW_RST:
		result = sd_cprm_cmnd(cr, srb);
		break;

	case SECURITY_PROTOCOL_IN:
	case SECURITY_PROTOCOL_OUT:
	case CHECK_MEDIA_TYPE:
	case MEDIA_SPEC_FUNCTION:
		result = sd_ge_cmnd(cr, srb);
		break;

#ifdef SUPPORT_MAGIC_GATE
	case CMD_MSPRO_MG_RKEY:
		result = mg_report_key(cr, srb);
		break;

	case CMD_MSPRO_MG_SKEY:
		result = mg_send_key(cr, srb);
		break;
#endif
	case VERIFY:
		result = verify(cr, srb);
		break;

	case FORMAT_UNIT:
	case MODE_SELECT:
	case MODE_SELECT_10:
		result = TRANSPORT_GOOD;
		break;

	default:
		set_sense_type(cr, lun, SENSE_TYPE_MEDIA_INVALID_CMD_FIELD);
		result = TRANSPORT_FAILED;
		break;
	}

	return result;
}

u8 media_not_present[] = {
	0x70, 0, 0x02, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0x3A, 0, 0, 0, 0, 0 };
u8 invalid_cmd_field[] = {
	0x70, 0, 0x05, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0x24, 0, 0, 0, 0, 0 };

static int check_rw_scsi(struct scsi_cmnd *srb)
{
	if (srb->cmnd[0] == READ_10 || srb->cmnd[0] == WRITE_10)
		return 1;
	if (srb->cmnd[0] == READ_6 || srb->cmnd[0] == WRITE_6)
		return 1;

	return 0;
}

static int is_pre_reading(struct card_reader *cr)
{
	struct card *card = cm_get_ready_card(cr->cm, SD_ID);

	if (card) {
		struct sd_info *sd_card = card->extra;
		if (PRE_READ == sd_card->rw_mode)
			return 1;
	}

	return 0;
}

static int is_pre_writing(struct card_reader *cr)
{
	struct card *card = cm_get_ready_card(cr->cm, SD_ID);

	if (card) {
		struct sd_info *sd_card = card->extra;
		if (PRE_WRITE == sd_card->rw_mode)
			return 1;
	}

	return 0;
}

#ifdef RTS_PM
int need_normal_power(struct scsi_cmnd *srb)
{
	if (srb->cmnd[0] == INQUIRY
		|| srb->cmnd[0] == TEST_UNIT_READY
		|| srb->cmnd[0] == READ_CAPACITY
		|| srb->cmnd[0] == ALLOW_MEDIUM_REMOVAL
		|| srb->cmnd[0] == 0x85
		|| srb->cmnd[0] == REQUEST_SENSE
		|| srb->cmnd[0] == MODE_SENSE
		|| srb->cmnd[0] == MODE_SENSE_10)
		return 0;

	if (srb->cmnd[0] == VENDOR_CMND) {
		if (srb->cmnd[1] == SET_LED) {
			return 1;
		} else if (srb->cmnd[1] == SCSI_APP_CMD) {
			u8 cmd = srb->cmnd[2];
			if (PP_READ10 == cmd
				|| PP_WRITE10 == cmd
				|| RESET_CARD == cmd
				|| RELEASE_CARD == cmd
				|| READ_CSD == cmd
				|| WRITE_CSD == cmd)
				return 1;
		}

		return 0;
	}

	return 1;
}

/**
 * scsi_pm - preprocess the SCSI command due to the PM factor
 *
 * if the system is in RTSX_STAT_SS state, the responds for cmd "TEST_UNIT_READY"
 * and "ALLOW_MEDIUM_REMOVAL" are faked and with -1 as return value, else, return 0.
 */
int scsi_pm(struct scsi_cmnd *srb, struct card_reader *cr)
{
	int ret = 0;

	if (need_normal_power(srb)) {
		/* TODO: how to clear idle_counter?
		 * Maybe should call cr_set_stat(cr, RTSX_STAT_RUN)
		 * directly.
		 */
		cr->idle_counter = 0;
		if (!rtsx_chk_stat(cr, RTSX_STAT_RUN)) {
			/* Update hw to full on status */
			cr_set_stat(cr, RTSX_STAT_RUN);
			cr_pm_update_pm_real(cr);
			/* wake up polling thread firstly. */
			cr_wakeup_polling_thread(cr);
		}
		if (check_rw_scsi(srb)) {
			int pre_read = is_pre_reading(cr);
			int pre_write = is_pre_writing(cr);
			int test_mode = CHECK_STATUS_FLAG(cr, LED_TEST_MODE);

			if ((!pre_read) && (!pre_write) && (!test_mode))
				cr_enable_auto_blink(cr);
		}
	} else if (rtsx_chk_stat(cr, RTSX_STAT_SS)) {
		if (srb->cmnd[0] == TEST_UNIT_READY) {
			if (check_fake_card_ready(cr, SCSI_LUN(srb))) {
				srb->result = SAM_STAT_GOOD;
			} else {
				srb->result = SAM_STAT_CHECK_CONDITION;
				memcpy(srb->sense_buffer,
				       media_not_present, SENSE_SIZE);
			}
		} else if (srb->cmnd[0] == ALLOW_MEDIUM_REMOVAL) {
			int prevent = srb->cmnd[4] & 0x1;
			if (prevent) {
				srb->result = SAM_STAT_CHECK_CONDITION;
				memcpy(srb->sense_buffer,
				       invalid_cmd_field, SENSE_SIZE);
			} else {
				srb->result = SAM_STAT_GOOD;
			}
		}
		ret = -1;
		if (srb->cmnd[0] == INQUIRY ||
				srb->cmnd[0] == TEST_UNIT_READY ||
				srb->cmnd[0] == READ_CAPACITY ||
				srb->cmnd[0] == ALLOW_MEDIUM_REMOVAL ||
				srb->cmnd[0] == REQUEST_SENSE ||
				srb->cmnd[0] == MODE_SENSE ||
				srb->cmnd[0] == MODE_SENSE_10 ||
				srb->cmnd[0] == 0x23 ||
				srb->cmnd[0] == FORMAT_UNIT ||
				srb->cmnd[0] == MODE_SELECT ||
				srb->cmnd[0] == VERIFY ||
				srb->cmnd[0] == VENDOR_CMND)
				ret = 0;


	}

	return ret;
}
#else
int scsi_pm(struct scsi_cmnd *srb, struct card_reader *cr)
{
	return 0;
}

#endif
