/* Driver for Realtek RTS52xx PCIe card reader
 *
 * Copyright(c) 2009 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   wwang (wei_wang@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>

#include "trans_pcie.h"
#include "cardreader.h"
#include "cr_regs.h"
#include "cards.h"
#include "rts_cr.h"


/*
 * tr_pcie_add_sg_tbl - add a sg entry to sg table.
 * @tr: points to the transport struct
 * @addr: address of host DMA buffer to transfer data
 * @len: buffer length in bytes
 * @option: option
 *
 * Add a sg entry to sg table.
 *
 * Note: The length field is 20-bit long. So if the buffer length is
 * longer than 0x80000, this function will divide the buffer into
 * several small buffers to ensure the length field won't overflow.
 */
static inline void tr_pcie_add_sg_tbl(struct transport *tr, u32 addr,
							u32 len, u8 option)
{
	struct rts_pcie *pcie = tr->extra;
	struct cr_option *opt = &tr->cr->option;
	u64 *sgb = (u64 *)(pcie->host_sg_tbl_ptr);
	u64 val = 0;
	u32 temp_len = 0;
	u8  temp_opt = 0;
	u32 max_len = 0x80000;

	if (opt->dma_len_ext_en)
		max_len = 0x2000000;

	do {
		if (len > max_len) {
			temp_len = max_len;
			temp_opt = option & (~SG_END);
		} else {
			temp_len = len;
			temp_opt = option;
		}
		if (opt->dma_len_ext_en)
			val = ((u64)addr << 32) |
				(((u64)temp_len & 0xFFFFF) << 12) |
				((((u64)temp_len & 0x3F00000) >> 20) << 6) |
				temp_opt;
		else
			val = ((u64)addr << 32) | ((u64)temp_len << 12) | temp_opt;

		if (pcie->sgi < (HOST_SG_TBL_BUF_LEN / 8)) {
			sgb[(pcie->sgi)++] = cpu_to_le64(val);
		}

		len -= temp_len;
		addr += temp_len;
	} while (len);
}

/**
 * tr_pcie_transfer_sglist_adma_partial - transfer sg list partially in adma mode
 * @tr: points to the transport struct
 * @card: this command is relevant to card or not
 * @sg: scatter-gather list
 * @num_sg: entry count of sg list
 * @index: next transfer will pick up from which sg entry
 * @offset: next transfer will pick up from the offset in the sg entry
 * @size: transfer size in bytes
 * @dma_dir: transfer direction (DMA_FROM_DEVICE or DMA_TO_DEVICE)
 * @timeout: time out in millisecond
 *
 * Transfer partial data in scatter-gather mode. In this mode,
 * ADMA option will be turned on.
 *
 * This function is usually called in MS card flow. In MS
 * read/write function, one transfer stage will be divided to several stages.
 * The *index and *offset variables are used to record the postion in
 * scatter-gather list that the next transfer will pick up.
 */
static int tr_pcie_transfer_sglist_adma_partial(struct transport *tr,
		u8 card, struct scatterlist *sg, int num_sg,
		unsigned int *index, unsigned int *offset, int size,
		enum dma_data_direction dma_dir, int timeout)
{
	struct card_reader *cr = tr->cr;
	struct rts_pcie *pcie = tr->extra;
	struct completion trans_done;
	u8 dir;
	int sg_cnt, i, resid;
	int err = 0;
	long timeleft;
	struct scatterlist *sg_ptr;
	u32 val = TRIG_DMA;

	if ((sg == NULL) || (num_sg <= 0) || !offset || !index) {
		return -EIO;
	}

	if (dma_dir == DMA_TO_DEVICE) {
		dir = HOST_TO_DEVICE;
	} else if (dma_dir == DMA_FROM_DEVICE) {
		dir = DEVICE_TO_HOST;
	} else {
		return -ENXIO;
	}
#if 0 /* edwin: mark */
	if (card == SD_CARD) {
		rtsx->check_card_cd = SD_EXIST;
	} else if (card == MS_CARD) {
		rtsx->check_card_cd = MS_EXIST;
	} else {
		rtsx->check_card_cd = 0;
	}
#endif
	spin_lock_irq(&pcie->reg_lock);

	/* set up data structures for the wakeup system */
	pcie->done = &trans_done;

	pcie->trans_state = STATE_TRANS_SG;
	pcie->trans_result = TRANS_NOT_READY;

	spin_unlock_irq(&pcie->reg_lock);

	sg_cnt = dma_map_sg(&(pcie->pci_dev->dev), sg, num_sg, dma_dir);

	resid = size;
	sg_ptr = sg;
	pcie->sgi = 0;
	/* Usually the next entry will be @sg@ + 1, but if this sg element is part */
	/* of a chained scatterlist, it could jump to the start of a new */
	/* scatterlist array. So here we use sg_next to move to the proper sg */
	for (i = 0; i < *index; i++) {
		sg_ptr = sg_next(sg_ptr);
	}
	for (i = *index; i < sg_cnt; i++) {
		dma_addr_t addr;
		unsigned int len;
		u8 option;

		addr = sg_dma_address(sg_ptr);
		len = sg_dma_len(sg_ptr);

		rts_trace(TRACE_TR_PCIE,  "DMA addr: 0x%x, Len: 0x%x\n",
						(unsigned int)addr, len);
		rts_trace(TRACE_TR_PCIE,  "*index = %d, *offset = %d\n",
							*index, *offset);

		addr += *offset;

		if ((len - *offset) > resid) {
			*offset += resid;
			len = resid;
			resid = 0;
		} else {
			resid -= (len - *offset);
			len -= *offset;
			*offset = 0;
			*index = *index + 1;
		}
		if ((i == (sg_cnt - 1)) || !resid) {
			option = SG_VALID | SG_END | SG_TRANS_DATA;
		} else {
			option = SG_VALID | SG_TRANS_DATA;
		}

		tr_pcie_add_sg_tbl(tr, (u32)addr, (u32)len, option);

		if (!resid) {
			break;
		}

		sg_ptr = sg_next(sg_ptr);
	}

	rts_trace(TRACE_TR_PCIE, "SG table count = %d\n", pcie->sgi);

	val |= (u32)(dir & 0x01) << 29;
	val |= ADMA_MODE;

	spin_lock_irq(&pcie->reg_lock);

	init_completion(&trans_done);

	cr_iowrite32(cr, HDBAR, pcie->host_sg_tbl_addr);
	cr_iowrite32(cr, HDBCTLR, val);

	spin_unlock_irq(&pcie->reg_lock);

	timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
	if (timeleft <= 0) {
		rts_trace(TRACE_TR_PCIE, "Timeout (%s %d)\n",
						__func__, __LINE__);
		rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n", cr->int_reg);
		err = -ETIMEDOUT;
		goto out;
	}

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_RESULT_FAIL) {
		err = -EIO;
		spin_unlock_irq(&pcie->reg_lock);
		goto out;
	}
	spin_unlock_irq(&pcie->reg_lock);


	/* Wait for TRANS_OK_INT */
	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_NOT_READY) {
		init_completion(&trans_done);
		spin_unlock_irq(&pcie->reg_lock);
		timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
		if (timeleft <= 0) {
			rts_trace(TRACE_TR_PCIE, "Timeout (%s %d)\n",
						__func__, __LINE__);
			rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n",
								cr->int_reg);
			err = -ETIMEDOUT;
			goto out;
		}
	} else {
		spin_unlock_irq(&pcie->reg_lock);
	}

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_RESULT_FAIL) {
		err = -EIO;
	} else if (pcie->trans_result == TRANS_RESULT_OK) {
		err = 0;
	}
	spin_unlock_irq(&pcie->reg_lock);

out:
	pcie->done = NULL;
	pcie->trans_state = STATE_TRANS_NONE;
	dma_unmap_sg(&(pcie->pci_dev->dev), sg, num_sg, dma_dir);

	if (err < 0) {
		cr_stop_cmd(cr, card);
	}

	if (err == -ETIMEDOUT) {
		;/* CATCH_TRIGGER1(chip); */
	}

	return err;
}

/*
 * tr_pcie_transfer_sglist_adma - transfer sg list in adma mode
 * @tr: points to the transport struct
 * @card: this command is relevant to card or not
 * @sg: scatter-gather list
 * @num_sg: entry count of sg list
 * @dma_dir: transfer direction (DMA_FROM_DEVICE or DMA_TO_DEVICE)
 * @timeout: time out in millisecond
 *
 * Transfer data in scatter-gather mode. In this mode, ADMA option will be turned on.
 */
static int tr_pcie_transfer_sglist_adma(struct transport *tr, u8 card,
					struct scatterlist *sg, int num_sg,
				enum dma_data_direction dma_dir, int timeout)
{
	struct card_reader *cr = tr->cr;
	struct rts_pcie *pcie = tr->extra;
	struct completion trans_done;
	u8 dir;
	int buf_cnt, i;
	int err = 0;
	long timeleft;
	struct scatterlist *sg_ptr;

	if ((sg == NULL) || (num_sg <= 0)) {
		return -EIO;
	}

	if (dma_dir == DMA_TO_DEVICE) {
		dir = HOST_TO_DEVICE;
	} else if (dma_dir == DMA_FROM_DEVICE) {
		dir = DEVICE_TO_HOST;
	} else {
		return -ENXIO;
	}
#if 0 /* edwin: mark out */
	if (card == SD_CARD) {
		rtsx->check_card_cd = SD_EXIST;
	} else if (card == MS_CARD) {
		rtsx->check_card_cd = MS_EXIST;
	} else {
		rtsx->check_card_cd = 0;
	}
#endif
	spin_lock_irq(&pcie->reg_lock);

	/* set up data structures for the wakeup system */
	pcie->done = &trans_done;

	pcie->trans_state = STATE_TRANS_SG;
	pcie->trans_result = TRANS_NOT_READY;

	spin_unlock_irq(&pcie->reg_lock);

	buf_cnt = dma_map_sg(&(pcie->pci_dev->dev), sg, num_sg, dma_dir);

	sg_ptr = sg;

	for (i = 0; i <= buf_cnt / (HOST_SG_TBL_BUF_LEN / 8); i++) {
		u32 val = TRIG_DMA;
		int sg_cnt, j;

		if (i == buf_cnt / (HOST_SG_TBL_BUF_LEN / 8)) {
			sg_cnt = buf_cnt % (HOST_SG_TBL_BUF_LEN / 8);
		} else {
			sg_cnt = (HOST_SG_TBL_BUF_LEN / 8);
		}

		pcie->sgi = 0;
		for (j = 0; j < sg_cnt; j++) {
			dma_addr_t addr = sg_dma_address(sg_ptr);
			unsigned int len = sg_dma_len(sg_ptr);
			u8 option;

			rts_trace(TRACE_TR_PCIE, "DMA addr: 0x%x, Len: 0x%x\n",
						(unsigned int)addr, len);

			if (j == (sg_cnt - 1)) {
				option = SG_VALID | SG_END | SG_TRANS_DATA;
			} else {
				option = SG_VALID | SG_TRANS_DATA;
			}

			tr_pcie_add_sg_tbl(tr, (u32)addr, (u32)len, option);

			sg_ptr = sg_next(sg_ptr);
		}

		rts_trace(TRACE_TR_PCIE, "SG table count = %d\n", pcie->sgi);

		val |= (u32)(dir & 0x01) << 29;
		val |= ADMA_MODE;

		spin_lock_irq(&pcie->reg_lock);

		init_completion(&trans_done);

		cr_iowrite32(cr, HDBAR, pcie->host_sg_tbl_addr);
		cr_iowrite32(cr, HDBCTLR, val);

		spin_unlock_irq(&pcie->reg_lock);

		timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
		if (timeleft <= 0) {
			rts_trace(TRACE_TR_PCIE, "Timeout (%s %d)\n",
						__func__, __LINE__);
			rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n",
								cr->int_reg);
			err = -ETIMEDOUT;
			goto out;
		}

		spin_lock_irq(&pcie->reg_lock);
		if (pcie->trans_result == TRANS_RESULT_FAIL) {
			err = -EIO;
			spin_unlock_irq(&pcie->reg_lock);
			goto out;
		}
		spin_unlock_irq(&pcie->reg_lock);

		sg_ptr += sg_cnt;
	}

	/* Wait for TRANS_OK_INT */
	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_NOT_READY) {
		init_completion(&trans_done);
		spin_unlock_irq(&pcie->reg_lock);
		timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
		if (timeleft <= 0) {
			rts_trace(TRACE_TR_PCIE, "Timeout (%s %d)\n",
						__func__, __LINE__);
			rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n",
								cr->int_reg);
			err = -ETIMEDOUT;
			goto out;
		}
	} else {
		spin_unlock_irq(&pcie->reg_lock);
	}

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_RESULT_FAIL) {
		err = -EIO;
	} else if (pcie->trans_result == TRANS_RESULT_OK) {
		err = 0;
	}
	spin_unlock_irq(&pcie->reg_lock);

out:
	pcie->done = NULL;
	pcie->trans_state = STATE_TRANS_NONE;
	dma_unmap_sg(&(pcie->pci_dev->dev), sg, num_sg, dma_dir);

	if (err < 0) {
		cr_stop_cmd(cr, card);
	}

	if (err == -ETIMEDOUT) {
		;/* CATCH_TRIGGER1(chip); */
	}

	return err;
}

/*
 * tr_pcie_transfer_buf - transfer data in linear buffer.
 * @tr: points to the transport struct
 * @card: this command is relevant to card or not
 * @buf: data buffer
 * @len: buffer length
 * @dma_dir: transfer direction (DMA_FROM_DEVICE or DMA_TO_DEVICE)
 * @timeout: time out in millisecond
 *
 * Transfer data in linear buffer.
 */
static int tr_pcie_transfer_buf(struct transport *tr, u8 card,
	void *buf, size_t len, enum dma_data_direction dma_dir, int timeout)
{
	struct card_reader *cr = tr->cr;
	struct rts_pcie *pcie = tr->extra;
	struct completion trans_done;
	dma_addr_t addr;
	u8 dir;
	int err = 0;
	u32 val = (1 << 31);
	long timeleft;

	if ((buf == NULL) || (len <= 0)) {
		return -EIO;
	}

	if (dma_dir == DMA_TO_DEVICE) {
		dir = HOST_TO_DEVICE;
	} else if (dma_dir == DMA_FROM_DEVICE) {
		dir = DEVICE_TO_HOST;
	} else {
		return -ENXIO;
	}

	addr = dma_map_single(&(pcie->pci_dev->dev), buf, len, dma_dir);
	if (!addr) {
		return -ENOMEM;
	}
#if 0 /* edwin: mark out */
	if (card == SD_CARD) {
		rtsx->check_card_cd = SD_EXIST;
	} else if (card == MS_CARD) {
		rtsx->check_card_cd = MS_EXIST;
	} else {
		rtsx->check_card_cd = 0;
	}
#endif
	val |= (u32)(dir & 0x01) << 29;
	val |= (u32)(len & 0x00FFFFFF);

	spin_lock_irq(&pcie->reg_lock);

	/* set up data structures for the wakeup system */
	pcie->done = &trans_done;

	init_completion(&trans_done);

	pcie->trans_state = STATE_TRANS_BUF;
	pcie->trans_result = TRANS_NOT_READY;

	cr_iowrite32(cr, HDBAR, addr);
	cr_iowrite32(cr, HDBCTLR, val);

	spin_unlock_irq(&pcie->reg_lock);

	/* Wait for TRANS_OK_INT */
	timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
	if (timeleft <= 0) {
		rts_trace(TRACE_TR_PCIE, "Timeout (%s %d)\n",
						__func__, __LINE__);
		rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n", cr->int_reg);
		err = -ETIMEDOUT;
		goto out;
	}

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_RESULT_FAIL) {
		err = -EIO;
	} else if (pcie->trans_result == TRANS_RESULT_OK) {
		err = 0;
	}
	spin_unlock_irq(&pcie->reg_lock);

out:
	pcie->done = NULL;
	pcie->trans_state = STATE_TRANS_NONE;
	dma_unmap_single(&(pcie->pci_dev->dev), addr, len, dma_dir);

	if (err < 0) {
		cr_stop_cmd(cr, card);
	}

	if (err == -ETIMEDOUT) {
		;/* CATCH_TRIGGER1(chip); */
	}

	return err;
}


/*
 * tr_pcie_transfer_data_partial - partially transfer data in linear buffer or adma mode
 * @tr: points to the transport struct
 * @card: this command is relevant to card or not
 * @buf: data buffer
 * @len: buffer length
 * @use_sg: use scatter-gather mode or not
 * @index: next transfer will pick up from which sg entry
 * @offset: next transfer will pick up from the offset in the sg entry
 * @dma_dir: transfer direction (DMA_FROM_DEVICE or DMA_TO_DEVICE)
 * @timeout: time out in millisecond
 */
int tr_pcie_transfer_data_partial(struct transport *tr, u8 card,
	void *buf, size_t len, int use_sg, unsigned int *index,
	unsigned int *offset, enum dma_data_direction dma_dir, int timeout)
{
	struct card_reader *cr = tr->cr;
	int err = 0;

	/* don't transfer data during abort processing */
	if (rtsx_chk_stat(cr, RTSX_STAT_ABORT)) {
		return -EIO;
	}

	if (use_sg) {
		err = tr_pcie_transfer_sglist_adma_partial(tr, card,
			(struct scatterlist *)buf, use_sg, index, offset,
						(int)len, dma_dir, timeout);
	} else {
		err = tr_pcie_transfer_buf(tr, card, buf, len,
							dma_dir, timeout);
	}

	if (err < 0) {
		if (RTSX_TST_DELINK(cr)) {
			RTSX_CLR_DELINK(cr);
			cr->need_reinit = SD_CARD | MS_CARD;
		}
	}

	return err;
}

/*
 * tr_pcie_transfer_data_partial - transfer data in linear buffer or adma mode
 * @tr: points to the transport struct
 * @card: this command is relevant to card or not
 * @buf: data buffer
 * @len: buffer length
 * @use_sg: use scatter-gather mode or not
 * @dma_dir: transfer direction (DMA_FROM_DEVICE or DMA_TO_DEVICE)
 * @timeout: time out in millisecond
 */
int tr_pcie_transfer_data(struct transport *tr, u8 card, void *buf,
	size_t len, int use_sg, enum dma_data_direction dma_dir, int timeout)
{
	struct card_reader *cr = tr->cr;
	int err = 0;

	rts_trace(TRACE_TR_PCIE, "use_sg = %d\n", use_sg);

	/* don't transfer data during abort processing */
	if (rtsx_chk_stat(cr, RTSX_STAT_ABORT)) {
		return -EIO;
	}

	if (use_sg) {
		err = tr_pcie_transfer_sglist_adma(tr, card,
			(struct scatterlist *)buf, use_sg, dma_dir, timeout);
	} else {
		err = tr_pcie_transfer_buf(tr, card, buf, len,
							dma_dir, timeout);
	}

	if (err < 0) {
		if (RTSX_TST_DELINK(cr)) {
			RTSX_CLR_DELINK(cr);
			cr->need_reinit = SD_CARD | MS_CARD;
		}
	}

	return err;
}

/*
 * tr_pcie_get_cmd_data - get commod data
 * @tr: points to the transport struct
 *
 * Returns: return pointer of commands
 */
u8 *tr_pcie_get_cmd_data(struct transport *tr)
{
#ifdef CMD_USING_SG
	struct rts_pcie *pcie = tr->extra;

	return (u8 *)(pcie->host_sg_tbl_ptr);
#else
	struct rts_pcie *pcie = tr->extra;

	return (u8 *)(pcie->host_cmds_ptr);
#endif
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
#define pci_get_bus_and_slot(bus, devfn)	\
	pci_get_domain_bus_and_slot(0, (bus), (devfn))
#endif

/*
 * tr_pci_rd_cfg_byte - Read PCIe configuration
 *
 */
int tr_pci_rd_cfg_byte(struct transport *tr, u8 bus, u8 dev, u8 func, u8 offset, u8 *val)
{
	struct pci_dev *pdev;
	u8 data;
	u8 devfn = (dev << 3) | func;

	pdev = pci_get_bus_and_slot(bus, devfn);
	if (!dev) {
		return -1;
	}

	pci_read_config_byte(pdev, offset, &data);
	if (val) {
		*val = data;
	}

	return 0;
}


static int tr_pcie_wr_reg(struct transport *tr, u16 addr, u8 mask, u8 data)
{
	int i;
	u32 val = 3 << 30;

	val |= (u32)(addr & 0x3FFF) << 16;
	val |= (u32)mask << 8;
	val |= (u32)data;

	tr_iowrite32(tr, HAIMR, val);

	for (i = 0; i < MAX_RW_REG_CNT; i++) {
		val = tr_ioread32(tr, HAIMR);
		if ((val & (1 << 31)) == 0) {
			if (data != (u8)val) {
				/* TRACE_RET(cr, STATUS_FAIL); */
				rts_trace(TRACE_ERROR, "write register fail\n");
				return STATUS_FAIL;
			}
			return STATUS_SUCCESS;
		}

		if (tr->cr->option.delay_for_rw_reg)
			mdelay(tr->cr->option.delay_for_rw_reg);
	}

	/* TRACE_RET(cr, STATUS_TIMEDOUT); */
	rts_trace(TRACE_ERROR, "write register timeout\n");
	return STATUS_TIMEDOUT;
}

static int tr_pcie_rd_reg(struct transport *tr, u16 addr, u8 *data)
{
	int i;
	u32 val = 2 << 30;

	if (data)
		*data = 0;

	val |= (u32)(addr & 0x3FFF) << 16;
	tr_iowrite32(tr, HAIMR, val);

	for (i = 0; i < MAX_RW_REG_CNT; i++) {
		val = tr_ioread32(tr, HAIMR);
		if ((val & (1 << 31)) == 0)
			break;
		if (tr->cr->option.delay_for_rw_reg)
			mdelay(tr->cr->option.delay_for_rw_reg);
	}

	if (i >= MAX_RW_REG_CNT) {
		/* TRACE_RET(cr, STATUS_TIMEDOUT); */
		rts_trace(TRACE_ERROR, "read register timeout\n");
		return STATUS_TRANS_SHORT;
	}

	if (data)
		*data = (u8)(val & 0xFF);

	return STATUS_SUCCESS;
}

inline void tr_pcie_init_cmd(struct transport *tr)
{
	struct rts_pcie *pcie = tr->extra;
	pcie->cmd_index = 0;
}

/**
 * tr_pcie_add_cmd - add a command to command buffer.
 * @chip: Realtek's card reader chip
 * @cmd_type: command type, including read/write/check register
 * @reg_addr: internal card controller register address
 * @mask: bit mask
 * @data: register data
 *
 * Add a command to command buffer.
 *
 * Usually, this function is called after tr_pcie_init_cmd, which
 * intializes the command index to zero. After all commands are added,
 * tr_pcie_send_cmd or tr_pcie_send_cmd_no_wait should be called to send those
 * commands to card reader chip.
 */
void tr_pcie_add_cmd(struct transport *tr, u8 cmd_type, u16 reg_addr,
							u8 mask, u8 data)
{
	struct rts_pcie *pcie = tr->extra;
	u32 *cb = (u32 *)(pcie->host_cmds_ptr);
	u32 val = 0;

	val |= (u32)(cmd_type & 0x03) << 30;
	val |= (u32)(reg_addr & 0x3FFF) << 16;
	val |= (u32)mask << 8;
	val |= (u32)data;

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->cmd_index < (HOST_CMDS_BUF_LEN / 4))
		cb[(pcie->cmd_index)++] = cpu_to_le32(val);
	spin_unlock_irq(&pcie->reg_lock);
}

/**
 * tr_pcie_send_cmd_no_wait - send commands to chip.
 * @chip: Realtek's card reader chip
 *
 * Trigger card reader chip to fetch commands from command buffer.
 * This funtion returns immediately.
 */
void tr_pcie_send_cmd_no_wait(struct transport *tr)
{
	struct rts_pcie *pcie = tr->extra;

	u32 val = 1 << 31;

	tr_iowrite32(tr, HCBAR, pcie->host_cmds_addr);

	val |= (u32)(pcie->cmd_index * 4) & 0x00FFFFFF;
	/* Hardware Auto Response */
	val |= 0x40000000;
	tr_iowrite32(tr, HCBCTLR, val);
}

/**
 * tr_pcie_stop_cmd - stop command transfer and DMA transfer
 * @chip: Realtek's card reader chip
 * @card: flash card type
 *
 * Stop command transfer and DMA transfer.
 * This function is called in error handler.
 */
void tr_pcie_stop_cmd(struct transport *tr, int card)
{
	int i;

	for (i = 0; i <= 8; i++) {
		int addr = HCBAR + i * 4;
		u32 reg;
		reg = tr_ioread32(tr, addr);
		rts_trace(TRACE_TR_PCIE, "BAR (0x%02x): 0x%08x\n", addr, reg);
	}

	tr_iowrite32(tr, HCBCTLR, STOP_CMD);
	tr_iowrite32(tr, HDBCTLR, STOP_DMA);

	for (i = 0; i < 16; i++) {
		u16 addr = 0xFE20 + (u16)i;
		u8 val;
		tr_pcie_rd_reg(tr, addr, &val);
		rts_trace(TRACE_TR_PCIE, "0x%04X: 0x%02x\n", addr, val);
	}

	tr_pcie_wr_reg(tr, REG_DMACTL, 0x80, 0x80);
	tr_pcie_wr_reg(tr, REG_RBCTL, 0x80, 0x80);
}

/**
 * tr_pcie_send_cmd - send commands to chip.
 * @chip: Realtek's card reader cr
 * @card: this command is relevant to card or not
 * @timeout: time out in millisecond
 *
 * Trigger card reader cr to fetch commands from command buffer.
 * This funtion will wait for transfer-finished interrupt.
 *
 * Returns zero if successful, or a negative error code on failure.
 */
int tr_pcie_send_cmd(struct transport *tr, u8 card, int timeout)
{
	struct rts_pcie *pcie = tr->extra;
	struct completion trans_done;
	u32 val = 1 << 31;
	long timeleft;
	int err = 0;
#if 0 /* edwin: comment out */
	if (card == SD_CARD) {
		rts5227->check_card_cd = SD_EXIST;
	} else if (card == MS_CARD) {
		rts5227->check_card_cd = MS_EXIST;
	} else {
		rts5227->check_card_cd = 0;
	}
#endif
	spin_lock_irq(&pcie->reg_lock);

	/* set up data structures for the wakeup system */
	pcie->done = &trans_done;
	pcie->trans_result = TRANS_NOT_READY;
	init_completion(&trans_done);
	pcie->trans_state = STATE_TRANS_CMD;

	tr_iowrite32(tr, HCBAR, pcie->host_cmds_addr);

	val |= (u32)(pcie->cmd_index * 4) & 0x00FFFFFF;
	/* Hardware Auto Response */
	val |= 0x40000000;
	tr_iowrite32(tr, HCBCTLR, val);

	spin_unlock_irq(&pcie->reg_lock);

	rts_trace(TRACE_TR_PCIE, "waiting .......cr->int_reg = 0x%x\n",
			tr_ioread32(tr, BIPR));

	/* Wait for TRANS_OK_INT */
	timeleft = wait_for_completion_interruptible_timeout(&trans_done,
							timeout * HZ / 1000);
	if (timeleft <= 0) {
		rts_trace(TRACE_TR_PCIE, "cr->int_reg = 0x%x\n", tr_ioread32(tr, BIPR));
		err = -ETIMEDOUT;
		/* TRACE_GOTO(cr, finish_send_cmd); */
		rts_trace(TRACE_ERROR, "send command timeout\n");
		goto finish_send_cmd;
	}

	spin_lock_irq(&pcie->reg_lock);
	if (pcie->trans_result == TRANS_RESULT_FAIL)
		err = -EIO;
	else if (pcie->trans_result == TRANS_RESULT_OK)
		err = 0;
	spin_unlock_irq(&pcie->reg_lock);
finish_send_cmd:
	pcie->done = NULL;
	pcie->trans_state = STATE_TRANS_NONE;

	if (err < 0)
		tr_pcie_stop_cmd(tr, card);

	return err;
}


/*
 * init_transport_pcie - Main entry to initialise PCIe transport
 */
int init_transport_pcie(struct transport *tr, void *extra)
{
	struct pci_dev *pci_dev = extra;
	struct rts_pcie *pcie = tr->extra;
	int err = 0;

	rts_trace(TRACE_TR_PCIE, "pci_dev->devfn: %d\n", pci_dev->devfn);
	err = pci_enable_device(pci_dev);
	if (err < 0) {
		printk(KERN_ERR "PCI enable device failed!\n");
		return err;
	}
	pci_set_master(pci_dev);

	err = pci_request_regions(pci_dev, CR_DRIVER_NAME);
	if (err < 0) {
		printk(KERN_ERR "PCI request regions for %s failed!\n",
			CR_DRIVER_NAME);
		pci_disable_device(pci_dev);
		return err;
	}

	spin_lock_init(&pcie->reg_lock);
	pcie->pci_dev = pci_dev;
	pcie->irq = -1;

	rts_trace(TRACE_TR_PCIE, "Resource length: 0x%x\n",
		(unsigned int)pci_resource_len(pci_dev, 0));

	/* using BAR1 if BAR1 exist  */
	if (0 != pci_resource_start(pci_dev, 1)) {
		pcie->addr = pci_resource_start(pci_dev, 1);
		pcie->remap_addr = ioremap_nocache(pcie->addr,
			pci_resource_len(pci_dev, 1));
	} else {
		pcie->addr = pci_resource_start(pci_dev, 0);
		pcie->remap_addr = ioremap_nocache(pcie->addr,
			pci_resource_len(pci_dev, 0));
	}
	if (pcie->remap_addr == NULL) {
		printk(KERN_ERR "ioremap error\n");
		err = -ENXIO;
		return err;
	}

	/* Using "unsigned long" cast here to eliminate gcc warning in 64-bit system */
	rts_trace(TRACE_TR_PCIE, "Original address: 0x%lx, remapped address:"
				"0x%lx\n", (unsigned long)(pcie->addr),
				(unsigned long)(pcie->remap_addr));

	pcie->rtsx_resv_buf = dma_alloc_coherent(&(pci_dev->dev),
		RTSX_RESV_BUF_LEN, &(pcie->rtsx_resv_buf_addr), GFP_KERNEL);
	if (pcie->rtsx_resv_buf == NULL) {
		printk(KERN_ERR "alloc dma buffer fail\n");
		err = -ENXIO;
		return err;
	}

	tr->iobuf = pcie->rtsx_resv_buf;

	pcie->func_no = pcie->pci_dev->devfn;
	pcie->pcie_start_addr = pcie->addr;
	pcie->pcie_start_remap_addr = pcie->remap_addr;
	pcie->host_cmds_ptr = pcie->rtsx_resv_buf;
	pcie->host_cmds_addr = pcie->rtsx_resv_buf_addr;
	pcie->host_sg_tbl_ptr = pcie->rtsx_resv_buf + HOST_CMDS_BUF_LEN;
	pcie->host_sg_tbl_addr = pcie->rtsx_resv_buf_addr + HOST_CMDS_BUF_LEN;

	return err;
}

/*
 * release_transport_pcie - Main entry to release PCIe transport
 */
int release_transport_pcie(struct transport *tr)
{
	struct rts_pcie *pcie = tr->extra;

	rts_trace(TRACE_TR_PCIE, "<--\n");
#if 1
	if (pcie->rtsx_resv_buf) {
		dma_free_coherent(&(pcie->pci_dev->dev), RTSX_RESV_BUF_LEN,
				pcie->rtsx_resv_buf, pcie->rtsx_resv_buf_addr);
		/* pcie->host_cmds_ptr = NULL; */
		/* pcie->host_sg_tbl_ptr = NULL; */
	}

	cr_free_irq(tr->cr);

	if (pcie->remap_addr)
		iounmap(pcie->remap_addr);

	pci_disable_device(pcie->pci_dev);
	pci_release_regions(pcie->pci_dev);
#endif
	rts_trace(TRACE_TR_PCIE, "-->\n");

	return 0;
}

inline int tr_pcie_enable_msi(struct transport *tr)
{
	struct rts_pcie *pcie = tr->extra;

	return pci_enable_msi(pcie->pci_dev);
}

inline void tr_pcie_synchronize_irq(struct transport *tr)
{
	struct rts_pcie *pcie = tr->extra;
	synchronize_irq(pcie->irq);
}

int tr_pci_find_capablility(struct transport *tr, int cap)
{
	struct rts_pcie *pcie = tr->extra;

	return pci_find_capability(pcie->pci_dev, cap);
}

inline int tr_pci_find_host_capablility(struct transport *tr, int cap)
{
	struct rts_pcie *pcie = tr->extra;
	struct pci_dev *bridge = cr_pci_upstream_bridge(pcie->pci_dev);

	return pci_find_capability(bridge, cap);
}
