/* Driver for Realtek card reader
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/module.h>

#include "trace.h"
#include "transport.h"
#include "trans_pcie.h"


/* deprecated */
struct transport realtek_transport_sets[REALTEK_MAX_TRANSPORTS] = {
	[REALTEK_TRANSPORT_PCIE] = {
		.name = "PCIe",
		.cb = {
			.init = init_transport_pcie,
			.release = release_transport_pcie,
		},
	},
};
EXPORT_SYMBOL_GPL(realtek_transport_sets);

int tr_transfer_data(struct transport *tr, u8 card, void *buf,
	size_t len, int use_sg, int dir, int timeout)
{
	return tr_pcie_transfer_data(tr, card, buf, len, use_sg, dir, timeout);

}

int tr_transfer_data_partial(struct transport *tr, u8 card,
	void *buf, size_t len, int use_sg, unsigned int *index,
	unsigned int *offset, int dir, int timeout)
{

	return tr_pcie_transfer_data_partial(tr, card, buf, len, use_sg, index,
			offset, dir, timeout);
}

inline u8 *tr_get_cmd_data(struct transport *tr)
{
	return tr_pcie_get_cmd_data(tr);

}

/* transport_alloc - Allocate struct transport
 * @preserve_size: extra size
 *
 * Returns: return address of struct transport
 * Description: if preserve_size is not equal to zero, and allocation is done successfully
 * the tr->extra will be assigned with the address of the preserved area
 */
struct transport *transport_alloc(int preserve_size)
{
	struct transport *tr = NULL;

	rts_trace(TRACE_TR, "<--\n");

	if (0 == preserve_size) {
		printk("extra size is ZERO for transport\n");
		goto err_alloc_transport;
	}

	tr = kzalloc(sizeof(struct transport), GFP_KERNEL);
	if (NULL == tr) {
		printk("ERROR on malloc for transport\n");
		goto err_alloc_transport;
	}

	tr->extra = kzalloc(preserve_size, GFP_KERNEL);
	if (NULL == tr->extra) {
		printk("ERROR on malloc for extra\n");
		goto err_alloc_extra;
	}

	rts_trace(TRACE_TR, "tr=%p, tr->extra=%p\n", tr, tr->extra);

	rts_trace(TRACE_TR, "-->\n");
	return tr;

err_alloc_extra:
	kfree(tr);
	tr = NULL;
err_alloc_transport:
	return tr;
}

/* transport_init - Initialise struct transport and alloc hw resource
 * @tr: transport structure
 * @tr_capa: transport configuration
 *
 * Returns: 0 for success, otherwise nonzero
 */
int transport_init(struct transport *tr, struct transport *tr_config,
		void *extra)
{
	if (!tr)
		return -EINVAL;

	mutex_init(&tr->trans_mutex);

	/* Now only support PCIe */
	rts_trace(TRACE_TR, "init PCIe transport\n");
	tr->name = "PCIe";
	return init_transport_pcie(tr, extra);
}

/* transport_release - Release resource allocated to transport
 * @tr: transport structure
 *
 * Returns: Nothing
 */
void transport_release(struct transport *tr)
{
	rts_trace(TRACE_TR, "release PCIe transport\n");

	release_transport_pcie(tr);

	if (tr) {
		if (tr->extra)
			kfree(tr->extra);
		kfree(tr);
	}
}

void tr_enable_host_aspm(struct transport *tr)
{
	/* TODO: */
}

void tr_disable_host_aspm(struct transport *tr)
{

	/* TODO: */
}

void tr_set_host_aspm(struct transport *tr, u8 val)
{

	/* TODO: */
}

void tr_get_host_aspm(struct transport *tr, u8 *val)
{
	/* TODO: */

}

inline void tr_notify_upper(struct transport *tr, void *arg, int val)
{

	/* TODO: */
}

#define TR_PCIE(tr) ((struct rts_pcie *)(tr->extra))

inline u32 tr_ioread32(struct transport *tr, u32 reg)
{
	return  ioread32(TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline void tr_iowrite32(struct transport *tr, u32 reg, u32 value)
{
	iowrite32(value, TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline u16 tr_ioread16(struct transport *tr, u32 reg)
{
	return  ioread16(TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline void tr_iowrite16(struct transport *tr, u32 reg, u16 value)
{
	iowrite16(value, TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline u8 tr_ioread8(struct transport *tr, u32 reg)
{
	return  ioread8(TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline void tr_iowrite8(struct transport *tr, u32 reg, u8 value)
{
	iowrite8(value, TR_PCIE(tr)->pcie_start_remap_addr + reg);
}

inline int tr_read_config8(struct transport *tr, int reg,  u8 *value)
{
	return pci_read_config_byte(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_read_config16(struct transport *tr, int reg,  u16 *value)
{
	return pci_read_config_word(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_read_config32(struct transport *tr, int reg,  u32 *value)
{
	return pci_read_config_dword(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_read_host_config8(struct transport *tr, int reg, u8 *value)
{
	struct pci_dev *bridge = cr_pci_upstream_bridge(TR_PCIE(tr)->pci_dev);
	if (!bridge)
		return -ENODEV;
	return pci_read_config_byte(bridge, reg, value);
}

inline int tr_write_config8(struct transport *tr, int reg, u8 value)
{
	return pci_write_config_byte(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_write_config16(struct transport *tr, int reg, u16 value)
{
	return pci_write_config_word(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_write_config32(struct transport *tr, int reg, u32 value)
{
	return pci_write_config_dword(TR_PCIE(tr)->pci_dev, reg, value);
}

inline int tr_write_host_config8(struct transport *tr, int reg, u8 value)
{
	struct pci_dev *bridge = cr_pci_upstream_bridge(TR_PCIE(tr)->pci_dev);
	if (!bridge)
		return -ENODEV;
	return pci_write_config_byte(bridge, reg, value);
}

inline void tr_init_cmd(struct transport *tr)
{
	return tr_pcie_init_cmd(tr);
}

inline void tr_add_cmd(struct transport *tr, u8 type, u16 addr, u8 mask, u8 data)
{
	return tr_pcie_add_cmd(tr, type, addr, mask, data);
}


inline void tr_send_cmd_no_wait(struct transport *tr)
{
	tr_pcie_send_cmd_no_wait(tr);
}

inline void tr_stop_cmd(struct transport *tr, int card)
{
	tr_pcie_stop_cmd(tr, card);
}

inline int tr_send_cmd(struct transport *tr, u8 card, int timeout)
{
	return tr_pcie_send_cmd(tr, card, timeout);
}

inline int tr_enable_msi(struct transport *tr)
{
	return tr_pcie_enable_msi(tr);
}

inline void tr_synchronize_irq(struct transport *tr)
{
	tr_pcie_synchronize_irq(tr);
}

inline int tr_find_pci_capability(struct transport *tr, int cap)
{
	return tr_pci_find_capablility(tr, cap);
}

inline int tr_find_host_pci_capability(struct transport *tr, int cap)
{
	return tr_pci_find_host_capablility(tr, cap);
}
