/* Driver for Realtek PCIe card reader
 *
 * Copyright(c) 2013 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>
#include <linux/dma-mapping.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/pm.h>

#include "cr_regs.h"
#include "pm.h"
#include "trace.h"
#include "trans_pcie.h"
#include "rts_cr.h"
#include "cards/sd.h"
#include "cardreader.h"
#include "cr_driver.h"
#include "cr_sysfs.h"
#include "rts_rpm.h"

extern struct cr_capability rts5227_capability;
extern struct cr_capability rtl8411b_capability;
extern struct cr_capability rts5249_capability;
extern struct cr_capability rts5250_capability;
extern struct cr_capability rts522A_capability;
extern struct cr_capability rts524A_capability;
extern struct cr_capability rts525A_capability;

#define rtsx_pci_id(id) \
	{0x10EC, (id), PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_OTHERS << 16, 0xFF0000}
#define rtsx_pci_id_class(id, class, class_mask)			\
	{0x10EC, (id), PCI_ANY_ID, PCI_ANY_ID, (class), (class_mask)}

static struct pci_device_id cr_pcie_ids[] = {
	rtsx_pci_id(0x5227),
	rtsx_pci_id(0x5287),
	rtsx_pci_id(0x5249),
	rtsx_pci_id(0x522A),
	rtsx_pci_id(0x524A),
	rtsx_pci_id(0x525A),
	rtsx_pci_id_class(0x5250, PCI_CLASS_SYSTEM_SDHCI << 8, 0xFFFF00),
	{ 0, },
};

static int cr_pm_suspend(struct device *dev)
{
	struct pci_dev *pci = to_pci_dev(dev);
	struct card_reader *cr = pci_get_drvdata(pci);

	rts_trace(TRACE_PM, "pm suspend\n");

	/* lock the device pointers */
	mutex_lock(&cr->tr->trans_mutex);

	cr_pm_bef_pow_dwn(cr, PM_S3);

	cr_free_irq(cr);

	pci_save_state(pci);
	pci_enable_wake(pci, PCI_D3hot, 1);
	pci_disable_device(pci);
	pci_set_power_state(pci, PCI_D3hot);

	mutex_unlock(&cr->tr->trans_mutex);

	return 0;
}

static int cr_pm_resume(struct device *dev)
{
	struct pci_dev *pci = to_pci_dev(dev);
	struct card_reader *cr = pci_get_drvdata(pci);

	rts_trace(TRACE_PM, "pm resume\n");

	/* lock the device pointers */
	mutex_lock(&cr->tr->trans_mutex);

	pci_set_power_state(pci, PCI_D0);
	pci_restore_state(pci);
	if (pci_enable_device(pci) < 0) {
		rts_trace(TRACE_PM,
			"%s: pci_enable_device failed, disabling device\n",
			CR_DRIVER_NAME);
		mutex_unlock(&cr->tr->trans_mutex);
		return -EIO;
	}
	pci_set_master(pci);

	if (cr_pm_do_resume(cr) < 0) {
		mutex_unlock(&cr->tr->trans_mutex);
		return -EIO;
	}

	cr_wakeup_polling_thread(cr);

	mutex_unlock(&cr->tr->trans_mutex);

	return 0;
}

static int cr_pm_poweroff(struct device *dev)
{
	struct pci_dev *pci = to_pci_dev(dev);
	struct card_reader *cr = pci_get_drvdata(pci);

	rts_trace(TRACE_PM, "pm poweroff\n");

	cr_pm_bef_pow_dwn(cr, PM_S1);

	cr_free_irq(cr);

	pci_disable_device(pci);

	return 0;
}

static int cr_runtime_suspend(struct device *dev)
{
	struct card_reader *cr = pci_get_drvdata(to_pci_dev(dev));
	struct transport *tr = cr->tr;

	rts_trace(TRACE_PM, "usage_count = %d, state = %d\n",
		dev_power_ucount(dev), cr->rtsx_stat);

	if (CHECK_STATUS_FLAG(cr, ENTERED_RTD3)) {
		rts_trace(TRACE_PM, "device is already suspend...\n");
		return 0;
	}

	mutex_lock(&tr->trans_mutex);
	SET_STATUS_FLAG(cr, ENTERED_RTD3);
	cm_status_runtime_save(cr->cm);
	cr_pm_bef_pow_dwn(cr, PM_S3);

	cr_free_irq(cr);

	rts_trace(TRACE_PM, "runtime suspend success\n");
	mutex_unlock(&tr->trans_mutex);

	return 0;
}

static int cr_runtime_resume(struct device *dev)
{
	struct card_reader *cr = pci_get_drvdata(to_pci_dev(dev));
	struct transport *tr = cr->tr;

	rts_trace(TRACE_PM, "usage_count = %d, state = %d\n",
		dev_power_ucount(dev), cr->rtsx_stat);

	if (!CHECK_STATUS_FLAG(cr, ENTERED_RTD3)) {
		rts_trace(TRACE_PM, "device is already running...");
		return 0;
	}

	mutex_lock(&tr->trans_mutex);
	cr_set_stat(cr, RTSX_STAT_RUN);
	CLEAR_STATUS_FLAG(cr, ENTERED_RTD3);
	cm_status_runtime_clear(cr->cm);
	cr_pm_do_resume(cr);
	cr_polling_func(cr);
	rts_trace(TRACE_PM,
		"-------------------- RTD3 resumed --------------------\n");
	mutex_unlock(&tr->trans_mutex);

	return 0;
}

static struct dev_pm_ops cr_pm_ops = {
	.suspend = cr_pm_suspend,
	.resume = cr_pm_resume,
	.poweroff = cr_pm_poweroff,
	.freeze = cr_pm_suspend,
	.restore = cr_pm_resume,
	.runtime_idle = cr_runtime_idle,
	.runtime_suspend = cr_runtime_suspend,
	.runtime_resume = cr_runtime_resume,
};

struct cr_driver cr_pcie_driver = {
	.pci_driver = {
		.name = "Realtek PCIe card reader",
		.id_table = cr_pcie_ids,
		.probe = rts_cr_pcie_probe,
		.remove = rts_cr_pcie_remove,
		.driver.groups = cr_driver_attr_groups,
		.driver.pm = &cr_pm_ops,
	},
};

static void cr_init_sysfs_param(void)
{
	cr_init_generic_sw_param(&cr_pcie_driver.sw_param);
	cr_init_generic_hw_param(&cr_pcie_driver.hw_param);
	/* cr_init_sw_param(&cr_pcie_driver.sw_param); */
	/* cr_init_hw_param(&cr_pcie_driver.hw_param); */
}

static int __init cr_init_driver(void)
{
	rts_trace(TRACE_PROBE, "Register Realtek card reader driver.\n");

	cr_init_sysfs_param();

	cr_register_capability(&rts5227_capability);
	cr_register_capability(&rtl8411b_capability);
	cr_register_capability(&rts5249_capability);
	cr_register_capability(&rts5250_capability);
	cr_register_capability(&rts522A_capability);
	cr_register_capability(&rts524A_capability);
	cr_register_capability(&rts525A_capability);

	return pci_register_driver(&cr_pcie_driver.pci_driver);
}

static void __exit cr_exit_driver(void)
{
	rts_trace(TRACE_PROBE, "Unregister Realtek card reader driver.\n");
	cr_ungister_capabililty(0x5227, 0x10EC);
	cr_ungister_capabililty(0x5287, 0x10EC);
	cr_ungister_capabililty(0x5249, 0x10EC);
	cr_ungister_capabililty(0x5250, 0x10EC);
	cr_ungister_capabililty(0x522A, 0x10EC);
	cr_ungister_capabililty(0x524A, 0x10EC);
	cr_ungister_capabililty(0x525A, 0x10EC);
	pci_unregister_driver(&cr_pcie_driver.pci_driver);
}

module_init(cr_init_driver);
module_exit(cr_exit_driver);

MODULE_DESCRIPTION(RTS_CR_DESC);
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
MODULE_DEVICE_TABLE(pci, cr_pcie_ids);
