/* Driver for Realtek PCIe card reader
 *
 * Copyright(c) 2009 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   wwang (wei_wang@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/workqueue.h>
#include <linux/dma-mapping.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <linux/module.h>
#include <linux/init.h>

#include "cr_system.h"
#include "pm.h"
#include "cr_regs.h"
#include "trace.h"
#include "trans_pcie.h"
#include "rts_cr.h"
#include "cards.h"
#include "cards/sd.h"
#include "cards/ms.h"
#include "rtl8411.h"

static int rtl8411_set_ocp_thd(struct card_reader *cr, u8 thd);
static int rtl8411_save_host_aspm_val(struct card_reader *cr);
static int rtl8411_do_driver_load_handshake(struct card_reader *cr);
static int rtl8411_do_driver_unload_handshake(struct card_reader *cr);

#define READ_NIC_IO	0
#define WRITE_NIC_IO	1
static int wait_nic_io_complete(struct card_reader *cr, int action)
{
	int i = 0;
	u32 lval = 0;
	u32 complete = 0;

	for (i = 0; i < 1000; i++) {
		cr_rd_cfg32(cr, 0x820, &lval);
		if (READ_NIC_IO == action)
			complete = !(lval & (1 << 31));
		else
			complete = !(lval & (1 << 30));

		if (complete)
			return 0;

		udelay(1);
	}

	return -ETIMEDOUT;
}

static int write_nic_io(struct card_reader *cr, u8 addr, u8 mask, u32 val)
{
	u32 cmd = (0x40 << 24) | ((mask & 0x0F) << 16) | addr;

	cr_wr_cfg32(cr, 0x824, 0xFFFFFFFF, val);
	cr_wr_cfg32(cr, 0x820, 0xFFFFFFFF, cmd);

	return wait_nic_io_complete(cr, WRITE_NIC_IO);
}

static int read_nic_io(struct card_reader *cr, u8 addr, u8 mask, u32 *val)
{
	u32 cmd = (0x80 << 24) | ((mask & 0x0F) << 16) | addr;
	int ret = 0;

	cr_wr_cfg32(cr, 0x820, 0xFFFFFFFF, cmd);
	ret = wait_nic_io_complete(cr, READ_NIC_IO);
	cr_rd_cfg32(cr, 0x824, val);

	return  ret;
}

void rtl8411_turn_on_led(struct card_reader *cr)
{
	cr_wr_reg(cr, RTL8411_CARD_GPIO, 0x01, 0x01);
}

void rtl8411_turn_off_led(struct card_reader *cr)
{
	cr_wr_reg(cr, RTL8411_CARD_GPIO, 0x01, 0x00);
}

void rtl8411_enable_auto_blink(struct card_reader *cr)
{
	cr_wr_reg(cr, RTL8411_CARD_AUTO_BLINK, 0x0F, 0x0D);
}

void rtl8411_disable_auto_blink(struct card_reader *cr)
{
	cr_wr_reg(cr, RTL8411_CARD_AUTO_BLINK, 0x08, 0x00);
}

static enum cr_QFN rtl8411B_get_QFN(struct card_reader *cr)
{
	u8 val;

	if ((cr_rd_reg(cr, RTL8411_PACKAGE_MODE, &val)))
		return QFN_INVALID;

	if (val & (1 << 1))
		return QFN48;
	else
		return QFN64;
}

static int rtl8411_extra_init_hw(struct card_reader *cr)
{
	u8 mask;

	if (CHECK_PID(cr, RTL8411B_PID)) {
		if (QFN48 == rtl8411B_get_QFN(cr))
			cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0xFF, 0xF5);
	}

	mask = RTL8411_CD_DISABLE_MASK | CD_AUTO_DISABLE;
	return cr_wr_reg(cr, REG_CD_PAD_CTL, mask, CD_ENABLE);
}

int rtl8411_init_hw(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	struct rts_pcie *pcie = cr->tr->extra;

	cr_iowrite32(cr, HCBAR, pcie->host_cmds_addr);

	cr_wr_reg(cr, REG_FPDCTL, DISABLE_AUTOPD_OCP, DISABLE_AUTOPD_OCP);

	if (check_dev_flag(cr, HANDSHAKE_EN))
		rtl8411_save_host_aspm_val(cr);
	rtl8411_init_aspm(cr);

	/* Power on SSC and Wait SSC power stable */
	cr_wr_reg(cr, REG_FPDCTL, SSC_POWER_DOWN, 0);
	udelay(200);
	cr_wr_reg(cr, REG_FUNC1_FORCE_CTL, RTL8411_U_CR_HW_PFM_EN
			| RTL8411_U_CR_SW_PFM_EN, 0);

	/* init Phy */

	/* Set mcu_cnt to 7 to ensure data can be sampled properly */
	cr_wr_reg(cr, REG_CLK_DIV, 0x07, 0x07);

	cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03, 0x00);
	/* Disable card clock */
	cr_wr_reg(cr, REG_CARD_CLK_EN, 0x1E, 0);

	cr_init_ocp(cr);

	/* Configure LED */
	if (check_dev_flag(cr, TURN_ON_LED_AT_START))
		cr_turn_on_led(cr);
	else
		cr_turn_off_led(cr);

	/* Reset delink mode */
	cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x0A, 0);

	/* Card driving select */
	cr_wr_reg(cr, REG_CARD_DRIVE_SEL, 0xFF, option->card_drive_sel);
	cr_wr_reg(cr, REG_SD30_DRIVE_SEL, 0x07, option->sd30_drive_sel_3v3);

	/* Enable SSC Clock */
	cr_wr_reg(cr, REG_SSC_CTL1, 0xFF, SSC_8X_EN | SSC_SEL_4M);
	cr_wr_reg(cr, REG_SSC_CTL2, 0xFF, 0x12);

	cr_wr_reg(cr, REG_CD_PAD_CTL,
		RTL8411_CD_DISABLE_MASK | CD_AUTO_DISABLE, CD_ENABLE);

	/* Disable cd_pwr_save (u_force_rst_core_en=0, u_cd_rst_core_en=0) */
	cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x16, 0x10);

	/* Clear Link Ready Interrupt Write 1 to clear */
	cr_wr_reg(cr, REG_IRQSTAT0, LINK_RDY_INT, LINK_RDY_INT);

	cr_wr_reg(cr, REG_PERST_GLITCH_WIDTH, 0xFF, 0x80);

	/* Update RC oscillator to 400k
	 * bit[0] F_HIGH: for RC oscillator, Rst_value is 1'b1
	 *		1: 2M  0: 400k */
	cr_wr_reg(cr, REG_RCCTL, 0x01, 0x00);

	if (option->write_clear_interrupt)
		cr_wr_reg(cr, REG_NFTS_TX_CTRL, 0x02, 0);

	/* u_force_clkreq_0
	 * If 1, CLKREQ# PIN will be forced to drive 0 to request clock */
	if (option->force_clkreq_0)
		cr_wr_reg(cr, REG_PETXCFG, 0x80, 0x80);
	else
		cr_wr_reg(cr, REG_PETXCFG, 0x80, 0x00);

	if (DEV_ASPM_DISABLE != option->dev_aspm_mode) {
		/* TODO: if SDIO func exist, configure 0xC0 */

		/* TODO: U_cr_aspm_rdy had beed removed */
		cr_wr_reg(cr, REG_FUNC1_FORCE_CTL, 0x08, 0x00);
		cr->aspm_level[0] = option->dev_aspm_val;
		cr->aspm_enabled = 1;
		if (option->config_host_aspm)
			cr_set_host_aspm(cr, cr->host_aspm_val);
	}

	/* Enable clk_request_n to enable clock power management */
	cr_write_config8(cr, 0x81, 1);

	/* wwang, 2010-06-22
	 * Enter L1 when host tx idle
	 * Set 0x70F to 0x5B using backdoor */
	cr_wr_cfg32(cr, 0x70C, 0xFF000000, 0x5B000000);


	/* Enable power gating for autodelink */
	/* cr_wr_reg(cr, REG_PWD_SUSPEND_EN, 0xFF, 0xFF); */
	/* cr_wr_reg(cr, REG_PWR_GATE_CTRL, PWR_GATE_EN, PWR_GATE_EN); */

	rtl8411_extra_init_hw(cr);
	cr_enable_bus_int(cr, cr->hw_param.interrupt_en);

	cr_check_card_in_init(cr);

	option->aux_pwr_exist = 1;

	return 0;
}

static int rtl8411_init_from_hw(struct card_reader *cr)
{
	u8 val;
	u32 cfg;
	struct cr_option *option = &(cr->option);

	option->aux_pwr_exist = 0;

	if ((cr_rd_cfg32(cr, 0x724, &cfg)))
		TRACE_RET(cr, STATUS_FAIL);
	rts_trace(TRACE_CR, "dw in 0x724: 0x%x\n", cfg);

	val = (u8)cfg;

	/*if (val & 0x02)
	  option->hw_bypass_sd = 0;
	  else
	  option->hw_bypass_sd = 1;*/

	/* 0x727 */
	val = (u8)(cfg >> 24);
	if (((val & 0x01) == 0) && option->use_hw_setting) {
		u8 card_spt_map = cr_get_card_spt_map_sw(cr);
		u32 sd_capability = cr_get_sd_capability_sw(cr);
		u8 cap = 0;
		u8 lun_mode[4] = {0xFF, 2, 1, 0};

		u8 sd_drive[4] = {
			0x01, /* Type D	*/
			0x02, /* Type C	*/
			0x05, /* Type A	*/
			0x03, /* Type B	*/
		};

		u8 ssc_depth[4] = {
			SSC_DEPTH_512K,
			SSC_DEPTH_1M,
			SSC_DEPTH_2M,
			SSC_DEPTH_4M,
		};

		u8 ssc_depth2[4] = {
			SSC_DEPTH_256K,
			SSC_DEPTH_512K,
			SSC_DEPTH_1M,
			SSC_DEPTH_2M,
		};

		/* 0x727 */
		option->lun_mode = lun_mode[(val >> 6) & 0x03];
		cap = (val >> 6) & 0x03;
		option->dev_aspm_val = (val >> 4) & 0x03;
		option->force_host_aspm_val = (val >> 4) & 0x03;
		if (check_dev_flag(cr, FORCE_HOST_ASPM_EN))
			cr->host_aspm_val = option->force_host_aspm_val;
		option->sd30_drive_sel_1v8 = sd_drive[(val >> 2) & 0x03];

		/* check if support MMC by 0x727[1] */
		if (val & (1 << 1))
			sd_capability &= ~SD_CAP_MMC;
		else
			sd_capability |= SD_CAP_MMC;

		/* 0x726 */
		val = (u8)(cfg >> 16);

		option->ssc_depth_sd_hs = ssc_depth2[(val >> 6) & 0x03];
		option->ssc_depth_mmc_52m = option->ssc_depth_sd_hs;

		if (val & (1 << 5))
			sd_capability |= SD_CAP_SDR50;
		else
			sd_capability &= ~SD_CAP_SDR50;
		if (val & (1 << 4))
			sd_capability |= SD_CAP_DDR50;
		else
			sd_capability &= ~SD_CAP_DDR50;
		if (val & (1 << 3))
			sd_capability |= SD_CAP_SDR104;
		else
			sd_capability &= ~SD_CAP_SDR104;

		if (0x03 == cap)
			card_spt_map = (1 << SD_ID) | (1 << MS_ID);
		else if (0x02 == cap)
			card_spt_map = 1 << SD_ID;
		else if (0x01 == cap)
			card_spt_map = 1 << MS_ID;
		else
			card_spt_map = 0;

		cr_set_card_spt_map_sw(cr, card_spt_map);
		cr_set_sd_capability_sw(cr, sd_capability);

		/* 0x726[0~2]: reserved */

		/* 0x725 */
		val = (u8)(cfg >> 8);

		if ((val & 0xE0) != 0xE0)
			option->asic_sd_sdr104_clk = 206 -
				((val >> 5) & 0x07) * 3;
		if ((val & 0x1C) != 0x1C)
			option->asic_sd_sdr50_clk = 98 -
				((val >> 2) & 0x07) * 2;
		if ((val & 0x03) != 0x03)
			option->asic_sd_ddr50_clk = (48 - (val & 0x03) * 2) * 2;

		/* 0x724 */
		val = (u8)cfg;

		option->ssc_depth_sd_sdr104 = ssc_depth[(val >> 6) & 0x03];
		option->ssc_depth_sd_sdr50 = option->ssc_depth_sd_sdr104;
		option->ssc_depth_sd_ddr50 = ssc_depth[(val >> 4) & 0x03];
		option->rtd3_en = !!(val & (1 << 2));
		option->sd30_drive_sel_3v3 = sd_drive[val & 0x03];
	}


	return 0;
}

static int rtl8411_init_from_cfg(struct card_reader *cr)
{
	u32 lval;
	struct cr_option *option = &(cr->option);

	/* for ltr */
	cr_read_config32(cr, 0x98, &lval);
	if (lval & (1 << 10))
		option->ltr_enabled = 1;

	/* for obff */
	if (lval & (1 << 13))
		option->obff_enabled = 1;

	/* for l1off */
	cr_read_config32(cr, 0x180, &lval);
	if (lval & (1 << 3))
		set_dev_flag(cr, ASPM_L1_1_EN);
	else
		clear_dev_flag(cr, ASPM_L1_1_EN);
	if (lval & (1 << 2))
		set_dev_flag(cr, ASPM_L1_2_EN);
	else
		clear_dev_flag(cr, ASPM_L1_2_EN);

	if (lval & (1 << 1))
		set_dev_flag(cr, PM_L1_1_EN);
	else
		clear_dev_flag(cr, PM_L1_1_EN);
	if (lval & (1 << 0))
		set_dev_flag(cr, PM_L1_2_EN);
	else
		clear_dev_flag(cr, PM_L1_2_EN);

	return 0;
}

int rtl8411_init_cr(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	int retval;
	u8 val = 0;

	rts_trace(TRACE_CR, "VID: 0x%04x, PID: 0x%04x, SSVID: 0x%04x,"
			"SSDID: 0x%04x\n", cr->vendor_id, cr->product_id,
						cr->ssvid, cr->ssdid);

	val = cr_ioread8(cr, 0x1C);
	if ((val & 0x10) == 0)
		cr->asic_code = 1;
	else
		cr->asic_code = 0;

	cr_rd_reg(cr, 0xFE90, &val);
	rts_trace(TRACE_CR, "0xFE90: 0x%x\n", val);
	cr->ic_version = val & 0x0F;

	cr->dev_pci_cap_offset = cr_find_pci_capability(cr, PCI_CAP_ID_EXP);

	cr_set_stat(cr, RTSX_STAT_INIT);

	/* FIXME: edwin: is this place appropriate to initialise this variable?
	 * or maybe the only choice */
	cr->sd_max_phase = MAX_PHASE;

	option->auto_delink_allowed = 1;

	if (!check_sd_speed_prior(option->sd_speed_prior))
		option->sd_speed_prior = 0x01040203;
	rts_trace(TRACE_CR, "sd_speed_prior = 0x%08x\n",
		option->sd_speed_prior);

	if (!check_sd_current_prior(option->sd_current_prior))
		option->sd_current_prior = 0x00010203;
	rts_trace(TRACE_CR, "sd_current_prior = 0x%08x\n",
		option->sd_current_prior);

	if ((option->mmc_ddr_tx_phase > 31) || (option->mmc_ddr_tx_phase < 0))
		option->mmc_ddr_tx_phase = 0;

	rtl8411_init_from_cfg(cr);
	rtl8411_init_from_hw(cr);

	/* If ss_en is set to 2, disable ss */
	if (option->ss_en == 2)
		option->ss_en = 0;

	cr_print_option(cr);

	retval = rtl8411_init_hw(cr);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	rtl8411_do_driver_load_handshake(cr);

	if (RTSX_TRACE_MASK & TRACE_DUMP_ALL) {
		cr_dump_phy(cr, 0x00, 0x1F);
		cr_dump_register(cr, 0xFC00, 0xFF7F);
	}

	return 0;
}

static void rtl8411_init_special_param(struct card_reader *cr)
{
	struct rtl8411_info *info = cr->extra;

	info->ldo_tuned18 = ASIC_1V8;
}

void rtl8411_init_sw_param(struct cr_option *option)
{
	option->dev_flags |= FORCE_HOST_ASPM_EN;
	option->dev_flags |= LTR_L1SS_PWR_GATE_EN;
	option->dev_flags |= RESET_CARD_AFTER_RESUME_EN;
	option->dev_flags |= HANDSHAKE_EN;
	option->support_ms_8bit = 1;
	option->phy_optimize = 1;
	option->obff_en = 0;
	option->aux_clk_freq = 0;
	option->clk_req_bypass_short = 0;
	option->sdr_tx_tuning_en = 1;
	option->ltr_l1off_snooze_sspwrgate = 0xBC;
	option->ltr_l1off_sspwrgate = 0xBF;
	option->ltr_l1off_latency = 0x9003;
	option->sd30_clk_drive_sel_1v8 = 0xB3;
	option->sd30_clk_drive_sel_3v3 = 0x96;
	option->sd30_cmd_drive_sel_1v8 = 0xB3;
	option->sd30_cmd_drive_sel_3v3 = 0x96;
	option->sd30_dat_drive_sel_1v8 = 0xB3;
	option->sd30_dat_drive_sel_3v3 = 0x96;
	option->card_drive_sel = 1;
	option->ssc_depth_sd_hs = SSC_DEPTH_4M;
}

void rtl8411_init_hw_param(struct hw_param *hw_param)
{
	hw_param->interrupt_en = TRANS_OK_INT_EN | TRANS_FAIL_INT_EN;
	hw_param->interrupt_en |= DELINK_INT_EN | SD_INT_EN | MS_INT_EN;
	hw_param->ocp_glitch = 0x03;
	hw_param->phy0x00_val = 0xBA43;
	hw_param->phy0x01_val = 0x713F;
	hw_param->phy0x03_val = 0xC152;
}

void rtl8411_init_options(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	struct hw_param *hw_param = &(cr->hw_param);

	cr_init_generic_options(cr);
	rtl8411_init_sw_param(option);
	rtl8411_init_hw_param(hw_param);

	if (option->ocp_en)
		hw_param->interrupt_en |= SD_OC_INT_EN;

	if (!option->adma_mode)
		hw_param->interrupt_en |= DATA_DONE_INT_EN;

	rtl8411_init_special_param(cr);
}

int rtl8411_init(struct card_reader *cr)
{
	if ((cr_request_irq(cr) < 0))
		return -1;

	if (rtl8411_init_cr(cr) != STATUS_SUCCESS) {
		rts_trace(TRACE_CR, "rtl8411_init_cr() Failed\n");
		return -EIO;
	}

	return 0;
}

int rtl8411_poweron_card(struct card_reader *cr, enum CARD_ID id)
{
	int err;

	cr_wr_reg(cr, REG_CARD_PWR_CTL, BPP_POWER_MASK,
			BPP_POWER_5_PERCENT_ON);
	err = cr_wr_reg(cr, RTL8411_LDO_CTL, BPP_LDO_POWB, BPP_LDO_SUSPEND);
	if (err < 0)
		return err;

	/* To avoid too large in-rush current */
	udelay(150);

	err = cr_wr_reg(cr, REG_CARD_PWR_CTL,
			BPP_POWER_MASK, BPP_POWER_10_PERCENT_ON);
	if (err < 0)
		return err;

	udelay(150);

	err = cr_wr_reg(cr, REG_CARD_PWR_CTL,
			BPP_POWER_MASK, BPP_POWER_15_PERCENT_ON);
	if (err < 0)
		return err;

	udelay(150);

	err = cr_wr_reg(cr, REG_CARD_PWR_CTL, BPP_POWER_MASK, BPP_POWER_ON);
	if (err < 0)
		return err;

	return cr_wr_reg(cr, RTL8411_LDO_CTL, BPP_LDO_POWB, BPP_LDO_ON);
}

int rtl8411_power_off_card(struct card_reader *cr, enum CARD_ID id)
{
	int err;

	err = cr_wr_reg(cr, REG_CARD_PWR_CTL, BPP_POWER_MASK, BPP_POWER_OFF);
	if (err < 0)
		return err;

	return cr_wr_reg(cr, RTL8411_LDO_CTL, BPP_LDO_POWB, BPP_LDO_SUSPEND);
}

static int rtl8411_disable_card_int_by_efuse(struct card_reader *cr, u32 card)
{
	u8 val = 0x00;

	cr_rd_reg(cr, RTL8411_EFUSE_CONTENT, &val);

	if (card & (1 << SD_ID))
		val |= (1 << 6);

	if (card & (1 << MS_ID))
		val |= (1 << 7);

	return cr_wr_reg(cr, RTL8411_EFUSE_CONTENT, 0xE0, val);
}

static int rtl8411_enable_card_int_by_efuse(struct card_reader *cr, u32 card)
{
	u8 val = 0;

	cr_rd_reg(cr, RTL8411_EFUSE_CONTENT, &val);

	if (card & (1 << SD_ID))
		val &= ~(1 << 6);

	if (card & (1 << MS_ID))
		val &= ~(1 << 7);

	return cr_wr_reg(cr, RTL8411_EFUSE_CONTENT, 0xE0, val);
}

u32 rtl8411_cd_deglitch(struct card_reader *cr)
{
	u32 card_exist;
	u8 card_spt_map = cr_get_card_spt_map_sw(cr);
	u8 support_ms = card_spt_map & (1 << MS_ID);
	u8 support_sd = card_spt_map & (1 << SD_ID);

	card_exist = cr_ioread32(cr, BIPR);
	card_exist &= CARD_EXIST;
	if (!card_exist) {
		/* Enable card CD */
		cr_wr_reg(cr, REG_CD_PAD_CTL,
				RTL8411_CD_DISABLE_MASK, CD_ENABLE);
		/* Enable card interrupt */
		rtl8411_enable_card_int_by_efuse(cr,
				(1 << MS_ID) | (1 << SD_ID));
		return 0;
	}

	if (hweight32(card_exist) > 1) {
		cr_wr_reg(cr, REG_CARD_PWR_CTL,
				BPP_POWER_MASK, BPP_POWER_5_PERCENT_ON);
		msleep(100);

		card_exist = cr_ioread32(cr, BIPR);
		if (card_exist & MS_EXIST)
			card_exist = MS_EXIST;
		else if (card_exist & SD_EXIST)
			card_exist = SD_EXIST;
		else
			card_exist = 0;

		cr_wr_reg(cr, REG_CARD_PWR_CTL, BPP_POWER_MASK, BPP_POWER_OFF);

		rts_trace(TRACE_CM,
				"After CD deglitch, card_exist = 0x%x\n",
				card_exist);
	}

	if ((card_exist & MS_EXIST) && support_ms) {
		/* Disable SD interrupt */
		rtl8411_disable_card_int_by_efuse(cr, 1 << SD_ID);
		cr_wr_reg(cr, REG_CD_PAD_CTL,
			RTL8411_CD_DISABLE_MASK, RTL8411_MS_CD_EN_ONLY);
	} else if ((card_exist & SD_EXIST) && support_sd) {
		/* Disable MS interrupt */
		rtl8411_disable_card_int_by_efuse(cr, 1 << MS_ID);
		cr_wr_reg(cr, REG_CD_PAD_CTL,
			RTL8411_CD_DISABLE_MASK, RTL8411_SD_CD_EN_ONLY);
	}

	return card_exist;
}

static int rtl8411B_ms_pull_ctl_disable(struct card_reader *cr)
{
	enum cr_QFN qfn = rtl8411B_get_QFN(cr);

	if (QFN48 == qfn) {
		cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0x6F, 0x65);
		cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else if (QFN64 == qfn) {
		cr_wr_reg(cr, REG_CARD_PULL_CTL1, 0xFF, 0x65);
		cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0x0F, 0x05);
		cr_wr_reg(cr, REG_CARD_PULL_CTL4, 0x0F, 0x09);
		cr_wr_reg(cr, REG_CARD_PULL_CTL5, 0x0F, 0x05);
		cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else {
		return -EIO;
	}

	return 0;
}

static int rtl8411A_ms_pull_ctl_disable(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_CARD_PULL_CTL1, 0xFF, 0x65);
	cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
	cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0xFF, 0x95);
	cr_wr_reg(cr, REG_CARD_PULL_CTL4, 0xFF, 0x09);
	cr_wr_reg(cr, REG_CARD_PULL_CTL5, 0xFF, 0x05);
	cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0xFF, 0x04);

	return 0;
}

int rtl8411_ms_pull_ctl_disable(struct card_reader *cr)
{
	/* This function would be called in enter_ss
	 * which is mutual exclusive to interrupt handler.
	 * So we can't use command buffer to access internal register */
	if (CHECK_PID(cr, RTL8411B_PID))
		return rtl8411B_ms_pull_ctl_disable(cr);
	else
		return rtl8411A_ms_pull_ctl_disable(cr);

	return 0;
}

static int rtl8411B_ms_pull_ctl_enable(struct card_reader *cr)
{
	enum cr_QFN qfn = rtl8411B_get_QFN(cr);
	int ret;

	cr_init_cmd(cr);

	if (QFN48 == qfn) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0x6F, 0x65);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else if (QFN64 == qfn) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL1, 0xFF, 0x65);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0x0F, 0x05);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL4, 0x0F, 0x05);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL5, 0x0F, 0x05);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else {
		return -EINVAL;
	}

	ret = cr_send_cmd(cr, MS_CARD, 100);
	if (ret < 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int rtl8411A_ms_pull_ctl_enable(struct card_reader *cr)
{
	int ret;

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL1, 0xFF, 0x65);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0x55);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0xFF, 0x95);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL4, 0xFF, 0x05);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL5, 0xFF, 0x05);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0xFF, 0x04);

	ret = cr_send_cmd(cr, MS_CARD, 100);
	if (ret < 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int rtl8411_ms_pull_ctl_enable(struct card_reader *cr)
{
	/*
	 * MS CD: pull up
	 * others: pull down
	 */
	if (CHECK_PID(cr, RTL8411B_PID))
		return rtl8411B_ms_pull_ctl_enable(cr);
	else
		return rtl8411A_ms_pull_ctl_enable(cr);
}

static int rtl8411B_sd_pull_ctl_disable(struct card_reader *cr)
{
	enum cr_QFN qfn = rtl8411B_get_QFN(cr);

	if (QFN48 == qfn) {
		cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0x6F, 0x65);
		cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else if (QFN64 == qfn) {
		cr_wr_reg(cr, REG_CARD_PULL_CTL1, 0xFF, 0x65);
		cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
		cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0x0F, 0x05);
		cr_wr_reg(cr, REG_CARD_PULL_CTL4, 0x0F, 0x09);
		cr_wr_reg(cr, REG_CARD_PULL_CTL5, 0x0F, 0x05);
		cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else {
		return -EIO;
	}

	return 0;
}

int rtl8411A_sd_pull_ctl_disable(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_CARD_PULL_CTL1, 0xFF, 0x65);
	cr_wr_reg(cr, REG_CARD_PULL_CTL2, 0xFF, 0x55);
	cr_wr_reg(cr, REG_CARD_PULL_CTL3, 0xFF, 0x95);
	cr_wr_reg(cr, REG_CARD_PULL_CTL4, 0xFF, 0x09);
	cr_wr_reg(cr, REG_CARD_PULL_CTL5, 0xFF, 0x05);
	cr_wr_reg(cr, REG_CARD_PULL_CTL6, 0xFF, 0x04);

	return STATUS_SUCCESS;
}

int rtl8411_sd_pull_ctl_disable(struct card_reader *cr)
{
	/* This function would be called in enter_ss
	 * which is mutual exclusive to interrupt handler.
	 * So we can't use command buffer to access internal register
	 */

	if (CHECK_PID(cr, RTL8411B_PID))
		return rtl8411B_sd_pull_ctl_disable(cr);
	else
		return rtl8411A_sd_pull_ctl_disable(cr);

	return 0;
}

static int rtl8411B_sd_pull_ctl_enable(struct card_reader *cr)
{
	enum cr_QFN qfn = rtl8411B_get_QFN(cr);
	int ret;

	cr_init_cmd(cr);

	if (QFN48 == qfn) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0xAA);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0x6F, 0x69);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0x0C, 0x08);
	} else if (QFN64 == qfn) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL1, 0xFF, 0xAA);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0xAA);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0x0F, 0x09);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL4, 0x0F, 0x09);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL5, 0x0F, 0x09);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0x0C, 0x04);
	} else {
		return -EINVAL;
	}

	ret = cr_send_cmd(cr, MS_CARD, 100);
	if (ret < 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int rtl8411A_sd_pull_ctl_enable(struct card_reader *cr)
{
	int retval;

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL1, 0xFF, 0xAA);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL2, 0xFF, 0xAA);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL3, 0xFF, 0xA9);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL4, 0xFF, 0x09);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL5, 0xFF, 0x09);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_PULL_CTL6, 0xFF, 0x04);

	retval = cr_send_cmd(cr, SD_CARD, 100);
	if (retval < 0)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

int rtl8411_sd_pull_ctl_enable(struct card_reader *cr)
{
	/* SD Data0~3: pull up
	 * SD CD: pull up
	 * SD WP: pull up
	 * SD CMD: pull up
	 * SD CLK: pull down
	 */

	if (CHECK_PID(cr, RTL8411B_PID))
		return rtl8411B_sd_pull_ctl_enable(cr);
	else
		return rtl8411A_sd_pull_ctl_enable(cr);

	return 0;
}

int rtl8411_sd_init_power(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	int retval;

	if ((sd_power_off_card3v3(cr)))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_change_bank_voltage(cr, SD_COMM_IO_3V3))
		TRACE_RET(cr, STATUS_FAIL);

	if (!option->ft2_fast_mode)
		msleep(250);

	retval = cr_enable_card_clk(cr, SD_ID);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	if (cr->asic_code) {
		retval = sd_pull_ctl_enable(cr);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		cr_wr_reg(cr, REG_FPGA_PULL_CTL,
				FPGA_SD_PULL_CTL_BIT | 0x20, 0);
	}


	retval = cr_poweron_card(cr, SD_ID);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	msleep(260);

	if (cr->option.ocp_en) {
		u8 ocp;
		cr_get_ocpstat(cr, &ocp);
		if (ocp) {
			struct card *card = cm_get_card(cr->cm, SD_ID);

			if (card)
				card->reset_flags |= RESET_MEET_OCP;
			rts_trace(TRACE_SD, "OCP! OCPSTAT:%#x\n", ocp);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	cr_wr_reg(cr, REG_CARD_OE, SD_OUTPUT_EN, SD_OUTPUT_EN);

	return STATUS_SUCCESS;
}

int rtl8411_release_sd_card(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct cr_option *option = &(cr->option);
	int ret;
	u32 val;

	rts_trace(TRACE_SD, "%s\n", __func__);

	clr_card_flags(card, (READY | FAIL | WP));
	card->reset_flags = 0;

	sd_card->sd_lock_status = 0;
	sd_card->sd_erase_status = 0;

	memset(sd_card->raw_csd, 0, 16);
	memset(sd_card->raw_scr, 0, 8);
	memset(sd_card->raw_cid, 0, 16);
	memset(&sd_card->ge, 0, sizeof(struct sd_ge));


	ret = sd_power_off_card3v3(cr);

	ret |= sd_change_bank_voltage(cr, SD_IO_3V3);

	if (CHK_SD30_SPEED(sd_card)) {
		rtl8411_sd30_drive_sel(cr, SD_IO_3V3);
		rtl8411_set_ocp_thd(cr, option->sd_400mA_ocp_thd);
	}

	/* Suspend LDO3318 */
	/* cr_wr_reg(cr, REG_PWR_GATE_CTRL,
			LDO3318_PWR_MASK, LDO_SUSPEND); */

	/* if SD card had been removed, enable MS CD */
	val = cr_ioread32(cr, BIPR);
	if (!(val & SD_EXIST)) {
		rtl8411_enable_card_int_by_efuse(cr, 1 << MS_ID);
		/* Enable card CD */
		cr_wr_reg(cr, REG_CD_PAD_CTL,
				RTL8411_CD_DISABLE_MASK, CD_ENABLE);
	}

	return ret;
}

int rtl8411_sd_disable_card(struct card_reader *cr, struct card *card)
{
	return STATUS_SUCCESS;
}

int rtl8411_sd_update_report_status(struct card_reader *cr, struct card *card)
{
	return 0;
}

int rtl8411_release_ms_card(struct card_reader *cr, struct card *card)
{
	struct ms_info *ms_card = card->extra;
	u32 val = 0;
	int ret = 0;

	rts_trace(TRACE_MS, "%s\n", __func__);

	if (!(check_card_flag(card, READY | EJECTED | FAIL | FAIL_RESET))) {
		rts_trace(TRACE_MS,
			"This is a fake interrupt about MS CARD removed.\n");
		goto out;
	}

#ifdef MS_DELAY_WRITE
	ms_card->delay_write.delay_write_flag = 0;
#endif
	ms_card->pro_under_formatting = 0;
	clr_card_flags(card, (READY | FAIL | WP));
	ms_free_l2p_tbl(cr, card);
	memset(ms_card->raw_sys_info, 0, sizeof(ms_card->raw_sys_info));
	memset(ms_card->raw_ms_id, 0, sizeof(ms_card->raw_ms_id));
#ifdef SUPPORT_PCGL_1P18
	memset(ms_card->raw_model_name, 0, sizeof(ms_card->raw_model_name));
#endif

	ret |= ms_power_off_card3v3(cr);

out:
	/* if MS card had been removed, enable SD CD */
	val = cr_ioread32(cr, BIPR);
	if (!(val & MS_EXIST)) {
		rtl8411_enable_card_int_by_efuse(cr, 1 << SD_ID);
		/* Enable card CD */
		cr_wr_reg(cr, REG_CD_PAD_CTL,
				RTL8411_CD_DISABLE_MASK, CD_ENABLE);
	}

	return ret;
}

int rtl8411_clear_ocpstat(struct card_reader *cr)
{
	u8 mask = SD_OCP_INT_CLR | SD_OC_CLR;
	u8 val = SD_OCP_INT_CLR | SD_OC_CLR;
	cr_wr_reg(cr, RTL8411_OCPCTL, mask, val);
	cr_wr_reg(cr, RTL8411_OCPCTL, mask, 0);

	return 0;
}

int rtl8411_enable_ocp(struct card_reader *cr)
{
	u8 val = SD_OCP_INT_EN | SD_DETECT_EN;

	return cr_wr_reg(cr, RTL8411_OCPCTL, 0x0C, val);
}

int rtl8411_disable_ocp(struct card_reader *cr)
{
	u8 mask = SD_OCP_INT_EN | SD_DETECT_EN;

	return cr_wr_reg(cr, RTL8411_OCPCTL, mask, 0);
}

int rtl8411_get_ocpstat(struct card_reader *cr, u8 *val)
{
	return cr_rd_reg(cr, RTL8411_OCPSTAT, val);
}

static int rtl8411_set_ocp_thd(struct card_reader *cr, u8 thd)
{
	cr_wr_reg(cr, RTL8411_OCPPARA2, SD_OCP_THD_MASK, thd);

	return 0;
}

int rtl8411_init_ocp(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);

	if (option->ocp_en) {
		u8 val = option->sd_400mA_ocp_thd;
		cr_wr_reg(cr, REG_FPDCTL, OC_POWER_DOWN, 0);

		/* Init OCPPARA1 */
		cr_wr_reg(cr, RTL8411_OCPPARA1, SD_OCP_TIME_MASK,
							SD_OCP_TIME_800);
		/* Init OCPPARA2 */
		cr_wr_reg(cr, RTL8411_OCPPARA2, SD_OCP_THD_MASK, val);

		/* Init OCPGLITCH */
		cr_wr_reg(cr, RTL8411_OCPGLITCH, 0x07, cr->hw_param.ocp_glitch);
		cr_enable_ocp(cr);
	} else {
		/* OC power down */
		cr_wr_reg(cr, REG_FPDCTL, OC_POWER_DOWN, OC_POWER_DOWN);
	}

	return 0;
}

int rtl8411_sd_change_bank_voltage(struct card_reader *cr, u8 voltage)
{
	struct rtl8411_info *info = cr->extra;
	u8 mask, val, shift;

	shift = RTL8411_TUNED18_SHIFT;
	mask = (RTL8411_TUNED18 << shift) | RTL8411_SDBUS_IS_3V3;

	if (voltage == SD_IO_3V3) {
		if (cr->asic_code)
			val = (ASIC_3V3 << shift) | RTL8411_SD30_3V3;
		else
			val = (FPGA_3V3 << shift) | RTL8411_SD30_3V3;
	} else if (voltage == SD_IO_1V8) {
		if (cr->asic_code)
			val = (info->ldo_tuned18 << shift) | RTL8411_SD30_1V8;
		else
			val = (FPGA_1V8 << shift) | RTL8411_SD30_1V8;
	} else if (voltage == SD_IO_1V9) {
		if (cr->asic_code)
			val = (ASIC_1V9 << shift) | RTL8411_SD30_1V8;
		else
			val = (FPGA_1V8 << shift) | RTL8411_SD30_1V8;
	} else {
		TRACE_RET(cr, STATUS_FAIL);
	}

	cr_wr_reg(cr, RTL8411_LDO_CTL, mask, val);

	return 0;
}

int rtl8411_sd_current_over_limit400_handler(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);

	cr_wr_reg(cr, RTL8411_OCPPARA2, SD_OCP_THD_MASK,
		option->sd_800mA_ocp_thd);

	return 0;
}

int rtl8411_sd30_drive_sel(struct card_reader *cr, u8 voltage)
{
	struct cr_option *option = &(cr->option);
	u8 val;

	if (voltage == SD_IO_3V3) {
		val = option->sd30_drive_sel_3v3;
	} else {
		struct card *card = cm_get_ready_card(cr->cm, SD_ID);

		val = option->sd30_drive_sel_1v8;

		if (card) {
			struct sd_info *sd_card = card->extra;
			if (CHECK_PID(cr, 0x5289) && CHK_SD_SDR104(sd_card))
				val = 0x04;
		}
	}

	cr_wr_reg(cr, REG_SD30_DRIVE_SEL, 0xFF, val);

	return 0;
}

int rtl8411_calc_ssc_clock_N(struct card_reader *cr,
		struct card *card, struct ssc_clock_N *ssc)
{
	ssc->N = (u8)((ssc->clk * 4) / 5 - 2);
	ssc->min_N = 80;
	ssc->max_N = 208;

	if (SD_ID == card->id) {
		struct sd_info *sd_card = card->extra;
		if (CHK_SD_DDR50(sd_card) || CHK_MMC_DDR52(sd_card))
			ssc->min_N = 40;
	}

	return 0;
}

static int cfg_set_bit(struct card_reader *cr, u16 addr, unsigned int index,
		u8 val)
{
	int ret;
	u32 lval, mask;

	ret = cr_rd_cfg32(cr, addr, &lval);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	lval &= ~((u32)1 << index);
	lval |= ((u32)(val & 0x01) << index);

	if (index < 8) {
		mask = 0x000000FF;
	} else if (index < 16) {
		mask = 0x0000FF00;
	} else if (index < 24) {
		mask = 0x00FF0000;
	} else {
		mask = 0xFF000000;
	}

	ret = cr_wr_cfg32(cr, addr, mask, lval);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

static int cfg_check_bit(struct card_reader *cr, u16 addr, unsigned int index)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, addr, &lval);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, -1);

	if (lval & ((u32)1 << index))
		return 1;

	return 0;
}

static int rtl8411_dekker_set_turn(struct card_reader *cr, u8 val)
{
	if (CHECK_PID(cr, RTL8411_PID))
		return cfg_set_bit(cr, 0x740, 24, val);
	else
		return cfg_set_bit(cr, 0x818, 23, val);
}

static int rtl8411_dekker_get_turn(struct card_reader *cr)
{
	if (CHECK_PID(cr, RTL8411_PID))
		return cfg_check_bit(cr, 0x740, 24);
	else
		return cfg_check_bit(cr, 0x818, 23);
}

static int rtl8411_dekker_check_flag0(struct card_reader *cr)
{
	if (CHECK_PID(cr, RTL8411_PID))
		return cfg_check_bit(cr, 0x740, 8);
	else
		return cfg_check_bit(cr, 0x818, 7);
}

static int rtl8411_dekker_set_flag1(struct card_reader *cr, u8 val)
{
	if (CHECK_PID(cr, RTL8411_PID))
		return cfg_set_bit(cr, 0x740, 16, val);
	else
		return cfg_set_bit(cr, 0x818, 15, val);
}

int rtl8411_cfg_try_lock(struct card_reader *cr, int timeout)
{
	int ret;
	int turn, flag0;

	if (timeout == 0)
		timeout = 31415926;

	ret = rtl8411_dekker_set_flag1(cr, 1);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	do {
		flag0 = rtl8411_dekker_check_flag0(cr);
		if (flag0 < 0)
			TRACE_RET(cr, STATUS_FAIL);
		if (flag0 == 0)
			break;

		turn = rtl8411_dekker_get_turn(cr);
		if (turn == 0) {
			ret = rtl8411_dekker_set_flag1(cr, 0);
			if (ret != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);

			while (timeout--) {
				turn = rtl8411_dekker_get_turn(cr);
				if (turn < 0)
					TRACE_RET(cr, STATUS_FAIL);
				if (turn == 1)
					break;
				mdelay(1);
			}
			if (turn == 0)
				TRACE_RET(cr, STATUS_TIMEDOUT);

			ret = rtl8411_dekker_set_flag1(cr, 1);
			if (ret != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);
		}

		mdelay(1);
	} while (timeout--);

	return 0;
}

int rtl8411_cfg_unlock(struct card_reader *cr)
{
	if ((rtl8411_dekker_set_turn(cr, 0)))
		TRACE_RET(cr, STATUS_FAIL);

	if ((rtl8411_dekker_set_flag1(cr, 0)))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int rtl8411_chk_nic_attach(struct card_reader *cr)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_check_bit(cr, 0x744, 8);
	else
		return cfg_check_bit(cr, 0x74C, 24);
}

int rtl8411_set_cr_attach(struct card_reader *cr, u8 val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_set_bit(cr, 0x744, 9, val);
	else
		return cfg_set_bit(cr, 0x74C, 25, val);
}


static int rtl8411B_get_root_aspm_val(struct card_reader *cr, u8 *val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x818, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	if (val)
		*val = (u8)((lval >> 24) & 0x03);

	return 0;
}

static int rtl8411A_get_root_aspm_val(struct card_reader *cr, u8 *val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x744, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	if (val)
		*val = (u8)(lval & 0x03);

	return 0;
}

int rtl8411_get_root_aspm_val(struct card_reader *cr, u8 *val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return rtl8411A_get_root_aspm_val(cr, val);
	else
		return rtl8411B_get_root_aspm_val(cr, val);
}

static int rtl8411B_set_root_aspm_val(struct card_reader *cr, u8 val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x818, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	lval &= 0xFCFFFFFF;
	lval |= ((val & 0x03) << 24);

	ret = cr_wr_cfg32(cr, 0x818, 0xFF000000, lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int rtl8411A_set_root_aspm_val(struct card_reader *cr, u8 val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x744, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	lval &= 0xFFFFFFFC;
	lval |= (val & 0x03);

	ret = cr_wr_cfg32(cr, 0x744, 0xFF, lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int rtl8411_set_root_aspm_val(struct card_reader *cr, u8 val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return rtl8411A_set_root_aspm_val(cr, val);
	else
		return rtl8411B_set_root_aspm_val(cr, val);
}

int rtl8411_chk_root_aspm_flag(struct card_reader *cr)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_check_bit(cr, 0x744, 2);
	else
		return cfg_check_bit(cr, 0x818, 26);
}

int rtl8411_set_root_aspm_flag(struct card_reader *cr, u8 val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_set_bit(cr, 0x744, 2, val);
	else
		return cfg_set_bit(cr, 0x818, 26, val);
}

static int rtl8411B_get_aspm_val(struct card_reader *cr, u8 *val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x818, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	if (val)
		*val = (u8)((lval >> 28) & 0x03);

	return 0;
}

static int rtl8411A_get_aspm_val(struct card_reader *cr, u8 *val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x744, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	if (val)
		*val = (u8)((lval >> 4) & 0x03);

	return 0;
}

int rtl8411_get_aspm_val(struct card_reader *cr, u8 *val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return rtl8411A_get_aspm_val(cr, val);
	else
		return rtl8411B_get_aspm_val(cr, val);
}


int rtl8411_chk_aspm_mod_flag(struct card_reader *cr)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_check_bit(cr, 0x744, 6);
	else
		return cfg_check_bit(cr, 0x818, 30);
}

int rtl8411_set_aspm_mod_flag(struct card_reader *cr, u8 val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return cfg_set_bit(cr, 0x744, 6, val);
	else
		return cfg_set_bit(cr, 0x818, 30, val);
}

static int rtl8411B_set_aspm_val(struct card_reader *cr, u8 val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x818, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	lval &= 0xCFFFFFFF;
	lval |= ((val & 0x03) << 28);

	ret = cr_wr_cfg32(cr, 0x818, 0xFF000000, lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int rtl8411A_set_aspm_val(struct card_reader *cr, u8 val)
{
	int ret;
	u32 lval;

	ret = cr_rd_cfg32(cr, 0x744, &lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	lval &= 0xFFFFFFCF;
	lval |= ((val & 0x03) << 4);

	ret = cr_wr_cfg32(cr, 0x744, 0xFF, lval);
	if (ret != 0)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int rtl8411_set_aspm_val(struct card_reader *cr, u8 val)
{
	if ((CHECK_PID(cr, RTL8411_PID)))
		return rtl8411A_set_aspm_val(cr, val);
	else
		return rtl8411B_set_aspm_val(cr, val);
}

static int rtl8411_save_host_aspm_val(struct card_reader *cr)
{
	int ret;
	int save_aspm = 1;
	u8 val;

	ret = rtl8411_cfg_try_lock(cr, 10000);
	if (ret != 0)
		TRACE_RET(cr, ret);

	ret = rtl8411_set_cr_attach(cr, 1);
	if (ret != 0)
		TRACE_GOTO(cr, Out);

	if (rtl8411_chk_nic_attach(cr) > 0) {
		if (rtl8411_chk_root_aspm_flag(cr) > 0)
			save_aspm = 0;
	}

	if (save_aspm) {
		cr_get_host_aspm(cr, &val);
		rts_trace(TRACE_PM, "Initial Host ASPM value: 0x%x\n", val);
		ret = rtl8411_set_root_aspm_val(cr, val);
		if (ret != 0)
			TRACE_GOTO(cr, Out);
		ret = rtl8411_set_root_aspm_flag(cr, 1);
		if (ret != 0)
			TRACE_GOTO(cr, Out);
	} else {
		ret = rtl8411_get_root_aspm_val(cr, &val);
		if (ret != 0)
			TRACE_GOTO(cr, Out);
		rts_trace(TRACE_PM, "Read root_aspm_val set by NIC: 0x%x\n",
			val);
	}

Out:
	(void)rtl8411_cfg_unlock(cr);
	return ret;
}

static int rtl8411_read_aspm_val_from_nic(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	int ret;
	u8 aspm_val;

	rts_trace(TRACE_PM, "aspm_mod_flag set!\n");

	ret = rtl8411_get_aspm_val(cr, &aspm_val);
	if (ret != 0)
		TRACE_RET(cr, ret);
	rts_trace(TRACE_PM, "aspm_val = 0x%x\n", aspm_val);

	if (cr->host_aspm_val != (option->dev_aspm_val & aspm_val)) {
		CLEAR_STATUS_FLAG(cr, HOST_ASPM_ENABLED);
		cr->host_aspm_val = option->dev_aspm_val & aspm_val;
	}
	rts_trace(TRACE_PM, "host_aspm_enabled: %d\n",
			CHECK_STATUS_FLAG(cr, HOST_ASPM_ENABLED));

	ret = rtl8411_set_aspm_mod_flag(cr, 0);
	if (ret != 0)
		TRACE_RET(cr, ret);

	return 0;
}

static int rtl8411_configure_host_aspm(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);
	int ret;

	ret = rtl8411_cfg_try_lock(cr, 10000);
	if (ret != 0)
		TRACE_RET(cr, ret);

	if (rtl8411_chk_nic_attach(cr) > 0) {
		if (rtl8411_chk_aspm_mod_flag(cr) > 0) {
			ret = rtl8411_read_aspm_val_from_nic(cr);
			if (ret != 0)
				TRACE_GOTO(cr, Out);
		}
	} else {
		if (cr->host_aspm_val != option->dev_aspm_val) {
			rts_trace(TRACE_PM, "After NIC detached, "
				"host_aspm_enabled = %d\n",
				CHECK_STATUS_FLAG(cr, HOST_ASPM_ENABLED));
			CLEAR_STATUS_FLAG(cr, HOST_ASPM_ENABLED);
			cr->host_aspm_val = option->dev_aspm_val;
		}
	}

Out:
	(void)rtl8411_cfg_unlock(cr);
	return ret;
}

static int rtl8411_restore_host_aspm_halt(struct card_reader *cr)
{
	int ret;
	u8 val;

	ret = rtl8411_cfg_try_lock(cr, 10000);
	if (ret != 0)
		TRACE_RET(cr, ret);

	if (rtl8411_chk_nic_attach(cr) == 0) {
		ret = rtl8411_get_root_aspm_val(cr, &val);
		if (ret != 0)
			TRACE_GOTO(cr, Out);

		cr_set_host_aspm(cr, val);

		ret = rtl8411_set_root_aspm_flag(cr, 0);
		if (ret != 0)
			TRACE_GOTO(cr, Out);
	}

	ret = rtl8411_set_cr_attach(cr, 0);
	if (ret != 0)
		TRACE_GOTO(cr, Out);

Out:
	(void)rtl8411_cfg_unlock(cr);
	return ret;
}

static int rtl8411_restore_host_aspm_sleep(struct card_reader *cr)
{
	int ret;
	u8 val;

	ret = rtl8411_get_root_aspm_val(cr, &val);
	if (ret != 0)
		TRACE_RET(cr, ret);

	cr_set_host_aspm(cr, val);

	ret = rtl8411_set_root_aspm_flag(cr, 0);
	if (ret != 0)
		TRACE_RET(cr, ret);

	ret = rtl8411_set_cr_attach(cr, 0);
	if (ret != 0)
		TRACE_RET(cr, ret);

	return 0;
}

int rtl8411_force_power_on(struct card_reader *cr, u8 ctl)
{
	int retval;
	u8 mask = 0;

	if (ctl & SSC_PDCTL) {
		mask |= SSC_POWER_DOWN;
		cr_wr_reg(cr, RTL8411_LDO_SSC_CTL,
				RTL8411_LDO_EN, RTL8411_LDO_EN);
	}

	if (cr->option.ocp_en && (ctl & OC_PDCTL))
		mask |= SD_OC_POWER_DOWN;

	if (mask) {
		retval = cr_wr_reg(cr, REG_FPDCTL, mask, 0);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

int rtl8411_force_power_down(struct card_reader *cr, u8 ctl)
{
	int retval;
	u8 mask = 0, val = 0;

	rts_trace(TRACE_PM, "<--\n");

	if (ctl & SSC_PDCTL) {
		mask |= SSC_POWER_DOWN;
		cr_wr_reg(cr, RTL8411_LDO_SSC_CTL, RTL8411_LDO_EN, 0);
	}

	if (cr->option.ocp_en && (ctl & OC_PDCTL))
		mask |= SD_OC_POWER_DOWN;

	if (mask) {
		val = mask;
		retval = cr_wr_reg(cr, REG_FPDCTL, mask, val);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

void rtl8411_before_power_down(struct card_reader *cr, int pm_stat)
{
	struct cr_option *option = &cr->option;
	struct rtl8411_info *info = cr->extra;


	rts_trace(TRACE_PM, "%s, pm_stat = %d\n", __func__, pm_stat);

	if (PM_S1 == pm_stat)
		rtl8411_do_driver_unload_handshake(cr);

	if (rtl8411_force_power_on(cr, SSC_PDCTL))
		return;

	if (check_dev_flag(cr, HANDSHAKE_EN)) {
		if (info->driver_unload)
			rtl8411_restore_host_aspm_halt(cr);
		else
			rtl8411_restore_host_aspm_sleep(cr);
	}

	cm_release_cards(cr->cm);
	cr->cur_card = INVAL_ID;
	cr->cm->cur_card = NULL;
	cr->release_map = 0;

	cr_disable_bus_int(cr);

	/* u_force_clkreq_0 */
	/* cr_wr_reg(cr, REG_PETXCFG, 0x08, 0x08); */
	rtl8411_disable_aspm(cr);

	/* TODO: U_cr_aspm_rdy had beed removed */
	cr_wr_reg(cr, REG_FUNC1_FORCE_CTL, 0x08, 0x08);

	if (pm_stat == PM_S1) {
		rts_trace(TRACE_PM, "Host enter S1\n");
		cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03,
						HOST_ENTER_S1);
		cr_wr_reg(cr, 0xFF01, 0xFF, 0);
		cr_wr_reg(cr, 0xFF02, 0xFF, 0);
		cr_wr_reg(cr, 0xFF03, 0x01, 0);
	} else if (pm_stat == PM_S3) {
		/* wwang, 2010-12-16
		 * Before entering S3, some card may be in sequential mode,
		 * so here we delay some time to ensure programming finished
		 */
		if (option->s3_pwr_off_delay > 0)
			msleep(option->s3_pwr_off_delay);

		rts_trace(TRACE_PM, "Host enter S3\n");
		/* cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03, HOST_ENTER_S3); */

		cr_wr_reg(cr, 0xFF01, 0xFF, 0);
		cr_wr_reg(cr, 0xFF02, 0xFF, 0);
		cr_wr_reg(cr, 0xFF03, 0x01, 0);
	}

	if (option->do_delink_before_power_down && option->auto_delink_en) {
		/* Enable autodelink */
		cr_wr_reg(cr, REG_CHANGE_LINK_STATE, 0x02, 2);
	}

	rtl8411_enable_card_int_by_efuse(cr, (MS_ID << 1) | (SD_ID << 1));
	cr_wr_reg(cr, REG_CD_PAD_CTL, RTL8411_CD_DISABLE_MASK, CD_ENABLE);

	cr_wr_reg(cr, REG_FUNC1_FORCE_CTL, RTL8411_U_CR_HW_PFM_EN
			| RTL8411_U_CR_SW_PFM_EN, RTL8411_U_CR_SW_PFM_EN);

	/* 2010-03-02: OS will handle dstate, so we mark it here. */
	rtl8411_force_power_down(cr, SSC_PDCTL | OC_PDCTL);

	cr_pm_set_flags(cr, PM_POWR_SAVING_MAX);

	if (CHECK_STATUS_FLAG(cr, ENTERED_RTD3))
		cr_patch_rtd3(cr);
}

int rtl8411_do_resume(struct card_reader *cr)
{
	if (cr_request_irq(cr) < 0)
		return -1;

	cr_wr_reg(cr, REG_HOST_SLEEP_STATE, 0x03, 0x00);
	rtl8411_init_cr(cr);

	return 0;
}

void rtl8411_init_aspm(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;
	u8 val = 0;
	int lctlr = cr->dev_pci_cap_offset + PCI_EXP_LNKCTL;

	if (check_dev_flag(cr, FORCE_HOST_ASPM_EN)) {
		cr->host_aspm_val = option->force_host_aspm_val;
	} else {
		cr_get_host_aspm(cr, &cr->host_aspm_val);
		cr->host_aspm_val &= 0xCF;
	}
	rts_trace(TRACE_PM, "cr->host_aspm_val = 0x%x\n", cr->host_aspm_val);

	if (option->config_host_aspm)
		cr_disable_host_aspm(cr);

	cr_read_config8(cr, lctlr, &val);
	val &= 0xFC;
	cr_write_config8(cr, lctlr, val);
}

static void rtl8411_invalidate_host_aspm(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);

	if (option->config_host_aspm) {
		rtl8411_configure_host_aspm(cr);
		if ((CHECK_STATUS_FLAG(cr, HOST_ASPM_ENABLED))) {
			rts_trace(TRACE_PM, "disable Host ASPM\n");
			CLEAR_STATUS_FLAG(cr, HOST_ASPM_ENABLED);
			cr_disable_host_aspm(cr);
		}
	}

}

static void rtl8411_activate_host_aspm(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);

	if (option->config_host_aspm) {
		rtl8411_configure_host_aspm(cr);
		if (!(CHECK_STATUS_FLAG(cr, HOST_ASPM_ENABLED))) {
			rts_trace(TRACE_PM, "Try to set Host ASPM to:"
					"%d\n", cr->host_aspm_val);
			SET_STATUS_FLAG(cr, HOST_ASPM_ENABLED);
			cr_set_host_aspm(cr, cr->host_aspm_val);
		}
	}
}

static void rtl8411_do_enable_aspm(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	rts_trace(TRACE_PM, "dev_aspm_val = %#x\n", option->dev_aspm_val);
	rts_trace(TRACE_PM, "dev_aspm_mode = %#x\n", option->dev_aspm_mode);

	if (!cr->aspm_enabled) {
		if (option->dev_aspm_mode != DEV_ASPM_DISABLE) {
			int lctlr = cr->dev_pci_cap_offset + PCI_EXP_LNKCTL;
			u8 val = 0;
			rts_trace(TRACE_PM, "Try to enable ASPM\n");
			cr_read_config8(cr, lctlr, &val);
			val &= 0xFC;
			val |= (option->dev_aspm_val & 0x03);
			cr_write_config8(cr, lctlr, val);
		}
		rtl8411_activate_host_aspm(cr);
		cr->aspm_enabled = 1;
	}

	return;
}

static void rtl8411B_enable_aspm(struct card_reader *cr)
{
	rtl8411_do_enable_aspm(cr);
}

static void rtl8411A_enable_aspm(struct card_reader *cr)
{
	rtl8411_do_enable_aspm(cr);
}

void rtl8411_enable_aspm(struct card_reader *cr)
{
	if (CHECK_PID(cr, RTL8411_PID))
		rtl8411A_enable_aspm(cr);
	else
		rtl8411B_enable_aspm(cr);
}

static void rtl8411_do_disable_aspm(struct card_reader *cr)
{
	struct cr_option *option = &cr->option;

	if (cr->aspm_enabled) {
		rtl8411_invalidate_host_aspm(cr);
		if (option->dev_aspm_mode != DEV_ASPM_DISABLE) {
			int lctlr = cr->dev_pci_cap_offset + PCI_EXP_LNKCTL;
			u8 val = 0;
			rts_trace(TRACE_PM, "Try to disable ASPM\n");
			if (option->config_host_aspm)
				cr_disable_host_aspm(cr);
			cr_read_config8(cr, lctlr, &val);
			val &= 0xFC;
			cr_write_config8(cr, lctlr, val);
			msleep(10);
		}
		cr->aspm_enabled = 0;
	}
}

static void rtl8411B_disable_aspm(struct card_reader *cr)
{

	cr_wr_reg(cr, REG_FUNC1_FORCE_CTL,
		RTL8411_U_CR_HW_PFM_EN | RTL8411_U_CR_SW_PFM_EN, 0);

	rtl8411_do_disable_aspm(cr);
}

static void rtl8411A_disable_aspm(struct card_reader *cr)
{
	rtl8411_do_disable_aspm(cr);
}

void rtl8411_disable_aspm(struct card_reader *cr)
{
	if (CHECK_PID(cr, RTL8411_PID))
		rtl8411A_disable_aspm(cr);
	else
		rtl8411B_disable_aspm(cr);
}

/* set_sdr104_exist_bit
 *
 * set 0x753[1] to indicate sdr104 exist and was accessing now,
 * so NIC driver would adjust performance for CR and NIC.
 */
static int set_sdr104_exist_bit(struct card_reader *cr)
{
	if (cr->cm->cur_card) {
		if (cr->cm->cur_card->id == SD_ID) {
			struct sd_info *sd_card = cr->cm->cur_card->extra;
			if (CHK_SD_SDR104(sd_card)) {
				u32 mask = 0xFF000000;
				u32 val;

				cr_rd_cfg32(cr, 0x750, &val);
				val |= (1 << 25);
				cr_wr_cfg32(cr, 0x750, mask, val);
			}
		}
	}

	return 0;
}

/* clear_sdr104_exist_bit - clear sdr104 exsit bit
 */
static void clear_sdr104_exist_bit(struct card_reader *cr)
{
	u32 mask = 0xFF000000;
	u32 val;

	cr_rd_cfg32(cr, 0x750, &val);
	val &= ~(1 << 25);
	cr_wr_cfg32(cr, 0x750, mask, val);
}

int rtl8411_open_global_obff_switch(struct card_reader *cr)
{
	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0xB0, 0x0F, 0xF0161880);

	write_nic_io(cr, 0xB0, 0x0F, 0xF0174880);

	write_nic_io(cr, 0x70, 0x0F, 0x0000E602);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F3F2);

	write_nic_io(cr, 0x70, 0x0F, 0x000000FA);

	write_nic_io(cr, 0x74, 0x0F, 0x800015C0);

	return 0;
}

int rtl8411_close_global_obff_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0x74, 0x0F, 0x0000F3F2);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval &= 0xFFFDFFFF;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F3F2);

	return 0;
}

static int rtl8411_open_global_l1off_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0x74, 0x0F, 0x0000F1E0);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval |= 0x00040000;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F1E0);
	write_nic_io(cr, 0x74, 0x0F, 0x0000F1E8);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval &= 0xFFFF00FF;
	lval |= 0x00006400;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F1E8);

	return 0;
}

static int rtl8411_close_global_l1off_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0x74, 0x0F, 0x0000F1E0);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval &= 0xFFFBFFFF;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F1E0);

	return 0;
}

static int rtl8411_open_global_ltr_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0x74, 0x0F, 0x0000F5C0);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval |= 0x00020000;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F5C0);

	return 0;
}

static int rtl8411_close_global_ltr_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	write_nic_io(cr, 0x74, 0x0F, 0x0000F5C0);

	read_nic_io(cr, 0x70, 0x0F, &lval);
	lval &= 0xFFFDFFFF;
	write_nic_io(cr, 0x70, 0x0F, lval);

	write_nic_io(cr, 0x74, 0x0F, 0x8000F5C0);

	return 0;
}

static int rtl8411_open_global_aspm_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	read_nic_io(cr, 0x50, 0x0F, &lval);
	lval |= 0x800000C0;
	write_nic_io(cr, 0x50, 0x0F, lval);

	read_nic_io(cr, 0x54, 0x0F, &lval);
	lval |= 0x00010000;
	write_nic_io(cr, 0x54, 0x0F, lval);

	read_nic_io(cr, 0xF0, 0x0F, &lval);
	lval |= 0x00008000;
	write_nic_io(cr, 0xF0, 0x0F, lval);

	return 0;
}

static int rtl8411_close_global_aspm_switch(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	read_nic_io(cr, 0x50, 0x0F, &lval);
	lval |= 0x000000C0;
	write_nic_io(cr, 0x50, 0x0F, lval);

	read_nic_io(cr, 0x50, 0x0F, &lval);
	lval &= 0x7FFFFFFF;
	write_nic_io(cr, 0x50, 0x0F, lval);

	read_nic_io(cr, 0x54, 0x0F, &lval);
	lval &= 0xFFFEFFFF;
	write_nic_io(cr, 0x54, 0x0F, lval);

	read_nic_io(cr, 0xF0, 0x0F, &lval);
	lval &= 0xFFFF7FFF;
	write_nic_io(cr, 0xF0, 0x0F, lval);

	return 0;
}

static int rtl8411_do_lan_power_saving(struct card_reader *cr)
{
	u32 lval;

	if (CHECK_PID(cr, RTL8411_PID))
		return 0;

	read_nic_io(cr, 0xD0, 0x0F, &lval);
	lval &= 0x00000040;
	write_nic_io(cr, 0xD0, 0x0F, lval);

	read_nic_io(cr, 0xF0, 0x0F, &lval);
	lval &= 0x00400000;
	write_nic_io(cr, 0xF0, 0x0F, lval);

	write_nic_io(cr, 0xB0, 0x03, 0xF0290004);

	write_nic_io(cr, 0xB0, 0x03, 0xF02B0070);

	return 0;
}

static int rtl8411_do_driver_load_handshake(struct card_reader *cr)
{
	int ret = 0;

	ret = rtl8411_cfg_try_lock(cr, 10000);
	if (ret)
		TRACE_RET(cr, ret);

	if (!rtl8411_chk_nic_attach(cr)) {
		struct cr_option *option = &(cr->option);
		u32 l1off = 0;
		u32 lval = 0;

		 rtl8411_open_global_aspm_switch(cr);

		 /* if(option->obff_enabled)
			 rtl8411_open_global_obff_switch(cr); */

		 l1off = check_dev_flag(cr, ASPM_L1_1_EN);
		 l1off |= check_dev_flag(cr, ASPM_L1_2_EN);
		 l1off |= check_dev_flag(cr, PM_L1_1_EN);
		 l1off |= check_dev_flag(cr, PM_L1_2_EN);

		 if (l1off)
			 rtl8411_open_global_l1off_switch(cr);

		 if (option->ltr_enabled)
			 rtl8411_open_global_ltr_switch(cr);

		 /* check if LAN is disabled */
		 cr_rd_cfg32(cr, 0x724, &lval);
		 if (!(lval & (1 << 3)))
			 rtl8411_do_lan_power_saving(cr);
	}

	rtl8411_set_cr_attach(cr, 1);

	rtl8411_cfg_unlock(cr);

	return 0;
}

static int rtl8411_do_driver_unload_handshake(struct card_reader *cr)
{
	int ret = 0;

	ret = rtl8411_cfg_try_lock(cr, 10000);
	if (ret)
		TRACE_RET(cr, ret);

	if (!rtl8411_chk_nic_attach(cr)) {
		struct cr_option *option = &(cr->option);
		u32 l1off = 0;

		rtl8411_close_global_aspm_switch(cr);

		 /* if(option->obff_enabled)
			 rtl8411_close_global_obff_switch(cr);*/

		 l1off = check_dev_flag(cr, ASPM_L1_1_EN);
		 l1off |= check_dev_flag(cr, ASPM_L1_2_EN);
		 l1off |= check_dev_flag(cr, PM_L1_2_EN);
		 l1off |= check_dev_flag(cr, PM_L1_2_EN);

		 if (l1off)
			 rtl8411_close_global_l1off_switch(cr);

		 if (option->ltr_enabled)
			 rtl8411_close_global_ltr_switch(cr);
	}

	rtl8411_set_cr_attach(cr, 0);

	rtl8411_cfg_unlock(cr);

	return 0;
}

void rtl8411_deal_extra_action(struct card_reader *cr)
{
	struct rtl8411_info *info = cr->extra;
	int sdr104_exist = 0;

	if (cr_pm_check_flags(cr, PM_FULL_ON)) {
		if (cr->cm->cur_card && cr->cm->cur_card->id == SD_ID) {
				struct sd_info *sd_card;
				sd_card = cr->cm->cur_card->extra;
				if (CHK_SD_SDR104(sd_card))
					sdr104_exist = 1;
		}
	}

	if (sdr104_exist != info->sdr104_exist) {
		if (sdr104_exist)
			set_sdr104_exist_bit(cr);
		else
			clear_sdr104_exist_bit(cr);
		info->sdr104_exist = sdr104_exist;
	}
}


int rtl8411_set_ltr_latency(struct card_reader *cr, u32 latency)
{
	cr_wr_reg(cr, REG_MSGTXDATA0, 0xFF, (u8)(latency & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA1, 0xFF, (u8)((latency >> 8) & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA2, 0xFF, (u8)((latency >> 16) & 0xFF));
	cr_wr_reg(cr, REG_MSGTXDATA3, 0xFF, (u8)((latency >> 24) & 0xFF));
	cr_wr_reg(cr, RTL8411_LTR_CTL, 0xC0, 0xC0);

	return 0;
}

int rtl8411_set_l1off_sub(struct card_reader *cr, u32 val)
{
	return 0;
}

void rtl8411_power_saving_max(struct card_reader *cr)
{
	u32 card_exist;

	card_exist = cr_ioread32(cr, BIPR);
	card_exist &= CARD_EXIST;

	if (card_exist) {
		cr_wr_reg(cr, REG_FUNC1_FORCE_CTL,
			RTL8411_U_CR_HW_PFM_EN | RTL8411_U_CR_SW_PFM_EN, 0);
	} else {
		cr_wr_reg(cr, REG_FUNC1_FORCE_CTL,
			RTL8411_U_CR_HW_PFM_EN | RTL8411_U_CR_SW_PFM_EN,
			RTL8411_U_CR_HW_PFM_EN);
	}
}
