/* Driver for Realtek card reader
 * Header file
 *
 * Copyright(c) 2012 Realtek Semiconductor Corp. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   edwin.rong(edwin_rong@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 *
 *   Tony.Nie(tony_nie@realsil.com.cn)
 *   No. 450, Shenhu Road, Suzhou Industry Park, Suzhou, China
 */

#include <linux/kernel.h>
#include <scsi/scsi.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_device.h>
#include <linux/module.h>

#include "cardreader.h"
#include "cardmanager.h"
#include "rts_cr.h"
#include "trace.h"
#include "cr_regs.h"
#include "cards.h"
#include "sd.h"
#include "sd40.h"

typedef int (*tuning_func)(struct card_reader *cr,
					struct card *card, u8 sample_point);
inline void sd_comm_clear_sd_error(struct card_reader *cr);
static int sd_comm_change_bank_voltage(struct card_reader *cr, u8 voltage);
inline int sd_comm_check_spec(struct card_reader *cr, struct card *c);
static int sd_comm_voltage_switch(struct card_reader *cr, struct card *c);
static int sd_comm_prepare_reset(struct card_reader *cr, struct card *card);
static int sd_comm_set_byte_count_per_blk(struct card_reader *cr, int bcpb);
static int sd_comm_set_blk_count(struct card_reader *cr, int count);
static int sd_comm_set_init_para(struct card_reader *cr, struct card *card);
static int sd_comm_set_clock_divider(struct card_reader *cr, u8 clk_div);
static int sd_comm_get_crc7(struct card_reader *cr, u8 *buf);
static int sd_comm_current_over_limit400_handler(struct card_reader *cr);
static int sd_comm_check_crc16_err(struct card_reader *cr);
static int sd_comm_set_clk_pol_ddr_mod(struct card_reader *cr,
					enum sd_comm_polarity_sol sol);
static int sd_comm_set_sample_push_timing(struct card_reader *cr,
							struct card *card);
static int sd_comm_sd30_drive_sel(struct card_reader *cr, u8 select);
static int sd_comm_change_phase(struct card_reader *cr, struct card *card,
						u8 sample_point, u8 tune_dir);
static int sd_comm_wait_data_idle(struct card_reader *cr);
static int sd_comm_mmc_switch_timing_bus(struct card_reader *cr,
					struct card *card, int switch_ddr);
static int sd_comm_dummy_clock(struct card_reader *cr);
static int sd_comm_auto_tune_clock(struct card_reader *cr, struct card *card);

int sd_send_data_cmd(struct card *card, struct sd_command *cmd);
static int sd_tuning_tx(struct card_reader *cr, struct card *c);
static int sd_reset_mmc_only(struct card_reader *cr, struct card *card);
static int sd_check_wp_state(struct card_reader *cr, struct card *c);
static int sd_ddr_tuning(struct card_reader *cr, struct card *c);
static int sd_sdr_tuning(struct card_reader *cr, struct card *c);
static int sd_switch_function(struct card_reader *cr, struct card *c,
					enum sd_bus_width bus_width);
static int sd_read_lba0(struct card_reader *cr, struct card *c);
static int sd_wait_data_buf_ready(struct card_reader *cr, struct card *card);
static int mmc_check_ext_csd(struct card *card);
static int mmc_process_sdr200_error(struct card_reader *cr, struct card *card);
static int mmc_select_sdr200(struct card *card);
static u32 do_tuning(struct card_reader *cr,
	struct card *card, tuning_func func, u32 *map, u8 tune_dir);

/* Helper functions about send sd command */

static int sd_go_idle_state(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	INIT_SD_CMD(cmd, GO_IDLE_STATE, 0, 0, SD_RSP_TYPE_R0, 0);

	return sd_send_cmd(cr, card, &cmd);

}

static int sd_send_op_cond(struct card *card, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SEND_OP_COND, SUPPORT_VOLTAGE | 0x40000000,
						0, SD_RSP_TYPE_R3, 5);
	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

static int sd_set_blocklen(struct card *card, u32 blocklen,  u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SET_BLOCKLEN, blocklen, 0, SD_RSP_TYPE_R1, 5);

	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

int sd_send_cid(struct card *card, u8 *cid)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, ALL_SEND_CID, 0, 0, SD_RSP_TYPE_R2, 16);
	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (cid)
		memcpy(cid, cmd.rsp + 1, 15);

	return 0;
}
EXPORT_SYMBOL_GPL(sd_send_cid);

static int sd_set_relative_addr(struct card *card, u32 addr, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SET_RELATIVE_ADDR, addr, 0, SD_RSP_TYPE_R6, 5);

	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

int sd_clear_status(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	struct sd_info *sd_card = card->extra;

	INIT_SD_CMD(cmd, SEND_STATUS, sd_card->sd_addr, 0, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(cr, card, &cmd);
}
EXPORT_SYMBOL_GPL(sd_clear_status);

int sd_send_status(struct card *card, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	struct sd_info *sd_card = card->extra;
	int ret = 0;

	INIT_SD_CMD(cmd, SEND_STATUS, sd_card->sd_addr, 0, SD_RSP_TYPE_R1, 5);
	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

static int sd_app_cmd(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	struct sd_info *sd_card = card->extra;

	INIT_SD_CMD(cmd, APP_CMD, sd_card->sd_addr, 0, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(cr, card, &cmd);
}

static int sd_send_if_cond(struct card *card, u32 arg, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SEND_IF_COND, arg, 0, SD_RSP_TYPE_R7, 5);
	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);
	return 0;
}

static int sd_app_send_op_cond(struct card *card, u32 arg, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SD_APP_OP_COND, arg, SD_APP_CMD, SD_RSP_TYPE_R3, 5);
	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

static int sd_app_set_bus_width(struct card *card, u32 width)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	INIT_SD_CMD(cmd, SET_BUS_WIDTH, width, SD_APP_CMD, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(cr, card, &cmd);
}

/*
 * sd_switch_func - send SWITCH_FUNC cmd
 */
static int sd_switch_func(struct card *card, u32 arg, u8 *rsp)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;
	int ret = 0;

	INIT_SD_CMD(cmd, SWITCH, arg, 0, SD_RSP_TYPE_R1b, 5);
	cmd.timeout = 3000;

	ret = sd_send_cmd(cr, card, &cmd);
	if (ret)
		return ret;

	if (rsp)
		memcpy(rsp, cmd.rsp, 5);

	return 0;
}

static int sd_bustest_w(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	INIT_SD_CMD(cmd, BUSTEST_W, 0, 0, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(cr, card, &cmd);
}

int sd_stop_transmission(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	INIT_SD_CMD(cmd, STOP_TRANSMISSION, 0, 0, SD_RSP_TYPE_R1b, 0);
	cmd.timeout = 3000;

	return sd_send_cmd(cr, card, &cmd);
}
EXPORT_SYMBOL_GPL(sd_stop_transmission);

static int sd_app_set_clr_card_detect(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	INIT_SD_CMD(cmd, SET_CLR_CARD_DETECT, 0, SD_APP_CMD, ACMD42_R, 0);

	return sd_send_cmd(cr, card, &cmd);
}

static int sd_write_multiple_block(struct card *card, u32 addr)
{
	struct sd_command cmd;

	INIT_SD_CMD(cmd, WRITE_MULTIPLE_BLOCK, addr, 0, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(card->cm->cr, card, &cmd);
}

void sd_choose_proper_clock(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct cr_option *option = &(cr->option);

	if (CHK_SD_UHSII(sd_card)) {
		rts_trace(TRACE_SD, "Exit:%s\n", __func__);
	}

	if (CHK_SD_SDR104(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_sd_sdr104_clk;
		else
			sd_card->sd_clock = option->fpga_sd_sdr104_clk;
	} else if (CHK_SD_DDR50(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_sd_ddr50_clk;
		else
			sd_card->sd_clock = option->fpga_sd_ddr50_clk;
	} else if (CHK_SD_SDR50(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_sd_sdr50_clk;
		else
			sd_card->sd_clock = option->fpga_sd_sdr50_clk;
	} else if (CHK_SD_HS(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_sd_hs_clk;
		else
			sd_card->sd_clock = option->fpga_sd_hs_clk;
	} else if (CHK_MMC_SDR200(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_mmc_sdr200_clk;
		else
			sd_card->sd_clock = option->fpga_mmc_sdr200_clk;
	} else if (CHK_MMC_52M(sd_card) || CHK_MMC_DDR52(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = option->asic_mmc_52m_clk;
		else
			sd_card->sd_clock = option->fpga_mmc_52m_clk;
	} else if (CHK_MMC_26M(sd_card)) {
		if (cr->asic_code)
			sd_card->sd_clock = 48;
		else
			sd_card->sd_clock = CLK_50;
	}
}
EXPORT_SYMBOL_GPL(sd_choose_proper_clock);

inline void sd_set_error_code(struct card_reader *cr, struct card *c,
								u8 err_code)
{
	struct sd_info *sd_card = c->extra;

	sd_card->err_code |= err_code;
}
EXPORT_SYMBOL_GPL(sd_set_error_code);

inline void sd_clear_error_code(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;

	sd_card->err_code = 0;
}
EXPORT_SYMBOL_GPL(sd_clear_error_code);

inline int sd_check_error_code(struct card_reader *cr, struct card *c,
								u8 err_code)
{
	struct sd_info *sd_card = c->extra;

	return (sd_card->err_code & err_code);
}
EXPORT_SYMBOL_GPL(sd_check_error_code);

#define IMPLEMENT_SD_PROTOCOL

int sd_check_data0_status(struct card_reader *cr, struct card *card)
{
	u8 stat;

	cr_rd_reg(cr, REG_SD_BUS_STAT, &stat);

	if (!(stat & SD_DAT0_STATUS)) {
		sd_set_error_code(cr, card, SD_BUSY);
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(sd_check_data0_status);

/*
 * sd_prepare_reset -- Prepare for card reset operation
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_prepare_reset(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->prepare_reset)
		return cr->sd_ops_prot->prepare_reset(cr, c);
	else
		return sd_comm_prepare_reset(cr, c);
}
EXPORT_SYMBOL_GPL(sd_prepare_reset);

/*
 * sd_set_byte_count_per_blk - Set byte count for each block
 * @cr: points to card reader structure
 * @bcpb: byte count for each block
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_byte_count_per_blk(struct card_reader *cr, int bcpb)
{
	if (cr->sd_ops_prot->set_byte_count_per_blk)
		return  cr->sd_ops_prot->set_byte_count_per_blk(cr, bcpb);
	else
		return sd_comm_set_byte_count_per_blk(cr, bcpb);
}

/*
 * sd_set_blk_count - Set block count
 * @cr: points to card reader structure
 * @count: block count
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_blk_count(struct card_reader *cr, int count)
{
	if (cr->sd_ops_prot->set_blk_count)
		return  cr->sd_ops_prot->set_blk_count(cr, count);
	else
		return sd_comm_set_blk_count(cr, count);
}

/*
 * sd_change_bank_voltage - Change bank voltage
 * @cr: points to card reader structure
 * @volt: voltage to be changed to, the value of which can be
 * SD_IO_3V3/SD_IO_1V8/SD_IO_1V9
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_change_bank_voltage(struct card_reader *cr, u8 volt)
{
	if (cr->sd_ops_prot->chg_bank_volt)
		return cr->sd_ops_prot->chg_bank_volt(cr, volt);
	else
		return sd_comm_change_bank_voltage(cr, volt);
}
EXPORT_SYMBOL_GPL(sd_change_bank_voltage);

inline int sd_check_spec(struct card_reader *cr, struct card *c)
{
	return sd_comm_check_spec(cr, c);
}
EXPORT_SYMBOL_GPL(sd_check_spec);


/*
 * sd_set_clock_divider - Set clock dividor
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_clock_divider(struct card_reader *cr, u8 divider)
{
	if (cr->sd_ops_prot->set_clk_div)
		return cr->sd_ops_prot->set_clk_div(cr, divider);
	else
		return sd_comm_set_clock_divider(cr, divider);

}
EXPORT_SYMBOL_GPL(sd_set_clock_divider);

/* sd_voltage_switch - Perform Voltage Switch behaviors
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_voltage_switch(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->volt_switch)
		return cr->sd_ops_prot->volt_switch(cr, c);
	 else
		return sd_comm_voltage_switch(cr, c);
}
EXPORT_SYMBOL_GPL(sd_voltage_switch);

/* sd_get_crc7 - Get value of CRC7
 * @cr: points to card reader structure
 * @buf: where to store CRC7
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_get_crc7(struct card_reader *cr, u8 *buf)
{
	if (cr->sd_ops_prot->get_crc7)
		return cr->sd_ops_prot->get_crc7(cr, buf);

	else
		return sd_comm_get_crc7(cr, buf);
}


/* sd_clear_sd_error - Clear SD error
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
inline void sd_clear_sd_error(struct card_reader *cr)
{
	if (cr->sd_ops_prot->clr_err)
		cr->sd_ops_prot->clr_err(cr);
	else
		sd_comm_clear_sd_error(cr);
}
EXPORT_SYMBOL_GPL(sd_clear_sd_error);

/* sd_current_over_limit400_handler - Handle scenario that current is over 400mA
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int sd_current_over_limit400_handler(struct card_reader *cr)
{
	if (cr->sd_ops_prot->current_over_limit400_handler)
		return cr->sd_ops_prot->current_over_limit400_handler(cr);
	else
		return sd_comm_current_over_limit400_handler(cr);
}

/* sd_check_crc16_error - Check if there is error in CRC16
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int sd_check_crc16_error(struct card_reader *cr)
{
	if (cr->sd_ops_prot->check_crc16_err)
		return cr->sd_ops_prot->check_crc16_err(cr);
	else
		return sd_comm_check_crc16_err(cr);
}

/* sd_set_clk_pol_ddr_mod - Choose SDCLK polarity solution for DDR mode
 * @cr: points to card reader structure
 * @sol: polarity solution: analog/digital/driver solution
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_clk_pol_ddr_mod(struct card_reader *cr,
						enum sd_comm_polarity_sol sol)
{
	if (cr->sd_ops_prot->set_clk_pol_ddr_mod)
		return cr->sd_ops_prot->set_clk_pol_ddr_mod(cr, sol);
	else
		return sd_comm_set_clk_pol_ddr_mod(cr, sol);
}

/* sd_set_sample_push_timing - Set sample push timing
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_sample_push_timing(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->set_sample_push_timing)
		return cr->sd_ops_prot->set_sample_push_timing(cr, c);
	else
		return sd_comm_set_sample_push_timing(cr, c);
}
EXPORT_SYMBOL_GPL(sd_set_sample_push_timing);

/* sd_init_power - Initalise power to sd card
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_init_power(struct card_reader *cr)
{
	if (cr->sd_ops_prot->init_power) {
		return cr->sd_ops_prot->init_power(cr);
	} else {
		rts_trace(TRACE_ERROR,
			"sd_ops_prot->init_power is not assigned\n");
		return -EINVAL;
	}
}
EXPORT_SYMBOL_GPL(sd_init_power);

/* sd_sd30_drive_sel - SD3.0 driving strength selection
 * @cr: points to card reader structure
 * @select: driving strength
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_sd30_drive_sel(struct card_reader *cr, u8 voltage)
{
	if (cr->sd_ops_prot->sd30_drive_sel)
		return cr->sd_ops_prot->sd30_drive_sel(cr, voltage);
	else
		return sd_comm_sd30_drive_sel(cr, voltage);
}
EXPORT_SYMBOL_GPL(sd_sd30_drive_sel);


/* sd_change_phase - Change SD phases
 * @cr: points to card reader structure
 * @c: points to card struct
 * @sample_point: Sample point
 * @tune_dir: direction to tune, receive/transmit?
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_change_phase(struct card_reader *cr, struct card *c,
				u8 sample_point, u8 tune_dir)
{
	if (cr->sd_ops_prot->change_phase)
		return cr->sd_ops_prot->change_phase(cr, c, sample_point,
				tune_dir);
	else
		return sd_comm_change_phase(cr, c, sample_point, tune_dir);
}
EXPORT_SYMBOL_GPL(sd_change_phase);

/* sd_wait_data_idle - Wait for the IDLE state of DATA state machine
 * @cr: points to card reader structure
 *
 * Returns: STATUS_SUCCESS for success, otherwise STATUS_TIMEDOUT
 */
int sd_wait_data_idle(struct card_reader *cr)
{
	if (cr->sd_ops_prot->wait_data_idle)
		return cr->sd_ops_prot->wait_data_idle(cr);
	else
		return sd_comm_wait_data_idle(cr);
}
EXPORT_SYMBOL_GPL(sd_wait_data_idle);

/* sd_is_mechanical_wp - Check SD Machanical Write-Protect Switch
 * @cr: points to card reader structure
 *
 * Returns: 1  for WP, otherwise non-WP
 */
static int sd_is_mechanical_wp(struct card_reader *cr)
{
	if (cr->sd_ops_prot->is_mechanical_wp)
		return cr->sd_ops_prot->is_mechanical_wp(cr);
	else
		return (cr_ioread32(cr, BIPR) & SD_WRITE_PROTECT);
}

/* sd_set_init_para - Set inital parameters including "set sample pushing
 * timing" and "switch clock"
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_set_init_para(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->set_init_para)
		return cr->sd_ops_prot->set_init_para(cr, c);
	else
		return sd_comm_set_init_para(cr, c);
}
EXPORT_SYMBOL_GPL(sd_set_init_para);

/* sd_mmc_switch_timing_bus - Switch MMC card bus timing
 * @cr: points to card reader structure
 * @c: points to card struct
 * @switch_ddr: whether switch bus timing for DDR mode
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_mmc_switch_timing_bus(struct card_reader *cr,
					struct card *c, int switch_ddr)
{
	if (cr->sd_ops_prot->mmc_switch_timing_bus)
		return cr->sd_ops_prot->mmc_switch_timing_bus(cr, c,
				switch_ddr);
	else
		return sd_comm_mmc_switch_timing_bus(cr, c, switch_ddr);
}
EXPORT_SYMBOL_GPL(sd_mmc_switch_timing_bus);

/* sd_dummy_clock - Trigger dummy clock to make the clock stable before
 * real operation
 * @cr: points to card reader structure
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_dummy_clock(struct card_reader *cr)
{
	if (cr->sd_ops_prot->dummy_clk)
		return cr->sd_ops_prot->dummy_clk(cr);
	else
		return sd_comm_dummy_clock(cr);
}
EXPORT_SYMBOL_GPL(sd_dummy_clock);

/* sd_auto_tune_clk - Automatically tune clock
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
int sd_auto_tune_clock(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->auto_tune_clk)
		return cr->sd_ops_prot->auto_tune_clk(cr, c);
	else
		return sd_comm_auto_tune_clock(cr, c);
}
EXPORT_SYMBOL_GPL(sd_auto_tune_clock);

static void add_cmd_4_legacy_cmd(struct card_reader *cr, struct card *card,
							struct sd_command *cmd)
{
	u16 addr;
	u32 arg = cmd->arg;
	u8 rsp_type = cmd->rsp_type;

	rts_trace(TRACE_SD, "SD/MMC CMD %d\n", cmd->opcode);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD0, 0xFF, (u8)
		(0x40 | cmd->opcode));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD1, 0xFF, (u8)(arg >> 24));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD2, 0xFF, (u8)(arg >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD3, 0xFF, (u8)(arg >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD4, 0xFF, (u8)arg);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, rsp_type);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
						0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER,	0xFF,
					SD_TM_CMD_RSP | SD_TRANSFER_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
		     SD_TRANSFER_END | SD_STAT_IDLE,
		     SD_TRANSFER_END | SD_STAT_IDLE);

	if (rsp_type == SD_RSP_TYPE_R2) {
		/* Read data from ping-pong buffer */
		for (addr = PPBUF_BASE2; addr < PPBUF_BASE2 + 16; addr++)
			cr_add_cmd(cr, READ_REG_CMD, addr, 0, 0);
	} else if (rsp_type != SD_RSP_TYPE_R0) {
		/* Read data from SD_CMDx registers */
		for (addr = REG_SD_CMD0; addr <= REG_SD_CMD4; addr++)
			cr_add_cmd(cr, READ_REG_CMD, addr, 0, 0);
	}

	cr_add_cmd(cr, READ_REG_CMD, REG_SD_STAT1, 0, 0);
}


int sd_do_send_cmd_legacy(struct card_reader *cr, struct card *card,
		struct sd_command *cmd)
{
	int ret = 0;
	u8 *ptr = NULL;

	cr_init_cmd(cr);

	add_cmd_4_legacy_cmd(cr, card, cmd);

	ret = cr_send_cmd(cr, SD_CARD, cmd->timeout);
	if (ret) {
		cmd->error = ret;
		return ret;
	}

	if (cmd->rsp_type != SD_RSP_TYPE_R0) {
		ptr = cr_get_cmd_data(cr) + 1;
		memcpy(cmd->rsp, ptr, sizeof(cmd->rsp));
	}

	return ret;
}

static inline int sd_retry_by_command_legacy(struct card_reader *cr,
				struct card *card, struct sd_command *cmd)
{
	return 0;
}

int sd_process_cmd_error_4_legacy(struct card_reader *cr,
				struct card *card, struct sd_command *cmd)
{
	int error = cmd->error;
	int rsp_type = cmd->rsp_type;
	u8 val;

	cr_rd_reg(cr, REG_SD_STAT1, &val);
	rts_trace(TRACE_SD, "SD_STAT1: 0x%x\n", val);

	cr_rd_reg(cr, REG_SD_STAT2, &val);
	rts_trace(TRACE_SD, "SD_STAT2: 0x%x\n", val);

	if (val & SD_RSP_80CLK_TIMEOUT) {
		sd_clear_sd_error(cr);
		sd_set_error_code(cr, card, SD_RSP_TIMEOUT);
		return error;
	}

	cr_rd_reg(cr, REG_SD_BUS_STAT, &val);
	rts_trace(TRACE_SD, "SD_BUS_STAT: 0x%x\n", val);

	if (error == -ETIMEDOUT) {
		if (rsp_type & SD_WAIT_BUSY_END) {
			if (sd_check_data0_status(cr, card)) {
				sd_clear_sd_error(cr);
				return error;
			}
		} else {
			sd_set_error_code(cr, card, SD_TO_ERR);
		}
	}

	sd_clear_sd_error(cr);

	return error;
}

static int sd_check_legacy_responce(struct card_reader *cr,
				struct card *card, struct sd_command *cmd)
{
	struct sd_info *sd_card = card->extra;
	u8 rsp_type = cmd->rsp_type;
	u8 *rsp = (u8 *) cmd->rsp;
	u8 sd_status1;

	cr_rd_reg(cr, REG_SD_STAT1, &sd_status1);

	/* Check (Start,Transmission) bit of Response */
	if ((rsp[0] & 0xC0) != 0) {
		sd_set_error_code(cr, card, SD_STS_ERR);
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Check CRC7 */
	if (!(rsp_type & SD_NO_CHECK_CRC7)) {
		if (sd_status1 & SD_CRC7_ERR) {
			sd_set_error_code(cr, card, SD_CRC_ERR);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	/* Check Status */
	if ((rsp_type == SD_RSP_TYPE_R1) ||
		(rsp_type == SD_RSP_TYPE_R1b)) {
		if ((cmd->opcode != SEND_RELATIVE_ADDR) &&
			(cmd->opcode != SEND_IF_COND)) {
			/* SEND_EXT_CSD= 8, But Never call SD_SendCMDGetRSP
			 * by SEND_EXT_CSD command. We shall check OUT_OF_RANGE
			 * error except STOP_TRANSMISSION command */
			if (cmd->opcode != STOP_TRANSMISSION) {
				if (rsp[1] & 0x80)
					TRACE_RET(cr, STATUS_FAIL);
			}
			/* exclude bit25 CARD_IS_LOCKED */
			if (rsp[1] & 0x7D) {
				rts_trace(TRACE_SD, "rsp[1]: 0x%02x\n", rsp[1]);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if (rsp[2] & 0xFF) {
				rts_trace(TRACE_SD, "rsp[2]: 0x%02x\n", rsp[2]);
				TRACE_RET(cr, STATUS_FAIL);
			}
			if (rsp[3] & 0x80) {
				rts_trace(TRACE_SD, "rsp[3]: 0x%02x\n", rsp[3]);
				TRACE_RET(cr, STATUS_FAIL);
			}
			/* When access Kington MMC4.0 cards, we shall check
			 * "READY_FOR_DATA" bit in "Card Status" of R1 Those
			 * cards will report READY_FOR_DATA=0 after
			 * MULTIPLE_WRITE_BLOCK command. Especially after "STOP"
			 * transfer from "MULTIPLE_READ_BLOCK" status(). */
			/* Get "READY_FOR_DATA" bit */
			if (rsp[3] & 0x01)
				sd_card->sd_data_buf_ready = 1;
			else
				sd_card->sd_data_buf_ready = 0;
		}
	}

	return 0;
}



int sd_retry_by_responce(struct card_reader *cr, struct sd_command *cmd)
{
	u8 sd_status1;

	if (cmd->rsp_type & SD_NO_CHECK_CRC7)
		return 0;

	if (!cmd->rsp)
		return 0;


	cr_rd_reg(cr, REG_SD_STAT1, &sd_status1);
	if (sd_status1 & SD_CRC_ERR) {
		if (cmd->opcode == WRITE_MULTIPLE_BLOCK)
			return 0;
		else
			return 1;
	}

	return 0;
}


static int sd_send_legacy_ccmd(struct card_reader *cr, struct card *card,
		struct sd_command *cmd)
{
	int ret;
	unsigned int rty_cnt = 0;

	sd_clear_error_code(cr, card);

	/* for ACMD, send CMD55 first */
	if (cmd->flags & SD_APP_CMD) {
		struct sd_command app_cmd;
		struct sd_info *sd_card = card->extra;

		INIT_SD_CMD(app_cmd, APP_CMD, sd_card->sd_addr,
						0, SD_RSP_TYPE_R1, 0);
		ret = sd_send_legacy_ccmd(cr, card, &app_cmd);
		if (ret)
			return ret;
	}

	do {
		ret = sd_do_send_cmd_legacy(cr, card, cmd);
		if (ret) {
			sd_process_cmd_error_4_legacy(cr, card, cmd);
			break;
		}

		if (cmd->rsp_type == SD_RSP_TYPE_R0)
			break;

		ret = sd_check_legacy_responce(cr, card, cmd);
		if (ret) {
			int need_retry = sd_retry_by_responce(cr, cmd);
			if (need_retry) {
				if (rty_cnt++ < cmd->max_retries) {
					msleep(20);
					continue;
				}
			}
		}

		break;
	} while (1);

	return ret;
}

static int sd_send_ccmd(struct card_reader *cr, struct card *card,
		struct sd_command *cmd)
{
	struct sd_info *sd_card = card->extra;

	if (CHK_SD_UHSII(sd_card)) {
		return sd40_send_sdtran_ccmd(card, cmd);
	} else {
		return sd_send_legacy_ccmd(cr, card, cmd);
	}

}

int sd_send_cmd(struct card_reader *cr, struct card *card,
		struct sd_command *cmd)
{
	rts_trace(TRACE_SD, "SD/MMC CMD %d, arg = 0x%08x\n",
					cmd->opcode & 0x3F, cmd->arg);

	if (cmd->flags & SD_CMD_WITH_DATA) {
		return sd_send_data_cmd(card, cmd);
	} else {
		if (cr->sd_ops_prot->send_ccmd)
			return cr->sd_ops_prot->send_ccmd(cr, card, cmd);
		else
			return sd_send_ccmd(cr, card, cmd);
	}
}
EXPORT_SYMBOL_GPL(sd_send_cmd);

static int set_ppbuf_4_legacy_dcmd(struct card *card, struct sd_command *cmd)
{
	struct card_reader *cr = card->cm->cr;
	int ret = 0;

	if (TO_SD == cmd->data.trans_dir) {
		if (cmd->data.buf_len) {
			ret = cr_wr_ppbuf(cr, cmd->data.buf, cmd->data.buf_len);
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return 0;
}

static int get_ppbuf_4_legacy_dcmd(struct card *card, struct sd_command *cmd)
{
	struct card_reader *cr = card->cm->cr;
	int ret = 0;

	if (FROM_SD == cmd->data.trans_dir) {
		if (cmd->data.buf_len) {
			ret = cr_rd_ppbuf(cr, cmd->data.buf, cmd->data.buf_len);
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return 0;
}


static void add_cmd_4_legacy_dcmd(struct card *card, struct sd_command *cmd)
{
	u32 arg = cmd->arg;
	u32 bytes = cmd->data.bytes;
	u16 blocks = cmd->data.blocks;
	struct card_reader *cr = card->cm->cr;

	cr_init_cmd(cr);

	rts_trace(TRACE_SD, "SD/MMC CMD %d\n", cmd->opcode & (~0x40));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD0, 0xFF,
					(u8)(0x40 | cmd->opcode));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD1, 0xFF, (u8)(arg >> 24));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD2, 0xFF, (u8)(arg >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD3, 0xFF, (u8)(arg >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD4, 0xFF, (u8)arg);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_L, 0xFF, (u8)bytes);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_H, 0xFF,
							(u8)(bytes >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_L, 0xFF, (u8)blocks);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_H, 0xFF,
							(u8)(blocks >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG1, 0x03, cmd->data.bus_width);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, SD_CALCULATE_CRC7
			| SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END
			| SD_CHECK_CRC7 | SD_RSP_LEN_6);

	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
					0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
				cmd->data.trans_mode | SD_TRANSFER_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
			SD_TRANSFER_END, SD_TRANSFER_END);
}

static int sd_send_legacy_dcmd(struct card *card, struct sd_command *cmd)
{
	struct card_reader *cr = card->cm->cr;
	int ret = 0;

	sd_clear_error_code(cr, card);

	if (!cmd->data.buf)
		cmd->data.buf_len = 0;
	if (cmd->flags & SD_APP_CMD) {
		if (sd_app_cmd(card))
			TRACE_RET(cr, STATUS_FAIL);
	}

	/* This function can't r/w data more than PPBUF_LEN bytes
	 * if we need fetch data back to host
	 */
	if (cmd->data.buf_len && (cmd->data.buf_len > PPBUF_LEN))
		TRACE_RET(cr, STATUS_FAIL);

	if (set_ppbuf_4_legacy_dcmd(card, cmd))
		TRACE_RET(cr, STATUS_FAIL);

	add_cmd_4_legacy_dcmd(card, cmd);

	ret = cr_send_cmd(cr, SD_CARD, cmd->timeout);
	if (ret) {
		sd_print_register(cr);
		if (ret == -ETIMEDOUT)
			sd_clear_status(card);
		TRACE_RET(cr, STATUS_FAIL);

	}

	if (get_ppbuf_4_legacy_dcmd(card, cmd))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

int sd_send_data_cmd(struct card *card, struct sd_command *cmd)
{
	struct sd_info *sd_card = card->extra;

	if (CHK_SD_UHSII(sd_card)) {
		return sd40_send_sdtran_dcmd(card, cmd);
	} else {
		return sd_send_legacy_dcmd(card, cmd);
	}
}

inline void sd_comm_clear_sd_error(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_CARD_STOP, SD_STOP | SD_CLR_ERR,
					SD_STOP | SD_CLR_ERR);
	cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
}

static int sd_comm_voltage_switch(struct card_reader *cr, struct card *card)
{
	struct cr_option *option = &(cr->option);
	int retval;
	struct sd_command cmd;
	u8 stat;

	/* Enable SD clock toggle */
	cr_wr_reg(cr, REG_SD_BUS_STAT,
			SD_CLK_TOGGLE_EN | SD_CLK_FORCE_STOP,
			SD_CLK_TOGGLE_EN);
	INIT_SD_CMD(cmd, VOLTAGE_SWITCH, 0, 0, SD_RSP_TYPE_R1, 0);
	retval = sd_send_cmd(cr, card, &cmd);
	if (retval != STATUS_SUCCESS)
		TRACE_GOTO(cr, SWITCH_VOLTAGE_FAIL);

	udelay(option->sd_voltage_switch_delay);

	/* SD_CMD, SD_DAT3~0 should be drived to low by card;
	 * If either one of SD_CMD,SD_DAT3~0 is not low,
	 * abort the voltage switch sequence; */
	cr_rd_reg(cr, REG_SD_BUS_STAT, &stat);
	if (stat & (SD_CMD_STATUS | SD_DAT3_STATUS |
			SD_DAT2_STATUS | SD_DAT1_STATUS |
			SD_DAT0_STATUS))
		TRACE_GOTO(cr, SWITCH_VOLTAGE_FAIL);

	rts_trace(TRACE_SD, "step1 pass:SD_BUS_STAT: 0x%x\n", stat);

	/* Stop toggle SD clock */
	cr_wr_reg(cr, REG_SD_BUS_STAT, 0xFF, SD_CLK_FORCE_STOP);

	if(card->reset_flags & UHSI_1V9_RERTY) {
		retval = sd_change_bank_voltage(cr, SD_IO_1V9);
	} else {
		retval = sd_change_bank_voltage(cr, SD_IO_1V8);
	}

	if (retval != STATUS_SUCCESS)
		TRACE_GOTO(cr, SWITCH_VOLTAGE_FAIL);
	msleep(50);
	/* Toggle SD clock again */
	cr_wr_reg(cr, REG_SD_BUS_STAT, 0xFF, SD_CLK_TOGGLE_EN);
	msleep(10);

	/* SD_CMD, SD_DAT3~0 should be pulled high by host */
	cr_rd_reg(cr, REG_SD_BUS_STAT, &stat);
	if ((stat & (SD_CMD_STATUS | SD_DAT3_STATUS |
			SD_DAT2_STATUS | SD_DAT1_STATUS |
			SD_DAT0_STATUS)) !=
			(SD_CMD_STATUS | SD_DAT3_STATUS |
			SD_DAT2_STATUS | SD_DAT1_STATUS |
			SD_DAT0_STATUS)) {
		rts_trace(TRACE_SD, "SD_BUS_STAT: 0x%x\n", stat);
		cr_wr_reg(cr, REG_SD_BUS_STAT,
				SD_CLK_TOGGLE_EN | SD_CLK_FORCE_STOP, 0);
		cr_wr_reg(cr, REG_CARD_CLK_EN, 0xFF, 0);
		TRACE_GOTO(cr, SWITCH_VOLTAGE_FAIL);
	}

	rts_trace(TRACE_SD, "step2 pass:SD_BUS_STAT: 0x%x\n", stat);

	/* Stop toggle SD clock in idle */
	cr_wr_reg(cr, REG_SD_BUS_STAT, SD_CLK_TOGGLE_EN | SD_CLK_FORCE_STOP, 0);

	return 0;

SWITCH_VOLTAGE_FAIL:

	cr_wr_reg(cr, REG_SD_BUS_STAT, SD_CLK_TOGGLE_EN | SD_CLK_FORCE_STOP, 0);
	rts_trace(TRACE_SD, "switch fail,SD_BUS_STAT: 0x%x\n", stat);
	return STATUS_FAIL;

}


int sd_check_csd(struct card_reader *cr, struct card *c, char check_wp)
{
	struct sd_info *sd_card = c->extra;
	int retval;
	int i;
	u8 csd_ver;
	struct sd_command cmd;
	u8 *csd = sd_card->raw_csd;

	for (i = 0; i < 6; i++) {
		if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
			sd_set_error_code(cr, c, SD_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}
		INIT_SD_CMD(cmd, SEND_CSD, sd_card->sd_addr,
						0, SD_RSP_TYPE_R2, 16);
		retval = sd_send_cmd(cr, c, &cmd);
		if (retval == STATUS_SUCCESS)
			break;
	}

	if (i == 6)
		TRACE_RET(cr, STATUS_FAIL);

	memcpy(sd_card->raw_csd, cmd.rsp + 1, 15);

	/* Get CRC7 */
	retval = sd_get_crc7(cr, sd_card->raw_csd + 15);
	if (STATUS_SUCCESS != retval)
		TRACE_RET(cr, retval);

	rts_trace(TRACE_SD, "CSD Response:\n");
	hex_dump(sd_card->raw_csd, 16);

	/* Get CSD Version */
	csd_ver = (csd[0] & 0xc0) >> 6;
	rts_trace(TRACE_SD, "csd_ver = %d\n", csd_ver);

	if (!CHK_SD_UHSII(sd_card)) {
		u8 trans_speed;
		trans_speed = csd[3];
		if ((trans_speed & 0x07) == 0x02) {	/* 10Mbits/s */
			if ((trans_speed & 0xf8) >= 0x30) {		/* >25Mbits/s */
				if (cr->asic_code)
					sd_card->sd_clock = 47;
				else
					sd_card->sd_clock = CR_CLK_50;
			} else if ((trans_speed & 0xf8) == 0x28) {	/* 20Mbits/s */
				if (cr->asic_code)
					sd_card->sd_clock = 39;
				else
					sd_card->sd_clock = CR_CLK_40;
			} else if ((trans_speed & 0xf8) == 0x20) {	/* 15Mbits/s */
				if (cr->asic_code)
					sd_card->sd_clock = 29;
				else
					sd_card->sd_clock = CR_CLK_30;
			} else if ((trans_speed & 0xf8) >= 0x10) {	/* 12Mbits/s */
				if (cr->asic_code)
					sd_card->sd_clock = 23;
				else
					sd_card->sd_clock = CR_CLK_20;
			} else if ((trans_speed & 0x08) >= 0x08) {	/* 10Mbits/s */
				if (cr->asic_code)
					sd_card->sd_clock = 19;
				else
					sd_card->sd_clock = CR_CLK_20;
			} else {
				TRACE_RET(cr, STATUS_FAIL);
			}
		}
	}

	if (CHK_MMC_SECTOR_MODE(sd_card)) {
		sd_card->capacity = 0;
	} else {
		/* For High-Capacity Card, CSD_STRUCTURE always be "0x1" */
		if ((!CHK_SD_HCXC(sd_card)) || (csd_ver == 0)) {
			/* Calculate total sector according to C_SIZE,
			 * C_SIZE_MULT & READ_BL_LEN */
			u8 blk_size, c_size_mult;
			u16 c_size;
			/* Get READ_BL_LEN */
			blk_size = csd[5] & 0x0F;
			/* Get C_SIZE */
			c_size =  ((u16)(csd[6] & 0x03) << 10)
					+ ((u16)csd[7] << 2)
					+ ((u16)(csd[8] & 0xC0) >> 6);
			/* Get C_SIZE_MUL */
			c_size_mult = (u8)((csd[9] & 0x03) << 1);
			c_size_mult += (csd[10] & 0x80) >> 7;
			/* Calculate total Capacity */
			sd_card->capacity = (((u32)(c_size + 1)) *
				(1 << (c_size_mult + 2))) << (blk_size - 9);
		} else {
			/* High Capacity Card and Use CSD2.0 Version */
			u32 total_sector = 0;
			total_sector = (((u32)csd[7] & 0x3f) << 16) |
				((u32)csd[8] << 8) | (u32)csd[9];
			/* Total Capacity= (C_SIZE+1) * 512K Byte =
			 * (C_SIZE+1)K Sector,1K = 1024 Bytes */
			sd_card->capacity = (total_sector + 1) << 10;
		}
	}

	/* We need check Write-Protected Status by Field PERM WP or TEMP WP */
	if (check_wp) {
		/* PERM WP or TEMP WP These field located at the same offset no
		 * matter SD2.0 or earlier. */
		if (csd[14] & 0x30) {
			set_card_flags(c, WP);
		}
		rts_trace(TRACE_SD, "CSD WP Status: 0x%x\n", csd[14]);
	}

	c->capacity = sd_card->capacity;
	rts_trace(TRACE_SD, "SD capacity: %d\n", c->capacity);

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(sd_check_csd);

int sd_select_card(struct card_reader *cr, struct card *c, int select)
{
	struct sd_info *sd_card = c->extra;
	int ret;
	struct sd_command cmd;

	if (select) {
		INIT_SD_CMD(cmd, SELECT_CARD, sd_card->sd_addr,
				0, SD_RSP_TYPE_R1, 0);
		ret = sd_send_cmd(cr, c, &cmd);
	} else {
		INIT_SD_CMD(cmd, DESELECT_CARD, 0,
				0, SD_RSP_TYPE_R0, 0);
		ret = sd_send_cmd(cr, c, &cmd);
	}

	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return ret;
}
EXPORT_SYMBOL_GPL(sd_select_card);

int sd_update_lock_status(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	int ret;
	u8 rsp[16];

	ret = sd_send_status(c, rsp);
	if (ret)
		TRACE_RET(cr, STATUS_FAIL);

	if (rsp[1] & 0x02)
		sd_card->sd_lock_status |= SD_LOCKED;
	else
		sd_card->sd_lock_status &= ~SD_LOCKED;

	rts_trace(TRACE_SD, "sd_lock_status = 0x%x\n", sd_card->sd_lock_status);

	/* LOCK_UNLOCK_FAILED */
	if (rsp[1] & 0x01)
		TRACE_RET(cr, STATUS_FAIL);

	return ret;
}
EXPORT_SYMBOL_GPL(sd_update_lock_status);

int sd_get_raw_scr(struct card_reader *cr, struct card *card, int bus_width)
{
	struct sd_info *sd_card = card->extra;
	struct sd_command cmd;
	u8 buf[8];

	INIT_SD_CMD(cmd, 0x40 | SEND_SCR, 0x0, SD_APP_CMD, 0, 0);
	init_sd_dcmd_data(cmd, buf, (u8)(sizeof(buf)), (u8)(sizeof(buf)), 1,
			(u8)bus_width, SD_TM_NORMAL_READ, FROM_SD, 250);

	if (sd_send_cmd(cr, card, &cmd)) {
		sd_clear_sd_error(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}

	memcpy(sd_card->raw_scr, buf, 8);
	return 0;
}

int sd_comm_check_spec(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	u8 spec_ver = sd_card->raw_scr[0] & 0xF;

	if (spec_ver == 0)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

static int sd_ge_card(struct sd_extr_reg *extr)
{
	const char *name = "GAMMAEMIDMEDIA";
	int i;

	for (i = 0; i < 15; i++) {
		if (name[i] != extr->fn[i])
			return 0;
	}

	return 1;
}

int sd_init_extr_param(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct sd_ge *ge = &sd_card->ge;
	struct sd_extr_header exth;
	struct sd_extr_reg extr;
	u32 arg = sd_extr_arg_init_single(0, 0, 0, 0, 512);
	int err = 0, i = 0;
	u8 *data;

	data = kzalloc(512, GFP_KERNEL);
	if (!data)
		return -ENOMEM;
	err = sd_read_extr_single(cr, card, arg, data);
	if (err)
		TRACE_RET(cr, STATUS_FAIL);

	exth.version = cr_get_le16(&data[0]);
	exth.len = cr_get_le16(&data[2]);
	exth.num = data[4];

	extr.pfc = cr_get_le16(&data[38]);
	memcpy(extr.fn, &data[40], 16);
	extr.next_addr = cr_get_le16(&data[56]);
	extr.reg_set_num = data[58];

	if (extr.reg_set_num > SD_EXTR_SET_NUM_MAX) {
		rts_trace(TRACE_SD, "error: reg_set_num 0x%02x to large\n",
			extr.reg_set_num);
		extr.reg_set_num = SD_EXTR_SET_NUM_MAX;
	}

	for (i = 0; i < extr.reg_set_num; i++) {
		/* Q: why here is big endian? */
		extr.reg_set[i] = cr_get_le32(&data[60 + 4 * i]);
		rts_trace(TRACE_SD, "extr.reg_set[i]=%#x\n", extr.reg_set[i]);
	}

	for (i = 0; i < exth.len; i += 4) {
		rts_trace(TRACE_SD, "%02x %02x %02x %02x\n",
			data[i], data[i+1], data[i+2], data[i+3]);
	}

	if (sd_ge_card(&extr)) {
		rts_trace(TRACE_SD, "extr.reg_set[0]=%#x\n", extr.reg_set[0]);
		ge->ge_card = 1;
		ge->fno = (extr.reg_set[0] >> SD_EXTR_SET_MEMFNO_SHIFT) &
			SD_EXTR_SET_MEMFNO_MASK;
		ge->start_addr = extr.reg_set[0] & SD_EXTR_SET_ADDR_MASK;
		rts_trace(TRACE_SD, "Gamma/EMID card found, fno: %d, start addr: 0x%04x\n",
			ge->fno, ge->start_addr);
	}

	kfree(data);

	return 0;
}

int sd_check_sdio_card_exist(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_command cmd;

	if ((cr_detect_card(cr, SD_ID)))
		return 0;

	INIT_SD_CMD(cmd, IO_SEND_OP_COND, 0, 0, SD_RSP_TYPE_R4, 5);
	if (!sd_send_cmd(cr, card, &cmd)) {
		int func = (cmd.rsp[1] >> 4) & 0x07;

		if (func) {
			rts_trace(TRACE_SD, "SDIO CARD func num:%d\n", func);
			SET_SDIO_EXIST(cr);
			cr->sdio_func_init = 0;
			cr_wr_reg(cr, 0xFE0F, 0x10, 0x00);
			return 1;
		}
	}

	return 0;
}
EXPORT_SYMBOL_GPL(sd_check_sdio_card_exist);

int sd_comm_query_switch_result(struct card_reader *cr,
					u8 func_group, u8 func_to_switch,
							u8 *buf, int buf_len)
{
	u8 support_mask = 0, query_switch = 0, switch_busy = 0;
	int support_offset = 0, query_switch_offset = 0, check_busy_offset = 0;

	if (func_group == SD_FUNC_GROUP_1) {
		support_offset = FUNCTION_GROUP1_SUPPORT_OFFSET;
		query_switch_offset = FUNCTION_GROUP1_QUERY_SWITCH_OFFSET;
		check_busy_offset = FUNCTION_GROUP1_CHECK_BUSY_OFFSET;

		switch (func_to_switch) {
		case HS_SUPPORT:
			support_mask = HS_SUPPORT_MASK;
			query_switch = HS_QUERY_SWITCH_OK;
			switch_busy = HS_SWITCH_BUSY;
			break;

		case SDR50_SUPPORT:
			support_mask = SDR50_SUPPORT_MASK;
			query_switch = SDR50_QUERY_SWITCH_OK;
			switch_busy = SDR50_SWITCH_BUSY;
			break;

		case SDR104_SUPPORT:
			support_mask = SDR104_SUPPORT_MASK;
			query_switch = SDR104_QUERY_SWITCH_OK;
			switch_busy = SDR104_SWITCH_BUSY;
			break;

		case DDR50_SUPPORT:
			support_mask = DDR50_SUPPORT_MASK;
			query_switch = DDR50_QUERY_SWITCH_OK;
			switch_busy = DDR50_SWITCH_BUSY;
			break;

		default:
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else if (func_group == SD_FUNC_GROUP_3) {
		support_offset = FUNCTION_GROUP3_SUPPORT_OFFSET;
		query_switch_offset = FUNCTION_GROUP3_QUERY_SWITCH_OFFSET;
		check_busy_offset = FUNCTION_GROUP3_CHECK_BUSY_OFFSET;

		switch (func_to_switch) {
		case DRIVING_TYPE_A:
			support_mask = DRIVING_TYPE_A_MASK;
			query_switch = TYPE_A_QUERY_SWITCH_OK;
			switch_busy = TYPE_A_SWITCH_BUSY;
			break;

		case DRIVING_TYPE_C:
			support_mask = DRIVING_TYPE_C_MASK;
			query_switch = TYPE_C_QUERY_SWITCH_OK;
			switch_busy = TYPE_C_SWITCH_BUSY;
			break;

		case DRIVING_TYPE_D:
			support_mask = DRIVING_TYPE_D_MASK;
			query_switch = TYPE_D_QUERY_SWITCH_OK;
			switch_busy = TYPE_D_SWITCH_BUSY;
			break;

		default:
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else if (func_group == SD_FUNC_GROUP_4) {
		support_offset = FUNCTION_GROUP4_SUPPORT_OFFSET;
		query_switch_offset = FUNCTION_GROUP4_QUERY_SWITCH_OFFSET;
		check_busy_offset = FUNCTION_GROUP4_CHECK_BUSY_OFFSET;

		switch (func_to_switch) {
		case CURRENT_LIMIT_400:
			support_mask = CURRENT_LIMIT_400_MASK;
			query_switch = CURRENT_LIMIT_400_QUERY_SWITCH_OK;
			switch_busy = CURRENT_LIMIT_400_SWITCH_BUSY;
			break;

		case CURRENT_LIMIT_600:
			support_mask = CURRENT_LIMIT_600_MASK;
			query_switch = CURRENT_LIMIT_600_QUERY_SWITCH_OK;
			switch_busy = CURRENT_LIMIT_600_SWITCH_BUSY;
			break;

		case CURRENT_LIMIT_800:
			support_mask = CURRENT_LIMIT_800_MASK;
			query_switch = CURRENT_LIMIT_800_QUERY_SWITCH_OK;
			switch_busy = CURRENT_LIMIT_800_SWITCH_BUSY;
			break;

		default:
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else {
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (func_group == SD_FUNC_GROUP_1) {
		if (!(buf[support_offset] & support_mask) ||
			((buf[query_switch_offset] & 0x0F) != query_switch)) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	/* Check 'Busy Status' If "Busy Status" Field is defined and Function1
	 * of Group1 shows 'busy' */
	if ((buf[DATA_STRUCTURE_VER_OFFSET] == 0x01) &&
		    ((buf[check_busy_offset] & switch_busy) == switch_busy)) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int sd_check_switch_mode(struct card_reader *cr,
			struct card *c, u8 mode, u8 func_group,
			u8 func_to_switch, enum sd_bus_width bus_width)
{
	struct sd_info *sd_card = c->extra;
	int retval;
	struct sd_command cmd;
	u32 arg = 0;
	u8 buf[64];

	rts_trace(TRACE_SD, "sd_check_switch_mode (mode = %d, func_group = %d,"
				"func_to_switch = %d)\n", mode, func_group,
								func_to_switch);

	arg = (u32) (mode << 24);
	if (func_group == SD_FUNC_GROUP_1) {
		arg |= (u32) (0xFF << 16);
		arg |= (u32) (0xFF << 8);
		arg |= (u32) (0xF0 + func_to_switch);
	} else if (func_group == SD_FUNC_GROUP_3) {
		arg |= (u32) (0xFF << 16);
		arg |= (u32) (0xF0 << 8);
		arg |= (u32) 0xFF;
	} else if (func_group == SD_FUNC_GROUP_4) {
		arg |= (u32) (0xFF << 16);
		arg |= (u32) ((0x0F | (func_to_switch << 4)) << 8);
		arg |= (u32) 0xFF;
	} else {
		arg |= (u32) (SD_CHECK_MODE << 24);
		arg |= (u32) (0xFF << 16);
		arg |= (u32) (0xFF << 8);
		arg |= (u32) 0xFF;
	}

	INIT_SD_CMD(cmd, 0x40 | SWITCH, arg, 0, 0, 0);
	init_sd_dcmd_data(cmd, buf, sizeof(buf), sizeof(buf), 1, bus_width,
				SD_TM_NORMAL_READ, FROM_SD, 250);
	if (sd_send_cmd(cr, c, &cmd)) {
		sd_clear_sd_error(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}
	RTS_DUMP(buf, 64);

	if (func_group == NO_ARGUMENT) {
		sd_card->func_group1_mask = buf[0x0D];
		sd_card->func_group2_mask = buf[0x0B];
		sd_card->func_group3_mask = buf[0x09];
		sd_card->func_group4_mask = buf[0x07];

		rts_trace(TRACE_SD, "func_group1_mask = 0x%02x\n", buf[0x0D]);
		rts_trace(TRACE_SD, "func_group2_mask = 0x%02x\n", buf[0x0B]);
		rts_trace(TRACE_SD, "func_group3_mask = 0x%02x\n", buf[0x09]);
		rts_trace(TRACE_SD, "func_group4_mask = 0x%02x\n", buf[0x07]);
	} else {
		/* Maximum current consumption, check whether current is
		 * acceptable; bit[511:496] = 0x0000 means some error
		 * happaned. */
		u16 cc = ((u16) buf[0] << 8) | buf[1];
		rts_trace(TRACE_SD, "Maximum current consumption: %dmA\n", cc);
		if ((cc == 0) || (cc > 800)) {
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = sd_comm_query_switch_result(cr, func_group,
						func_to_switch, buf, 64);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if ((cc > 400) || (func_to_switch > CURRENT_LIMIT_400)) {
			retval = sd_current_over_limit400_handler(cr);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, retval);
		}
	}

	return STATUS_SUCCESS;
}

static u8 sd_comm_downgrade_switch_mode(u8 func_group,
							u8 func_to_switch)
{
	if (func_group == SD_FUNC_GROUP_1) {
		if (func_to_switch > HS_SUPPORT) {
			func_to_switch--;
		}
	} else if (func_group == SD_FUNC_GROUP_4) {
		if (func_to_switch > CURRENT_LIMIT_200) {
			func_to_switch--;
		}
	}

	return func_to_switch;
}

int sd_comm_check_switch(struct card_reader *cr, struct card *c,
					u8 func_group, u8 func_to_switch,
					enum sd_bus_width bus_width)
{
	int retval;
	int i;
	int switch_good = 0;

	for (i = 0; i < 3; i++) {
		if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
			sd_set_error_code(cr, c, SD_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}
		retval = sd_check_switch_mode(cr, c, SD_CHECK_MODE,
				func_group, func_to_switch, bus_width);

		if (retval == STATUS_SUCCESS) {
			/* u8 stat; */
			retval = sd_check_switch_mode(cr, c,
						SD_SWITCH_MODE, func_group,
						func_to_switch, bus_width);

			if (retval == STATUS_SUCCESS) {
				switch_good = 1;
				break;
			}

			retval = sd_check_crc16_error(cr);
			if (STATUS_SUCCESS != retval)
				TRACE_RET(cr, retval);
		}

		func_to_switch = sd_comm_downgrade_switch_mode(func_group,
							func_to_switch);
		msleep(20);
	}

	if (!switch_good) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int sd_switch_function_legacy(struct card_reader *cr,
				struct card *c, enum sd_bus_width bus_width)
{
	struct sd_info *sd_card = c->extra;
	struct cr_option *option = &(cr->option);

	int retval;
	int i;
	u8 func_to_switch = 0;
	u32 host_cap = cr_get_sd_capability_sw(cr);

	/* Get supported functions */
	retval = sd_check_switch_mode(cr, c, SD_CHECK_MODE,
			NO_ARGUMENT, NO_ARGUMENT, bus_width);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	sd_card->func_group1_mask &= ~(sd_card->sd_switch_fail);
	rts_trace(TRACE_SD, "After masked with sd_switch_fail,"
			"func_group1_mask = 0x%x\n", sd_card->func_group1_mask);

	/* Function Group 1: Access Mode; */
	for (i = 0; i < 4; i++) {
		switch ((u8)(option->sd_speed_prior >> (i*8))) {
		case SDR104_SUPPORT:
			if ((sd_card->func_group1_mask & SDR104_SUPPORT_MASK)
					&& (host_cap & SD_CAP_SDR104)) {
				if (!cr_chk_card_status(cr, CARD_THERMAL_JUMP_TO_SDR50))
					func_to_switch = SDR104_SUPPORT;
			}
			break;

		case DDR50_SUPPORT:
			if ((sd_card->func_group1_mask & DDR50_SUPPORT_MASK)
					&& (host_cap & SD_CAP_DDR50)) {
				func_to_switch = DDR50_SUPPORT;
			}
			break;

		case SDR50_SUPPORT:
			if ((sd_card->func_group1_mask & SDR50_SUPPORT_MASK)
					&& (host_cap & SD_CAP_SDR50)) {
				func_to_switch = SDR50_SUPPORT;
			}
			break;

		case HS_SUPPORT:
			if (sd_card->func_group1_mask & HS_SUPPORT_MASK)
				func_to_switch = HS_SUPPORT;
			break;

		default:
			continue;
		}

		if (func_to_switch)
			break;
	}

	rts_trace(TRACE_SD, "SD_FUNC_GROUP_1: func_to_switch = 0x%02x",
							func_to_switch);

	/* We reset SD30 card with SDR mode if required.  Since block length
	 * shall always be even in DDR50 mode, and length of password should
	 * also be set even, which impose a unfriendly restrict for end user
	 * when SD30 card running at DDR50 mode. */
	if ((sd_card->sd_lock_status & SD_SDR_RST) &&
			(DDR50_SUPPORT == func_to_switch) &&
			(sd_card->func_group1_mask & SDR50_SUPPORT_MASK)) {
		func_to_switch = SDR50_SUPPORT;
		rts_trace(TRACE_SD,
			"Using SDR50 instead of DDR50 for SD Lock\n");
	}

	if (func_to_switch) {
		retval = sd_comm_check_switch(cr, c, SD_FUNC_GROUP_1,
				func_to_switch, bus_width);
		if (retval != STATUS_SUCCESS) {
			if (func_to_switch == SDR104_SUPPORT) {
				sd_card->sd_switch_fail = SDR104_SUPPORT_MASK;
			} else if (func_to_switch == DDR50_SUPPORT) {
				sd_card->sd_switch_fail =
					SDR104_SUPPORT_MASK |
					DDR50_SUPPORT_MASK;
			} else if (func_to_switch == SDR50_SUPPORT) {
				sd_card->sd_switch_fail =
					SDR104_SUPPORT_MASK |
					DDR50_SUPPORT_MASK |
					SDR50_SUPPORT_MASK;
			}
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (func_to_switch == SDR104_SUPPORT) {
			SET_SD_SDR104(sd_card);
			sd_card->sd_default_tx_phase =
				option->sd_sdr104_default_tx_phase;
			sd_card->sd_default_rx_phase =
				option->sd_sdr104_default_rx_phase;
		} else if (func_to_switch == DDR50_SUPPORT) {
			SET_SD_DDR50(sd_card);
			sd_card->sd_default_tx_phase =
				option->sd_ddr50_default_tx_phase;
			sd_card->sd_default_rx_phase =
				option->sd_ddr50_default_rx_phase;
		} else if (func_to_switch == SDR50_SUPPORT) {
			SET_SD_SDR50(sd_card);
			sd_card->sd_default_tx_phase =
				option->sd_sdr50_default_tx_phase;
			sd_card->sd_default_rx_phase =
				option->sd_sdr50_default_rx_phase;
		} else {
			SET_SD_HS(sd_card);
		}
	}

	if (CHK_SD_DDR50(sd_card)) {
		retval = sd_set_clk_pol_ddr_mod(cr, SD_COMM_POL_SOL_DRV);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
		retval = sd_set_sample_push_timing(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	if (!func_to_switch || (func_to_switch == HS_SUPPORT)) {
		/* Do not try to switch current limit if the card doesn't
		 * support UHS mode or we don't want it to support UHS mode */
		return STATUS_SUCCESS;
	}


	/* Function Group 4: Current Limit; */
	func_to_switch = 0xFF;

	for (i = 0; i < 4; i++) {
		switch ((u8)(option->sd_current_prior >> (i*8))) {
		case CURRENT_LIMIT_800:
			if (sd_card->func_group4_mask &
					CURRENT_LIMIT_800_MASK) {
				func_to_switch = CURRENT_LIMIT_800;
			}
			break;

		case CURRENT_LIMIT_600:
			if (sd_card->func_group4_mask &
					CURRENT_LIMIT_600_MASK) {
				func_to_switch = CURRENT_LIMIT_600;
			}
			break;

		case CURRENT_LIMIT_400:
			if (sd_card->func_group4_mask &
					CURRENT_LIMIT_400_MASK) {
				func_to_switch = CURRENT_LIMIT_400;
			}
			break;

		case CURRENT_LIMIT_200:
			if (sd_card->func_group4_mask &
					CURRENT_LIMIT_200_MASK) {
				func_to_switch = CURRENT_LIMIT_200;
			}
			break;

		default:
			continue;
		}

		if (func_to_switch != 0xFF)
			break;
	}

	rts_trace(TRACE_SD, "SD_FUNC_GROUP_4: func_to_switch = 0x%02x",
			func_to_switch);

	if (func_to_switch <= CURRENT_LIMIT_800) {
		sd_card->current_limit = CURRENT_LIMIT_200;
		retval = sd_comm_check_switch(cr, c, SD_FUNC_GROUP_4,
				func_to_switch, bus_width);
		/* We don't care switch result here */
		if (retval != STATUS_SUCCESS) {
			if (sd_check_error_code(cr, c, SD_NO_CARD)) {
				TRACE_RET(cr, STATUS_FAIL);
			}
		} else {
			sd_card->current_limit = func_to_switch;
		}

		rts_trace(TRACE_SD, "Switch current limit finished! (%d)\n",
				retval);
	}

	if (CHK_SD_DDR50(sd_card)) {
		retval = sd_set_clk_pol_ddr_mod(cr, SD_COMM_POL_SOL_ANG);
		if (retval != STATUS_SUCCESS) {
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int sd_switch_function_uhsii(struct card_reader *cr,
				struct card *c, enum sd_bus_width bus_width)
{

	struct sd_info *sd_card = c->extra;
	int retval;
	u8 func_to_switch = 0;

	/* Get supported functions */
	retval = sd_check_switch_mode(cr, c, SD_CHECK_MODE,
		NO_ARGUMENT, NO_ARGUMENT, bus_width);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	/* only check POWER LIMIT 1.44 now.
	 * SPEC: Part 1 Physical Layer Specification Ver4.10
	 * TABLE 4-11:Available Functions of CMD6
	 */
	if ((!(c->reset_flags & UHSII_TO_BYPASS_CURRENT)) &&
		(sd_card->func_group4_mask & POWER_LIMIT_1_44_W_MASK)) {
		func_to_switch = POWER_LIMIT_1_44_W;
		rts_trace(TRACE_SD, "SD_FUNC_GROUP_4: func_to_switch = 0x%02x",
								func_to_switch);
		retval = sd_comm_check_switch(cr, c, SD_FUNC_GROUP_4,
						func_to_switch, bus_width);
		/* We don't care switch result here */
		if (retval != STATUS_SUCCESS) {
			if (sd_check_error_code(cr, c, SD_NO_CARD))
				TRACE_RET(cr, STATUS_FAIL);
		}

		sd_current_over_limit400_handler(cr);
		rts_trace(TRACE_SD, "Switch power limit finished! (%d)\n",
									retval);
	} else {
		rts_trace(TRACE_SD, "Keep the dafault value of func_group_4\n");
	}

	return STATUS_SUCCESS;
}

static int sd_switch_function(struct card_reader *cr, struct card *c,
					enum sd_bus_width bus_width)
{
	struct sd_info *sd_card = c->extra;

	if (CHK_SD_UHSII(sd_card))
		return sd_switch_function_uhsii(cr, c, bus_width);
	else
		return sd_switch_function_legacy(cr, c, bus_width);
}

static int sd_sdr_tuning_rx_cmd(struct card_reader *cr,
					struct card *c, u8 sample_point)
{
	struct sd_info *sd_card = c->extra;
	struct sd_command cmd;

	if ((sd_change_phase(cr, c, sample_point, TUNE_RX)))
		TRACE_RET(cr, STATUS_FAIL);

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN,
		SD_RSP_80CLK_TIMEOUT_EN);

	if (CHK_MMC_SDR200(sd_card)) {
		enum sd_bus_width width = SD_BUS_WIDTH_4;
		int len = 0;
		if (CHK_MMC_8BIT(sd_card)) {
			width = SD_BUS_WIDTH_8;
			len = 0x80;
		} else {
			width = SD_BUS_WIDTH_4;
			len = 0x40;
		}

		INIT_SD_CMD(cmd, 0x40 | SEND_TUNING_BLOCK, 0, 0, 0, 0);
		init_sd_dcmd_data(cmd, NULL, 0, len, 1, width,
					SD_TM_AUTO_TUNING, FROM_SD, 500);
	} else {
		INIT_SD_CMD(cmd, 0x40 | SEND_TUNING_PATTERN, 0, 0, 0, 0);
		init_sd_dcmd_data(cmd, NULL, 0, 0x40, 1, SD_BUS_WIDTH_4,
					SD_TM_AUTO_TUNING, FROM_SD, 100);
	}

	if (sd_send_cmd(cr, c, &cmd)) {
		/* Wait till SD DATA IDLE */
		sd_wait_data_idle(cr);
		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);
		TRACE_RET(cr, STATUS_FAIL);
	}

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);

	return STATUS_SUCCESS;
}

static int sd_ddr_tuning_rx_cmd(struct card_reader *cr,
					struct card *c, u8 sample_point)
{
	int retval;
	struct sd_command cmd;

	retval = sd_change_phase(cr, c, sample_point, TUNE_RX);
	if (retval)
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "sd ddr tuning rx\n");
	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN,
		SD_RSP_80CLK_TIMEOUT_EN);

	INIT_SD_CMD(cmd, 0x40 | SD_STATUS, 0, SD_APP_CMD, 0, 0);
	init_sd_dcmd_data(cmd, NULL, 0, 64, 1, SD_BUS_WIDTH_4,
				SD_TM_NORMAL_READ, FROM_SD, 100);
	if (sd_send_cmd(cr, c, &cmd)) {
		/* Wait till SD DATA IDLE */
		sd_wait_data_idle(cr);
		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);
		TRACE_RET(cr, STATUS_FAIL);
	}

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);

	return STATUS_SUCCESS;
}

static int mmc_ddr_tunning_rx_cmd(struct card_reader *cr,
					struct card *c, u8 sample_point)
{
	struct sd_info *sd_card = c->extra;
	struct sd_command cmd;
	int retval;
	enum sd_bus_width bus_width;

	if (CHK_MMC_8BIT(sd_card))
		bus_width = SD_BUS_WIDTH_8;
	else if (CHK_MMC_4BIT(sd_card))
		bus_width = SD_BUS_WIDTH_4;
	else
		bus_width = SD_BUS_WIDTH_1;

	retval = sd_change_phase(cr, c, sample_point, TUNE_RX);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	rts_trace(TRACE_SD, "mmc ddr tuning rx\n");
	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN,
		SD_RSP_80CLK_TIMEOUT_EN);

	INIT_SD_CMD(cmd, 0x40 | SEND_EXT_CSD, 0, 0, 0, 0);
	init_sd_dcmd_data(cmd, NULL, 0, 0x200, 1, bus_width,
				SD_TM_NORMAL_READ, FROM_SD, 100);
	if (sd_send_cmd(cr, c, &cmd)) {
		/* Wait till SD DATA IDLE */
		sd_wait_data_idle(cr);
		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);
		TRACE_RET(cr, STATUS_FAIL);
	}

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);

	return STATUS_SUCCESS;
}

typedef  u64 phase_t;
int find_longest_continue_bits(phase_t p, int len, int *__max_pos,
		int *__max_len)
{
	int max_pos = -1;
	int max_len = 0;
	int cursor = 0;

	for (cursor = 0; cursor < len;) {
		int index = cursor;
		int cur_len = 0;

		if (!(p & (1ULL << index))) {
			cursor++;
			continue;
		}

		index++;
		index %= len;
		while (p & (1ULL << index)) {
			index++;
			index %= len;

			if (index == cursor)
				break;
		}

		if (index <= cursor)
			cur_len = (index + len) - cursor;
		else
			cur_len = index - cursor;

		if (max_len < cur_len) {
			max_len = cur_len;
			max_pos = cursor;
		}

		if (index <= cursor)
			break;
		else
			cursor = index;
	}

	*__max_len = max_len;
	*__max_pos = max_pos;

	return 0;
}
static u8 sd_search_final_phase(u32 phase_map, int total_phase,
					int *__phase, int *__max_len)
{
	struct timing_phase_path *path = NULL;
	int i, j, cont_path_cnt;
	int new_block, max_len, final_path_idx;
	int final_phase = -1;

	path = kmalloc(total_phase * sizeof(*path), GFP_KERNEL);
	if (path == NULL) {
		rts_trace(TRACE_ERROR,
			"Allocating memroy failed for timing phase path\n");
		return -ENOMEM;
	}

	/* Parse phase_map, take it as a bit-ring */
	cont_path_cnt = 0;
	new_block = 1;
	j = 0;

	for (i = 0; i < total_phase; i++) {
		if (phase_map & (1 << i)) {
			if (new_block) {
				new_block = 0;
				j = cont_path_cnt++;
				path[j].start = i;
				path[j].end = i;
			} else {
				path[j].end = i;
			}
		} else {
			new_block = 1;
			if (cont_path_cnt) {
				/* Calculate path length and middle point */
				int idx = cont_path_cnt - 1;
				path[idx].len = path[idx].end -
							path[idx].start + 1;
				path[idx].mid = path[idx].start +
							path[idx].len / 2;
			}
		}
	}

	if (cont_path_cnt == 0) {
		rts_trace(TRACE_SD, "No continuous phase path\n");
		*__phase = -1;
		*__max_len = 0;
		goto exit;
	} else {
		/* Calculate last continuous path length and middle point */
		int idx = cont_path_cnt - 1;
		path[idx].len = path[idx].end - path[idx].start + 1;
		path[idx].mid = path[idx].start + path[idx].len / 2;
	}

	/* Connect the first continuous path and the last one if they are
	 * adjacent */
	if ((path[0].start == 0)
		&& (path[cont_path_cnt - 1].end == (total_phase - 1))) {
		path[0].start = path[cont_path_cnt - 1].start -
			total_phase; /* Using negative index */
		path[0].len += path[cont_path_cnt - 1].len;
		path[0].mid = path[0].start + path[0].len / 2;
		/* Convert negative middle point index to positive one */
		if (path[0].mid < 0)
			path[0].mid += total_phase;
		cont_path_cnt--;
	}

	/* Choose the longest continuous phase path */
	max_len = 0;
	final_phase = 0;
	final_path_idx = 0;
	for (i = 0; i < cont_path_cnt; i++) {
		if (path[i].len > max_len) {
			max_len = path[i].len;
			final_phase = (u8)path[i].mid;
			final_path_idx = i;
		}

		rts_trace(TRACE_SD, "path[%d].start = %d\n", i, path[i].start);
		rts_trace(TRACE_SD, "path[%d].end = %d\n", i, path[i].end);
		rts_trace(TRACE_SD, "path[%d].len = %d\n", i, path[i].len);
		rts_trace(TRACE_SD, "path[%d].mid = %d\n", i, path[i].mid);
		rts_trace(TRACE_SD, "\n");
	}

	*__phase = final_phase;
	*__max_len = max_len;

exit:
	if (path)
		kfree(path);
	return 0;
}

static int sd_final_phase_shift_4_tx(struct card *card,
				int total_phase, int phase, int max_len)
{
	struct sd_info *sd_card = card->extra;
	int final_phase = phase;

	if (CHK_SD_SDR104(sd_card) || CHK_MMC_SDR200(sd_card)) {
		/* SD30 SPEC 6.7.3.2, Table 6-16 SDR104 mode: tIS min 1.40ns;
		 * tIH min 0.8ns The worst data valid windows length to satisfy
		 * the spec is (1.40 + 0.80) = 2.2ns.  In this case, if choose
		 * the mid phase, tIS will be overspec(less than 1.40ns).  So
		 * the phase should be selected according to the value(tIS: tIH)
		 * SDR104, clock freq 208M, clock period 1/208M = 4.8ns; There
		 * are 32 phase for TX tuning.  So each phase time = (4.8ns/32)
		 * = 0.15ns; In order to satisfy (tIS min 1.40ns), setup time
		 * should be (1.40ns/0.15ns) = 10 phase at least to satify (tIH
		 * min 0.8ns), hold time should be (0.8ns/0.15ns) = 6 phase at
		 * least. */
		if (max_len >= 28) {
			/* guarantee hold time 12 phase */
			final_phase -= ((max_len / 2) - 12);
		} else if (max_len >= 18) {
			/* guarantee hold time 8 phase */
			final_phase -= ((max_len / 2) - 8);
		} else if (max_len >= 16) {
			/* only guarantee hold time 6 phase */
			final_phase -= ((max_len / 2) - 6);
		} else if (max_len >= 10) {
			/* only guarantee hold time 5 phase */
			final_phase -= ((max_len / 2) - 5);
		}

		if (final_phase < 0)
			final_phase += total_phase;
		} else if (CHK_SD_SDR50(sd_card) || CHK_SD_DDR50(sd_card)
			|| CHK_MMC_DDR52(sd_card)) {
		/* SD30 SPEC 6.7.3.2, Table 6-16 SDR50 mode: tIS min 3.00ns; tIH
		 * min 0.8ns The worst data valid windows length to satisfy the
		 * spec is (3.00 + 0.80) = 3.8ns.  In this case, if choose the
		 * mid phase, tIS will be overspec(less than 3.0ns).  So the
		 * phase should be selected according to the value(tIS: tIH)
		 * SDR104, clock freq 100M, clock period 1/100M = 10ns; There
		 * are 32 phase for TX tuning.  So each phase time = (10ns/32) =
		 * 0.30ns; In order to satisfy (tIS min 3.0ns), setup time
		 * should be (3.0ns/0.15ns) = 10 phase at least to satify (tIH
		 * min 0.8ns), hold time should be (0.8ns/0..3ns) = 3 phase at
		 * least */
		final_phase -= 3;
		if (final_phase < 0)
			final_phase += total_phase;

		if (CHK_SD_DDR50(sd_card) || CHK_MMC_DDR52(sd_card)) {
			final_phase += 8;
			final_phase %= total_phase;
		}
	}

	return final_phase;
}

static int sd_final_phase_shift_4_rx(struct card *card,
				int total_phase, int phase, int max_len)
{
	struct sd_info *sd_card = card->extra;
	int final_phase = phase;

	if (CHK_SD_SDR104(sd_card) || CHK_MMC_SDR200(sd_card)) {
		if (max_len >= 18) {
			/* guarantee hold time 6 phase */
			final_phase += ((max_len / 2) - 6);
		} else if (max_len >= 16) {
			/* only guarantee hold time 5 phase */
			final_phase += ((max_len / 2) - 5);
		} else if (max_len >= 12) {
			/* only guarantee hold time 4 phase */
			final_phase += ((max_len / 2) - 4);
		} else if (max_len >= 10) {
			/* only guarantee hold time 3 phase */
			final_phase += ((max_len / 2) - 3);
		}

		final_phase %= total_phase;
	} else if (CHK_SD_SDR50(sd_card) || CHK_SD_DDR50(sd_card)
				|| CHK_MMC_DDR52(sd_card)) {
		final_phase += 3;
		final_phase %= total_phase;
	}

	return final_phase;
}

static int  sd_get_final_phase(struct card *card, u32 phase_map, u8 tune_dir)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int max_len = 0, final_phase = -1, start = -1;
	int total_phase = cr->sd_max_phase + 1;
	int phase_min_cnt = cr->option.phase_min_cnt;

	if (phase_map == 0xFFFFFFFF) {
		if (tune_dir == TUNE_RX)
			final_phase = (int)sd_card->sd_default_rx_phase;
		else
			final_phase = (int)sd_card->sd_default_tx_phase;
		goto Search_Finish;
	}

	find_longest_continue_bits((phase_t) phase_map,
			total_phase, &start, &max_len);
	rts_trace(TRACE_SD, "max phase start:%d\n", start);
	final_phase = start + (max_len / 2);
	final_phase %= total_phase;

	{
		/* compare new and old algorithm */
		int  __final_phase = -1, __max_len = 0;

		sd_search_final_phase(phase_map, total_phase,
						&__final_phase, &__max_len);

		rts_trace(TRACE_SD, "TEST new algorithm:\n");
		rts_trace(TRACE_SD, "\tNEW - phase_map:%#x, max_len:%d final_phase:%d\n",
				phase_map, max_len, final_phase);
		rts_trace(TRACE_SD, "\tOLD - phase_map:%#x, max_len:%d final_phase:%d\n",
				phase_map, __max_len, __final_phase);

	}

	if (final_phase < 0) {
		rts_trace(TRACE_SD, "No phase can be choosed\n");
		goto Search_Finish;
	}

	if (0 == phase_min_cnt) {
		rts_trace(TRACE_SD, "Don't shift final phase\n");
		goto Search_Finish;
	} else if ((max_len < phase_min_cnt) && (tune_dir == TUNE_TX)) {
		rts_trace(TRACE_ERROR, "max phase len:%d < phase_min_cnt:%d\n",
				max_len, phase_min_cnt);
		rts_trace(TRACE_ERROR, "Tuning Fail\n");
		final_phase = -1;
		goto Search_Finish;
	}

	rts_trace(TRACE_SD, "shift final phase\n");

	if (TUNE_TX == tune_dir) {
		final_phase = sd_final_phase_shift_4_tx(card,
					total_phase, final_phase, max_len);
	} else if (TUNE_RX == tune_dir) {
		final_phase = sd_final_phase_shift_4_rx(card,
					total_phase, final_phase, max_len);
	}

Search_Finish:
	rts_trace(TRACE_SD, "Final choosen phase: %d\n", final_phase);
	return final_phase;

}

static int sd_tuning_rx(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	int retval;
	u32 phase_map;
	int final_phase;
	int (*tuning_cmd)(struct card_reader *cr, struct card *c,
							u8 sample_point);
	if (CHK_SD(sd_card)) {
		if (CHK_SD_DDR50(sd_card)) {
			tuning_cmd = sd_ddr_tuning_rx_cmd;
		} else {
			tuning_cmd = sd_sdr_tuning_rx_cmd;
		}
	} else {
		if (CHK_MMC_DDR52(sd_card))
			tuning_cmd = mmc_ddr_tunning_rx_cmd;
		else if (CHK_MMC_SDR200(sd_card))
			tuning_cmd = sd_sdr_tuning_rx_cmd;
		else
			TRACE_RET(cr, STATUS_FAIL);
	}

	if (do_tuning(cr, c, tuning_cmd, &phase_map, TUNE_RX))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "RX phase_map = 0x%08x\n", phase_map);

	/* note: some card Tx pass but Rx all zero, try using fix Tx.
	 */
	if ((!phase_map) && (!(c->reset_flags & UHSI_FIX_TX_PHASE))){
		c->reset_flags |= UHSI_FIX_TX_PHASE;
		TRACE_RET(cr, STATUS_FAIL);
	}

	final_phase = sd_get_final_phase(c, phase_map, TUNE_RX);
	if (final_phase < 0)
		TRACE_RET(cr, STATUS_FAIL);

	retval = sd_change_phase(cr, c, (u8) final_phase, TUNE_RX);

	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

static int sd_sdr_tuning(struct card_reader *cr, struct card *c)
{
	int retval;
	struct sd_info *sd_card = c->extra;
	struct cr_option *option = &cr->option;

	if ((option->sdr_tx_tuning_en) &&
		!(c->reset_flags & UHSI_FIX_TX_PHASE)) {
		retval = sd_tuning_tx(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);

		retval = sd_tuning_rx(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		retval = sd_change_phase(cr, c, sd_card->sd_default_tx_phase,
			TUNE_TX);
		if (retval != STATUS_SUCCESS) {
			if (c->reset_flags & UHSI_FIX_TX_PHASE) {
				c->reset_flags &= ~(UHSI_FIX_TX_PHASE);
			}
			TRACE_RET(cr, STATUS_FAIL);
		}

		retval = sd_tuning_rx(cr, c);
		if (retval != STATUS_SUCCESS) {
			if (c->reset_flags & UHSI_FIX_TX_PHASE) {
				c->reset_flags &= ~(UHSI_FIX_TX_PHASE);
			}
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return STATUS_SUCCESS;
}

static int sd_ddr_tuning(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;

	if (!(check_sd_flag(cr, SD_COMM_DDR_TX_PHASE_SET_BY_USER))) {
		if (sd_tuning_tx(cr, c))
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		if (sd_change_phase(cr, c,
				(u8)sd_card->sd_default_tx_phase, TUNE_TX))
			TRACE_RET(cr, STATUS_FAIL);
	}

	if ((sd_tuning_rx(cr, c)))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int mmc_ddr_tuning(struct card_reader *cr, struct card *c)
{
	struct cr_option *option = &(cr->option);
	int retval;

	if (!(check_sd_flag(cr, SD_COMM_MMC_DDR_TX_PHASE_SET_BY_USER))) {
		if (sd_tuning_tx(cr, c))
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		retval = sd_change_phase(cr, c,
					(u8)option->mmc_ddr_tx_phase, TUNE_TX);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	retval = sd_tuning_rx(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}

static int mmc_sdr_tuning(struct card_reader *cr, struct card *c)
{
	int retval;
	struct sd_info *sd_card = c->extra;
	struct cr_option *option = &cr->option;

	if (option->sdr_tx_tuning_en) {
		retval = sd_tuning_tx(cr, c);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		u8 phase = 0;

		if (CHK_MMC_SDR200(sd_card))
			phase = (u8) (option->mmc_sdr200_tx_phase);
		if (sd_change_phase(cr, c, phase, TUNE_TX))
			TRACE_RET(cr, STATUS_FAIL);
	}

	retval = sd_tuning_rx(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}
int sd_wait_state_data_ready(struct card_reader *cr,
		struct card *c, u8 state, u8 data_ready, int polling_cnt)
{
	int retval, i;
	u8 rsp[16];

	for (i = 0; i < polling_cnt; i++) {
		/* Send CMD13 to get the R1 response from SD card */
		retval = sd_send_status(c, rsp);
		if (retval)
			TRACE_RET(cr, STATUS_FAIL);

		/* Check "Current State" was matched with byStateForPolling or
		 * not */
		if (((rsp[3] & 0x1E) == state)
			&& ((rsp[3] & 0x01) == data_ready)) {
			return STATUS_SUCCESS;
		}
	}

	TRACE_RET(cr, STATUS_FAIL);
}

int sd_speed_class_control(struct card_reader *cr, struct card *card,
		u32 arg)
{
	struct sd_command cmd;
	INIT_SD_CMD(cmd, SPEED_CLASS_CONTROL, arg,
		0, SD_RSP_TYPE_R1b, 0);
	return sd_send_cmd(cr, card, &cmd);
}

static int sd_check_wp_state(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	u16 sd_card_type;
	u8 buf[64];
	struct sd_command cmd;

	rts_trace(TRACE_SD, "ACMD13:\n");

	/* Check ROM card or OTP through ACMD 13 */
	INIT_SD_CMD(cmd, 0x40 | SD_STATUS, 0, SD_APP_CMD, 0, 0);
	init_sd_dcmd_data(cmd, buf, sizeof(buf), sizeof(buf), 1, SD_BUS_WIDTH_4,
				SD_TM_NORMAL_READ, FROM_SD, 250);
	if (sd_send_cmd(cr, c, &cmd)) {
		sd_clear_sd_error(cr);
		sd_clear_status(c);
		TRACE_RET(cr, STATUS_FAIL);
	}

	RTS_DUMP(buf, 64);

	sd_card_type = ((u16)buf[2] << 8) | buf[3];
	rts_trace(TRACE_SD, "sd_card_type = 0x%04x\n", sd_card_type);
	/* ROM card or OTP */
	if ((sd_card_type == 0x0001) || (sd_card_type == 0x0002))
		set_card_flags(c, WP);

	if (sd_is_mechanical_wp(cr))
		set_card_flags(c, WP);

	sd_card->sd_au_size = (buf[10] & 0xF0) >> 4;
	sd_card->sd_erase_size = ((u16)buf[11] << 8) | buf[12];
	sd_card->sd_erase_timeout = (buf[13] & 0xFC) >> 2;
	sd_card->sd_erase_offset = buf[13] & 0x03;
	sd_card->sd_uhs_au_size = buf[14] & 0x0F;

	rts_trace(TRACE_SD, "sd_au_size = 0x%02x\n",
		sd_card->sd_au_size);
	rts_trace(TRACE_SD, "sd_uhs_au_size = 0x%02x\n",
		sd_card->sd_uhs_au_size);
	rts_trace(TRACE_SD, "sd_erase_size = 0x%04x\n",
		sd_card->sd_erase_size);
	rts_trace(TRACE_SD, "sd_erase_timeout = 0x%02x\n",
		sd_card->sd_erase_timeout);
	rts_trace(TRACE_SD, "sd_erase_offset = 0x%02x\n",
		sd_card->sd_erase_offset);

	return 0;
}

static int sd_comm_change_bank_voltage(struct card_reader *cr, u8 voltage)
{
	u16 val = 0;
	int retval;

	if (voltage == SD_IO_3V3) {
		if (cr->asic_code) {
			cr_rd_phy16(cr, 0x08, &val);
			val &= 0xFC3F;
			val |= 0x03C0;
			retval = cr_wr_phy16(cr, 0x08, val);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);
		} else {
			cr_wr_reg(cr, REG_SD_PAD_CTL, SD_IO_USING_1V8, 0);
		}
	} else if (voltage == SD_IO_1V8) {
		if (cr->asic_code) {
			u16 phy;
			cr_rd_phy16(cr, 0x11, &phy);
			phy &= 0xFFF3;
			retval = cr_wr_phy16(cr, 0x11, phy);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);

			cr_rd_phy16(cr, 0x08, &val);
			val &= 0xFC3F;
			val |= 0x0080;
			retval = cr_wr_phy16(cr, 0x08, val);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);
		} else {
			cr_wr_reg(cr, REG_SD_PAD_CTL,
					SD_IO_USING_1V8, SD_IO_USING_1V8);
		}
	} else if (voltage == SD_IO_1V9) {
		if (cr->asic_code) {
			retval = cr_update_phy16(cr, 0x11, 0xFFF3, 0);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);

			retval = cr_update_phy16(cr, 0x08, 0xFC3F, 0x0100);
			if (retval != STATUS_SUCCESS)
				TRACE_RET(cr, STATUS_FAIL);
		} else {
			cr_wr_reg(cr, REG_SD_PAD_CTL,
					SD_IO_USING_1V8, SD_IO_USING_1V8);
		}
	} else {
		TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}


static int sd_comm_reset_mmc(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	int retval, i = 0, j = 0, k = 0;
	int switch_ddr = 1, reset_emmc45_fail = 0;
	u8 rsp[16];
	u8 spec_ver = 0;

	c->reset_flags &= ~(RESET_UHSII_NOW);
	c->reset_flags &= ~(RESET_E_MMC45_NOW);

	if (check_sd_flag(cr, SD_COMM_RESET_MMC_FIRST)) {
		if (sd_init_power(cr))
			TRACE_RET(cr, STATUS_FAIL);
	}

	if (sd_card->sd_lock_status & SD_UNLOCK_POW_ON)
		goto MMC_UNLOCK_ENTRY;

DDR_TUNING_FAIL:
	/* FIXME: edwin: prepare_reset() herein inclues sd_dummy_clock()
	 * Is it OK here, to be checked */
	retval = sd_prepare_reset(cr, c);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, retval);
	}
	SET_MMC(sd_card);
RTY_MMC_RST:
	retval = sd_go_idle_state(c);
	if (retval != STATUS_SUCCESS)
	       TRACE_RET(cr, STATUS_FAIL);

	do {
		/* To check whether card is extracted while identifying In some
		 * situation, cmd1 is still success even if card extracted, so
		 * we check it here. */
		if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
			sd_set_error_code(cr, c, SD_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}

		/* CMD 1
		 * For MMC4.2, we shall set Bit[30,29] to [1,0] to indicate
		 * we support sector addressing mode */
		retval = sd_send_op_cond(c, rsp);
		if (retval != STATUS_SUCCESS) {
			if (sd_check_error_code(cr, c, SD_BUSY) ||
				sd_check_error_code(cr, c, SD_TO_ERR)) {
				k++;
				if (k < 20) {
					sd_clear_error_code(cr, c);
					goto RTY_MMC_RST;
				} else {
					TRACE_RET(cr, STATUS_FAIL);
				}
			} else {
				j++;
				if (j < 100) {
					sd_clear_error_code(cr, c);
					goto RTY_MMC_RST;
				} else {
					TRACE_RET(cr, STATUS_FAIL);
				}
			}
		}

		msleep(20);
		i++;
	} while (!(rsp[1] & 0x80) && (i < 255)); /* Not complete power on */

	if (i == 255) {
		/* Time out */
		TRACE_RET(cr, STATUS_FAIL);
	} else {
		c->reset_flags |= CONCRETE_MMC_CARD;
	}

	if ((rsp[1] & 0x60) == 0x40) {
		SET_MMC_SECTOR_MODE(sd_card);
	} else {
		CLR_MMC_SECTOR_MODE(sd_card);
	}

	/* CMD 2 */
	retval = sd_send_cid(c, sd_card->raw_cid);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	rts_trace(TRACE_SD, "CID Response:\n");
	hex_dump(sd_card->raw_cid, 16);

	/* CMD 3 */
	sd_card->sd_addr = 0x00100000;
	retval = sd_set_relative_addr(c, sd_card->sd_addr, rsp);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	/* Get CSD register for Calculating Timing,Capacity
	 * Check CSD to determaine as if this is the SD ROM card  */
	retval = sd_check_csd(cr, c, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Get MMC Spec_Ver in the CSD register */
	spec_ver = (sd_card->raw_csd[0] & 0x3C) >> 2;
	rts_trace(TRACE_SD, "MMC spec version :%d\n", spec_ver);

	/* Select MMC card */
	retval = sd_select_card(cr, c, 1);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* Set block length 512 bytes for all block commands */
	retval = sd_set_blocklen(c, 0x200, rsp);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

MMC_UNLOCK_ENTRY:
	/* Get SD lock status */
	retval = sd_update_lock_status(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	/* If we want to use the NORMAL_READ_MODE, we shall not set "Clock
	 * Divider" */
	retval = sd_set_clock_divider(cr, SD_CLK_DIVIDE_0);
	if (retval != STATUS_SUCCESS) {
		TRACE_RET(cr, STATUS_FAIL);
	}
	c->bus_width = 1;

	/* mmc_dont_switch_bus will be set if crc error occurred in mmc r/w */
	if (!sd_card->mmc_dont_switch_bus) {
		if (spec_ver == 4) {
			int go_sdr200 = 0;
			u32 sd_cap = cr_get_sd_capability_sw(cr);

			mmc_check_ext_csd(c);

			go_sdr200 = !!(CHK_MMC_SDR200(sd_card));
			go_sdr200 &= !(reset_emmc45_fail);
			go_sdr200 &= !!(sd_cap & SD_CAP_MMC_SDR200);

			if (CHK_MMC_SDR200(sd_card) && !go_sdr200) {
				CLR_MMC_SDR200(sd_card);
				SET_MMC_DDR52(sd_card);
			}

			if (!switch_ddr && CHK_MMC_DDR52(sd_card)) {
				CLR_MMC_DDR52(sd_card);
				SET_MMC_52M(sd_card);
			}

			if (go_sdr200 &&
				check_sd_flag(cr, EMMC45_DONT_SWITCH_POWER)) {
				c->reset_flags |= RESET_E_MMC45_NOW;
				rts_trace(TRACE_SD, "no need change power\n");
			}

			if (go_sdr200 && !(c->reset_flags & RESET_E_MMC45_NOW)) {
				rts_trace(TRACE_SD, "%s GO MMC4.5\n", __func__);
				c->reset_flags |= RESET_E_MMC45_NOW;
				/* poweron card 1.8 */
				if (sd_init_power(cr))
					TRACE_RET(cr, STATUS_FAIL);
				msleep(50);
				goto DDR_TUNING_FAIL;
			} else if (c->reset_flags & RESET_E_MMC45_NOW) {
				if (mmc_select_sdr200(c)) {
					reset_emmc45_fail = 1;
					mmc_process_sdr200_error(cr, c);
					goto DDR_TUNING_FAIL;
				}
			} else {
				rts_trace(TRACE_SD, "%s GO MMC4.4\n", __func__);
				/* MMC 4.x Cards */
				sd_mmc_switch_timing_bus(cr, c, switch_ddr);
			}
		}

		/* For MMC4.2, we shall return fail if get capacity of MMC4.2
		 * card fail */
		if (CHK_MMC_SECTOR_MODE(sd_card) && (c->capacity == 0)) {
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (CHK_MMC_SDR200(sd_card)) {
			int tuning = CHK_MMC_8BIT(sd_card)
				|| CHK_MMC_4BIT(sd_card);
			if (sd_set_init_para(cr, c))
				TRACE_RET(cr, STATUS_FAIL);
			if (tuning) {
				if (mmc_sdr_tuning(cr, c)) {
					reset_emmc45_fail = 1;
					mmc_process_sdr200_error(cr, c);
					goto RTY_MMC_RST;
				}
			}
		} else if (switch_ddr && CHK_MMC_DDR52(sd_card)) {
			retval = sd_set_init_para(cr, c);
			if (retval != STATUS_SUCCESS) {
				TRACE_RET(cr, STATUS_FAIL);
			}
			retval = mmc_ddr_tuning(cr, c);
			if (retval != STATUS_SUCCESS) {
				retval = sd_init_power(cr);
				if (retval != STATUS_SUCCESS) {
					TRACE_RET(cr, STATUS_FAIL);
				}
				switch_ddr = 0;
				goto DDR_TUNING_FAIL;
			}
			retval = sd_wait_state_data_ready(cr, c,
							0x08, 1, 1000);
			if (retval == STATUS_SUCCESS) {
				retval = sd_read_lba0(cr, c);
				if (retval != STATUS_SUCCESS) {
					retval = sd_init_power(cr);
					if (retval != STATUS_SUCCESS) {
						TRACE_RET(cr, STATUS_FAIL);
					}
					switch_ddr = 0;
					goto DDR_TUNING_FAIL;
				}
			}
		}
	}

	if (sd_card->sd_lock_status & SD_UNLOCK_POW_ON)
		sd_set_blk_count(cr, 0x0200); /* blk counts: 512*/

	/* For MMC Micro Adapter, this adapter has WP function just like the SD
	 * card So, we need to Check SD Machanical Write-Protect status
	 * here.  */
	if (sd_is_mechanical_wp(cr)) set_card_flags(c, WP);
	rts_trace(TRACE_SD, "-->\n");

	return STATUS_SUCCESS;
}

static int sd_reset_mmc(struct card_reader *cr, struct card *c)
{
	u32 sd_cap = cr_get_sd_capability_sw(cr);

	if (!(sd_cap & SD_CAP_MMC))
		return -1;

	if (c->reset_flags & CONCRETE_SD_CARD)
		return -1;

	if (c->reset_flags & RESET_MEET_OCP)
		return -1;

	if (cr->sd_ops_prot->reset_mmc)
		return cr->sd_ops_prot->reset_mmc(cr, c);
	else
		return sd_comm_reset_mmc(cr, c);
}

/*
 * sd_verify_if_conf - send cmd8 to card
 */
static int sd_verify_if_conf(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	u8 rsp[16];

	/* Send CMD8 to beginning Initialization
	 * Choose Voltage Supplied = 2.7V~3.6V, Check Pattern = 0xAA */
	if ((sd_send_if_cond(card, 0x1AA, rsp))) {
		SET_SD_SC(sd_card);
		return -1;
	}

	/* Check R7 Format */
	if ((rsp[4] == 0xAA) && ((rsp[3] & 0x0f) == 0x01)) {
		/* Voltage Supported and Pattern Check OK */
		/* hi_cap_flow = 1; */
		/* Bit[30] of Argument shall set to '1' when we support
		 * High Capacity Card */
		/* SET_SD_HCXC(sd_card); */
		CLR_SD_SC(sd_card);
	} else {
		/* fix me: if get invalid response, how can we do? */
		SET_SD_SC(sd_card);
	}

	return 0;
}

/*
 * sd_verify_op_conf - send ACMD41 to card
 */
static int sd_verify_op_conf(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	u32 voltage = 0;
	u8 rsp[16];
	int ret = 0;
	int i = 0;

	if (CHK_SD_SC(sd_card)) {
		voltage = SUPPORT_VOLTAGE;
	} else {
		voltage = SUPPORT_VOLTAGE | SUPPORT_HIGH_AND_EXTENDED_CAPACITY;
		if (!(card->reset_flags & GO_SD20_DIRECT))
			voltage |= SUPPORT_MAX_POWER_PERMANCE | SUPPORT_1V8;
	}

	do {
		ret = sd_app_send_op_cond(card, voltage, rsp);
		if (ret)
			return ret;
		if (rsp[1] & 0x80)
			break;
		msleep(20);
	} while (i++ < 255);

	if (i >= 255)
		return -ETIMEDOUT;
	else
		card->reset_flags |= CONCRETE_SD_CARD;

	if (!CHK_SD_SC(sd_card)) {
		/* Check CCS bit to determine as if this is a High-Capacity
		 * Card */
		if (rsp[1] & 0x20) {
			rts_trace(TRACE_SD, "SD UHS-II Card\n");
			SET_SD_HCXC(sd_card);
		} else if (rsp[1] & 0x40) {
			rts_trace(TRACE_SD, "SD HCXC Card\n");
			SET_SD_HCXC(sd_card);
		} else  {
			rts_trace(TRACE_SD, "SDSC Card\n");
			SET_SD_SC(sd_card);
		}

		if (CHK_SD_HCXC(sd_card)
				&& !(card->reset_flags & GO_SD20_DIRECT))
			sd_card->r.s18a = (rsp[1] & 0x01) ? 1 : 0;
		else
			sd_card->r.s18a = 0;
	} else {
		SET_SD_SC(sd_card);
		sd_card->r.s18a = 0;
	}

	return 0;
}

/*
 * sd_asign_relative_addr - send cmd3 to card
 */
static int sd_asign_relative_addr(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int i = 0;
	u8 rsp[16];

	/*In the case that the RCA equals 0, then the host is recommanded
	 * to Re-send CMD3*/
	for (i = 0; i < 3; i++) {
		if ((sd_set_relative_addr(card, 0, rsp)))
			TRACE_RET(cr, STATUS_FAIL);

		sd_card->sd_addr = (u32)rsp[1] << 24;
		sd_card->sd_addr += (u32)rsp[2] << 16;

		if ((CHK_SD_UHSII(sd_card))) {
			if ((sd_card->sd_addr >> 16) != sd_card->sd40_did) {
				rts_trace(TRACE_SD,
					"SD40 CMD3 RCA(%#x)!= NODE ID(%#X)\n",
					sd_card->sd_addr, sd_card->sd40_did);
				sd_card->sd_addr = 0;
			}
		}

		if (sd_card->sd_addr)
			break;
	}

	if (0 == sd_card->sd_addr)
		return -1;

	return 0;
}

/*
 * sd_do_reset_legacy_card - Start Initialization Process of SD Card
 */
static int sd_do_reset_legacy_card(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int retry = 0;

	do {
		if ((sd_go_idle_state(card)))
			TRACE_RET(cr, STATUS_FATAL_ERROR);

		msleep(20);

		/* check if SDSC card */
		sd_verify_if_conf(card);

		if (CHK_SD_SC(sd_card)) {
			if ((sd_go_idle_state(card)))
				TRACE_RET(cr, STATUS_FATAL_ERROR);
			msleep(20);
		}

		if ((sd_verify_op_conf(card))) {
			if (cr_detect_card(cr, SD_ID)) {
				sd_set_error_code(cr, card, SD_NO_CARD);
				TRACE_RET(cr, STATUS_FATAL_ERROR);
			}
			continue;
		}
		rts_trace(TRACE_SD, "support_1v8 = %d\n", sd_card->r.s18a);

		/* CMD11: Switch Voltage */
		if (sd_card->r.s18a) {
			if ((sd_voltage_switch(cr, card))) {
				if(card->reset_flags & UHSI_1V9_RERTY) {
					card->reset_flags &= ~UHSI_1V9_RERTY;
					card->reset_flags |= GO_SD20_DIRECT;
				} else {
					card->reset_flags |= UHSI_1V9_RERTY;
				}

				TRACE_RET(cr, STATUS_FAIL);
			}
		}

		/* CMD 2 */
		if ((sd_send_cid(card, sd_card->raw_cid)))
			TRACE_RET(cr, STATUS_FATAL_ERROR);
		rts_trace(TRACE_SD, "CID Response:\n");
		hex_dump(sd_card->raw_cid, 16);

		if (sd_asign_relative_addr(card))
			TRACE_RET(cr, STATUS_FATAL_ERROR);

		/* Get CSD register for Calculating Timing,Capacity
		 * Check CSD to determaine as if this is the SD ROM card  */
		if ((sd_check_csd(cr, card, 1)))
			TRACE_RET(cr, STATUS_FATAL_ERROR);

		/* Select SD card */
		if ((sd_select_card(cr, card, 1)))
			TRACE_RET(cr, STATUS_FATAL_ERROR);

		break;

	} while (retry++ < 3);

	if (retry >= 3)
		return STATUS_FATAL_ERROR;

	return 0;
}

static int sd_do_switch_action(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int ret;

	if (card->reset_flags & GO_SD20_DIRECT) {
		/* Set sd_switch_fail here, because we needn't switch
		 * to UHS mode */
		sd_card->sd_switch_fail = SDR104_SUPPORT_MASK;
		sd_card->sd_switch_fail |= DDR50_SUPPORT_MASK;
		sd_card->sd_switch_fail |= SDR50_SUPPORT_MASK;
	}

	/* check SCR "SD_SPEC" field before sending CMD6...  */
	/* Check the card whether follow SD1.1 spec or higher */
	ret = sd_check_spec(cr, card);
	if (ret == STATUS_SUCCESS) {
		ret = sd_switch_function(cr, card, sd_card->r.switch_bus_width);
		if (ret != STATUS_SUCCESS) {
			sd_change_bank_voltage(cr, SD_COMM_IO_3V3);
			sd_card->r.sd_dont_switch = 1;
			TRACE_RET(cr, STATUS_FAIL);
		}
	} else {
		if (sd_card->r.s18a) {
			sd_change_bank_voltage(cr, SD_COMM_IO_3V3);
			sd_card->r.sd_dont_switch = 1;
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return 0;
}

static int sd_do_tuning_action(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	int read_lba0 = 1;
	int ret;

	if ((sd_set_init_para(cr, card)))
		TRACE_RET(cr, STATUS_FATAL_ERROR);

	if (CHK_SD_DDR50(sd_card))
		ret = sd_ddr_tuning(cr, card);
	else
		ret = sd_sdr_tuning(cr, card);

	if (ret != STATUS_SUCCESS) {
		sd_change_bank_voltage(cr, SD_COMM_IO_3V3);
		card->reset_flags |= GO_SD20_DIRECT;
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* SEND_STATUS (CMD13) to clear potential error status */
	sd_clear_status(card);

	if (CHK_SD_DDR50(sd_card)) {
		/* for Panasonic UHS-I SD3.0 DDR mode issue.
		 * If the card's current state was "prog", the card will not
		 * reply response to host. So before issuing CMD17,
		 * we shall wait Card's Current State was in "tran"
		 * and "READY_FOR_DATA" was set */
		ret = sd_wait_state_data_ready(cr, card, 0x08, 1, 1000);
		if (ret != STATUS_SUCCESS)
			read_lba0 = 0;
	}

	if (read_lba0) {
		/* Read LBA0 here to improve compatibility with some card,
		 * such as Panasonic RP-SDY08G card */
		ret = sd_read_lba0(cr, card);
		if (ret != STATUS_SUCCESS) {
			sd_change_bank_voltage(cr, SD_COMM_IO_3V3);
			if (card->reset_flags & UHSI_FIX_TX_PHASE) {
				card->reset_flags &= ~(UHSI_FIX_TX_PHASE);
			}
			card->reset_flags |= GO_SD20_DIRECT;
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	return 0;
}

static int sd_check_cprm_state(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	u32 protected_area_size = 0;
	u8 cprm_security_sp = 0;
	u8 buf[64];
	struct sd_command cmd;

	rts_trace(TRACE_SD, "ACMD13:\n");

	/* Check ROM card or OTP through ACMD 13 */
	INIT_SD_CMD(cmd, 0x40 | SD_STATUS, 0, SD_APP_CMD, 0, 0);
	init_sd_dcmd_data(cmd, buf, sizeof(buf), sizeof(buf), 1, SD_BUS_WIDTH_4,
				SD_TM_NORMAL_READ, FROM_SD, 250);
	if (sd_send_cmd(cr, c, &cmd)) {
		sd_clear_sd_error(cr);
		sd_clear_status(c);
		TRACE_RET(cr, STATUS_FAIL);
	}

	RTS_DUMP(buf, 64);

	protected_area_size = ((u32)buf[4] << 24) | ((u32)buf[5] << 16)
				| ((u32)buf[6] << 8) | ((u32)buf[7]);
	cprm_security_sp = (sd_card->raw_scr[1] & 0x70) >> 4;

	rts_trace(TRACE_SD, "protected_area_size = %#x\n", protected_area_size);
	rts_trace(TRACE_SD, "cprm_security_sp = %#x\n", cprm_security_sp);

	if (0 == protected_area_size || 0 == cprm_security_sp)
		sd_card->sd_cprm_sp = 0;
	else
		sd_card->sd_cprm_sp = 1;
	rts_trace(TRACE_SD, "sd_cprm_sp = %#x\n", sd_card->sd_cprm_sp);

	return 0;
}

static int sd_reset_legacy_card(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	int retval;
	int i = 0;
	int retry = 3;

	SET_SD(sd_card);
	c->reset_flags &= ~(RESET_UHSII_NOW);

	for (i = 0; i < retry; i++) {
		rts_trace(TRACE_SD, "reset flags %#x\n", c->reset_flags);

		if ((sd_init_power(cr)))
			TRACE_RET(cr, STATUS_FAIL);

		if (!(sd_card->sd_lock_status & SD_UNLOCK_POW_ON)) {
			if ((sd_prepare_reset(cr, c)))
				TRACE_RET(cr, STATUS_FAIL);

			if (sd_dummy_clock(cr))
				TRACE_RET(cr, STATUS_FAIL);

			/* TODO: check if SDIO card */

			retval = sd_do_reset_legacy_card(c);
			if (retval) {
				if (retval == STATUS_FATAL_ERROR)
					TRACE_RET(cr , STATUS_FATAL_ERROR);
				else
					continue;
			}
		}

		/* Get SD lock status */
		if ((sd_update_lock_status(cr, c)))
			TRACE_RET(cr, STATUS_FAIL);

		if (sd_card->sd_lock_status & SD_LOCKED) {
			sd_card->sd_lock_status |= (SD_LOCK_1BIT_MODE |
					SD_PWD_EXIST);
			return STATUS_SUCCESS;
		} else if (!(sd_card->sd_lock_status & SD_UNLOCK_POW_ON)) {
			sd_card->sd_lock_status &= ~SD_PWD_EXIST;
		}

		/* ACMD42: Disconnect 50Kohm pull-up registor on CD/DAT3 pin */
		if ((sd_app_set_clr_card_detect(c)))
			TRACE_RET(cr, STATUS_FAIL);

		if (sd_card->r.s18a) {
			if ((sd_app_set_bus_width(c, 2)))
				TRACE_RET(cr, STATUS_FAIL);
			sd_card->r.switch_bus_width = SD_BUS_WIDTH_4;
		} else {
			sd_card->r.switch_bus_width = SD_BUS_WIDTH_1;
		}

		/* Set block length 512 bytes for all block commands */
		if ((sd_set_blocklen(c, 0x200, NULL)))
			TRACE_RET(cr, STATUS_FAIL);

		/* If we want to use the NORMAL_READ_MODE, we shall not set
		 * "Clock Divider" */
		if ((sd_set_clock_divider(cr, SD_CLK_DIVIDE_0)))
			TRACE_RET(cr, STATUS_FAIL);

		if (sd_get_raw_scr(cr, c, sd_card->r.switch_bus_width))
			TRACE_RET(cr, STATUS_FAIL);

		/* Not support class ten */
		if (!(sd_card->raw_csd[4] & 0x40))
			sd_card->r.sd_dont_switch = 1;

		if (!sd_card->r.sd_dont_switch) {
			if ((sd_do_switch_action(c)))
				continue;
		}

		if (sd_card->r.s18a) {
			/* Set pad driving strength to 16mA */
			retval = sd_sd30_drive_sel(cr, SD_IO_1V8);
		} else {
			/* ACMD6: Set bus width For Transcend SD 4.0G card
			 * issue: Card read fail when enable 4 bit data bus
			 * before sending CMD6; So now enable SD 4 bit data bus
			 * after sending CMD6. */
			/* Enable 4 bit data bus */
			if ((sd_app_set_bus_width(c, 2)))
				TRACE_RET(cr, STATUS_FAIL);
		}

		/* clear 1 bit mode status */
		sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;
		if (!(c->reset_flags & GO_SD20_DIRECT)
					&& CHK_SD30_SPEED(sd_card)) {
			retval = sd_do_tuning_action(c);
			if (retval) {
				if (c->reset_flags & UHSI_FIX_TX_PHASE) {
					c->reset_flags &= ~(GO_SD20_DIRECT);
				}
				if (retval == STATUS_FATAL_ERROR)
					TRACE_RET(cr, STATUS_FATAL_ERROR);
				else
					continue;
			}
		}

		retval = sd_check_wp_state(cr, c);
		if (retval != STATUS_SUCCESS) {
			if (CHK_SD30_SPEED(sd_card)
				&& !(c->reset_flags & GO_SD20_DIRECT)) {
				c->reset_flags |= GO_SD20_DIRECT;
				if (c->reset_flags & UHSI_FIX_TX_PHASE) {
					c->reset_flags &= ~(UHSI_FIX_TX_PHASE);
				}
				rts_trace(TRACE_SD,
					"ACMD13 fail, try to reset card in sd 2.0 mode\n");

				continue;
			}
			TRACE_RET(cr, STATUS_FAIL);
		}

		retval = sd_check_cprm_state(cr, c);
		if (retval != STATUS_SUCCESS) {
			if (CHK_SD30_SPEED(sd_card)
				&& !(c->reset_flags & GO_SD20_DIRECT)) {
				c->reset_flags |= GO_SD20_DIRECT;
				if (c->reset_flags & UHSI_FIX_TX_PHASE) {
					c->reset_flags &= ~(UHSI_FIX_TX_PHASE);
				}
				rts_trace(TRACE_SD,
					"ACMD13 fail, try to reset card in sd 2.0 mode\n");

				continue;
			}
			TRACE_RET(cr, STATUS_FAIL);
		}

		if (sd_cmd_support(c, READ_EXTR_SINGLE)) {
			sd_init_extr_param(cr, c);
		}

		break;
	}

	if (i >= retry)
		TRACE_RET(cr, STATUS_FAIL);

	c->bus_width = 4;
	if (sd_card->sd_lock_status & SD_UNLOCK_POW_ON)
		sd_set_blk_count(cr, 0x0200); /* blk counts: 512*/

	return STATUS_SUCCESS;
}

/*
 * sd_do_reset_uhsii_card - Start Initialization Process of SD Card
 */
static int sd_do_reset_uhsii_card(struct card *card)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;

	/* don't issue CMD0 for UHSII */
	/* if ((sd_go_idle_state(card)))
		TRACE_RET(cr, STATUS_FAIL);*/

	msleep(20);

	/* check if SDSC card */
	sd_verify_if_conf(card);

	if ((sd_verify_op_conf(card))) {
		if (cr_detect_card(cr, SD_ID)) {
			sd_set_error_code(cr, card, SD_NO_CARD);
			TRACE_RET(cr, STATUS_FATAL_ERROR);
		}
		TRACE_RET(cr, STATUS_FAIL);
	}
	rts_trace(TRACE_SD, "support_1v8 = %d\n", sd_card->r.s18a);

	/* CMD 2 */
	if ((sd_send_cid(card, sd_card->raw_cid)))
		TRACE_RET(cr, STATUS_FAIL);
	rts_trace(TRACE_SD, "CID Response:\n");
	hex_dump(sd_card->raw_cid, 16);

	if (sd_asign_relative_addr(card))
		TRACE_RET(cr, STATUS_FAIL);

	/* Get CSD register for Calculating Timing,Capacity
	 * Check CSD to determaine as if this is the SD ROM card  */
	if ((sd_check_csd(cr, card, 1)))
		TRACE_RET(cr, STATUS_FAIL);

	/* Select SD card */
	if ((sd_select_card(cr, card, 1)))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int sd_do_init_sd40(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;

	if (0 == sd40_initialization(cr, card)) {
		SET_SD_UHSII_FDX15(sd_card);
		SET_SD_HCXC(sd_card);
		cr_set_card_status(cr, CARD_TYPE_UHSII);
	} else {
		rts_trace(TRACE_ERROR, "SD40 Init Fail\n");

		SET_SD(sd_card);

		sd40_uninit_hw(cr, card);

		TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

/* some UHSII of Legacy card may not be compatible with CR, so pathc it */
static int sd_patch_spcial_card(struct card_reader *cr, struct card *card)
{
	if (cr->sd_ops_prot->patch_special_card)
		return cr->sd_ops_prot->patch_special_card(cr, card);

	return 0;
}

static int sd_reset_uhsii_card(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;
	int retval;

	rts_trace(TRACE_SD, "%s\n", __func__);

	SET_SD(sd_card);
	c->reset_flags |= RESET_UHSII_NOW;

	if ((sd_init_power(cr)))
		TRACE_RET(cr, STATUS_FAIL);

	if (!(sd_card->sd_lock_status & SD_UNLOCK_POW_ON)) {
		if ((sd_prepare_reset(cr, c)))
			TRACE_RET(cr, STATUS_FAIL);

		if (sd_do_init_sd40(cr, c))
			TRACE_RET(cr, STATUS_FAIL);

		if (check_sd40_flag(cr, SD40_GO_LEGACY_WAY)) {
			rts_trace(TRACE_SD, "****Go SD40 Legacy way\n");
		} else {
			rts_trace(TRACE_SD, "****Do not go legacy way\n");
			return 0;
		}

		/* don't issue CMD5 for UHSII */
		/* if (sd_check_sdio_card_exist(c))
			TRACE_RET(cr, STATUS_FAIL);*/

		retval = sd_do_reset_uhsii_card(c);
		if (retval)
			TRACE_RET(cr, retval);
	}

	if (sd_get_raw_scr(cr, c, SD_BUS_WIDTH_4))
		TRACE_RET(cr, STATUS_FAIL);

	/* Get SD lock status */
	if ((sd_update_lock_status(cr, c)))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_card->sd_lock_status & SD_LOCKED) {
		sd_card->sd_lock_status |= (SD_LOCK_1BIT_MODE | SD_PWD_EXIST);
		return STATUS_SUCCESS;
	} else if (!(sd_card->sd_lock_status & SD_UNLOCK_POW_ON)) {
		sd_card->sd_lock_status &= ~SD_PWD_EXIST;
	}

	retval = sd_switch_function(cr, c, 0);
	if (retval) {
		rts_trace(TRACE_SD, "UHS-II switch power fail\n");
		TRACE_RET(cr, STATUS_FAIL);
	} else {
		rts_trace(TRACE_SD, "UHS-II switch power success\n");
	}

	/* Not support class ten */
	if (!(sd_card->raw_csd[4] & 0x40))
		sd_card->r.sd_dont_switch = 1;

	/* clear 1 bit mode status */
	sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;

	retval = sd_check_wp_state(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	c->bus_width = 4;

	/* restore max_blklen: 0x200 */
	if ((CHK_SD_UHSII(sd_card))) {
		u16 addr;
		u8 val;

		addr = (SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_SET_L;
		addr = (addr << 2) + 2;
		cr_read_host_sd40_iospace(cr, addr, &val, 1);
		val &= 0x0F;
		cr_write_host_sd40_iospace(cr, addr, &val, 1);

		addr = (SD40_IOADR_CFG_BASE << 8) + SD40_IOADR_LINK_SET_L;
		addr = (addr << 2) + 3;
		val = 0x20;
		cr_write_host_sd40_iospace(cr, addr, &val, 1);
	}

	if (sd_card->sd_lock_status & SD_UNLOCK_POW_ON)
		sd_set_blk_count(cr, 0x0200); /* blk counts: 512*/

	sd_patch_spcial_card(cr, c);

	retval = sd_check_cprm_state(cr, c);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_cmd_support(c, READ_EXTR_SINGLE))
		sd_init_extr_param(cr, c);

	return STATUS_SUCCESS;
}

int sd_reset_sd(struct card_reader *cr, struct card *c)
{
	u32 sd_cap = cr_get_sd_capability_sw(cr);
	struct cr_option *option = &(cr->option);

	if (!(sd_cap & (SD_CAP_HS | SD_CAP_UHSI | SD_CAP_UHSII)))
		return -1;

	if (c->reset_flags & CONCRETE_MMC_CARD)
		return -1;

	if (c->reset_flags & RESET_MEET_OCP)
		return -1;

	if (cr->sd_ops_prot->reset_sd) {
		return cr->sd_ops_prot->reset_sd(cr, c);
	} else {
		u32 cap_uhsii = sd_cap & SD_CAP_UHSII;
		u32 go_legacy = c->reset_flags & GO_SD_LEGACY_DIRECT;

		if (cap_uhsii && (!go_legacy)) {
			int ret = 0;

			ret = sd_reset_uhsii_card(cr, c);
			if (0 == ret) {
				return 0;
			} else {
				/* continue resetting UHSII or legacy SD */
				if (!check_sd_flag(cr, SD_COMM_RESET_MMC_FIRST))
					c->reset_flags |= CONTINUE_RESETTING_SD;

				sd40_uninit_hw(cr, c);
				if (STATUS_FATAL_ERROR == ret)
					return ret;
				if ((c->reset_retry >= 1) || !(option->sd40_initial_retry_enable)) {
					c->reset_flags |= GO_SD_LEGACY_DIRECT;
				} else {
					c->reset_flags |=
						PATCH_SPECIAL_SD40_PHY;
				}

				return ret;
			}
		}

		if ((sd_cap & SD_CAP_HS) && !(sd_cap & SD_CAP_UHSI))
			c->reset_flags |= GO_SD20_DIRECT;

		if (sd_cap & (SD_CAP_HS | SD_CAP_UHSI))
			return sd_reset_legacy_card(cr, c);
	}

	return -1;
}
EXPORT_SYMBOL_GPL(sd_reset_sd);

static int sd_reset_mmc_first(struct card_reader *cr, struct card *card)
{
	int ret = 0;

	ret = sd_reset_mmc(cr, card);
	if (ret) {
		if (sd_check_error_code(cr, card, SD_NO_CARD))
			return -1;
		ret = sd_reset_sd(cr, card);
		if (ret)
			sd_change_bank_voltage(cr, SD_COMM_IO_3V3);
		return ret;
	}

	return ret;
}

static int sd_reset_sd_first(struct card_reader *cr, struct card *card)
{
	if (0 == sd_reset_sd(cr, card)) {
		return 0;
	} else {
		if (card->reset_flags & CONTINUE_RESETTING_SD) {
			card->reset_flags &= ~CONTINUE_RESETTING_SD;
			return -1;
		}

		if (check_sd_flag(cr, SD40_NO_MMC_FLOW)) {
			rts_trace(TRACE_SD,
				"*****SD reset fail, return directly\n");
			return -1;
		}
		if (sd_check_error_code(cr, card, SD_NO_CARD))
			return -1;
		if (sd_change_bank_voltage(cr, SD_COMM_IO_3V3))
			return -1;

		return sd_reset_mmc(cr, card);
	}

}

static void sd_init_sd_info(struct sd_info *sd_card)
{
	sd_card->sd_default_rx_phase = 15;
	sd_card->sd_default_tx_phase = 15;
}

int sd_reset_card(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct cr_option *option = &(cr->option);
	u32 sd_cap = cr_get_sd_capability_sw(cr);
	int ret;

	memset(sd_card, 0, sizeof(struct sd_info));
	sd_init_sd_info(sd_card);
	card->capacity = 0;

	rts_trace(TRACE_SD, "SD capability:%#x\n", sd_cap);
	rts_trace(TRACE_SD, "SD reset flags:%#x\n", card->reset_flags);

	if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
		rts_trace(TRACE_SD, "SD card not exist\n");
		sd_set_error_code(cr, card, SD_NO_CARD);
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (cr_enable_card_clk(cr, SD_ID))
		TRACE_RET(cr, STATUS_FAIL);

	if (option->ignore_sd) {
		if (sd_pull_ctl_enable(cr))
			TRACE_RET(cr, STATUS_FAIL);
		if (cr_select_card(cr, card))
			TRACE_RET(cr, STATUS_FAIL);
		SET_SDIO_EXIST(cr);
	}

	/* if (sd_init_power(cr))
		TRACE_RET(cr, STATUS_FAIL); */

	if (check_sd_flag(cr, SD_COMM_RESET_MMC_FIRST))
		ret = sd_reset_mmc_first(cr, card);
	else
		ret = sd_reset_sd_first(cr, card);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_set_clock_divider(cr, SD_CLK_DIVIDE_0))
		TRACE_RET(cr, STATUS_FAIL);

	sd_set_byte_count_per_blk(cr, BCPB_512B);
	card->capacity = sd_card->capacity;

	if (sd_set_init_para(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_init_sd_cfg_reg(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "sd_card->sd_type = 0x%x\n", sd_card->sd_type);
	rts_trace(TRACE_SD, "SD capacity: 0x%X\n", card->capacity);

	return 0;
}
EXPORT_SYMBOL_GPL(sd_reset_card);

static int sd_comm_prepare_reset(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int retval;

	if (cr->asic_code)
		sd_card->sd_clock = 29;
	else
		sd_card->sd_clock = CLK_30;

	SET_SD(sd_card);
	sd_card->sd_addr = 0;
	sd_card->rw_mode = NORMAL_RW;
	sd_card->sd_data_buf_ready = 0;
	card->capacity = 0;
	sd_card->sd_lock_status = 0;
	sd_card->sd_erase_status = 0;
	card->capacity = 0;

	/* init for UHSII */
	sd_card->sd40_2L_HD = 0;
	sd_card->sd40_did = 0;
	sd_card->sd40_sid_tid = 0;
	sd_card->hibernate_mode_en = 0;
	sd_card->lan_mode = DM_FD;

	retval = sd_set_init_para(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, retval);

	/* Initialize SD_CFG1 register */
	cr_wr_reg(cr, REG_SD_CFG1, 0xFF,
		SD_CLK_DIVIDE_128 | SD_20_MODE | SD_BUS_WIDTH_1);
	cr_wr_reg(cr, REG_SD_SAMPLE_POINT_CTL, 0xFF, SD20_RX_POS_EDGE);
	cr_wr_reg(cr, REG_SD_PUSH_POINT_CTL, 0xFF, 0);
	cr_wr_reg(cr, REG_CARD_STOP, SD_STOP | SD_CLR_ERR,
						SD_STOP | SD_CLR_ERR);

	/* Reset SD_CFG3 register */
	cr_wr_reg(cr, REG_SD_CFG3, SD30_CLK_END_EN, 0);
	cr_wr_reg(cr, REG_SD_STOP_SDCLK_CFG,
		SD30_CLK_STOP_CFG_EN | SD30_CLK_STOP_CFG1 | SD30_CLK_STOP_CFG0,
		0);

	if (cr_select_card(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int sd_comm_set_byte_count_per_blk(struct card_reader *cr, int bcpb)
{
	u8 cl, ch = 0;

	cl = bcpb & 0xff;
	ch = (bcpb >> 8) & 0xff;
	rts_trace(TRACE_SD, "REG_SD_BYTE_CNT_L = 0x%x\n", cl);
	rts_trace(TRACE_SD, "REG_SD_BYTE_CNT_H = 0x%x\n", ch);
	cr_wr_reg(cr, REG_SD_BYTE_CNT_L, 0xFF, cl);
	cr_wr_reg(cr, REG_SD_BYTE_CNT_H, 0xFF, ch);

	return STATUS_SUCCESS;
}

static int sd_comm_set_blk_count(struct card_reader *cr, int count)
{
	u8 cl, ch = 0;

	cl = count & 0xff;
	ch = (count >> 8) & 0xff;

	rts_trace(TRACE_SD, "REG_SD_BLOCK_CNT_L = 0x%x\n", cl);
	rts_trace(TRACE_SD, "REG_SD_BLOCK_CNT_H = 0x%x\n", ch);

	cr_wr_reg(cr, REG_SD_BLOCK_CNT_H, 0xFF, ch);
	cr_wr_reg(cr, REG_SD_BLOCK_CNT_L, 0xFF, cl);

	return 0;
}

static int sd_comm_set_init_para(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int retval;

	if (CHK_SD_UHSII(sd_card)) {
		rts_trace(TRACE_SD, "Exit %s\n", __func__);
		return 0;
	}

	retval = sd_set_sample_push_timing(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	sd_choose_proper_clock(cr, card);
	retval = cr_switch_clk(cr, card, sd_card->sd_clock);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return 0;
}

static int sd_comm_set_clock_divider(struct card_reader *cr, u8 clk_div)
{
	cr_wr_reg(cr, REG_SD_CFG1, SD_CLK_DIVIDE_MASK, clk_div);
	return 0;
}

static int sd_comm_get_crc7(struct card_reader *cr, u8 *buf)
{
	cr_rd_reg(cr, REG_SD_CMD5, buf);
	return 0;
}


static int sd_comm_current_over_limit400_handler(struct card_reader *cr)
{
	struct cr_option *option = &(cr->option);

	cr_wr_reg(cr, REG_OCPPARA2, SD_OCP_THD_MASK, option->sd_800mA_ocp_thd);
	/*cr_wr_reg(cr, REG_CARD_PWR_CTL, PMOS_STRG_MASK, PMOS_STRG_800mA);*/

	return 0;
}

static int sd_comm_check_crc16_err(struct card_reader *cr)
{
	u8 stat = 0;

	cr_rd_reg(cr, REG_SD_STAT1, &stat);
	if (stat & SD_CRC16_ERR) {
		rts_trace(TRACE_SD, "SD CRC16 error"
				"when switching mode\n");
		TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

static int sd_comm_set_clk_pol_ddr_mod(struct card_reader *cr,
						enum sd_comm_polarity_sol sol)
{
	u8 val = 0;

	switch (sol) {
	case SD_COMM_POL_SOL_ANG:
		val = 0x00;
		break;
	case SD_COMM_POL_SOL_DTL:
		val = 0x02;
		break;
	case SD_COMM_POL_SOL_DRV:
		val = 0x04;
		break;
	default:
		rts_trace(TRACE_SD, "the solution %d is not prohibited\n", sol);
		return STATUS_FAIL;
	}

	cr_wr_reg(cr, REG_SD_PUSH_POINT_CTL, 0x06, val);

	return STATUS_SUCCESS;
}

static int sd_set_sample_push_timing_sd20(struct card *c)
{
	struct card_reader *cr = c->cm->cr;
	struct sd_info *sd_card = c->extra;
	u8 val = 0;

	cr_wr_reg(cr, REG_SD_CFG1, 0x0C, SD_20_MODE);
	cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, CLK_LOW_FREQ);
	cr_wr_reg(cr, REG_CARD_CLK_SOURCE, 0xFF,
			CRC_FIX_CLK | SD30_VAR_CLK0 | SAMPLE_VAR_CLK1);
	cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, 0);

	if (check_sd_flag(cr, SD_PUSH_POINT_AUTO))
		val = SD20_TX_NEG_EDGE;
	else if (check_sd_flag(cr, SD_PUSH_POINT_DELAY))
		val = SD20_TX_14_AHEAD;
	else
		val = SD20_TX_NEG_EDGE;

	cr_wr_reg(cr, REG_SD_PUSH_POINT_CTL, SD20_TX_SEL_MASK, val);

	if (check_sd_flag(cr, SD_SAMPLE_POINT_AUTO)) {
		if (cr->asic_code) {
			if (CHK_SD_HS(sd_card) || CHK_MMC_52M(sd_card))
				val = SD20_RX_14_DELAY;
			else
				val = SD20_RX_POS_EDGE;
		} else {
			val = SD20_RX_14_DELAY;
		}
	} else if (check_sd_flag(cr, SD_SAMPLE_POINT_DELAY)) {
		val = SD20_RX_14_DELAY;
	} else {
		val = SD20_RX_POS_EDGE;
	}

	cr_wr_reg(cr, REG_SD_SAMPLE_POINT_CTL, SD20_RX_SEL_MASK, val);

	return 0;
}

static int sd_set_sample_push_timing_sd30(struct card *c)
{
	struct card_reader *cr = c->cm->cr;
	struct sd_info *sd_card = c->extra;

	if (CHK_SD_SDR104(sd_card)
			|| CHK_SD_SDR50(sd_card) || CHK_MMC_SDR200(sd_card)) {
		cr_wr_reg(cr, REG_SD_CFG1, 0x0C | SD_ASYNC_FIFO_NOT_RST,
				SD_30_MODE | SD_ASYNC_FIFO_NOT_RST);
		cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, CLK_LOW_FREQ);
		cr_wr_reg(cr, REG_CARD_CLK_SOURCE, 0xFF,
				CRC_VAR_CLK0 | SD30_FIX_CLK | SAMPLE_VAR_CLK1);
		cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, 0);
	} else if (CHK_SD_DDR50(sd_card) || CHK_MMC_DDR52(sd_card)) {
		cr_wr_reg(cr, REG_SD_CFG1, 0x0C | SD_ASYNC_FIFO_NOT_RST,
				SD_DDR_MODE | SD_ASYNC_FIFO_NOT_RST);
		cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, CLK_LOW_FREQ);
		cr_wr_reg(cr, REG_CARD_CLK_SOURCE, 0xFF,
				CRC_VAR_CLK0 | SD30_FIX_CLK | SAMPLE_VAR_CLK1);
		cr_wr_reg(cr, REG_CLK_CTL, CLK_LOW_FREQ, 0);
		cr_wr_reg(cr, REG_SD_PUSH_POINT_CTL, DDR_VAR_TX_CMD_DAT,
				DDR_VAR_TX_CMD_DAT);
		cr_wr_reg(cr, REG_SD_SAMPLE_POINT_CTL,
			DDR_VAR_RX_DAT | DDR_VAR_RX_CMD,
			DDR_VAR_RX_DAT | DDR_VAR_RX_CMD);
	}

	return 0;
}

static int sd_comm_set_sample_push_timing(struct card_reader *cr,
							struct card *card)
{
	struct sd_info *sd_card = card->extra;

	if (CHK_SD_UHSII(sd_card)) {
		rts_trace(TRACE_SD, "**Exit %s\n", __func__);
		return 0;
	}

	if (CHK_SD30_SPEED(sd_card)
			|| CHK_MMC_DDR52(sd_card) || CHK_MMC_SDR200(sd_card))
		sd_set_sample_push_timing_sd30(card);
	else
		sd_set_sample_push_timing_sd20(card);

	return 0;
}

static int sd_comm_sd30_drive_sel(struct card_reader *cr, u8 voltage)
{
	struct cr_option *option = &(cr->option);
	u8 sd30_clk_drive_sel;
	u8 sd30_cmd_drive_sel;
	u8 sd30_dat_drive_sel;

	if (voltage == SD_IO_3V3) {
		sd30_clk_drive_sel = option->sd30_clk_drive_sel_3v3;
		sd30_cmd_drive_sel = option->sd30_cmd_drive_sel_3v3;
		sd30_dat_drive_sel = option->sd30_dat_drive_sel_3v3;
	} else {
		sd30_clk_drive_sel = option->sd30_clk_drive_sel_1v8;
		sd30_cmd_drive_sel = option->sd30_cmd_drive_sel_1v8;
		sd30_dat_drive_sel = option->sd30_dat_drive_sel_1v8;
	}

	cr_wr_reg(cr, SD30_CLK_DRIVE_SEL, 0xFF, sd30_clk_drive_sel);
	cr_wr_reg(cr, SD30_CMD_DRIVE_SEL, 0xFF, sd30_cmd_drive_sel);
	cr_wr_reg(cr, SD30_DAT_DRIVE_SEL, 0xFF, sd30_dat_drive_sel);

	return STATUS_SUCCESS;
}

static int sd_reset_dcm(struct card_reader *cr, u8 tune_dir)
{
	if (tune_dir == TUNE_RX) {
		cr_wr_reg(cr, REG_DCM_DRP_CTL, 0xFF,
					DCM_RESET | DCM_RX);
		cr_wr_reg(cr, REG_DCM_DRP_CTL, 0xFF, DCM_RX);
	} else {
		cr_wr_reg(cr, REG_DCM_DRP_CTL, 0xFF,
					DCM_RESET | DCM_TX);
		cr_wr_reg(cr, REG_DCM_DRP_CTL, 0xFF, DCM_TX);
	}

	return STATUS_SUCCESS;
}

static int sd_comm_change_phase(struct card_reader *cr, struct card *card,
						u8 sample_point, u8 tune_dir)
{
	struct sd_info *sd_card = card->extra;
	u16 SD_VP_CTL, SD_DCMPS_CTL;
	u8 val;
	int retval;
	int ddr_rx = 0;

	rts_trace(TRACE_SD,
		"sd_change_phase (sample_point = %d, tune_dir = %d)\n",
		sample_point, tune_dir);

	if (tune_dir == TUNE_RX) {
		SD_VP_CTL = REG_SD_VPRX_CTL;
		SD_DCMPS_CTL = REG_SD_DCMPS_RX_CTL;
		if (CHK_SD_DDR50(sd_card))
			ddr_rx = 1;
	} else {
		SD_VP_CTL = REG_SD_VPTX_CTL;
		SD_DCMPS_CTL = REG_SD_DCMPS_TX_CTL;
	}

	if (cr->asic_code) {
		cr_wr_reg(cr, REG_CLK_CTL, CHANGE_CLK, CHANGE_CLK);
		cr_wr_reg(cr, SD_VP_CTL, 0x1F, sample_point);
		cr_wr_reg(cr, REG_SD_VPCLK0_CTL, PHASE_NOT_RESET, 0);
		cr_wr_reg(cr, REG_SD_VPCLK0_CTL,
					PHASE_NOT_RESET, PHASE_NOT_RESET);
		cr_wr_reg(cr, REG_CLK_CTL, CHANGE_CLK, 0);
	} else {

		cr_rd_reg(cr, SD_VP_CTL, &val);
		rts_trace(TRACE_SD, "SD_VP_CTL: 0x%x\n", val);
		cr_rd_reg(cr, SD_DCMPS_CTL, &val);
		rts_trace(TRACE_SD, "SD_DCMPS_CTL: 0x%x\n", val);

		if (ddr_rx) {
			cr_wr_reg(cr, SD_VP_CTL, PHASE_CHANGE, PHASE_CHANGE);
			udelay(50);
			cr_wr_reg(cr, SD_VP_CTL, 0xFF,
				PHASE_CHANGE | PHASE_NOT_RESET | sample_point);
		} else {
			cr_wr_reg(cr, REG_CLK_CTL, CHANGE_CLK, CHANGE_CLK);
			udelay(50);
			cr_wr_reg(cr, SD_VP_CTL, 0xFF,
					PHASE_NOT_RESET | sample_point);
		}
		udelay(100);

		cr_init_cmd(cr);
		cr_add_cmd(cr, WRITE_REG_CMD, SD_DCMPS_CTL,
					DCMPS_CHANGE, DCMPS_CHANGE);
		cr_add_cmd(cr, CHECK_REG_CMD, SD_DCMPS_CTL,
					DCMPS_CHANGE_DONE, DCMPS_CHANGE_DONE);
		retval = cr_send_cmd(cr, SD_CARD, 100);
		if (retval != STATUS_SUCCESS)
			TRACE_GOTO(cr, Fail);
		val = *cr_get_cmd_data(cr);
		if (val & DCMPS_ERROR)
			TRACE_GOTO(cr, Fail);
		if ((val & DCMPS_CURRENT_PHASE) != sample_point)
			TRACE_GOTO(cr, Fail);
		cr_wr_reg(cr, SD_DCMPS_CTL, DCMPS_CHANGE, 0);
		if (ddr_rx)
			cr_wr_reg(cr, SD_VP_CTL, PHASE_CHANGE, 0);
		else
			cr_wr_reg(cr, REG_CLK_CTL, CHANGE_CLK, 0);
		udelay(50);
	}

	cr_wr_reg(cr, REG_SD_CFG1, SD_ASYNC_FIFO_NOT_RST, 0);

	return STATUS_SUCCESS;

Fail:
	cr_rd_reg(cr, SD_VP_CTL, &val);
	rts_trace(TRACE_SD, "SD_VP_CTL: 0x%x\n", val);
	cr_rd_reg(cr, SD_DCMPS_CTL, &val);
	rts_trace(TRACE_SD, "SD_DCMPS_CTL: 0x%x\n", val);
	cr_wr_reg(cr, SD_DCMPS_CTL, DCMPS_CHANGE, 0);
	cr_wr_reg(cr, SD_VP_CTL, PHASE_CHANGE, 0);
	msleep(10);
	sd_reset_dcm(cr, tune_dir);

	return STATUS_FAIL;
}

static int sd_comm_wait_data_idle(struct card_reader *cr)
{
	int retval = STATUS_TIMEDOUT;
	int i;
	u8 val = 0;

	for (i = 0; i < 100; i++) {
		cr_rd_reg(cr, REG_SD_DATA_STATE, &val);
		if (val & SD_DATA_IDLE) {
			retval = STATUS_SUCCESS;
			break;
		}
		udelay(100);
	}
	rts_trace(TRACE_SD, "SD_DATA_STATE: 0x%02x\n", val);

	return retval;
}

static int sd_sdr_tuning_tx_cmd(struct card_reader *cr,
					struct card *card, u8 sample_point)
{
	int retval;

	retval = sd_change_phase(cr, card, sample_point, TUNE_TX);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN,
					SD_RSP_80CLK_TIMEOUT_EN);
	retval = sd_clear_status(card);
	if (retval != STATUS_SUCCESS) {
		if (sd_check_error_code(cr, card, SD_RSP_TIMEOUT)) {
			/* Tunning TX fail */
			cr_wr_reg(cr, REG_SD_CFG3,
					SD_RSP_80CLK_TIMEOUT_EN, 0);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}
	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);

	return STATUS_SUCCESS;
}

static int sd_ddr_tuning_tx_cmd(struct card_reader *cr,
					struct card *card, u8 sample_point)
{
	struct sd_info *sd_card = card->extra;
	int retval;
	u8 bus_width;

	retval = sd_change_phase(cr, card, sample_point, TUNE_TX);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	if (CHK_SD(sd_card)) {
		bus_width = SD_BUS_WIDTH_4;
	} else {
		if (CHK_MMC_8BIT(sd_card))
			bus_width = SD_BUS_WIDTH_8;
		else if (CHK_MMC_4BIT(sd_card))
			bus_width = SD_BUS_WIDTH_4;
		else
			bus_width = SD_BUS_WIDTH_1;
	}

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN,
					SD_RSP_80CLK_TIMEOUT_EN);

	retval = sd_clear_status(card);
	if (retval != STATUS_SUCCESS) {
		if (sd_check_error_code(cr, card, SD_RSP_TIMEOUT)) {
			/* Tunning TX fail */
			cr_wr_reg(cr, REG_SD_CFG3,
					SD_RSP_80CLK_TIMEOUT_EN, 0);
			TRACE_RET(cr, STATUS_FAIL);
		}
	}

	cr_wr_reg(cr, REG_SD_CFG3, SD_RSP_80CLK_TIMEOUT_EN, 0);

	return STATUS_SUCCESS;
}


static u32 do_tuning(struct card_reader *cr,
	struct card *card, tuning_func func, u32 *map, u8 tune_dir)
{
	struct sd_info *sd_card = card->extra;
	int i = 0, j = 0;
	u32 raw_phase_map[3];
	int total_phase = cr->sd_max_phase + 1;
	int ddr_mode = !!(CHK_SD_DDR50(sd_card) || CHK_MMC_DDR52(sd_card));
	int shift = 0;

	if (ddr_mode && (TUNE_TX == tune_dir))
		shift = 1;
	else
		shift = 0;

	for (i = 0; i < 3; i++) {
		raw_phase_map[i] = 0;
		/* for DDR tuning, the phase must be from 0 to 31
		 * please refer to Mantis:4533
		 */
		for (j = 0; j < total_phase; j++) {
			if (cr_detect_card(cr, SD_ID) !=
					STATUS_SUCCESS) {
				sd_set_error_code(cr, card, SD_NO_CARD);
				cr_wr_reg(cr, REG_SD_CFG3,
						SD_RSP_80CLK_TIMEOUT_EN, 0);
				TRACE_RET(cr, STATUS_FAIL);
			}

			if ((j == (total_phase / 2)) && shift)
				cr_swap_sdclk_polarity(cr);

			if (0 == func(cr, card, (u8)(j << shift)))
				raw_phase_map[i] |= 1 << j;
		}

		if (shift)
			cr_swap_sdclk_polarity(cr);
	}

	*map = raw_phase_map[0] & raw_phase_map[1] & raw_phase_map[2];

	for (i = 0; i < 3; i++) {
		rts_trace(TRACE_SD, "raw_phase_map[%d] = 0x%08x\n",
				i, raw_phase_map[i]);
	}

	return 0;
}

/* sd_tuning_tx - Tune clock for the transmit direction
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int sd_tuning_tx(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int ret = 0;
	u32 phase_map;
	int final_phase;
	tuning_func tuning_cmd;

	if (CHK_SD(sd_card)) {
		if (CHK_SD_DDR50(sd_card))
			tuning_cmd = sd_ddr_tuning_tx_cmd;
		else
			tuning_cmd = sd_sdr_tuning_tx_cmd;
	} else {
		if (CHK_MMC_DDR52(sd_card))
			tuning_cmd = sd_ddr_tuning_tx_cmd;
		else if (CHK_MMC_SDR200(sd_card))
			tuning_cmd = sd_sdr_tuning_tx_cmd;
		else
			TRACE_RET(cr, STATUS_FAIL);
	}

	if (do_tuning(cr, card, tuning_cmd, &phase_map, TUNE_TX))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "TX(FST) phase_map = 0x%08x\n", phase_map);

	if (!phase_map) {
		card->reset_flags |= UHSI_FIX_TX_PHASE;
		TRACE_RET(cr, STATUS_FAIL);
	}

	if (0xFFFFFFFF != phase_map) {
		final_phase = sd_get_final_phase(card, phase_map, TUNE_TX);
		if (final_phase < 0)
			TRACE_RET(cr, STATUS_FAIL);
	} else {
		u8 val = 0;
		if (CHK_SD_SDR104(sd_card) || CHK_MMC_SDR200(sd_card))
			val = 0xDD;
		else
			val = 0xEE;
		cr_wr_reg(cr, REG_SD_DUTY_CTL, 0xFF, val);
		ret = do_tuning(cr, card, tuning_cmd, &phase_map, TUNE_TX);
		rts_trace(TRACE_SD, "TX(SEC) phase_map = 0x%08x\n", phase_map);
		cr_wr_reg(cr, REG_SD_DUTY_CTL, 0xFF, 0);
		if (ret)
			TRACE_RET(cr, STATUS_FAIL);

		final_phase = sd_get_final_phase(card, phase_map, TUNE_TX);
		if (final_phase < 0)
			TRACE_RET(cr, STATUS_FAIL);
		if (CHK_SD_SDR104(sd_card) || CHK_MMC_SDR200(sd_card))
			final_phase += 2;
		else
			final_phase += 3;

		final_phase %= (cr->sd_max_phase + 1);
	}

	if (sd_change_phase(cr, card, (u8) final_phase, TUNE_TX))
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

/* sd_read_lba0 - Read LBA0
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: STATUS_SUCCESS for success, otherwise nonzero
 */
static int sd_read_lba0(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct sd_command cmd;
	u8 bus_width;

	if (CHK_SD(sd_card)) {
		bus_width = SD_BUS_WIDTH_4;
	} else {
		if (CHK_MMC_8BIT(sd_card))
			bus_width = SD_BUS_WIDTH_8;
		else if (CHK_MMC_4BIT(sd_card))
			bus_width = SD_BUS_WIDTH_4;
		else
			bus_width = SD_BUS_WIDTH_1;
	}

	INIT_SD_CMD(cmd, 0x40 | READ_SINGLE_BLOCK, 0, 0, 0, 0);
	init_sd_dcmd_data(cmd, NULL, 0, 512, 1, bus_width,
				SD_TM_NORMAL_READ, FROM_SD, 100);
	if (sd_send_cmd(cr, card, &cmd)) {
		sd_clear_sd_error(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}

	return 0;
}

static int mmc_test_switch_bus(struct card_reader *cr,
						struct card *card, u8 width)
{
	struct sd_info *sd_card = card->extra;
	struct sd_command cmd;
	int retval;
	u8 buf[8] = {0}, bus_width, *ptr;
	u16 len;

	retval = sd_bustest_w(card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	if (width == MMC_8BIT_BUS) {
		buf[0] = 0x55;
		buf[1] = 0xAA;
		len = 8;
		bus_width = SD_BUS_WIDTH_8;
	} else {
		buf[0] = 0x5A;
		len = 4;
		bus_width = SD_BUS_WIDTH_4;
	}

	INIT_SD_CMD(cmd, 0, 0, 0, 0, 0);
	/* INIT_SD_CMD(cmd, BUSTEST_W, 0, 0, 0, 0); */
	init_sd_dcmd_data(cmd, buf, len, len, 1, bus_width,
				SD_TM_AUTO_WRITE_3, TO_SD, 100);

	if (sd_send_cmd(cr, card, &cmd)) {
		u8 val1 = 0, val2 = 0;
		cr_rd_reg(cr, REG_SD_STAT1, &val1);
		cr_rd_reg(cr, REG_SD_STAT2, &val2);
		sd_clear_sd_error(cr);
		/*
		 * Only when CRC7, CRC16, writeCRC, and cmd RSP timeout
		 * are set, return failed
		 */
		if ((val1 & 0xE0) || val2)
			TRACE_RET(cr, STATUS_FAIL);
	}

	/* BUSTEST_R needn't check CRC16, so we don't call sd_read_data here */
	rts_trace(TRACE_SD, "SD/MMC CMD %d\n", BUSTEST_R);
	cr_init_cmd(cr);
	/* CMD14 */
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD0, 0xFF,
							0x40 | BUSTEST_R);
	if (width == MMC_8BIT_BUS)
		cr_add_cmd(cr, WRITE_REG_CMD,
				REG_SD_BYTE_CNT_L, 0xFF, 0x08);
	else
		cr_add_cmd(cr, WRITE_REG_CMD,
				REG_SD_BYTE_CNT_L, 0xFF, 0x04);

	cr_add_cmd(cr, WRITE_REG_CMD,
				REG_SD_BLOCK_CNT_L, 0xFF, 1);
	cr_add_cmd(cr, WRITE_REG_CMD,
				REG_SD_BLOCK_CNT_H, 0xFF, 0);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF,
			SD_CALCULATE_CRC7 | SD_NO_CHECK_CRC16 |
			SD_NO_WAIT_BUSY_END | SD_CHECK_CRC7 |
			SD_RSP_LEN_6);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE,
					0x01, PINGPONG_BUFFER);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
			SD_TM_NORMAL_READ | SD_TRANSFER_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
			SD_TRANSFER_END, SD_TRANSFER_END);
	cr_add_cmd(cr, READ_REG_CMD, PPBUF_BASE2, 0, 0);
	if (width == MMC_8BIT_BUS)
		cr_add_cmd(cr, READ_REG_CMD,
					PPBUF_BASE2 + 1, 0, 0);
	retval = cr_send_cmd(cr, SD_CARD, 100);
	if (retval < 0) {
		sd_clear_sd_error(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}
	ptr = cr_get_cmd_data(cr) + 1;
	/* Switch corresponding bus mode if data pattern is correct! */
	if (width == MMC_8BIT_BUS) {
		rts_trace(TRACE_SD, "BUSTEST_R [8bits]: 0x%02x 0x%02x\n",
							ptr[0], ptr[1]);
		if ((ptr[0] == 0xAA) && (ptr[1] == 0x55)) {
			u8 rsp[5];
			u32 arg;
			if (CHK_MMC_DDR52(sd_card))
				arg = 0x03B70600;
			else
				arg = 0x03B70200;
			/* Switch MMC to  8-bit mode */
			retval = sd_switch_func(card, arg, rsp);
			if ((retval == STATUS_SUCCESS) &&
				!(rsp[4] & MMC_SWITCH_ERR))
				return STATUS_SUCCESS;
		}
	} else {
		rts_trace(TRACE_SD, "BUSTEST_R [4bits]: 0x%02x\n", ptr[0]);
		if (ptr[0] == 0xA5) {
			u8 rsp[5];
			u32 arg;
			if (CHK_MMC_DDR52(sd_card))
				arg = 0x03B70500;
			else
				arg = 0x03B70100;
			/* Switch MMC to  4-bit mode */
			retval = sd_switch_func(card, arg, rsp);
			if ((retval == STATUS_SUCCESS) &&
				!(rsp[4] & MMC_SWITCH_ERR))
				return STATUS_SUCCESS;
		}
	}

	TRACE_RET(cr, STATUS_FAIL);
}

static int mmc_select_sdr200(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	u32 sd_cap = cr_get_sd_capability_sw(cr);
	u8 rsp[8];
	int ret = 0;

	/* if ((ret = sd_switch_func(card, 0x03B90200, rsp))) */
		/* TRACE_RET(cr, STATUS_FAIL); */

	/*sd_choose_proper_clock(cr, card);
	ret = cr_switch_clk(cr, card, sd_card->sd_clock);
	if (ret != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);*/

	/* Test Bus Procedure */
	if (sd_cap & SD_CAP_MMC_8BIT) {
		ret = mmc_test_switch_bus(card->cm->cr, card, MMC_8BIT_BUS);
		if (!ret) {
			SET_MMC_8BIT(sd_card);
			card->bus_width = 8;
			sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;
			ret = sd_switch_func(card, 0x03B90200, rsp);
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
			return 0;
		} else {
			CLR_MMC_4BIT(sd_card);
		}
	}

	if (sd_cap & SD_CAP_MMC_4BIT) {
		ret = mmc_test_switch_bus(card->cm->cr, card, MMC_4BIT_BUS);
		if (!ret) {
			SET_MMC_4BIT(sd_card);
			card->bus_width = 4;
			sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;
			ret = sd_switch_func(card, 0x03B90200, rsp);
			if (ret)
				TRACE_RET(cr, STATUS_FAIL);
			return 0;
		} else {
			CLR_MMC_4BIT(sd_card);
		}
	}

	card->bus_width = 1;

	return 0;
}

static int sd_comm_mmc_switch_timing_bus(struct card_reader *cr,
					struct card *card, int switch_ddr)
{
	struct sd_info *sd_card = card->extra;
	u32 sd_cap = cr_get_sd_capability_sw(cr);
	u8 rsp[8];
	int retval = 0;

	if (CHK_MMC_SDR200(sd_card))
		retval = sd_switch_func(card, 0x03B90200, rsp);
	else
		retval = sd_switch_func(card, 0x03B90100, rsp);

	if ((retval != STATUS_SUCCESS) || (rsp[4] & MMC_SWITCH_ERR)) {
		CLR_MMC_HS(sd_card);
		CLR_MMC_SDR200(sd_card);
	}

	/* Set MMC/SD Clocks to run normal clock */
	sd_choose_proper_clock(cr, card);
	retval = cr_switch_clk(cr, card, sd_card->sd_clock);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	/* Test Bus Procedure */
	if (sd_cap & SD_CAP_MMC_8BIT) {
		retval = mmc_test_switch_bus(cr, card, MMC_8BIT_BUS);
		if (retval == STATUS_SUCCESS) {
			SET_MMC_8BIT(sd_card);
			card->bus_width = 8;
			sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;
			return STATUS_SUCCESS;
		}
	}

	if (sd_cap & SD_CAP_MMC_4BIT) {
		retval = mmc_test_switch_bus(cr, card, MMC_4BIT_BUS);
		if (retval == STATUS_SUCCESS) {
			SET_MMC_4BIT(sd_card);
			card->bus_width = 4;
			sd_card->sd_lock_status &= ~SD_LOCK_1BIT_MODE;
			return STATUS_SUCCESS;
		}
	}

	CLR_MMC_8BIT(sd_card);
	CLR_MMC_4BIT(sd_card);

	return STATUS_SUCCESS;
}

static int sd_comm_dummy_clock(struct card_reader *cr)
{
	cr_wr_reg(cr, REG_SD_BUS_STAT, SD_CLK_TOGGLE_EN, SD_CLK_TOGGLE_EN);
	msleep(5);
	cr_wr_reg(cr, REG_SD_BUS_STAT, SD_CLK_TOGGLE_EN, 0x00);

	return 0;
}

/* sd_wait_data_buf_ready - Wait for data is ready in buffer
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 *
 * Description: When accessing Kingston MMC4.0 cards, we shall check
 * "READY_FOR_DATA" bit in "Card Status" of R1 those cards will report
 * READY_FOR_DATA=0 after MULTIPLE_WRITE_BLOCK command. Especially after "STOP"
 * transfer from "MULTIPLE_READ_BLOCK" status().
 */
static int sd_wait_data_buf_ready(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int i, retval;

	for (i = 0; i < WAIT_DATA_READY_RTY_CNT; i++) {
		if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
			sd_set_error_code(cr, card, SD_NO_CARD);
			TRACE_RET(cr, STATUS_FAIL);
		}
		sd_card->sd_data_buf_ready = 0;
		retval = sd_clear_status(card);
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
		if (sd_card->sd_data_buf_ready)
			return sd_clear_status(card);
	}
	sd_set_error_code(cr, card, SD_TO_ERR);
	TRACE_RET(cr, STATUS_FAIL);
}


int sd_switch_clock(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int retval;
	int re_tuning = 0;


	if (CHK_SD_UHSII(sd_card)) {
		rts_trace(TRACE_SD, "Exit from %s\n", __func__);
		return 0;
	}

	retval = cr_select_card(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	if ((CHK_SD30_SPEED(sd_card)
			|| CHK_MMC_DDR52(sd_card) || CHK_MMC_SDR200(sd_card))) {
		if (sd_card->need_retune &&
			(sd_card->sd_clock != cr->cur_clk)) {
			re_tuning = 1;
			sd_card->need_retune = 0;
		}
	}
	retval = cr_switch_clk(cr, card, sd_card->sd_clock);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	if (re_tuning) {
		if (CHK_SD(sd_card)) {
			if (CHK_SD_DDR50(sd_card))
				retval = sd_ddr_tuning(cr, card);
			else
				retval = sd_sdr_tuning(cr, card);
		} else {
			if (CHK_MMC_DDR52(sd_card))
				retval = mmc_ddr_tuning(cr, card);
			else if (CHK_MMC_SDR200(sd_card))
				retval = mmc_sdr_tuning(cr, card);
		}
		if (retval != STATUS_SUCCESS)
			TRACE_RET(cr, STATUS_FAIL);
	}

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL_GPL(sd_switch_clock);

static int sd_comm_auto_tune_clock(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int retval;

	if (cr->asic_code) {
		if (sd_card->sd_clock > 30)
			sd_card->sd_clock -= 20;
	} else {
		switch (sd_card->sd_clock) {
		case CLK_200:
			sd_card->sd_clock = CLK_150;
			break;

		case CLK_150:
			sd_card->sd_clock = CLK_120;
			break;

		case CLK_120:
			sd_card->sd_clock = CLK_100;
			break;

		case CLK_100:
			sd_card->sd_clock = CLK_80;
			break;

		case CLK_80:
			sd_card->sd_clock = CLK_60;
			break;

		case CLK_60:
			sd_card->sd_clock = CLK_50;
			break;

		default:
			break;
		}
	}

	retval = sd_switch_clock(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	return STATUS_SUCCESS;
}

/* sd_reset_mmc_only - Only reset MMC card
 * @cr: points to card reader structure
 * @c: points to card struct
 *
 * Returns: 0 for success, otherwise nonzero
 */
static int sd_reset_mmc_only(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	int retval;

	sd_card->sd_type = 0;
	sd_card->rw_mode = NORMAL_RW;
	sd_card->sd_data_buf_ready = 0;
	card->capacity = 0;
	sd_card->sd_switch_fail = 0;
#ifdef SUPPORT_SD_LOCK
	sd_card->sd_lock_status = 0;
	sd_card->sd_erase_status = 0;
#endif
	retval = cr_enable_card_clk(cr, SD_ID);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	retval = sd_init_power(cr);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	retval = sd_reset_mmc(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	retval = sd_set_clock_divider(cr, SD_CLK_DIVIDE_0);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);
	cr_wr_reg(cr, REG_SD_BYTE_CNT_L, 0xFF, 0);
	cr_wr_reg(cr, REG_SD_BYTE_CNT_H, 0xFF, 2);
	retval = sd_set_init_para(cr, card);
	if (retval != STATUS_SUCCESS)
		TRACE_RET(cr, STATUS_FAIL);

	if (sd_init_sd_cfg_reg(cr, card))
		TRACE_RET(cr, STATUS_FAIL);

	rts_trace(TRACE_SD, "In reset_mmc_only, sd_card->sd_type = 0x%x\n",
							sd_card->sd_type);

	return STATUS_SUCCESS;
}

static void sd_print_srb_info(struct scsi_cmnd *srb, u32 start_sector,
		u16 sector_cnt)
{
	if (srb->sc_data_direction == DMA_FROM_DEVICE) {
		rts_trace(TRACE_SD, "sd_rw: Read %d %s from 0x%x\n", sector_cnt,
				(sector_cnt > 1) ? "sectors" : "sector",
				start_sector);
	} else {
		rts_trace(TRACE_SD, "sd_rw: Write %d %s to 0x%x\n", sector_cnt,
				(sector_cnt > 1) ? "sectors" : "sector",
				start_sector);
	}
}

static int sd_check_card_state_befor_rw(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int ret;

	if (!check_card_flag(card, READY)) {
		sd_card->rw_mode = NORMAL_RW;
		if (3 == card->rw_retry_cnt)
			card->reset_flags |= GO_SD_LEGACY_DIRECT;
		ret = sd_reset_card(cr, card);
		if (ret == STATUS_SUCCESS) {
			set_card_flags(card, READY);
			clr_card_flags(card, FAIL);
		} else {
			clr_card_flags(card, READY);
			set_card_flags(card, FAIL);
			card->capacity = 0;
			card->rw_need_retry = 1;
			return ret;
		}
	}

	return 0;
}

int sd_comm_stop_sequential(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int ret  = 0;

	if ((DMA_FROM_DEVICE == sd_card->prev_dir
			&& PRE_READ == sd_card->rw_mode)
			|| (DMA_TO_DEVICE == sd_card->prev_dir
			&& PRE_WRITE == sd_card->rw_mode)) {
		u32 reg, value;

		value = cr_ioread32(cr, HCBCTLR);
		cr_iowrite32(cr, HCBCTLR, value | STOP_CMD);
		value = cr_ioread32(cr, HDBCTLR);
		cr_iowrite32(cr, HDBCTLR, value | STOP_DMA);

		if (CHK_SD_UHSII(sd_card)) {
			if (CHECK_PID(cr, 0x524a) || CHECK_PID(cr, 0x525a))
				cr_wr_reg(cr, SD40_STOP_CTL ,
					SD40_LINK_RST, SD40_LINK_RST);
			cr_wr_reg(cr, REG_SD40_CTRL,
					SD40_MAC_RST, SD40_MAC_RST);
		}

		sd_clear_sd_error(cr);
		cr_wr_reg(cr, REG_DMACTL, DMA_RST, DMA_RST);

		reg = cr->hw_param.interrupt_en & (~DATA_DONE_INT_EN);
		cr_enable_bus_int(cr, reg);
	}

	/*
	 * For PQI MMC 32MB burn in fail:
	 * Use our own burn in tool, random burn in,
	 * write error will occur. The stop of previous read command
	 * is sent too early so that the following write command will
	 * fail. The phenomina is write one page data to card and
	 * no CRC status from card. So we add one dummy cmd13 between
	 * read command and stop command random rw,
	 * so prev_sec_cnt < 0x80
	 */
	if ((sd_card->prev_sec_cnt < 0x80)
		&& (sd_card->prev_dir == DMA_FROM_DEVICE)
		&& !CHK_SD_UHS(sd_card)
		&& !CHK_SD_HS(sd_card)
		&& !CHK_MMC_HS(sd_card))
		sd_clear_status(card);

	ret = sd_stop_transmission(card);
	if (ret) {
		card->rw_need_retry = 1;
		sd_set_error_code(cr, card, SD_STS_ERR);
		sd_set_error_code(cr, card, SD_CMD12_ERR);
		return ret;
	}

	/* For a SDXC card:
	 * Sometimes this card responds cmd12 too quickly, and this would
	 * cause next cmd18 fail. So delay 50us here
	 */

	udelay(50);

	sd_card->rw_mode = NORMAL_RW;
	ret = cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
	if (ret) {
		sd_set_error_code(cr, card, SD_IO_ERR);
		return ret;
	}
	/*
	 * For PRETEC MMC 32MB card, we shall delay a moment
	 * after "STOP" command before next R/W command.
	 */
	/* random rw, so prev_sec_cnt < 0x80 */
	if ((sd_card->prev_sec_cnt < 0x80) && !CHK_SD30_SPEED(sd_card)
			&& !CHK_SD_HS(sd_card) && !CHK_MMC_HS(sd_card))
		sd_clear_status(card);

	return 0;
}
int sd_stop_sequential(struct card *card)
{
	struct card_reader *cr = card->cm->cr;

	if (cr->sd_ops_prot->stop_sequential)
		return cr->sd_ops_prot->stop_sequential(card);
	else
		return sd_comm_stop_sequential(card);
}
EXPORT_SYMBOL_GPL(sd_stop_sequential);

int scsi_get_sector_info(struct scsi_cmnd *srb, u32 *lba, u16 *num);

static int sd_prepare_write(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u32 data_addr;
	u16 sector_cnt;
	u8 cfg2;
	int ret;

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	/*
	 * High-Capacity Card Addressing Mode is in "Block"
	 * while Standard Capacity Card is in "Byte",
	 * So We should check Capacity mode to
	 * determine parameter "Address"
	 */
	if (!CHK_SD_HCXC(sd_card) && !CHK_MMC_SECTOR_MODE(sd_card))
		data_addr = start_sector << 9;
	else
		data_addr = start_sector;
	/*
	 * When accessing Kingston MMC4.0 cards, we shall check
	 * "READY_FOR_DATA" bit in "Card Status" of R1
	 * Those cards will report READY_FOR_DATA=0 after MULTIPLE_WRITE_BLOCK
	 * command. Especially after "STOP" transfer from
	 * "MULTIPLE_READ_BLOCK" status().
	 */
	ret = sd_wait_data_buf_ready(cr, card);
	if (ret) {
		card->rw_need_retry = 1;
		sd_set_error_code(cr, card, SD_TO_ERR);
		return ret;
	}

	/*
	 * Not auto write 2, cause some card need
	 * more time between command and data
	 */
	ret = sd_write_multiple_block(card, data_addr);
	if (ret) {
		card->rw_need_retry = 1;
		return ret;
	}

	cr_init_cmd(cr);

	cfg2 = SD_NO_CALCULATE_CRC7 | SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END;
	cfg2 |= SD_NO_CHECK_CRC7 | SD_RSP_LEN_0;
	if (!CHK_SD30_SPEED(sd_card))
		cfg2 |= SD_NO_CHECK_WAIT_CRC_TO;
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, cfg2);

	cr_enable_dma(cr, sector_cnt * 512, DMA_512, srb->sc_data_direction);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
			SD_TM_AUTO_WRITE_3 | SD_TRANSFER_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
			SD_TRANSFER_END, SD_TRANSFER_END);
	cr_send_cmd_no_wait(cr);

	return 0;
}

static int sd_prepare_read(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u32 data_addr;
	u16 sector_cnt;
	u8 cfg2;
	int ret;

	rts_trace(TRACE_SD, "SD/MMC CMD %d\n", READ_MULTIPLE_BLOCK);

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	/*
	 * High-Capacity Card Addressing Mode is in "Block"
	 * while Standard Capacity Card is in "Byte",
	 * So We should check Capacity mode to
	 * determine parameter "Address"
	 */
	if (!CHK_SD_HCXC(sd_card) && !CHK_MMC_SECTOR_MODE(sd_card))
		data_addr = start_sector << 9;
	else
		data_addr = start_sector;


	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD0, 0xFF,
					0x40 | READ_MULTIPLE_BLOCK);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD1, 0xFF,
					(u8)(data_addr >> 24));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD2, 0xFF,
					(u8)(data_addr >> 16));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD3, 0xFF,
					(u8)(data_addr >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD4, 0xFF,
						(u8)data_addr);

	cfg2 = SD_CALCULATE_CRC7 | SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END
		| SD_CHECK_CRC7 | SD_RSP_LEN_6;
	if (!CHK_SD30_SPEED(sd_card))
		cfg2 |= SD_NO_CHECK_WAIT_CRC_TO;
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, cfg2);

	cr_enable_dma(cr, sector_cnt * 512, DMA_512, srb->sc_data_direction);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
			SD_TM_AUTO_READ_2 | SD_TRANSFER_START);
	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
			SD_TRANSFER_END, SD_TRANSFER_END);

	cr_send_cmd_no_wait(cr);

	return 0;
}

static int sd_prepare_rw_normal(struct card *card, struct scsi_cmnd *srb)
{
	if (srb->sc_data_direction == DMA_FROM_DEVICE)
		return sd_prepare_read(card, srb);
	else
		return sd_prepare_write(card, srb);
}

static int sd_prepare_rw_sequential(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u16 sector_cnt;
	int ret;
	u8 cfg2;

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	cr_init_cmd(cr);

	cfg2 = SD_NO_CALCULATE_CRC7 | SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END;
	cfg2 |= SD_NO_CHECK_CRC7 | SD_RSP_LEN_0;
	if (!CHK_SD30_SPEED(sd_card))
		cfg2 |= SD_NO_CHECK_WAIT_CRC_TO;
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, cfg2);

	cr_enable_dma(cr, sector_cnt * 512, DMA_512, srb->sc_data_direction);

	if (srb->sc_data_direction == DMA_FROM_DEVICE) {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
				SD_TM_AUTO_READ_3 | SD_TRANSFER_START);
	} else {
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
				SD_TM_AUTO_WRITE_3 | SD_TRANSFER_START);
	}

	cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER, SD_TRANSFER_END,
							SD_TRANSFER_END);
	cr_send_cmd_no_wait(cr);

	return 0;
}

static int sd_prepare_rw_base(struct card *card, struct scsi_cmnd *srb)
{
	struct card_reader *cr = card->cm->cr;
	struct sd_info *sd_card = card->extra;
	u32 start_sector;
	u16 sector_cnt;
	int ret;

	ret = scsi_get_sector_info(srb, &start_sector, &sector_cnt);
	if (ret)
		return ret;

	if (srb->sc_data_direction == DMA_FROM_DEVICE) {
		rts_trace(TRACE_SD, "sd_rw: Read %d %s from 0x%x\n", sector_cnt,
				(sector_cnt > 1) ? "sectors" : "sector",
				start_sector);
	} else {
		rts_trace(TRACE_SD, "sd_rw: Write %d %s to 0x%x\n", sector_cnt,
				(sector_cnt > 1) ? "sectors" : "sector",
				start_sector);
	}
	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_L, 0xFF, 0x00);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_H, 0xFF, 0x02);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_L, 0xFF, (u8)sector_cnt);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_H,
						0xFF, (u8)(sector_cnt >> 8));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE, 0x01, RING_BUFFER);
	if (CHK_MMC_8BIT(sd_card))
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG1,
						0x03, SD_BUS_WIDTH_8);
	else if (CHK_MMC_4BIT(sd_card) || CHK_SD(sd_card))
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG1,
						0x03, SD_BUS_WIDTH_4);
	else
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG1,
						0x03, SD_BUS_WIDTH_1);

	return cr_send_cmd(cr, SD_CARD, 200);
}

static int sd_prepare_legacy_rw(struct card *card, struct scsi_cmnd *srb)
{
	int ret = 0;
	struct sd_info *sd_card = card->extra;

	ret = sd_prepare_rw_base(card, srb);
	if (ret) {
		sd_clear_sd_error(card->cm->cr);
		card->rw_need_retry = 1;
		sd_set_error_code(card->cm->cr, card, SD_TO_ERR);
		return ret;
	}

	if (SEQ_RW == sd_card->rw_mode) {
		return sd_prepare_rw_sequential(card, srb);
	} else {
		sd_card->rw_mode = SEQ_RW;
		return sd_prepare_rw_normal(card, srb);
	}

	return 0;
}

static int sd_prepare_rw(struct card *card, struct scsi_cmnd *srb)
{
	struct sd_info *sd_card = card->extra;

	if (CHK_SD_UHSII(sd_card))
		return sd_prepare_uhsii_rw(card, srb);
	else
		return sd_prepare_legacy_rw(card, srb);
}

static void sd_process_transfer_error_uhsii(struct card *card, int error)
{
	struct card_reader *cr = card->cm->cr;

	sd40_process_transfer_error(cr, card);
	card->rw_need_retry = 1;
}

static void sd_process_transfer_error_legacy(struct card *card, int error)
{
	struct card_reader *cr = card->cm->cr;
	u8 stat = 0;

	sd_print_register(cr);

	cr_rd_reg(cr, REG_SD_STAT1, &stat);

	sd_clear_sd_error(cr);

	if (sd_stop_transmission(card)) {
		card->rw_need_retry = 1;
		sd_set_error_code(cr, card, SD_STS_ERR);
		return;
	}

	if (stat & (SD_CRC7_ERR | SD_CRC16_ERR | SD_CRC_WRITE_ERR)) {
		card->rw_need_retry = 1;
		rts_trace(TRACE_SD, "SD CRC error, tune clock!\n");
		sd_set_error_code(cr, card, SD_CRC_ERR);
		return;
	}

	if (-ETIMEDOUT == error) {
		card->rw_need_retry = 1;
		sd_set_error_code(cr, card, SD_TO_ERR);
		return;
	}

}

static void sd_process_transfer_error(struct card *card, int error)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;

	if (CHK_SD_UHSII(sd_card))
		sd_process_transfer_error_uhsii(card, error);
	else
		sd_process_transfer_error_legacy(card, error);

	/* In card rw,if card unplugged surprisingly, don't need to retry rw */
	if (cr_detect_card(cr, SD_ID) != STATUS_SUCCESS) {
		card->rw_need_retry = 0;
		rts_trace(TRACE_SD, "No card exist, exit sd_rw\n");
		return;
	}
}

static int sd_process_rw_fail_legacy(struct card *card)
{

	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int retry_cnt = card->rw_retry_cnt;
	int ret = 0;

	if (sd_check_error_code(cr, card, SD_CRC_ERR)) {
		if (CHK_MMC_4BIT(sd_card) || CHK_MMC_8BIT(sd_card)) {
			sd_card->mmc_dont_switch_bus = 1;
			sd_reset_mmc_only(cr, card);
			sd_card->mmc_dont_switch_bus = 0;
		} else if (CHK_SD30_SPEED(sd_card)) {
			if (0 == retry_cnt) {
				sd_card->need_retune = 1;
				sd_auto_tune_clock(cr, card);
			} else if (1 == retry_cnt) {
				ret = sd_reset_card(cr, card);
			} else if (2 == retry_cnt) {
				card->reset_flags |= GO_SD20_DIRECT;
				ret = sd_reset_card(cr, card);
			}
		} else if (CHK_SD_HS(sd_card)) {
			if (0 == retry_cnt) {
				cr_wr_reg(cr, REG_SD_SAMPLE_POINT_CTL,
					SD20_RX_SEL_MASK, SD20_RX_POS_EDGE);
				sd_card->need_retune = 1;
				sd_auto_tune_clock(cr, card);
			} else if (1 == retry_cnt) {
				ret = sd_reset_card(cr, card);
			} else if (2 == retry_cnt) {
				sd_card->need_retune = 1;
				sd_auto_tune_clock(cr, card);
				ret = sd_reset_card(cr, card);
			}
		} else {
			ret = sd_reset_card(cr, card);
		}
	} else {
		if (CHK_SD30_SPEED(sd_card)) {
			card->reset_flags |= GO_SD20_DIRECT;
		}
		ret = sd_reset_card(cr, card);
	}

	return ret;
}

static int sd_process_rw_fail_uhsii(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int retry_cnt = card->rw_retry_cnt;
	int ret = 0;

	if (CHK_SD_UHSII(sd_card) && (DM_2L_HD == sd_card->lan_mode)) {
		card->uhsii_hd_rw_fail++;
	} else if (CHK_SD_UHSII(sd_card)) {
		card->uhsii_fd_rw_fail++;
	}

	if (sd_check_error_code(cr, card, SD_CRC_ERR)) {
		if (DM_2L_HD == sd_card->lan_mode) {
			sd_card->lan_mode = DM_FD;
		} else {
			card->reset_flags |= GO_SD_LEGACY_DIRECT;
			sd40_uninit_hw(cr, card);
			ret = sd_reset_card(cr, card);
		}
	} else if (sd_check_error_code(cr, card, SD_CMD12_ERR)) {
		rts_trace(TRACE_SD, "Forbit UHSII unspecified len mode\n");
		card->runtime_flags
			|= FORBIT_UHSII_UNSPECIFIED_LEN_MODE;
		sd40_uninit_hw(cr, card);
		ret = sd_reset_card(cr, card);
	} else {
		int mode = card->runtime_flags
			& FORBIT_UHSII_UNSPECIFIED_LEN_MODE;
		int hd = (DM_2L_HD == sd_card->lan_mode) ? 1 : 0;

		if ((card->uhsii_hd_rw_fail >= 8) && hd) {
			rts_trace(TRACE_SD, "UHSII HD RW fail number:%d\n",
						card->uhsii_hd_rw_fail);
			rts_trace(TRACE_SD, "Down to FD mode.\n");
			card->reset_flags |= FORBIT_USHII_HD_MODE;
			card->reset_flags &= ~PATCH_SPECIAL_SD40_PHY;
			sd40_uninit_hw(cr, card);
			ret = sd_reset_card(cr, card);
		} else if (((retry_cnt == 0) || ((retry_cnt == 1 && mode)))
			&& card->uhsii_fd_rw_fail <= 3) {
			rts_trace(TRACE_SD, "UHSII HD RW fail number:%d\n",
						card->uhsii_hd_rw_fail);
			rts_trace(TRACE_SD, "UHSII FD RW fail number:%d\n",
						card->uhsii_fd_rw_fail);
			rts_trace(TRACE_SD, "Patch Special UHSII PHY\n");
			card->reset_flags |= PATCH_SPECIAL_SD40_PHY;
			sd40_uninit_hw(cr, card);
			ret = sd_reset_card(cr, card);
		} else if (DM_2L_HD == sd_card->lan_mode) {
			rts_trace(TRACE_SD, "Down to FD mode.\n");
			card->reset_flags |= FORBIT_USHII_HD_MODE;
			card->reset_flags &= ~PATCH_SPECIAL_SD40_PHY;
			sd40_uninit_hw(cr, card);
			ret = sd_reset_card(cr, card);
		} else {
			rts_trace(TRACE_SD, "UHSII FD RW fail number:%d\n",
						card->uhsii_fd_rw_fail);
			rts_trace(TRACE_SD, "Down to UHSI.\n");
			card->reset_flags |= GO_SD_LEGACY_DIRECT;
			sd40_uninit_hw(cr, card);
			ret = sd_reset_card(cr, card);
		}
	}

	return ret;
}

static void sd_process_rw_fail(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	int ret = 0;

	if (NORMAL_RW == sd_card->rw_mode && CHK_SD_UHSII(sd_card))
		cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, 0);

	sd_card->rw_mode = NORMAL_RW;

	/* In card rw,if card unplugged surprisingly, don't need to retry rw */
	if ((cr_detect_card(cr, SD_ID) != STATUS_SUCCESS)
		|| test_bit(card->id, &(card->cm->cr->need_release))) {
		card->rw_need_retry = 0;
		rts_trace(TRACE_SD, "No card exist, exit sd_rw\n");
	}

	if (check_sd_flag(cr, SD_FORBID_RETRY_OF_RW))
		card->rw_need_retry = 0;

	if (!card->rw_need_retry)
		return;

	cr_patch_dma(cr);

	if (CHK_SD_UHSII(sd_card))
		ret = sd_process_rw_fail_uhsii(card);
	else
		ret = sd_process_rw_fail_legacy(card);


	if (0 != ret) {
		clr_card_flags(card, READY);
		set_card_flags(card, FAIL);
		card->capacity = 0;
	}

	return;
}

int sd_check_seq(struct card *card, struct scsi_cmnd *srb)
{
	u32 start_sector;
	u16 sector_cnt;
	struct sd_info * sd_card = card->extra;

	scsi_get_sector_info(srb, &start_sector, &sector_cnt);

	if (sd_card->prev_dir != srb->sc_data_direction) {
		sd_card->seq_num = 0;
		return 0;
	}

	if ((sd_card->prev_sec_addr + sd_card->prev_sec_cnt) != start_sector) {
		sd_card->seq_num = 0;
		return 0;
	}

	if (CHK_SD_UHSII(sd_card) &&
		(card->cm->cr->option.sd40_seq_speedup_en)) {
		rts_trace(TRACE_SD, "sector_cnt=%#x \n", sector_cnt);
		rts_trace(TRACE_SD, "sd_card->seq_num=%#x \n", sd_card->seq_num);
		if (sector_cnt%2) {
			sd_card->seq_num = 0;
			if (PRE_READ != sd_card->rw_mode &&
				PRE_WRITE != sd_card->rw_mode &&
				SEQ_RW != sd_card->rw_mode)
				return 0;
		} else {
			if (DMA_FROM_DEVICE == srb->sc_data_direction)
				sd_card->seq_num += 1;
			else if (DMA_TO_DEVICE == srb->sc_data_direction) {
				if (0 == (sector_cnt % sd_card->n_fcu))
					sd_card->seq_num += 1;
				else
					sd_card->seq_num = 0;
			}
		}
		rts_trace(TRACE_SD, "sector_cnt=%#x \n", sector_cnt);
		rts_trace(TRACE_SD, "sd_card->seq_num=%#x \n", sd_card->seq_num);
	}
	return 1;
}



static int sd_comm_rw(struct card_reader *cr, struct card *card,
		struct scsi_cmnd *srb, u32 start_sector, u16 sector_cnt)
{
	struct sd_info *sd_card = card->extra;
	int retval;
	int timeout = 0;

	sd_print_srb_info(srb, start_sector, sector_cnt);

	if (sd_check_card_state_befor_rw(card))
		TRACE_RET(cr, STATUS_FAIL);

	sd_clear_error_code(cr, card);

	retval = sd_switch_clock(cr, card);
	if (retval != STATUS_SUCCESS) {
		sd_set_error_code(cr, card, SD_IO_ERR);
		TRACE_GOTO(cr, RW_FAIL);
	}

	if (SEQ_RW == sd_card->rw_mode) {
		if (!sd_check_seq(card, srb)) {
			if (sd_stop_sequential(card))
				goto RW_FAIL;
			sd_card->rw_mode = NORMAL_RW;
		}
	}

	retval = sd_prepare_rw(card, srb);
	if (retval) {
		rts_trace(TRACE_ERROR, "Prepare for r/w fail\n");
		goto RW_FAIL;
	}

	if (CHK_SD_UHSII(sd_card))
		timeout = cr->option.sd40_timeout;
	else
		timeout = cr->option.sd_timeout;

	retval = cr_transfer_data(cr, SD_CARD, scsi_sglist(srb),
			scsi_bufflen(srb), scsi_sg_count(srb),
			srb->sc_data_direction, timeout);
	if (retval < 0) {
		sd_process_transfer_error(card, retval);
		goto RW_FAIL;
	}

	sd_card->prev_sec_addr = start_sector;
	sd_card->prev_sec_cnt = sector_cnt;
	sd_card->prev_dir = srb->sc_data_direction;
	if (PRE_READ == sd_card->rw_mode) {
		if (DMA_FROM_DEVICE == srb->sc_data_direction)
			sd_card->read_sectors += sector_cnt;
	}

	if (PRE_WRITE == sd_card->rw_mode) {
		if (DMA_TO_DEVICE == srb->sc_data_direction)
			sd_card->write_sectors += sector_cnt;
	}

	if (NORMAL_RW == sd_card->rw_mode && CHK_SD_UHSII(sd_card))
		cr_wr_reg(cr, REG_SD40_EBSY, 0xC3, 0);

	return 0;

RW_FAIL:
	sd_process_rw_fail(card);
	TRACE_RET(cr, STATUS_FAIL);
}

static int sd_rw(struct card_reader *cr, struct card *c,
			struct scsi_cmnd *srb, u32 lba, u16 num)
{
	if (cr->sd_ops_prot->rw)
		return cr->sd_ops_prot->rw(cr, c, srb, lba, num);
	else
		return sd_comm_rw(cr, c, srb, lba, num);
}

static int sd_disable(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->disable) {
		return cr->sd_ops_prot->disable(cr, c);
	} else {
		/* DO NOTHING. IT WILL BE DEPRECATED */
		return 0;
	}
}

void sd_stop_seq_mode(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;

	/* TODO: for pre-read */

	if (SEQ_RW != sd_card->rw_mode)
		return;

	if ((!CHK_SD_UHSII(sd_card))
		|| ((CHK_SD_UHSII(sd_card))
		&& !(sd_card->tmode & LM_LEN_SPECIF))) {
		int ret;

		ret = sd_switch_clock(cr, card);
		if (ret != STATUS_SUCCESS)
			return;
		ret = sd_stop_transmission(card);
		if (ret != STATUS_SUCCESS)
			sd_set_error_code(cr, card, SD_STS_ERR);
		ret = sd_wait_state_data_ready(cr, card, 0x08, 1, 1000);
		if (ret != STATUS_SUCCESS)
			sd_set_error_code(cr, card, SD_STS_ERR);
		sd_card->rw_mode = NORMAL_RW;
		cr_wr_reg(cr, REG_RBCTL, RB_FLUSH, RB_FLUSH);
	} else {
		/* TODO: UHSII */
	}
}
EXPORT_SYMBOL_GPL(sd_stop_seq_mode);


int sd_pull_ctl_disable(struct card_reader *cr)
{
	if (cr->sd_ops_prot->pull_ctl_disable) {
		return cr->sd_ops_prot->pull_ctl_disable(cr);
	} else {
		rts_trace(TRACE_ERROR, "pull_ctrl_disable not Implement\n");
		return -1;
	}

}
EXPORT_SYMBOL_GPL(sd_pull_ctl_disable);

int sd_pull_ctl_enable(struct card_reader *cr)
{
	if (cr->sd_ops_prot->pull_ctl_enable) {
		return cr->sd_ops_prot->pull_ctl_enable(cr);
	} else {
		rts_trace(TRACE_ERROR, "pull_ctrl_enable not Implement\n");
		return -1;
	}
}
EXPORT_SYMBOL_GPL(sd_pull_ctl_enable);

/*
 * sd_power_off_card3v3 should be placed in cardreader.c
 * and rename to cr_power_off_sd3v3
 */
int sd_power_off_card3v3(struct card_reader *cr)
{
	if (cr_disable_card_clk(cr, SD_ID))
		TRACE_RET(cr, STATUS_FAIL);

	cr_wr_reg(cr, REG_CARD_OE, SD_OUTPUT_EN, 0);

	if (cr_poweroff_card(cr, SD_ID))
		TRACE_RET(cr, STATUS_FAIL);

	msleep(50);

	if (cr->asic_code) {
		sd_pull_ctl_disable(cr);
	} else {
		u8 mask = FPGA_SD_PULL_CTL_BIT | 0x20;
		u8 val = FPGA_SD_PULL_CTL_BIT;
		cr_wr_reg(cr, REG_FPGA_PULL_CTL, mask, val);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(sd_power_off_card3v3);


static int sd_comm_cleanup_work(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;

	if (SEQ_RW == sd_card->rw_mode) {
		rts_trace(TRACE_SD, "SD: stop transmission SEQ_RW\n");
		sd_stop_seq_mode(cr, card);
	} else if (PRE_READ == sd_card->rw_mode) {
		rts_trace(TRACE_SD, "SD: stop transmission PRE_READ\n");
		sd_stop_sequential(card);
	} else if (PRE_WRITE == sd_card->rw_mode) {
		rts_trace(TRACE_SD, "SD: stop transmission PRE_WRITE\n");
		sd_stop_sequential(card);
	}

	return 0;
}

int sd_cleanup_work(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->cleanup)
		return cr->sd_ops_prot->cleanup(cr, c);
	else
		return sd_comm_cleanup_work(cr, c);
}
EXPORT_SYMBOL_GPL(sd_cleanup_work);

int sd_release(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->release) {
		return cr->sd_ops_prot->release(cr, c);
	} else {
		rts_trace(TRACE_ERROR,
			"sd_ops_prot->release is not assigned\n");
		return -EINVAL;
	}
}
EXPORT_SYMBOL_GPL(sd_release);

static int sd_update_status(struct card_reader *cr, struct card *c)
{
	if (cr->sd_ops_prot->update_status) {
		return cr->sd_ops_prot->update_status(cr, c);
	} else {
		/* DO NOTHING. IT WILL BE DEPRECATED */
		return 0;
	}
}


int sd_card_reset(struct card *c)
{
	return sd_reset_card(c->cm->cr, c);
}

int sd_card_disable(struct card *c)
{
	return sd_disable(c->cm->cr, c);
}

int sd_card_rw(struct scsi_cmnd *srb, struct card *c, u32 lba, u16 num)
{
	return sd_rw(c->cm->cr, c, srb, lba, num);
}

int sd_card_cleanup(struct card *c)
{
	return sd_cleanup_work(c->cm->cr, c);
}

int sd_card_release(struct card *c)
{
	return sd_release(c->cm->cr, c);
}

int sd_card_update_status(struct card *c)
{
	return sd_update_status(c->cm->cr, c);
}

void sd_print_register(struct card_reader *cr)
{
	u8 val;

	cr_rd_reg(cr, REG_SD_STAT1, &val);
	rts_trace(TRACE_SD, "SD_STAT1: 0x%x\n", val);
	cr_rd_reg(cr, REG_SD_STAT2, &val);
	rts_trace(TRACE_SD, "SD_STAT2: 0x%x\n", val);
	cr_rd_reg(cr, REG_SD_BUS_STAT, &val);
	rts_trace(TRACE_SD, "SD_BUS_STAT: 0x%x\n", val);
	cr_rd_reg(cr, REG_SD_CMD0, &val);
	rts_trace(TRACE_SD, "SD_CMD0: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_CMD1, &val);
	rts_trace(TRACE_SD, "SD_CMD1: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_CMD2, &val);
	rts_trace(TRACE_SD, "SD_CMD2: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_CMD3, &val);
	rts_trace(TRACE_SD, "SD_CMD3: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_CMD4, &val);
	rts_trace(TRACE_SD, "SD_CMD4: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_CMD5, &val);
	rts_trace(TRACE_SD, "SD_CMD5: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_TRANSFER, &val);
	rts_trace(TRACE_SD, "SD_TRANSFER: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_BLOCK_CNT_L, &val);
	rts_trace(TRACE_SD, "SD_BLOCK_CNT_L: 0x%02x\n", val);
	cr_rd_reg(cr, REG_SD_BLOCK_CNT_H, &val);
	rts_trace(TRACE_SD, "SD_BLOCK_CNT_H: 0x%02x\n", val);
}

int check_sd_speed_prior(u32 sd_speed_prior)
{
	int i, fake_para = 0;

	/* Check the legality of sd_speed_prior */
	for (i = 0; i < 4; i++) {
		u8 tmp = (u8)(sd_speed_prior >> (i*8));
		if ((tmp < 0x01) || (tmp > 0x04)) {
			fake_para = 1;
			break;
		}
	}

	return !fake_para;
}
EXPORT_SYMBOL_GPL(check_sd_speed_prior);

int check_sd_current_prior(u32 sd_current_prior)
{
	int i, fake_para = 0;

	/* Check the legality of sd_speed_prior */
	for (i = 0; i < 4; i++) {
		u8 tmp = (u8)(sd_current_prior >> (i * 8));
		if (tmp > 0x03) {
			fake_para = 1;
			break;
		}
	}

	return !fake_para;
}
EXPORT_SYMBOL_GPL(check_sd_current_prior);

static int mmc_get_ext_csd(struct card *card, u8 *buf, int len)
{
	struct sd_command cmd;
	u32 bytes = (len < 512) ? len : 512;

	INIT_SD_CMD(cmd, SEND_EXT_CSD, 0x0, 0, 0, 0);
	init_sd_dcmd_data(cmd, buf, len, bytes, 1, SD_BUS_WIDTH_1,
					SD_TM_NORMAL_READ, FROM_SD, 250);

	if (sd_send_cmd(card->cm->cr, card, &cmd)) {
		sd_clear_sd_error(card->cm->cr);
		sd_clear_status(card);
		TRACE_RET(card->cm->cr, STATUS_FAIL);
	}

	return 0;
}

static int mmc_check_ext_csd(struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct card_reader *cr = card->cm->cr;
	struct cr_option *option = &cr->option;
	int ret = 0;
	u8 *buf = NULL;
	u8 card_type = 0;

	CLR_MMC_HS(sd_card);

	buf = kmalloc(512, GFP_KERNEL);
	if (!buf) {
		rts_trace(TRACE_ERROR, "Can't alloc Mem\n");
		return -ENOMEM;
	}

	if (mmc_get_ext_csd(card, buf, 512)) {
		kfree(buf);
		TRACE_RET(cr, STATUS_FAIL);
	}

	/* For MMC4.2, We shall calculate MMC4.2 card's Capacity
	 * by EXT_CSD register
	 */
	if (CHK_MMC_SECTOR_MODE(sd_card)) {
		u8 *p = &buf[212];
		sd_card->capacity = p[0];
		sd_card->capacity |= p[1] << 8;
		sd_card->capacity |= p[2] << 16;
		sd_card->capacity |= p[3] << 24;
		card->capacity = sd_card->capacity;
	}
	rts_trace(TRACE_SD, "mmc capacity:%#x\n", card->capacity);
	rts_trace(TRACE_SD, "MMC version :%d\n", buf[192]);

	/* check card type */
	if (CHK_MMC_SECTOR_MODE(sd_card))
		card_type = buf[196] & 0x3F;
	else
		card_type = buf[196] & 0x03;

	if (card_type & 0x10) {
		SET_MMC_SDR200(sd_card);
		sd_card->sd_default_tx_phase = (u8) option->mmc_sdr200_tx_phase;
		sd_card->sd_default_rx_phase =
			option->sd_sdr104_default_rx_phase;
	} else if (card_type & 0x04) {
		SET_MMC_DDR52(sd_card);
		sd_card->sd_default_tx_phase =
			option->mmc_ddr_tx_phase;
		sd_card->sd_default_rx_phase =
			option->sd_ddr50_default_rx_phase;
	} else if (card_type & 0x02) {
		SET_MMC_52M(sd_card);
		sd_card->sd_default_tx_phase =
			option->sd_sdr50_default_tx_phase;
		sd_card->sd_default_rx_phase =
			option->sd_sdr50_default_rx_phase;
	} else {
		SET_MMC_26M(sd_card);
	}

	rts_trace(TRACE_SD, "MMC card type:0x%x\n", card_type);

	if (buf)
		kfree(buf);

	return ret;
}

static int mmc_process_sdr200_error(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;

	card->reset_flags &= ~RESET_E_MMC45_NOW;
	CLR_MMC_SDR200(sd_card);
	sd_init_power(cr);
	if (!check_sd_flag(cr, EMMC45_DONT_SWITCH_POWER))
		sd_change_bank_voltage(cr, SD_COMM_IO_3V3);

	return 0;
}

int sd_cmd_support(struct card *card, int opcode)
{
	struct sd_info *sd_card = card->extra;
	u8 cmd_support = sd_card->raw_scr[3] & 0xF; /* be: bit 32-35 */
	u8 sd_spec4 = sd_card->raw_scr[2] & 0x04; /* be: bit 42 */

	rts_trace(TRACE_SD, "cmd_support=%#x\n", cmd_support);
	switch (opcode) {
	case SPEED_CLASS_CONTROL:
		return cmd_support & 0x1;
	case SET_BLOCK_COUNT:
		return cmd_support & 0x2;
	case READ_EXTR_SINGLE:
	case WRITE_EXTR_SINGLE:
		return ((cmd_support & 0x4)&&sd_spec4);
	case READ_EXTR_MULTI:
	case WRITE_EXTR_MULTI:
		return ((cmd_support & 0x8)&&sd_spec4);
	default:
		return 0;
	}

	return 0;
}

u32 sd_extr_arg_init_single(u32 mio, u32 fno, u32 mw, u32 addr, u32 len)
{
	u32 arg = 0;
	arg |= (mio & SD_EXTR_ARG_MIO_MASK) << SD_EXTR_ARG_MIO_SHIFT;
	if (mio & SD_EXTR_ARG_MIO_MASK)
		arg |= (fno & SD_EXTR_ARG_IOFNO_MASK) <<
			SD_EXTR_ARG_IOFNO_SHIFT;
	else
		arg |= (fno & SD_EXTR_ARG_MEMFNO_MASK) <<
			SD_EXTR_ARG_MEMFNO_SHIFT;

	arg |= (mw & SD_EXTR_ARG_MW_MASK) << SD_EXTR_ARG_MW_SHIFT;
	arg |= (addr & SD_EXTR_ARG_ADDR_MASK) << SD_EXTR_ARG_ADDR_SHIFT;
	if (mw & SD_EXTR_ARG_MW_MASK)
		arg |= (len & SD_EXTR_ARG_MASK_MASK);
	else
		arg |= ((len - 1) & SD_EXTR_ARG_LEN_MASK);

	rts_trace(TRACE_SD, "EXT single arg: 0x%08x, len=%d\n", arg, len);
	return arg;
}

u32 sd_extr_arg_init_multi(u32 mio, u32 fno, u32 bus, u32 addr, u32 buc)
{
	u32 arg = 0;

	arg |= (mio & SD_EXTR_ARG_MIO_MASK) << SD_EXTR_ARG_MIO_SHIFT;
	if (mio & SD_EXTR_ARG_MIO_MASK)
		arg |= (fno & SD_EXTR_ARG_IOFNO_MASK) <<
			SD_EXTR_ARG_IOFNO_SHIFT;
	else
		arg |= (fno & SD_EXTR_ARG_MEMFNO_MASK) <<
			SD_EXTR_ARG_MEMFNO_SHIFT;

	arg |= (bus & SD_EXTR_ARG_BUS_MASK) << SD_EXTR_ARG_BUS_SHIFT;
	arg |= (addr & SD_EXTR_ARG_ADDR_MASK) << SD_EXTR_ARG_ADDR_SHIFT;
	arg |= ((buc - 1) & SD_EXTR_ARG_BUC_MASK);

	return arg;
}

u32 sd_extr_arg_single_len(u32 arg)
{
	if ((arg >> SD_EXTR_ARG_MW_SHIFT) & SD_EXTR_ARG_MW_MASK)
		return 1;

	if ((arg >> SD_EXTR_ARG_MIO_SHIFT) & SD_EXTR_ARG_MIO_MASK)
		return (arg & SD_EXTR_ARG_MASK_MASK);
	else
		return (arg & SD_EXTR_ARG_LEN_MASK) + 1;
}

u32 sd_extr_arg_multi_addr(u32 arg)
{
	return arg >> SD_EXTR_ARG_ADDR_SHIFT & SD_EXTR_ARG_ADDR_MASK;
}

u32 sd_extr_arg_multi_buc(u32 arg)
{
	return (arg & 0x1FF) + 1;
}

u32 sd_extr_arg_multi_bus(u32 arg)
{
	return (arg >> SD_EXTR_ARG_BUS_SHIFT) & SD_EXTR_ARG_BUS_MASK;
}

/**
 * sd_read_extr_single - read extension SD register(single block)
 */
int sd_read_extr_single(struct card_reader *cr, struct card *card, u32 arg,
		u8 *__buf)
{
	struct sd_command cmd;
	u8 *buf = NULL;
	u32 len = sd_extr_arg_single_len(arg);

	int err = 0;

	buf = kmalloc(512, GFP_KERNEL);
	if (!buf) {
		rts_trace(TRACE_ERROR, "Can't alloc Mem\n");
		return -ENOMEM;
	}

	INIT_SD_CMD(cmd, 0x40 | READ_EXTR_SINGLE, arg, 0, SD_RSP_TYPE_R1, 0);
	init_sd_dcmd_data(cmd, buf, 512, 512, 1, SD_BUS_WIDTH_4,
			SD_TM_NORMAL_READ, FROM_SD, 1000);

	err = sd_send_cmd(cr, card, &cmd);
	if (!err) {
		rts_trace(TRACE_SD, "len=%#x\n", len);
		RTS_DUMP(buf, 16);
		memcpy(__buf, buf, len);
	} else {
		sd_clear_sd_error(cr);
	}

	kfree(buf);

	return err;
}

int sd_write_extr_single(struct card_reader *cr, struct card *card, u32 arg,
		u8 *__buf)
{
	struct sd_command cmd;
	u8 *buf = NULL;
	int len = sd_extr_arg_single_len(arg);
	int err = 0;

	buf = kmalloc(512, GFP_KERNEL);
	if (!buf) {
		rts_trace(TRACE_ERROR, "Can't alloc Mem\n");
		return -ENOMEM;
	}

	memcpy(buf, __buf, len);
	INIT_SD_CMD(cmd, 0x40 | WRITE_EXTR_SINGLE, arg, 0, SD_RSP_TYPE_R1, 0);
	init_sd_dcmd_data(cmd, buf, 512, 512, 1, SD_BUS_WIDTH_4,
			SD_TM_AUTO_WRITE_2, TO_SD, 1000);

	err = sd_send_cmd(cr, card, &cmd);
	if (err)
		sd_clear_sd_error(cr);
	kfree(buf);

	return err;
}

static int sd_extr_write_multiple_block(struct card *card, u32 addr)
{
	struct sd_command cmd;

	INIT_SD_CMD(cmd, WRITE_EXTR_MULTI, addr, 0, SD_RSP_TYPE_R1, 0);

	return sd_send_cmd(card->cm->cr, card, &cmd);
}

int sd_legacy_rw_extr_multi(struct card_reader *cr, struct card *card,
		u32 arg, u8 *buf, int to_sd)
{
	int err = 0;
	u8 bus_width = SD_BUS_WIDTH_4;

	u32 buc = sd_extr_arg_multi_buc(arg);
	u32 bus = sd_extr_arg_multi_bus(arg);
	u8 cfg2 = 0;
	u32 len = bus ? buc * 32 * 1024 : buc * 512;
	u16 block_count = (u16)(len / 512);
	struct sd_info *sd_card = NULL;

	u8 opcode = to_sd ? WRITE_EXTR_MULTI : READ_EXTR_MULTI;
	int dma_dir = to_sd ? DMA_TO_DEVICE : DMA_FROM_DEVICE;
	int trans_mode = to_sd ? SD_TM_AUTO_WRITE_3 : SD_TM_AUTO_READ_2;

	if (card)
		sd_card = card->extra;

	if (sd_card && (sd_card->sd_lock_status & SD_LOCK_1BIT_MODE) == 0) {
		if (CHK_MMC_8BIT(sd_card))
			bus_width = SD_BUS_WIDTH_8;
		else if (CHK_SD(sd_card) || CHK_MMC_4BIT(sd_card))
			bus_width = SD_BUS_WIDTH_4;
		else
			bus_width = SD_BUS_WIDTH_1;
	}

	rts_trace(TRACE_SD, "SD/MMC CMD %d\n", opcode);

	cr_init_cmd(cr);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_H, 0xFF, 0x02);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BYTE_CNT_L, 0xFF, 0x00);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_H, 0xFF,
		(u8)(block_count >> 8 & 0xFF));
	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_BLOCK_CNT_L, 0xFF,
		(u8)(block_count & 0xFF));

	cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG1, 0x03, bus_width);
	cr_add_cmd(cr, WRITE_REG_CMD, REG_CARD_DATA_SOURCE, 0x01, RING_BUFFER);


	if (DMA_FROM_DEVICE == dma_dir)
	{
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD0, 0xFF,
				0x40 | opcode);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD1, 0xFF,
				(u8)(arg >> 24));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD2, 0xFF,
				(u8)(arg >> 16));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD3, 0xFF,
				(u8)(arg >> 8));
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CMD4, 0xFF,
				(u8)arg);
		cr_enable_dma(cr, len, DMA_512, dma_dir);

		cfg2 = SD_CALCULATE_CRC7 | SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END
			| SD_CHECK_CRC7 | SD_RSP_LEN_6;
		if (!CHK_SD30_SPEED(sd_card))
			cfg2 |= SD_NO_CHECK_WAIT_CRC_TO;
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, cfg2);

		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
			trans_mode | SD_TRANSFER_START);
		cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER, SD_TRANSFER_END,
			SD_TRANSFER_END);
		cr_send_cmd_no_wait(cr);
	} else if (DMA_TO_DEVICE == dma_dir) {
		err = cr_send_cmd(cr, SD_CARD, 200);
		if (err < 0) {
			sd_clear_sd_error(cr);
			sd_set_error_code(cr, card, SD_TO_ERR);
			TRACE_RET(cr, err);
		}
		/*
		 * When accessing Kingston MMC4.0 cards, we shall check
		 * "READY_FOR_DATA" bit in "Card Status" of R1
		 * Those cards will report READY_FOR_DATA=0 after MULTIPLE_WRITE_BLOCK
		 * command. Especially after "STOP" transfer from
		 * "MULTIPLE_READ_BLOCK" status().
		 */
		err = sd_wait_data_buf_ready(cr, card);
		if (err) {
			sd_set_error_code(cr, card, SD_TO_ERR);
			return err;
		}

		/*
		 * Not auto write 2, cause some card need
		 * more time between command and data
		 */
		err = sd_extr_write_multiple_block(card, arg);
		if (err) {
			return err;
		}

		cr_init_cmd(cr);

		cfg2 = SD_NO_CALCULATE_CRC7 | SD_CHECK_CRC16 | SD_NO_WAIT_BUSY_END;
		cfg2 |= SD_NO_CHECK_CRC7 | SD_RSP_LEN_0;
		if (!CHK_SD30_SPEED(sd_card))
			cfg2 |= SD_NO_CHECK_WAIT_CRC_TO;
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_CFG2, 0xFF, cfg2);
		cr_enable_dma(cr, len, DMA_512, dma_dir);
		cr_add_cmd(cr, WRITE_REG_CMD, REG_SD_TRANSFER, 0xFF,
				trans_mode | SD_TRANSFER_START);
		cr_add_cmd(cr, CHECK_REG_CMD, REG_SD_TRANSFER,
				SD_TRANSFER_END, SD_TRANSFER_END);
		cr_send_cmd_no_wait(cr);
	}

	err = cr_transfer_data(cr, SD_CARD, buf, len, 0, dma_dir, 10000);
	if (err < 0) {
		sd_clear_sd_error(cr);
		TRACE_RET(cr, STATUS_FAIL);
	}
	return err;
}


int sd_uhsii_rw_extr_multi(struct card_reader *cr, struct card *card, u32 arg,
		u8 *buf, int to_sd)
{
	struct sd_info *sd_card = card->extra;
	struct sd40_tlp tlp;
	u8 tmode = 0;

	memset(&tlp, 0, sizeof(tlp));

	tlp.header[0] = HD_DCMD | APP_PACKET | sd_card->sd40_did;
	tlp.header[1] = sd_card->sd40_sid_tid;

	tmode = DAM_INC_ADDR | TLUM_BLOCK;
	tmode |=  LM_LEN_SPECIF | sd_card->lan_mode;
	tlp.arg[0] = tmode;
	if (to_sd) {
		tlp.arg[0] |= ARG_DIR_W;
		tlp.arg[1] = WRITE_EXTR_MULTI & 0x3F;
	} else {
		tlp.arg[0] |= ARG_DIR_R;
		tlp.arg[1] = READ_EXTR_MULTI & 0x3F;
	}

	tlp.ext_arg.sd_arg = arg;
	tlp.scatter_gather = 1;
	tlp.sg = (struct scatterlist*)buf;
	tlp.timeout = 1000;

	return sd40_send_tlp(card, &tlp);
}

/**
 * sd_read_extr_multi - read extension SD register(multiple block)
 */
inline int sd_read_extr_multi(struct card_reader *cr, struct card *card,
		u32 arg, u8 * buf)
{
	struct sd_info * sd_card = NULL;
	if (card)
		sd_card = card->extra;
	if ((sd_card) && CHK_SD_UHSII(sd_card))
		return sd_uhsii_rw_extr_multi(cr, card, arg, buf, 0);
	else
		return sd_legacy_rw_extr_multi(cr, card, arg, buf, 0);
}

/**
 * sd_write_extr_single - write extension SD register(multiple block)
 */
inline int sd_write_extr_multi(struct card_reader *cr, struct card *card,
		u32 arg, u8 * buf)
{
	struct sd_info * sd_card = NULL;
	if (card)
		sd_card = card->extra;
	if ((sd_card) && CHK_SD_UHSII(sd_card))
		return sd_uhsii_rw_extr_multi(cr, card, arg, buf, 1);
	else
		return sd_legacy_rw_extr_multi(cr, card, arg, buf, 1);
}

u8 sd_ge_extr(struct card_reader *cr, struct card *card, u8 addr)
{
	struct sd_info *sd_card = card->extra;
	struct sd_ge *ge = &sd_card->ge;
	u32 arg = sd_extr_arg_init_single(0, ge->fno, 0, ge->start_addr + addr, 1);
	u8 val = 0;
	sd_read_extr_single(cr, card, arg, &val);
	rts_trace(TRACE_SD, "sd_ge_extr return %#x\n", val);
	return val;
}

int sd_ge_go_to_ready(struct card_reader *cr, struct card *card)
{
	struct sd_info *sd_card = card->extra;
	struct sd_ge *ge = &sd_card->ge;
	u32 arg = sd_extr_arg_init_single(0, ge->fno, 0, ge->start_addr + SD_GE_ADDR_OPERATION, 512);
	u8 val[512] = {0};
	val[0] = 1;
	return sd_write_extr_single(cr, card, arg, val);
}

int sd_init_sd_cfg_reg(struct card_reader *cr, struct card *c)
{
	struct sd_info *sd_card = c->extra;

	if (CHK_SD30_SPEED(sd_card) || CHK_MMC_DDR52(sd_card)) {
		/* Initialize SD_CFG3 register */
		cr_wr_reg(cr, REG_SD_CFG3, SD30_CLK_END_EN, SD30_CLK_END_EN);
		cr_wr_reg(cr, REG_SD_STOP_SDCLK_CFG,
			SD30_CLK_STOP_CFG_EN | SD30_CLK_STOP_CFG1 | SD30_CLK_STOP_CFG0,
			SD30_CLK_STOP_CFG_EN | SD30_CLK_STOP_CFG1 | SD30_CLK_STOP_CFG0);
	}

	return 0;
}
